Package: power4mome
Title: Power Analysis for Moderation and Mediation
Version: 0.2.1
Authors@R: 
    c(person(given = "Shu Fai",
           family = "Cheung",
           role = c("aut", "cre"),
           email = "shufai.cheung@gmail.com",
           comment = c(ORCID = "0000-0002-9871-9448")),
      person(given = "Sing-Hang",
           family = "Cheung",
           role = c("aut"),
           comment = c(ORCID = "0000-0001-5182-0752")),
      person(given = "Wendie",
             family ="Yang",
             role = c("aut"),
             comment = c(ORCID = "0009-0000-8388-6481")))
Description: Power analysis and sample size determination
  for moderation, mediation, and moderated mediation in models
  fitted by structural equation modelling using the 'lavaan'
  package by Rosseel (2012) <doi:10.18637/jss.v048.i02> or
  by multiple regression. The package 'manymome' by
  Cheung and Cheung (2024) <doi:10.3758/s13428-023-02224-z>
  is used to specify the indirect paths or conditional
  indirect paths to be tested.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/testthat/parallel: true
Depends: R (>= 4.4.0)
URL: https://sfcheung.github.io/power4mome/
BugReports: https://github.com/sfcheung/power4mome/issues
Imports: lavaan, stats, manymome (>= 0.3.1.4), pbapply, parallel,
        pgnorm, lmhelprs (>= 0.4.2), psych, yaml, graphics, methods,
        cli, mice
Config/Needs/website: rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-03-23 13:19:25 UTC; shufa
Author: Shu Fai Cheung [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9871-9448>),
  Sing-Hang Cheung [aut] (ORCID: <https://orcid.org/0000-0001-5182-0752>),
  Wendie Yang [aut] (ORCID: <https://orcid.org/0009-0000-8388-6481>)
Maintainer: Shu Fai Cheung <shufai.cheung@gmail.com>
Repository: CRAN
Date/Publication: 2026-03-24 05:10:02 UTC
Built: R 4.6.0; ; 2026-04-25 13:49:19 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(c,power4test_by_es)
S3method(c,power4test_by_n)
S3method(plot,n_region_from_power)
S3method(plot,power4test_by_es)
S3method(plot,power4test_by_n)
S3method(plot,power_curve)
S3method(plot,q_power_mediation)
S3method(plot,x_from_power)
S3method(predict,power_curve)
S3method(print,n_region_from_power)
S3method(print,power4test)
S3method(print,power4test_by_es)
S3method(print,power4test_by_n)
S3method(print,power_curve)
S3method(print,q_power_mediation)
S3method(print,rejection_rates_df)
S3method(print,sim_data)
S3method(print,sim_out)
S3method(print,summary.n_region_from_power)
S3method(print,summary.x_from_power)
S3method(print,test_out_list)
S3method(print,test_summary)
S3method(print,test_summary_list)
S3method(print,x_from_power)
S3method(rejection_rates,default)
S3method(rejection_rates,n_region_from_power)
S3method(rejection_rates,power4test)
S3method(rejection_rates,power4test_by_es)
S3method(rejection_rates,power4test_by_n)
S3method(rejection_rates,q_power_mediation)
S3method(rejection_rates,x_from_power)
S3method(summary,n_region_from_power)
S3method(summary,q_power_mediation)
S3method(summary,x_from_power)
export(R_for_bz)
export(Rs_bz_supported)
export(arg_x_from_power)
export(as.power4test_by_es)
export(as.power4test_by_n)
export(cut_patterns)
export(do_test)
export(find_par_names)
export(fit_model)
export(gen_boot)
export(gen_mc)
export(missing_values)
export(model_matrices_pop)
export(n_from_power)
export(n_region_from_power)
export(ordinal_variables)
export(pba_diagnosis)
export(pool_sim_data)
export(pop_es_yaml)
export(power4test)
export(power4test_by_es)
export(power4test_by_n)
export(power_curve)
export(ptable_pop)
export(q_power_mediation)
export(q_power_mediation_parallel)
export(q_power_mediation_serial)
export(q_power_mediation_simple)
export(rbeta_rs)
export(rbeta_rs2)
export(rbinary_rs)
export(rejection_rates)
export(rexp_rs)
export(rlnorm_rs)
export(rpgnorm_rs)
export(rt_rs)
export(runif_rs)
export(scale_scores)
export(sim_data)
export(sim_out)
export(summarize_tests)
export(test_cond_indirect)
export(test_cond_indirect_effects)
export(test_group_equal)
export(test_index_of_mome)
export(test_indirect_effect)
export(test_k_indirect_effects)
export(test_moderation)
export(test_parameters)
export(x_from_power)
importFrom(grDevices,adjustcolor)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,text)
importFrom(graphics,title)

<!-- badges: start -->
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![Project Status: Active - The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN status](https://www.r-pkg.org/badges/version/power4mome?color=blue)](https://CRAN.R-project.org/package=power4mome)
[![CRAN: Release Date](https://www.r-pkg.org/badges/last-release/power4mome?color=blue)](https://cran.r-project.org/package=power4mome)
[![Code size](https://img.shields.io/github/languages/code-size/sfcheung/power4mome.svg)](https://github.com/sfcheung/power4mome)
[![Last Commit at Main](https://img.shields.io/github/last-commit/sfcheung/power4mome.svg)](https://github.com/sfcheung/power4mome/commits/main)
[![R-CMD-check](https://github.com/sfcheung/power4mome/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sfcheung/power4mome/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

(Version 0.2.1, updated on 2026-03-23, [release history](https://sfcheung.github.io/power4mome/news/index.html))

# power4mome <a href="https://sfcheung.github.io/power4mome/"><img src="man/figures/logo.png" align="right" height="120" alt="power4mome website" /></a>

This package is for power analysis and
sample size determination for
moderation, mediation, and moderated mediation.

It includes functions for power
analysis and sample size determination for
moderation, mediation, and moderated mediation
effects in models fitted by
structural equation modeling (SEM) or multiple
linear regression. For SEM, both latent variable
models and path models of observed variables
are supported.

For more information on this package,
please visit its GitHub page:

https://sfcheung.github.io/power4mome/

A get-started guide illustrates how to
use this package:

https://sfcheung.github.io/power4mome/articles/power4mome.html

Code templates are also available for
common mediation, moderation, and
moderated mediation models:

https://sfcheung.github.io/power4mome/articles/

Major new features can be found [below](#major-new-features).

# Philosophy

The package was developed with this
philosophy:

- Easy to specify the population models,
  even with latent variables.

- As few manually-set numbers as possible
  when specifying the population models,
  with convenient default values.

- As few restrictions as possible
  on the form of the models, even when
  the models have latent factors.

- As automatic as possible in finding
  the sample size with the target power.

To achieve this comes with some costs,
and some of the goals conflict with other
goals: Being flexible usually means
being less user-friendly, and being
easy to specify the model usually means
not supporting some models.

Therefore, we also try to

- balance these goals, and

- allow the functions to be used in
  different ways, to accommodate
  scenarios that prioritize these goals
  differently.

# Major New Features

These are some major new features added
in `power4mome` 0.2.0 (see the
[release history](https://sfcheung.github.io/power4mome/news/index.html)
for a full list of changes).

## All-In-One Quick Functions

All-in-one "quick" functions for
common mediation models. Instead of
specifying the model and effect size
and then call `power4test()`, these
quick functions combine all the steps
into one. Learn more about them
in [these examples](https://sfcheung.github.io/power4mome/articles/index.html#common-mediation-models).

## Multigroup Power Analysis

Support for multigroup models has been
substantially improved. For example,
power analysis for detecting the group
difference in an indirect path can be
conducted. Several test functions have
also been updated to support testing
group differences.
See this [article](https://sfcheung.github.io/power4mome/articles/index.html#multigroup-models)
for an illustration. More demonstrations
will be added.

## Data Processors

Real studies rarely involve continuous
variables. Datasets with missing data
is also common. The argument `process_data`
has been added to `power4test()` and
friends for generating more realistic
data for power analysis. For example,
missing data can be simulated, and
ordinal variables can be created from
the continuous variables to simulate
binary or Likert scale items. Power analysis
can then be conducted by taking missing
data into account.
See these [articles](https://sfcheung.github.io/power4mome/articles/index.html#advanced-features)
for illustrations.

# Installation

The stable version at CRAN can be installed by `install.packages()`:

```r
install.packages("power4mome")
```

The latest developmental version of this package can be installed by `remotes::install_github`:

```r
remotes::install_github("sfcheung/power4mome")
```

# Background

Some of us the developers have developed the package
[`manymome`](https://sfcheung.github.io/manymome/)
(Cheung & Cheung, 2024)
for computing and testing effects in
models with mediation, mediation,
or moderated mediation. The tests are
usually done by simulation-based methods
such as Monte Carlo or bootstrap
confidence intervals, due to the complicated
sampling distributions of the effects.
Therefore, there are no simple ways to
determine the power of the test analytically
and accurately. The computation becomes
more complicated when latent variables
are involved, necessitating a simulation-based
method to estimate the sample size.

There are already many excellent packages
out there for estimating power in
structural equation modeling in general,
and some are also specifically for mediation
or moderated mediation. We are not intended
to replace with them or reinvent the wheel.
We just want to have a tool that meet
our own needs:

(a) It leverages on the flexibility of `manymome`
in testing an indirect effect or
conditional effect with little limitations
on the model.

(b) It allows users (us and our collaborators) to
specify the population model as easy (quickly) as
typical power analysis programs.

We ourselves know how to do the power
estimation on our own by simulation,
if necessary. However, time is usually a concern,
and we would like to
have a tool that, though specifically
designed with mediation, moderation,
and moderated mediation in mind and may
be limited in scope (though it is a
"big" scope), is easy for our daily use
in estimating power.

So here it is, `power4mome`, developed
with we ourselves as the users, but we
believe are also useful for others who
need to do power analysis for mediation,
moderation, and moderated mediation.

# Not Just That ...

But `power4mome` is not just for mediation,
moderation, and moderated mediation. We
avoided writing the functions just
for these effects, and have left room for
testing other effects, as hinted in some
examples in the help pages. They may be
introduced later. For now, supporting
effects that can be tested by `manymome`
is our priority.

# Issues

If you have any suggestions and found any bugs, please feel
free to open a GitHub issue:

https://github.com/sfcheung/power4mome/issues

Thanks.

# Reference

Cheung, S. F., & Cheung, S.-H. (2024).
*manymome*: An R package for computing
the indirect effects, conditional
effects, and conditional indirect
effects, standardized or unstandardized,
and their bootstrap confidence intervals,
in many (though not all) models.
*Behavior Research Methods, 56*(5),
4862--4882.
https://doi.org/10.3758/s13428-023-02224-z

#' @title All-in-One Power Estimation For Mediation Models
#'
#' @description All-in-one functions for
#' estimating power or finding the
#' region with target power for common
#' mediation models.
#'
#' @details
#'
#' # All-in-One Functions for Common Mediation Models
#'
#' These functions are wrappers
#' that call [power4test()] and
#' [n_region_from_power()] to (a)
#' estimate the level of power for
#' a mediation model, given the population
#' effects and the sample size, and
#' (b) find the region of sample sizes
#' with the levels of power not
#' significantly different from the
#' target power.
#'
#' They are convenient functions that
#' set the argument values automatically
#' for common mediation models before
#' calling [power4test()] and
#' [n_region_from_power()]. Please refer
#' to the help pages of these two
#' functions for the details on how
#' the estimation and the search are
#' conducted.
#'
#' For some arguments not described in
#' details here, please refer to the
#' help pages of [power4test()]
#' and [n_region_from_power()],
#'
#' # Simple Mediation Model
#'
#' The function [q_power_mediation_simple()]
#' can be used for the power analysis of
#' a simple mediation model with only
#' one mediator.
#'
#' This function will fit
#' the following model:
#'
#' \preformatted{"m ~ x
#'  y ~ m + x"}
#'
#' # Serial Mediation Model
#'
#' The function [q_power_mediation_serial()]
#' can be used for the power analysis of
#' a serial mediation model with only
#' any number of mediators.
#'
#' This is the model being fitted if
#' the model has two mediators:
#'
#' \preformatted{"m1 ~ x
#'  m2 ~ m1 + x
#'  y ~ m2 + m1 + x"}
#'
#' # Parallel Mediation Model
#'
#' The function [q_power_mediation_parallel()]
#' can be used for the power analysis of
#' a parallel mediation model with only
#' any number of mediators.
#'
#' This is the model being fitted if
#' the model has two mediators:
#'
#' \preformatted{"m1 ~ x
#'  m2 ~ x
#'  y ~ m2 + m1 + x"}
#'
#' # An Arbitrary Mediation Model
#'
#' The function [q_power_mediation()],
#' an advanced function,
#' can be used for the power analysis of
#' an arbitrary mediation model.
#' The model and the population effect
#' sizes are specified as in [power4test()].
#'
#' This is an example of a model with
#' both parallel paths and serial paths:
#'
#' \preformatted{model <-
#'  "
#'  m1 ~ x
#'  m21 ~ m1
#'  m22 ~ m1
#'  y ~ m21 + m22 + x
#'  "}
#'
#' \preformatted{pop_es <-
#' "
#' m1 ~ x: m
#' m21 ~ m1: m
#' m22 ~ m1: m
#' y ~ m21: m
#' y ~ m22: m
#' "}
#'
#' Knowledge of using [power4test()]
#' is required to use this advanced
#' function.
#'
#' If this advanced function is used,
#' users need to specify `test_fun`
#' as when using [power4test()], and
#' need to set `test_args` correctly
#'
#' @return
#' A named list of outputs.
#' If `mode` is `power`, then a
#' `power4test` object is set to the
#' element `power4test`.
#' If `mode` is `region`, then a
#' `n_region_from_power` object is
#' set of the element `n_region_from_power`.
#' If `mode` is `n_from_power`, then a
#' `n_from_power` object is
#' set to the `n_from_power` element.

#'
#' @inheritParams power4test
#' @inheritParams n_region_from_power
#'
#' @param n The sample size for the
#' first run of [power4test()]. Must be
#' set to a sample size if `mode` is
#' `"power"`. For the other modes,
#' if `NULL`, the default, it will be
#' determined internally.
#'
#' @param nrep The number of replications
#' to generate the simulated datasets.
#' Default is `NULL` and will be determined
#' internally.
#'
#' @param test_fun A function to do the
#' test. See 'Details' of [power4test()]
#' for the requirement of this function.
#'
#' @param test_more_args A named list of
#' additional arguments to be passed
#' to the test function
#' ([test_indirect_effect()] for
#' simple and serial mediation models,
#' and [test_k_indirect_effects()]
#' for parallel mediation models).
#' Similar to `test_args` in
#' [power4test()].
#'
#' @param iseed The seed for the random
#' number generator. Used by
#' [power4test()].
#'
#' @param seed The seed for the random
#' number generator. Used by
#' [n_region_from_power()].
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used when calling
#' other functions, if appropriate.
#'
#' @param algorithm The algorithm to be
#' used in mode `"region"` and `"n"`.
#' If `NULL`, then it will be determined
#' internally based on the `mode`. (The
#' default may be different from that
#' of [n_region_from_power()] and
#' [n_from_power()])
#'
#' @param ... For `q_power_mediation_*`,
#' these are optional arguments to
#' be passed to [power4test()] and
#' [n_region_from_power()].
#' For the `print` method, these
#' are optional arguments to
#' be passed to [rejection_rates()]
#' as well as other print methods
#' (see [print.power4test()] and
#' [print.n_region_from_power()]).
#' For the `plot` method, these
#' are optional arguments to
#' be passed to [plot.n_region_from_power()].
#' For the `summary` method, these
#' are optional arguments to
#' be passed to [summary.n_region_from_power()].
#'
#' @param mode For `q_power_mediation_*,
#' this is the mode of what to do.
#' If `"power"`, then
#' only [power4test()] will be called,
#' and the level of power will be
#' estimated. If `"region"`, then
#' the region of sample sizes with
#' levels of power not significantly
#' different from the target power will
#' be searched by calling
#' [n_region_from_power()]. For the
#' `print` method, this indicates
#' what to print. If `"region"`
#' and the output of [n_region_from_power()]
#' is available, it will print the results
#' of [n_region_from_power()] if available.
#' If `"n"`, then the output of
#' [n_from_power()] will be printed if
#' available.
#' If `"power"`, then the output of
#' [power4test()] will be printed.
#' If
#' `"all"`, then all available outputs
#' will be printed.
#'
#' @seealso See [power4test()] and
#' [n_region_from_power()] for full
#' details on how these functions
#' work.
#'
#' @examples
#'
#' \dontrun{
#'
#' # An arbitrary mediation model
#'
#' model <-
#' "
#' m1 ~ x
#' m21 ~ m1
#' m22 ~ m1
#' y ~ m21 + m22
#' "
#' pop_es <-
#' "
#' m1 ~ x: m
#' m21 ~ m1: m
#' m22 ~ m1: m
#' y ~ m21: m
#' y ~ m22: m
#' "
#'
#' # NOTE: In real power analysis:
#' # - Set R to an appropriate value.
#' # - Remove nrep or set nrep to the desired value.
#' # - Remove parallel or set it to TRUE to enable parallel processing.
#' # - Remove progress or set it to TRUE to see the progress.
#'
#' outa1 <- q_power_mediation(
#'     model = model,
#'     pop_es = pop_es,
#'     n = 100,
#'     R = 199,
#'     test_fun = test_k_indirect_effects,
#'     test_more_args = list(x = "x",
#'                           y = "y",
#'                           omnibus = "all"),
#'     seed = 1234,
#'     mode = "region",
#'     nrep = 20,
#'     parallel = FALSE,
#'     progress = FALSE
#'   )
#' outa1
#' summary(outa1)
#' plot(outa1)
#'
#' }
#'
#' @export
q_power_mediation <- function(
  model = NULL,
  pop_es = NULL,
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_fun = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
) {
  # A general quick functions for other
  # wrappers.
  ci_type <- match.arg(ci_type)
  mode <- match.arg(mode)

  if (mode == "power") {
    if (is.null(n)) {
      stop("'n' must be set to a sample size if mode is 'power'")
    }
  }

  if (progress) {
    tmp <- switch(
              mode,
              power = paste("Only the power for n =",
                            n,
                            "will be estimated."),
              region = paste("The region of sample sizes",
                             "for a power of",
                              sprintf("%3.2f", target_power),
                             "will be searched."),
              n = paste("A sample size",
                             "with estimated power close to",
                              sprintf("%3.2f", target_power),
                             "will be searched.")
    )
    cat("\n",
        tmp,
        "\n\n")
  }

  # ==== Set n if mode is region or n ====

  if (is.null(n)) {
    tmp <- match.call()
    if (mode == "region") {
      tmp2 <- eval(formals(n_region_from_power)$x_interval)
      if (is.null(tmp$x_interval)) {
        if (is.null(tmp2)) {
          n <- 200
        } else {
          n <- mean(tmp2)
        }
      } else {
        n <- mean(eval(tmp$x_interval))
      }
    }
    if (mode == "n") {
      tmp2 <- eval(formals(n_from_power)$x_interval)
      if (is.null(tmp$x_interval)) {
        if (is.null(tmp2)) {
          n <- 200
        } else {
          n <- mean(tmp2)
        }
      } else {
        n <- mean(eval(tmp$x_interval))
      }
    }
  }

  if (!parallel &&
      progress) {
    cat("\nIt is recommended to set 'parallel' to TRUE",
        "to speed up the computation.\n\n")
  }

  # model cannot be NULL for this general function.
  # The wrappers will be responsible for
  # setting model and pop_es.

  if (is.null(model)) {
    stop("'model' cannot be NULL calling this function directly")
  }
  if (is.null(pop_es)) {
    stop("'pop_es' cannot be NULL if calling this function directly")
  }

  if (is.null(test_fun)) {
    stop("'test_fun' cannot be NULL if calling this function directly")
  }

  ddd <- list(...)

  # ==== Process optional arguments =====

  power4test_args <- names(formals(power4test))
  ddd_power4test <- ddd[names(ddd) %in% power4test_args]

  n_region_from_power_args <- names(formals(n_region_from_power))
  ddd_n_from_region <- ddd[names(ddd) %in% n_region_from_power_args]

  n_from_power_args <- names(formals(n_from_power))
  ddd_n_from_power <- ddd[names(ddd) %in% n_from_power_args]

  # No need. This is the new default behavior of n_from_region
  # if (is.null(ddd_n_from_region$final_nrep)) {
  #   ddd_n_from_region$final_nrep <- nrep
  # }

  # ==== Set test arguments ====

  test_args0 <- list(
                    mc_ci = isTRUE(ci_type == "mc"),
                    boot_ci = isTRUE(ci_type == "boot")
                  )
  test_args0 <- utils::modifyList(
                  test_args0,
                  test_more_args
                )

  if (is.null(nrep)) {
    if (mode == "power") {
      nrep <- 400
    } else if (mode == "region") {
      if (is.null(algorithm)) {
        # Do not override the default, for now
        algorithm <- NULL
      }
      nrep <- 400
    } else if (mode == "n") {
      if (is.null(algorithm)) {
        algorithm <- "probabilistic_bisection"
      }
      if (algorithm == "probabilistic_bisection") {
        nrep <- 51
      }
    }
  }

  # ==== Call power4test() =====

  p4t_args <- list(
                nrep = nrep,
                model = model,
                pop_es = pop_es,
                n = n,
                number_of_indicators = number_of_indicators,
                reliability = reliability,
                loading_difference = loading_difference,
                reference = reference,
                R = R,
                ci_type = ci_type,
                test_fun = test_fun,
                test_args = test_args0,
                iseed = iseed %||% seed,
                parallel = parallel,
                progress = progress,
                ncores = max(1, parallel::detectCores(logical = FALSE) - 1)
              )

  p4t_args <- utils::modifyList(
                p4t_args,
                ddd_power4test
              )

  out <- do.call(
            power4test,
            p4t_args
          )

  # ==== Find the region ====

  if (mode == "region") {

    # Do this only if mode is not "power",
    # i.e., "region"

    n_reg_args <- list(
              object = out,
              target_power = target_power,
              progress = progress,
              simulation_progress = simulation_progress,
              max_trials = max_trials,
              algorithm = algorithm,
              seed = seed
            )

    n_reg_args <- utils::modifyList(
                  n_reg_args,
                  ddd_n_from_region
                )

    out2 <- do.call(
              n_region_from_power,
              n_reg_args
            )

  } else {
    out2 <- NULL
  }

  # ==== Find the n ====

  if (mode == "n") {

    n_args <- list(
              object = out,
              target_power = target_power,
              progress = progress,
              simulation_progress = simulation_progress,
              max_trials = max_trials,
              algorithm = algorithm,
              seed = seed
            )

    n_args <- utils::modifyList(
                  n_args,
                  ddd_n_from_power
                )
    out3 <- do.call(
              n_from_power,
              n_args
            )
  } else {
    out3 <- NULL
  }

  outz <- list(power4test = out,
               n_region_from_power = out2,
               n_from_power = out3)
  class(outz) <- c("q_power_mediation",
                   class(outz))
  outz
}

#' @param x The object
#' for the relevant methods.
#'
#' @return
#' The `print` method of `q_power_mediation`
#' returns `x` invisibly. Called for
#' its side effect.
#'
#' @rdname q_power_mediation
#' @export
print.q_power_mediation <- function(
  x,
  mode = c("all", "region", "n", "power"),
  ...
) {
  mode <- match.arg(mode)

  # ==== Print power4test ====

  if (mode %in% c("power", "all")) {
    cat("\n========== power4test Results ==========\n\n")
    print(x$power4test,
          ...)
    cat("\n========== power4test Power ==========\n\n")
    print(rejection_rates(x$power4test,
                          ...),
          ...)
  }

  # ==== Print n_region_from_power ====

  if (mode %in% c("region", "all")) {
    if (is.null(x$n_region_from_power)) {
      # cat("\n'mode' is not 'region' and results not available.\n\n")
    } else {
      cat("\n========== n_region_from_power Results ==========\n\n")
      print(x$n_region_from_power,
            ...)
    }
  }

  # ==== Print n_from_power ====

  if (mode %in% c("n", "all")) {
    if (is.null(x$n_from_power)) {
      # cat("\n'mode' is not 'n_from_power' and results not available.\n\n")
    } else {
      cat("\n========== n_from_power Results ==========\n\n")
      print(x$n_from_power,
            call = FALSE,
            ...)
    }
  }

  invisible(x)
}

#' @return
#' The `plot`-method of `q_power_mediation`
#' returns `NULL`.
#' It will plot either the output of
#' [n_region_from_power()]
#' (mode `"region"`) or the output of
#' [n_from_power()] (mode `"n"`).
#' If no output from
#' either of these functions is available,
#' nothing wil be plotted.
#'
#' @rdname q_power_mediation
#' @export
plot.q_power_mediation <- function(
  x,
  ...
) {
  # x will have only one of the followings
  if (!is.null(x$n_region_from_power)) {
    plot(x$n_region_from_power,
         ...)
  }
  if (!is.null(x$n_from_power)) {
    plot(x$n_from_power,
         ...)
  }
  invisible(NULL)
}

#' @param object For the `summary`
#' method of [q_power_mediation()]
#' outputs.
#'
#' @return
#' The `summary` method for
#' `q_power_mediation` objects returns
#' the output of [summary()] for
#' either the output of [n_region_from_power()]
#' (mode `"region"`) or the output of
#' [n_from_power()] (mode `"n"`).
#' Return `NULL` if no output from
#' either of these functions is available.
#'
#' @rdname q_power_mediation
#' @export
summary.q_power_mediation <- function(
                                object,
                                ...) {
  if (!is.null(object$n_region_from_power)) {
    # stop("'mode' is not 'region' and summary cannot be generated.")
    out <- summary(object$n_region_from_power,
                   ...)
    return(out)
  }
  if (!is.null(object$n_from_power)) {
    # stop("'mode' is not 'region' and summary cannot be generated.")
    out <- summary(object$n_from_power,
                   ...)
    return(out)
  }
  NULL
}

#' @param a For a simple mediation
#' model, this is the population effect
#' size for the path from `x` to `m` .
#'
#' @param b For a simple mediation
#' model, this is the population effect
#' size for the path from `m` to `y` .
#'
#' @param cp For a simple mediation
#' model, this is the population effect
#' size for the direct path from
#' `c` to `y` .
#'
#' @examples
#'
#' # Simple mediation model
#'
#' # NOTE: In real power analysis:
#' # - Set R to an appropriate value.
#' # - Remove nrep or set nrep to the desired value.
#' # - Remove parallel or set it to TRUE to enable parallel processing.
#' # - Remove progress or set it to TRUE to see the progress.
#'
#' out <- q_power_mediation_simple(
#'     a = "m",
#'     b = "m",
#'     cp = "n",
#'     n = 50,
#'     R = 79,
#'     seed = 1234,
#'     nrep = 10,
#'     parallel = FALSE,
#'     progress = FALSE
#'   )
#' out
#'
#' # If mode = "region" is added, can call the following
#' # summary(out)
#' # plot(out)
#'
#' @rdname q_power_mediation
#' @export
q_power_mediation_simple <- function(
  a = "m",
  b = "m",
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
) {

  # ==== Setup the model =====

  model <- paste0(
            c("m ~ x",
              "y ~ m + x",
              ""),
            collapse = "\n"
          )
  pop_es <-
    sprintf(
      paste0(
        c(
          "m ~ x: %1$s",
          "y ~ m: %2$s",
          "y ~ x: %3$s",
          ""
        ),
        collapse = "\n"
      ),
      a, b, cp
    )

  # ==== Process number of indicators and reliability =====

  # Do not do the test for now because some of the variables
  # may be observed variables

  # if (length(number_of_indicators) > 1) {
  #     if (length(number_of_indicators) != 3) {
  #       stop("number_of_indicators must be one or three values")
  #     }
  #     if (!setequal(
  #           names(number_of_indicators),
  #           c("x", "m", "y"))) {
  #       stop("Names of number_of_indicators must 'x', 'm', and 'y'")
  #     }
  #   }
  # if (length(reliability) > 1) {
  #   if (length(reliability) != 3) {
  #     stop("reliability must be one or three values")
  #   }
  #   if (!setequal(
  #         names(reliability),
  #         c("x", "m", "y"))) {
  #     stop("Names of reliability must 'x', 'm', and 'y'")
  #   }
  # }

  # ==== Set test arguments ====

  # Set model specific arguments

  test_args0 <- utils::modifyList(
                  test_more_args,
                  list(x = "x",
                       m = "m",
                       y = "y")
                )

  # ==== Call the general function ====

  out <- q_power_mediation(
    model = model,
    pop_es = pop_es,
    number_of_indicators = number_of_indicators,
    reliability = reliability,
    loading_difference = loading_difference,
    reference = reference,
    test_fun = "test_indirect_effect",
    test_more_args = test_args0,
    target_power = target_power,
    nrep = nrep,
    n = n,
    R = R,
    ci_type = ci_type,
    seed = seed,
    iseed = iseed,
    parallel = parallel,
    progress = progress,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    algorithm = algorithm,
    ...,
    mode = mode
  )
  out
}

#' @param ab For a serial mediation
#' model, this is a numeric vector
#' of the population effect
#' sizes along the path
#' `x->m1->m2->...->y`.
#'
#' @param ab_other Should be one single
#' value. This is the population effect
#' sizes of all other paths not along
#' `x->m1->m2->...->y`, except for the
#' direct path from `x` to `y`.
#'
#' @examples
#'
#' \dontrun{
#' # Serial mediation model
#'
#' # NOTE: In real power analysis:
#' # - Set R to an appropriate value.
#' # - Remove nrep or set nrep to the desired value.
#' # - Remove parallel or set it to TRUE to enable parallel processing.
#' # - Remove progress or set it to TRUE to see the progress.
#'
#' outs <- q_power_mediation_serial(
#'     ab = c("s", "m", "l"),
#'     ab_others = "n",
#'     cp = "s",
#'     n = 50,
#'     R = 199,
#'     seed = 1234,
#'     mode = "region",
#'     nrep = 20,
#'     parallel = FALSE,
#'     progress = FALSE
#'   )
#' outs
#' summary(outs)
#' plot(outs)
#' }
#'
#' @rdname q_power_mediation
#' @export
q_power_mediation_serial <- function(
  ab = c("m", "m"),
  ab_other = "n",
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
) {

  # ==== Setup the model =====

  p <- length(ab)

  m_names <- c(paste0("m", seq_len(p - 1)),
               "y")

  all_m_paths <- character(0)
  x_previous <- "x"
  for (i in m_names) {
    all_m_paths <- c(all_m_paths,
                  paste0(i,
                         " ~ ",
                         x_previous)
                )
    x_previous <- union(x_previous, i)
  }

  model_m <- character(0)
  m_previous <- character(0)
  for (i in m_names) {
    model_m <- c(model_m,
                 paste0(i,
                        " ~ ",
                        paste0(c(m_previous,
                                 "x"),
                               collapse = " + "))
                )
    m_previous <- union(m_previous, i)
  }
  model <- paste0(
            model_m,
            collapse = "\n"
          )

  m_paths <- paste0(m_names,
         " ~ ",
         c("x", m_names[-p])
        )
  m_other_paths <- setdiff(
                      all_m_paths,
                      m_paths
                    )
  es_ab <- paste0(
            m_paths,
            ": ",
            ab
          )
  es_m_other <- paste0(
            setdiff(m_other_paths,
                    "y ~ x"),
            ": ",
            ab_other
          )
  es_y <- paste0(
            "y ~ x: ",
            cp
          )
  pop_es <- paste0(
      c(es_ab,
        es_m_other,
        es_y),
      collapse = "\n"
    )

  # ==== Process number of indicators and reliability =====

  # Do not do the test for now because some of the variables
  # may be observed variables

  # if (length(number_of_indicators) > 1) {
  #     if (length(number_of_indicators) != 3) {
  #       stop("number_of_indicators must be one or the number of factors")
  #     }
  #     if (!setequal(
  #           names(number_of_indicators),
  #           c("x", "m", "y"))) {
  #       stop("Names of number_of_indicators must 'x', 'm', and 'y'")
  #     }
  #   }
  # if (length(reliability) > 1) {
  #   if (length(reliability) != 3) {
  #     stop("reliability must be one or three values")
  #   }
  #   if (!setequal(
  #         names(reliability),
  #         c("x", "m", "y"))) {
  #     stop("Names of reliability must 'x', 'm', and 'y'")
  #   }
  # }


  # ==== Set the test arguments ====

  test_more_args <- utils::modifyList(
                      test_more_args,
                      list(x = "x",
                           y = "y",
                           m = c(m_names[-p]))
                    )

  # ==== Call the general function ====

  out <- q_power_mediation(
    model = model,
    pop_es = pop_es,
    number_of_indicators = number_of_indicators,
    reliability = reliability,
    loading_difference = loading_difference,
    reference = reference,
    test_fun = "test_indirect_effect",
    test_more_args = test_more_args,
    target_power = target_power,
    nrep = nrep,
    n = n,
    R = R,
    ci_type = ci_type,
    seed = seed,
    iseed = iseed,
    parallel = parallel,
    progress = progress,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    algorithm = algorithm,
    ...,
    mode = mode
  )
  out
}

#' @param as For a parallel mediation
#' model, this is a numeric vector
#' of the population effect
#' sizes for the paths from `x` to
#' the mediators: `x->m1`, `x->m2`,
#' ... `x->mp`, for a parallel mediation
#' model with `p` mediators.
#'
#' @param bs For a parallel mediation
#' model, this is a numeric vector
#' of the population effect
#' sizes for the paths from the
#' mediators to `y`: `m1->y`, `m2->y`,
#' ... `mp->y`, for a parallel mediation
#' model with `p` mediators.
#'
#' @param omnibus `"all_sig"`, the
#' default, then
#' the test is declared significant if
#' *all* paths are significant. If
#' `"at_least_one_sig"`, then only
#' one row of test is stored, and the
#' test is declared significant if
#' at least one of the paths is
#' significant. If `"at_least_k_sig"`,
#' then only one row of test is stored,
#' and the test is declared significant
#' if at least `k` of the paths is
#' significant, `k` determined by the
#' argument `at_least_k`.
#'
#' @param at_least_k The minimum number
#' of paths required to be significant
#' for the omnibus test to be considered
#' significant. Used when
#' `omnibus` is `"at_least_k_sig"`.
#'
#' @examples
#'
#' \dontrun{
#' # Parallel mediation model
#'
#' # NOTE: In real power analysis:
#' # - Set R to an appropriate value.
#' # - Remove nrep or set nrep to the desired value.
#' # - Remove parallel or set it to TRUE to enable parallel processing.
#' # - Remove progress or set it to TRUE to see the progress.
#'
#' outp <- q_power_mediation_parallel(
#'     as = c("s", "m"),
#'     bs = c("m", "s"),
#'     cp = "n",
#'     n = 100,
#'     R = 199,
#'     seed = 1234,
#'     mode = "region",
#'     nrep = 20,
#'     parallel = FALSE,
#'     progress = FALSE
#'   )
#' outp
#' summary(outp)
#' plot(outp)
#' }
#'
#' @rdname q_power_mediation
#' @export
q_power_mediation_parallel <- function(
  as = c("m", "m"),
  bs = c("m", "m"),
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  omnibus = c("all_sig", "at_least_one_sig", "at_least_k_sig"),
  at_least_k = 1,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
) {

  omnibus <- match.arg(omnibus)

  # ==== Setup the model =====

  if (length(as) != length(bs)) {
    stop("'as' and 'bs' must have the same number of elements.")
  }

  p <- length(bs)

  m_names <- paste0("m", seq_len(p))
  model_m <- paste0(m_names, " ~ x")
  model_y <- paste0("y ~ ",
                    paste0(m_names, collapse = " + "),
                    " + x")
  model <- paste0(
            c(model_m,
              model_y),
            collapse = "\n"
          )
  f <- function(m, a) {
      paste0(m, " ~ x: ",
             a)
    }
  es_m <- mapply(
              f,
              m = m_names,
              a = as,
              SIMPLIFY = TRUE
            )
  es_y <- paste0(
            "y ~ ",
            m_names,
            ": ",
            bs
          )
  es_cp <- paste0("y ~ x: ", cp)
  pop_es <- paste0(
      c(es_m,
        es_y,
        es_cp),
      collapse = "\n"
    )

# ==== Process number of indicators and reliability =====

  # Do not do the test for now because some of the variables
  # may be observed variables

  # if (length(number_of_indicators) > 1) {
  #     if (length(number_of_indicators) != 3) {
  #       stop("number_of_indicators must be one or three values")
  #     }
  #     if (!setequal(
  #           names(number_of_indicators),
  #           c("x", "m", "y"))) {
  #       stop("Names of number_of_indicators must 'x', 'm', and 'y'")
  #     }
  #   }
  # if (length(reliability) > 1) {
  #   if (length(reliability) != 3) {
  #     stop("reliability must be one or three values")
  #   }
  #   if (!setequal(
  #         names(reliability),
  #         c("x", "m", "y"))) {
  #     stop("Names of reliability must 'x', 'm', and 'y'")
  #   }
  # }

  # ==== Set the test arguments ====

  test_more_args <- utils::modifyList(
                      test_more_args,
                      list(x = "x",
                           y = "y",
                           omnibus = omnibus,
                           at_least_k = at_least_k)
                    )

  # ==== Call the general function ====

  out <- q_power_mediation(
    model = model,
    pop_es = pop_es,
    number_of_indicators = number_of_indicators,
    reliability = reliability,
    loading_difference = loading_difference,
    reference = reference,
    test_fun = "test_k_indirect_effects",
    test_more_args = test_more_args,
    target_power = target_power,
    nrep = nrep,
    n = n,
    R = R,
    ci_type = ci_type,
    seed = seed,
    iseed = iseed,
    parallel = parallel,
    progress = progress,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    algorithm = algorithm,
    ...,
    omnibus = omnibus,
    at_least_k = at_least_k,
    mode = mode
  )
  out
}

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Parallel mediation =====

options(power4mome.bz = TRUE)
system.time(
outp <- q_power_mediation_parallel(
    as = c("s", "m"),
    bs = c("m", "s"),
    cp = "n",
    n = 50,
    nrep = 10,
    R = 199,
    seed = 1234,
    mode = "region",
    max_trials = 2,
    progress = progress0,
    parallel = FALSE,
    simulation_progress = progress0,
    tolerance = .20
  )
)
# plot(outp)
expect_no_error(capture.output(print(outp)))
expect_no_error(summary(outp))
expect_equal(outp$n_region_from_power$above$args$final_nrep,
             10)

})

skip_on_cran()

test_that("q_power_mediation: R_for_bz", {

progress0 <- !testthat::is_testing()

# ===== Simple mediation =====

system.time(
out1 <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    nrep = 10,
    R = R_for_bz(100),
    seed = 1234,
    mode = "region",
    max_trials = 1,
    progress = progress0,
    parallel = FALSE,
    simulation_progress = progress0,
    tolerance = .20
  )
)

system.time(
out2 <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    nrep = 10,
    R = R_for_bz(200),
    seed = 1234,
    mode = "region",
    max_trials = 1,
    progress = progress0,
    parallel = FALSE,
    simulation_progress = progress0,
    tolerance = .20
  )
)

expect_equal(attr(out1$power4test, "args")$R,
             79)
expect_equal(attr(out2$power4test, "args")$R,
             199)

})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Simple mediation =====

system.time(
out <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 100,
    final_nrep = 60,
    R = 99,
    seed = 1234,
    mode = "n",
    x_interval = c(100, 120),
    progress = progress0,
    parallel = FALSE
  )
)
if (!inherits(out, "error")) {
expect_no_error(capture.output(print(out)))
expect_no_error(summary(out))
}
})

skip("A long test with parallel processing. Test interactively")

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Simple mediation =====

system.time(
out <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    final_nrep = 2000,
    R = 199,
    seed = 1234,
    mode = "n",
    progress = progress0,
    parallel = TRUE
  )
)
expect_no_error(capture.output(print(out)))
expect_no_error(summary(out))

})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Parallel mediation =====

system.time(
outp <- q_power_mediation_parallel(
    as = c("m", "m"),
    bs = c("m", "m"),
    cp = "n",
    n = 125,
    final_nrep = 60,
    R = 99,
    seed = 2345,
    mode = "n",
    x_interval = c(125, 155),
    progress = progress0,
    parallel = FALSE
  )
)
if (!inherits(outp, "error")) {
expect_no_error(capture.output(print(outp)))
expect_no_error(summary(outp))
}

})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Arbitrary model =====

options(power4mome.bz = TRUE)
model <-
"
m1 ~ x
m21 ~ m1
m22 ~ m1
y ~ m21 + m22
"
pop_es <-
"
m1 ~ x: m
m21 ~ m1: m
m22 ~ m1: m
y ~ m21: m
y ~ m22: m
"

system.time(
outa <- q_power_mediation(
    model = model,
    pop_es = pop_es,
    n = 100,
    R = 199,
    test_fun = test_indirect_effect,
    test_more_args = list(x = "x",
                          y = "y",
                          m = c("m1", "m21")),
    seed = 1234,
    mode = "region",
    nrep = 10,
    max_trials = 2,
    parallel = FALSE,
    progress = progress0,
    simulation_progress = progress0,
    tolerance = .20
  )
)
expect_no_error(capture.output(print(outa)))
expect_no_error(summary(outa))
expect_equal(outa$n_region_from_power$above$args$final_nrep,
             10)

})

skip("A long test with parallel processing. Test interactively")

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Parallel mediation =====

system.time(
outp <- q_power_mediation_parallel(
    as = c("s", "m"),
    bs = c("m", "s"),
    cp = "n",
    n = 50,
    final_nrep = 2000,
    R = 199,
    seed = 2345,
    mode = "n",
    progress = progress0,
    parallel = TRUE
  )
)
# plot(outp)
expect_no_error(capture.output(print(outp)))
expect_no_error(summary(outp))


})

skip_on_cran()

test_that("q_power_mediation: rejection", {

# ===== Simple mediation =====

system.time(
out <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    nrep = 10,
    R = 199,
    seed = 1234,
    mode = "region",
    max_trials = 1,
    progress = !is_testing(),
    parallel = FALSE,
    simulation_progress = !is_testing(),
    tolerance = .20
  )
)

system.time(
out2 <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "s",
    target_power = .80,
    R = 199,
    final_nrep = 60,
    final_R = 199,
    seed = 1234,
    mode = "n",
    progress = !is_testing(),
    control = list(total_nrep = 200),
    parallel = FALSE
  )
)

system.time(
out3 <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "s",
    nrep = 10,
    R = 199,
    n = 100,
    seed = 1234,
    mode = "power",
    progress = !is_testing(),
    parallel = FALSE
  )
)

rejection_rates(out)
rejection_rates(out2)
rejection_rates(out3)

chk1 <- rejection_rates(out, all_columns = TRUE)
chk2 <- rejection_rates(out$n_region_from_power, all_columns = TRUE)
expect_equal(chk1$reject,
             chk2$reject)

chk1 <- rejection_rates(out2, all_columns = TRUE)
chk2 <- rejection_rates(out2$n_from_power, all_columns = TRUE)
expect_equal(chk1$reject,
             chk2$reject)

chk1 <- rejection_rates(out3, all_columns = TRUE)
chk2 <- rejection_rates(out3$power4test, all_columns = TRUE)
expect_equal(chk1$reject,
             chk2$reject)
})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Serial mediation =====

system.time(
outs <- tryCatch(q_power_mediation_serial(
    ab = c("m", "m", "l"),
    ab_others = "nil",
    cp = "nil",
    n = 80,
    final_nrep = 60,
    R = 99,
    seed = 1234,
    mode = "n",
    x_interval = c(80, 130),
    progress = progress0,
    parallel = FALSE
  ),
  error = function(e) e)
)
if (!inherits(outs, "error")) {
expect_no_error(capture.output(print(outs)))
expect_no_error(summary(outs))
}

})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Simple mediation =====

options(power4mome.bz = TRUE)
system.time(
out <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    nrep = 10,
    R = 199,
    seed = 1234,
    mode = "region",
    max_trials = 1,
    progress = progress0,
    parallel = FALSE,
    simulation_progress = progress0,
    tolerance = .20
  )
)
expect_no_error(capture.output(print(out)))
expect_no_error(summary(out))
expect_equal(out$n_region_from_power$below$args$final_nrep,
             10)

})

skip_on_cran()

test_that("q_power_mediation", {

progress0 <- !testthat::is_testing()

# ===== Serial mediation =====

options(power4mome.bz = TRUE)
system.time(
outs <- tryCatch(q_power_mediation_serial(
    ab = c("s", "m", "l"),
    ab_others = "n",
    cp = "s",
    n = 50,
    nrep = 10,
    R = 199,
    seed = 1234,
    mode = "region",
    max_trials = 1,
    progress = progress0,
    parallel = FALSE,
    simulation_progress = progress0,
    tolerance = .20
  ),
  error = function(e) e)
)
if (!inherits(outs, "error")) {
expect_no_error(capture.output(print(outs)))
expect_no_error(summary(outs))
expect_equal(outs$n_region_from_power$below$args$final_nrep,
             10)
}

})

set_n_range_by_x <- function(
                        object,
                        target_power = .80,
                        k = 4,
                        n_max = 1000,
                        object_by_org = NULL,
                        what = NULL,
                        goal = NULL,
                        tol = NULL,
                        ci_level = .95) {
  # No need for other arguments because they will be retrieved from object_by_org
  reject0by <- rejection_rates(object_by_org,
                               level = ci_level,
                               add_se = TRUE,
                               all_columns = TRUE)
  # Always return a value because closest_ok is TRUE
  i0 <- find_solution(
                object_by_org,
                target_power = target_power,
                ci_level = ci_level,
                what = what,
                tol = tol,
                goal = goal,
                final_nrep = 0,
                closest_ok = TRUE,
                weight_by = "nrep",
                debug = TRUE
              )
  n0 <- reject0by$n[i0]
  what0 <- switch(
              what,
              point = "reject",
              lb = "reject_ci_lo",
              ub = "reject_ci_hi"
            )
  est0 <- reject0by[i0, what0, drop = TRUE]
  # Display est0 a little bit
  if (est0 == target_power) {
    est0 <- target_power * .95
  }
  if (est0 < target_power) {
    i1a <- reject0by[, what0, drop = TRUE] > target_power
    i1b <- reject0by$n > n0
    i1 <- i1a & i1b
  } else {
    i1a <- reject0by[, what0, drop = TRUE] < target_power
    i1b <- reject0by$n < n0
    i1 <- i1a & i1b
  }
  if (any(i1)) {
    i1 <- which(i1)[1]
    n1 <- reject0by$n[i1]
  } else {
    n1 <- n0 * target_power / est0
  }
  n1 <- ceiling(n1)
  n0 <- min(n0, n_max)
  n1 <- min(n1, n_max)
  # Displace n0 and n1
  n_out <- range(n0, n1)
  n_out
}

set_n_range <- function(object,
                        target_power = .80,
                        k = 4,
                        n_max = 1000) {
  n0 <- attr(object, "args")$n
  # No need for other arguments because only `reject` is used
  reject0 <- rejection_rates(object)
  power0 <- reject0$reject[1]
  if (n0 > n_max) {
    stop("Initial sample size (",
          n0,
          ") is equal to or greater than 'n_max' (",
          n_max,
          "). Please increase 'n_max'.")
  }
  if (power0 < target_power) {
    b <- power0 / n0
    n_end <- min(round(target_power / b),
                 n_max,
                 n0 * 2)
    n_out <- seq(from = n0,
                 to = n_end,
                 length.out = k)
    n_out <- round(n_out)
    return(n_out)
  } else {
    # If power0 == target_power,
    # Be conservative and decrease power by a small amount,
    # if (power0 == target_power) {
    #   power0 <- target_power * .80
    # }
    b <- power0 / n0
    n_end <- min(round(target_power / b),
                 n_max,
                 n0 / 2)
    n_out <- seq(from = n_end,
                 to = n0,
                 length.out = k)
    n_out <- round(n_out)
    return(n_out)
  }
}

estimate_n <- function(power_n_fit,
                       target_power = .80,
                       interval = c(50, 2000),
                       extendInt = "no") {
  f <- function(n) {
    # stats::predict(power_n_fit,
    #                newdata = list(n = n)) - target_power
    stats::predict(power_n_fit,
                   newdata = list(x = n)) - target_power
  }
  n_target <- tryCatch(stats::uniroot(f,
                             interval = interval,
                             extendInt = extendInt),
                       error = function(e) e)
  if (inherits(n_target, "error")) {
    # Return NA if error occurred. E.g.,
    # - Root not in the interval.
    return(NA)
  }
  n_target <- round(n_target$root)
  # Negative sample size should be handled by the
  # calling function, e.g,, estimate_n_range().
  return(n_target)
}

estimate_n_range <- function(power_n_fit,
                             target_power = .80,
                             k = 5,
                             tolerance = .20,
                             power_min = .01,
                             power_max = .99,
                             interval = c(50, 2000),
                             extendInt = "upX",
                             n_to_exclude = NULL) {
  power_j <- seq(from = max(target_power - tolerance, power_min),
                 to = min(target_power + tolerance, power_max),
                 length.out = k)
  if (isFALSE(target_power %in% power_j)) {
    power_j <- sort(c(power_j, target_power))
  }
  out <- sapply(power_j,
                function(x) {
                  estimate_n(power_n_fit = power_n_fit,
                             target_power = x,
                             interval = interval,
                             extendInt = extendInt)
                })
  out <- ceiling(out)

  # If NA, have to do random sampling
  i <- is.na(out)
  if (any(i)) {
    # Duplication is OK because it will be fixed later
    out[i] <- sample(seq(interval[1],
                         interval[2]),
                     size = sum(i),
                     replace = FALSE)
  }

  # Check invalid Ns

  i <- check_n(ns = out,
               interval = interval,
               n_to_exclude = n_to_exclude,
               extendInt = extendInt)

  if (isFALSE(any(i))) {
    return(out)
  }
  # Replace invalid Ns by random Ns
  # Do not use the full interval
  # But can include Ns already considered
  new_interval1 <- min(n_to_exclude, interval)
  new_interval2 <- max(ceiling(max(n_to_exclude) * 1.25),
                       interval[2])
  n_pool <- setdiff(seq(new_interval1, new_interval2),
                    c(n_to_exclude, out[!i]))
  for (q in 1:10) {
    out[i] <- out[i] + sample(c(seq(1, q),
                                -seq(1, q)),
                              size = sum(i),
                              replace = TRUE)
    i <- check_n(out,
                 interval = c(new_interval1, new_interval2),
                 n_to_exclude = n_to_exclude,
                 extendInt = extendInt)
    if (isFALSE(any(i))) {
      # All ns OK
      break
    }
  }
  if (any(i)) {
    # Have to do random sample
    n_pool <- setdiff(seq(new_interval1, new_interval2),
                      c(n_to_exclude, out[!i]))
    n_new <- sample(n_pool, size = sum(i))
    out[i] <- n_new
  }
  return(out)
}

check_n <- function(ns,
                    interval,
                    n_to_exclude,
                    extendInt,
                    hard_min = 5) {
  i <- rep(FALSE, length(ns))
  if (isFALSE(extendInt %in% c("yes", "upX"))) {
    i[ns > interval[2]] <- TRUE
  }
  if (isFALSE(extendInt %in% c("yes", "downX"))) {
    i[ns < interval[1]] <- TRUE
  }

  # Ns used are considered invalid
  i[ns %in% n_to_exclude] <- TRUE

  # Duplicated are considered invalid
  i[duplicated(ns)] <- TRUE

  i[is.na(ns)] <- TRUE

  i[ns < hard_min] <- TRUE

  i
}

# ==== Diagnostic function ====

pba_diag <- function(a_out) {
  (x_tmp <- ceiling(q_dfun(a_out$dfun_out, prob = .50)))
  (x_lo <- q_dfun(a_out$dfun_out, .05))
  (x_hi <- q_dfun(a_out$dfun_out, .95))
  parold <- par(no.readonly = TRUE)
  layout(matrix(1:6, nrow = 3, byrow = TRUE))
  plot(a_out$fit_1)
  abline(h = .80, col = "blue", lwd = 1)
  abline(h = a_out$f_power, col = "red", lwd = 1)
  abline(v = x_tmp, col = "red", lwd = 1)
  abline(v = c(x_lo, x_hi), col = "black", lwd = 1, lty = "dotted")
  plot(a_out$x_history, type = "l")
  abline(h = x_tmp, col = "blue", lwd = 1)
  abline(h = c(x_lo, x_hi), col = "black", lwd = 1, lty = "dotted")
  plot(a_out$f_history, type = "l")
  abline(h = 0, col = "blue", lwd = 1)
  tmp <- c(q_dfun(a_out$dfun_out, .01),
           q_dfun(a_out$dfun_out, .99))
  plot(a_out$dfun_out, type = "l", xlim = tmp)
  abline(v = x_tmp, col = "red", lwd = 1)
  abline(v = c(x_lo, x_hi), col = "black", lwd = 1, lty = "dotted")
  plot(a_out$fit_1,
      xlim = c(x_lo * .9, x_hi * 1.1))
  abline(h = .80, col = "blue", lwd = 1)
  abline(h = a_out$f_power, col = "red", lwd = 1)
  abline(v = x_tmp, col = "red", lwd = 1)
  abline(v = c(x_lo, x_hi), col = "black", lwd = 1, lty = "dotted")
  hdr_power_history <- a_out$hdr_power_history
  tmp <- sapply(unlist(hdr_power_history, recursive = FALSE),
                range)
  plot(seq_along(hdr_power_history),
       y = rep(a_out$f_power, length(hdr_power_history)),
       ylim = range(tmp),
       type = "l")
  # points(a_out$f_power + a_out$f_history)
  points(a_out$reject_by_power_curve_history)
  for (i in seq_along(hdr_power_history)) {
    for (y in hdr_power_history[[i]]) {
      arrows(
          x0 = i,
          y0 = y[1],
          x1 = i,
          y1 = y[2],
          lty = "dotted",
          code = 3,
          angle = 90,
          length = ifelse(y[1] == y[2], 0, .05)
        )
    }
  }
  tmp <- which(a_out$final_check_history)
  abline(v = tmp, col = "red")
  par(parold)
  hdr_h <- a_out$hdr_power_history
  tmp <- sapply(hdr_h,
            \(x) ifelse(length(x) == 1,
                        diff(x[[1]]),
                        NA)
            )
  tmp2 <- sapply(hdr_h,
            \(x) ifelse(length(x) == 1,
                        (x[[1]][1] <= a_out$f_power) &
                        (x[[1]][1] >= a_out$f_power),
                        NA)
            )
  print(tmp <= a_out$hdr_power_tol)
}

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("probabilistic bisection: n: reuse trials", {

# ==== n ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  parallel = FALSE,
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90)

## ==== Close enough ====

set.seed(4321)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  final_nrep = 2000)
pba_diag(a_out)
plot(a_out$time_passed_history, type = "l")

a_out_by_x_1 <- a_out$by_x_1
names(a_out_by_x_1)

## ==== All settings identical ====

set.seed(4321)
a_out2 <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = a_out_by_x_1,
                                  x_interval = c(721, 800),
                                  goal = "close_enough",
                                  final_nrep = 2000)
tryCatch(pba_diag(a_out2),
         error = function(e) e)

# The solution is in the bound, but
# only the trial with nrep == initial_nrep will be used.

set.seed(4321)
a_out2b <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = a_out_by_x_1,
                                  x_interval = c(795, 2000),
                                  goal = "close_enough",
                                  final_nrep = 2000)
tryCatch(pba_diag(a_out2b),
         error = function(e) e)

## ==== Different final_nrep ====

set.seed(4321)
a_out3 <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = a_out_by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  final_nrep = 1000)
a_out3$x_out
tryCatch(pba_diag(a_out3),
         error = function(e) e)

## ==== Reuse outputs ====

set.seed(3)
a_out4 <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = a_out_by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  final_nrep = 2000)
a_out4$x_out
tryCatch(pba_diag(a_out4),
         error = function(e) e)
plot(a_out4$time_passed_history, type = "l")

expect_equal(attr(a_out4$by_x_1[[a_out4$i2]], "args")$iseed,
             attr(a_out$by_x_1[[a_out$i2]], "args")$iseed)

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("probabilistic bisection: n, indirect", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

# ==== n ====

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 199,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
rejection_rates(out)

by_x_1 <- power4test_by_n(out,
                          n = 90)

## ==== Close enough ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = ceiling(a_out$x_out),
              R = 1000,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)

## ==== ub ====

set.seed(4321)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  what = "ub",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = ceiling(a_out$x_out),
              R = 1000,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== lb ====

set.seed(4312)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  x_interval = c(100, 1000),
                                  goal = "close_enough",
                                  what = "lb",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = ceiling(a_out$x_out),
              R = 1000,
              nrep = 2000,
              iseed = 3456)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

# Solution already in interval

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(775, 800))
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(600, 775))
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# ==== es ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: l
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 199,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE,
                                   test_method = "pvalue"),
                  iseed = 1234,
                  parallel = TRUE)
rejection_rates(out)

by_x_1 <- power4test_by_es(out,
                           pop_es_name = "y~m",
                           pop_es_values = c(.10))

## ==== Close enough ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "y~m",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  x_interval = c(.00, .50),
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "y~m"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              R = 1000,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== ub ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "y~m",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  what = "ub",
                                  x_interval = c(.30, .60),
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "y~m"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              R = 1000,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== lb ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "y~m",
                                  R = 199,
                                  by_x_1 = by_x_1,
                                  what = "lb",
                                  x_interval = c(.30, .60),
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "y~m"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              R = 2000,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("probabilistic bisection: n", {

# ==== n ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  parallel = FALSE,
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90)

## ==== Close enough ====

set.seed(4321)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(50, 2000),
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 2000,
              iseed = 147)
rejection_rates(tmp_out)

## ==== ub ====

set.seed(2345)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(50, 2000),
                                  what = "ub",
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== lb ====

set.seed(271828)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  what = "lb",
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 2000,
              iseed = 12345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

# Solution already in interval

# This should nearly never happen because
# initial_nrep should be less than final_nrep.

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(795, 800),
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(600, 775))
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)


# ==== es ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  iseed = 1234)

by_x_1 <- power4test_by_es(out,
                           pop_es_name = "m~x",
                           pop_es_values = c(.10))

## ==== Close enough ====

set.seed(345)
a_out <- power_algorithm_prob_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   goal = "close_enough",
                                   x_interval = c(.10, .40),
                                   final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "m~x"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== ub ====

set.seed(3579)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "m~x",
                                  by_x_1 = by_x_1,
                                  what = "ub",
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "m~x"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

## ==== lb ====

set.seed(123456)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "m~x",
                                  by_x_1 = by_x_1,
                                  x_interval = c(.10, .50),
                                  what = "lb",
                                  goal = "close_enough",
                                  final_nrep = 2000)
rejection_rates(a_out$by_x_1)
pba_diag(a_out)

(tmp_es <- setNames(a_out$x_out, "m~x"))
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 2000,
              iseed = 2345)
rejection_rates(tmp_out)
a_out$f_power
a_out$f_what
a_out$f_goal

})

#' @title Generate Monte Carlo Estimates
#'
#' @description Get a list of the output
#' of [lavaan::sem()] and generate
#' Monte Carlo estimates of model
#' parameters.
#'
#' @details
#' The function [gen_mc()] simply calls
#' [manymome::do_mc()]
#' on each output of
#' [lavaan::sem()] in `fit_all`. The
#' simulated
#' estimates can then be used to test
#' effects such as indirect effects,
#' usually by functions from the
#' `manymome` package, such as
#' [manymome::indirect_effect()].
#'
#' # The role of `gen_mc()`
#'
#' This function is used by the
#' all-in-one function [power4test()].
#' Users usually do not call this
#' function directly, though
#' developers can use this function to
#' customize the workflow of the
#' power analysis.
#'
#' @seealso See [power4test()] for
#' the all-in-one function that uses
#' this function.
#'
#' @param fit_all The output of
#' [fit_model()] or an object of the
#' class `fit_out`.
#'
#' @param R The number of replications
#' to generate the Monte Carlo estimates
#' for each fit output.
#'
#' @param ... Optional arguments to be
#' passed to [manymome::do_mc()] when
#' generating the Monte Carlo estimates.
#'
#' @param iseed The seed for the random
#' number generator. Default is `NULL`
#' and the seed is not changed.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to generate
#' Monte Carlo estimates for the
#' fit outputs. Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' will be displayed.
#' Default is `FALSE.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param compute_implied_stats Whether
#' implied statistics are computed
#' in each Monte Carlo replication. Usually
#' not needed and so default to `FALSE`.
#'
#' @param cl A cluster, such as one created
#' by [parallel::makeCluster()]. If `NULL`,
#' a cluster will be created, but will be
#' stopped on exit. If set to an existing
#' cluster, it will not be stopped when
#' the function exits; users need to
#' stop it manually.
#'
#' @return
#' An `mc_list` object, which is a list
#' of the output of [manymome::do_mc()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "m ~ x
#'  y ~ m + x"
#'
#' # Specify the effect sizes (population parameter values)
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate several simulated datasets
#'
#' data_all <- sim_data(nrep = 5,
#'                      model = mod,
#'                      pop_es = es,
#'                      n = 100,
#'                      iseed = 1234)
#'
#' # Fit the population model to each datasets
#'
#' fit_all <- fit_model(data_all)
#'
#' # Generate Monte Carlo estimates for each replication
#'
#' mc_all <- gen_mc(fit_all,
#'                  R = 100,
#'                  iseed = 4567)
#'
#' mc_all
#'
#' @export
#'
gen_mc <- function(fit_all,
                   R = 100,
                   ...,
                   iseed = NULL,
                   parallel = FALSE,
                   progress = FALSE,
                   ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                   compute_implied_stats = FALSE,
                   cl = NULL) {
  # This check can be removed after manymome is updated on CRAN.
  # The version requirement should be added to DESCRIPTION later.
  manymome_implied <- (utils::packageVersion("manymome") >= "0.2.7.1")
  if (isFALSE(manymome_implied)) {
    stop("manymome 0.2.7.1 or later is required. ",
         "If not available from CRAN, can be installed from ",
         "GitHub: remotes::install_github('sfcheung/manymome')")
  }
  out <- do_FUN(X = fit_all,
                FUN = gen_mc_i,
                R = R,
                ...,
                compute_implied_stats = compute_implied_stats,
                iseed = iseed,
                parallel = parallel,
                progress = progress,
                ncores = ncores,
                cl = cl)
  class(out) <- c("mc_list", class(out))
  return(out)
}

#' @noRd
gen_mc_i <- function(fit_i,
                     R = 100,
                     seed = NULL,
                     compute_implied_stats = FALSE,
                     ...) {
  if (inherits(fit_i, "lavaan")) {
    mc_out <- tryCatch(manymome::do_mc(fit = fit_i,
                                       R = R,
                                       ...,
                                       compute_implied_stats = compute_implied_stats,
                                       parallel = FALSE,
                                       progress = FALSE),
                       error = function(e) e)
  } else {
    mc_out <- tryCatch(stop("The fit is not a lavaan object, ",
                            "probably an error in model fitting."),
                       error = function(e) e)
  }
  mc_out
}

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out and do_test: k models", {

mod <-
"m ~ x
 y ~ m + x"
es <-
c("y ~ m" = "m",
  "m ~ x" = "m",
  "y ~ x" = "n")
data_all <- sim_data(nrep = 3,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)
fit_all <- fit_model(data_all)
fit_all2 <- fit_model(data_all,
                      model = c("m ~ x", "y ~ m"))
fit_all3 <- fit_model(data_all,
                      model = c("m ~ x", "y ~ 0*m"))
# mc_all <- gen_mc(fit_all,
#                  R = 100,
#                  iseed = 4567)

sim_all <- sim_out(data_all = data_all,
                   fit = fit_all,
                   fit2 = fit_all2,
                   fit3 = fit_all3)

lrt_test <- function(fit,
                     fit2) {
  lrt_out <- lavaan::lavTestLRT(fit,
                                fit2)
  lrt_out
}

lrt_results <- function(lrt_out,
                        alpha = .05) {
  pvalue <- lrt_out[2, "Pr(>Chisq)"]
  out <- c(est = lrt_out[2, "Chisq diff"],
           pvalue = pvalue,
           sig = ifelse(pvalue < alpha, 1, 0))
  out
}

test_all <- do_test(sim_all,
                    test_fun = lrt_test,
                    map_names = c(fit = "fit",
                                  fit2 = "fit2"),
                    results_fun = lrt_results,
                    parallel = FALSE,
                    progress = !is_testing())

chk <- lavTestLRT(fit_all[[2]],
                  fit_all2[[2]])
expect_equal(unname(test_all[[2]]$test_results["pvalue"]),
             chk[2, "Pr(>Chisq)"],
             ignore_attr = FALSE)

# Check update models

sim_all1 <- sim_out(data_all = data_all,
                    fit = fit_all)
sim_all1 <- sim_out(data_all = sim_all1,
                    fit2 = fit_all2)
sim_all1 <- sim_out(data_all = sim_all1,
                    fit3 = fit_all3)

expect_true(identical(sim_all,
                      sim_all1))

# Check NULL

sim_all2 <- sim_out(data_all = data_all)

expect_true(is.null(sim_all2[[1]]$extra))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("gen_mc: Error and warning catching", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = ".30",
                         "m ~ x" = ".40",
                         "y ~ x" = ".10")

data_i <- sim_data(nrep = 3,
                   model = model_simple_med,
                   pop_es = model_simple_med_es,
                   n = 50,
                   iseed = 1234,
                   parallel = FALSE,
                   progress = !is_testing())

# Add an error in fit
data_i[[2]]$model_final <- "y1 ~ x1"

fit_out <- fit_model(data_i,
                     parallel = FALSE,
                     progress = !is_testing())

mc_out <- gen_mc(fit_out,
                 R = 100,
                 seed = 1234)
expect_s3_class(mc_out[[2]],
                "error")

})

fix_many_lm_model <- function(model) {
  # Try to use lavaan
  ptable <- lavaan::lavaanify(model)
  ptable <- ptable[ptable$op == "~", ]
  all_y <- lavaan::lavNames(ptable, "eqs.y")
  all_models <- lapply(
      all_y,
      function(y) {
        pt1 <- ptable[ptable$lhs == y, ]
        all_x_i <- pt1$rhs
        model_i <- paste(
                    y,
                    "~",
                    paste(all_x_i,
                          collapse = " + ")
                  )
      })
  all_models <- unlist(all_models)
  all_models <- paste0(
                  all_models,
                  collapse = "\n"
                )
  all_models <- paste0(all_models, "\n")
  all_models
}

#' @noRd
# Does not work for model specified in more than one line.
fix_many_lm_model_old <- function(model) {
  # Remove lines that cannot be fitted by lm()
  tmp <- strsplit(
            model,
            split = "\n",
            fixed = TRUE)
  tmp <- unlist(tmp)
  tmp2 <- sapply(
            tmp,
            function(x) {
              isFALSE(inherits(
                tryCatch(stats::terms(stats::as.formula(x)),
                        silent = TRUE,
                        error = function(e) e),
                "error"))
            })
  out <- tmp[tmp2]
  out <- paste(out,
               collapse = "\n")
  out
}

# Adapted from manymome

#' @title Asymmetric Bootstrap p-Value
#'
#' @description Compute the asymmetric
#' bootstrap *p*-value from a vector of
#' bootstrap estimates.
#'
#' @details It computes the *p*-value
#' based on the method presented in
#' Asparouhov and Muthén (2021).
#'
#' @return
#' Numeric. The *p*-value. `NA` if
#' it cannot be computed.
#'
#' @param x A numeric vector. The
#' bootstrap estimates.
#'
#' @param h0 The value for the null
#' hypothesis. Default is zero.
#'
#' @param min_size Integer. The
#' bootstrap *p*-value will be computed
#' only if `x` has at least `min_size`
#' valid values.
#'
#' @references
#' Asparouhov, A., & Muthén, B. (2021). Bootstrap p-value computation.
#' Retrieved from https://www.statmodel.com/download/FAQ-Bootstrap%20-%20Pvalue.pdf
#'
#' @examples
#' x1 <- rnorm(n, 2, 4)
#' est2p(x1)
#'
#' @noRd

est2p <- function(x,
                  h0 = 0,
                  min_size = 100,
                  warn = FALSE) {
    # Based on the method in
    # https://www.statmodel.com/download/FAQ-Bootstrap%20-%20Pvalue.pdf
    x <- x[!is.na(x)]
    if (length(x) == 0) return(NA)
    if (length(x) < min_size) {
        if (warn) {
          warning(paste("Bootstrap p-value not computed. Less than ",
                        min_size,
                        "bootstrap estimates."))
        }
        return(NA)
      }
    b <- length(x)
    m0 <- sum((x < h0))
    out <- 2 * min(m0 / b, 1 - m0 / b)
    out
  }

#' @title Power Curve
#'
#' @description Estimate the relation
#' between power and a characteristic,
#' such as sample size or population
#' effect size (population value of
#' a model parameter).
#'
#' @details
#' The function [power_curve()]
#' retrieves the information
#' from the output of
#' [power4test_by_n()] or
#' [power4test_by_es()], and
#' estimate the power curve: the
#' relation between the characteristic
#' varied, sample size for
#' [power4test_by_n()] and the
#' population effect size for
#' [power4test_by_es()], and the
#' rejection rate of the test conducted
#' by [power4test_by_n()] or
#' [power4test_by_es()]. This
#' rejection rate is the power when the
#' null hypothesis is false (e.g., the
#' population value of the effect size
#' being tested is nonzero).
#'
#' The model fitted is *not* intended to
#' be a precise model for the relation
#' across a wide range. It is only a
#' crude estimate based on the limited
#' number of values of the
#' characteristic (e.g., sample size)
#' examined, which can be as small as
#' four or even smaller. The model is
#' intended to be
#' used for only for the range covered,
#' and for estimating the probable
#' sample size or effect size with a
#' desirable level of power. This value
#' should then be studied by higher
#' precision through simulation
#' using functions such as
#' [power4test()].
#'
#' These are the models to be tried,
#' in the following order:
#'
#' - One or nonlinear models, to be
#'   fitted by [stats::nls()]. If
#'   several models are specified, all
#'   will be fitted and the one with the
#'   smallest deviance will be used.
#'
#' - If all the nonlinear models failed,
#'   for whatever reason, a logistic
#'   regression will be fitted by
#'   [stats::glm()] to predict the
#'   binary significant test results.
#'
#' - If the logistic model also failed,
#'   for whatever reason, a simple
#'   linear regression model will be
#'   fitted. Although the power curve is
#'   nonlinear across a wide range of,
#'   say, sample size, a linear model
#'   can still be a good enough
#'   approximation for a narrow range of
#'   the predictor.
#'
#' The output can then be plotted to
#' visualize the power curve, using
#' the `plot` method ([plot.power_curve()])
#' for the output
#' of [power_curve()].
#'
#' This function can be used directly,
#' but is also used internally by
#' functions such as [x_from_power()].
#'
#' ## 'nls_options'
#'
#' These are possible arguments.
#'
#' - `min_points`: If the object has
#'   less than `min_points` rejection
#'   rates, `nls` will not be used.
#'
#' - `regions`: A numeric vector with
#'   values from 0 to 1. It defines
#'   regions in which there must be
#'   at least one rejection rate. If this
#'   criterion is not met, `nls` will
#'   not be used. This is to ensure that
#'   `nls` will not be used when the
#'   rejection rates are clustered around
#'   a very narrow region.
#'
#'
#' @return
#' It returns a list which is a
#' `power_curve` object, with the
#' following elements:
#'
#' - `fit`: The model fitted, which is the
#'    output of [stats::nls()],
#'    [stats::glm()], or [stats::lm()].
#'
#' - `reject_df`: The table of reject
#'  rates and other characteristics,
#'  which is generated by
#'  [rejection_rates()].
#'
#'  - `predictor`: The predictor or the
#'  power curve, ether `"n"` (sample
#'  size) or `"es"` (population effect
#'  size).
#'
#'  - `call`: The call used to run this
#'  function.
#'
#' @param object An object of the class
#' `power4test_by_n` or `power4test_by_es`,
#' which is the output of [power4test_by_n()]
#' or [power4test_by_es()].
#'
#' @param formula A formula of the model
#' for [stats::nls()]. It can also be
#' a list of formulas, and the models
#' will be fitted successively by
#' [stats::nls()], with the first
#' model fitted successfully adopted.
#' The response variable in the formula must be named
#' `reject`, and the predictor named
#' `x`. Whether `x` represents `n` or
#' `es` depends on the class of `object`.
#' If `NULL`, the default, it will be
#' determined internally based on the type of
#' `object`.
#'
#' @param start Either a named vector
#' of the start value(s) of parameter(s)
#' in `formula`, or a list of named
#' vectors of the starting value(s)
#' of the list of formula(s). If `NULL`,
#' the default, they will be determined
#' internally.
#'
#' @param lower_bound Either a named vector
#' of the lower bound(s) of parameter(s)
#' in `formula`, or a list of named
#' vectors of the lower bound(s)
#' for the list of formula(s). They will
#' be passed to `lower` of [stats::nls()].
#' If `NULL`, the default, it will be
#' determined internally based on the type of
#' `object`.
#'
#' @param upper_bound Either a named vector
#' of the upper bound(s) of parameter(s)
#' in `formula`, or a list of named
#' vectors of the upper bound(s)
#' for the list of formula(s). They will
#' be passed to `upper` of [stats::nls()].
#' If `NULL`, the default, it will be
#' determined internally based on the type of
#' `object`.
#'
#' @param nls_args A named list of
#' arguments to be used when calling
#' [stats::nls()]. Used to override
#' internal default, such as the
#' algorithm (default is `"port"`).
#' Use this argument with cautions.
#'
#' @param nls_control A named list of
#' arguments to be passed the `control`
#' argument of [stats::nls()]. The values will
#' override internal default values,
#' and also override `nls_args`.
#' Use this argument with cautions.
#'
#' @param verbose Logical. Whether
#' messages will be printed when
#' trying different models.
#'
#' @param models Models to try. Support
#' `"nls"` (fitted by [nls()]),
#' `"logistic"` (fitted by [glm()]), and
#' `"lm"` (fitted by [lm()]). By default,
#' all three models will be attempted,
#' in this order.
#'
#' @param nls_options A named list of
#' options to be used by [power_curve()]
#' to configure the use of `"nls"`. For
#' advanced use. See 'Details' for
#' available options.
#'
#' @seealso [power4test_by_n()] and [power4test_by_es()]
#' for the output supported by
#' [power_curve()], [plot.power_curve()]
#' for the `plot` method and
#' [predict.power_curve()]
#' for the `predict` method of the output
#' of [power_curve()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the effect sizes (population parameter values)
#'
#' model_simple_med_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: s
#' "
#'
#' # Simulate datasets to check the model
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' sim_only <- power4test(nrep = 10,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 50,
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234,
#'                        parallel = FALSE,
#'                        progress = FALSE)
#'
#' # By n: Do a test for different sample sizes
#'
#' out1 <- power4test_by_n(sim_only,
#'                         nrep = 10,
#'                         test_fun = test_parameters,
#'                         test_args = list(par = "y~x"),
#'                         n = c(25, 50, 100),
#'                         by_seed = 1234,
#'                         parallel = FALSE,
#'                         progress = FALSE)
#'
#' pout1 <- power_curve(out1)
#' pout1
#' plot(pout1)
#'
#' # By pop_es: Do a test for different population values of a model parameter
#'
#' out2 <- power4test_by_es(sim_only,
#'                          nrep = 10,
#'                          test_fun = test_parameters,
#'                          test_args = list(par = "y~x"),
#'                          pop_es_name = "y ~ x",
#'                          pop_es_values = c(0, .3, .5),
#'                          by_seed = 1234,
#'                          parallel = FALSE,
#'                          progress = FALSE)
#'
#' pout2 <- power_curve(out2)
#' pout2
#' plot(pout2)
#'
#' @export
power_curve <- function(object,
                        formula = NULL,
                        start = NULL,
                        lower_bound = NULL,
                        upper_bound = NULL,
                        nls_args = list(),
                        nls_control = list(),
                        verbose = FALSE,
                        models = c("nls", "logistic", "lm"),
                        nls_options = list(min_points = 4,
                                           regions = c(0, .45, .75, .85, 1))) {
  if (missing(models)) {
    models <- eval(formals(power_curve)$models)
  } else {
    models <- match.arg(models,
                        several.ok = TRUE)
  }

  # reject ~ I((x - c0)^e) / (b + I((x - c0)^e))
  # The formula used depends on the nature of the predictors
  # The predictor should always be named 'x' in the formula.
  # The outcome is always 'reject'
  # `formula` can be a list of possible NLS models.

  class0 <- class(object)[1]
  if (!(class0 %in% c("power4test_by_n",
                      "power4test_by_es"))) {
    stop("'object' is neither a power4test_by_n or power4test_by_n object.")
  }

  if (class0 == "power4test_by_n") {
    if (is.null(formula)) {
      formula <- list(
                    reject ~ 1 - I(exp((a - x) / b)),
                    reject ~ I((x - c0)^e) / (b + I((x - c0)^e))
                  )
    }
    if (is.null(start)) {
      start <- list(
                  c(a = 0, b = 10),
                  c(b = 2, c0 = 100, e = 1)
                )
    }
    if (is.null(lower_bound)) {
      lower_bound <- list(
                        c(a = 0, b = .05),
                        c(b = 0, c0 = 0, e = 1)
                      )
    }
    if (is.null(upper_bound)) {
      upper_bound <- list(
                        c(a = Inf, b = Inf),
                        c(b = Inf, c0 = Inf, e = Inf)
                      )
    }
  }

  if (class0 == "power4test_by_es") {
    if (is.null(formula)) {
      formula <- list(reject ~ 1 - I(exp((a - x) / b)),
                      reject ~ 1 - 1 / I((1 + (x / d)^a)^b),
                      reject ~ 1 - exp(x / a) / I((1  + exp(x / a))^b),
                      reject ~ 1 - 2 / (exp(x / d) + exp(-x / d)),
                      reject ~ 1 / (1 + a * exp(-b * x)))
    }
    if (is.null(start)) {
      start <- list(c(a = 0, b = 10),
                    c(a = 2, b = 4, d = 4),
                    c(a = 1, b = 2),
                    c(d = 1),
                    c(a = 1, b = 1))
    }
    if (is.null(lower_bound)) {
      lower_bound <- list(NULL, NULL, NULL, NULL, NULL)
    }
    if (is.null(upper_bound)) {
      upper_bound <- list(NULL, NULL, NULL, NULL, NULL)
    }
  }

  # No need for other arguments. They should be stored in object.
  reject0 <- rejection_rates(object,
                             all_columns = TRUE)

  predictor <- switch(class0,
                      power4test_by_n = "n",
                      power4test_by_es = "es")

  reject0$x <- reject0[[predictor]]

  model_found <- FALSE

  fit <- NA

  nls_ok <- 1L
  if (length(reject0$reject) < nls_options$min_points) {
    nls_ok <- nls_ok - 1L
  }
  if (!isTRUE(check_regions(
                    reject0$reject,
                    nls_options$regions))) {
    nls_ok <- nls_ok - 1L
  }

  if ((nls_ok == 1L) &&
      ("nls" %in% models)) {

    # === nls ===
    nls_args_fixed <- fix_nls_args(formula = formula,
                                   start = start,
                                   lower_bound = lower_bound,
                                   upper_bound = upper_bound,
                                   nls_args = nls_args,
                                   nls_control = nls_control)
    # Try each formula
    fit_deviance <- Inf
    for (i in seq_along(nls_args_fixed$formula)) {
      nls_args1 <- utils::modifyList(nls_args_fixed$nls_args,
                                     list(data = reject0,
                                          control = nls_args_fixed$nls_contorl1,
                                          nrep = reject0$nrep))
      nls_args_fixed_i <- list(formula = nls_args_fixed$formula[[i]],
                               start = nls_args_fixed$start[[i]],
                               lower = nls_args_fixed$lower_bound[[i]],
                               upper = nls_args_fixed$upper_bound[[i]])
      for (xx in names(nls_args_fixed_i)) {
        if (is.null(nls_args_fixed_i[[xx]])) {
          nls_args_fixed_i[[xx]] <- NULL
        }
      }
      nls_args1 <- utils::modifyList(nls_args1,
                                     nls_args_fixed_i)
      # Do nls0
      fit_i <- tryCatch(suppressWarnings(do.call(do_nls,
                                                 nls_args1)),
                        error = function(e) e)
      if (inherits(fit_i, "nls")) {
        fit_i_d <- stats::deviance(fit_i)
        if (fit_i_d < fit_deviance) {
          fit <- fit_i
          fit_deviance <- fit_i_d
        }
      }
    }

    if (inherits(fit, "nls")) {
      model_found <- TRUE
    } else {
      if (verbose) {
        cat("- 'nls()' estimation failed. Switch to logistic regression.\n")
      }
    }

  } else if ("nls" %in% models) {
    if (verbose) {
      cat("- 'nls()' estimation skipped when less than 4 values of predictor examined.\n")
    }
  }

  # === Logistic ===

  if (!model_found &&
      ("logistic" %in% models)) {
    # Do logistic
    fit <- do_logistic(reject_df = reject0)

    if (inherits(fit, "glm")) {
      model_found <- TRUE
    } else {
      if (verbose) {
        cat("- Logistic regression failed. Switch to linear regression.\n")
      }
    }
  }

  # === OLS Regression ===

  if (!model_found &&
      ("lm" %in% models)) {
  # Last resort: OLS regression
    fit <- do_lm(reject_df = reject0,
                weights = reject0$nrep)

    if (inherits(fit, "lm")) {
      model_found <- TRUE
    } else {
      if (verbose) {
        cat("- OLS regression failed. No power curve estimated.\n")
      }
    }
  }

  # TODO:
  # - Consider using `splinefun()` as a last resort.

  out <- list(fit = fit,
              reject_df = reject0,
              predictor = predictor,
              call = match.call())
  class(out) <- c("power_curve", class(out))
  return(out)
}

#' @rdname power_curve
#'
#' @param x A `power_curve` object.
#'
#' @param data_used Logical. Whether
#' the rejection rates data frame
#' used to fit the model is printed.
#'
#' @param digits,right,row.names
#' Arguments of the same names used
#' by the `print` method of a
#' `data.frame` object. Used when `data_used`
#' is `TRUE` and the rejection rates
#' data frame is printed.
#'
#' @param ... For the `print` method of
#' `power_curve` objects, they are optional
#' arguments to be passed to
#' [print.data.frame()] when printing
#' the rejection rates data frame.
#'
#' @return
#' The `print` method of `power_curve`
#' object returns `x` invisibly. Called
#' for its side-effect.
#'
#' @export
print.power_curve <- function(x,
                              data_used = FALSE,
                              digits = 3,
                              right = FALSE,
                              row.names = FALSE,
                              ...) {
  cat("Call:\n")
  my_call <- x$call
  if (!is.symbol(my_call$object)) {
    my_call$object <- as.symbol("<hidden>")
  }
  if (!is.symbol(my_call[[1]])) {
    my_call[[1]] <- as.symbol("power_curve")
  }
  print(my_call)
  cat("\nPredictor: ",
      x$predictor,
      " (",
      switch(x$predictor,
             n = "Sample Size",
             es = "Effect Size"),
      ")\n",
      sep = "")
  cat("\nModel:\n")
  tmp <- x$fit
  tmp$data <- "(Omitted)"
  print(tmp)
  if (data_used) {
    cat("\nData Used:\n")
    print(x$reject_df,
          digits = digits,
          right = right,
          row.names = row.names,
          ...)
  }
  invisible(x)
}

#' @noRd
do_nls <- function(...,
                   nrep = NULL) {
  args <- list(...)
  # Try weights
  if (!is.null(nrep)) {
    args1 <- utils::modifyList(args,
                              list(weights = nrep))
    fit <- tryCatch(suppressWarnings(do.call(stats::nls,
                                     args1)),
                   error = function(e) e)
    if (inherits(fit, "nls")) {
      return(fit)
    }
  }

  # Do not use weights
  fit <- tryCatch(suppressWarnings(do.call(stats::nls,
                                   args)),
                  error = function(e) e)
  if (inherits(fit, "nls")) {
    return(fit)
  }
  # fit is an error. Return it
  return(fit)
}

#' @noRd
do_logistic <- function(reject_df) {
  # The predictor is always 'x'.

  # Expand to one row per replication
  reject1 <- reject_df[, c("x", "reject", "nrep")]
  reject1$sig <- round(reject1$reject * reject1$nrep)
  reject1$ns <- reject1$nrep - reject1$sig
  tmp <- mapply(function(x, y) {
                  c(rep(1, x), rep(0, y - x))
                },
                x = reject1$sig,
                y = reject1$nrep,
                SIMPLIFY = FALSE)
  tmp <- unlist(tmp)
  reject1 <- data.frame(x = rep(reject1$x,
                                times = reject1$nrep),
                        sig = tmp)
  # Rename sig to reject,
  # to be consistent with other mdoels
  reject1$reject <- reject1$sig

  # Do logistic regression
  fit <- tryCatch(stats::glm(reject ~ x,
                             data = reject1,
                             family = "binomial"),
                  error = function(e) e,
                  warning = function(w) w)
  # Also catch warning such as
  # - "fitted probabilities numerically 0 or 1 occurred>"

  # Always return the fit
  # Let the calling function to handle error
  return(fit)
}

#' @noRd
do_lm <- function(reject_df,
                  weights) {

  # Use weights
  fit <- tryCatch(stats::lm(reject ~ x,
                            data = reject_df,
                            weights = weights),
                  error = function(e) e)
  if (inherits(fit, "lm")) {
    return(fit)
  }
  # Do not use weights
  fit <- tryCatch(stats::lm(reject ~ x,
                            data = reject_df),
                  error = function(e) e)

  # Return fit, error or not
  return(fit)
}

#' @noRd

fix_nls_args <- function(formula,
                         start,
                         lower_bound,
                         upper_bound,
                         nls_args,
                         nls_control) {

  # Set up the per-formula arguments
  # - If not a list, coerce to a list:
  #   one element for each formula.

  if (!is.list(formula)) {
    formula <- list(formula)
  }
  if (!is.list(start)) {
    start <- list(start)
  }
  if (!is.list(lower_bound)) {
    lower_bound <- list(lower_bound)
  }
  if (!is.list(upper_bound)) {
    upper_bound <- list(upper_bound)
  }
  # Default argument values
  nls_contorl0 <- list(maxiter = 1000)
  nls_contorl1 <- utils::modifyList(nls_contorl0,
                                    nls_control)

  # Use "port" because zero residual cases are possible
  # But can be overridden
  nls_args0 <- list(algorithm = "port")
  nls_args1 <- utils::modifyList(nls_args0,
                                 nls_args)

  out <- list(formula = formula,
              start = start,
              lower_bound = lower_bound,
              upper_bound = upper_bound,
              nls_args = nls_args1,
              nls_control = nls_control)

  return(out)

}

#' @noRd
check_regions <- function(
                   x,
                   regions) {
  # Determine whether all regions have
  # at least one point.
  regions <- unique(c(regions, 0, 1))
  regions <- regions[(regions <= 1) & (regions >= 0)]
  regions <- sort(regions)
  k <- length(regions) - 1
  tmp <- findInterval(
            x = x,
            vec = regions,
            rightmost.closed = TRUE)
  setequal(unique(tmp), seq_len(k))
}

skip("Internal: Only valid under a specific configuration")

library(testthat)

test_that("x_from_power: escape a loop", {

options(power4mome.bz = TRUE)

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: l
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 400,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 199,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 2345,
                  parallel = TRUE)

n_power_region <- n_region_from_power(out,
                                      seed = 1357)
n_power_region
plot(n_power_region)

tmp <- n_power_region$above$power4test_trials
class(tmp)
tmp2 <- tmp[1:7]
class(tmp2) <- class(tmp)

tmp3 <- power_curve(
          tmp2,
          models = "nls",
          verbose = TRUE
        )
plot(tmp3)

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: n", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 500,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
out_power <- rejection_rates(out)
out_power

tmp <- x_from_power(out,
                    x = "n",
                    progress = TRUE,
                    final_nrep = 100,
                    seed = 1234,
                    save_sim_all = TRUE,
                    algorithm = "power_curve")
tmp
summary(tmp)
plot(tmp)
tmp_pc <- power_curve(tmp$power4test_trials)
tmp_pc
plot(tmp_pc)
plot(tmp,
     lwd = 4,
     col = "green",
     what = c("ci",
              "final_x",
              "final_power"))
plot(tmp,
     lwd = 4,
     col = "darkgreen",
     what = c("ci",
              "final_x",
              "final_power"),
     pars_ci_final_x = list(lwd = 10),
     pars_final_x = list(lwd = 2, col = "red"))

plot(tmp,
     pars_power_curve = list(col = "blue", lwd = 4, type = "o"))

## Close Enough

tmp <- x_from_power(out,
                    x = "n",
                    progress = TRUE,
                    final_nrep = 100,
                    seed = 1234,
                    save_sim_all = TRUE,
                    goal = "close_enough",
                    what = "point",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE))
summary(tmp)
plot(tmp)

tmpb <- x_from_power(out,
                    x = "n",
                    progress = TRUE,
                    final_nrep = 100,
                    seed = 1234,
                    save_sim_all = TRUE,
                    goal = "close_enough",
                    what = "point",
                    algorithm = "bisection")
summary(tmpb)
plot(tmpb)

# Case 2

out2 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 1000,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = "m",
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

tmp2 <- x_from_power(out2,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     algorithm = "power_curve")
tmp2
summary(tmp2)
plot(tmp2)

# Case 3

mod <-
"
m1 ~ x
m2 ~ m1 + x
m3 ~ m2 + m1 + x
y ~ m3 + m2 + m1 + x
"

mod_es <-
"
m1 ~ x: s
m2 ~ m1: s
m3 ~ m2: s
y ~ m3: s
y ~ x: l
"

out3 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 200,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = c("m1", "m2", "m3"),
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

rejection_rates(out3)

tmp3 <- x_from_power(out3,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     algorithm = "power_curve")

tmp3
summary(tmp3)
plot(tmp3)



})

skip_on_cran()

library(testthat)

test_that("pwoer_curve", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = ".2")

sim_only <- power4test(nrep = 50,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = !is_testing())

# By n

out1 <- power4test_by_n(sim_only,
                        nrep = 50,
                        test_fun = test_parameters,
                        test_args = list(par = "y~x"),
                        n = c(25, 100, 200, 250, 500),
                        by_seed = 1234,
                        parallel = FALSE,
                        progress = !is_testing())

rejection_rates(out1)

expect_no_error(pout1 <- power_curve(out1,
                                     verbose = FALSE))
expect_no_error(capture.output(print(pout1)))

expect_no_error(capture.output(print(pout1, data_used = TRUE)))

# expect_no_error(plot(pout1))

# By es

out2 <- power4test_by_es(sim_only,
                             nrep = 50,
                             test_fun = test_parameters,
                             test_args = list(par = "y~x"),
                             pop_es_name = "y ~ x",
                             pop_es_values = seq(0, .7, .15),
                             by_seed = 1234,
                             parallel = FALSE,
                             progress = !is_testing())

rejection_rates(out2,
                all_columns = TRUE)

expect_no_error(pout2 <- power_curve(out2,
                                     verbose = FALSE))
expect_no_error(capture.output(print(out2)))

expect_no_error(capture.output(print(pout2, data_used = TRUE)))

# expect_no_error(plot(pout2))

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: es", {

# Case 1

mod <-
"
m1 ~ x
m2 ~ m1 + x
m3 ~ m2 + m1 + x
y ~ m3 + m2 + m1 + x
"

mod_es <-
"
.ind.(x->m1->m2->m3->y): .05
y ~ x: nil
"

out3 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 50,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = c("m1", "m2", "m3"),
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

print(out3,
      data_long = TRUE)
rejection_rates(out3, all_columns = TRUE)

tmp3 <- x_from_power(out3,
                     x = "es",
                     pop_es_name = ".ind.(x->m1->m2->m3->y)",
                     progress = TRUE,
                     final_nrep = 80,
                     max_trials = 4,
                     x_interval = c(0, .90),
                     seed = 1234,
                     algorithm = "power_curve")

tmp3
summary(tmp3)
plot(tmp3)
tmp3$power_curve
power_curve(tmp3$power4test_trials)

})

set_es_range_by_x <- function(
                          object,
                          pop_es_name,
                          target_power = .80,
                          k = 4,
                          es_max = .7,
                          es_min = 0,
                          object_by_org = NULL,
                          what = NULL,
                          goal = NULL,
                          tol = NULL,
                          ci_level = .95) {
  # Other arguments will be retrieved from object_by_org
  reject0by <- rejection_rates(object_by_org,
                               level = ci_level,
                               add_se = TRUE,
                               all_columns = TRUE)
  # Always return a value because closest_ok is TRUE
  i0 <- find_solution(
                object_by_org,
                target_power = target_power,
                ci_level = ci_level,
                what = what,
                tol = tol,
                goal = goal,
                final_nrep = 0,
                closest_ok = TRUE,
                weight_by = "nrep",
                debug = TRUE
              )
  es0 <- reject0by$es[i0]
  what0 <- switch(
              what,
              point = "reject",
              lb = "reject_ci_lo",
              ub = "reject_ci_hi"
            )
  est0 <- reject0by[i0, what0, drop = TRUE]
  # Displace est0 a little bit
  if (est0 == target_power) {
    est0 <- target_power * .95
  }
  if (est0 < target_power) {
    i1a <- reject0by[, what0, drop = TRUE] > target_power
    i1b <- reject0by$es > es0
    i1 <- i1a & i1b
  } else {
    i1a <- reject0by[, what0, drop = TRUE] < target_power
    i1b <- reject0by$es < es0
    i1 <- i1a & i1b
  }
  if (any(i1)) {
    i1 <- which(i1)[1]
    es1 <- reject0by$es[i1]
  } else {
    es1 <- es0 * target_power / est0
  }
  es0 <- min(es0, es_max)
  es1 <- min(es1, es_max)
  es_out <- range(es0, es1)
  es_out
}

set_es_range <- function(object,
                         pop_es_name,
                         target_power = .80,
                         k = 4,
                         es_max = .7,
                         es_min = 0) {
  # TODO:
  # - Add support for multigroup models.
  es0 <- pop_es(object,
                pop_es_name = pop_es_name)
  es0_sign <- sign(es0)
  es0_abs <- abs(es0)
  # No need for other arguments because only `reject` is used
  reject0 <- rejection_rates(object)
  power0 <- reject0$reject[1]
  if (es0 > es_max) {
    # Use x_max because es_max and es_min are internal arguments
    stop("Initial population value (",
          es0,
          ") is equal to or greater than 'x_max' (",
          es_max,
          "). Please increase 'x_max'.")
  }
  if (es0 < es_min) {
    # Use x_max because es_max and es_min are internal arguments
    stop("Initial population value (",
          es0,
          ") is equal to or less than 'x_min' (",
          es_min,
          "). Please decrease 'x_min'.")
  }

  if (power0 == target_power) {
    # If power0 == target_power,
    # Be conservative and decrease power by a small amount
    power0 <- target_power * .99
  }

  b <- power0 / es0_abs
  es_bound <- ifelse(es0_sign >= 0,
                      abs(es_max),
                      abs(es_min))
  es_end <- min(target_power / b,
                es_bound)
  es_out <- seq(from = es0_abs,
                to = es_end,
                length.out = k)
  if (es0_sign >= 0) {
    return(es_out)
  } else {
    es_out <- sort(-es_out)
    return(es_out)
  }
}

#' @noRd
pop_es <- function(object,
                   pop_es_name) {
  # TODO:
  # - Add support for multigroup models

  # Support objects generated by power4test_by_es()
  tmp1 <- attr(object, "pop_es_name")
  if (identical(tmp1, pop_es_name)) {
    tmp2 <- attr(object, "pop_es_value")
    if (!is.null(tmp2)) {
      return(tmp2)
    }
  }

  is_ind <- FALSE
  pop_es_name_comp <- tryCatch(expand_to_components(pop_es_name),
                               error = function(e) e)
  if (!inherits(pop_es_name_comp, "error")) {
    if (length(pop_es_name_comp) > 1) {
      is_ind <- TRUE
    }
  }
  ptable <- object$sim_all[[1]]$ptable
  ngroups <- max(ptable$group)
  if (ngroups != 1) {
    stop("Do not support multigroup models for now.")
  }
  ptable$lavlabel <- lavaan::lav_partable_labels(ptable)
  ptable$lavlabel2 <- paste0(ptable$lhs,
                             ptable$op,
                             ptable$rhs)
  if (!is_ind) {
    pop_es_name0 <- gsub(" ", "", pop_es_name)
    i0 <- match(pop_es_name0, ptable$lavlabel)
    i1 <- match(pop_es_name0, ptable$lavlabel2)
    i <- c(i0, i1)
    i <- unique(i)
    i <- i[!is.na(i)]
    if (length(i) == 0) {
      stop("'pop_es_name' not a valid name of a parameter in the model.")
    }
    out <- ptable$start[i]
  } else {
    pop_es_name_comp <- gsub(" ", "",
                             pop_es_name_comp,
                             fixed = TRUE)
    i0 <- match(pop_es_name_comp, ptable$lavlabel)
    i1 <- match(pop_es_name_comp, ptable$lavlabel2)
    i <- c(i0, i1)
    i <- unique(i)
    i <- i[!is.na(i)]
    if (length(i) == 0) {
      stop("'pop_es_name' is not a valid path in the model.")
    }
    out <- prod(ptable$start[i])
  }
  out
}

#' @noRd
estimate_es <- function(power_es_fit,
                        target_power = .80,
                        interval = c(0, .70),
                        extendInt = "no") {
  f <- function(es) {
    stats::predict(power_es_fit,
                   newdata = list(x = es)) - target_power
  }
  es_target <- tryCatch(stats::uniroot(f,
                             interval = interval,
                             extendInt = extendInt),
                       error = function(e) e)
  if (inherits(es_target, "error")) {
    # Return NA if error occurred. E.g.,
    # - Root not in the interval.
    return(NA)
  }
  es_target <- es_target$root
  return(es_target)
}

#' @noRd
estimate_es_range <- function(power_es_fit,
                              target_power = .80,
                              k = 5,
                              tolerance = .20,
                              power_min = .01,
                              power_max = .99,
                              interval = c(0, .70),
                              extendInt = "upX",
                              es_to_exclude = NULL) {
  # target power is guaranteed to be in the choice
  if (((k %% 2) == 0) && (k >= 4)) {
    # If k is even and k >= 4, be conservative
    a <- seq(from = max(target_power - tolerance, power_min),
             to = target_power,
             length.out = k / 2)
    b <- seq(from = target_power,
             to = min(target_power - tolerance, power_max),
             length.out = k / 2 + 1)[-1]
    power_j <- c(a, b)
  } else if (k == 2) {
    power_j <- c(target_power,
                 min(target_power - tolerance, power_max))
  } else if (((k %% 2) == 0) && (k >= 3)) {
    # k is odd and k >= 3
    power_j <- seq(from = max(target_power - tolerance, power_min),
                   to = min(target_power + tolerance, power_max),
                   length.out = k)
  } else if (k == 1) {
    # k = 1
    power_j <- target_power
  } else {
    # Just in case
    power_j <- seq(from = max(target_power - tolerance, power_min),
                   to = min(target_power + tolerance, power_max),
                   length.out = k)
  }

  # # No longer needed
  # if (isFALSE(target_power %in% power_j)) {
  #   power_j <- sort(c(power_j, target_power))
  # }
  out <- sapply(power_j,
                function(x) {
                  estimate_es(power_es_fit = power_es_fit,
                              target_power = x,
                              interval = interval,
                              extendInt = extendInt)
                })

  # If NA, have to do random sampling
  i <- is.na(out)
  if (any(i)) {
    # Duplication is OK because it will be fixed later
    out[i] <- stats::runif(n = sum(i),
                           min = interval[1],
                           max = interval[2])
  }

  # Check invalid es values

  i <- check_es(ess = out,
                interval = interval,
                es_to_exclude = es_to_exclude,
                extendInt = extendInt)

  if (isFALSE(any(i))) {
    return(out)
  }
  # Replace invalid ESs by random ESs
  # Do not use the full interval
  # But can include ESs already considered
  new_interval1 <- min(es_to_exclude, interval)
  new_interval2 <- max(max(es_to_exclude),
                       interval[2])
  interval_width <- (new_interval2 - new_interval1)
  es_pool <- setdiff(seq(new_interval1,
                         new_interval2,
                         length = 500),
                     c(es_to_exclude, out[!i]))
  for (q in 1:10) {
    out[i] <- out[i] + stats::runif(n = sum(i),
                                    min = -interval_width / 100,
                                    max = interval_width / 100)
    i <- check_es(out,
                  interval = c(new_interval1, new_interval2),
                  es_to_exclude = es_to_exclude,
                  extendInt = extendInt)
    if (isFALSE(any(i))) {
      # All ns OK
      break
    }
  }
  if (any(i)) {
    # Have to do random sample
    es_pool <- setdiff(seq(new_interval1,
                           new_interval2,
                           length = 500),
                       c(es_to_exclude, out[!i]))
    es_new <- sample(es_pool, size = sum(i))
    out[i] <- es_new
  }
  return(out)
}

#' @noRd
check_es <- function(ess,
                     interval,
                     es_to_exclude,
                     extendInt) {
  i <- rep(FALSE, length(ess))
  if (isFALSE(extendInt %in% c("yes", "upX"))) {
    i[ess > interval[2]] <- TRUE
  }
  if (isFALSE(extendInt %in% c("yes", "downX"))) {
    i[ess < interval[1]] <- TRUE
  }

  # ESs used are considered invalid
  i[ess %in% es_to_exclude] <- TRUE

  # Duplicated are considered invalid
  i[duplicated(ess)] <- TRUE

  i[is.na(ess)] <- TRUE

  i
}

#' @noRd

reject_ci <- function(
      nreject,
      nvalid,
      level = .95,
      method = c("wilson", "norm")
    ) {
  method <- match.arg(method)
  out <- switch(method,
          norm = reject_ci_norm(
                    nreject = nreject,
                    nvalid = nvalid,
                    level = level
                  ),
          wilson = reject_ci_wilson(
                    nreject = nreject,
                    nvalid = nvalid,
                    level = level
                  ))
  out
}

#' @noRd
reject_ci_norm <- function(
      nreject,
      nvalid,
      level
    ) {
  reject <- nreject / nvalid
  reject_se <- sqrt(reject * (1 - reject) / nvalid)
  a <- abs(stats::qnorm(1 - (1 - level) / 2))
  reject_ci_lo <- reject - a * reject_se
  reject_ci_hi <- reject + a * reject_se
  cbind(cilo = reject_ci_lo,
        cihi = reject_ci_hi)
}

#' @noRd
reject_ci_wilson <- function(
      nreject,
      nvalid,
      level
    ) {
  z0 <- abs(stats::qnorm((1 - level) / 2, lower.tail = FALSE))
  a1 <- (nreject + z0^2 / 2) / (nvalid + z0^2)
  a2 <- (z0 / (nvalid + z0^2)) * sqrt(nreject * (nvalid - nreject) / nvalid + z0^2 / 4)
  reject_ci_lo <- a1 - a2
  reject_ci_hi <- a1 + a2
  cbind(cilo = reject_ci_lo,
        cihi = reject_ci_hi)
}

library(testthat)

test_that("Check solution in objects", {

p_ci <- function(x,
                 nrep,
                 ci_level) {
  a <- abs(stats::qnorm((1 - ci_level) / 2))
  se <- sqrt(x * (1 - x) / nrep)
  cilb <- x - a * se
  ciub <- x + a * se
  c(cilb, ciub)
}

# Use Wilson CI
p_ci <- function(x,
                 nrep,
                 ci_level = .95) {
  reject_ci_wilson(round(x * nrep),
                   nrep,
                   ci_level)
}


reject_out <- data.frame(nrep = c(60, 60, 50),
                         reject = c(.60, .70, .80))
reject_out[, c("cilo", "cihi")] <- p_ci(reject_out$reject,
                                        reject_out$nrep,
                                        ci_level = .95)
class(reject_out) <- "rejection_rates_df"

chk1 <- check_solution_in_by_x(
  reject_out,
  target_power = .80,
  final_nrep = 60,
  ci_level = .95,
  what = "point",
  tol = 1e-2,
  goal = "ci_hit"
)

expect_equal(chk1,
             c(FALSE, TRUE, FALSE))

chk2 <- check_solution_in_by_x(
  reject_out,
  target_power = .60,
  ci_level = .95,
  final_nrep = 60,
  what = "point",
  tol = 1e-2,
  goal = "ci_hit"
)

expect_equal(chk2,
             c(TRUE, TRUE, FALSE))

chk3 <- check_solution_in_by_x(
  reject_out,
  target_power = .890,
  final_nrep = 50,
  ci_level = .95,
  what = "ub",
  tol = 1e-2,
  goal = "close_enough"
)

expect_equal(chk3,
             c(FALSE, FALSE, TRUE))

chk4 <- check_solution_in_by_x(
  reject_out,
  target_power = .58,
  ci_level = .95,
  final_nrep = 60,
  what = "lb",
  tol = 1e-2,
  goal = "close_enough"
)

expect_equal(chk4,
             c(FALSE, TRUE, FALSE))

})

library(testthat)

test_that("bisection: n", {

ci_from_f <- function(f_i,
                      nrep,
                      ci_level = .95) {
  a <- abs(stats::qnorm((1 - ci_level) / 2))
  se_i <- sqrt(f_i * (1 - f_i) / nrep)
  cilb <- f_i - a * se_i
  ciub <- f_i + a * se_i
  c(cilb, ciub)
}

# Use Wilson CI
ci_from_f <- function(f_i,
                      nrep,
                      ci_level = .95) {
  reject_ci_wilson(round(f_i * nrep),
                   nrep,
                   ci_level)
}

ci_from_f(.70, 100)
expect_false(check_solution(
               .70,
               target_power = .80,
               nrep = 100,
               final_nrep = 100,
               what = "point",
               tol = 1e-2,
               goal = "ci_hit"
             ))

ci_from_f(.70, 10)
expect_true(check_solution(
              .70,
              target_power = .80,
              nrep = 10,
              final_nrep = 10,
              what = "point",
              tol = 1e-2,
              goal = "ci_hit"
            ))

expect_false(check_solution(
               .70,
               target_power = .80,
               nrep = 10,
               final_nrep = 10,
               what = "point",
               tol = 1e-2,
               goal = "close_enough"
             ))

expect_true(check_solution(
              .79,
              target_power = .80,
              nrep = 10,
              final_nrep = 10,
              what = "point",
              tol = .02,
              goal = "close_enough"
            ))

ci_from_f(.65, 25)
expect_true(check_solution(
              .65,
              target_power = .80,
              nrep = 25,
              final_nrep = 25,
              what = "ub",
              tol = .02,
              goal = "close_enough"
            ))
expect_false(check_solution(
               .60,
               target_power = .80,
               nrep = 10,
               final_nrep = 10,
               what = "ub",
               tol = .02,
               goal = "close_enough"
             ))

ci_from_f(.90, 50)
expect_true(check_solution(
              .90,
              target_power = .80,
              nrep = 50,
              final_nrep = 50,
              what = "lb",
              tol = .02,
              goal = "close_enough"
            ))
expect_false(check_solution(
               .90,
               target_power = .80,
               nrep = 100,
               final_nrep = 100,
               what = "lb",
               tol = .02,
               goal = "close_enough"
            ))

})

library(testthat)

test_that("Adjusted target power", {
out <- target_power_adjusted(
          .70,
          goal = "close_enough",
          what = "ub",
          nrep = 1000,
          tolerance = .03)
out
chk <- reject_ci_wilson(
          nreject = ceiling(out * 1000),
          nvalid = 1000,
          level = .95)
expect_equal(chk[2],
             .70,
             tolerance = 1e-2)

out <- target_power_adjusted(
          .70,
          goal = "close_enough",
          what = "lb",
          nrep = 1000,
          tolerance = .03)
out
chk <- reject_ci_wilson(
          nreject = ceiling(out * 1000),
          nvalid = 1000,
          level = .95)
expect_equal(chk[1],
             .70,
             tolerance = 1e-2)

out <- target_power_adjusted(
          .70,
          goal = "close_enough",
          what = "point",
          nrep = 1000,
          tolerance = .03)
out
expect_equal(out,
             .70,
             tolerance = 1e-2)

out <- target_power_adjusted(
          .70,
          goal = "ci_hit",
          what = "point",
          nrep = 1000,
          tolerance = .03)
out
expect_equal(out,
             .70,
             tolerance = 1e-2)
})

#' @title Test Group Constraints
#'
#' @description Test the model fit change
#' when one or more between-group
#' constraints are imposed.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing
#' the difference in model fit
#' when one or more between-group
#' constraints are imposed
#' , by
#' setting it to the `test_fun`
#' argument.
#'
#'
#' @return
#' In its normal usage, it returns
#' a one-row data frame with the
#' following columns:
#'
#' - `est`: The chi-square difference.
#'
#' - `cilo` and `cihi`: `NA`. Not used.
#'
#' - `sig`: Whether the chi-square
#' difference test is significant
#'
#' - `test_label`: The constraints
#'  imposted.
#'
#' @inheritParams test_k_indirect_effects
#'
#' @inheritParams test_indirect_effect
#'
#' @param fit The fit object. Must be
#' the output of [lavaan::lavaan()] or
#' its wrappers, such as [lavaan::sem()]
#' and [lavaan::cfa()]. The model must
#' be a multigroup model.
#'
#' @param group.equal The same argument
#' used by `lavaan`. A character vector
#' with one or more of these values:
#' `"regressions"`, `"loadings"`,
#' `"lv.covariances"`, `"lv.variances"`,
#' `"intercepts"`, `"means"`, `"thresholds"`,
#' `"residual.covariances"`, `"composite.weights"`,
#' and `"residuals"`.
#'
#' @param group.partial The same argument
#' used by `lavaan`. The parameters that
#' should be free across groups. Used with
#' `group.equal` to exclude some parameters
#' from those requested to be equal
#' across groups by `group.equal`.
#'
#' @param ... Optional arguments to be
#' passed to [lavaan::lavTestLRT()].
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' y ~ m: l
#' m ~ x:
#'   - nil
#'   - s
#' y ~ x: nil
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        iseed = 1234)
#'
#' # Do the tests in each replication
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_group_equal,
#'                        test_args = list(group.equal = "regressions"))
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#'
#' @export
test_group_equal <- function(fit = fit,
                             group.equal = NULL,
                             group.partial = NULL,
                             check_post_check = TRUE,
                             ...,
                             fit_name = "fit",
                             get_map_names = FALSE,
                             get_test_name = FALSE) {
  map_names <- c(fit = fit_name)
  args <- list(...)
  if (get_map_names) {

    # ==== Return map_names ====

    return(map_names)
  }
  if (get_test_name) {

    # ==== Return test_name ====

    return("test_group_equal")

  }

  # ==== Prepare test_label ====

  tmp <- character(0)
  if (!is.null(group.equal)) {
    tmp <- c(paste0("equal(",
                    paste0(group.equal, collapse = ","),
                    ")"))
  }
  if (!is.null(group.partial)) {
    tmp <- c(tmp,
             c(paste0("partial(",
                    paste0(group.partial, collapse = ","),
                    ")")))
  }
  tmp <- paste0(tmp,
                collapse = ";")

  test_label <- tmp

  # ==== Check the type of fit ====

  if (inherits(fit, "lavaan")) {
    fit_type <- "lavaan"
  } else {
    stop("fit is not a supported object.")
  }

  ngroups <- lavaan::lavInspect(fit,
                                "ngroups")
  if (ngroups == 1) {
    stop("Only support multigroup models.")
  }

  # ==== Is the lavaan fit OK? ====

  fit_ok <- lavaan::lavInspect(fit, "converged") &&
            (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
              !check_post_check))

  # ==== Refit the model ====

  if (fit_ok) {
    suppressWarnings(
      fit_update <- methods::getMethod("update",
                  signature = "lavaan",
                  where = asNamespace("lavaan"))(
                    fit,
                    group.equal = group.equal,
                    group.partial = group.partial
                  )
    )
    update_ok <- lavaan::lavInspect(fit_update, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit_update, "post.check") ||
                !check_post_check))
  } else {
    update_ok <- FALSE
  }

  # ==== Fit not OK. Return NAs ====

  if (!update_ok) {
    out1 <- data.frame(test_label = test_label,
                       est = NA,
                       cilo = NA,
                       cihi = NA,
                       sig = NA,
                       pvalue = NA)
    return(out1)
  }

  # ==== Compare the models ====

  out0 <- lavaan::lavTestLRT(
              fit,
              fit_update,
              ...
            )

  out1 <- data.frame(test_label = test_label,
                     est = out0[2, "Chisq diff"],
                     cilo = NA,
                     cihi = NA,
                     sig = NA,
                     pvalue = out0[2, "Pr(>Chisq)"])

  tmp_level <- args$level %||% .95
  out1$sig <- ifelse(out1$pvalue < (1 - tmp_level),
                     yes = 1,
                     no = 0)

  # ==== Prepare the output ====

  attr(out1, "test_label") <- "test_label"
  return(out1)

}

#' @title Plot a Power Curve
#'
#' @description Plotting the results
#' in a 'power_curve' object, such as the
#' estimated power against sample size,
#' or the results of [power4test_by_n()]
#' or [power4test_by_es()].
#'
#' @details
#' The `plot` method of `power_curve`
#' objects currently plots the relation
#' between estimated power and
#' the predictor. Other elements
#' can be requested (see the argument
#' `what`), and they can be customized
#' individually.
#'
#' @return
#' The `plot`-methods return `x`
#' invisibly. They
#' are called for their side effects.
#'
#' @param x The object to be plotted.
#' It can be a `power_curve` object,
#' the output of [power_curve()]. It
#' can also be the output of
#' [power4test_by_n()] or
#' [power4test_by_es()].
#'
#' @param what A character vector of
#' what to include in the
#' plot. Possible values are
#' `"ci"` (confidence intervals
#' for the estimated sample size) and
#' `"power_curve"` (the crude power
#' curve, if available).
#' The default values depend on the
#' type of `x`.
#'
#' @param main The title of the plot.
#'
#' @param xlab,ylab The labels for the
#' horizontal and vertical axes,
#' respectively.
#'
#' @param pars_ci A named list of
#' arguments to be passed to [arrows()]
#' to customize the drawing of the
#' confidence intervals.
#'
#' @param type An argument of the
#' default plot method [plot.default()].
#' Default is `"l"`. See [plot.default()]
#' for other options.
#'
#' @param ylim A two-element numeric
#' vector of the range of the vertical
#' axis.
#'
#' @param ci_level The level of
#' confidence of the confidence intervals,
#' if requested. Default is `.95`, denoting
#' 95%.
#'
#' @param ... Optional arguments.
#' Passed to [plot()] when drawing
#' the base plot.
#'
#' @seealso [power_curve()],
#' [power4test_by_n()], and
#' [power4test_by_es()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the effect sizes (population parameter values)
#'
#' model_simple_med_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: s
#' "
#'
#' # Simulate datasets to check the model
#'
#' sim_only <- power4test(nrep = 10,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 50,
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234,
#'                        parallel = FALSE,
#'                        progress = FALSE)
#'
#' # By n: Do a test for different sample sizes
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' out1 <- power4test_by_n(sim_only,
#'                         nrep = 10,
#'                         test_fun = test_parameters,
#'                         test_args = list(par = "y~x"),
#'                         n = c(25, 50, 100),
#'                         by_seed = 1234,
#'                         parallel = FALSE,
#'                         progress = FALSE)
#'
#' pout1 <- power_curve(out1)
#' pout1
#' plot(pout1)
#'
#' # By pop_es: Do a test for different population values of a model parameter
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' out2 <- power4test_by_es(sim_only,
#'                              nrep = 10,
#'                              test_fun = test_parameters,
#'                              test_args = list(par = "y~x"),
#'                              pop_es_name = "y ~ x",
#'                              pop_es_values = c(0, .3, .5),
#'                              by_seed = 1234,
#'                              parallel = FALSE,
#'                              progress = FALSE)
#'
#' pout2 <- power_curve(out2)
#' plot(pout2)
#'
#' @importFrom graphics abline arrows par points text title
#'
#' @export
plot.power_curve <- function(x,
                             what = c("ci", "power_curve"),
                             main = paste0("Power Curve ",
                                           "(Predictor: ",
                                           switch(x$predictor, n = "Sample Size", es = "Effect Size"),
                                           ")"),
                             xlab = switch(x$predictor, n = "Sample Size", es = "Effect Size"),
                             ylab = "Estimated Power",
                             pars_ci = list(),
                             type = "l",
                             ylim = c(0, 1),
                             ci_level = .95,
                             ...) {

  what <- match.arg(what, several.ok = TRUE)

  # No need for other arguments. They should be stored in object.
  reject_df <- rejection_rates_add_ci(x$reject_df,
                                      level = ci_level)

  # === Draw the base plot: Power vs. N

  predictor <- x$predictor

  plot(x = reject_df$x,
       y = reject_df$reject,
       main = main,
       xlab = xlab,
       ylab = ylab,
       type = type,
       ylim = ylim,
       ...)

  # === Add CIs?

  if ("ci" %in% what) {

    tmp_args <- utils::modifyList(pars_ci,
                                  list(reject_df = reject_df))
    do.call(plot_power_curve_ci,
            tmp_args)
  }

  # === Draw the power curve?

  if ("power_curve" %in% what) {
    tmp_args <- utils::modifyList(pars_ci,
                                  list(x = x))
    do.call(plot_power_curve_curve,
            tmp_args)
  }

  invisible(x)
}

#' @noRd
plot_power_curve_ci <- function(reject_df,
                                length = .1,
                                angle = 90,
                                code = 3,
                                col = "grey50",
                                ...) {
  # object should be a data frame with CIs already computed
  # Some CIs may be of zero width
  i <- !(reject_df$reject_ci_lo == reject_df$reject_ci_hi)
  if (any(i)) {
    arrows(x0 = reject_df$x[i],
           y0 = reject_df$reject_ci_lo[i],
           x1 = reject_df$x[i],
           y1 = reject_df$reject_ci_hi[i],
           length = length,
           angle = angle,
           code = code,
           col = col,
           ...)
  }
}

#' @noRd

plot_power_curve_curve <- function(x,
                                   type = "l",
                                   lwd = 2,
                                   col = "red",
                                   curve_points = 20,
                                   ...) {
  # x is a power_curve object
  reject_df <- x$reject_df
  x_new <- seq(min(reject_df$x),
               max(reject_df$x),
               length.out = curve_points)
  fit <- x$fit
  if (inherits(fit, "nls") || inherits(fit, "lm")) {
    y_new <- stats::predict(x,
                            newdata = list(x = x_new))
    points(x = x_new,
           y = y_new,
           type = type,
           lwd = lwd,
           col = col,
           ...)
  }
}

#' @rdname plot.power_curve
#' @export
plot.power4test_by_n <- function(
                             x,
                             what = c("ci", "power_curve"),
                             main = "Estimated Power vs. Sample Size",
                             xlab = "Sample Size",
                             ylab = "Estimated Power",
                             pars_ci = list(),
                             type = "l",
                             ylim = c(0, 1),
                             ci_level = .95,
                             ...) {
  what <- match.arg(what, several.ok = TRUE)
  # No need for other arguments. They should be stored in object.
  reject_df <- rejection_rates(x,
                               level = ci_level,
                               all_columns = TRUE)
  reject_df$x <- reject_df$n
  x0 <- list(predictor = "n",
             reject_df = reject_df)

  plot.power_curve(x = x0,
                   what = what,
                   main = main,
                   xlab = xlab,
                   ylab = ylab,
                   pars_ci = pars_ci,
                   type = type,
                   ylim = ylim,
                   ci_level = ci_level,
                   ...)
  invisible(x)
}

#' @rdname plot.power_curve
#' @export
plot.power4test_by_es <- function(
                             x,
                             what = c("ci", "power_curve"),
                             main = paste0("Estimated Power vs. Effect Size / Parameter (",
                                           attr(x[[1]], "pop_es_name"), ")"),
                             xlab = paste0("Effect Size / Parameter (",
                                           attr(x[[1]], "pop_es_name"), ")"),
                             ylab = "Estiamted Power",
                             pars_ci = list(),
                             type = "l",
                             ylim = c(0, 1),
                             ci_level = .95,
                             ...) {
  what <- match.arg(what, several.ok = TRUE)
  # No need for other arguments. They should be stored in object.
  reject_df <- rejection_rates(x,
                               level = ci_level,
                               all_columns = TRUE)
  reject_df$x <- reject_df$es
  x0 <- list(predictor = "es",
             reject_df = reject_df)

  plot.power_curve(x = x0,
                   main = main,
                   xlab = xlab,
                   ylab = ylab,
                   pars_ci = pars_ci,
                   type = type,
                   ylim = ylim,
                   ci_level = ci_level,
                   ...)
  invisible(x)
}

#' @title Rejection Rates
#'
#' @description Get all rejection rates
#' of all tests stored in a `power4test`
#' object or other supported objects.
#'
#' @details
#' For a `power4test` object,
#' [rejection_rates] loops over the tests stored
#' in a `power4test` object and retrieves
#' the rejection rate of each test.
#'
#' @return
#' The `rejection_rates` method returns
#' a `rejection_rates_df` object,
#' with a `print` method.
#'
#' If the input (`object`) is a
#' `power4test` object, the
#' `rejection_rates_df` object is
#' a data-frame like object with the
#' number of
#' rows equal to the number of tests.
#' Note that some tests, such as
#' the test by [test_parameters()],
#' conduct one test for each parameter.
#' Each such test is counted as one
#' test.
#'
#' The data frame has at least these columns:
#'
#' - `test`: The name of the test.
#'
#' - `label`: The label for each
#'  test, or `"Test"` if a test only
#'  does one test (e.g., [test_indirect_effect()]).
#'
#' - `pvalid`: The proportion of valid
#'  tests across all replications.
#'
#' - `reject`: The rejection rate for
#'  each test. If the null hypothesis
#'  is false, then this is the power.
#'
#' @param object The object
#' from which the rejection rates
#' are to be computed, such as
#' a `power4test` object,
#' a `power4test_by_n` object,
#' or a `power4test_by_es` object.
#'
#' @param ... Optional arguments. For
#' the `print` method, these arguments
#' will be passed to the `print` method
#' of `data.frame` objects [print.data.frame()].
#' For the `rejection_rates` method
#' of `x_from_power` and `n_region_from_power` objects, these are
#' arguments to be passed to the
#' `rejection_rates` method
#' for `power4test_by_n`
#' and `power4test_by_es` objects.
#' For the `rejection_rates` method for
#' `q_power_mediation` objects (the
#' output of [q_power_mediation()]
#' and friends), they are optional arguments
#' to be passed to the corresponding methods.
#'
#' @seealso [power4test()],
#' [power4test_by_n()], and
#' [power4test_by_es()], which are
#' supported by this method.
#'
#' @examples
#'
#' # Specify the population model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the effect sizes (population parameter values)
#'
#' model_simple_med_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate some datasets to check the model
#'
#' sim_only <- power4test(nrep = 4,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 100,
#'                        R = 50,
#'                        ci_type = "boot",
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test 'test_indirect_effect' on each datasets
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_indirect_effect,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         boot_ci = TRUE,
#'                                         mc_ci = FALSE))
#'
#' # Do the test 'test_parameters' on each datasets
#' # and add the results to 'test_out'
#'
#' test_out <- power4test(object = test_out,
#'                        test_fun = test_parameters)
#'
#' # Compute and print the rejection rates for stored tests
#'
#' rejection_rates(test_out)
#'
#' # See the help pages of power4test_by_n() and power4test_by_es()
#' # for other examples.
#'
#' @export
rejection_rates <- function(object,
                            ...) {
  UseMethod("rejection_rates")
}

#' @export
#' @rdname rejection_rates
rejection_rates.default <- function(object,
                                    ...) {
  stop("The object is not of a supported class.")
}

#' @param all_columns If `TRUE`, all
#' columns stored by a test will be
#' extracted. Default is `FALSE` and
#' only essential columns related to
#' power will be printed.
#'
#' @param ci If `TRUE`, confidence
#' intervals for the rejection rates
#' (column `reject` or `sig`) will
#' be computed. The method is determined
#' by the option `power4mome.ci_method`.
#' If `NULL` or `"wilson"`, Wilson's
#' (1927) method is used. If
#' `"norm"`, normal approximation
#' is used.
#'
#' @param level The level of confidence
#' for the confidence intervals, if
#' `ci` is `TRUE`. Default is .95,
#' denoting 95%.
#'
#' @param se If `TRUE`, standard errors
#' for the rejection rates
#' (column `reject` or `sig`) will
#' be computed. Normal approximation
#' is used to compute the standard
#' errors.
#'
#' @param collapse Whether a single
#' decision (significant vs. not significant)
#' is made across all tests for a test
#' that consists of several tests
#' (e.g., the tests of several parameters).
#' If `"none"`, tests will be summarized
#' individually. If `"all_sig"`, then
#' the set of tests is considered significant
#' if all individual tests are significant.
#' If `"at_least_one_sig"`, then the set of
#' tests is considered significant if
#' at least one of the tests is significant.
#' If `"at_least_k_sig"`, then the set of
#' tests is considered significant if
#' at least `k` tests are significant,
#' `k` set by the argument `at_least_k`.
#' If `NULL`, will use the value stored
#' in `object` (default is `"none"`).
#'
#' @param at_least_k Used by `collapse`,
#' the number of tests required to be
#' significant for the set of tests to
#' be considered significant. If `NULL`,
#' will use the value stored in `object`
#' (default is 1).
#'
#' @param merge_all_tests If `TRUE`, all
#' the tests in each replication will be
#' merged into one test.
#' If `NULL`, will use the value stored
#' in `object` (default is `FALSE`).
#'
#' @param p_adjust_method The method to be
#' passed to [p.adjust()] to adjust the
#' *p*-values when testing the effects.
#' Default is `"none"` and the *p*-values
#' will not be adjusted. Ignored if some tests
#' do not have *p*-values stored.
#' NOTE: Use this only if all tests
#' can be  conducted using *p*-values.
#' If `NULL`, will use the value stored in
#' `object` (default is `"none"`).
#'
#' @param alpha The level of significance
#' to use when using `p_adjust_method`.
#' The significance results (the column
#' `sig`) will be updated using the
#' adjusted *p*-values. Used only if
#' `p_adjust_method` is not `"none"`.
#' If `NULL`, will use the value stored
#' in `object` (default is .05).
#'
#' @param nrep_if_diff If `TRUE` and
#' the numbers of replications across
#' rows are different, `nrep` will be
#' included in the output. If `FALSE`,
#' `nrep` will not be included unless
#' `all_columns` is `TRUE`.
#'
#' @param keep_nrep If `TRUE`, the column
#' `nrep` will be kept.
#'
#' @references
#' Wilson, E. B. (1927). Probable inference, the law of
#' succession, and statistical inference.
#' *Journal of the American Statistical Association, 22*(158),
#' 209-212. \doi{10.1080/01621459.1927.10502953}
#'
#' @export
#' @rdname rejection_rates
rejection_rates.power4test <- function(object,
                                       all_columns = FALSE,
                                       ci = TRUE,
                                       level = .95,
                                       se = FALSE,
                                       collapse = NULL,
                                       at_least_k = NULL,
                                       merge_all_tests = NULL,
                                       p_adjust_method = NULL,
                                       alpha = NULL,
                                       keep_nrep = FALSE,
                                       ...) {
  stored_args <- attr(object, "args")$rejection_rates_args
  collapse <- collapse %||% (stored_args$collapse %||% "none")
  at_least_k <- at_least_k %||% (stored_args$at_least_k %||% 1)
  merge_all_tests <- merge_all_tests %||% (stored_args$merge_all_tests %||% FALSE)
  p_adjust_method <- p_adjust_method %||% (stored_args$p_adjust_method %||% "none")
  alpha <- alpha %||% (stored_args$alpha %||% .05)

  out0 <- summarize_tests(object,
                          collapse = collapse,
                          at_least_k = at_least_k,
                          merge_all_tests = merge_all_tests,
                          p_adjust_method = p_adjust_method,
                          alpha = alpha)
  tests_merged <- out0[[1]]$test_attributes$tests
  out1 <- lapply(out0,
                 rejection_rates_i,
                 all_columns = all_columns,
                 ci = ci,
                 level = level,
                 se = se)
  if (all_columns) {
    out2 <- do.call(rbind_adv,
                    out1)
  } else {
    out2 <- do.call(rbind,
                    out1)
  }
  if (!all_columns &&
      !keep_nrep) {
    out2$nrep <- NULL
  }
  rownames(out2) <- NULL
  class(out2) <- c("rejection_rates_df",
                   class(out2))
  extra_attr <- lapply(
                    out1,
                    \(x) attr(x, "extra")
                  )
  attr(out2, "extra_list") <- extra_attr
  attr(out2, "tests_merged") <- tests_merged
  out2
}

#' @noRd
rejection_rates_i <- function(object_i,
                                  all_columns = FALSE,
                                  ci = ci,
                                  level = level,
                                  se = se) {
  if (is.vector(object_i$mean)) {
    out <- rejection_rates_i_vector(object_i,
                                        all_columns = all_columns,
                                        ci = ci,
                                        level = level,
                                        se = se)
    return(out)
  }
  if (length(dim(object_i$mean)) == 2) {
    # Likely a data frame
    out <- rejection_rates_i_data_frame(object_i,
                                            all_columns = all_columns,
                                            ci = ci,
                                            level = level,
                                            se = se)
    return(out)
  }
  stop("Something is wrong. The tests results are not of the supported format.")
}

#' @noRd
rejection_rates_i_vector <- function(object_i,
                                         all_columns = FALSE,
                                         ci = TRUE,
                                         level = .95,
                                         se = FALSE) {
  test_args <- object_i$test_attributes
  test_name <- test_args$test_name
  if (all_columns) {
    out_i <- data.frame(test = test_name,
                        test_label = "Test",
                        pvalid = object_i$nvalid["sig"] / object_i$nrep,
                        nvalid = object_i$nvalid["sig"],
                        nrep = object_i$nrep,
                        rbind(object_i$mean),
                        row.names = NULL)
    tmp <- gsub("sig",
                "reject",
                colnames(out_i),
                fixed = TRUE)
    colnames(out_i) <- tmp
  } else {
    out_i <- data.frame(test = test_name,
                        test_label = "Test",
                        est = object_i$mean["est"],
                        pvalid = object_i$nvalid["sig"] / object_i$nrep,
                        nvalid = object_i$nvalid["sig"],
                        nrep = object_i$nrep,
                        reject = object_i$mean["sig"],
                        row.names = NULL)
  }
  if (se || ci) {
    out_i <- rejection_rates_add_ci(out_i,
                                    level = level,
                                    add_reject = FALSE,
                                    add_se = se)
    if (!ci) {
      out_i$reject_ci_lo <- NULL
      out_i$reject_ci_hi <- NULL
    }
  }
  if (!all_columns) {
    # nvalid used only for adding CI and SE
    out_i$nvalid <- NULL
  }
  attr(out_i, "extra") <- attr(object_i, "extra")
  out_i
}

#' @noRd
rejection_rates_i_data_frame <- function(object_i,
                                         all_columns = FALSE,
                                         ci = TRUE,
                                         level = .95,
                                         se = FALSE) {
  test_args <- object_i$test_attributes
  test_name <- test_args$test_name
  out_i0 <- object_i$mean
  p <- nrow(out_i0)
  pvalid <- object_i$nvalid[, "sig", drop = TRUE] / object_i$nrep
  if (all_columns) {
    out_i <- data.frame(test = test_name,
                        pvalid = pvalid,
                        nvalid = object_i$nvalid[, "sig", drop = TRUE],
                        nrep = object_i$nrep,
                        object_i$mean,
                        row.names = NULL)
    tmp <- gsub("sig",
                "reject",
                colnames(out_i),
                fixed = TRUE)
    colnames(out_i) <- tmp
    class(out_i) <- class(object_i$mean)
  } else {
    out_i <- data.frame(test = test_name,
                        test_label = out_i0$test_label,
                        est = object_i$mean[, "est", drop = TRUE],
                        pvalid = pvalid,
                        nvalid = object_i$nvalid[, "sig", drop = TRUE],
                        nrep = object_i$nrep,
                        reject = object_i$mean[, "sig", drop = TRUE],
                        row.names = NULL)
  }
  if (se || ci) {
    out_i <- rejection_rates_add_ci(out_i,
                                    level = level,
                                    add_reject = FALSE,
                                    add_se = se)
    if (!ci) {
      out_i$reject_ci_lo <- NULL
      out_i$reject_ci_hi <- NULL
    }
  }
  if (!all_columns) {
    # nvalid used only for adding CI and SE
    out_i$nvalid <- NULL
  }
  attr(out_i, "extra") <- attr(object_i, "extra")
  out_i
}

#' @noRd

rbind_adv <- function(...) {
  # Need this function only if all_columns
  dfs <- list(...)
  cnames <- lapply(dfs,
                   colnames)
  cnames_all <- Reduce(union, cnames)
  # If all_columns
  # - test
  # - test_label
  # - pvalid
  # - nvalid
  # - nrep
  # - ...
  # - se, reject_ci_lo, reject_ci_hi (if se || ci)
  c1 <- lapply(cnames,
               \(x) setdiff(x, c("test",
                                 "test_label",
                                 "pvalid",
                                 "nvalid",
                                 "nrep",
                                 "sig",
                                 "reject_se",
                                 "reject_ci_lo",
                                 "reject_ci_hi")))
  c2 <- Reduce(union, c1)
  c3 <- c("test",
          "test_label",
          "pvalid",
          "nvalid",
          "nrep",
          c2)
  cnames_final <- c(c3, setdiff(cnames_all, c3))
  tmpfct <- function(x) {
    # Add NA
    d1 <- setdiff(cnames_final, colnames(x))
    if (length(d1) == 0) return(x)
    for (y in d1) {
      x[y] <- NA
    }
    x <- x[cnames_final]
    return(x)
  }
  dfs_out <- lapply(dfs,
                    tmpfct)
  dfs_out1 <- do.call(rbind,
                      dfs_out)
  dfs_out1
}

#' @return
#' The `rejection_rates` method
#' for `power4test_by_es` objects
#' returns an object of the
#' class `rejection_rates_df_by_es`,
#' which is a subclass of
#' `rejection_rates_df`.
#' It is a data frame which is
#' similar to the output of
#' [rejection_rates()], with two
#' columns added for the effect size (`pop_es_name` and
#' `pop_es_values`)
#' for each test.
#'
#' @details
#' The `rejection_rates` method for
#' `power4test_by_es` objects
#' is used to compute the rejection
#' rates from a `power4test_by_es`
#' object, with effect sizes added to
#' the output.
#'
#' @rdname rejection_rates
#' @export
rejection_rates.power4test_by_es <- function(object,
                                             all_columns = FALSE,
                                             ci = TRUE,
                                             level = .95,
                                             se = FALSE,
                                             nrep_if_diff = TRUE,
                                             ...) {
  out <- rejection_rates_by_es(object_by_es = object,
                               all_columns = all_columns,
                               ci = ci,
                               level = level,
                               se = se,
                               nrep_if_diff = nrep_if_diff,
                               ...)
  class(out) <- c("rejection_rates_df_by_es",
                  "rejection_rates_df",
                  class(out))
  out
}

#' @return
#' The `rejection_rates` method
#' for `power4test_by_n` objects
#' returns an object of the
#' class `rejection_rates_df_by_n`,
#' which is a subclass of
#' `rejection_rates_df`.
#' It is a data frame which is
#' similar to the output of
#' a `power4test` object, with a
#' column `n` added for the sample size
#' for each test.
#'
#' @details
#' The `rejection_rates` method for
#' `power4test_by_n` objects
#' is used to compute the rejection
#' rates, with sample sizes added to
#' the output.
#'
#' @rdname rejection_rates
#' @export
rejection_rates.power4test_by_n <- function(object,
                                            all_columns = FALSE,
                                            ci = TRUE,
                                            level = .95,
                                            se = FALSE,
                                            nrep_if_diff = TRUE,
                                            ...) {
  out <- rejection_rates_by_n(object_by_n = object,
                              all_columns = all_columns,
                              ci = ci,
                              level = level,
                              se = se,
                              nrep_if_diff = nrep_if_diff,
                              ...)
  class(out) <- c("rejection_rates_df_by_n",
                  "rejection_rates_df",
                  class(out))
  out
}

#' @return
#' The `rejection_rates` method
#' for `x_from_power` objects
#' retrieves the stored
#' `power4test_by_n` or
#' `power4test_by_es` object,
#' and then
#' runs `rejection_rates` on it
#' and returns the result.
#'
#' @details
#' The `rejection_rates` method for
#' `x_from_power` objects
#' is used to compute the rejection
#' rates for stored trials. It supports
#' the output of [x_from_power()] and
#' its wrappers, such as [n_from_power()].
#'
#' @rdname rejection_rates
#' @export
rejection_rates.x_from_power <- function(
  object,
  ...
) {
  x <- object$power4test_trials
  rejection_rates(object = x,
                  ...)
}


#' @return
#' The `rejection_rates` method
#' for `n_region_from_power` objects
#' retrieves the stored
#' `power4test_by_n` object from
#' the `above` element (the search
#' for the region with power significantly
#' above the target power)
#' and then
#' runs `rejection_rates` on it
#' and returns the result.
#'
#' @details
#' The `rejection_rates` method for
#' `n_region_from_power` objects
#' is used to compute the rejection
#' rates for stored trials. It supports
#' the output of [n_region_from_power()].
#' It is sufficient to retrieve the
#' trials in searching for the upper bound
#' because they also include the trials
#' used in searching for the lower bound.
#'
#' @rdname rejection_rates
#' @export
rejection_rates.n_region_from_power <- function(
  object,
  ...
) {
  x <- object$above
  rejection_rates(object = x,
                  ...)
}


#' @return
#' The `rejection_rates` method
#' for `q_power_mediation` objects
#' retrieves the trials from
#' and then
#' runs `rejection_rates` on them
#' and returns the result. If `mode`
#' is `"n"`, then the stored output
#' from [n_from_power()] is used.
#' If `mode` is `"region"`, then
#' the stored output from [n_region_from_power()]
#' is used. If `mode` is `"power"`, then
#' the stored output from [power4test()]
#' is used.
#'
#' @details
#' The `rejection_rates` method for
#' `q_power_mediation` objects
#' is used to compute the rejection
#' rates for stored trials.
#'
#' @rdname rejection_rates
#' @export
rejection_rates.q_power_mediation <- function(
  object,
  ...
) {
  if (!is.null(object$n_region_from_power)) {
    out <- rejection_rates(
              object = object$n_region_from_power,
              ...
            )
  } else if (!is.null(object$n_from_power)) {
    out <- rejection_rates(
              object = object$n_from_power,
              ...
            )
  } else {
    out <- rejection_rates(
              object = object$power4test,
              ...
            )
  }
  return(out)
}

#' @param x The `rejection_rates_df`
#' object to be printed.
#'
#' @param digits The number of digits
#' to be printed
#' after the decimal.
#'
#' @param annotation Logical. Whether
#' additional notes will be printed.
#'
#' @param abbreviate_col_names Logical.
#' Whether some column names will be
#' abbreviated.
#'
#' @return
#' The `print` method of a
#' `rejection_rates_df` object returns
#' the object invisibly. It is called
#' for its side-effect.
#'
#' @rdname rejection_rates
#' @export
print.rejection_rates_df <- function(x,
                                     digits = 3,
                                     annotation = TRUE,
                                     abbreviate_col_names = TRUE,
                                     ...) {
  x0 <- x
  class(x0) <- "data.frame"

  # Handle special columns
  if ("nvalid" %in% colnames(x0)) {
    x0$nvalid <- as.character(x0$nvalid)
  }
  if ("nrep" %in% colnames(x0)) {
    x0$nrep <- as.character(x0$nrep)
  }

  # In by_n
  if ("n" %in% colnames(x0)) {
    x0$n <- as.character(x0$n)
  }

  test_i <- unique(x0$test)
  if (length(test_i) == 1) {
    x0$test <- NULL
  }
  test_label_i <- unique(x0$test_label)
  if (length(test_label_i) == 1) {
    x0$test_label <- NULL
  }

  abbr_names <- c(pvalid = "p.v",
                  nvalid = "n.v",
                  reject_se = "r.se",
                  reject_ci_lo = "r.cilo",
                  reject_ci_hi = "r.cihi")

  # Abbreviate column names
  if (abbreviate_col_names) {
    cnames <- colnames(x0)
    for (i in seq_along(abbr_names)) {
      cnames <- gsub(names(abbr_names)[i],
                     abbr_names[i],
                     cnames,
                     fixed = TRUE)
    }
    colnames(x0) <- cnames
  }

  x1 <- format_num_cols(x0,
                        digits = digits)

  if (length(test_i) == 1) {
    cat("[test]:", test_i, "\n")
  }
  if (length(test_label_i) == 1) {
    cat("[test_label]:", test_label_i, "\n")
  }

  print(x1,
        ...)

  if (annotation) {

    # Any rejection rates estimated by Boos-Zhang-2000?
    extra_list <- attr(x,
                      "extra_list")
    is_bz <- sapply(extra_list,
                    \(x) x$bz_extrapolated)

    cat("Notes:\n")

    if ("n" %in% colnames(x1)) {
      catwrap(paste0("- ",
                     "n",
                     ": The sample size in a trial."),
              exdent = 2)
    }
    if ("par" %in% colnames(x1)) {
      catwrap(paste0("- ",
                     "par",
                     ": The parameter being varied."),
              exdent = 2)
    }
    if ("es" %in% colnames(x1)) {
      catwrap(paste0("- ",
                     "es",
                     ": The population value of 'par' in a trial."),
              exdent = 2)
    }
    tmp <- ifelse(abbreviate_col_names,
                  abbr_names["pvalid"],
                  "pvalid")
    if (tmp %in% colnames(x1)) {
      catwrap(paste0("- ",
                     tmp,
                     ": The proportion of valid replications."),
              exdent = 2)
    }
    tmp <- ifelse(abbreviate_col_names,
                  abbr_names["nvalid"],
                  "nvalid")
    if (tmp %in% colnames(x1)) {
      catwrap(paste0("- ",
                     tmp,
                     ": The number of valid replications."),
              exdent = 2)
    }
    if ("nrep" %in% colnames(x1)) {
      tmp <- "nrep"
      catwrap(paste0("- ",
                     tmp,
                     ": The number of replications."),
              exdent = 2)
    }
    if ("est" %in% colnames(x1)) {
      catwrap(paste0("- est: The mean of the estimates in a test across replications."),
              exdent = 2)
    }
    if ("cilo" %in% colnames(x1)) {
      catwrap(paste0("- cilo, cihi: The mean of the lower/upper limits of ",
                     "the confidence intervals in a test ",
                     "across replications."),
              exdent = 2)
    }
    if ("reject" %in% colnames(x1)) {
      catwrap(paste0("- reject: The proportion of 'significant' replications, ",
                     "that is, the rejection rate. ",
                     "If the null hypothesis is true, this is the Type I error rate. ",
                     "If the null hypothesis is false, this is the power."),
              exdent = 2)
      if (any(is_bz)) {
        catwrap(paste0("- Some or all values in 'reject' are estimated using ",
                      "the extrapolation method by Boos and Zhang (2000)."),
                exdent = 2)
      }
    }
    tmp <- ifelse(abbreviate_col_names,
                  abbr_names["reject_se"],
                  "reject_se")
    if (tmp %in% colnames(x1)) {
      catwrap(paste0("- ",
                     tmp,
                     ": The standard error of the rejection rate, ",
                     "based on normal approximation."),
              exdent = 2)
    }
    tmp1 <- ifelse(abbreviate_col_names,
                   abbr_names["reject_ci_lo"],
                   "reject_ci_lo")
    tmp2 <- ifelse(abbreviate_col_names,
                   abbr_names["reject_ci_hi"],
                   "reject_ci_hi")
    if (tmp1 %in% colnames(x1)) {
      tmp <- paste0(tmp1, ",", tmp2)
      ci_method <- getOption("power4mome.ci_method", default = "wilson")
      tmp3 <- switch(ci_method,
                     wilson = "Wilson's (1927) method",
                     norm = "normal approximation")
      catwrap(paste0("- ",
                     tmp,
                     ": The confidence interval ",
                     "of the rejection rate, based on ",
                     tmp3,
                     "."),
              exdent = 2)
      if (any(is_bz) &&
          (ci_method == "wilson")) {
        catwrap(paste0(
                  "- Wilson's (1927) method is used to ",
                  "approximate the confidence intervals of the rejection rates ",
                  "estimated by the method of Boos and Zhang (2000)."
                ),
                exdent = 2)
      }
    }
    tmp <- attr(x, "tests_merged")
    if (!is.null(tmp)) {
      catwrap(paste0(
                "- Test(s) merged:"
              ),
              exdent = 2)
      catwrap(paste("-", tmp),
              indent = 1)
    }
    catwrap(paste0("- Refer to the tests for the meanings of other columns."),
            exdent = 2)
  }
  invisible(x)
}

skip_on_cran()

library(testthat)

test_that("x_from_power: n", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 5,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = c("y~m", "y~x")),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_error(tmp <- x_from_power(out,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve"))

# No error if collapsed to one test

out <- power4test(nrep = 5,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = c("y~m", "y~x"),
                                   omnibus = "all_sig"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp <- x_from_power(out,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve"))

})

library(testthat)

test_that("All parameters: Multigroup: Compare groups", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(exclude_var = TRUE,
                                        op = "~",
                                        compare_groups =  TRUE))
tmp <- rejection_rates(test_out)

expect_true(all(grepl("1==2", tmp$test_label)))

})

library(testthat)

test_that("All parameters: lm", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "s",
                         "y ~ x" = "m")

sim_only <- power4test(nrep = 10,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 200,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

chk0 <- summarize_tests(test_out)
chk1 <- summarize_tests(test_out, collapse = "all_sig")
chk2 <- summarize_tests(test_out, collapse = "at_least_one_sig")
chk3 <- summarize_tests(test_out, collapse = "at_least_k_sig", at_least_k = 1)
chk4 <- summarize_tests(test_out, collapse = "at_least_k_sig", at_least_k = 2)
chk5 <- summarize_tests(test_out, collapse = "at_least_k_sig", at_least_k = 3)

expect_equal(chk1[[1]]$mean$sig,
             chk5[[1]]$mean$sig)
expect_equal(chk2[[1]]$mean$sig,
             chk3[[1]]$mean$sig)
expect_true(chk3[[1]]$mean$sig >= chk4[[1]]$mean$sig)
expect_true(chk4[[1]]$mean$sig >= chk5[[1]]$mean$sig)

chk0 <- rejection_rates(test_out)
chk1 <- rejection_rates(test_out, collapse = "all_sig")
chk2 <- rejection_rates(test_out, collapse = "at_least_one_sig")
chk3 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 1)
chk4 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 2)
chk5 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 3)

expect_equal(chk1$reject,
             chk5$reject)
expect_equal(chk2$reject,
             chk3$reject)
expect_true(chk3$reject >= chk4$reject)
expect_true(chk4$reject >= chk5$reject)

})

skip_on_cran()

library(testthat)

test_that("x_from_power: n region: rejection_rates", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())

expect_no_error(tmp <- n_region_from_power(out,
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection"))

chk1 <- rejection_rates(tmp, all_columns = TRUE)
chk2 <- rejection_rates(tmp$above$power4test_trials, all_columns = TRUE)

expect_equal(chk1$reject,
             chk2$reject)

})

skip_on_cran()

library(testthat)

test_that("x_from_power: n", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

out0 <- out

out <- power4test(out,
                  test_fun = test_parameters,
                  test_args = list(pars = "y~x"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp <- x_from_power(out,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection"))
expect_true(all(grepl("all_sig", tmp$rejection_rates$test_label)))

expect_no_error(tmp <- x_from_power(out,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection",
                    rejection_rates_args = list(collapse = "at_least_one_sig")))
expect_true(all(grepl("one_sig", tmp$rejection_rates$test_label)))

expect_no_error(tmp <- x_from_power(out,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection",
                    rejection_rates_args = list(collapse = "at_least_k_sig",
                                                at_least_k = 2)))
expect_true(all(grepl("at_least_2", tmp$rejection_rates$test_label)))

expect_error(tmp2 <- x_from_power(tmp,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve",
                    rejection_rates_args = list(collapse = "at_least_k_sig",
                                                at_least_k = 1)),
             "rejection_rates_args")

rejection_rates(out)
outby <- power4test_by_n(out,
                         n = c(80, 250, 500),
                         progress = !is_testing())
rejection_rates(outby)

tmp <- x_from_power(outby,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection")
expect_true(all(grepl("all_sig", tmp$rejection_rates$test_label)))

# No merging when there is only one test

expect_no_error(tmp <- x_from_power(out0,
                    x = "n",
                    target_power = .80,
                    final_nrep = 60,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "bisection"))
expect_false(all(grepl("all_sig", tmp$rejection_rates$test_label)))


})

library(testthat)
suppressMessages(library(lavaan))

test_that("summarize tests: merge tests", {

model <-
"
m ~ x + z + x:z
y ~ m + x
"

model_es <-
"
m ~ x:z: s
m ~ x: s
y ~ m: s
y ~ x: s
"

out <- power4test(
            nrep = 10,
            model = model,
            pop_es = model_es,
            n = 100,
            test_fun = test_parameters,
            test_args = list(pars = "m~x"),
            ci_type = "mc",
            R = 100,
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out)

out2 <- power4test(
            out,
            test_fun = test_parameters,
            test_args = list(pars = c("m~z", "y~m")),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out2)

out3 <- power4test(
            out2,
            test_fun = test_index_of_mome,
            test_args = list(x = "x",
                             y = "y",
                             m = "m",
                             w = "z"),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

tmp2 <- collapse_all_tests(out3)

expect_true("pvalue" %in% colnames(tmp2[[1]]$test_results))

tmp3 <- collapse_all_tests(out3,
                           p_adjust_method = "BH")

expect_true("pvalue" %in% colnames(tmp3[[1]]$test_results))
expect_true("pvalue_org" %in% colnames(tmp3[[1]]$test_results))

p_org <- tmp3[[1]]$test_results$pvalue_org
p_adj <- tmp3[[1]]$test_results$pvalue

expect_equal(p.adjust(p_org, method = "BH"),
             p_adj)

tmp <- summarize_tests(out3,
                       merge_all_tests = TRUE,
                       p_adjust_method = "BH")

expect_output(print(tmp),
              "merged")
expect_true(length(tmp) == 1)

tmp2 <- summarize_tests(out3,
                       merge_all_tests = TRUE)

expect_all_true(tmp[[1]]$mean$pvalue >= tmp2[[1]]$mean$pvalue)

tmp <- summarize_tests(out3,
                       collapse = "at_least_k",
                       at_least_k = 1,
                       merge_all_tests = TRUE,
                       p_adjust_method = "BH")
expect_true(nrow(tmp[[1]]$mean) == 1)

expect_output(print(tmp),
              "merged")

tmp2 <- summarize_tests(out3,
                       collapse = "at_least_k",
                       at_least_k = 1,
                       merge_all_tests = TRUE)
tmp <- rejection_rates(out3,
                merge_all_tests = TRUE,
                p_adjust_method = "BH")

expect_output(print(tmp),
              "merged")

tmp2 <- rejection_rates(out3,
                merge_all_tests = TRUE)

expect_output(print(tmp2),
              "merged")

})

skip_on_cran()

library(testthat)

test_that("Boos-Zhang: Set of Rs", {

opt_old <- options(power4mome.bz = TRUE)

mod <-
"
m1 ~ a1*x
m2 ~ a2*x
y ~ b1*m1 + b2*m2 + x
a1b1 := a1 * b1
a2b2 := a2 * b2
"

mod_es <- c("y ~ m1" = "l",
            "m1 ~ x" = "m",
            "y ~ m2" = "s",
            "m2 ~ x" = "l",
            "y ~ x" = "n")
k <- c(y = 3,
       m1 = 3,
       m2 = 3,
       x = 3)
rel <- c(y = .70,
         m1 = .70,
         m2 = .70,
         x = .70)

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 200,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       R = 237,
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = !is_testing())

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        test_method = "pvalue"),
                       parallel = FALSE,
                       progress = !is_testing())
(rr <- rejection_rates(test_ind))
expect_output(print(rr),
              "Boos and Zhang")
(chk <- test_summary(test_ind))
expect_true(any(grepl("bz_", colnames(chk[[1]]))))

# Alpha/level not supported

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        level = .90,
                                        test_method = "pvalue"),
                       parallel = FALSE,
                       progress = !is_testing())
(chk <- test_summary(test_ind))
expect_false(any(grepl("bz_", colnames(chk[[1]]))))

options(opt_old)
})

skip_on_cran()

library(testthat)

test_that("Boos-Zhang: test_indirect_effect", {

opt_old <- options(power4mome.bz = TRUE)

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       R = 119,
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = !is_testing())

test_ind <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        mc_ci = TRUE,
                                        test_method = "pvalue"),
                       progress = !is_testing())

expect_output(print(rejection_rates(test_ind)),
              "Boos and Zhang")
(chk <- test_summary(test_ind))
expect_true("nlt0" %in% names(chk[[1]]))

# Alpha/level not supported

test_ind <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        mc_ci = TRUE,
                                        level = .90,
                                        test_method = "pvalue"),
                       progress = !is_testing())

(rr <- rejection_rates(test_ind))
expect_true(is.null(attr(rr, "extra")$bz_model))

options(opt_old)
})

skip_on_cran()

library(testthat)

test_that("probabilistic bisection: n", {

# ==== n ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  parallel = FALSE,
                  progress = FALSE,
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90,
                          progress = FALSE)

## ==== Close enough ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(700, 1000),
                                  goal = "close_enough",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )

tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 100,
              iseed = 147,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < .80) &&
            (tmp$reject_ci_hi > .80))

## ==== ub ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(500, 700),
                                  goal = "close_enough",
                                  what = "ub",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )
tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 100,
              iseed = 2345,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < a_out$f_power) &&
            (tmp$reject_ci_hi > a_out$f_power))

## ==== lb ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  x_interval = c(800, 1000),
                                  goal = "close_enough",
                                  what = "lb",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )
tmp_out <- power4test(
              out,
              n = a_out$x_out,
              nrep = 100,
              iseed = 2345,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < a_out$f_power) &&
            (tmp$reject_ci_hi > a_out$f_power))

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by es", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       parallel = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE))

power_all_test_only_new_es <- power4test(object = test_out,
                                         pop_es = c("y ~ m" = ".10"))

out <- power4test_by_es(test_out,
                            pop_es_name = "y ~ m",
                            pop_es_values = c(.10, .20),
                            by_seed = 1357,
                            progress = !is_testing())
out_reject <- rejection_rates(out)

out2 <- power4test_by_es(object = sim_only,
                             pop_es_name = "y ~ m",
                             pop_es_values = c(.10, .20),
                             test_fun = test_indirect_effect,
                             test_args = list(x = "x",
                                              m = "m",
                                              y = "y",
                                              boot_ci = TRUE,
                                              mc_ci = FALSE),
                             by_seed = 1357,
                             progress = !is_testing())

out_reject2 <- rejection_rates(out2)
out_reject2

expect_equal(out[[2]]$sim_all[[1]]$mm_lm_dat_out[1, ],
             out2[[2]]$sim_all[[1]]$mm_lm_dat_out[1, ])

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by n", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 3,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       parallel = FALSE,
                       iseed = 1234)

out1 <- power4test_by_n(sim_only,
                        n = c(100, 300),
                        test_fun = test_indirect_effect,
                        test_args = list(x = "x",
                                          m = "m",
                                          y = "y",
                                          boot_ci = TRUE,
                                          mc_ci = FALSE),
                        progress = !is_testing(),
                        by_seed = 1234)

out2 <- power4test_by_n(out1,
                        n = c(50, 150),
                        by_seed = 2345,
                        progress = !is_testing())

out_reject1 <- rejection_rates(out1)
out_reject1
out_reject2 <- rejection_rates(out2)
out_reject2

expect_equal(out_reject1$n,
             c(100, 300))
expect_equal(out_reject2$n,
             c(50, 150))

out <- c(out1, out2)
out_reject <- rejection_rates(out)
expect_equal(out_reject$n,
             c(50, 100, 150, 300))

})

skip_on_cran()

library(testthat)

test_that("x_from_power: n region", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp2 <- n_region_from_power(out,
                    target_power = .75,
                    final_nrep = 100,
                    seed = 234,
                    progress = !is_testing(),
                    simulation_progress = !is_testing()))

expect_output(print(tmp2), "Approximate region")
expect_output(print(summary(tmp2)), "Upper Region")

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: bisection", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 500,
                  ci_type = "mc",
                  test_fun = test_parameters,
                  test_args = list(par = "y~m"),
                  iseed = 1234,
                  parallel = TRUE)
out_power <- rejection_rates(out)
out_power

system.time(
tmp2 <- x_from_power(out,
                     x = "es",
                     pop_es_name = "y ~ m",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     save_sim_all = TRUE,
                     algorithm = "bisection",
                     what = "ub",
                     goal = "close_enough")
)

tmp2

summary(tmp2)

plot(tmp2)


system.time(
tmp3 <- x_from_power(out,
                     x = "es",
                     pop_es_name = "y ~ m",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     save_sim_all = TRUE,
                     algorithm = "bisection",
                     what = "lb",
                     goal = "close_enough")
)

tmp3

summary(tmp3)

plot(tmp3)


})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: bisection", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 500,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
out_power <- rejection_rates(out)
out_power

system.time(
tmp2 <- x_from_power(out,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     save_sim_all = TRUE,
                     algorithm = "bisection",
                     what = "ub",
                     goal = "close_enough")
)

tmp2

summary(tmp2)

plot(tmp2)


system.time(
tmp3 <- x_from_power(out,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     save_sim_all = TRUE,
                     algorithm = "bisection",
                     what = "lb",
                     goal = "close_enough")
)

tmp3

summary(tmp3)

plot(tmp3)


# Case 2

out2 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 1000,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = "m",
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

system.time(
tmp2 <- x_from_power(out2,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     algorithm = "bisection",
                     what = "ub",
                     goal = "close_enough",
                     x_interval = c(100, 2000))
)

tmp2

summary(tmp2)

plot(tmp2)

# Case 3

mod <-
"
m1 ~ x
m2 ~ m1 + x
m3 ~ m2 + m1 + x
y ~ m3 + m2 + m1 + x
"

mod_es <-
"
m1 ~ x: s
m2 ~ m1: s
m3 ~ m2: s
y ~ m3: s
y ~ x: l
"

out3 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 200,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = c("m1", "m2", "m3"),
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

rejection_rates(out3)

system.time(
tmp2 <- x_from_power(out3,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     seed = 1234,
                     algorithm = "bisection",
                     what = "ub",
                     goal = "close_enough")
)

tmp2

summary(tmp2)

plot(tmp2)


})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: es", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 60,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "y~m"),
                  iseed = 1234,
                  parallel = FALSE)
out_power <- rejection_rates(out)
out_power

tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "y ~ m",
                    progress = TRUE,
                    final_nrep = 80,
                    max_trials = 4,
                    seed = 1234,
                    algorithm = "power_curve")
tmp
summary(tmp)
plot(tmp)
plot(tmp,
     lwd = 4,
     col = "green",
     what = c("ci",
              "final_x",
              "final_power"))
plot(tmp,
     lwd = 4,
     col = "darkgreen",
     what = c("ci",
              "final_x",
              "final_power"),
     pars_ci_final_x = list(lwd = 10),
     pars_final_x = list(lwd = 2, col = "red"))

plot(tmp,
     pars_power_curve = list(col = "blue", lwd = 4, type = "o"))

# Case 2

mod_es <-
"
m ~ x: s
y ~ m: .05
y ~ x: s
"
out2 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 60,
                   fit_model_args = list(fit_function = "lm"),
                   test_fun = test_parameters,
                   test_args = list(par = "y~m"),
                   iseed = 1234,
                   parallel = FALSE)

tmp2 <- x_from_power(out2,
                     x = "es",
                     pop_es_name = "y ~ m",
                     progress = TRUE,
                     final_nrep = 80,
                     max_trials = 4,
                     seed = 1234,
                     algorithm = "power_curve")
tmp2
summary(tmp2)
plot(tmp2)

# Case 3

mod <-
"
m1 ~ x
m2 ~ m1 + x
m3 ~ m2 + m1 + x
y ~ m3 + m2 + m1 + x
"

mod_es <-
"
m1 ~ x: m
m2 ~ m1: m
m3 ~ m2: m
y ~ m3: m
y ~ x: l
"

out3 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 150,
                   R = 1000,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = c("m1", "m2", "m3"),
                                    y = "y",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)

rejection_rates(out3)

tmp3 <- x_from_power(out3,
                     x = "es",
                     pop_es_name = "m1 ~ x",
                     progress = TRUE,
                     final_nrep = 80,
                     max_trials = 4,
                     x_interval = c(0, .90),
                     seed = 1234,
                     algorithm = "power_curve")

tmp3
summary(tmp3)
plot(tmp3)


})

skip_on_cran()

library(testthat)

test_that("x_from_power: use by_es", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

out <- power4test_by_es(out,
                        pop_es_name = "y ~ m",
                        pop_es_values = c(.11, .12),
                        by_seed = 2345,
                        progress = !is_testing())

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "y ~ m",
                    target_power = .80,
                    final_nrep = 50,
                    max_trials = 1,
                    seed = 1234,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve",))
expect_true(all(c(.11, .12) %in% tmp$x_tried))

})

library(testthat)

test_that("gen_objective", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  progress = !is_testing(),
                  iseed = 1234)

f0 <- gen_objective(out,
                    x = "n",
                    pop_es_name = NULL,
                    target_power = .80,
                    ci_level = .95,
                    progress = !is_testing(),
                    digits = 3,
                    nrep = 20,
                    R = NULL,
                    what = "point",
                    simulation_progress = !is_testing(),
                    save_sim_all = TRUE,
                    store_output = TRUE)
set.seed(1234)
f_i <- f0(20)
as.numeric(f_i)
output_i <- attr(f_i, "output")
expect_identical(as.numeric(f_i),
                 rejection_rates(output_i)$reject - .80)

set.seed(1234)
f_i <- f0(20,
          what = "lb")
as.numeric(f_i)
output_i <- attr(f_i, "output")
rejection_rates(output_i)$reject_ci_lo
expect_identical(as.numeric(f_i),
                 rejection_rates(output_i)$reject_ci_lo - .80)

set.seed(1234)
f_i <- f0(20,
          what = "ub")
as.numeric(f_i)
output_i <- attr(f_i, "output")
rejection_rates(output_i)$reject_ci_hi
expect_identical(as.numeric(f_i),
                 rejection_rates(output_i)$reject_ci_hi - .80)

# With power_i or output_i

expect_equal(f0(20, out_i = output_i, what = "ub", progress = !is_testing()),
             as.numeric(f_i))
expect_equal(f0(20, power_i = rejection_rates(output_i)$reject, what = "ub", progress = !is_testing()),
             as.numeric(f_i))


})

skip_on_cran()

library(testthat)

test_that("x_from_power: es: power_curve", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "point",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE)))
expect_no_error(capture.output(print(summary(tmp))))
expect_true(abs(tmp$power_final - tmp$target_power) < tmp$technical$tol)

tmp2 <- x_from_power(tmp,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "point",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE))
expect_identical(tmp2,
                 tmp)

tmp3 <- x_from_power(tmp$power4test_trials,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "point",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE))
expect_true(all.equal(tmp3$x_tried,
                      tmp$x_tried))
expect_true(all.equal(tmp3$x_final,
                      tmp$x_final))

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .60,
                    tolerance = .10,
                    final_nrep = 100,
                    max_trials = 2,
                    seed = 23456,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "point",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE)))
expect_no_error(capture.output(print(summary(tmp))))
expect_true(abs(tmp$power_final - tmp$target_power) < tmp$technical$tol)

})

library(testthat)

test_that("All parameters: lm", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "s")

sim_only <- power4test(nrep = 10,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       progress = !is_testing())

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

chk0 <- rejection_rates(test_out, collapse = "all_sig")
chk1 <- rejection_rates(test_out, collapse = "at_least_one_sig")
chk2 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 1)
chk3 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 2)
chk4 <- rejection_rates(test_out, collapse = "at_least_k_sig", at_least_k = 3)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(omnibus = "all_sig"),
                       progress = FALSE)
(chk <- test_summary(test_out))
expect_equal(chk0$reject,
             chk[[1]]$sig)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(omnibus = "at_least_one_sig"),
                       progress = FALSE)
(chk <- test_summary(test_out))
expect_equal(chk1$reject,
             chk[[1]]$sig)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(omnibus = "at_least_k_sig",
                                        at_least_k = 1),
                       progress = FALSE)
(chk <- test_summary(test_out))
expect_equal(chk1$reject,
             chk[[1]]$sig)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(omnibus = "at_least_k_sig",
                                        at_least_k = 2),
                       progress = FALSE)
(chk <- test_summary(test_out))
expect_equal(chk3$reject,
             chk[[1]]$sig)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(omnibus = "at_least_k_sig",
                                        at_least_k = 3),
                       progress = FALSE)
(chk <- test_summary(test_out))
expect_equal(chk4$reject,
             chk[[1]]$sig)
expect_equal(chk0$reject,
             chk[[1]]$sig)


})

skip_on_cran()

library(testthat)
suppressMessages(library(lavaan))

test_that("do_test: Error and warning catching", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 8,
       m = 8,
       x = 8)
rel <- c(y = .30,
         m = .30,

         x = .30)

suppressWarnings(
  out_i <- power4test(
            nrep = 2,
            model = model_simple_med,
            pop_es = model_simple_med_es,
            n = 50,
            number_of_indicators = k,
            R = 100,
            # fit_model_args = list(control = list(iter.max = 100)),
            ci_type = "mc",
            reliability = rel,
            test_fun = test_indirect_effect,
            test_args = list(x = "x",
                             m = "m",
                             y = "y"),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing()
          )
)
# TODO:
# - Find a better way to handle failed replications
# expect_true(rejection_rates(out_i)$nvalid < 2)
expect_true(is.na(rejection_rates(out_i)$pvalid))

suppressWarnings(
  out_i <- power4test(
            nrep = 2,
            model = model_simple_med,
            pop_es = model_simple_med_es,
            n = 50,
            number_of_indicators = k,
            R = 100,
            fit_model_args = list(control = list(iter.max = 100)),
            ci_type = "mc",
            reliability = rel,
            test_fun = test_indirect_effect,
            test_args = list(x = "x",
                             m = "m",
                             y = "y"),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing()
          )
)
# TODO:
# - Find a better way to handle failed replications
# expect_true(rejection_rates(out_i, all_columns = TRUE)$nvalid == 0)
expect_true(is.na(rejection_rates(out_i)$pvalid))

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by n", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 3,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

out <- power4test_by_n(test_out,
                       n = c(100, 110),
                       by_seed = 1234,
                       progress = !is_testing())
out_reject <- rejection_rates(out)
out_reject

out2 <- power4test_by_n(object = sim_only,
                        n = c(100, 110),
                        test_fun = test_indirect_effect,
                        test_args = list(x = "x",
                                         m = "m",
                                         y = "y",
                                         boot_ci = TRUE,
                                         mc_ci = FALSE),
                        by_seed = 1234,
                        progress = !is_testing())
out_reject2 <- rejection_rates(out2)
out_reject2

expect_equal(out[[2]]$sim_all[[1]]$mm_lm_dat_out[1, ],
             out2[[2]]$sim_all[[1]]$mm_lm_dat_out[1, ])

})

library(testthat)

test_that("cond indirect effects: mg: compare groups", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
  - m
y ~ x: nil
"

# Generate the data

dat1 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = c(100, 50, 75),
          R = 100,
          ci_type = "mc",
          progress = !is_testing(),
          iseed = 1234
        )

# Multigroup tests

out1 <- power4test(
          dat1,
          test_fun = test_cond_indirect_effects,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE,
                           compare_groups = TRUE),
          progress = !is_testing(),
          iseed = 1234
        )

tmp1 <- rejection_rates(out1)
expect_true(all(grepl("Group", tmp1$test_label)))

out1b <- power4test(
          dat1,
          test_fun = test_cond_indirect_effects,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE),
          progress = !is_testing(),
          iseed = 1234
        )

tmp1b <- rejection_rates(out1b)

expect_equal(tmp1$est[c(1, 3)],
             diff(tmp1b$est))

})

test_that("cond indirect effects: mg: compare groups: two groups", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m:
  - m
  - sm
m ~ x:
  - s
  - m
y ~ x: nil
"

# Generate the data

dat1 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = c(100, 75),
          R = 100,
          ci_type = "mc",
          progress = !is_testing(),
          iseed = 1234
        )

# Multigroup tests

out1 <- power4test(
          dat1,
          test_fun = test_cond_indirect_effects,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE,
                           compare_groups = TRUE),
          progress = !is_testing(),
          iseed = 1234
        )

tmp1 <- rejection_rates(out1)
expect_true(all(grepl("Group", tmp1$test_label)))

out1b <- power4test(
          dat1,
          test_fun = test_cond_indirect_effects,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE),
          progress = !is_testing(),
          iseed = 1234
        )

tmp1b <- rejection_rates(out1b)

expect_equal(diff(tmp1b$est),
             tmp1$est)

})

library(testthat)

test_that("Get all rejection rates", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 4,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE),
                       progress = !is_testing())
test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       progress = !is_testing())
test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(op = "~"),
                       progress = !is_testing())
test_out <- power4test(object = test_out,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE),
                       progress = !is_testing())

expect_no_error(rejection_rates(test_out))
expect_no_error(rejection_rates(test_out,
                                    all_columns = TRUE))

tmp <- rejection_rates(test_out,
                           ci = FALSE)
expect_false("reject_ci_lo" %in% colnames(tmp))

tmp <- rejection_rates(test_out,
                           se = TRUE)
expect_true("reject_se" %in% colnames(tmp))

tmp <- rejection_rates(test_out,
                           ci = FALSE,
                           se = TRUE)
expect_true("reject_se" %in% colnames(tmp))

tmp <- rejection_rates(test_out,
                           all_columns = TRUE,
                           ci = FALSE)
expect_false("reject_ci_lo" %in% colnames(tmp))

tmp <- rejection_rates(test_out,
                           all_columns = TRUE,
                           se = TRUE)
expect_true("reject_se" %in% colnames(tmp))

tmp <- rejection_rates(test_out,
                           all_columns = TRUE,
                           ci = FALSE,
                           se = TRUE)
expect_true("reject_se" %in% colnames(tmp))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("test_group_equal", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_group_equal,
                       test_args = list(group.equal = c("regressions"),
                                        group.partial = c("y~m")))

tmp <- rejection_rates(test_out)

fit0 <- test_out$sim_all[[2]]$extra$fit
mod0 <- attr(test_out$sim_all[[2]]$ptable, "model")
fit1 <- update(fit0,
          group.equal = "regressions",
          group.partial = "y~m")
lrt_out <- lavTestLRT(fit0, fit1)
test0 <- test_out$test_all[[1]][[2]]$test_results

expect_equal(test0[, "pvalue"],
             lrt_out[2, "Pr(>Chisq)"])

est <- parameterEstimates(fit1, se = FALSE, ci = FALSE)

expect_equal(est[est$lhs == "m" & est$rhs == "x" & est$group == 1, "est"],
             est[est$lhs == "m" & est$rhs == "x" & est$group == 2, "est"])
expect_false(isTRUE(all.equal(
             est[est$lhs == "y" & est$rhs == "m" & est$group == 1, "est"],
             est[est$lhs == "y" & est$rhs == "m" & est$group == 2, "est"]
             )))

})

skip("Not used")

# Do not use these tests. They are "too low level."

skip_on_cran()

library(testthat)

test_that("bisection", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

####### n

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  progress = !is_testing(),
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90,
                          progress = !is_testing())

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   final_nrep = 20,
                                   x_interval = c(200, 2000),
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_false(.80 %in% range(a_out$ci_out))

# Close enough

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(600, 700),
                                   goal = "close_enough",
                                   tol = .07,
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$power_out - .80) < .07)

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(890, 950),
                                   extendInt = "yes",
                                   goal = "close_enough",
                                   tol = .05,
                                   max_trials = 3,
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$power_out - .80) < .05)

# ub

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   what = "ub",
                                   goal = "close_enough",
                                   final_nrep = 20,
                                   x_interval = c(100, 1000),
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$ci_out[2] - .80) < .02)

# lb

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   what = "lb",
                                   goal = "close_enough",
                                   extendInt = "yes",
                                   final_nrep = 20,
                                   tol = .2,
                                   x_interval = c(100, 1000),
                                   max_trials = 3,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$ci_out[1] - .80) < .2)

# Solution already in interval

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(775, 800),
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_false(.80 %in% range(a_out$ci_out))

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(600, 775),
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_false(.80 %in% range(a_out$ci_out))

####### es

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  progress = !is_testing(),
                  iseed = 1234)

by_x_1 <- power4test_by_es(out,
                           pop_es_name = "m~x",
                           pop_es_values = c(.10),
                           progress = !is_testing())

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   x_interval = c(0, .50),
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_false(.80 %in% range(a_out$ci_out))

# Close enough

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   goal = "close_enough",
                                   tol = .10,
                                   x_interval = c(0, .50),
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$power_out - .80) < .10)

# ub

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   what = "ub",
                                   goal = "close_enough",
                                   x_interval = c(0, .50),
                                   tol = .05,
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$ci_out[2] - .80) < .02)

# lb

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   what = "lb",
                                   goal = "close_enough",
                                   x_interval = c(0, .50),
                                   tol = .10,
                                   final_nrep = 20,
                                   progress = !is_testing(),
                                   simulation_progress = !is_testing())
rejection_rates(a_out$by_x_1)
expect_true(abs(a_out$ci_out[1] - .80) < .10)

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("x_from_power: probabilistic bisection", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 199,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
out_power <- rejection_rates(out)
out_power

system.time(
tmp2 <- x_from_power(out,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 1000,
                     final_R = 199,
                     seed = 1234,
                     save_sim_all = FALSE,
                     algorithm = "probabilistic_bisection",
                     what = "point",
                     goal = "close_enough")
)

tmp2

summary(tmp2)

plot(tmp2)


system.time(
tmp3 <- x_from_power(out,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 1000,
                     seed = 1234,
                     save_sim_all = FALSE,
                     algorithm = "probabilistic_bisection",
                     what = "lb",
                     goal = "close_enough")
)

tmp3

summary(tmp3)

plot(tmp3)


# Case 2

out2 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 1000,
                   R = 199,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = "m",
                                    y = "y",
                                    mc_ci = TRUE,
                                    test_method = "pvalue"),
                   iseed = 1234,
                   parallel = TRUE)

system.time(
tmp2 <- x_from_power(out2,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 1000,
                     final_R = 1000,
                     seed = 1234,
                     algorithm = "probabilistic_bisection",
                     what = "ub",
                     goal = "close_enough",
                     x_interval = c(100, 2000))
)

tmp2

summary(tmp2)

plot(tmp2)

# Case 3

mod <-
"
m1 ~ x
m2 ~ m1 + x
m3 ~ m2 + m1 + x
y ~ m3 + m2 + m1 + x
"

mod_es <-
"
m1 ~ x: s
m2 ~ m1: s
m3 ~ m2: s
y ~ m3: s
y ~ x: l
"

out3 <- power4test(nrep = 50,
                   model = mod,
                   pop_es = mod_es,
                   n = 200,
                   R = 199,
                   ci_type = "mc",
                   test_fun = test_indirect_effect,
                   test_args = list(x = "x",
                                    m = c("m1", "m2", "m3"),
                                    y = "y",
                                    mc_ci = TRUE,
                                    test_method = "pvalue"),
                   iseed = 1234,
                   parallel = TRUE)

rejection_rates(out3)

system.time(
tmp2 <- x_from_power(out3,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 1000,
                     final_R = 1000,
                     seed = 1234,
                     algorithm = "probabilistic_bisection",
                     what = "ub",
                     goal = "close_enough")
)

tmp2

summary(tmp2)

plot(tmp2)


})

skip("Graphs: Test interactively")

library(testthat)

test_that("x_from_power: n region: plot", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = FALSE)
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp2 <- n_region_from_power(out,
                    target_power = .75,
                    final_nrep = 100,
                    seed = 234,
                    progress = FALSE,
                    simulation_progress = FALSE))

expect_no_error(plot(tmp2))

})

library(testthat)

test_that("Power by es: MG", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       parallel = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(op = "~"))

rejection_rates(test_out)

out1 <- power4test(
          object = test_out,
          pop_es = c("y ~ m.g2" = .30)
        )

out2 <- power4test(
          object = test_out,
          pop_es = c("y ~ m" = "s")
        )

pt1 <- out1$sim_all[[1]]$ptable
pt2 <- out2$sim_all[[1]]$ptable

expect_equal(pt1$start[(pt1$lhs == "y") & (pt1$op == "~") & (pt1$rhs == "m") & (pt1$group == 2)],
             .3)
expect_equal(pt2$start[(pt2$lhs == "y") & (pt2$op == "~") & (pt2$rhs == "m") & (pt2$group == 1)],
             .1)

out1 <- power4test_by_es(
            test_out,
            pop_es_name = "y ~ m",
            pop_es_values = c(.15, .25),
            by_seed = 1357,
            progress = !is_testing())

pt1 <- out1[[1]]$sim_all[[1]]$ptable
pt2 <- out1[[2]]$sim_all[[1]]$ptable

expect_equal(pt1$start[(pt1$lhs == "y") & (pt1$op == "~") & (pt1$rhs == "m") & (pt1$group == 1)],
             .15)
expect_equal(pt2$start[(pt2$lhs == "y") & (pt2$op == "~") & (pt2$rhs == "m") & (pt2$group == 1)],
             .25)

out2 <- power4test_by_es(
            test_out,
            pop_es_name = "y ~ m.g2",
            pop_es_values = c(.35, .45),
            by_seed = 1357,
            progress = !is_testing())

pt1 <- out2[[1]]$sim_all[[1]]$ptable
pt2 <- out2[[2]]$sim_all[[1]]$ptable

expect_equal(pt1$start[(pt1$lhs == "y") & (pt1$op == "~") & (pt1$rhs == "m") & (pt1$group == 2)],
             .35)
expect_equal(pt2$start[(pt2$lhs == "y") & (pt2$op == "~") & (pt2$rhs == "m") & (pt2$group == 2)],
             .45)

})

skip_on_cran()

library(testthat)

test_that("x_from_power: es", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 30,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

# Failed in the search

tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "y ~ m",
                    target_power = .90,
                    final_nrep = 50,
                    x_interval = c(.0, .4),
                    extendInt = "no",
                    max_trials = 2,
                    seed = 1234,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(is.na(tmp$power_final))
expect_true(is.na(tmp$x_final))

# Can extend the search

tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "y ~ m",
                    target_power = .90,
                    final_nrep = 50,
                    x_interval = c(.0, .4),
                    extendInt = "yes",
                    max_trials = 3,
                    seed = 12,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(!is.na(tmp$power_final))
expect_true(!is.na(tmp$x_final))

# Can extend the search

mod_es <-
"
m ~ x: s
y ~ m: l
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 110,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "y ~ m",
                    target_power = .20,
                    final_nrep = 50,
                    x_interval = c(.4, .6),
                    extendInt = "yes",
                    max_trials = 3,
                    seed = 12,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(!is.na(tmp$power_final))
expect_true(!is.na(tmp$x_final))
})

library(testthat)

test_that("rejection_rates_args", {

model <-
"
m ~ x + z + x:z
y ~ m + x
"

model_es <-
"
m ~ x:z: s
m ~ x: m
y ~ m: m
y ~ x: s
"

out <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es,
            n = 100,
            test_fun = test_parameters,
            test_args = list(pars = "m~x"),
            ci_type = "mc",
            R = 100,
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out)

out2 <- power4test(
            out,
            test_fun = test_parameters,
            test_args = list(pars = c("m~z", "y~m")),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

(tmp <- rejection_rates(out2))

expect_equal(nrow(tmp),
             3)

out3 <- power4test(
            out2,
            test_fun = test_index_of_mome,
            test_args = list(x = "x",
                             y = "y",
                             m = "m",
                             w = "z"),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing(),
            rejection_rates_args = list(merge_all_tests = TRUE,
                                        collapse = "all_sig"))

(tmp <- rejection_rates(out3))

expect_equal(nrow(tmp),
             1)

(tmp <- rejection_rates(out3,
                        merge_all_tests = FALSE))

expect_equal(nrow(tmp),
             3)

(tmp <- rejection_rates(out3,
                        collapse = "none"))

expect_equal(nrow(tmp),
             4)

(tmp <- rejection_rates(out3,
                        collapse = "none",
                        merge_all_tests = FALSE))

expect_equal(nrow(tmp),
             4)

expect_error(out4 <- power4test(
                        out3,
                        rejection_rates_args = list(merge_all_tests = FALSE,
                                                    collapse = "none"))
            )

out <- power4test_by_n(out3,
                       n = c(100, 110),
                       by_seed = 1234,
                       progress = !is_testing())

(tmp <- rejection_rates(out))

expect_equal(tmp$est,
             c(NaN, NaN))

(tmp <- rejection_rates(out,
                        merge_all_tests = FALSE))

expect_equal(nrow(tmp),
             6)

(tmp <- rejection_rates(out,
                        collapse = "none",
                        merge_all_tests = FALSE))

expect_equal(nrow(tmp),
             8)

})

skip_on_cran()

library(testthat)

test_that("Boos-Zhang", {

opt_old <- options(power4mome.bz = TRUE)

mod <-
"
m ~ x + w1 + x:w1
y ~ m + w2 + m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 119,
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = !is_testing())

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w2 = 1, w1 = 0),
                                        mc_ci = TRUE,
                                        test_method = "pvalue"),
                       parallel = FALSE,
                       progress = !is_testing())

(rr <- rejection_rates(test_ind))
(chk <- test_summary(test_ind))
expect_output(print(rr),
              "Boos and Zhang")
expect_true(any(grepl("nlt0", names(chk[[1]]))))

# Alpha/level not supported

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w2 = 1, w1 = 0),
                                        mc_ci = TRUE,
                                        level = .90,
                                        test_method = "pvalue"),
                       parallel = FALSE,
                       progress = !is_testing())

(rr <- rejection_rates(test_ind))
expect_true(is.null(attr(rr, "extra")$bz_model))

options(opt_old)
})

skip_on_cran()

library(testthat)

test_that("probabilistic bisection: es", {

# ==== es ====

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  progress = FALSE,
                  iseed = 1234)

by_x_1 <- power4test_by_es(out,
                           pop_es_name = "m~x",
                           pop_es_values = c(.10),
                           progress = FALSE)

## ==== Close enough ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "m~x",
                                  by_x_1 = by_x_1,
                                  x_interval = c(.20, .30),
                                  goal = "close_enough",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )

tmp_es <- setNames(a_out$x_out, "m~x")
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 100,
              iseed = 147,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < .80) &&
            (tmp$reject_ci_hi > .80))

## ==== ub ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "m~x",
                                  by_x_1 = by_x_1,
                                  x_interval = c(.10, .30),
                                  goal = "close_enough",
                                  what = "ub",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )

tmp_es <- setNames(a_out$x_out, "m~x")
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 100,
              iseed = 147,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < a_out$f_power) &&
            (tmp$reject_ci_hi > a_out$f_power))

## ==== lb ====

set.seed(1234)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "es",
                                  pop_es_name = "m~x",
                                  by_x_1 = by_x_1,
                                  x_interval = c(.20, .40),
                                  goal = "close_enough",
                                  what = "lb",
                                  final_nrep = 100,
                                  progress = FALSE,
                                  max_trials = 50,
                                  variants = list(initial_nrep = 50)
                                )

tmp_es <- setNames(a_out$x_out, "m~x")
tmp_out <- power4test(
              out,
              pop_es = tmp_es,
              nrep = 100,
              iseed = 147,
              progress = FALSE)
tmp <- rejection_rates(tmp_out)
expect_true((tmp$reject_ci_lo < a_out$f_power) &&
            (tmp$reject_ci_hi > a_out$f_power))

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by n: by_seed", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       parallel = FALSE,
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

out1 <- power4test_by_n(test_out,
                        n = c(100, 110),
                        by_seed = 1234,
                        progress = !is_testing())
out1_reject <- rejection_rates(out1)
out1_reject

out2 <- power4test_by_n(test_out,
                        n = c(100, 110),
                        by_seed = 1234,
                        progress = !is_testing())
out2_reject <- rejection_rates(out2)
out2_reject

expect_identical(out1_reject,
                 out2_reject)

})

skip_on_cran()

library(testthat)

test_that("indirect effects", {

mod <-
"
m1 ~ a1*x
m2 ~ a2*x
y ~ b1*m1 + b2*m2 + x
a1b1 := a1 * b1
a2b2 := a2 * b2
"

mod_es <- c("y ~ m1" = "l",
            "m1 ~ x" = "m",
            "y ~ m2" = "s",
            "m2 ~ x" = "l",
            "y ~ x" = "n")
k <- c(y = 3,
       m1 = 3,
       m2 = 3,
       x = 3)
rel <- c(y = .70,
         m1 = .70,
         m2 = .70,
         x = .70)

sim_only <- power4test(nrep = 10,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       R = 200,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 2345)

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE))

(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

chk0 <- rejection_rates(test_ind, collapse = "all_sig")
chk1 <- rejection_rates(test_ind, collapse = "at_least_one_sig")

test_indb <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        test_method = "pvalue"))
rejection_rates(test_indb)

(chkb <- test_summary(test_indb))
expect_equal(chk[[1]]$sig[1],
             chkb[[1]]$sig[1])

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        omnibus = "all_sig",
                                        mc_ci = TRUE))

test_ind$test_all[[1]][[1]][[1]]
(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

expect_equal(chk0$reject,
             chk[[1]]$sig)

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        omnibus = "at_least_one_sig",
                                        mc_ci = TRUE))
test_ind$test_all[[1]][[1]][[1]]
(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)
expect_equal(chk1$reject,
             chk[[1]]$sig)

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        omnibus = "at_least_k_sig",
                                        at_least_k = 2,
                                        mc_ci = TRUE))
test_ind$test_all[[1]][[1]][[1]]
(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)
expect_equal(chk0$reject,
             chk[[1]]$sig)

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        omnibus = "at_least_k_sig",
                                        at_least_k = 1,
                                        mc_ci = TRUE))
test_ind$test_all[[1]][[1]][[1]]
(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)
expect_equal(chk1$reject,
             chk[[1]]$sig)


})

skip_on_cran()

library(testthat)

test_that("x_from_power: n", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 30,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

# Failed in the search

tmp <- x_from_power(out,
                    x = "n",
                    target_power = .90,
                    final_nrep = 50,
                    x_interval = c(10, 35),
                    extendInt = "no",
                    max_trials = 2,
                    seed = 1234,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(is.na(tmp$power_final))
expect_true(is.na(tmp$x_final))

# Can extend the search

tmp <- x_from_power(out,
                    x = "n",
                    target_power = .90,
                    final_nrep = 50,
                    x_interval = c(10, 60),
                    extendInt = "yes",
                    max_trials = 3,
                    seed = 12,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(!is.na(tmp$power_final))
expect_true(!is.na(tmp$x_final))

# Can extend the search

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 110,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

tmp <- x_from_power(out,
                    x = "n",
                    target_power = .50,
                    final_nrep = 20,
                    x_interval = c(100, 120),
                    extendInt = "yes",
                    max_trials = 2,
                    seed = 12345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    algorithm = "power_curve")
expect_no_error(capture.output(print(summary(tmp))))
expect_true(!is.na(tmp$power_final))
expect_true(!is.na(tmp$x_final))
})

skip_on_cran()

library(testthat)

test_that("Boos-Zhang", {

opt_old <- options(power4mome.bz = TRUE)

mod <-
"
m ~ x + w + x:w
y ~ m
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w" = "n",
            "m ~ x:w" = "l")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 119,
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = !is_testing())

test_out1 <- power4test(object = sim_only,
                         test_fun = test_index_of_mome,
                         test_args = list(x = "x",
                                          m = "m",
                                          y = "y",
                                          w = "w",
                                          mc_ci = TRUE,
                                          test_method = "pvalue"),
                         parallel = FALSE,
                         progress = !is_testing())

(rr <- rejection_rates(test_out1))
(chk <- test_summary(test_out1))
expect_output(print(rr),
              "Boos and Zhang")
expect_true("nlt0" %in% names(chk[[1]]))

# Alpha/level not supported

test_out1 <- power4test(object = sim_only,
                         test_fun = test_index_of_mome,
                         test_args = list(x = "x",
                                          m = "m",
                                          y = "y",
                                          w = "w",
                                          mc_ci = TRUE,
                                          level = .90,
                                          test_method = "pvalue"),
                         parallel = FALSE,
                         progress = !is_testing())

(rr <- rejection_rates(test_out1))
expect_true(is.null(attr(rr, "extra")$bz_model))

options(opt_old)
})

skip_on_cran()

library(testthat)

test_that("power4test: mg", {

model_simple_med <-
"
m ~ c(a1, a2) * x
y ~ m + x
a := a1 - a2
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - l
  - s
y ~ x: nil
"

# Generate the data

out1 <- power4test(
          nrep = 50,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = 200,
          test_fun = test_parameters,
          test_args = list(par = c("a")),
          progress = !is_testing(),
          iseed = 1234,
          n_ratio = c(1, .5)
        )

rejection_rates(out1)

expect_no_error(tmp2 <- n_region_from_power(out1,
                    target_power = .75,
                    final_nrep = 100,
                    seed = 234,
                    progress = !is_testing(),
                    simulation_progress = !is_testing()))

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("bisection: n", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

####### n

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90)

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1)
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# Close enough

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(600, 700),
                                   goal = "close_enough",
                                   tol = .05)
a_out$solution_found
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(600, 700),
                                   extendInt = "yes",
                                   goal = "close_enough",
                                   tol = .05)
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)


# ub

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   what = "ub",
                                   goal = "close_enough")
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(v = a_out$x_out)
abline(h = .80)

# lb

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   what = "lb",
                                   goal = "close_enough")
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# Solution already in interval

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(775, 800))
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "n",
                                   by_x_1 = by_x_1,
                                   x_interval = c(600, 775))
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)


####### es

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  iseed = 1234)

by_x_1 <- power4test_by_es(out,
                           pop_es_name = "m~x",
                           pop_es_values = c(.10))

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1)
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# Close enough

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   goal = "close_enough",
                                   tol = .30)
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# ub

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   what = "ub",
                                   goal = "close_enough")
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)


# lb

set.seed(1234)
a_out <- power_algorithm_bisection(object = out,
                                   x = "es",
                                   pop_es_name = "m~x",
                                   by_x_1 = by_x_1,
                                   what = "lb",
                                   goal = "close_enough")
rejection_rates(a_out$by_x_1)
plot(a_out$fit_1)
abline(h = .80)

# Solution already in interval

})

library(testthat)

test_that("cond indirect effects: mg", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

# Generate the data

dat1 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = c(100, 50),
          R = 100,
          ci_type = "mc",
          progress = !is_testing(),
          iseed = 1234
        )

# Multigroup tests

out1 <- power4test(
          dat1,
          test_fun = test_cond_indirect_effects,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE),
          progress = !is_testing(),
          iseed = 1234
        )

tmp1 <- rejection_rates(out1)
expect_true(all(grepl("Group", tmp1$test_label)))

# One single conditional indirect effects

out2a <- power4test(
          dat1,
          test_fun = test_cond_indirect,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE,
                           group = "Group1"),
          progress = !is_testing(),
          iseed = 1234
        )

tmp2a <- rejection_rates(out2a)

out2b <- power4test(
          dat1,
          test_fun = test_cond_indirect,
          test_args = list(x = "x",
                           m = "m",
                           y = "y",
                           mc_ci = TRUE,
                           group = "Group2"),
          progress = !is_testing(),
          iseed = 1234
        )

tmp2b <- rejection_rates(out2b)

expect_true(all.equal(tmp1$est,
                      c(tmp2a$est, tmp2b$est)))

})

skip("Internal")

library(testthat)

test_that("bisection: difficult cases", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: m
y ~ m: l
y ~ x: nil
"

out <- power4test(nrep = 2,
                  model = mod,
                  pop_es = mod_es,
                  n = 50000,
                  iseed = 1234)

out <- power4test(nrep = 100,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 1000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)

out_es_a <- x_from_power(out,
                         x = "es",
                         pop_es_name = "m~x",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         algorithm = "power_curve",
                         goal = "ci_hit")
summary(out_es_a)
plot(out_es_a)

out_es_b <- x_from_power(out,
                         x = "es",
                         pop_es_name = "y~m",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         algorithm = "power_curve",
                         goal = "ci_hit")

summary(out_es_b)

out_es_b <- x_from_power(out,
                         x = "es",
                         pop_es_name = "y~m",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         algorithm = "power_curve",
                         goal = "ci_hit")

summary(out_es_b)


out_es_ab <- x_from_power(out,
                         x = "es",
                         pop_es_name = ".ind.(x->m->y)",
                         target_power = .80,
                         final_nrep = 400,
                         algorithm = "power_curve",
                         goal = "ci_hit",
                         seed = 4567)

summary(out_es_ab)

out_by_es_ab <- power4test_by_es(out,
                                 pop_es_name = ".ind.(x->m->y)",
                                 pop_es_values = seq(.05, .15, .01))
rejection_rates(out_by_es_ab)

# Close enough


mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: m
y ~ m: l
y ~ x: nil
"

out <- power4test(nrep = 2,
                  model = mod,
                  pop_es = mod_es,
                  n = 50000,
                  iseed = 1234)

out <- power4test(nrep = 100,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 1000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)


out_es_a <- x_from_power(out,
                         x = "es",
                         pop_es_name = "m~x",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         what = "ub",
                         goal = "close_enough")

summary(out_es_a)
plot(out_es_a)

# Try Muller

out_es_a <- x_from_power(out,
                         x = "es",
                         pop_es_name = "m~x",
                         target_power = .75,
                         final_nrep = 400,
                         seed = 2345,
                         what = "ub",
                         goal = "close_enough",
                         control = list(muller = TRUE))

summary(out_es_a)
plot(out_es_a)

out_es_b <- x_from_power(out,
                         x = "es",
                         pop_es_name = "m~x",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         what = "lb",
                         goal = "close_enough")

summary(out_es_b)
plot(out_es_b)

# Try Muller

out_es_b <- x_from_power(out,
                         x = "es",
                         pop_es_name = "m~x",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         what = "lb",
                         goal = "close_enough",
                         control = list(muller = TRUE))


summary(out_es_b)
plot(out_es_b)


mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: s
y ~ x: nil
"

out <- power4test(nrep = 2,
                  model = mod,
                  pop_es = mod_es,
                  n = 50000,
                  iseed = 1234)

out <- power4test(nrep = 100,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 1000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)

out_es_b <- x_from_power(out,
                         x = "n",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         what = "lb",
                         goal = "close_enough",
                         control = list(muller = TRUE))

out_es_b <- x_from_power(out,
                         x = "n",
                         target_power = .80,
                         final_nrep = 400,
                         seed = 4567,
                         what = "lb",
                         goal = "close_enough")


summary(out_es_b)
plot(out_es_b)

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by es: by_seed", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       parallel = FALSE,
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE))

power_all_test_only_new_es <- power4test(object = test_out,
                                         pop_es = c("y ~ m" = ".10"))

out1 <- power4test_by_es(test_out,
                             pop_es_name = "y ~ m",
                             pop_es_values = c(.10, .20),
                             by_seed = 1234,
                             progress = !is_testing())
out1_reject <- rejection_rates(out1)

out2 <- power4test_by_es(test_out,
                             pop_es_name = "y ~ m",
                             pop_es_values = c(.10, .20),
                             by_seed = 1234,
                             progress = !is_testing())
out2_reject <- rejection_rates(out2)

expect_identical(out1_reject,
                 out2_reject)

})

skip_on_cran()

library(testthat)

test_that("x_from_power: es: power_curve", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !is_testing())
out_power <- rejection_rates(out)
out_power

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "ub",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE)))
expect_no_error(capture.output(print(summary(tmp))))
expect_true(abs(tmp$ci_final[2] - tmp$target_power) < tmp$technical$tol)

tmp2 <- x_from_power(tmp,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "ub",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE))
expect_identical(tmp2,
                 tmp)

tmp3 <- x_from_power(tmp$power4test_trials,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .70,
                    tolerance = .10,
                    final_nrep = 50,
                    max_trials = 2,
                    seed = 2345,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "ub",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE))
expect_true(all.equal(tmp3$x_tried,
                      tmp$x_tried))
expect_true(all.equal(tmp3$x_final,
                      tmp$x_final))

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .60,
                    tolerance = .10,
                    final_nrep = 100,
                    max_trials = 2,
                    seed = 23456,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "ub",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE)))
expect_no_error(capture.output(print(summary(tmp))))
expect_true(abs(tmp$ci_final[2] - tmp$target_power) < tmp$technical$tol)

expect_no_error(tmp <- x_from_power(out,
                    x = "es",
                    pop_es_name = "m ~ x",
                    target_power = .50,
                    tolerance = .10,
                    final_nrep = 100,
                    max_trials = 2,
                    seed = 23456,
                    progress = !is_testing(),
                    simulation_progress = !is_testing(),
                    what = "lb",
                    goal = "close_enough",
                    algorithm = "power_curve",
                    internal_args = list(keep_algorithm = TRUE)))
expect_no_error(capture.output(print(summary(tmp))))
expect_true(abs(tmp$ci_final[1] - tmp$target_power) < tmp$technical$tol)


})

#' @title Power By Effect Sizes
#'
#' @description Estimate power for a
#' set of effect sizes (population
#' values of a model parameter).
#'
#' @details
#' The function [power4test_by_es()]
#' regenerates
#' datasets for a set of effect sizes
#' (population values of a model parmeter)
#' and does the stored tests in each of
#' them.
#'
#' Optionally, it can also be run
#' on a object with no stored tests.
#' In this case, additional arguments
#' must be set to instruct [power4test()]
#' on the tests to be conducted.
#'
#' It is usually used to examine the
#' power over a sets of effect sizes
#' (population values).
#'
#' @return
#' The function
#' [power4test_by_es()] returns a
#' `power4test_by_es` object, which is a
#' list of `power4test` objects, one for
#' each population value of the parameter.
#'
#' @inheritParams power4test_by_n
#'
#' @param object A `power4test` object,
#' or a `power4test_by_es` object.
#' If it is a `power4test_by_es` object,
#' the first element, which is a
#' `power4test` object, will be used
#' as the value of this argument.
#'
#' @param pop_es_name The name of the
#' parameter. See the help page
#' of [ptable_pop()] on the names for
#' the argument `pop_es`.
#'
#' @param pop_es_values A numeric
#' vector of the population values
#' of the parameter specified in
#' `pop_es_names`.
#'
# @param progress <- Inherited
#'
#' @param ... For [power4test_by_es()],
#' they are arguments to be passed
#' to [power4test()]. For [c.power4test_by_es()],
#' they are [power4test_by_es()] outputs
#' to be combined together.
#' For the `print` method of the output
#' of [power4test_by_es()], they are
#' arguments to be passed to the
#' `print` method of the output of
#' [power4test()] ([print.power4test()]).
#'
# @param by_seed <- Inherited
#'
#' @param by_nrep If set to a number,
#' it will be used to generate the
#' number of replications (`nrep`) for
#' each call to [power4test()].
#' If set to a numeric vector of the
#' same length as `pop_es_values`, then these are
#' the `nrep` values for each of the
#' calls, allowing for different numbers
#' of replications for the population values.
#' If `NULL`, the default, then the
#' original `nrep` will be used.
#' This argument is used by
#' [x_from_power()] for efficiency, and
#' is rarely used when calling
#' this function directly.
#'
# @param save_sim_all <- Inherited
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_med_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 100,
#'                        R = 40,
#'                        ci_type = "boot",
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_indirect_effect,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         boot_ci = TRUE,
#'                                         mc_ci = FALSE))
#'
#' out <- power4test_by_es(test_out,
#'                             pop_es_name = "y ~ m",
#'                             pop_es_values = c(.10, .20))
#' out_reject <- rejection_rates(out)
#' out_reject
#'
#' @export
power4test_by_es <- function(object,
                             pop_es_name = NULL,
                             pop_es_values = NULL,
                             progress = TRUE,
                             ...,
                             by_seed = NULL,
                             by_nrep = NULL,
                             save_sim_all = TRUE) {
  if (!inherits(object, "power4test") &&
      !inherits(object, "power4test_by_es")) {
    stop("Only support 'power4test' or 'power4test_by_es' objects.")
  }
  # Use the first object if the object is a power4test_by_es object:
  if (inherits(object, "power4test_by_es")) {
    object <- object[[1]]
  }
  # if (!is.null(object$sim_all[[1]]$group_name)) {
  #   stop("Does not support multigroup models for now.")
  # }
  if (is.null(pop_es_name) || is.null(pop_es_values)) {
    stop("Both 'pop_es_name' and 'pop_es_values' must be set.")
  }
  out <- list()

  if (length(by_seed) == length(pop_es_values)) {
    seeds <- by_seed
  } else {
    if (!is.null(by_seed)) {
      set.seed(by_seed)
    }
    seeds <- sample.int(99999999,
                        size = length(pop_es_values))
  }

  if (length(by_nrep) == length(pop_es_values)) {
    new_nrep <- by_nrep
  } else {
    if (!is.null(by_nrep)) {
      new_nrep <- rep(by_nrep, length(pop_es_values))
    } else {
      new_nrep <- NULL
    }
  }
  # TODO
  # - Think about to handle MG models,
  #   for which pop_values can be vectors.
  for (i in seq_along(pop_es_values)) {
    x <- pop_es_values[i]
    tmp <- ifelse(x < .001,
                  sprintf("%5.3e", x),
                  formatC(x, digits = 3, format = "f"))
    p_name <- paste0(pop_es_name,
                     " = ",
                     tmp)
    if (progress) {
      cat("\nUpdating the simulation for new value:",
          p_name,
          "\n")
    }
    tmp <- as.character(x)
    names(tmp) <- pop_es_name
    if (is.null(new_nrep)) {
      tmp_out <- tryCatch(power4test(
                            object = object,
                            pop_es = tmp,
                            progress = progress,
                            iseed = seeds[i],
                            ...,
                          ),
                          error = function(e) e)
    } else {
      tmp_out <- tryCatch(power4test(
                            object = object,
                            pop_es = tmp,
                            nrep = new_nrep[i],
                            progress = progress,
                            iseed = seeds[i],
                            ...
                          ),
                          error = function(e) e)
    }

    if (inherits(tmp_out, "error")) {

      if (progress) {
        msg <- "Error in estimating power for this value:"
        msg <- c(msg, paste0("- ", names(tmp), ": ", tmp))
        msg <- c(msg, "The error message:")
        msg <- c(msg, paste0("- ", tmp_out$message))
        msg <- c(msg, "This value is skipped.")
        cat(c("", msg, ""), sep = "\n")
      }

      next
    }

    if (!save_sim_all) {
      # Keep the first element
      tmp_out$sim_all <- tmp_out$sim_all[1]
    }
    out[[p_name]] <- tmp_out
    attr(out[[p_name]], "pop_es_name") <- pop_es_name
    attr(out[[p_name]], "pop_es_value") <- x
  }
  class(out) <- c("power4test_by_es", class(out))
  out
}


#' @rdname power4test_by_es
#'
#' @param sort WHen combining the
#' objects, whether they will be sorted
#' by population values. Default is `TRUE`.
#'
# @param skip_checking_models <- Inherited
#'
#' @return
#' The method [c.power4test_by_es()] returns
#' a `power4test_by_es` object with
#' all the elements (tests for different
#' values of `pop_es_values`) combined.
#'
#' @details
#' The `c` method of `power4test_by_es`
#' objects
#' is used to combine tests from different
#' runs of [power4test_by_es()].
#' @export
c.power4test_by_es <- function(...,
                               sort = TRUE,
                               skip_checking_models = FALSE) {
  # Check whether they have the same ptable
  tmp <- list(...)
  if ((length(tmp) > 1) && !skip_checking_models) {
    ptables <- lapply(tmp,
                      \(x) {x[[1]]$sim_all[[1]]$ptable})
    for (i in seq_along(ptables)[-1]) {
      chk <- identical(ptables[[i]][, c("lhs", "op", "rhs", "group", "user", "block")],
                       ptables[[i - 1]][, c("lhs", "op", "rhs", "group", "user", "block")])
      if (!chk) {
        stop("Not all objects are based on the same model.")
      }
    }
  }
  out <- NextMethod()
  out["sort"] <- NULL
  out["skip_checking_models"] <- NULL

  all_pop_es_name <- sapply(out,
                            \(x) {attr(x, "pop_es_name")})
  all_pop_es_name <- unique(all_pop_es_name)
  if (length(all_pop_es_name) != 1) {
    stop("Not all objects have the same parameters changes: ",
         paste0(all_pop_es_name, collapse = ","))
  }

  all_pop_es_values <- sapply(out,
                              \(x) {attr(x, "pop_es_value")})
  class(out) <- c("power4test_by_es", class(out))
  if (!sort) {
    return(out)
  }
  i <- order(all_pop_es_values)
  out <- out[i]
  class(out) <- c("power4test_by_es", class(out))
  return(out)
}

#' @param object_by_es A `power4test_by_es`
#' object, which is an output of
#' [power4test_by_es()].
#'
#' @param all_columns If `TRUE`, all
#' columns stored by a test will be
#' printed. Default is `FALSE` and
#' only essential columns related to
#' power will be printed.
#'
#' @return
#' The function
#' returns a data frame which is
#' similar to the output of
#' [rejection_rates()], with
#' columns added for the effect size (`pop_es_name` and
#' `pop_es_values`)
#' for each test.
#'
#' @description
#' The function
#' is used to extract the rejection
#' rates form a `power4test_by_es`
#' object, with effect sizes added to
#' the output.
#'
#' @noRd
rejection_rates_by_es <- function(object_by_es,
                                  all_columns = FALSE,
                                  nrep_if_diff = TRUE,
                                  ...) {
  tmpfct <- function(x) {
    out_i <- rejection_rates(x,
                             all_columns = all_columns,
                             keep_nrep = nrep_if_diff,
                             ...)
    pn <- attr(x, "pop_es_name")
    pv <- attr(x, "pop_es_value")
    out_i <- data.frame(par = pn,
                        es = pv,
                        out_i)
    out_i
  }
  out <- mapply(tmpfct,
                x = object_by_es,
                SIMPLIFY = FALSE)
  # Keep common columns
  cnames <- lapply(out,
                   \(x) tryCatch(colnames(x),
                                 error = function(e) names(x)))
  cnames1 <- Reduce(intersect,
                    cnames)
  out <- lapply(out,
                \(x) tryCatch(x[, cnames1],
                              error = function(e) x[cnames1]))
  out <- do.call(rbind,
                 out)
  if (nrep_if_diff &&
      !all_columns) {
    nrep_range <- try(suppressWarnings(diff(range(out$nrep, na.rm = TRUE))), silent = TRUE)
    if (isFALSE(nrep_range > 1)) {
      out$nrep <- NULL
    }
  }
  rownames(out) <- NULL
  out
}

#' @rdname power4test_by_es
#'
#' @param original_object The object
#' to be converted to a `power4test_by_es`
#' object.
#'
#' @return
#' The function [as.power4test_by_es()] returns
#' a `power4test_by_es` object converted
#' from the input object.
#'
#' @details
#' The function [as.power4test_by_es()]
#' is used to convert a `power4test`
#' object to a `power4test_by_es`
#' object, if it is not already one.
#' Useful when concatenating
#' `power4test` objects with
#' `power4test_by_es` objects.
#' @export
as.power4test_by_es <- function(original_object,
                                pop_es_name) {
  if (inherits(original_object, "power4test")) {
    x0 <- pop_es(original_object,
                 pop_es_name = pop_es_name)
    original_object <- list(original_object)
    class(original_object) <- c("power4test_by_es", class(original_object))
    names(original_object) <- paste0(pop_es_name,
                                     " = ",
                                     as.character(x0))
    attr(original_object[[1]], "pop_es_name") <- pop_es_name
    attr(original_object[[1]], "pop_es_value") <- x0
    return(original_object)
  }
  if (inherits(original_object, "power4test_by_es")) {
    return(original_object)
  }
  stop("original_object not of a supported class.")
}


#' @rdname power4test_by_es
#'
#' @param x The object
#' to be printed.
#'
#' @param digits The numbers of digits
#' displayed after the decimal.
#'
#' @param print_all If `TRUE`, all
#' elements in `x`, that is, the results
#' of all sample sizes examined, will
#' be printed. If `FALSE`, then only
#' those of the first value of the
#' parameter
#' will be printed.
#'
#' @return
#' The `print`-method of `power4test_by_es`
#' objects returns the object invisibly.
#' It is called for its side-effect.
#'
#' @export
print.power4test_by_es <- function(
    x,
    print_all = FALSE,
    digits = 3,
    ...
  ) {
  pop_es_name <- attr(x[[1]], "pop_es_name")
  x_tried <- sapply(
                x,
                function(xx) {attr(xx, "pop_es_value")}
              )
  x_tried_str <- formatC(
                      x_tried,
                      digits = digits,
                      format = "f"
                    )
  x_tried_all_str <- paste0(
                        x_tried_str,
                        collapse = ", "
                      )
  catwrap(
      c(paste("The parameter examined:", pop_es_name),
        paste("The value(s) examined:", x_tried_all_str))
      )

  if (print_all) {
    cat("\nThe output for each `power4test` analysis:\n")
    ii <- seq_along(x)
  } else {
    cat("\nThe output of the first element:\n")
    ii <- 1
  }

  for (i in ii) {
    x_i <- attr(x[[i]], "pop_es_value")
    x_i_str <- formatC(
                  x_i,
                  digits = digits,
                  format = "f"
                )
    xx <- x[[i]]
    class(xx) <- "power4test"
    cat(header_str("",
                  hw = .84,
                  sym = "-",
                  prefix = "\n",
                  suffix = "\n",
                  sep_i = ""))
    cat(header_str(paste0("Parameter: ", pop_es_name),
                  hw = .8,
                  sym = "-",
                  suffix = "\n"))
    cat(header_str(paste0("Value: ", x_i_str),
                  hw = .8,
                  sym = "-",
                  suffix = "\n"))
    cat(header_str("",
                  hw = .84,
                  sym = "-",
                  suffix = "\n",
                  sep_i = ""))
    print(xx,
          digits = digits,
          ...)
  }
  if (!print_all) {
    cat("\n")
    catwrap(
        "Print with 'print_all = TRUE' to print all elements."
        )
  }
  invisible(x)
}

skip_on_cran()

library(testthat)

test_that("Power by es", {

mod <-
"
m ~ x
y ~ m + x
"
mod_es <-
"
y ~ x: s
.ind.x->m->y: .09
"
out <- power4test(nrep = 5,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 1000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  progress = !is_testing(),
                  parallel = FALSE)
tmp <- power4test_by_es(out,
                        pop_es_name = ".ind.x->m->y",
                        pop_es_values = seq(.75, .90, .05),
                        by_nrep = 10,
                        progress = !is_testing())
expect_true(length(tmp) == 2)
})

#' @title Test All Free Parameters
#'
#' @description Test all free parameters,
#' including user-defined parameters,
#' for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing all
#' free and user-defined model
#' parameters, by
#' setting it to the `test_fun`
#' argument.
#'
#' For models fitted by `lavaan`,
#' it uses [lavaan::parameterEstimates()]
#' to do the test. If bootstrapping was
#' requested (by setting `se = "boot"`),
#' then it supports bootstrap
#' confidence intervals returned by
#' [lavaan::parameterEstimates()].
#'
#' It has preliminary, though limited,
#' supported for models fitted by
#' [stats::lm()] (through
#' [lmhelprs::many_lm()]). Tests are
#' conducted by ordinary least squares
#' confidence intervals based on
#' the *t* statistic, reported by
#' [stats::confint()] applied to
#' the output of [stats::lm()].
#'
#' # Find the names of parameters
#'
#' To use the argument `pars`, the
#' names as appeared in the function
#' [coef()] must be used. For the
#' output of `lavaan`, this can
#' usually be inferred from the
#' parameter syntax (e.g., `y~x`,
#' no space). If not sure, call
#' [coef()] on the output of `lavaan`.
#' If a parameter is labelled, then
#' the label should be used in `par`.
#'
#' If not sure, the function
#' [find_par_names()] can be used to
#' find valid names.
#'
#' @return
#' In its normal usage, it returns
#' the output returned by
#' [lavaan::parameterEstimates()]
#' or [lmhelprs::lm_list_to_partable()],
#' with the following modifications:
#'
#' - `est`: The parameter estimates,
#'  even if standardized estimates
#'  are requested (not `est.std`).
#'
#' - `cilo` and `cihi`: The
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively (not
#'  `ci.lower` and `ci.upper`).
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' - `test_label`: A column of labels
#'  generated by
#'  [lavaan::lav_partable_labels()],
#'  which are usually the labels used by
#'  `coef()` to label the parameters.
#'
#' @inheritParams test_k_indirect_effects
#'
#' @inheritParams test_indirect_effect
#'
#' @param fit The fit object, to be
#' passed to [lavaan::parameterEstimates()],
#' [lavaan::standardizedSolution()],
#' or [lmhelprs::lm_list_to_partable()].
#'
#' @param standardized Logical. If `TRUE`,
#' [lavaan::standardizedSolution()] will
#' be used. Can be used only with models
#' fitted by `lavaan`.
#'
#' @param pars Optional. If set to
#' a character vector, only parameters
#' with `test_label` equal to values in
#' `pars` will be returned. See the
#' help page
#' on valid names.
#'
#' @param op Optional. If set to a
#' character vector, only parameters with
#' operators (e.g., `"~"`, `"=~"`) will
#' be returned. If both `pars` and `op`
#' are specified, only parameters meeting
#' *both* requirements will be returned.
#'
#' @param exclude_var Logical. If `TRUE`,
#' exclude variances and error variances
#' from the test.
#'
#' @param compare_groups Logical. If `TRUE`,
#' the likelihood ratio test (by [lavaan::lavTestLRT()])
#' will be used to test the pairwise-equality
#' constraints for all selected free
#' parameters that appear in all groups.
#' These tests will be reported instead of
#' the tests for individual parameters.
#'
#' @param remove.nonfree Logical. If
#' `TRUE`, the default, only free
#' parameters will be returned. Ignored
#' if `standardized` is `TRUE` or
#' if the model is not fitted by
#' `lavaan`.
#'
#' @param ... Additional arguments to
#' be passed to [lavaan::parameterEstimates()],
#' [lavaan::standardizedSolution()],
#' or [lmhelprs::lm_list_to_partable()].
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the tests in each replication
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters)
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#' # Do the tests in each replication: Standardized solution
#' # Delta method SEs will be used to do the tests
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters,
#'                        test_args = list(standardized = TRUE))
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#' # Do the tests in each replication: Parameters with the selected operator
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters,
#'                        test_args = list(op = "~"))
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#' @export
test_parameters <- function(fit = fit,
                            standardized = FALSE,
                            pars = NULL,
                            op = NULL,
                            remove.nonfree = TRUE,
                            check_post_check = TRUE,
                            exclude_var = FALSE,
                            compare_groups = FALSE,
                            ...,
                            omnibus = c("no", "all_sig", "at_least_one_sig", "at_least_k_sig"),
                            at_least_k = 1,
                            p_adjust_method = "none",
                            fit_name = "fit",
                            get_map_names = FALSE,
                            get_test_name = FALSE) {
  omnibus <- match.arg(omnibus)
  map_names <- c(fit = fit_name)
  args <- list(...)
  if (get_map_names) {

    # ==== Return map_names ====

    return(map_names)
  }
  if (get_test_name) {

    # ==== Prepare test_name ====

    tmp <- character(0)
    if (!is.null(pars)) {
      tmp0 <- paste0("pars: ",
                     paste0(pars,
                            collapse = ","))
      tmp <- c(tmp,
               tmp0)
    }
    if (!is.null(op)) {
      tmp0 <- paste0("op: ",
                     paste0(op,
                            collapse = ","))
      tmp <- c(tmp,
               tmp0)
    }
    if (length(tmp) >= 1) {
      tmp <- paste0("(",
                    paste0(tmp,
                           collapse = "; "),
                    ")")
    } else {
      tmp <- character(0)
    }

    # ==== Return test_name ====

    if (standardized) {
      return(paste("test_parameters: CIs (standardized)", tmp))
    } else {
      return(paste("test_parameters: CIs", tmp))
    }
  }

  # ==== Check the type of fit ====

  if (inherits(fit, "lm_list")) {
    fit_type <- "lm_list"
  } else if (inherits(fit, "lavaan")) {
    fit_type <- "lavaan"
  } else {
    stop("fit is not a supported object.")
  }

  if (fit_type == "lm_list") {
    if (is.null(op) && is.null(pars)) {
      op <- "~"
    }
  }

  # ==== Is the lavaan fit OK? ====

  if (inherits(fit, "lavaan")) {
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
  } else {
    fit_ok <- TRUE
  }
  if (standardized) {

    # ==== Standardized solution ====

    if (fit_type != "lavaan") {
      stop('Standardized solution supported only for `lavaan` output.')
    }
    est <- lavaan::standardizedSolution(object = fit,
                                        pvalue = TRUE,
                                        ci = TRUE)
    if (!fit_ok) {
      est$est.std <- as.numeric(NA)
      est$ci.lower <- as.numeric(NA)
      est$ci.upper <- as.numeric(NA)
      est$pvalue <- as.numeric(NA)
      est$se <- as.numeric(NA)
    }
  } else {

    # ==== Raw solution ====

    if (fit_type == "lm_list") {

       # ==== lm_list ====

      # TODO:
      # - Find a better way to handle level
      if (!is.null(args$level)) {
        ci_args <- list(level = args$level)
      } else {
        ci_args <- eval(formals(lmhelprs::lm_list_to_partable)$ci_args)
      }
      est <- lmhelprs::lm_list_to_partable(object = fit,
                                           ci = TRUE,
                                           ci_args = ci_args)
      est <- est[, c("lhs", "op", "rhs", "est", "se", "pvalue",
                     "ci.lower", "ci.upper")]
    } else {

       # ==== lavaan ====

      est <- lavaan::parameterEstimates(object = fit,
                                        pvalue = TRUE,
                                        ci = TRUE,
                                        remove.nonfree = remove.nonfree,
                                        ...)
      if (!fit_ok) {
        est$est.std <- as.numeric(NA)
        est$ci.lower <- as.numeric(NA)
        est$ci.upper <- as.numeric(NA)
        est$pvalue <- as.numeric(NA)
        est$se <- as.numeric(NA)
      }
    }
  }

  # ==== Fix the column names ====

  enames <- colnames(est)
  enames <- gsub("ci.lower",
                 "cilo",
                 x = enames,
                 fixed = TRUE)
  enames <- gsub("ci.upper",
                 "cihi",
                 x = enames,
                 fixed = TRUE)
  if (standardized) {
    enames <- gsub("est.std",
                   "est",
                   x = enames,
                   fixed = TRUE)
  }
  colnames(est) <- enames

  # ==== Handle sig based on fit_ok ====

  if (!fit_ok) {
    est$sig <- as.numeric(NA)
  } else {
    est$sig <- ifelse((est$cilo > 0) | (est$cihi < 0),
                      yes = 1,
                      no = 0)
  }
  test_label <- lavaan::lav_partable_labels(est)
  out <- cbind(test_label = test_label,
               est)

  # ==== Exclude variances and error variances? ====

  if (exclude_var) {
    i <- (out$lhs == out$rhs) & out$op == "~~"
    out <- out[!i, ]
  }

  # ==== Find parameters using `op` ====

  if (!is.null(op)) {
    j <- which(out$op %in% op)
    if (!isTRUE(length(j) > 0)) {
      stop("'op' set but not found in the test results.")
    }
    out <- out[j, ]
  }

  # ==== Find parameters using `pars` ====

  if (!is.null(pars)) {
    j <- out$test_label %in% pars
    if (!is.null(out$label)) {
      j <- j | (out$label %in% pars)
    }
    j <- which(j)
    if (!isTRUE(length(j) > 0)) {
      stop("'pars' set but not found in the test results.")
    }
    out <- out[j, ]
  }

  # ==== Compare groups? ====

  if (compare_groups && (fit_type == "lavaan")) {
    out <- constrain_all_parameters(
              fit = fit,
              est = out
            )
    if (!fit_ok) {
      out$sig <- as.numeric(NA)
    } else {
      tmp_level <- args$level %||% formals(lavaan::parameterEstimates)$level
      out$sig <- ifelse(out$pvalue < (1 - tmp_level),
                        yes = 1,
                        no = 0)
    }
    out$cilo <- NA
    out$cihi <- NA
    out <- cbind(test_label = paste0(lavaan::lav_partable_labels(out),
                                     " (", out$gps, ")"),
                 out)
    out$gps <- NULL
  }

  # ==== Adjust p-values? ====

  if (p_adjust_method != "none") {

    out$pvalue_org <- out$pvalue
    out$pvalue <- stats::p.adjust(
                        out$pvalue_org,
                        method = p_adjust_method
                      )
  }

  # ==== Omnibus test? ====

  if (omnibus == "no") {

    # ==== No omnibus test ====

    attr(out, "test_label") <- "test_label"

    # ==== Prepare the output ====

    class(out) <- class(est)
    return(out)
  } else {

    # ==== Handle omnibus ====

    out2 <- out[1, ]
    out2[1, ] <- as.numeric(NA)
    tmp <- switch(omnibus,
                  all_sig = "All sig",
                  at_least_one_sig = "1+ sig",
                  at_least_k_sig = paste0(at_least_k,
                                          "+ sig)"))
    out2[1, "test_label"] <- tmp
    tmp <- switch(omnibus,
                  all_sig = as.numeric(isTRUE(all(out$sig == 1))),
                  at_least_one_sig = as.numeric(isTRUE(any(out$sig == 1))),
                  at_least_k_sig = as.numeric(isTRUE(sum(out$sig == 1) >= at_least_k)))
    out2$sig <- tmp
    if (any(is.na(out2$sig))) {
      out2$sig <- as.numeric(NA)
    }

    # ==== Prepare the output ====

    attr(out2, "test_label") <- "test_label"
    return(out2)
  }
}

#' @param object A `power4test` object.
#'
#' @param fit_name The name of the fit
#' results for which the parameter names
#' will be displayed. Default is `"fit"`.
#'
#' @examples
#'
#' # Finding valid parameter names
#'
#' find_par_names(sim_only)
#'
#' @rdname test_parameters
#' @export
find_par_names <- function(object,
                           fit_name = "fit") {
  if (!inherits(object, "power4test")) {
    stop("Only support a 'power4test' object.")
  }
  out <- tryCatch(methods::getMethod("coef",
                                     signature = "lavaan",
                                     where = asNamespace("lavaan"))(object$sim_all[[1]]$extra[[fit_name]]),
                  error = function(e) e)
  if (inherits(out, "error")) {
    stop("Error in getting the coefficients.")
  }
  names(out)
}

#' @noRd
# Find parameters that appear in all groups
pars_to_compare <- function(est) {
  est$pars_id <- seq_len(nrow(est))
  # It is intentional not to use lavaan parameter labels
  est$lor <- paste0(est$lhs, est$op, est$rhs)
  pars_gp1 <- est$pars_id[est$group == 1]
  pars_gp2 <- est$pars_id[est$group == 2]
  pars_mg <- intersect(
                est$lor[pars_gp1],
                est$lor[pars_gp2]
              )
  pars_mg
}

#' @noRd
# Test constraining lor to be equal across groups
test_par_eq <- function(
                  fit,
                  lor,
                  gpa,
                  gpb) {
  pt <- lavaan::parameterTable(fit)

  # Need to fit locally because update() may not work
  slot_opt <- fit@Options
  slot_dat <- fit@Data

  pt$lor <- paste0(pt$lhs, pt$op, pt$rhs)
  i_a <- which((pt$lor == lor) & (pt$group == gpa))
  i_b <- which((pt$lor == lor) & (pt$group == gpb))
  plabel_i_a <- pt$plabel[i_a]
  plabel_i_b <- pt$plabel[i_b]

  pt0 <- lavaan::lav_partable_complete(
            list(lhs = plabel_i_a,
                 op = "==",
                 rhs = plabel_i_b))
  pt0 <- as.data.frame(pt0)
  pt0$id <- max(pt$id) + 1
  pt0$user <- 2
  pt0$block <- 0
  pt0$group <- 0
  pt0$free <- 0

  names_add <- setdiff(colnames(pt),
                       colnames(pt0))
  for (xx in names_add) {
    pt0[, xx] <- NA
  }
  pt1 <- rbind(pt, pt0[, colnames(pt)])
  pt1$lor <- NULL
  pt1[pt0$id, "plabel"] <- ""

  slot_opt$se = "none"

  suppressWarnings(
    fit_update <- lavaan::lavaan(
            model = pt1,
            slotOptions = slot_opt,
            slotData = slot_dat
          )
  )

  if (lavaan::lavInspect(fit, "converged") &&
      (suppressWarnings(lavaan::lavInspect(fit, "post.check")))) {
    suppressWarnings(
      out <- lavaan::lavTestLRT(fit_update, fit)
    )
    est <- out[2, "Chisq diff"]
    pvalue <- out[2, "Pr(>Chisq)"]
  } else {
    pvalue <- NA
  }
  out1 <- data.frame(lhs = pt$lhs[i_a],
                     op = pt$op[i_a],
                     rhs = pt$rhs[i_a],
                     gps = paste0(gpa, "==", gpb),
                     est = est,
                     pvalue = pvalue)
  out1
}

#' @noRd
constrain_all_parameters <- function(
  fit,
  est
) {
  if (missing(est)) {
    est <- lavaan::parameterEstimates(fit,
                                      se = FALSE,
                                      remove.nonfree = TRUE)
  }
  pars0 <- pars_to_compare(est)
  ngroups <- lavaan::lavInspect(fit, "ngroups")
  gp_all <- utils::combn(ngroups,
                         m = 2,
                         simplify = FALSE)
  f <- function(gp_i,
                fit,
                pars0) {
    out <- lapply(pars0,
                  test_par_eq,
                  fit = fit,
                  gpa = gp_i[1],
                  gpb = gp_i[2])
    do.call(rbind,
            out)
  }
  out <- lapply(gp_all,
                f,
                fit = fit,
                pars0 = pars0)
  do.call(rbind,
          out)
}

#' @title Random Variable From a Beta Distribution
#'
#' @description Generate random numbers
#' from a beta distribution, rescaled to
#' have user-specified population mean
#' and standard deviation.
#'
#' @details
#' First, specify the two parameters,
#' `shape1` and `shape2`, and the
#' desired population mean and standard
#' deviation. The random numbers, drawn
#' from a beta distribution by
#' [stats::rbeta()] will then be
#' rescaled with the desired population
#' mean and standard deviation.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param shape1 `shape1` for
#' [stats::rbeta()].
#'
#' @param shape2 `shape2` for
#' [stats::rbeta()].
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rbeta_rs(n = 5000,
#'               shape1 = .5,
#'               shape2 = .5,
#'               pmean = 3,
#'               psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

rbeta_rs <- function(n = 10,
                     shape1 = .5,
                     shape2 = .5,
                     pmean = 0,
                     psd = 1) {
  bmean <- shape1 / (shape1 + shape2)
  bsd   <- sqrt((shape1 * shape2) / ((shape1 + shape2 + 1) * (shape1 + shape2)^2))
  x <- stats::rbeta(n,
                    shape1,
                    shape2)
  x <- pmean + psd * (x - bmean) / bsd
  x
}

library(testthat)

test_that("r* functions", {

set.seed(1234)
x <- rbeta_rs(n = 50000,
              shape1 = .5,
              shape2 = .5,
              pmean = 3,
              psd = 1)
expect_equal(mean(x),
             3,
             tolerance = 1e-2)
expect_equal(sd(x),
             1,
             tolerance = 1e-2)

set.seed(1234)
x <- rbeta_rs2(n = 50000,
               blow = 3,
               bhigh = 10,
               bmean = 5,
               bsd = 1)
expect_equal(mean(x),
             5,
             tolerance = 1e-2)
expect_equal(sd(x),
             1,
             tolerance = 1e-2)
expect_true(max(x) < 10)
expect_true(min(x) > 3)

set.seed(1234)
x <- rbinary_rs(n = 50000,
                p1 = .2,
                pmean = 3,
                psd = 4)
expect_equal(mean(x),
             3,
             tolerance = 1e-2)
expect_equal(sd(x),
             4,
             tolerance = 1e-2)
expect_equal(table(x) / length(x),
             c(.80, .20),
             ignore_attr = TRUE,
             tolerance = 1e-2)

set.seed(1234)
x <- rexp_rs(n = 50000,
             pmean = 3,
             psd = 4)
expect_equal(mean(x),
             3,
             tolerance = 1e-1)
expect_equal(sd(x),
             4,
             tolerance = 1e-1)
expect_true(median(x) < mean(x))

set.seed(1234)
x <- runif_rs(n = 50000,
              min = -3,
              max = 4,
              pmean = 3,
              psd = 2)
expect_equal(mean(x),
             3,
             tolerance = 1e-1)
expect_equal(sd(x),
             2,
             tolerance = 1e-1)

set.seed(1234)
x <- rt_rs(n = 50000,
           df = 5,
           pmean = 3,
           psd = 2)
expect_equal(mean(x),
             3,
             tolerance = 1e-1)
expect_equal(sd(x),
             2,
             tolerance = 1e-1)

set.seed(1234)
x <- rpgnorm_rs(n = 50000,
                p = 1,
                pmean = 3,
                psd = 2)
expect_equal(mean(x),
             3,
             tolerance = 1e-1)
expect_equal(sd(x),
             2,
             tolerance = 1e-1)

set.seed(1234)
x <- rlnorm_rs(n = 50000,
               pmean = 3,
               psd = 2)
expect_equal(mean(x),
             3,
             tolerance = 1e-1)
expect_equal(sd(x),
             2,
             tolerance = 1e-1)
expect_true(min(x) > 0)

})

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

library(testthat)
library(power4mome)

test_check("power4mome")

#' @noRd

alg_bisection <- function(
    object,
    x,
    pop_es_name,
    ...,
    target_power = .80,
    x_max,
    x_min,
    progress,
    x_include_interval,
    x_interval = switch(
                  x,
                  n = c(100, 1000),
                  es = c(.10, .50)),
    simulation_progress,
    save_sim_all,
    is_by_x,
    object_by_org,
    final_nrep,
    final_R,
    ci_level = .95,
    extendInt = c("no", "yes", "downX", "upX"),
    max_trials = 10,
    R = NULL,
    digits = 3,
    lower_hard = switch(x, n = 10, es = 0),
    upper_hard = switch(x, n = 10000, es = .999),
    extend_maxiter = 5,
    what = c("point", "ub", "lb"),
    goal = c("ci_hit", "close_enough"),
    tol = .02,
    variants = list()
) {

  # ==== Pre-search setup ====

  a_out <- power_algorithm_bisection_pre_i(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    x_max = x_max,
    x_min = x_min,
    progress = progress,
    x_include_interval = x_include_interval,
    x_interval = x_interval,
    simulation_progress = simulation_progress,
    save_sim_all = save_sim_all,
    is_by_x = is_by_x,
    object_by_org = object_by_org,
    final_nrep = final_nrep,
    final_R = final_R,
    what = what,
    goal = goal,
    tol = tol,
    ci_level = ci_level
  )

  # ==== Process output ====

  x_interval_updated <- a_out$x_interval_updated
  by_x_1 <- a_out$by_x_1
  fit_1 <- a_out$fit_1

  # TODO:
  # - Check whether it works
  if (is_by_x) {
    tmp <- setdiff(names(object_by_org), names(by_x_1))
    by_x_1 <- c(by_x_1,
                object_by_org[tmp],
                skip_checking_models = TRUE)
  }

  rm(a_out)

  lower_hard <- min(x_interval)
  upper_hard <- max(x_interval)

  # ==== Start the search ====

  a_out <- power_algorithm_bisection(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    ci_level = ci_level,
    x_interval = x_interval_updated,
    extendInt = extendInt,
    progress = progress,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    final_nrep = final_nrep,
    R = R,
    save_sim_all = save_sim_all,
    by_x_1 = by_x_1,
    fit_1 = fit_1,
    is_by_x = is_by_x,
    digits = digits,
    lower_hard = lower_hard,
    upper_hard = upper_hard,
    what = what,
    goal = goal,
    tol = tol,
    variants = variants
  )

  # ==== Return the output ====

  a_out
}

#' @noRd
power_algorithm_bisection <- function(object,
                                      x,
                                      pop_es_name = NULL,
                                      ...,
                                      target_power = .80,
                                      ci_level = .95,
                                      x_interval = switch(x, n = c(100, 1000),
                                                             es = c(.10, .50)),
                                      extendInt = c("no", "yes", "downX", "upX"),
                                      progress = TRUE,
                                      simulation_progress = TRUE,
                                      max_trials = 10,
                                      final_nrep = 400,
                                      R = NULL,
                                      power_model = NULL,
                                      power_curve_start = NULL,
                                      lower_bound = NULL,
                                      upper_bound = NULL,
                                      nls_control = list(),
                                      nls_args = list(),
                                      save_sim_all = FALSE,
                                      start = mean(x_interval),
                                      by_x_1 = NULL,
                                      fit_1 = NULL,
                                      is_by_x = FALSE,
                                      digits = 3,
                                      lower_hard = switch(x, n = 10, es = 0),
                                      upper_hard = switch(x, n = 10000, es = .999),
                                      extend_maxiter = 3,
                                      what = c("point", "ub", "lb"),
                                      goal = c("ci_hit", "close_enough"),
                                      tol = .02,
                                      variants = list()) {
  extendInt <- match.arg(extendInt)
  what <- match.arg(what)
  goal <- match.arg(goal)
  status <- NULL
  changes_ok <- TRUE
  f_history <- vector("numeric", max_trials)
  f_history[] <- NA
  x_history <- vector("numeric", max_trials)
  x_history[] <- NA
  x_interval_history <- matrix(NA,
                               nrow = max_trials,
                               ncol = 2)
  colnames(x_interval_history) <- c("lower", "upper")
  f_interval_history <- matrix(NA,
                               nrow = max_trials,
                               ncol = 2)
  colnames(f_interval_history) <- c("lower", "upper")
  reject_history <- vector("numeric", max_trials)
  reject_history[] <- NA

  i <- NA

  # what: The value to be examined.
  # goal:
  # - ci_hit: Only relevant for what == "point"
  # - close_enough: Can be used for all what.

  x_type <- x

  # ==== Default for variants ====

  variants0 <- list(delta_tol = NULL,
                    last_k = 3,
                    use_power_curve_assist = TRUE,
                    use_power_curve_min_points = 2,
                    power_curve_args = list(),
                    use_power_curve_hybrid = TRUE,
                    muller = FALSE,
                    min_interval_width = c(n = 2,
                                           es = .001))
  variants <- utils::modifyList(variants0,
                                variants)
  proxy_power <- NULL
  if (variants$use_power_curve_assist) {
    proxy_power <- tryCatch(target_power_adjusted(
                      target_power = target_power,
                      goal = goal,
                      what = what,
                      tolerance = 0,
                      nrep = final_nrep,
                      level = ci_level
                    ),
                    error = function(e) e)
    if (inherits(proxy_power, "error")) {
      # Failed to find the proxy power. Do not use power_curve
      variants$use_power_curve_assist <- FALSE
    }
  }

  # ==== Set default for delta_tol ====

  if (is.null(variants$delta_tol)) {
    delta_tol <- switch(
                    x,
                    n = 1,
                    es = .001
                  )
  } else {
    delta_tol <- variants$delta_tol
  }

  last_k <- variants$last_k

  # Create the objective function

  # Tbe output is:
  # - a scalar,
  # - with the attribute "output" storing the full by_* output.
  # - The scalar depends on 'what'

  # ==== Generate the objective function ====

  f <- gen_objective(object = object,
                     x = x,
                     pop_es_name = pop_es_name,
                     target_power = target_power,
                     ci_level = ci_level,
                     progress = progress,
                     digits = digits,
                     nrep = final_nrep,
                     R = R,
                     what = what,
                     simulation_progress = simulation_progress,
                     save_sim_all = save_sim_all,
                     store_output = TRUE)

  # ==== Get the initial interval ====

  # Find f.lower and f.upper

  # Arguments for rejection rates should be retrieved from the object
  by_x_ci <- rejection_rates_add_ci(by_x_1,
                                    level = ci_level)
  x_tried <- get_x_tried(by_x_ci,
                         x = x)
  reject_tried <- by_x_ci$reject
  lower <- min(x_interval)
  upper <- max(x_interval)

  if (progress) {
    cat("\n")
    print_interval(lower = lower,
                   upper = upper,
                   digits = digits,
                   x_type = x_type,
                   prefix = "Initial interval:")
    cat("\n")
  }

  # Compute f.lower and f.upper if not available

  # f.lower

  # ==== Compute f for the lower bound  ====

  tmp <- in_x_tried(lower,
                    object = by_x_ci,
                    x = x)
  if (!is.na(tmp)) {

    # ==== Extract from by_x_1  ====

    output_tmp <- by_x_1[tmp]
    class(output_tmp) <- class(by_x_1)
    f.lower <- f(x_i = lower,
                 power_i = reject_tried[tmp],
                 progress = FALSE)
    attr(f.lower, "output") <- output_tmp
  } else {
    if (progress) {
      cat("\nDo the simulation for the lower bound:\n")
    }

    # ==== Estimate f for the lower bound  ====

    f.lower <- f(x_i = lower,
                 x = x,
                 pop_es_name = pop_es_name,
                 target_power = target_power,
                 ci_level = ci_level,
                 progress = progress,
                 digits = digits,
                 nrep = final_nrep,
                 R = R,
                 what = what,
                 simulation_progress = simulation_progress,
                 save_sim_all = save_sim_all,
                 store_output = TRUE)
    by_x_1 <- c(by_x_1, attr(f.lower, "output"),
                skip_checking_models = TRUE)
  }

  # f.upper

  # ==== Compute f for the upper bound  ====

  tmp <- in_x_tried(upper,
                    object = by_x_ci,
                    x = x)
  if (!is.na(tmp)) {

    # ==== Extract from by_x_1  ====

    output_tmp <- by_x_1[tmp]
    class(output_tmp) <- class(by_x_1)
    f.upper <- f(x_i = upper,
                 power_i = reject_tried[tmp],
                 progress = FALSE)
    attr(f.upper, "output") <- output_tmp
  } else {
    if (progress) {
      cat("\nDo the simulation for the upper bound:\n")
    }

    # ==== Estimate f for the upper bound  ====

    f.upper <- f(x_i = upper,
                 x = x,
                 pop_es_name = pop_es_name,
                 target_power = target_power,
                 ci_level = ci_level,
                 progress = progress,
                 digits = digits,
                 nrep = final_nrep,
                 R = R,
                 what = what,
                 simulation_progress = simulation_progress,
                 save_sim_all = save_sim_all,
                 store_output = TRUE)
    by_x_1 <- c(by_x_1, attr(f.upper, "output"),
                skip_checking_models = TRUE)
  }

  if (progress) {
    cat("\n")
    print_interval(lower = lower,
                   upper = upper,
                   digits = digits,
                   x_type = x_type,
                   prefix = "Initial interval:")
    cat("\n")
    cat("- Rejection Rates:\n")
    tmp <- rejection_rates(by_x_1,
                           level = ci_level)
    print(tmp, annotation = FALSE)
    cat("\n")
  }

  do_search <- TRUE

  # ==== Is one of the bound a solution?  ====

  # Check whether lower or upper is already a solution

  output_lower <- attr(f.lower, "output")
  # Arguments for rejection rates should be retrieved from the object
  # No need for other arguments. Only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_lower,
            all_columns = TRUE
          )
  reject_lower <- tmp$reject
  nrep_lower <- tmp$nrep

  output_upper <- attr(f.upper, "output")
  # Arguments for rejection rates should be retrieved from the object
  # No need for other arguments. Only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_upper,
            all_columns = TRUE
          )
  reject_upper <- tmp$reject
  nrep_upper <- tmp$nrep

  tmp <- check_solution_bounds(
            f.lower = f.lower,
            f.upper = f.upper,
            target_power = target_power,
            final_nrep = final_nrep,
            ci_level = ci_level,
            what = what,
            goal = goal,
            tol = tol
          )

  ok_lower <- tmp["ok_lower"]
  ok_upper <- tmp["ok_upper"]

  if (ok_lower || ok_upper) {

    # ==== Solution in the interval. Skip the search ====

    # One of them is a solution. No need to check the interval,
    # even if the interval is below or above the root.

    do_search <- FALSE

  } else {

    # ==== Extend the interval if necessary ====

    # Fix the interval
    # The original interval is returned if it is OK

    interval_updated <- extend_interval(f = f,
                                        x = x,
                                        pop_es_name = pop_es_name,
                                        target_power = target_power,
                                        ci_level = ci_level,
                                        nrep = final_nrep,
                                        R = R,
                                        what = what,
                                        goal = goal,
                                        tol = tol,
                                        simulation_progress = simulation_progress,
                                        save_sim_all = save_sim_all,
                                        progress = progress,
                                        x_type = x_type,
                                        by_x_1 = by_x_1,
                                        lower = lower,
                                        upper = upper,
                                        f.lower = f.lower,
                                        f.upper = f.upper,
                                        lower_hard = lower_hard,
                                        upper_hard = upper_hard,
                                        extendInt = extendInt,
                                        extend_maxiter = extend_maxiter,
                                        trace = as.numeric(progress),
                                        digits = digits,
                                        store_output = TRUE,
                                        overshoot = switch(x,
                                                           n = .5,
                                                           es = .05),
                                        variants = variants,
                                        proxy_power = proxy_power)
    by_x_1 <- interval_updated$by_x_1 %||% by_x_1
    ci_hit <- FALSE
    solution_found <- FALSE

    lower <- interval_updated$lower
    upper <- interval_updated$upper

    # ==== Update bounds' outputs, if necessary ====

    # The outputs of lower and upper are always in by_x_1

    # Store the output back into f.lower, if not yet stored

    f.lower <- interval_updated$f.lower
    if (is.null(attr(f.lower, "output"))) {
      tmp <- in_x_tried(test_x = lower,
                        object = by_x_1,
                        x = x)
      tmp <- by_x_1[tmp]
      class(tmp) <- class(by_x_1)
      attr(f.lower, "output") <- tmp
    }

    # Store the output back into f.upper, if not yet stored

    f.upper <- interval_updated$f.upper
    if (is.null(attr(f.upper, "output"))) {
      tmp <- in_x_tried(test_x = upper,
                        object = by_x_1,
                        x = x)
      tmp <- by_x_1[tmp]
      class(tmp) <- class(by_x_1)
      attr(f.upper, "output") <- tmp
    }

    # ==== Fix the start value, if necessary ====

    # Fix the start value if it is outside the new interval

    if ((start <= lower) || (start >= upper)) {
      start <- mean(c(lower, upper))
    }

    # ==== Store output(s) to by_x_1, if necessary ====

    # Store the output of f.lower to by_x_1, if not yet stored

    tmp <- in_x_tried(test_x = lower,
                      object = by_x_1,
                      x = x)
    if (is.na(tmp)) {
      by_x_1 <- c(by_x_1, attr(interval_updated$f.lower, "output"),
                  skip_checking_models = TRUE)
    }

    # Store the output of f.upper to by_x_1, if not yet stored

    tmp <- in_x_tried(test_x = upper,
                      object = by_x_1,
                      x = x)
    if (is.na(tmp)) {
      by_x_1 <- c(by_x_1, attr(interval_updated$f.upper, "output"),
                  skip_checking_models = TRUE)
    }

    # ==== Does the updated interval has a solution?  ====

    # Check whether the updated lower or upper is already a solution

    output_lower <- attr(f.lower, "output")
    # Arguments for rejection rates should be retrieved from the object
    # No need for other arguments. Only `reject` and `nrep` are used.
    tmp <- rejection_rates(
             output_lower,
             all_columns = TRUE
           )
    reject_lower <- tmp$reject
    nrep_lower <- tmp$nrep

    output_upper <- attr(f.upper, "output")
    # Arguments for rejection rates should be retrieved from the object
    # No need for other arguments. Only `reject` and `nrep` are used.
    tmp <- rejection_rates(
             output_upper,
             all_columns = TRUE
           )
    reject_upper <- tmp$reject
    nrep_upper <- tmp$nrep

    tmp <- check_solution_bounds(
              f.lower = f.lower,
              f.upper = f.upper,
              target_power = target_power,
              final_nrep = final_nrep,
              ci_level = ci_level,
              what = what,
              goal = goal,
              tol = tol
            )

    ok_lower <- tmp["ok_lower"]
    ok_upper <- tmp["ok_upper"]

    if ((interval_updated$extend_status > 0) &&
        (!ok_lower && !ok_upper)) {

      # ==== No solution and interval invalid. Skip the search ====

      # Interval not OK and no bounds are the solution

      if (progress) {
        cat("\n")
        cat(names(interval_updated$extend_status), ".\n")
        cat("... and none of the bounds are solution.\n")
        cat("Try another interval.\n\n")

        # Should quit

        do_search <- FALSE
      }

    }

  }

  # ==== Is one of the bounds a solution? ====

  if (ok_lower || ok_upper) {

    # ==== Yes. Skip the search ====

    if (progress) {
      cat("One of the bounds in the interval is already a solution.\n\n")
    }

    do_search <- FALSE
    ci_hit <- switch(goal,
                     ci_hit = TRUE,
                     close_enough = NA)
    solution_found <- TRUE

    status <- bisection_status_message(0, status)

    # The lower bound take precedence

    if (ok_lower) {
      x_i <- lower
      out_i <- f.lower
      output_i <- output_lower
      reject_i <- reject_lower
    } else {
      x_i <- upper
      out_i <- f.upper
      output_i <- output_upper
      reject_i <- reject_upper
    }
    # by_x_1 <- c(by_x_1, output_i)
    # No need. lower and upper always in by_x_1
  }

  # ==== Do the search ====

  if (do_search) {

    # Do the bisection search

    x_i <- start
    if (x_type == "n") {
      x_i <- ceiling(x_i)
    }
    f.lower_i <- f.lower
    f.upper_i <- f.upper
    lower_i <- lower
    upper_i <- upper
    status <- 0

    # ==== Start the loop ====

    i <- 1
    while (i <= max_trials) {

      if (progress) {
        cat("\nIteration #", i, "\n")
      }

      # ==== Compute f(x) ====

      out_i <- f(x_i = x_i,
                 x = x,
                 pop_es_name = pop_es_name,
                 target_power = target_power,
                 ci_level = ci_level,
                 progress = progress,
                 digits = digits,
                 nrep = final_nrep,
                 R = R,
                 what = what,
                 simulation_progress = simulation_progress,
                 save_sim_all = save_sim_all,
                 store_output = TRUE)

      output_i <- attr(out_i, "output")
      by_x_1 <- c(by_x_1, output_i,
                  skip_checking_models = TRUE)
      # Arguments for rejection rates should be retrieved from the object
      # No need for other arguments. Only `reject` is used.
      reject_i <- rejection_rates(output_i)$reject

      if (progress) {
        cat("- Rejection Rates:\n")
        # Arguments for rejection rates should be retrieved from the object
        tmp <- rejection_rates(by_x_1,
                               level = ci_level)
        print(tmp, annotation = FALSE)
        cat("\n")
      }

      # TODO:
      # - Check NA, error, etc.
      # Convergence?
      # final_nrep and nrep are always the same
      # for now for bisection

      # ==== Is x a solution? ====

      ok <- check_solution(
              f_i = reject_i,
              target_power = target_power,
              nrep = final_nrep,
              final_nrep = final_nrep,
              ci_level = ci_level,
              what = what,
              goal = goal,
              tol = tol
            )

      f_history[i] <- as.numeric(out_i)
      x_history[i] <- x_i
      x_interval_history[i, ] <- c(lower_i, upper_i)
      f_interval_history[i, ] <- c(f.lower_i, f.upper_i)
      reject_history[i] <- reject_i

      if (ok) {

        # ==== Yes. Solution found ====

        # Solution found

        ci_hit <- switch(goal,
                         ci_hit = TRUE,
                         close_enough = NA)
        solution_found <- TRUE

        status <- bisection_status_message(0, status)

        break

      }

      # ==== Check changes ====

      changes_ok <- check_changes(
              x_history = x_history,
              delta_tol = delta_tol,
              last_k = last_k
            )

      if (!changes_ok) {

        status <- bisection_status_message(2, status)
        break

      }

      # ==== No solution. Update the interval ====

      # TODO:
      # - Handle nonnegative rate

      # == Check rate ==

      # rate_ok <- check_rate(f_history,
      #                       delta_slope_tol = -.1,
      #                       last_k = last_k)

      # == Rate OK. Update as usual ==

      if (x_type == "n") {
        x_i <- ceiling(x_i)
      }
      if (sign(out_i) == sign(f.lower_i)) {
        lower_i <- x_i
        f.lower_i <- out_i
      } else {
        upper_i <- x_i
        f.upper_i <- out_i
      }

      tmp <- variants$min_interval_width[x_type]
      if (abs(lower_i - upper_i) <= tmp) {

        # ==== Interval too narrow. Extend it ====

        if (progress) {
          print_interval(lower = lower_i,
                         upper = upper_i,
                         digits = digits,
                         x_type = x_type,
                         prefix = "Current interval: ")
          cat("- Interval too narrow. Extend it ... \n")
          cat("\n")
        }

        lower_hard_i <- lower_i * .70
        upper_hard_i <- upper_i * 1.30
        if (x_type == "n") {
          lower_hard_i <- ceiling(lower_hard_i)
          upper_hard_i <- ceiling(upper_hard_i)
        }

        interval_updated_i <- extend_interval(
              f = f,
              x = x_type,
              pop_es_name = pop_es_name,
              target_power = target_power,
              ci_level = ci_level,
              nrep = final_nrep,
              R = R,
              what = what,
              goal = goal,
              tol = tol,
              simulation_progress = simulation_progress,
              save_sim_all = save_sim_all,
              progress = progress,
              x_type = x_type,
              by_x_1 = by_x_1,
              lower = lower_i,
              upper = upper_i,
              f.lower = f.lower_i,
              f.upper = f.upper_i,
              lower_hard = lower_hard_i,
              upper_hard = upper_hard_i,
              extendInt = "yes",
              extend_maxiter = extend_maxiter,
              trace = as.numeric(progress),
              digits = digits,
              store_output = TRUE,
              overshoot = switch(x,
                                 n = .9,
                                 es = .90),
              min_x_diff = tmp * 4,
              variants = variants,
              proxy_power = proxy_power
            )

        by_x_1 <- interval_updated_i$by_x_1
        lower_i <- interval_updated_i$lower
        upper_i <- interval_updated_i$upper
        f.lower_i <- interval_updated_i$f.lower
        f.upper_i <- interval_updated_i$f.upper
      }

      if (isTRUE(variants$muller) &&
          (i >= 3)) {

        # ==== Use Muller's Method ====

        # NOTE: Usable but do not use.
        # This method was found to fail to converge
        # even in some simple cases.

        tmp1 <- x_history[(i - 3 + 1):i]
        tmp2 <- f_history[(i - 3 + 1):i]
        x_muller <- tryCatch(
            root_muller_i(f = f,
                          xm2 = tmp1[1],
                          xm1 = tmp1[2],
                          x0i = tmp1[3],
                          ym2 = tmp2[1],
                          ym1 = tmp2[2],
                          y0i = tmp2[3]
                        ),
                      error = function(e) e)
        if (!inherits(x_muller, "error")) {
          x_i <- x_muller
        } else {
          x_i <- mean(c(lower_i, upper_i))
        }
      } else {

        # ==== Mean as the next start ====

        # Use power_curve() to accelerate the search

        power_curve_used <- FALSE

        if (variants$use_power_curve_assist &&
            (length(by_x_1) >= variants$use_power_curve_min_points)) {
          fit_1 <- tryCatch(do.call(
                              power_curve,
                              c(list(object = by_x_1),
                                variants$power_curve_args)
                            ),
                            error = function(e) e)
          if (inherits(fit_1, "power_curve")) {
            # Arguments for rejection rates should be retrieved from the object
            # No need for other arguments because only `x` is used.
            x_i_0 <- estimate_x_range(
                        power_x_fit = fit_1,
                        x = x,
                        target_power = proxy_power,
                        k = 1,
                        tolerance = 0,
                        power_min = 0,
                        power_max = 1,
                        interval = c(lower_i, upper_i),
                        extendInt = "no",
                        x_to_exclude = rejection_rates(by_x_1)[, x, drop = TRUE]
                      )
            if ((x_i_0 > lower_i) &&
                (x_i_0 < upper_i)) {
              power_curve_used <- TRUE
              if (variants$use_power_curve_hybrid) {
                x_i <- mean(c(x_i_0,
                              mean(c(lower_i, upper_i))))
              } else {
                x_i <- x_i_0
              }
            } else {
              x_i <- mean(c(lower_i, upper_i))
            }
          } else {
            x_i <- mean(c(lower_i, upper_i))
          }
        } else {
          x_i <- mean(c(lower_i, upper_i))
        }

      }

      if (x_type == "n") {
        x_i <- ceiling(x_i)
      }

      if (progress) {
        print_interval(lower = lower_i,
                       upper = upper_i,
                       digits = digits,
                       x_type = x_type)
        if (power_curve_used) {
          cat("Power curve used to find the next x ...\n")
        }
        cat("Updated x:", x_i, "\n")
      }

      i <- i + 1

    }

    # ==== End the loop ====

  } else {

    # No iteration

  }

  # Prepare the output

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.

  # ** x_tried **
  # The sample sizes or population values
  # examined.

  # ** x_out **
  # The output in the last trial.

  # ** power_out **
  # The power in the last trial.

  # ** nrep_out **
  # The nrep in the last trial.

  # ** ci_out **
  # The ci in the last trial.

  # ** by_x_out **
  # The by_* output in the last trial

  # ** i2 **
  # The row number of the last trial
  # in by_x_1.

  # ** solution_found **
  # TRUE if an acceptable solution is found

  # ==== Prepare the output ====

  if (!do_search &&
      (!ok_lower && !ok_upper)) {

    # ==== No solution. Set _out to NA ====

    status <- bisection_status_message(1, status)

    # No solution for whatever reason

    x_out <- NA
    power_out <- NA
    nrep_out <- NA
    ci_out <- NA
    by_x_out <- NA
    i2 <- NA

  } else {

    # ==== Solution found. Store it ====

    # Store the last results,
    # regardless of solution

    out_i <- as.numeric(out_i)
    by_ci_i <- rejection_rates_add_ci(output_i,
                                      level = ci_level)
    ci_i <- unlist(by_ci_i[1, c("reject_ci_lo", "reject_ci_hi")])
    x_tried <- get_x_tried(by_x_1,
                           x = x)
    i2 <- match(x_i, x_tried)
    by_x_ci <- rejection_rates_add_ci(by_x_1,
                                      level = ci_level)
    x_out <- x_i
    power_out <- unlist(by_ci_i[1, "reject"])
    nrep_out <- unlist(by_ci_i[1, "nrep"])
    ci_out <- ci_i
    by_x_out <- by_ci_i

  }

  # Available regardless of do_search

  # ==== Update the power curve ====

  fit_1 <- power_curve(by_x_1,
                      formula = power_model,
                      start = power_curve_start,
                      lower_bound = lower_bound,
                      upper_bound = upper_bound,
                      nls_control = nls_control,
                      nls_args = nls_args,
                      verbose = progress)
  x_tried <- get_x_tried(by_x_1,
                         x = x)

  if (progress) {
    if (solution_found) {
      cat("Solution found.\n")
    } else {
      cat("Solution not found.\n")
    }
  }

  # ==== Return the output ====

  # The final x_i are always returned if available
  # ci_hit is used to decide whether
  # final x_i is a solution.
  out <- list(by_x_1 = by_x_1,
              fit_1 = fit_1,
              ci_hit = ci_hit,
              x_tried = x_tried,
              x_out = x_out,
              power_out = power_out,
              nrep_out = nrep_out,
              ci_out = ci_out,
              by_x_out = by_x_out,
              i2 = i2,
              solution_found = solution_found,
              status = status,
              iteration = i,
              x_history = x_history[!is.na(x_history)],
              x_interval_history = x_interval_history[stats::complete.cases(x_interval_history), ],
              f_interval_history = f_interval_history[stats::complete.cases(f_interval_history), ],
              reject_history = reject_history[!is.na(reject_history)],
              f_history = f_history[!is.na(f_history)],
              tol = tol,
              delta_tol = delta_tol,
              last_k = last_k,
              what = what,
              goal = goal,
              ci_level = ci_level)
  out

}

#' @noRd
extend_interval <- function(f,
                            ...,
                            lower,
                            upper,
                            f.lower = f(lower, ...),
                            f.upper = f(upper, ...),
                            lower_hard,
                            upper_hard,
                            extendInt = c("no", "yes", "downX", "upX"),
                            extend_maxiter = 3,
                            trace = 0,
                            digits = 3,
                            by_x_1 = NULL,
                            overshoot = .5,
                            min_x_diff = 0,
                            variants = list(use_power_curve_assist = FALSE),
                            proxy_power = NULL,
                            min_abs_f = 0,
                            extend_only = FALSE) {
  if (trace) {
    cat("\n== Enter extending interval ...\n")
  }
  status_msg <- c("Interval OK" = 0,
                  "Interval not OK but extendInd is no" = 1,
                  "Interval above the solution but extendInd is not yes or downX" = 2,
                  "Interval below the solution but extendInd is not yes or upX" = 3,
                  "Interval above the solution but the lower bound hits lower_hard" = 4,
                  "Interval above the solution but the upper bound hits upper_hard" = 5,
                  "Interval not OK but extend_maxiter reached" = 6,
                  "One of the bounds is a solution" = -1)
  extend_status <- NA
  args <- list(...)
  # x is always supplied
  x_type <- args$x

  # ==== by_x_1 not supplied ====

  # This rarely happens.

  if (is.null(by_x_1)) {
    if (is.null((attr(f.lower, "output")))) {
      f.lower <- f(lower, ...)
    }
    if (is.null((attr(f.upper, "output")))) {
      f.upper <- f(upper, ...)
    }
    tmp <- switch(
              x_type,
              n = as.power4test_by_n(f.lower),
              es = as.power4test_by_es(f.lower,
                                       pop_es_name = args$pop_es_name)
            )
    by_x_1 <- c(tmp,
                f.upper)
  }

  # ==== Interval already valid? ====

  if ((sign(f.lower) != sign(f.upper)) &&
      (abs(lower - upper) >= min_x_diff) &&
      (abs(as.numeric(f.lower)) >= min_abs_f) &&
      (abs(as.numeric(f.upper)) >= min_abs_f)) {

    # ==== Yes. Exit ====

    # No need to extend
    if (trace) {
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = "The interval is already valid:")
      cat("== Exit extending interval ...\n")
    }
    return(list(lower = lower,
                upper = upper,
                f.lower = f.lower,
                f.upper = f.upper,
                interval_ok = TRUE,
                extend_status = status_msg[status_msg == 0],
                extendInt = extendInt))
  }

  # ==== Interval invalid ====

  if (extendInt == "no") {

    # ==== "No". Exit ====

    if (trace) {
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = "extendInt is 'no'. The interval is not extended:")
      cat("== Exit extending interval ...\n")
    }
    return(list(lower = lower,
                upper = upper,
                f.lower = f.lower,
                f.upper = f.upper,
                interval_ok = FALSE,
                extend_status = status_msg[status_msg == 1],
                extendInt = extendInt))
  }

  # ==== Handle bounds with equal function values ====

  rate_i <- .9
  rate_tmp <- 1
  max_tmp <- 5
  while (isTRUE(all.equal(
                  as.numeric(f.lower),
                  as.numeric(f.upper)
                )
              ) ||
          (max_tmp == 0)
          ) {
    max_tmp <- max_tmp - 1
    # To handle bounds with equal f values.
    # If f.upper == f.lower,
    if ((upper > 0) && (lower > 0)) {
      # Only extend lower
      tmp <- (args$target_power - as.numeric(f.lower)) /
             (args$target_power + as.numeric(f.lower))
      tmp <- tmp * rate_tmp
      lower <- max(lower_hard,
                   ifelse(x_type == "n",
                          floor(lower * tmp),
                          lower * tmp),
                   na.rm = TRUE)
      lower <- force_new_x(
                    lower,
                    x_tried = c(get_x_tried(object = by_x_1,
                                            x = x_type),
                                upper),
                    x_interval = range(lower_hard, upper_hard),
                    x_type = x_type
                  )
      f.lower <- f(lower, ...)
      by_x_1 <- c(by_x_1, attr(f.lower, "output"),
                  skip_checking_models = TRUE)
    } else if (((upper < 0) && (lower < 0))) {
      # Only extend upper
      tmp <- (args$target_power - as.numeric(f.upper)) /
             (args$target_power + as.numeric(f.upper))
      tmp <- tmp * 1 / rate_tmp
      upper <- min(upper_hard,
                   ifelse(x_type == "n",
                          floor(upper * tmp),
                          upper * tmp),
                   na.rm = TRUE)
      upper <- force_new_x(
                    upper,
                    x_tried = c(get_x_tried(object = by_x_1,
                                            x = x_type),
                                lower),
                    x_interval = range(lower_hard, upper_hard),
                    x_type = x_type
                  )
      f.upper <- f(upper, ...)
      by_x_1 <- c(by_x_1, attr(f.upper, "output"),
                  skip_checking_models = TRUE)
    } else {
      # This possibility should not happen. Just in case.
      # widen the interval slightly
      tmp_upper <- tmp * 1 / rate_tmp
      tmp_lower <- tmp * rate_tmp

      upper <- min(upper_hard,
                   ifelse(x_type == "n",
                          floor(upper * tmp_upper),
                          upper * tmp_upper),
                   na.rm = TRUE)
      lower <- max(lower_hard,
                   ifelse(x_type == "n",
                          floor(lower * tmp_lower),
                          lower * tmp_lower),
                   na.rm = TRUE)
      upper <- force_new_x(
                    upper,
                    x_tried = c(get_x_tried(object = by_x_1,
                                            x = x_type),
                                lower),
                    x_interval = range(lower_hard, upper_hard),
                    x_type = x_type
                  )
      lower <- force_new_x(
                    lower,
                    x_tried = c(get_x_tried(object = by_x_1,
                                            x = x_type),
                                upper),
                    x_interval = range(lower_hard, upper_hard),
                    x_type = x_type
                  )
      f.upper <- f(upper, ...)
      f.lower <- f(lower, ...)
      by_x_1 <- c(by_x_1, attr(f.upper, "output"),
                  skip_checking_models = TRUE)
      by_x_1 <- c(by_x_1, attr(f.lower, "output"),
                  skip_checking_models = TRUE)
    }
    rate_tmp <- rate_tmp * rate_i
  }

  # ==== How should the interval be extended? ====

  extend_which <- check_extend_x(
                    upper = upper,
                    lower = lower,
                    f.upper = as.numeric(f.upper),
                    f.lower = as.numeric(f.lower)
                  )
  extend_up <- (extend_which == "extend_up")
  extend_down <- (extend_which == "extend_down")

  if (extend_down && (!(extendInt %in% c("yes", "downX")))) {

    # ==== Should extend down but not requested. Exist ====

    if (trace) {
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = paste0("Should extend downward but extendInt not 'yes' or 'downX'.\n",
                                     "The interval is not extended:"))
      cat("== Exit extending interval ...\n")
    }
    return(list(lower = lower,
                upper = upper,
                f.lower = f.lower,
                f.upper = f.upper,
                interval_ok = FALSE,
                extend_status = status_msg[status_msg == 2],
                extendInt = extendInt))
  }
  if (extend_up && (!(extendInt %in% c("yes", "upX")))) {

    # ==== Should extend up but not requested. Exist ====

    if (trace) {
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = paste0("Should extend upward but extendInt not 'yes' or 'upX'.\n",
                                     "The interval is not extended:"))
      cat("== Exit extending interval ...\n")
    }
    return(list(lower = lower,
                upper = upper,
                f.lower = f.lower,
                f.upper = f.upper,
                interval_ok = FALSE,
                extend_status = status_msg[status_msg == 3],
                extendInt = extendInt))
  }
  interval_ok <- FALSE

  # ==== Extend the interval ====

  if (((extendInt %in% c("yes", "downX")) && extend_down) ||
      ((extendInt %in% c("yes", "upX")) && extend_up)) {
    # Extend simple linear extrapolation
    if (trace) {
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = "Current interval:")
      cat(
        switch(extend_which,
               extend_down =
                "Interval above the solution. Extend the lower bound ...\n",
               extend_up =
                "Interval below the solution. Extend the upper bound ...\n")
      )
    }

    # ==== Loop for extension ====
    i <- 1
    overshoot_i <- overshoot
    solution_in_bounds <- FALSE
    while ((i <= extend_maxiter) &&
           isFALSE(solution_in_bounds) &&
           ((sign(f.lower) == sign(f.upper)) ||
            (abs(lower - upper) < min_x_diff))) {
        overshoot_i <- overshoot_i * i
        out_i <- extend_i(
            f = f,
            args = args,
            i = i,
            lower = lower,
            upper = upper,
            f.lower = f.lower,
            f.upper = f.upper,
            interval_ok = interval_ok,
            extend_status = extend_status,
            trace = trace,
            status_msg = status_msg,
            lower_hard = lower_hard,
            upper_hard = upper_hard,
            x_type = x_type,
            digits = digits,
            overshoot = overshoot_i,
            which = switch(extend_which,
                           extend_down = "lower",
                           extend_up = "upper"),
            by_x_1 = by_x_1,
            variants = variants,
            proxy_power = proxy_power
          )
        if (extend_only) {
          if (lower > out_i$lower) {
            lower <- out_i$lower
            f.lower <- out_i$f.lower
          }
          if (upper < out_i$upper) {
            upper <- out_i$upper
            f.upper <- out_i$f.upper
          }
        } else {
          lower <- out_i$lower
          upper <- out_i$upper
          f.lower <- out_i$f.lower
          f.upper <- out_i$f.upper
        }
        if (trace) {
          cat("\n(Extending the interval) Iteration:", i, "\n")
          print_interval(lower = lower,
                         upper = upper,
                         digits = digits,
                         x_type = x_type)
        }
        if (is.na(in_x_tried(
                    lower,
                    by_x_1,
                    x = x_type
              ))) {
          by_x_1 <- c(by_x_1, attr(f.lower, "output"),
                      skip_checking_models = TRUE)
        }
        if (is.na(in_x_tried(
                    upper,
                    by_x_1,
                    x = x_type
              ))) {
          by_x_1 <- c(by_x_1, attr(f.upper, "output"),
                      skip_checking_models = TRUE)
        }
        interval_ok <- out_i$interval_ok
        extend_status <- out_i$extend_status
        extend_which <- check_extend_x(
                          upper = upper,
                          lower = lower,
                          f.upper = as.numeric(f.upper),
                          f.lower = as.numeric(f.lower),
                          upper_hard = upper_hard,
                          lower_hard = lower_hard
                        )
        extend_up <- (extend_which == "extend_up")
        extend_down <- (extend_which == "extend_down")
        i <- i + 1

        # ==== Is one of the bounds a solution?

        tmp <- check_solution_bounds(
                  f.lower = f.lower,
                  f.upper = f.upper,
                  target_power = args$target_power,
                  final_nrep = args$nrep,
                  ci_level = args$ci_level,
                  what = args$what,
                  goal = args$goal,
                  tol = args$tol
                )

        solution_in_bounds <- any(tmp)

      }
    }

  # ==== Check the extended interval ====

  interval_ok <- (sign(f.lower) != sign(f.upper)) &&
                 (abs(lower - upper) >= min_x_diff)

  if (solution_in_bounds) {
    extend_status <- status_msg[status_msg == -1]
    if (trace) {
      cat("One of the bounds is a solution.\n")
    }
  } else {
    if (!interval_ok) {
      if (is.na(extend_status)) {
        # extend_maxiter reached
        extend_status <- status_msg[status_msg == 6]
        if (trace) {
          cat(names(extend_status), ".\n", sep = "")
        }
      }
    } else {
      extend_status <- status_msg[status_msg == 0]
      if (trace) {
        cat(names(extend_status), ".\n", sep = "")
      }
    }
  }

  if (trace) {
    print_interval(lower = lower,
                   upper = upper,
                   digits = digits,
                   x_type = x_type,
                   prefix = "Final extended interval:")
  }

  if (trace) {
    cat("== Exit extending interval ...\n")
  }

  # ==== Return the extended interval ====

  return(list(lower = lower,
              upper = upper,
              f.lower = f.lower,
              f.upper = f.upper,
              interval_ok = interval_ok,
              extend_status = extend_status,
              extendInt = extendInt,
              by_x_1 = by_x_1,
              solution_in_bounds = solution_in_bounds))
}

#' @noRd
extend_i <- function(
                  f,
                  args,
                  i,
                  lower,
                  upper,
                  f.lower,
                  f.upper,
                  interval_ok,
                  extend_status,
                  trace,
                  status_msg,
                  lower_hard,
                  upper_hard,
                  x_type,
                  digits,
                  overshoot,
                  which = c("lower", "upper"),
                  by_x_1 = NULL,
                  variants = list(),
                  proxy_power = NULL) {
  which <- match.arg(which)
  if (((which == "lower") && (lower == lower_hard)) ||
      ((which == "upper") && (upper == upper_hard))) {
    if (!is.null(by_x_1)) {
      # if (which == "lower") {
      #   lower <- force_new_x(
      #                 lower,
      #                 x_tried = c(get_x_tried(object = by_x_1,
      #                                         x = x_type),
      #                             upper),
      #                 x_interval = range(lower_hard, upper_hard),
      #                 x_type = x_type
      #               )
      # } else {
      #   upper <- force_new_x(
      #                 upper,
      #                 x_tried = c(get_x_tried(object = by_x_1,
      #                                         x = x_type),
      #                             lower),
      #                 x_interval = range(lower_hard, upper_hard),
      #                 x_type = x_type
      #               )
      # }
    } else {
      # Hit the hard limit
      interval_ok <- FALSE
      tmp <- switch(which,
                    lower = 4,
                    upper = 5)
      extend_status <- status_msg[status_msg == tmp]
      if (trace) {
        cat(names(extend_status), ".\n", sep = "")
      }
    }
  }
  if (FALSE) {
    # Placeholder: For keeping the indentation.
  } else {
    slope <- (as.numeric(f.upper) - as.numeric(f.lower)) / (upper - lower)
    intercept <- -slope * upper +  as.numeric(f.upper)
    if (which == "lower") {

      # ==== Extend the lower bound ====

      upper <- lower
      f.upper <- f.lower
      # Handle upper == lower
      if (upper == lower) {
        lower <- ifelse(sign(lower) > 0, overshoot, (1 + overshoot)) * lower
      } else {
        lower <- ifelse(sign(lower) > 0, overshoot, (1 + overshoot)) * -intercept / slope
      }

      if (variants$use_power_curve_assist &&
          (length(by_x_1) >= variants$use_power_curve_min_points)) {

        fit_1 <- tryCatch(do.call(
                            power_curve,
                            c(list(object = by_x_1),
                              variants$power_curve_args)
                          ),
                          error = function(e) e)
        if (inherits(fit_1, "power_curve")) {
          # Only `x` is used. No need for other arguments
          lower0 <- estimate_x_range(
                      power_x_fit = fit_1,
                      x = x_type,
                      target_power = proxy_power,
                      k = 1,
                      tolerance = 0,
                      power_min = 0,
                      power_max = 1,
                      interval = c(lower_hard, upper_hard),
                      extendInt = "no",
                      x_to_exclude = rejection_rates(by_x_1)[, x_type, drop = TRUE]
                    )
          if (!is.na(lower0) && is.numeric(lower0)) {
            lower <- ifelse(sign(lower0) > 0, overshoot, (1 + overshoot)) * lower0
          }
        }
      }

      if (lower > upper) {
        lower <- mean(c(lower_hard, upper))
      }
      if (x_type == "n") {
        lower <- ceiling(lower)
      }
      lower <- max(lower, lower_hard)
      if (!is.null(by_x_1)) {
        lower <- force_new_x(
                      lower,
                      x_tried = c(get_x_tried(object = by_x_1,
                                              x = x_type),
                                  upper),
                      x_interval = range(lower_hard, upper_hard),
                      x_type = x_type
                    )
      }
      f.lower <- do.call(f,
                          c(list(x_i = lower),
                           args))
    } else {

      # ==== Extend the upper bound ====

      lower <- upper
      f.lower <- f.upper
      # Handle upper == lower
      if (upper == lower) {
        upper <- ifelse(sign(upper) < 0, overshoot, (1 + overshoot)) * upper
      } else {
        upper <- ifelse(sign(upper) < 0, overshoot, (1 + overshoot)) * -intercept / slope
      }

      if (variants$use_power_curve_assist &&
          (length(by_x_1) >= variants$use_power_curve_min_points)) {

        fit_1 <- tryCatch(do.call(
                            power_curve,
                            c(list(object = by_x_1),
                              variants$power_curve_args)
                          ),
                          error = function(e) e)
        if (inherits(fit_1, "power_curve")) {
          upper0 <- estimate_x_range(
                      power_x_fit = fit_1,
                      x = x_type,
                      target_power = proxy_power,
                      k = 1,
                      tolerance = 0,
                      power_min = 0,
                      power_max = 1,
                      interval = c(lower_hard, upper_hard),
                      extendInt = "no",
                      x_to_exclude = rejection_rates(by_x_1)[, x_type, drop = TRUE]
                    )
          if (!is.na(upper0) && is.numeric(upper0)) {
            upper <- ifelse(sign(upper0) < 0, overshoot, (1 + overshoot)) * upper0
          }
        }
      }

      if (upper < lower) {
        upper <- mean(c(lower, upper_hard))
      }
      if (x_type == "n") {
        upper <- ceiling(upper)
      }
      upper <- min(upper, upper_hard)
      if (!is.null(by_x_1)) {
        upper <- force_new_x(
                      upper,
                      x_tried = c(get_x_tried(object = by_x_1,
                                              x = x_type),
                                  lower),
                      x_interval = range(lower_hard, upper_hard),
                      x_type = x_type
                    )
      }
      f.upper <- do.call(f,
                         c(list(x_i = upper),
                           args))
    }
    # Fix the interval
    if (upper < lower) {
      tmp <- lower
      lower <- upper
      upper <- tmp
      tmp <- f.lower
      f.lower <- f.upper
      f.upper <- tmp
      rm(tmp)
    }
    # if (trace) {
    #   cat("\n\n(Extending the interval) Iteration:", i, "\n\n")
    #   print_interval(lower = lower,
    #                   upper = upper,
    #                   digits = digits,
    #                   x_type = x_type)
    # }
  }
  list(
    lower = lower,
    upper = upper,
    f.lower = f.lower,
    f.upper = f.upper,
    interval_ok = interval_ok,
    extend_status = extend_status
  )
}

#' @noRd

check_extend_x <- function(
                    upper,
                    lower,
                    f.upper,
                    f.lower,
                    upper_hard = Inf,
                    lower_hard = -Inf) {
  slope <- (f.upper - f.lower) / (upper - lower)
  extend_up <- ((slope > 0) && (f.upper < 0)) ||
                ((slope < 0) && (f.upper > 0))
  extend_down <- ((slope > 0) && (f.upper > 0)) ||
                  ((slope < 0) && (f.upper < 0))
  if (extend_up) {
    out <- "extend_up"
  } else if (extend_down) {
    out <- "extend_down"
  } else {
    # If f.upper and f.lower are equal,
    if (upper == upper_hard) {
      out <- "extend_down"
    } else if (lower == lower_hard) {
      out <- "extend_up"
    } else {
      out <- "extend_down"
    }
  }
  out
}

#' @noRd
print_interval <- function(lower,
                           upper,
                           digits,
                           x_type,
                           prefix = "New interval:") {
  tmp <- formatC(c(lower, upper),
                  digits = switch(x_type,
                                  n = 0,
                                  es = digits),
                  format = "f")
  tmp <- paste0("[",paste(tmp, collapse = ", "), "]")
  cat(prefix, tmp, "\n")
}

#' @noRd
gen_objective <- function(object,
                          x,
                          pop_es_name,
                          target_power,
                          ci_level,
                          progress,
                          digits,
                          nrep,
                          R,
                          what = c("point", "ub", "lb"),
                          simulation_progress,
                          save_sim_all,
                          store_output,
                          target_nrep,
                          progress_type = c("cat", "cli")) {
  what <- match.arg(what)
  # Create the objective function
  f <- function(x_i,
                ...,
                x = x,
                pop_es_name = NULL,
                target_power = .80,
                ci_level = .95,
                progress = TRUE,
                digits = 3,
                nrep = 10,
                R = NULL,
                what = "point",
                simulation_progress = TRUE,
                save_sim_all = FALSE,
                store_output = TRUE,
                out_i = NULL,
                power_i = NULL,
                target_nrep = NA,
                progress_type,
                pb_id) {
    if (is.na(target_nrep)) {
      target_nrep <- nrep
    }
    if (x == "n") {
      x_i <- ceiling(x_i)
    }
    # TODO:
    # - Adaptive nrep?
    # Static nrep
    if (progress) {
      tmp <- switch(x,
                    n = as.character(x_i),
                    es = formatC(x_i,
                                  digits = digits,
                                  format = "f"))
      if (progress_type == "cat") {
        cat("\nTry x =", tmp, "\n")
      }
      if (progress_type == "cli") {
        msg <- paste0("Try ", x, ":", tmp)
        # TODO:
        # - Fix this progress status
        # cli::cli_progress_update(id = pb_id)
      }
    }

    # If out_i is supplied
    # it will be used and many arguments will be ignored.
    # If power_i is supplied, even out_i will be ignored
    # Convenient for using f() to compute the function value
    # But be careful of conflicting arguments.

    # ==== How to get power ====

    if (is.null(out_i) && is.null(power_i)) {

      # ==== Estimate Power (No out_i and no powre_i) ====

      out_i <- switch(x,
                      n = power4test_by_n(object,
                                          n = x_i,
                                          R = R,
                                          progress = simulation_progress,
                                          by_nrep = nrep,
                                          save_sim_all = save_sim_all),
                      es = power4test_by_es(object,
                                            pop_es_name = pop_es_name,
                                            pop_es_values = x_i,
                                            R = R,
                                            progress = simulation_progress,
                                            by_nrep = nrep,
                                            save_sim_all = save_sim_all))
      # No need for other arguments because only `reject` is used
      power_i <- rejection_rates(out_i)$reject
    } else {
      if (is.null(power_i)) {

        # ==== Get Power (out_i supplied) ====

        # power_i is supplied. Ignore out_i
        # No need for other arguments because only `reject` is used
        power_i <- rejection_rates(out_i)$reject
        out_i <- NULL
      } else {

        # ==== power_i supplied ====

        # out_i is supplied
        # power_i will be used
      }
    }

    # ==== Compute CI ====

    se_i <- sqrt(power_i * (1 - power_i) / target_nrep)
    ci_i <- reject_ci(
              nreject = round(power_i * target_nrep),
              nvalid = target_nrep,
              level = ci_level,
              method = getOption("power4mome.ci_method", default = "wilson"))
    ci_i <- as.vector(ci_i)

    if (progress) {
      if (progress_type == "cat") {
        tmp1 <- formatC(power_i, digits = digits, format = "f")
        tmp2 <- paste0("[",
                       paste0(formatC(ci_i, digits = digits, format = "f"),
                              collapse = ","),
                      "]")
        cat("\nEstimated power at ", x_i, ": ",
            tmp1,
            ", ", formatC(ci_level*100,
                          digits = max(0, digits - 2),
                          format = "f"), "% confidence interval: ",
            tmp2,
            "\n",
            sep = "")
      }
      if (progress_type == "cli") {
        tmp1 <- formatC(power_i, digits = digits, format = "f")
        tmp2 <- paste0("[",
                       paste0(formatC(ci_i, digits = digits, format = "f"),
                              collapse = ","),
                       "]")
        msg <- paste0("Estimated power at ",
                      x_i, ": ",
                      tmp1, " ",
                      formatC(ci_level * 100,
                              digits = max(0, digits - 2),
                              format = "f"), "% CI ",
                      tmp2)
        # TODO:
        # - Fix this progress status
        # cli::cli_progress_update(id = pb_id)
      }
    }

    # ==== Return function value ====

    out2_i <- switch(what,
                     point = power_i - target_power,
                     ub = ci_i[2] - target_power,
                     lb = ci_i[1] - target_power)
    if (store_output) {
      attr(out2_i, "output") <- out_i
    }
    out2_i
  }

  # ==== Set default values for the objective function ====

  # TODO:
  # - There should be a better way to do this
  formals(f)$x <- x
  if (!is.null(pop_es_name)) {
    formals(f)$pop_es_name <- pop_es_name
  }
  formals(f)$target_power <- target_power
  formals(f)$ci_level <- ci_level
  formals(f)$progress <- progress
  formals(f)$digits <- digits
  formals(f)$nrep <- nrep
  if (!is.null(R)) {
    formals(f)$R <- R
  }
  formals(f)$what <- what
  formals(f)$simulation_progress <- simulation_progress
  formals(f)$save_sim_all <- save_sim_all
  formals(f)$store_output <- store_output
  formals(f)$target_nrep <- NA
  formals(f)$progress_type <- "cat"

  f
}

#' @noRd
power_algorithm_bisection_pre_i <- function(object,
                                            x,
                                            pop_es_name,
                                            target_power,
                                            x_max,
                                            x_min,
                                            progress,
                                            x_include_interval,
                                            x_interval,
                                            simulation_progress,
                                            save_sim_all,
                                            is_by_x,
                                            object_by_org,
                                            power_model,
                                            start,
                                            lower_bound,
                                            upper_bound,
                                            nls_control,
                                            nls_args,
                                            final_nrep,
                                            final_R,
                                            what,
                                            goal,
                                            tol,
                                            ci_level
                                            ) {

  # ==== Initial values ====

  # This method only needs an initial interval
  if (inherits(object_by_org, "power4test_by_n") ||
      inherits(object_by_org, "power4test_by_es")) {
    x_i <- set_x_range_by_x(
                      object,
                      x = x,
                      pop_es_name = pop_es_name,
                      target_power = target_power,
                      k = 2,
                      x_max = x_max,
                      x_min = x_min,
                      object_by_org = object_by_org,
                      what = what,
                      goal = goal,
                      tol = tol,
                      ci_level = ci_level)
  } else {
    x_i <- set_x_range(object,
                      x = x,
                      pop_es_name = pop_es_name,
                      target_power = target_power,
                      k = 2,
                      x_max = x_max,
                      x_min = x_min)
  }

  # For bisection, no need to exclude the value in the input objects
  # Exclude the value in the input object

  # For bisection, only two initial values are needed

  # ==== Set x_interval ====

  if (x_include_interval) {
    # For bisection, should exclude them initially,
    # and let extend_interval() to do the job
    x_i <- sort(c(x_interval, x_i))
  }
  x_i <- sort(unique(x_i))
  if (length(x_i) == 1) {
    x_interval_min <- min(x_interval)
    x_interval_max <- max(x_interval)
    if ((x_i < x_interval_min) ||
        (x_i > x_interval_max)) {
      x_i <- x_interval
    } else {
      tmp <- mean(x_interval)
      if (x_i < tmp) {
        x_i <- c(x_i, x_interval_max)
      } else if (x_i > tmp) {
        x_i <- c(x_i, x_interval_min)
      } else {
        x_i <- c(mean(c(x_i, x_interval_min)),
                 mean(c(x_i, x_interval_max)))
        x_i <- range(x_i)
      }
    }
  } else {
    x_i <- range(x_i)
  }

  # ==== Update by_x (by_x_i) ====

  by_x_i <- switch(x,
          n = as.power4test_by_n(object),
          es = as.power4test_by_es(object,
                                    pop_es_name = pop_es_name)
        )

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.
  by_x_1 <- by_x_i

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.
  # fit_1 <- fit_i
  fit_1 <- NULL

  # TODO:
  # - How about adaptive nrep for bisection?

  # ==== Return the output ====

  out <- list(x_i = NULL,
              by_x_i = NULL,
              fit_i = NULL,
              by_x_1 = by_x_1,
              fit_1 = fit_1,
              nrep_seq = NULL,
              final_nrep_seq = NULL,
              R_seq = NULL,
              x_interval_updated = x_i)

  return(out)
}

#' @noRd

bisection_status_message <- function(x,
                                       status_old) {
  # Do not override existing status
  if (!is.null(status_old)) {
    return(status_old)
  }
  status_msgs <- c(
        "Solution found." = 0,
        "Maximum iteration (max_trials) reached." = 1,
        "Changes in the last_k iterations less than 'delta_tol'." = 2,
        "Search terminated. Changes in the last_k iterations less than 'delta_tol'." = 3
      )
  status_msgs[status_msgs == x]
}

#' @noRd

random_interval <- function(interval_i) {
  interval_i <- stats::na.omit(interval_i)
  interval_i0 <- interval_i[nrow(interval_i), ]
  tmp <- expand.grid(lower = interval_i[, "lower"],
                     upper = interval_i[, "upper"])
  tmp <- split(tmp,
               seq_len(nrow(tmp)))
  tmp2 <- sapply(tmp,
            function(xx) {
              all(xx == interval_i0)
            })
  if (all(tmp2)) return(NULL)
  tmp <- tmp[!tmp2]
  tmp_i <- sample.int(length(tmp),
                      size = 1)
  return(tmp[[tmp_i]])
}

#' @noRd
check_solution_bounds <- function(
  f.lower,
  f.upper,
  target_power,
  final_nrep,
  ci_level,
  what,
  goal,
  tol
) {
  output_lower <- attr(f.lower, "output")
  # No need for other arguments because only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_lower,
            all_columns = TRUE
          )
  reject_lower <- tmp$reject
  nrep_lower <- tmp$nrep
  ok_lower <- check_solution(
                f_i = reject_lower,
                target_power = target_power,
                nrep = nrep_lower,
                final_nrep = final_nrep,
                ci_level = ci_level,
                what = what,
                goal = goal,
                tol = tol
              )
  output_upper <- attr(f.upper, "output")
  # No need for other arguments because only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_upper,
            all_columns = TRUE
          )
  reject_upper <- tmp$reject
  nrep_upper <- tmp$nrep
  ok_upper <- check_solution(
                f_i = reject_upper,
                target_power = target_power,
                nrep = nrep_upper,
                final_nrep = final_nrep,
                ci_level = ci_level,
                what = what,
                goal = goal,
                tol = tol
              )
  c(ok_lower = ok_lower,
    ok_upper = ok_upper)
}

skip("Not used")

# Do not use these tests. They are "too low level."

library(testthat)

test_that("extend_interval: n", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  progress = !is_testing(),
                  iseed = 1234)
out_by <- as.power4test_by_n(out)
f0 <- gen_objective(out,
                    x = "n",
                    pop_es_name = NULL,
                    target_power = .80,
                    ci_level = .95,
                    progress = !is_testing(),
                    digits = 3,
                    nrep = 20,
                    R = NULL,
                    what = "point",
                    simulation_progress = !is_testing(),
                    save_sim_all = TRUE,
                    store_output = TRUE)
set.seed(1234)
f.lower <- f0(20)
f.upper <- f0(200)
as.numeric(f.lower)
as.numeric(f.upper)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 20,
                       upper = 200,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "no",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             1,
             ignore_attr = TRUE)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 20,
                       upper = 200,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "yes",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             0,
             ignore_attr = TRUE)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 20,
                       upper = 200,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "downX",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             3,
             ignore_attr = TRUE)

set.seed(12)
f.lower <- f0(1200)
f.upper <- f0(1250)
as.numeric(f.lower)
as.numeric(f.upper)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 1200,
                       upper = 1250,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "no",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             1,
             ignore_attr = TRUE)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 1200,
                       upper = 1250,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 5000,
                       extendInt = "yes",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             0,
             ignore_attr = TRUE)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 1200,
                       upper = 1250,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "downX",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             0,
             ignore_attr = TRUE)

chk <- extend_interval(f = f0,
                       nrep = attr(out, "args")$nrep,
                       x_type = "n",
                       lower = 1200,
                       upper = 1250,
                       f.lower = f.lower,
                       f.upper = f.upper,
                       lower_hard = 10,
                       upper_hard = 1000,
                       extendInt = "upX",
                       trace = !is_testing(),
                       by_x_1 = out_by)
expect_equal(chk$extend_status,
             2,
             ignore_attr = TRUE)
})

# #' @param xs_per_trial The initial number
# #' of values (sample sizes or population
# #' values) to consider in each
# #' trial. Should be an integer at least
# #' 1. Rounded
# #' up if not an integer.

# #' @param power_min,power_max The minimum
# #' and maximum values, respectively,
# #' of power
# #' when determining the values to
# #' try in each trail. Default is .01.

# #' @param initial_nrep The initial
# #' number of replications. If set
# #' to `NULL`, the `nrep` used in
# #' `object` will be used. If higher
# #' than `final_nrep`, it will be
# #' converted to one-fourth of `final_nrep`.
# #' If lower than the `nrep` in `object`
# #' after the conversion,
# #' then set to `nrep` in the `object`.

# #' @param power_model The nonlinear
# #' model to be used when estimating
# #' the relation between power and
# #' `x`. Should be a formula
# #' acceptable by [stats::nls()],
# #' with `reject` on the left-hand side,
# #' and `x`
# #' on the right-hand
# #' side, with one or more parameters.
# #' Can also be set to a list of
# #' models.
# #' Users rarely need to change the
# #' default value. If `NULL`, the default,
# #' then the default model(s) will be
# #' determined by [power_curve()].
# #'
# #' @param start A named numeric vector
# #' of the starting values for `power_model`
# #' when fitted by [stats::nls()]. If
# #' `power_model` is a list, this should
# #' be a list of the same length.
# #' Users rarely need to change the
# #' default values.
# #'
# #' @param lower_bound A named numeric vector
# #' of the lower bounds for parameters
# #' in `power_model`
# #' when fitted by [stats::nls()]. If
# #' `power_model` is a list, this should
# #' be a list of the same length.
# #' Users rarely need to change the
# #' default values.
# #'
# #' @param upper_bound A named numeric vector
# #' of the upper bounds for parameters
# #' in `power_model`
# #' when fitted by [stats::nls()]. If
# #' `power_model` is a list, this should
# #' be a list of the same length.
# #' Users rarely need to change the
# #' default values.
# #'
# #' @param nls_args A named list of
# #' arguments to be used when calling
# #' [stats::nls()]. Used to override
# #' internal default, such as the
# #' algorithm (default is `"port"`).
# #' Use this argument with cautions.
# #'
# #' @param nls_control A named list of
# #' arguments to be passed the `control`
# #' argument of [stats::nls()] when
# #' estimating the relation between
# #' power and `x`. The values will
# #' override internal default values,
# #' and also override `nls_args`.
# #' Use this argument with cautions.

# #' @param initial_R The initial number of
# #' Monte Carlo simulation or
# #' bootstrapping samples. The `R` in calling
# #' [power4test()], [power4test_by_n()],
# #' or [power4test_by_es()]. If set to `NULL`,
# #' the `R` used in
# #' `object` will be used.
# #' If higher
# #' than `final_R`, it will be
# #' converted to one-fourth of `final_R`.
# #' If lower than the `R` in `object`
# #' after the conversion,
# #' then set to `R` in `object``.

# #' @param nrep_steps How many steps
# #' the number of replications will be
# #' increased to `final_nrep`, if the
# #' initial number of replications
# #' (`nrep` in [power4test()]) is
# #' less than `final_nrep`. The number
# #' of replications will be successively
# #' increased by this number of steps
# #' to increase the precision in estimating
# #' the power. Should be at least 1.
# #' Increasing this number will result
# #' in more trials and take longer to
# #' run, but will try more values.
# #' Rounded up if not an integer.

# #' @param final_xs_per_trial The final number
# #' of values (sample sizes or population
# #' values) to consider in the last
# #' trial or last few trials. Should be an integer at least
# #' 1. Rounded
# #' up if not an integer.

#' @noRd

alg_power_curve <- function(
  object,
  x,
  pop_es_name,
  ...,
  target_power,
  xs_per_trial = 3,
  x_max,
  x_min,
  nrep0 = 100,
  R0 = 250,
  progress,
  x_include_interval,
  x_interval,
  simulation_progress,
  save_sim_all,
  is_by_x,
  object_by_org,
  power_model,
  start,
  lower_bound,
  upper_bound,
  nls_control,
  nls_args,
  models,
  final_nrep,
  nrep_steps = 1,
  final_R,
  final_xs_per_trial = 1,
  pre_i_xs = 5,
  pre_i_nrep = 50,
  pre_i_R = ifelse(is.null(R0),
                   NULL,
                   min(200, R0)),
  max_trials,
  ci_level,
  power_min = .01,
  power_max = .90,
  extendInt,
  power_tolerance_in_interval,
  power_tolerance_in_final,
  what = c("point", "ub", "lb"),
  goal = c("ci_hit", "close_enough"),
  tol = .02,
  variants = list()
) {

  what <- match.arg(what)
  goal <- match.arg(goal)

  if (is.null(models)) {
    models <- c("nls", "logistic", "lm")
  }

  # ==== Sanity check ====

  if (final_xs_per_trial < 1) {
    stop("'final_xs_per_trial' (",
         final_xs_per_trial,
         ") is less than 1.")
  }
  final_xs_per_trial <- ceiling(final_xs_per_trial)

  nrep_steps <- ceiling(nrep_steps)
  if (nrep_steps < 0) {
    stop("'nrep_steps' must be at least 1 (after rounding, if necessary).")
  }

  if (xs_per_trial < 1) {
    stop("'xs_per_trial' (",
         xs_per_trial,
         ") is less than 1.")
  }
  xs_per_trial <- ceiling(xs_per_trial)

  if (power_min <= 0 || power_max >= 1) {
    stop("'power_min' and 'power_max' must be between 0 and 1.")
  }

  if (power_max < target_power || power_min > target_power) {
    stop("'target_power' must be between 'power_min' and 'power_max'.")
  }

  # ==== Get nrep and R ====

  nrep_org <- attr(object, "args")$nrep
  if (nrep0 > final_nrep) {
    nrep0 <- ceiling(final_nrep / 4)
    if ((nrep0 < 100) && (nrep_org <= final_nrep)) {
      nrep0 <- nrep_org
    }
  }

  R_org <- attr(object, "args")$R

  if (is.null(R0)) {
    R0 <- R_org
  } else {
    if (R0 > final_R) {
      R0 <- ceiling(final_R / 4)
      if ((R0 < 100) && (R_org <= final_R)) {
        R0 <- R_org
      }
    }
    R0 <- ceiling(R0)
  }

  # ==== proxy_power (adjusted power) ====

  proxy_power <- target_power_adjusted(
                    target_power = target_power,
                    goal = goal,
                    what = what,
                    tolerance = tol,
                    nrep = final_nrep,
                    level = ci_level
                  )

  # ==== Pre-search setup ====

  a_out <- power_algorithm_search_by_curve_pre_i(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    xs_per_trial = xs_per_trial,
    x_max = x_max,
    x_min = x_min,
    nrep0 = nrep0,
    R0 = R0,
    progress = progress,
    x_include_interval = x_include_interval,
    x_interval = x_interval,
    simulation_progress = simulation_progress,
    save_sim_all = save_sim_all,
    is_by_x = is_by_x,
    object_by_org = object_by_org,
    power_model = power_model,
    start = start,
    lower_bound = lower_bound,
    upper_bound = upper_bound,
    nls_control = nls_control,
    nls_args = nls_args,
    models = models,
    final_nrep = final_nrep,
    nrep_steps = nrep_steps,
    final_R = final_R,
    final_xs_per_trial = final_xs_per_trial,
    pre_i_xs = pre_i_xs,
    pre_i_nrep = pre_i_nrep,
    pre_i_R = pre_i_R,
    what = what,
    goal = goal,
    ci_level = ci_level,
    tol = tol,
    proxy_power = proxy_power
  )

  # ==== Process output ====

  by_x_1 <- a_out$by_x_1
  fit_1 <- a_out$fit_1
  nrep_seq <- a_out$nrep_seq
  final_nrep_seq <- a_out$final_nrep_seq
  R_seq <- a_out$R_seq
  xs_per_trial_seq <- a_out$xs_per_trial_seq

  # TODO:
  # - Check whether it works
  if (is_by_x) {
    tmp <- setdiff(names(object_by_org), names(by_x_1))
    by_x_1 <- c(by_x_1,
                object_by_org[tmp],
                skip_checking_models = TRUE)
  }

  rm(a_out)

  # ==== Start the search ====

  a_out <- power_algorithm_search_by_curve(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    xs_per_trial_seq = xs_per_trial_seq,
    ci_level = ci_level,
    power_min = power_min,
    power_max = power_max,
    x_interval = x_interval,
    extendInt = extendInt,
    progress = progress,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    final_nrep = final_nrep,
    power_model = power_model,
    start = start,
    lower_bound = lower_bound,
    upper_bound = upper_bound,
    nls_control = nls_control,
    nls_args = nls_args,
    save_sim_all = save_sim_all,
    power_tolerance_in_interval = power_tolerance_in_interval,
    power_tolerance_in_final = power_tolerance_in_final,
    by_x_1 = by_x_1,
    fit_1 = fit_1,
    nrep_seq = nrep_seq,
    final_nrep_seq = final_nrep_seq,
    R_seq = R_seq,
    final_xs_per_trial = final_xs_per_trial,
    what = what,
    goal = goal,
    tol = tol,
    proxy_power = proxy_power,
    variants = variants)

  # ==== Return the output ====

  a_out
}

#' @noRd

power_algorithm_search_by_curve <- function(object,
                                            x,
                                            pop_es_name,
                                            target_power,
                                            xs_per_trial_seq,
                                            ci_level,
                                            power_min,
                                            power_max,
                                            x_interval,
                                            extendInt,
                                            progress = TRUE,
                                            simulation_progress,
                                            max_trials = 10,
                                            final_nrep,
                                            power_model,
                                            start,
                                            lower_bound,
                                            upper_bound,
                                            nls_control,
                                            nls_args,
                                            models,
                                            save_sim_all,
                                            power_tolerance_in_interval,
                                            power_tolerance_in_final,
                                            by_x_1,
                                            fit_1,
                                            nrep_seq,
                                            final_nrep_seq,
                                            R_seq,
                                            final_xs_per_trial,
                                            what = "point",
                                            goal = "ci_hit",
                                            tol = .02,
                                            proxy_power,
                                            variants = list()) {

  # goal: ci_hit
  #   what: point
  #     - No need to change
  # goal: close_enough
  #   what: point:
  #     - Supported
  #   what: ub, lb
  #     - TODO: Use proxy_power from target_power_adjusted()

  # ci_hit does NOT denote solution found.
  # - Used only when goal == "ci_hit"
  # solution_found refers to goal met.
  # - Used in both "ci_hit" and "close_enough"
  ci_hit <- FALSE
  solution_found <- FALSE

  i2 <- NULL
  target_in_range <- FALSE
  status <- NULL
  changes_ok <- TRUE

  x_history <- vector("numeric", max_trials)
  x_history[] <- NA
  reject_history <- vector("numeric", max_trials)
  reject_history[] <- NA

  # ==== Default for variants ====

  variants0 <- list(delta_tol = NULL,
                    last_k = 3)
  variants <- utils::modifyList(variants0,
                                variants)

  # ==== Set default for delta_tol ====

  if (is.null(variants$delta_tol)) {
    delta_tol <- switch(
                    x,
                    n = 1,
                    es = .001
                  )
  } else {
    delta_tol <- variants$delta_tol
  }

  last_k <- variants$last_k

  # power_tolerance_in_interval
  # ci_hit-point: Used in the tolerance of estimate_x_range()
  # close_enough: Can use ci_hit approach because
  #               it can help estimate the power curve.

  # ==== Start the Loop ====

  for (j in seq_len(max_trials)) {

    if (progress) {
      cat("--- Trial", j, "---\n\n")
      tmp <- format(Sys.time(), "%Y-%m-%d %X")
      cat("- Start at", tmp, "\n")
    }

    # ==== Determine values to try ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    # The ci_hit approach can be used for close_enough

    if (ci_hit) {
      # After the first trial,
      # Check whether at least one CI hit the target power.
      # If yes, reduce the range of power levels when
      # selecting the values to try.
      # Necessary because the number of replication may have
      # increased, requiring narrower CI.
      power_tolerance_in_interval <- max(abs(ci_out - power_out))
    }

    # ** x_j **
    # The vector of values to be tried in this trial
    # Determined using by latest power curve (fit_1)
    x_tried <- switch(x,
                      n = as.numeric(names(by_x_1)),
                      es = sapply(by_x_1,
                                  \(x) {attr(x, "pop_es_value")},
                                  USE.NAMES = FALSE))

    # ==== Update x-interval to try ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    # The ci_hit approach can be used for close_enough

    if (target_in_range) {
      # ==== Target power in range ====

      # ci_hit-point: OK
      # close_enough-point: OK
      # close_enough-lb/ub: OK

      # The ci_hit approach can be used for close_enough

      # Always include the intersection, if target_in_range

      x_j <- estimate_x_range(power_x_fit = fit_1,
                              x = x,
                              target_power = proxy_power,
                              k = max(xs_per_trial_seq[1],
                                      1),
                              tolerance = power_tolerance_in_interval,
                              power_min = power_min,
                              power_max = power_max,
                              interval = x_interval,
                              extendInt = extendInt,
                              x_to_exclude = x_tried)
      tmp <- switch(x,
                    n = ceiling(x_between_i),
                    es = x_between_i)
      x_j <- c(x_j, tmp)
      x_j <- unique(x_j)
      if (length(x_j) < xs_per_trial_seq[1]) {
        # estimate_x_range generated x_between_i
        # Call it again to get all k values

        # ci_hit-point: OK
        # close_enough-point: OK
        # close_enough-lb/ub: OK

        # The ci_hit approach can be used for close_enough

        x_j <- estimate_x_range(power_x_fit = fit_1,
                                x = x,
                                target_power = proxy_power,
                                k = xs_per_trial_seq[1],
                                tolerance = power_tolerance_in_interval,
                                power_min = power_min,
                                power_max = power_max,
                                interval = x_interval,
                                extendInt = extendInt,
                                x_to_exclude = x_tried)
      }
    } else {
      # ==== Target power not in range ====

      # ci_hit-point: OK
      # close_enough-point: OK
      # close_enough-lb/ub: OK

      # The ci_hit approach can be used for close_enough

      x_j <- estimate_x_range(power_x_fit = fit_1,
                              x = x,
                              target_power = proxy_power,
                              k = xs_per_trial_seq[1],
                              tolerance = power_tolerance_in_interval,
                              power_min = power_min,
                              power_max = power_max,
                              interval = x_interval,
                              extendInt = extendInt,
                              x_to_exclude = x_tried)
    }

    # ==== Adjust nrep based on extrapolated power ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    # The ci_hit approach can be used for close_enough

    # Adjust the numbers of replication for each value.
    # A value with estimated power closer to the
    # target power will have a higher number of replication
    # power_j <- predict_fit(fit_1,
    #                        newdata = list(n = n_j))
    power_j <- stats::predict(fit_1,
                              newdata = list(x = x_j))
    nrep_j <- nrep_from_power(power_j = power_j,
                              target_power = proxy_power,
                              tolerance = power_tolerance_in_final,
                              nrep_min = nrep_seq[1],
                              nrep_max = final_nrep_seq[1])

    if (progress) {
      x_j_str <- formatC(x_j,
                         digits = switch(x, n = 0, es = 3),
                         format = "f")
      cat("- Value(s) to try:",
          paste0(x_j_str, collapse = ", "),
          "\n")
      cat("- Numbers of replications:",
          paste0(nrep_j, collapse = ", "),
          "\n")
    }

    # ==== Do the simulation for each value  ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    # ** by_x_j **
    # The results for this trial (based on n_j)
    by_x_j <- switch(x,
                     n = power4test_by_n(object,
                                         n = x_j,
                                         R = R_seq[1],
                                         progress = simulation_progress,
                                         by_nrep = nrep_j,
                                         save_sim_all = save_sim_all),
                     es = power4test_by_es(object,
                                           pop_es_name = pop_es_name,
                                           pop_es_values = x_j,
                                           R = R_seq[1],
                                           progress = simulation_progress,
                                           by_nrep = nrep_j,
                                           save_sim_all = save_sim_all))

    # Add the results to by_x_1
    by_x_1 <- c(by_x_1, by_x_j,
                skip_checking_models = TRUE)

    if (progress) {
      cat("\n- Rejection Rates:\n\n")
      tmp <- rejection_rates(by_x_1,
                             level = ci_level)
      print(tmp,
            annotation = FALSE)
      cat("\n")
    }

    # ==== Update the power curve ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    fit_1 <- power_curve(by_x_1,
                         formula = power_model,
                         start = start,
                         lower_bound = lower_bound,
                         upper_bound = upper_bound,
                         nls_control = nls_control,
                         nls_args = nls_args,
                         verbose = progress,
                         models = models)

    # ==== Is target power in the range of current power levels? ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    # Get the rejection rates of all values tried.
    # No need for other arguments because only `reject` is used.
    tmp1 <- rejection_rates_add_ci(by_x_1,
                                   level = ci_level)
    # tmp1$reject <- tmp1$sig
    tmp2 <- range(tmp1$reject)

    # Is the desired value likely already in the range
    # of values examined, based on the estimated power?
    target_in_range <- (proxy_power > tmp2[1]) &&
                       (proxy_power < tmp2[2])

    if (target_in_range) {

      # ==== Current solution: x with closest power ====

      # ci_hit-point: OK
      # close_enough-point: OK
      # close_enough-lb/ub: OK

      # The desired value probably within the range examined

      x_between_i <- x_from_y_rejection_rates(
                        reject_df = tmp1,
                        target_power = proxy_power,
                        x = x
                      )

      x_out_i <- find_solution(
                    object = by_x_1,
                    ci_level = ci_level,
                    target_power = target_power,
                    what = what,
                    tol = tol,
                    goal = goal,
                    final_nrep = 0,
                    closest_ok = TRUE,
                    if_ties = "min"
                  )

      # If ties, the smallest value will be used

      # ** x_out, power_out, nrep_out, ci_out, by_x_out **
      # The results of the candidate solution,
      # - The value with power closest to the target power
      x_out <- switch(x,
                      n = tmp1$n[x_out_i],
                      es = tmp1$es[x_out_i])
      power_out <- tmp1$reject[x_out_i]
      nrep_out <- tmp1$nrep[x_out_i]
      by_x_ci <- rejection_rates_add_ci(by_x_1,
                                        level = ci_level)
      ci_out <- unlist(by_x_ci[x_out_i, c("reject_ci_lo", "reject_ci_hi")])
      by_x_out <- by_x_1[[x_out_i]]

      if (progress) {
        x_out_str <- formatC(x_out,
                             digits = switch(x, n = 0, es = 4),
                             format = "f")
        cat("- Value with closest power:", x_out_str, "\n")
        cat("- Estimated power for ",
            x_out_str,
            ": ",
            formatC(power_out, digits = 4, format = "f"),
            "\n",
            sep = "")
      }

    } else {

      # ==== Current solution: By power curve ====

      # ci_hit-point: OK
      # close_enough-point: OK
      # close_enough-lb/ub: OK

      # The desired value may not be within the range examined
      # Use the latest power curve to estimate the desired value.
      # Inaccurate, but help approaching the target value.

      # ** x_out, power_out, nrep_out, ci_out, by_x_out **
      # Considered a candidate solution.
      x_tried <- switch(x,
                        n = as.numeric(names(by_x_1)),
                        es = sapply(by_x_1,
                                    \(x) {attr(x, "pop_es_value")},
                                    USE.NAMES = FALSE))
      x_out <- estimate_x_range(power_x_fit = fit_1,
                                x = x,
                                target_power = proxy_power,
                                k = 1,
                                tolerance = 0,
                                power_min = power_min,
                                power_max = power_max,
                                interval = x_interval,
                                extendInt = extendInt,
                                x_to_exclude = x_tried)
      by_x_out <- switch(x,
                         n = power4test_by_n(object,
                                             n = x_out,
                                             nrep = nrep_seq[1],
                                             R = R_seq[1],
                                             progress = simulation_progress,
                                             save_sim_all = save_sim_all),
                         es = power4test_by_es(object,
                                               pop_es_name = pop_es_name,
                                               pop_es_values = x_out,
                                               nrep = nrep_seq[1],
                                               R = R_seq[1],
                                               progress = simulation_progress,
                                               save_sim_all = save_sim_all))
      nrep_out <- nrep_seq[1]
      # No need for additional arguments because only `reject` is used
      power_out <- rejection_rates(by_x_out)[1, "reject"]
      by_x_out_ci <- rejection_rates_add_ci(by_x_out,
                                            level = ci_level)
      ci_out <- unlist(by_x_out_ci[1, c("reject_ci_lo", "reject_ci_hi")])

      if (progress) {
        x_out_str <- formatC(x_out,
                            digits = switch(x, n = 0, es = 4),
                            format = "f")
        cat("- Extrapolated value:", x_out_str, "\n")
        cat("- Estimated power for ",
            x_out_str,
            ": ",
            formatC(power_out, digits = 4, format = "f"),
            "\n",
            sep = "")
      }

      # Add the results for the new value to
      # the collection of results
      by_x_1 <- c(by_x_1, by_x_out,
                  skip_checking_models = TRUE)

      # Update by_n_out
      by_x_out <- by_x_out[[1]]

      # Update the power curve
      fit_1 <- power_curve(by_x_1,
                           formula = power_model,
                           start = stats::coef(fit_1),
                           lower_bound = lower_bound,
                           upper_bound = upper_bound,
                           nls_control = nls_control,
                           nls_args = nls_args,
                           verbose = progress,
                           models = models)

    }

    if (progress) {
      cat("- Power Curve:\n")
      print(fit_1)
      cat("\n")
    }

    x_history[j] <- x_out
    reject_history[j] <- power_out

    # ==== Goal met? ci_hit ====

    # ci_hit-point: OK.
    # - No need to use proxy_power

    if (goal == "ci_hit") {

      # ==== Any CI hits target power? ====

      # Check results accumulated so far
      by_x_ci <- rejection_rates_add_ci(by_x_1,
                                        level = ci_level)
      i0 <- (by_x_ci$reject_ci_lo < target_power) &
            (by_x_ci$reject_ci_hi > target_power)

      # Is there at least one CI hitting the target power?
      if (any(i0)) {

        # ==== At least one CI hits target power ====

        # At least one CI hits the target power

        ci_hit <- TRUE

        i2 <- find_ci_hit(by_x_1,
                          ci_level = ci_level,
                          target_power = target_power,
                          final_nrep = final_nrep)
        if (!is.na(i2) && !is.null(i2)) {

          # ==== Solution found ====

          # ci_hit && final_nrep reached

          solution_found <- TRUE

          # Updated *_out objects
          by_x_out <- by_x_1[[i2]]
          x_out <- switch(x,
                          n = by_x_ci$n[i2],
                          es = by_x_ci$es[i2])
          power_out <- by_x_ci$reject[i2]
          nrep_out <- by_x_ci$nrep[i2]
          ci_out <- unlist(by_x_ci[i2, c("reject_ci_lo", "reject_ci_hi")])
        } else {

          # ==== CI hits but final_nrep not reached ====

          # No CI with final_nrep hit
          # Get the closet solution

          i2 <- find_ci_hit(by_x_1,
                            ci_level = ci_level,
                            target_power = target_power,
                            final_nrep = 0)
          # Updated *_out objects
          by_x_out <- by_x_1[[i2]]
          x_out <- switch(x,
                          n = by_x_ci$n[i2],
                          es = by_x_ci$es[i2])
          power_out <- by_x_ci$reject[i2]
          nrep_out <- by_x_ci$nrep[i2]
          ci_out <- unlist(by_x_ci[i2, c("reject_ci_lo", "reject_ci_hi")])
        }

      } else {

        # ==== No CI hits target power ====

        # No CI hits the target power

        ci_hit <- FALSE
      }

      # ==== CI hits? ====

      if (ci_hit) {

        # Is the nrep of the candidate already equal to
        # target nrep for the final solution?

        if (solution_found) {

          # ==== CI hits and solution found. Exit the loop ====

          # Desired accuracy (based on nrep) achieved.
          # Exit the loop and finalize the results.

          if (progress) {
            cat("- Estimated power's CI include the target power (",
                formatC(target_power, digits = 4, format = "f"), "). ",
                "(CI: [", paste0(formatC(ci_out, digits = 4, format = "f"), collapse = ","), "])",
                "\n",
                sep = "")
          }

          status <- power_curve_status_message(0, status)

          break

        } else {

          # ==== CI hits but no solution. Next set of values in _seq ====

          # Move to the next step in the sequences
          # E.g., increase nrep.

          if (length(nrep_seq) > 1) {
            nrep_seq <- nrep_seq[-1]
            final_nrep_seq <- final_nrep_seq[-1]

            if (progress) {
              cat("- Minimum number of replications changed to",
                  nrep_seq[1], "\n\n")
            }

            R_seq <- R_seq[-1]
            xs_per_trial_seq <- xs_per_trial_seq[-1]
          }
        }
      } else {

        # ==== No CI hits ====

        # No value has CI hitting the target power.

        if (progress) {
          cat("- Estimated power's CI does not include the target power (",
              formatC(target_power, digits = 4, format = "f"), "). ",
              "(CI: [", paste0(formatC(ci_out, digits = 4, format = "f"), collapse = ","), "])",
              "\n",
              sep = "")
        }
      }

    }

    # ==== Goal met? close_enough ====

    # close_enough-point: OK
    # close_enough-lb/ub: OK

    if (goal == "close_enough") {

      # ==== Any power levels close enough to the target power? ====

      by_x_ci <- rejection_rates_add_ci(by_x_1,
                                        level = ci_level)
      i0 <- find_solution(
                by_x_1,
                target_power = target_power,
                ci_level = ci_level,
                what = what,
                tol = tol,
                goal = goal,
                final_nrep = final_nrep,
                closest_ok = FALSE,
                if_ties = "min")

      if (!is.null(i0)) {

        # ==== Close enough ====

        # close enough && final_nrep reached

        solution_found <- TRUE

        # Updated *_out objects
        by_x_out <- by_x_1[[i0]]
        x_out <- switch(x,
                        n = by_x_ci$n[i0],
                        es = by_x_ci$es[i0])
        power_out <- by_x_ci$reject[i0]
        nrep_out <- by_x_ci$nrep[i0]
        ci_out <- unlist(by_x_ci[i0, c("reject_ci_lo", "reject_ci_hi")])

        # Exit the loop and finalize the results.

        if (progress) {
          cat("Solution found. Finalizing the results ...\n")
        }

        status <- power_curve_status_message(0, status)

        break

      } else {

        # ==== Not close enough ====

        # No power with final_nrep close enough
        # Get the closet solution

        i0 <- find_solution(
                  by_x_1,
                  target_power = target_power,
                  ci_level = ci_level,
                  what = what,
                  tol = tol,
                  goal = goal,
                  final_nrep = 0,
                  closest_ok = TRUE,
                  if_ties = "min")

        # Updated *_out objects
        by_x_out <- by_x_1[[i0]]
        x_out <- switch(x,
                        n = by_x_ci$n[i0],
                        es = by_x_ci$es[i0])
        power_out <- by_x_ci$reject[i0]
        nrep_out <- by_x_ci$nrep[i0]
        ci_out <- unlist(by_x_ci[i0, c("reject_ci_lo", "reject_ci_hi")])

        if (progress) {
          cat("Solution not found. Next trial ...\n")
        }

      }

    }

    # ==== Check changes ====

    # ci_hit-point: OK
    # close_enough-point: OK
    # close_enough-lb/ub: OK

    changes_ok <- check_changes(
            x_history = x_history,
            delta_tol = delta_tol,
            last_k = last_k
          )

    if (!changes_ok) {

      status <- power_curve_status_message(2, status)
      break

    }

  }

  # ==== End the Loop ====

  # ci_hit-point: OK
  # close_enough-point: OK
  # close_enough-lb/ub: OK

  if (!solution_found) {
    status <- power_curve_status_message(1, status)
  }

  # ==== Return the output ====

  # ci_hit-point: OK
  # close_enough-point: OK
  # close_enough-lb/ub: OK

  if (goal == "close_enough") {

    # ==== Prepare the output (close_enough) ====

    if (solution_found) {

      # Store the last results,
      # regardless of solution

      # Other *_out objects guaranteed to be available

      i2 <- i0

    } else {
      x_out <- NA
      power_out <- NA
      nrep_out <- NA
      ci_out <- NA
      by_x_out <- NA
      i2 <- NA
    }
  }

  # what %in% c("lb", "ub")
  # - *_out: Still the output at x

  out <- list(by_x_1 = by_x_1,
              fit_1 = fit_1,
              ci_hit = switch(goal,
                         ci_hit = ci_hit,
                         close_enough = NA),
              x_tried = x_tried,
              x_out = x_out,
              power_out = power_out,
              nrep_out = nrep_out,
              ci_out = ci_out,
              by_x_out = by_x_out,
              i2 = i2,
              solution_found = solution_found,
              status = status,
              iteration = j,
              x_history = x_history[!is.na(x_history)],
              reject_history = reject_history[!is.na(reject_history)],
              tol = tol,
              delta_tol = delta_tol,
              last_k = last_k,
              what = what,
              goal = goal)
  out
}

#' @noRd

power_algorithm_search_by_curve_pre_i <- function(object,
                                                  x,
                                                  pop_es_name,
                                                  target_power,
                                                  xs_per_trial,
                                                  x_max,
                                                  x_min,
                                                  nrep0,
                                                  R0,
                                                  progress,
                                                  x_include_interval,
                                                  x_interval,
                                                  simulation_progress,
                                                  save_sim_all,
                                                  is_by_x,
                                                  object_by_org,
                                                  power_model,
                                                  start,
                                                  lower_bound,
                                                  upper_bound,
                                                  nls_control,
                                                  nls_args,
                                                  models,
                                                  final_nrep,
                                                  nrep_steps,
                                                  final_R,
                                                  final_xs_per_trial,
                                                  pre_i_xs = 5,
                                                  pre_i_nrep = 50,
                                                  pre_i_R = ifelse(is.null(R0),
                                                                   NULL,
                                                                   min(200, R0)),
                                                  what = "point",
                                                  goal = "ci_hit",
                                                  ci_level = .95,
                                                  tol = .02,
                                                  proxy_power) {

  # goal: ci_hit
  #   what: point
  #     - No need to change
  # goal: close_enough
  #   what: point
  #     - No need to change. The pre-search setup does not use tolerance.
  #   what: ub, lb
  #     - TODO: Use adjusted_power from target_power_adjusted()

  if (progress) {
    cat("\n--- Pre-iteration Crude Search ---\n\n")
    tmp <- format(Sys.time(), "%Y-%m-%d %X")
    cat("- Start at", tmp, "\n")
  }

  # ==== Initial values ====

  x_i <- set_x_range(object,
                    x = x,
                    pop_es_name = pop_es_name,
                    target_power = proxy_power,
                    k = pre_i_xs,
                    x_max = x_max,
                    x_min = x_min)

  # ==== Add x_interval (if x_include_interval) ====

  if (x_include_interval) {
    # Include the lowest and highest values in interval
    x_i <- sort(c(x_interval, x_i))
  }
  x_i <- sort(unique(x_i))

  # ==== Exclude existing values ====

  x0 <- switch(x,
              n = attr(object, "args")$n,
              es = pop_es(object,
                          pop_es_name = pop_es_name))
  x_i <- setdiff(x_i, x0)

  if (progress) {
    x_i_str <- formatC(x_i,
                      digits = switch(x, n = 0, es = 3),
                      format = "f")
    cat("- Value(s) to try: ",
        paste0(x_i_str, collapse = ", "),
        "\n")
    cat("- Crude search with",
        pre_i_nrep,
        "replications\n")
    cat("- R =",
        pre_i_R,
        "\n")
  }

  # ==== Estimate power ====

  # ** by_x_i **
  # The current (i-th) set of values examined,
  # along with their results.
  by_x_i <- switch(x,
                  n = power4test_by_n(object,
                                      n = x_i,
                                      nrep = pre_i_nrep,
                                      R = pre_i_R,
                                      progress = simulation_progress,
                                      save_sim_all = save_sim_all),
                  es = power4test_by_es(object,
                                        pop_es_name = pop_es_name,
                                        pop_es_values = x_i,
                                        nrep = pre_i_nrep,
                                        R = pre_i_R,
                                        progress = simulation_progress,
                                        save_sim_all = save_sim_all))

  # ==== Update by_x (by_x_i) ====

  # Add the input object to the list
  if (is_by_x) {
    # Object is an output of *_by_n() or *_by_es()
    by_x_i <- c(by_x_i,
                object_by_org,
                skip_checking_models = TRUE)
  } else {

    tmp <- switch(x,
            n = as.power4test_by_n(object),
            es = as.power4test_by_es(object,
                                     pop_es_name = pop_es_name)
          )
    by_x_i <- c(by_x_i, tmp,
                skip_checking_models = TRUE)
  }

  if (progress) {
    cat("\n- Rejection Rates:\n\n")
    tmp <- rejection_rates(by_x_i,
                           level = ci_level)
    print(tmp,
          annotation = FALSE)
    cat("\n")
  }

  # ==== Update power curve (fit_i) ====

  # ** fit_i **
  # The current power curve, based on by_x_i
  fit_i <- power_curve(by_x_i,
                      formula = power_model,
                      start = start,
                      lower_bound = lower_bound,
                      upper_bound = upper_bound,
                      nls_control = nls_control,
                      nls_args = nls_args,
                      verbose = progress,
                      models = models)

  if (progress) {
    cat("- Power Curve:\n")
    # Can use the print method of power_curve objects
    print(fit_i)
    cat("\n")
  }

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.
  by_x_1 <- by_x_i

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.
  fit_1 <- fit_i

  # === Initialize the Sequences ===
  # The sequence will be updated when nrep_step is initiated,
  # to successively increase precision and speed by
  # - increasing the number of replication,
  # - increasing the number of resampling, and
  # - decreasing the number of values to try.

  # The sequence of the numbers of replication
  # new_nrep <- rejection_rates(by_x_1,
  #                             all_columns = TRUE)$nrep

  # ==== Generate sequences of values ====

  # new_nrep <- ceiling(mean(new_nrep))
  new_nrep <- nrep0
  nrep_seq <- ceiling(seq(from = new_nrep,
                          to = final_nrep,
                          length.out = nrep_steps + 1))
  final_nrep_seq <- ceiling(seq(from = ceiling(mean(c(new_nrep, final_nrep))),
                                to = final_nrep,
                                length.out = nrep_steps + 1))

  # The sequence of the Rs (for boot and MC CI)
  # R0 <- attr(object, "args")$R
  if (!is.null(R0)) {
    R_seq <- ceiling(seq(from = R0,
                        to = final_R,
                        length.out = nrep_steps + 1))
  } else {
    R_seq <- NULL
  }

  # The sequence of the numbers of values per trial
  xs_per_trial_seq <- ceiling(seq(from = xs_per_trial,
                                  to = final_xs_per_trial,
                                  length.out = nrep_steps + 1))

  # ==== Return the output ====

  out <- list(x_i = x_i,
              by_x_i = by_x_i,
              fit_i = fit_i,
              by_x_1 = by_x_1,
              fit_1 = fit_1,
              nrep_seq = nrep_seq,
              final_nrep_seq = final_nrep_seq,
              R_seq = R_seq,
              xs_per_trial_seq = xs_per_trial_seq)

  return(out)
}

#' @noRd

power_curve_status_message <- function(x,
                                       status_old) {
  # Do not override existing status
  if (!is.null(status_old)) {
    return(status_old)
  }
  status_msgs <- c(
        "Solution found." = 0,
        "Maximum iteration (max_trials) reached." = 1,
        "Changes in the two iterations less than 'delta_tol'." = 2
      )
  status_msgs[status_msgs == x]
}

#' @title Random Variable From a t Distribution
#'
#' @description Generate random numbers
#' from a *t* distribution, rescaled to
#' have user-specified population mean
#' and standard deviation.
#'
#' @details
#' First, specify the parameter `df`
#' and the desired population
#' mean and standard deviation. The
#' random numbers, drawn from the
#' generalized normal distribution by
#' [stats::rt()], will then be
#' rescaled with the desired population
#' mean and standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param df `df` for [stats::rt()].
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rt_rs(n = 5000,
#'            df = 5,
#'            pmean = 3,
#'            psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

rt_rs <- function(n = 10,
                  df = 5,
                  pmean = 0,
                  psd = 1) {
  #  df0 <- match.call()$df
    df0 <- df
    bmean <- 0
    bsd   <- sqrt(df0 / (df0 - 2))
    x <- stats::rt(n, df0)
    x <- pmean + psd * (x - bmean) / bsd
    x
  }

#' @title Sample Size and Effect Size Determination
#'
#' @description It searches by simulation
#' the sample size (given other factors,
#' such as effect sizes) or effect size
#' (given other factors, such as sample
#' size) with power to
#' detect an effect close to a target
#' value.
#'
#' @details
#' This is how to use [x_from_power()]:
#'
#' - Specify the model by [power4test()],
#'   with `do_the_test = FALSE`, and set
#'   the magnitude of the effect sizes
#'   to the minimum levels to detect.
#'
#' - Add the test using [power4test()]
#'   using `test_fun` and `test_args`
#'   (see the help page of [power4test()]
#'   for details). Run it on the
#'   starting sample size or
#'   effect size.
#'
#' - Call [x_from_power()] on the output
#'   of [power4test()] returned from
#'   the previous step. This
#'   function will iteratively repeat
#'   the analysis on either other sample
#'   sizes, or other values for a
#'   selected model parameter (the
#'   effect sizes),
#'   trying to achieve a goal (`goal`) for
#'   a value of interest (`what`).
#'
#' If the `goal` is `"ci_hit"`, the
#' search will try to find a value (a sample
#' size, or a population value of
#' the selected model parameter) with
#' a power level close enough to the
#' target power, defined by having its
#' confidence interval for the power
#' including the target power.
#'
#' If the `goal` is `"close_enough"`,
#' then the search will try to find a
#' value of `x` with its level of
#' power (`"point"`), the upper bound
#' of the confidence interval for this
#' level of power (`"ub"`), or the
#' lower bound of the confidence interval
#' fro this level of power (`"lb"`)
#' "close enough" to the target level of
#' power, defined by having an absolute
#' difference less than the `tolerance`.
#'
#' If several values of `x` (sample
#' size or the population value of
#' a model parameter) have already been
#' examined by [power4test_by_n()] or
#' [power4test_by_es()], the output
#' of these two functions can also be
#' used as `object` by [x_from_power()].
#'
#' Usually, the default values of the
#' arguments should be sufficient.
#'
#' The results can be viewed using
#' [summary()], and the output has
#' a `plot` method ([plot.x_from_power()]) to
#' plot the relation between power and
#' values (of `x`) examined.
#'
#' A detailed illustration on how to
#' use this function for sample size can be found
#' from this page:
#'
#' <https://sfcheung.github.io/power4mome/articles/x_from_power_for_n.html>
#'
#' # Algorithms
#'
#' Three algorithms are currently available,
#' the simple (though sometimes inefficient)
#' bisection method, a method that
#' makes use of the estimated crude power
#' curve, and probabilistic bisection
#' algorithm (Waeber et al., 2013;
#' see Chalmers, 2024, for applying
#' this algorithm to power analysis).
#'
#' Unlike typical root-finding problems,
#' the prediction of the level of power
#' is stochastic. Moreover, the computational
#' cost is high when Monte Carlo or
#' bootstrap confidence intervals are
#' used to do a test because the estimation
#' of the power for one single value of
#' `x` can sometimes take one minute or
#' longer. Therefore, in addition to
#' the simple bisection method, two methods,
#' named *power curve* method
#' (which belongs to the family of
#' surrogate function approximation method
#' reviewed in Chalmers, 2024) and
#' probabilistic bisection method
#' (Chalmers, 2024; Waeber et al., 2013) were also
#' developed for this
#' scenario.
#'
#' ## Bisection Method
#'
#' This method (called informat bisection
#' in Chalmers, 2024), enabled by
#' `algorithm = "bisection"`,
#' basically starts with
#' an interval that probably encloses the
#' value of `x` that meets the goal,
#' and then successively narrows this
#' interval. A point inside this
#' interval, usually the mid-point but
#' can also be approximated by another
#' method (e.g., the power curve method),
#' is used as the estimate.
#' Though simple, there are cases in
#' which it can be slow. Nevertheless,
#' preliminary examination suggests that
#' this method is good enough for
#' scenarios in which only an approximate
#' sample size or range of sample sizes is
#' needed.
#'
#' The internal workflow of
#' this method implemented in
#' [x_from_power()] can be found in
#' this technical vignette: <https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_bisection.html>.
#'
#' ## Power Curve Method
#'
#' This method (belongs to the
#' surrogate function approximation family
#' reviewed in Chalmers, 2024),
#' enabled by `algorithm = "power_curve"`,
#' starts with a crude
#' power curve based on a few points.
#' This tentative model is then used
#' to suggest the values to examine in
#' the next iteration. Unlike some other
#' implementations
#' of this family of methods, the form, not
#' just the parameters, of the
#' model can change across iterations,
#' as more and more data points are
#' available.
#'
#' This method is the default method
#' for some scenarios, such as
#' `x = "es"` with `goal = "ci_hit"`
#' because the relation
#' between the power and the population
#' value of a parameter varies across
#' parameters, unlike the relation
#' between power and sample size, which
#' is monotonic. Therefore,
#' taking into account the working
#' power curve may help finding the
#' desired value of `x`.
#'
#' Before version 0.1.1.33, this
#' method can be used only with
#' the goal `"ci_hit"`. Since
#' version 0.1.1.34, it supports all
#' goals, like the bisection method.
#'
#' The internal workflow of
#' this method implemented in
#' [x_from_power()] can be found in
#' this technical vignette: <https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_power_curve.html>.
#'
#' ## Probabilistic Bisection
#'
#' This method, proposed by Waeber
#' and others (2013) for stochastic
#' root-solving, has been adapted by
#' Chalmers (2024) for power analysis.
#' Similar to bisection, this method
#' starts with an interval, with a
#' initial probability for each value
#' (or range of values), such as sample sizes,
#' as the sample size with the target power.
#' A value is then selected to estimate
#' power (the median, by default).
#' Based on the estimated power
#' for this value,
#' the distribution of probabilities
#' is updated. This process is repeated
#' until some termination criteria are met.
#'
#' Unlike bisection, each iteration can
#' be conducted with a small number of
#' replications (e.g., 50). The method
#' accumulate evidence in a Bayesian
#' approach, and so the certainty of the solution
#' is based on the accumulation of evidence
#' from successive iterations, not on
#' having strong evidence from a few iterations.
#'
#' In `x_from_power`, the version
#' of probabilistic bisection proposed
#' by Waeber et al. (2013) was implemented,
#' with minor changes.
#'
#' Most importantly, when some termination
#' criteria are met, the candidate value
#' of `x` (e.g., a sample size) will be
#' checked using a larger number of
#' replications (set by `final_nrep`)
#' to ensure that the estimated power
#' is indeed close enough to the target
#' power (based on `tolerance` value,
#' determined internally but can also be
#' set directly). If yes, it will be
#' returned as the solution. If no, then
#' the search will continue, until a
#' maximum number of candidate values
#' has been checked.
#'
#' Although the bisection method can be fast
#' in some situations (e.g., when the
#' interval is narrow and the solution
#' happens to be inside the interval),
#' the probabilistic bisection method
#' is nearly guaranteed to converge to
#' the solution (as long as the solution
#' is inside the interval). Therefore,
#' this is the default algorithm in
#' some scenarios.
#'
#' The internal workflow of
#' this method implemented in
#' [x_from_power()] can be found in
#' this technical vignette: <https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_pba.html>.

#'
#' @return
#' The function [x_from_power()]
#' returns an `x_from_power` object,
#' which is a list with the following
#' elements:
#'
#' - `power4test_trials`: The output of
#' [power4test_by_n()] for all sample
#' sizes examined, or of
#' [power4test_by_es()] for all
#' population values of the selected
#' parameter examined.
#'
#' - `rejection_rates`: The output of
#' [rejection_rates()].
#'
#' - `x_tried`: The sample sizes or
#' population values
#' examined.
#'
#' - `power_tried`: The estimated
#' rejection rates for all the values
#' examined.
#'
#' - `x_final`: The sample size or
#' population value in the
#' solution. `NA` if a solution not found.
#'
#' - `power_final`: The estimated power
#' of the value in the solution.
#' `NA` if a solution not found.
#'
#' - `i_final`: The position of the
#' solution in `power4test_trials`.
#' `NA` if a solution not found.
#'
#' - `ci_final`: The confidence interval
#' of the estimated power in the solution.
#' The method is determined
#' by the option `power4mome.ci_method`.
#' If `NULL` or `"wilson"`, Wilson's
#' (1927) method is used. If
#' `"norm"`, normal approximation
#' is used.
#'
#' - `ci_level`: The level of confidence
#' of `ci_final`.
#'
#' - `nrep_final`: The number of
#' replications (`nrep`) when estimating
#' the power in the solution.
#'
#' - `power_curve`: The output of
#' [power_curve()] when estimating the
#' power curve.
#'
#' - `target_power`: The requested
#' target power.
#'
#' - `power_tolerance`: The allowed
#' difference between the solution's
#' estimated power and the target
#' power. Determined by the number
#' of replications and the level of
#' confidence of the confidence intervals.
#'
#' - `x_estimated`: The value
#' (sample size or population value)
#' with the target power, estimated by
#' `power_curve`. This is used, when
#' solution not found, to determine the
#' range of the values to search when
#' calling the function again.
#'
#' - `start`: The time and date when
#' the process started.
#'
#' - `end`: The time and date when the
#' process ended.
#'
#' - `time_spent`: The time spent in
#' doing the search.
#'
#' - `args`: A named list of the arguments
#' of [x_from_power()] used in the search.
#'
#' - `call`: The call when this function
#' is called.
#'
#' @param object A `power4test` object,
#' which is the output of [power4test()].
#' Can also be a `power4test_by_n` object,
#' the output
#' of [power4test_by_n()], or
#' a `power4test_by_es` object, the
#' output of
#' [power4test_by_es()]. For these
#' two types of objects, the attempt
#' with power closest to the
#' `target_power` will be used as
#' `object`, and all other attempts in
#' them will be included in the estimation
#' of subsequent attempts and the final
#' output. Last, it can also be the
#' output of a previous call to
#' [x_from_power()], and the stored
#' trials will be retrieved.
#'
#' @param x For [x_from_power()],
#' `x` set the value to
#' be searched. Can be `"n"`, the sample
#' size, or `"es"`, the population value
#' of a parameter (set by `pop_es_name`).
#' For the `print` method of `x_from_power`
#' objects, this is the output of
#' [x_from_power()].
#'
#' @param pop_es_name The name of the
#' parameter. Required if `x` is `"es"`.
#' See the help page
#' of [ptable_pop()] on the names for
#' the argument `pop_es`.
#'
#' @param target_power The target power,
#' a value greater than 0 and less than
#' one.
#'
#' @param what The value for which is
#' searched: the estimate power (`"point"`),
#' the upper bound of the confidence
#' interval (`"ub"`), or the lower bound
#' of the confidence interval (`"lb"`).
#'
#' @param goal The goal of the search.
#' If `"ci_hit"`, then the goal is to
#' find a value of `x` with the
#' confidence interval of the estimated
#' power including the target power.
#' If `"close_enough"`, then the goal
#' is to find a value of `x` with the
#' value in `what` "close enough" to
#' the target power, defined by having
#' an absolute difference with the
#' target power less than `tolerance`.
#'
#' @param tolerance Used when the goal
#' is `"close_enough"`.
#' If `NULL`, set automatically based
#' on the algorithm used.
#'
#' @param ci_level The level of confidence
#' of the confidence intervals computed
#' for the estimated power. Default is
#' .95, denoting 95%.
#'
#' @param x_interval A vector of
#' two values, the minimum value
#' and the maximum values of `x`, in
#' the search for the values
#' (sample sizes or population values).
#' If `NULL`, default when `x = "es"`,
#' it will be determined internally.
#'
#' @param extendInt Whether `x_interval`
#' can be expanded when estimating the
#' the values to try. The value will
#' be passed to the argument of the
#' same name in [stats::uniroot()].
#' If `x` is `"n"`, then the default
#' value is `"upX"`.
#' That is, a value higher than
#' the maximum in `x_interval` is
#' allowed, if predicted by the tentative
#' model. Otherwise, the default value
#' is `"no"`. See the help page of
#' [stats::uniroot()] for further
#' information.
#'
#' @param progress Logical. Whether
#' the searching progress is reported.
#'
#' @param simulation_progress Logical.
#' Whether the progress in each call
#' to [power4test()], [power4test_by_n()],
#' or [power4test_by_es()]
#' is shown. To be passed to
#' the `progress` argument of these
#' functions.
#' If `NULL`, set automatically based
#' on the algorithm used.
#'
#' @param max_trials The maximum number
#' of trials in searching the value
#' with the target power. Rounded
#' up if not an integer.
#' If `NULL`, set automatically based
#' on the algorithm used.
#'
#' @param final_nrep The number of
#' replications in the final stage,
#' also the maximum number of replications
#' in each call to [power4test()],
#' [power4test_by_n()], or
#' [power4test_by_es()].
#' If `object`
#' is an output of [power4test()]
#' or [x_from_power()] and
#' this argument is not set, `final_nrep`
#' will be set to `nrep` or `final_nrep`
#' stored in
#' `object`.
#'
#' @param final_R The number of
#' Monte Carlo simulation or
#' bootstrapping samples in the final
#' stage. The `R` in calling
#' [power4test()], [power4test_by_n()],
#' or [power4test_by_es()]
#' will be stepped up to this value
#' when approaching the target
#' power. Do not need to be very large
#' because the goal is to estimate
#' power by replications, not for high
#' precision in one single replication.
#' If `object`
#' is an output of [power4test()]
#' or [x_from_power()] and
#' this argument is not set, `final_R`
#' will be set to `R` or `final_R`
#' stored in
#' `object`.
#'
#' @param seed If not `NULL`, [set.seed()]
#' will be used to make the process
#' reproducible. This is not always
#' possible if many stages of
#' parallel processing is involved.
#'
#' @param x_include_interval Logical.
#' Whether the minimum and maximum
#' values in `x_interval` are mandatory
#' to be included in the values
#' to be searched.
#'
#' @param power_curve_args A named
#' list of arguments to be passed
#' [power_curve()] when estimating
#' the relation between power and `x`
#' (sample size or effect size). Please
#' refer to [power_curve()] on available
#' arguments. There is one except:
#' `power_model` is mapped to
#' the `formula` argument of
#' [power_curve()].
#'
#' @param save_sim_all If `FALSE`,
#' the default, the data in each
#' `power4test` object for each
#' value of `x` is not saved,
#' to reduce the size of the output.
#' If set to `TRUE`, the size of the
#' output can be very large in size.
#'
#' @param algorithm The algorithm for
#' finding `x`. Can be `"power_curve"`,
#' `"bisection"`, or
#' `"probabilistic_bisection"`. The default algorithm
#' depends on `x`.
#'
#' @param control A named list of
#' additional
#' arguments to be passed to the
#' algorithm to be used. For advanced
#' users.
#'
#' @param check_es_interval If `TRUE`,
#' the default, and `x` is `"es"`,
#' a conservative probable
#' range of valid values for the selected
#' parameter will be determined, and it
#' will be used instead of `x_interval`.
#' If the range spans both positive and
#' negative values, only the interval
#' of the same sign as the population
#' value in `object` will be used.
#'
#' @param internal_args A named list
#' of internal arguments. For internal
#' testing. Do not use it.
#'
#' @param rejection_rates_args
#' Argument values to be used when
#' [rejection_rates()] is called, used
#' to decide how rejection rates will
#' be estimated. Only one single test
#' is supported by [x_from_power()].
#' Therefore, `merge_all_tests` is
#' always `TRUE` and cannot be changed.
#' The argument `collapse` can be
#' `"all_sig"`, `"at_least_one_sig"`,
#' or `"at_least_k_sig"` (it cannot be
#' `"none"`). Please refer to
#' [rejection_rates()] for other
#' possible arguments. These values,
#' if set, will overwrite any stored
#' settings in `object`.
#'
#' @references
#'
#' Chalmers, R. P. (2024). Solving variables
#' with Monte Carlo simulation experiments:
#' A stochastic root-solving approach.
#' *Psychological Methods*.
#' Advance online publication.
#' \doi{10.1037/met0000689}
#'
#' Waeber, R., Frazier, P. I., &
#' Henderson, S. G. (2013). Bisection
#' search with noisy responses.
#' *SIAM Journal on Control and Optimization*, *51*(3),
#' 2261--2279. \doi{10.1137/120861898}
#'
#' Wilson, E. B. (1927). Probable inference, the law of
#' succession, and statistical inference.
#' *Journal of the American Statistical Association, 22*(158),
#' 209-212. \doi{10.1080/01621459.1927.10502953}
#'
#' @seealso [power4test()], [power4test_by_n()],
#' and [power4test_by_es()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' # Generate the datasets
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 2345)
#'
#' # Do a test
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters,
#'                        test_args = list(pars = "m~x"))
#'
#' # Determine the sample size with a power of .80 (default)
#'
#' # In real analysis, to have more stable results:
#' # - Use a larger final_nrep (e.g., 400).
#'
#' # If the default values are OK, this call is sufficient:
#' # power_vs_n <- x_from_power(test_out,
#' #                            x = "n",
#' #                            seed = 4567)
#' power_vs_n <- x_from_power(test_out,
#'                            x = "n",
#'                            progress = TRUE,
#'                            target_power = .80,
#'                            final_nrep = 5,
#'                            max_trials = 1,
#'                            seed = 1234)
#' summary(power_vs_n)
#' plot(power_vs_n)
#'
#'
#' @importFrom graphics abline arrows par points text title
#'
#' @export
x_from_power <- function(object,
                         x = arg_x_from_power(object, "x", arg_in = "call") %||% "n",
                         pop_es_name = arg_x_from_power(object, "pop_es_name", arg_in = "call"),
                         target_power = .80,
                         what = arg_x_from_power(object, "what") %||% "point",
                         goal = arg_x_from_power(object, "goal") %||% {switch(what,
                                       point = "ci_hit",
                                       ub = "close_enough",
                                       lb = "close_enough")},
                         ci_level = .95,
                         tolerance = NULL,
                         x_interval = switch(x,
                                             n = c(50, 2000),
                                             es = NULL),
                         extendInt = NULL,
                         progress = TRUE,
                         simulation_progress = NULL,
                         max_trials = NULL,
                         final_nrep = attr(object, "args")$nrep %||% (object$nrep_final %||% 400),
                         final_R = attr(object, "args")$R %||% (object$args$final_R %||% 1000),
                         seed = NULL,
                         x_include_interval = FALSE,
                         check_es_interval = TRUE,
                         power_curve_args = list(power_model = NULL,
                                                 start = NULL,
                                                 lower_bound = NULL,
                                                 upper_bound = NULL,
                                                 nls_control = list(),
                                                 nls_args = list()),
                         save_sim_all = FALSE,
                         algorithm = NULL,
                         control = list(),
                         internal_args = list(),
                         rejection_rates_args = list(collapse = "all_sig",
                                                     at_least_k = 1,
                                                     p_adjust_method = "none",
                                                     alpha = .05)
                         ) {

  # Inputs
  # - Target power
  # Steps
  # - Get the initial sample size and power.
  # - Find the k sample sizes to explore.
  # - Fit the curve.
  # - See if the target power is in this range.
  #   - NO: Expand.
  #   - YES: Narrow.
  # - Stop when convergence achieved.
  # Outputs
  # - Initial power4test object.
  # - Final power4test object.
  # - Final model by nls.

  what <- match.arg(what,
                    c("point", "ub", "lb"))
  goal <- match.arg(goal,
                    c("ci_hit", "close_enough"))

  internal_args0 <- list(keep_algorithm = TRUE)

  internal_args <- utils::modifyList(
                      internal_args0,
                      internal_args
                    )

  # ==== Update the object if x_from_power ====

  if (inherits(object, "x_from_power")) {
    # Make object to be match some arguments
    if ((object$what != what) ||
        (object$goal != goal)) {
      # what or goal changed.
      object$solution_found <- FALSE
    }
    object$what <- what
    object$goal <- goal
    # No need to update them.
    # They will be updated using
    # what and goal in this call
    # object$call$what <- what
    # object$call$goal <- goal
  }

  # ==== Fix some arguments ====

  changed_to_point <- FALSE
  if ((goal == "ci_hit") &&
      (what != "point")) {
    changed_to_point <- TRUE
    what <- "point"
  }

  x <- match.arg(x,
                 choices = c("n", "es"))

  if (!is.null(algorithm)) {
    algorithm <- match.arg(algorithm,
                          c("bisection",
                            "power_curve",
                            "probabilistic_bisection"))
  }

  # merge_all_tests is always TRUE
  # Ignored by rejection_rates() if there is only one test
  tmp <- list(collapse = "all_sig",
              at_least_k = 1,
              p_adjust_method = "none",
              alpha = .05)
  rejection_rates_args <- utils::modifyList(
                            tmp,
                            rejection_rates_args
                          )
  rejection_rates_args$merge_all_tests <- TRUE

  # ==== Set default algorithm ====

  # what and goal take precedence

  if (goal == "ci_hit") {
    if (is.null(algorithm)) {
      if ((final_nrep >= 1000) &&
          (x == "n")) {
        algorithm <- "probabilistic_bisection"
      } else {
        algorithm <- switch(x,
                            n = "bisection",
                            es = "power_curve")
      }
    }
  }

  changed_to_bisection <- FALSE
  if (goal == "close_enough") {
    # internal_args$keep_algorithm is now ignored
    if (is.null(algorithm)) {
      if (final_nrep >= 1000) {
        algorithm <- "probabilistic_bisection"
      } else {
        algorithm <- "bisection"
      }
    } else {
      algorithm <- match.arg(algorithm,
                             c("bisection",
                               "probabilistic_bisection",
                               "power_curve"))
    }
  }

  if (algorithm == "probabilistic_bisection") {
    # PBA does not support ci_hit separately.
    # Instead, close_enough is used to emulated ci_hit
    goal <- "close_enough"
  }


  # what: The value to be examined.
  # goal:
  # - ci_hit: Only relevant for what == "point"
  # - close_enough: Can be used for all what.

  a <- reject_ci_wilson(
          nreject = ceiling(target_power * final_nrep),
          nvalid = final_nrep,
          level  = ci_level)
  power_tolerance_in_interval <- max(abs(target_power - a))
  power_tolerance_in_final <- max(abs(target_power - a))

  # ==== Sanity checks ====

  if (target_power <= 0 || target_power >= 1) {
    stop("'target power' (",
         target_power,
         ") not between 0 and 1.")
  }

  if (x == "n") {
    if (min(x_interval) < 0) {
      stop("The minimum value of 'n_interval' cannot be negative")
    }
  }

  if ((length(x_interval) != 2) &&
      !is.null(x_interval)) {
    stop("'x_interval' must be a vector with exactly two values.")
  }
  if ((x_interval[2] < x_interval[1]) &&
      !is.null(x_interval)) {
    stop("'x_interval' must be of the form c(minimum, maximum).")
  }

  if (is.null(extendInt)) {
    if (x == "n") {
      extendInt <- match.arg(extendInt,
                             choices = c("yes", "no", "upX", "downX"))
    }
    if (x == "es") {
      extendInt <- match.arg(extendInt,
                             choices = c("no", "yes", "upX", "downX"))
    }
  }

  if (!is.null(max_trials)) {
    max_trials <- ceiling(max_trials)
    if (max_trials < 1) {
      stop("'max_trials' must be at least 1 (after rounding, if necessary).")
    }
  }

  # ==== Set tolerance if NULL ====

  if (is.null(tolerance)) {
    tolerance <- set_tolerance(
                    algorithm = algorithm,
                    target_power = target_power,
                    goal = goal,
                    what = what,
                    final_nrep = final_nrep,
                    ci_level = ci_level
                  )
  }

  # ==== Set max_trials if NULL ====

  if (is.null(max_trials)) {
    max_trials <- set_max_trials(
                    algorithm = algorithm
                  )
  }

  # ==== Set max_trials if NULL ====

  if (is.null(simulation_progress)) {
    simulation_progress <- switch(
      algorithm,
      probabilistic_bisection = FALSE,
      TRUE
    )
  }

  if (!is.null(seed)) {
    set.seed(seed)
  }

  # ==== Process the object ====

  # Is it a compatible x_from_power object?
  # - If yes and solution not found,
  #   - extract the stored by_x object
  # - If yes and solution found,
  #   - The object is returned as is.

  is_x_from_power <- FALSE
  if (inherits(object, "x_from_power")) {

    # ==== x_from_power object ====

    # Throw an error if incompatible

    # TODO:
    # - Will run if goal and/or what changed

    check_x_from_power_as_input(object,
                                x = x,
                                pop_es_name = pop_es_name,
                                final_nrep = final_nrep,
                                ci_level = ci_level,
                                rejection_rates_args = rejection_rates_args)

    # Check these for compatibility:

    is_x_from_power <- TRUE
    if (object$solution_found &&
        (object$nrep_final == final_nrep) &&
        (object$ci_level == ci_level) &&
        (object$target_power == target_power)) {
      if (progress) {
        cat("\n--- Solution Already Found ---\n\n")
        cat("Solution already found in object. It is returned as is.\n")
      }
      return(object)
    }
    object <- object$power4test_trials
  }

  # Is it a by_x object??
  # - If yes and solution not found,
  #   - extract the power4test object closest to the solution
  # - If yes and solution found,
  #   - This is possible because the by_x object
  #     may be generated directly by user,
  #     not extracted from an x_from_power object.

  is_by_x <- FALSE
  if (inherits(object, "power4test_by_n") ||
      inherits(object, "power4test_by_es")) {

    # ==== by_x object ====

    is_by_x <- TRUE
    object_by_org <- object

    # Update rejection_rates_args

    object_by_org <- set_rejection_rates_args_by_x(
                        object_by_org,
                        rejection_rates_args = rejection_rates_args
                      )

    # Whether a solution exists will be checked later

    i_org <- find_solution(
               object_by_org,
               target_power = target_power,
               ci_level = ci_level,
               what = what,
               tol = tolerance,
               goal = goal,
               final_nrep = 1,
               closest_ok = TRUE,
               if_ties = "min")
    object <- object_by_org[[i_org]]
  } else {
    object_by_org <- NA
  }

  # ==== Update rejection_rates_args ====

  tmp <-  attr(object, "args")
  tmp2 <- tmp$rejection_rates_args
  tmp2 <- utils::modifyList(
                tmp2,
                rejection_rates_args,
                keep.null = TRUE
              )
  tmp$rejection_rates_args <- tmp2
  attr(object, "args") <- tmp
  rm(tmp)

  # The object to be used below is always a power4test object

  time_start <- Sys.time()

  if (progress) {
    cat("\n--- Setting ---\n\n")

    cat("Algorithm: ", algorithm, "\n")
    cat("Goal: ", goal, "\n")
    tmp <- switch(what,
                  point = "(Estimated Power)",
                  ub = "(Upper bound of the confidence interval)",
                  lb = "(Lower bound of the confidence interval)")
    cat("What: ", what, " ", tmp, "\n")

    if (changed_to_bisection) {
      catwrap(paste0("Note: Only bisection is supported when goal is 'close_enough'. ",
                     "Algorithm switched automatically to bisection."),
              exdent = 0)
    }
    if (changed_to_point) {
      cat("\n")
      catwrap(paste0("Note: The goal 'ci_hit' only supports 'point'; ",
                     "'what' is changed to 'point'."),
              exdent = 0)
    }

    if (progress) {
      cat("\n--- Progress  ---\n\n")
      catwrap("- Set 'progress = FALSE' to suppress displaying the progress.",
              exdent = 2)
      if (simulation_progress) {
        catwrap(paste0("- Set 'simulation progress = FALSE' ",
                       "to suppress displaying the progress ",
                       "in the simulation."),
                exdent = 2)
      }
    }
  }

  # Initialize these flags because a solution
  # may already be present before doing the search

  ci_hit <- FALSE
  solution_found <- FALSE
  status <- NULL
  technical <- NULL

  # === Check Existing Solution ===

  # If the input object is a by_x object,
  # - Check if the solution is already in one of the
  #   power4test object.
  #   - If yes, skip the search and create the
  #     x_from_power object.

  # ==== Solution already in by_x? ====

  if (is_by_x) {

    i_org_hit <- find_solution(
                   object_by_org,
                   ci_level = ci_level,
                   target_power = target_power,
                   what = what,
                   tol = tolerance,
                   goal = goal,
                   final_nrep = final_nrep,
                   closest_ok = FALSE,
                   if_ties = "min")

    if (!is.na(i_org_hit) && !is.null(i_org_hit)) {

      # ==== Solution in by_x. Skip the search ====

      # Solution already in the input.
      # DO not do the search

      if (progress) {
        cat("\n--- Solution Already Found ---\n\n")
        cat("Solution already found in the object. Search will be skipped.\n")
      }

      # ==== Create the output ====

      # Prepare the objects as if the search has completed

      ci_hit <- ifelse(goal == "ci_hit",
                       TRUE,
                       NA)
      solution_found <- TRUE
      i2 <- i_org_hit
      if (is_x_from_power || is_by_x) {
        by_x_1 <- object_by_org
        tmp1 <- rejection_rates(object_by_org,
                                level = ci_level,
                                all_columns = TRUE)
        fit_1 <- power_curve(by_x_1,
                             formula = power_curve_args$power_model,
                             start = power_curve_args$start,
                             lower_bound = power_curve_args$lower_bound,
                             upper_bound = power_curve_args$upper_bound,
                             nls_control = power_curve_args$nls_control,
                             nls_args = power_curve_args$nls_args,
                             verbose = progress)
        ci_hit <- ci_hit
        x_tried <- get_x_tried(tmp1,
                               x = x)
        x_out <- switch(x,
                        n = tmp1$n[i2],
                        es = tmp1$es[i2])
        power_out <- tmp1$reject[i2]
        nrep_out <- tmp1$nrep[i2]
        ci_out <- unlist(tmp1[i2, c("reject_ci_lo", "reject_ci_hi")])
        by_x_out <- by_x_1[[i2]]

        # ==== Reconstruct technical ====

        # TODO:
        # - (Not urgent) Not an ideal solution.
        #   Other functions should not rely on $technical
        #   because the search may have been skipped.

        technical <- list(tol = tolerance)

      }
    }
  }

  # x_interval is the actual range of values
  # to be searched, not necessarily the one
  # set by users.

  # ==== Fix x_interval ====

  if (solution_found) {
    x_interval <- range(x_tried)
  } else {
    if (check_es_interval) {

      # ==== Find probable es interval ====

      x_interval <- fix_es_interval(object = object,
                                    x = x,
                                    pop_es_name = pop_es_name,
                                    x_interval = x_interval,
                                    progress = progress)
    }
  }

  x_max <- max(x_interval)
  x_min <- min(x_interval)

  # ==== Check the tests ====
  tmp <- object$test_all[[1]][[1]]$test_results
  if (!is.null(dim(tmp))) {
    if (dim(tmp)[1] > 1) {
      tmp2 <- utils::capture.output(print(tmp))
      tmp3 <- paste(
                c("Does not support a test with more than one result:",
                tmp2),
                collapse = "\n")
      stop(tmp3)
    }
  }

  # ==== Call the algorithm ====

  if ((algorithm == "power_curve") && !solution_found) {

    a_out <- do.call(alg_power_curve,
      list(
        object = object,
        x = x,
        pop_es_name = pop_es_name,
        target_power = target_power,
        x_max = x_max,
        x_min = x_min,
        progress = progress,
        x_include_interval = x_include_interval,
        x_interval = x_interval,
        simulation_progress = simulation_progress,
        save_sim_all = save_sim_all,
        is_by_x = is_by_x,
        object_by_org = object_by_org,
        power_model = power_curve_args$power_model,
        start = power_curve_args$start,
        lower_bound = power_curve_args$lower_bound,
        upper_bound = power_curve_args$upper_bound,
        nls_control = power_curve_args$nls_control,
        nls_args = power_curve_args$nls_args,
        models = power_curve_args$models,
        final_nrep = final_nrep,
        final_R = final_R,
        max_trials = max_trials,
        ci_level = ci_level,
        extendInt = extendInt,
        power_tolerance_in_interval = power_tolerance_in_interval,
        power_tolerance_in_final = power_tolerance_in_final,
        what = what,
        goal = goal,
        tol = tolerance,
        variants = control
      ))

    by_x_1 <- a_out$by_x_1
    fit_1 <- a_out$fit_1
    ci_hit <- a_out$ci_hit
    x_tried <- a_out$x_tried
    x_out <- a_out$x_out
    power_out <- a_out$power_out
    nrep_out <- a_out$nrep_out
    ci_out <- a_out$ci_out
    by_x_out <- a_out$by_x_out
    i2 <- a_out$i2
    solution_found <- a_out$solution_found
    status <- a_out$status
    technical <- list(iteration = a_out$iteration,
                      x_history = a_out$x_history,
                      reject_history = a_out$reject_history,
                      delta_tol = a_out$delta_tol,
                      last_k = a_out$last_k,
                      tol = a_out$tol)

    rm(a_out)

  }

  if ((algorithm == "probabilistic_bisection") && !solution_found) {

    a_out <- do.call(alg_prob_bisection,
      list(
          object = object,
          x = x,
          pop_es_name = pop_es_name,
          target_power = target_power,
          x_max = x_max,
          x_min = x_min,
          progress = progress,
          x_include_interval = x_include_interval,
          x_interval = x_interval,
          simulation_progress = simulation_progress,
          save_sim_all = save_sim_all,
          is_by_x = is_by_x,
          object_by_org = object_by_org,
          final_nrep = final_nrep,
          final_R = final_R,
          ci_level = ci_level,
          extendInt = extendInt,
          max_trials = max_trials,
          R = attr(object, "args")$R,
          digits = 3,
          what = what,
          goal = goal,
          tol = tolerance,
          variants = control
        ))

    by_x_1 <- a_out$by_x_1
    fit_1 <- a_out$fit_1
    ci_hit <- a_out$ci_hit
    x_tried <- a_out$x_tried
    x_out <- a_out$x_out
    power_out <- a_out$power_out
    nrep_out <- a_out$nrep_out
    ci_out <- a_out$ci_out
    by_x_out <- a_out$by_x_out
    i2 <- a_out$i2
    solution_found <- a_out$solution_found
    status <- a_out$status
    technical <- list(iteration = a_out$iteration,
                      x_history = a_out$x_history,
                      x_interval_history = a_out$x_interval_history,
                      f_interval_history = a_out$f_interval_history,
                      reject_history = a_out$reject_history,
                      reject_by_power_curve_history = a_out$reject_by_power_curve_history,
                      f_history = a_out$f_history,
                      dfun_history = a_out$dfun_history,
                      p_c_history = a_out$p_c_history,
                      hdr_x_history = a_out$hdr_x_history,
                      hdr_power_history = a_out$hdr_power_history,
                      final_check_history = a_out$final_check_history,
                      seed_history = a_out$seed_history,
                      time_passed_history = a_out$time_passed_history,
                      delta_tol = a_out$delta_tol,
                      last_k = a_out$last_k,
                      tol = a_out$tol,
                      f_power = a_out$f_power,
                      f_goal = a_out$f_goal,
                      hdr_power_tol = a_out$hdr_power_tol)

    rm(a_out)

  }


  if ((algorithm == "bisection") && !solution_found) {

    a_out <- do.call(alg_bisection,
      list(
          object = object,
          x = x,
          pop_es_name = pop_es_name,
          target_power = target_power,
          x_max = x_max,
          x_min = x_min,
          progress = progress,
          x_include_interval = x_include_interval,
          x_interval = x_interval,
          simulation_progress = simulation_progress,
          save_sim_all = save_sim_all,
          is_by_x = is_by_x,
          object_by_org = object_by_org,
          final_nrep = final_nrep,
          final_R = final_R,
          ci_level = ci_level,
          extendInt = extendInt,
          max_trials = max_trials,
          R = attr(object, "args")$R,
          digits = 3,
          what = what,
          goal = goal,
          tol = tolerance,
          variants = control
        ))

    by_x_1 <- a_out$by_x_1
    fit_1 <- a_out$fit_1
    ci_hit <- a_out$ci_hit
    x_tried <- a_out$x_tried
    x_out <- a_out$x_out
    power_out <- a_out$power_out
    nrep_out <- a_out$nrep_out
    ci_out <- a_out$ci_out
    by_x_out <- a_out$by_x_out
    i2 <- a_out$i2
    solution_found <- a_out$solution_found
    status <- a_out$status
    technical <- list(iteration = a_out$iteration,
                      x_history = a_out$x_history,
                      x_interval_history = a_out$x_interval_history,
                      f_interval_history = a_out$f_interval_history,
                      reject_history = a_out$reject_history,
                      f_history = a_out$f_history,
                      delta_tol = a_out$delta_tol,
                      last_k = a_out$last_k,
                      tol = a_out$tol)

    rm(a_out)

  }

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.


  if (progress) {
    cat("\n========== Final Stage ==========\n\n")
    tmp <- format(Sys.time(), "%Y-%m-%d %X")
    cat("- Start at", tmp, "\n")

    cat("- Rejection Rates:\n\n")
    tmp <- rejection_rates(by_x_1,
                           level = ci_level)
    print(tmp)
    cat("\n")
    cat("- Estimated Power Curve:\n\n")
    print(fit_1)
    cat("\n")
  }

  # ==== Is solution found ====

  # Is solution found?
  # - The maximum number of replications reached.

  if (goal == "ci_hit") {

    # ==== Goal: ci_hit ====

    if (isTRUE(ci_hit) && (nrep_out == final_nrep)) {

      # ==== Solution found ====

      # ** x_final, by_x_final, power_final, ci_final, nrep_final, i_final **
      # The solution.

      # ==== Create _final objects ====

      x_final <- x_out
      by_x_final <- by_x_out
      power_final <- power_out
      ci_final <- ci_out
      nrep_final <- nrep_out
      i_final <- i2
    } else {

      # ==== Solution not found ====

      # No solution found.
      # Set the NAs to denote this.

      # ==== Set _final objects to NA====

      x_final <- NA
      by_x_final <- NA
      power_final <- NA
      ci_final <- NA
      nrep_final <- NA
      i_final <- NA
    }
  }

  if (goal == "close_enough") {

    # ==== Goal: close_enough ====

    if (isTRUE(solution_found) && (nrep_out == final_nrep)) {

      # ==== Solution found ====

      # ** x_final, by_x_final, power_final, ci_final, nrep_final, i_final **
      # The solution.

      # ==== Create _final objects ====

      x_final <- x_out
      by_x_final <- by_x_out
      power_final <- power_out
      ci_final <- ci_out
      nrep_final <- nrep_out
      i_final <- i2
    } else {

      # ==== Solution not found ====

      # No solution found.
      # Set the NAs to denote this.

      # ==== Set _final objects to NA====

      x_final <- NA
      by_x_final <- NA
      power_final <- NA
      ci_final <- NA
      nrep_final <- NA
      i_final <- NA
    }
  }

  # ==== Compute extrapolated x? ====

  # ** x_x **
  # The estimated value based on power_curve.
  # Used as a suggestion when no solution was found.
  x_x <- NA

  if (solution_found) {

    # ==== Yes only if solution_found ====

    x_tried <- switch(x,
                      n = as.numeric(names(by_x_1)),
                      es = sapply(by_x_1,
                                  \(x) {attr(x, "pop_es_value")},
                                  USE.NAMES = FALSE))
    x_x <- estimate_x_range(power_x_fit = fit_1,
                            x = x,
                            target_power = target_power,
                            k = 1,
                            tolerance = 0,
                            power_min = .01,
                            power_max = .99,
                            interval = x_interval,
                            extendInt = "yes",
                            x_to_exclude = x_tried)
  }

  if (progress) {
    if (solution_found && (nrep_out == final_nrep)) {
      cat("\n")
      x_out_str <- formatC(x_out,
                           digits = switch(x, n = 0, es = 4),
                           format = "f")
      cat("- Final Value:", x_out_str, "\n\n")
      cat("- Final Estimated Power:",
          formatC(power_final, digits = 4, format = "f"), "\n")
      cat("- Confidence Interval: [",
          paste0(formatC(ci_final, digits = 4, format = "f"), collapse = "; "),
          "]\n", sep = "")
      cat("- CI Level: ",
          formatC(ci_level*100, digits = 2, format = "f"), "%", "\n", sep = "")
    } else {
      cat("\n")
      cat("- None of the value(s) examined",
          "in the interval meet the goal.\n")
      if (!is.na(x_x) &&
          (goal == "ci_hit")) {
        x_x_str <- formatC(x_x,
                           digits = switch(x, n = 0, es = 4),
                           format = "f")
        cat("- The estimated required value is ",
            x_x_str,
            ".\n", sep = "")
      }
      cat("\nTry changing the setup, such as:\n")
      tmp <- switch(x,
                    n = paste0("[",
                               paste0(x_interval, collapse = ", "),
                               "]"),
                    es = paste0("[",
                               paste0(formatC(x_interval,
                                              digits = 3,
                                              format = "f"),
                                      collapse = ", "),
                               "]"))
      catwrap(paste0("- Changing the 'seed' to another value."),
              exdent = 2)
      catwrap(paste0("- Changing 'x_interval' to a wider range to examine. ",
                     "(Current 'x_interval' is ",
                     tmp, ".)"),
              exdent = 2)
      catwrap(paste0("- Increasing the maximum number of trials by ",
                     "setting 'max_trials' to a larger number. ",
                     "(Current 'max_trials' is ",
                     max_trials,
                     ".)"),
              exdent = 2)
      if (goal == "close_enough") {
        catwrap(paste0("- Increasing the tolerance for the goal 'close_enough'. ",
                      "(Current 'tolerance' is ",
                      tolerance,
                      ".)"),
                exdent = 2)
      }
    }
  }

  # ==== Finalize the Output ====

  my_call <- as.list(match.call())[-1]
  args <- formals()
  args <- utils::modifyList(args,
                            my_call)
  args$object <- NULL
  reject_1 <- rejection_rates(by_x_1,
                              level = ci_level)
  time_end <- Sys.time()

  out <- list(x = x,
              pop_es_name = pop_es_name,
              power4test_trials = by_x_1,
              rejection_rates = reject_1,
              x_tried = switch(x,
                               n = reject_1$n,
                               es = reject_1$es),
              power_tried = reject_1$reject,
              x_final = x_final,
              power_final = power_final,
              i_final = i_final,
              ci_final = ci_final,
              ci_level = ci_level,
              nrep_final = nrep_final,
              power_curve = fit_1,
              target_power = target_power,
              power_tolerance = power_tolerance_in_final,
              x_estimated = x_x,
              start = time_start,
              end = time_end,
              time_spent = difftime(time_end, time_start),
              solution_found = solution_found,
              what = what,
              goal = goal,
              args = args,
              status = status,
              technical = technical,
              algorithm = algorithm,
              call = match.call(),
              rejection_rates_args = rejection_rates_args)
  class(out) <- c("x_from_power", class(out))
  return(out)
}

#' @rdname x_from_power
#'
#' @details
#'
#' The function [n_from_power()] is just
#' a wrapper of [x_from_power()], with
#' `x` set to `"n"`.
#'
#' @export
n_from_power <- function(object,
                         pop_es_name = NULL,
                         target_power = .80,
                         what = formals(x_from_power)$what,
                         goal = formals(x_from_power)$goal,
                         ci_level = .95,
                         tolerance = NULL,
                         x_interval = c(50, 2000),
                         extendInt = NULL,
                         progress = TRUE,
                         simulation_progress = NULL,
                         max_trials = NULL,
                         final_nrep = formals(x_from_power)$final_nrep,
                         final_R = formals(x_from_power)$final_R,
                         seed = NULL,
                         x_include_interval = FALSE,
                         check_es_interval = TRUE,
                         power_curve_args = list(power_model = NULL,
                                                 start = NULL,
                                                 lower_bound = NULL,
                                                 upper_bound = NULL,
                                                 nls_control = list(),
                                                 nls_args = list()),
                         save_sim_all = FALSE,
                         algorithm = NULL,
                         control = list(),
                         internal_args = list(),
                         rejection_rates_args = list(collapse = "all_sig",
                                                     at_least_k = 1,
                                                     p_adjust_method = "none",
                                                     alpha = .05)
                         ) {
  what <- match.arg(eval(what),
                    c("point", "ub", "lb"))
  goal <- match.arg(eval(goal),
                    c("ci_hit", "close_enough"))
  if ((goal == "ci_hit") &&
      (what != "point")) {
    what <- "point"
  }
  my_call <- match.call()
  my_call$x <- "n"
  my_call$what <- what
  my_call$goal <- goal
  my_call$final_nrep <- eval(final_nrep)
  my_call$final_R <- eval(final_R)
  my_call[[1]] <- quote(power4mome::x_from_power)
  out <- eval(my_call,
              envir = parent.frame())
  out
}

#' @rdname x_from_power
#'
#' @details
#'
#' The function [n_region_from_power()] is just
#' a wrapper of [x_from_power()], with
#' `x` set to `"n"`, with two passes, one
#' with `what = "ub"` and one with
#' `what = "lb"`.
#'
#' @return
#' The function [n_region_from_power()]
#' returns a named list of two output of
#' [n_from_power()], of the class
#' `n_region_from_power`. The output
#' with `what = "ub"` is named `"below"`,
#' and the output with `what = "lb"` is
#' namd `"above"`.
#'
#' @export
n_region_from_power <- function(
                         object,
                         pop_es_name = NULL,
                         target_power = .80,
                         ci_level = .95,
                         tolerance = NULL,
                         x_interval = c(50, 2000),
                         extendInt = NULL,
                         progress = TRUE,
                         simulation_progress = NULL,
                         max_trials = NULL,
                         final_nrep = formals(x_from_power)$final_nrep,
                         final_R = formals(x_from_power)$final_R,
                         seed = NULL,
                         x_include_interval = FALSE,
                         check_es_interval = TRUE,
                         power_curve_args = list(power_model = NULL,
                                                 start = NULL,
                                                 lower_bound = NULL,
                                                 upper_bound = NULL,
                                                 nls_control = list(),
                                                 nls_args = list()),
                         save_sim_all = FALSE,
                         algorithm = NULL,
                         control = list(),
                         internal_args = list(),
                         rejection_rates_args = list(collapse = "all_sig",
                                                     at_least_k = 1,
                                                     p_adjust_method = "none",
                                                     alpha = .05)
                         ) {
  my_call <- match.call()
  my_call$final_nrep <- eval(final_nrep)
  my_call$final_R <- eval(final_R)
  my_call$x <- "n"
  my_call[[1]] <- quote(power4mome::x_from_power)
  if (progress) {
    tmp <- strwrap(paste0("Find the approximate region with power ",
                         "significantly below ",
                         target_power,
                         " ..."))
    cat("\n=========== Phase 1: Upper Bound ===========\n\n")
    cat(tmp, sep = "\n")
  }
  my_call$what <- "ub"
  my_call$goal <- "close_enough"
  out_ub <- eval(
              my_call,
              envir = parent.frame())
  if (progress) {
    tmp <- strwrap(paste0("Find the approximate region with power ",
                         "significantly above ",
                         target_power,
                         " ..."))
    cat("\n=========== Phase 2: Lower Bound ===========\n\n")
    cat(tmp, sep = "\n")
  }
  my_call$what <- "lb"
  my_call$goal <- "close_enough"
  tmp <- out_ub
  tmp$what <- "lb"
  tmp$goal <- "close_enough"
  tmp$solution_found <- FALSE
  tmp$call$what <- "lb"
  tmp$call$goal <- "close_enough"
  my_call2 <- my_call
  my_call2$object <- tmp
  out_lb <- eval(
              my_call2,
              envir = parent.frame())
  out_lb$call <- my_call
  out <- list(
          below = out_ub,
          above = out_lb
        )
  class(out) <- c("n_region_from_power", class(out))
  attr(out, "call") <- match.call()
  out
}


#' @rdname x_from_power
#'
#'
#' @param digits The number of digits
#' after the decimal when printing
#' the results.
#'
#' @param call Logical. Whether the
#' call is printed.
#'
#' @param ... Optional arguments.
#' Not used for now.
#'
#' @details
#' The `print` method only prints
#' basic information. Call the
#' `summary` method of `x_from_power` objects
#' ([summary.x_from_power()]) and its
#' `print` method for detailed results
#'
#' @return
#' The `print`-method of `x_from_power`
#' objects returns the object `x`
#' invisibly.
#' It is called for its side effect.
#'
#' @export
print.x_from_power <- function(x,
                               digits = 3,
                               call = TRUE,
                               ...) {
  my_call <- x$call
  if (call) {
    cat("Call:\n")
    print(my_call)
  }
  if (!is.symbol(my_call$object)) {
    my_call$object <- as.symbol("<hidden>")
  }
  if (!is.symbol(my_call[[1]])) {
    my_call[[1]] <- as.symbol("x_from_power")
  }
  cat("\n")
  solution_found <- !is.na(x$x_final)
  predictor <- x$x

  goal <- x$goal
  what <- x$what
  algorithm <- x$algorithm
  ci_level_str <- paste0(formatC(
                          x$ci_level * 100,
                          digits = 2,
                          format = "f"),
                        "%")

  tmp1 <- c("Predictor(x):" =
            switch(predictor,
             n = "Sample Size",
             es = "Effect Size"))
  tmp1b <- c("Parameter:" =
            switch(predictor,
             n = "N/A",
             es = x$pop_es_name))
  tmp2 <- c("goal:" = goal)
  tmp3 <- c("what:" = what)
  tmp4 <- c("algorithm:" = algorithm)
  tmp5 <- c("Level of confidence:" = ci_level_str)
  tmp6 <- c("Target Power:" =
              formatC(x$target_power, digits = digits, format = "f"))

  tmp <- data.frame("Setting" = c(
      tmp1,
      tmp1b,
      tmp2,
      tmp3,
      tmp4,
      tmp5,
      tmp6
    ))

  print(tmp)

  if (solution_found) {
    x_final_str <- formatC(x$x_final,
                           digits = switch(predictor,
                                           n = 0,
                                           es = digits),
                           format = "f")
    cat("\n- Final Value of",
        switch(x$x,
               n = " Sample Size (n): ",
               es = paste0("'", x$pop_es_name, "': ")),
        x_final_str,
        "\n\n",
        sep = "")
    ci_str <- paste0(
        "[",
        formatC(x$ci_final[1], digits = digits, format = "f"),
        ", ",
        formatC(x$ci_final[2], digits = digits, format = "f"),
        "]")
    cat("- Final Estimated Power (CI): ",
        formatC(x$power_final, digits = digits, format = "f"),
        " ",
        ci_str,
        "\n",
        sep = "")
  } else {
    cat("\n- Solution not found.\n")
  }
  cat("\nCall `summary()` for detailed results.\n")
  invisible(x)
}


#' @rdname x_from_power
#'
#' @return
#' The `print`-method of `x_from_power_region`
#' objects returns the object `x`
#' invisibly.
#' It is called for its side effect.
#'
#' @export
print.n_region_from_power <- function(
                              x,
                              digits = 3,
                              ...) {
  my_call <- attr(x, "call")
  cat("Call:\n")
  if (!is.symbol(my_call$object)) {
    my_call$object <- as.symbol("<hidden>")
  }
  if (!is.symbol(my_call[[1]])) {
    my_call[[1]] <- as.symbol("n_region_from_power")
  }
  print(my_call)
  cat("\n")
  x_below <- x$below
  x_above <- x$above
  solution_found_below <- !is.na(x_below$x_final)
  solution_found_above <- !is.na(x_above$x_final)
  predictor <- x_below$x

  goal <- x_below$goal
  what <- x_below$what
  algorithm <- x_below$algorithm
  ci_level_str <- paste0(formatC(
                          x_below$ci_level * 100,
                          digits = 2,
                          format = "f"),
                        "%")

  tmp1 <- c("Predictor(x)" = "Sample Size")
  tmp2 <- c("Goal:" = "Power significantly below or above the target")
  tmp4 <- c("algorithm:" = algorithm)
  tmp5 <- c("Level of confidence:" = ci_level_str)
  target_power_str <- formatC(x_below$target_power, digits = digits, format = "f")
  tmp6 <- c("Target Power:" = target_power_str)

  tmp <- data.frame("Setting" = c(
      tmp1,
      tmp2,
      tmp4,
      tmp5,
      tmp6
    ))

  print(tmp, right = FALSE)

  if (solution_found_below) {

    x_final_below_str <- formatC(x_below$x_final,
                           digits = switch(predictor,
                                           n = 0,
                                           es = digits),
                           format = "f")
    ci_below_str <- paste0(
        "[",
        formatC(x_below$ci_final[1], digits = digits, format = "f"),
        ", ",
        formatC(x_below$ci_final[2], digits = digits, format = "f"),
        "]")

  }

  if (solution_found_above) {

    x_final_above_str <- formatC(x_above$x_final,
                           digits = switch(predictor,
                                           n = 0,
                                           es = digits),
                           format = "f")
    ci_above_str <- paste0(
        "[",
        formatC(x_above$ci_final[1], digits = digits, format = "f"),
        ", ",
        formatC(x_above$ci_final[2], digits = digits, format = "f"),
        "]")

  }

  solution_found <- solution_found_below &&
                    solution_found_above

  if (solution_found_below ||
      solution_found_above) {

    cat("\nSolution: \n")

    cat("\nApproximate region of sample sizes with power:\n")
    if (solution_found) {
      tmp <- paste0("- not significantly different from ",
                    target_power_str,
                    ": ",
                    x_final_below_str,
                    " to ",
                    x_final_above_str)
      cat(strwrap(tmp, exdent = 2), sep = "\n")
    }
    if (solution_found_below) {
      tmp <- paste0("- significantly lower than ",
                    target_power_str,
                    ": ",
                    x_final_below_str)
      cat(strwrap(tmp, exdent = 2), sep = "\n")
    }
    if (solution_found_above) {
      tmp <- paste0("- significantly higher than ",
                    target_power_str,
                    ": ",
                    x_final_above_str)
      cat(strwrap(tmp, exdent = 2), sep = "\n")
    }

    cat("\nConfidence intervals of the estimated power:\n")
    if (solution_found_below) {
      tmp <- paste0("- for the lower bound (",
                    x_final_below_str,
                    "): ",
                    ci_below_str)
      cat(strwrap(tmp, exdent = 2), sep = "\n")
    }
    if (solution_found_above) {
      tmp <- paste0("- for the upper bound (",
                    x_final_above_str,
                    "): ",
                    ci_above_str)
      cat(strwrap(tmp, exdent = 2), sep = "\n")
    }

    if (!solution_found_below) {
      cat("Solution not found for the lower region.")
    }
    if (!solution_found_above) {
      cat("Solution not found for the upper region.")
    }
  } else {
    cat("\n- Solution not found.\n")
  }
  cat("\nCall `summary()` for detailed results.\n")
  invisible(x)
}

#' @rdname x_from_power
#'
#' @details
#'
#' The function [arg_x_from_power()]
#' is a helper to set argument values
#' if `object` is an output
#' of [x_from_power()] or similar
#' functions.
#'
#' @return
#' The function [arg_x_from_power()]
#' returns the requested argument if
#' available. If not available, it
#' returns `NULL`.
#'
#' @param arg The name of element to
#' retrieve.
#'
#' @param arg_in The name of the element
#' from which an element is to be
#' retrieved.
#'
#' @export
arg_x_from_power <- function(
                      object,
                      arg,
                      arg_in = NULL) {
  if (inherits(object, "x_from_power")) {
    if (is.null(arg_in)) {
      return(object[[arg]])
    } else {
      return(object[[arg_in]][[arg]])
    }
  } else {
    return(NULL)
  }
}

#' @rdname x_from_power
#'
#' @details
#'
#' The function [pba_diagnosis()]
#' generates simple diagnostic plots
#' for the search history of probabilistic
#' bisection algorithm. This function is for advanced users
#' to examine the search history of
#' the probabilistic bisection algorithm.
#' It is for diagnostic purpose and has
#' limited support for customizing the plots.
#'
#'
#'
#' @return
#' The function [pba_diagnosis()]
#' returns `NULL` invisibly. Called for
#' its side-effect.
#'
#' @param a_out The output of
#' [x_from_power()] and friends. The
#' algorithm used must be `"probabilistic_bisection"`.
#'
#' @param p_interval The range of
#' the plot that "zooms" around the solution
#' (or the median of in the final posterior
#' probability distribution), expressed
#' in terms of the area of the distribution.
#'
#' @param posterior_xlim The range of
#' the first plot of search history and the
#' plot of posterior probability distribution,
#' expressed
#' in terms of the area of the distribution.
#'
#' @param which If `a_out` is a list with
#' more than one output of [x_from_power()],
#' such as the output of [n_region_from_power()],
#' `which` must be set to the name of
#' one such output (`"below"` or `"above"`
#' for the output of [n_region_from_power()],
#' or the output of `q_power_mediation_*`).
#'
#' @export
pba_diagnosis <- function(
  a_out,
  p_interval = c(.05, .95),
  posterior_xlim = c(.01, .99),
  which = NULL
) {
  if (!inherits(a_out, "x_from_power")) {
    i <- sapply(a_out,
                inherits,
                what = c("x_from_power", "n_region_from_power"))
    if (any(i)) {
      a_out <- a_out[[which(i)]]
    } else {
      warning("No supported object found")
      return(invisible(NULL))
    }
    if (inherits(a_out, "n_region_from_power")) {
      tmp <- names(a_out)
      if (is.null(which) ||
          !isFALSE(is.na(match(which, tmp))) ||
          length(which) != 1) {
        stop("'which' should be set to one of these: ",
             paste0(sQuote(tmp), collapse = ","))
      } else {
        a_out <- a_out[[which]]
      }
    }
  }
  if (a_out$algorithm != "probabilistic_bisection") {
    warning("The algorithm used is not probabilistic_bisection.")
    invisible(NULL)
  }
  solution <- a_out$x_final
  technical <- a_out$technical
  i0 <- length(technical$x_history)
  dfun_out <- technical$dfun_history[[i0]]
  x_lo <- q_dfun(dfun_out, p_interval[1])
  x_hi <- q_dfun(dfun_out, p_interval[2])
  parold <- par(no.readonly = TRUE)
  on.exit(par(parold))
  graphics::layout(matrix(1:6,
         nrow = 3,
         byrow = TRUE))

  # ==== Plot of power_curve ====

  # plot(a_out$power_curve)
  tmp <- c(q_dfun(dfun_out, posterior_xlim[1]),
           q_dfun(dfun_out, posterior_xlim[2]))
  plot(a_out,
       xlim = tmp)
  abline(h = technical$f_power,
         col = "red",
         lwd = 1)
  abline(h = a_out$target_power,
         col = "blue",
         lwd = 1)
  if (!is.na(solution)) {
    abline(v = solution,
           col = "red",
           lwd = 1)
  }
  abline(v = c(x_lo, x_hi),
         col = "black",
         lwd = 1,
         lty = "dotted")

  # ==== x_history ====

  plot(technical$x_history,
       type = "l",
       main = "History of x",
       ylab = a_out$x,
       xlab = "Iteration")
  points(technical$x_history)
  if (!is.na(solution)) {
    abline(h = solution,
           col = "red",
           lwd = 1)
  }
  abline(h = c(x_lo, x_hi),
         col = "black",
         lwd = 1,
         lty = "dotted")
  tmp <- which(technical$final_check_history)
  abline(v = tmp,
         col = "red",
         lty = "dotted")

  # ==== Posterior Probability Distribution ====

  tmp <- c(q_dfun(dfun_out, posterior_xlim[1]),
           q_dfun(dfun_out, posterior_xlim[2]))
  plot(dfun_out,
       type = "l",
       xlim = tmp,
       main = "Posterior Probability Distribution",
       xlab = a_out$x,
       ylab = NULL)
  if (!is.na(solution)) {
    abline(v = solution,
           col = "red",
           lwd = 1)
  }
  abline(v = c(x_lo, x_hi),
         col = "black",
         lwd = 1,
         lty = "dotted")
  abline(h = 0,
         col = "black",
         lwd = 0.5,
         lty = "dotted")

  # ==== HDR history ====

  hdr_power_history <- technical$hdr_power_history
  tmp <- sapply(unlist(hdr_power_history,
                       recursive = FALSE),
                range)
  plot(seq_along(hdr_power_history),
       y = rep(a_out$f_power,
               length(hdr_power_history)),
       ylim = range(tmp),
       type = "l",
       main = "Highest Density Regions",
       ylab = a_out$x,
       xlab = "Iteration")
  points(a_out$reject_by_power_curve_history)
  abline(h = a_out$target_power,
         col = "blue",
         lwd = 1,
         lty = "dotted")
  for (i in seq_along(hdr_power_history)) {
    for (y in hdr_power_history[[i]]) {
      arrows(
          x0 = i,
          y0 = y[1],
          x1 = i,
          y1 = y[2],
          lty = "solid",
          code = 3,
          angle = 90,
          length = ifelse(y[1] == y[2], 0, .05)
        )
    }
  }
  tmp <- which(technical$final_check_history)
  abline(v = tmp,
         col = "red",
         lty = "dotted")

  # ==== Power curve around the solution ====

  tmp <- c(q_dfun(dfun_out, posterior_xlim[1]),
           q_dfun(dfun_out, posterior_xlim[2]))

  if (!is.na(solution)) {
    x_tmp <- solution
  } else {
    x_tmp <- q_dfun(dfun_out)
  }
  tmp2 <- max(abs(c(x_lo, x_hi) - x_tmp) / abs(tmp - x_tmp))
  tmp3 <- x_tmp + (tmp - x_tmp) * tmp2
  plot(a_out$power_curve,
       xlim = tmp3)
  abline(h = technical$f_power,
         col = "red",
         lwd = 1)
  abline(h = a_out$target_power,
         col = "blue",
         lwd = 1)
  if (!is.na(solution)) {
    abline(v = solution,
           col = "red",
           lwd = 1)
  }
  abline(v = c(x_lo, x_hi),
         col = "black",
         lwd = 1,
         lty = "dotted")
  tmp <- which(technical$final_check_history)
  abline(v = tmp,
         col = "red")

  # ==== f_history ====

  plot(technical$f_history,
       type = "l",
       main = "History of Function Values",
       xlab = "Iteration",
       ylab = "Function Value")
  points(technical$f_history)
  abline(h = 0,
         col = "blue",
         lwd = 1,
         lty = "dotted")
  tmp <- which(technical$final_check_history)
  abline(v = tmp,
         col = "red",
         lty = "dotted")

  invisible(NULL)
}

skip("Internal")

library(testthat)

test_that("es_region_from_power: check_extend: short", {

# A case the failed in previous versions due to invalid
# ranges.

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: m
y ~ m: m
y ~ x: m
"

options(power4mome.bz = TRUE)

out <- power4test(nrep = 50,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(pars = "y~m"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = FALSE)

out1 <- x_from_power(
            out,
            x = "es",
            pop_es_name = "y ~ m",
            what = "ub",
            target_power = .80,
            seed = 1234,
            progress = TRUE,
            simulation_progress = TRUE)
# plot(out1)

out1b <- x_from_power(
            out1,
            seed = 1234,
            progress = TRUE,
            simulation_progress = TRUE)
# plot(out1b)

out2 <- x_from_power(
            out1,
            what = "lb",
            seed = 4567)
# plot(out2)

})

skip("Internal")

library(testthat)

test_that("check_extend: Start with very small n", {

mod <-
"
m ~ x
y ~ m + x
"

pop_es <-
"
m ~ x: m
y ~ m: m
y ~ x: m
"

options(power4mome.bz = TRUE)
out <- power4test(
          nrep = 400,
          model = mod,
          pop_es = pop_es,
          n = 51,
          reliability = .70,
          number_of_indicators = 6,
          R = 199,
          ci_type = "mc",
          test_fun = test_indirect_effect,
          test_args = list(x = "x",
                            m = "m",
                            y = "y",
                            mc_ci = TRUE
                          ),
          iseed = 1234,
          parallel = TRUE,
          ncores = max(1, parallel::detectCores(logical = FALSE) - 1)
        )

out1 <- n_from_power(out,
                     what = "ub",
                     seed = 1234)
plot(out1)

out2 <- n_region_from_power(
          out,
          seed = 1234
        )
plot(out2)

})

test_that("check_extend: Start with very large n", {

mod <-
"
m ~ x
y ~ m + x
"

pop_es <-
"
m ~ x: m
y ~ m: m
y ~ x: m
"

options(power4mome.bz = TRUE)
outb <- power4test(
          nrep = 400,
          model = mod,
          pop_es = pop_es,
          n = 1999,
          reliability = .70,
          number_of_indicators = 6,
          R = 199,
          ci_type = "mc",
          test_fun = test_indirect_effect,
          test_args = list(x = "x",
                            m = "m",
                            y = "y",
                            mc_ci = TRUE
                          ),
          iseed = 1234,
          parallel = TRUE,
          ncores = max(1, parallel::detectCores(logical = FALSE) - 1)
        )

outb1 <- n_from_power(outb,
                     what = "ub",
                     seed = 1234)
plot(outb1)

outb2 <- n_region_from_power(
          outb,
          seed = 1234
        )
plot(outb2)

})

skip("Internal")

library(testthat)

test_that("PBA: diagnostic function", {

# Case 1

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

out <- power4test(nrep = 51,
                  model = mod,
                  pop_es = mod_es,
                  n = 600,
                  R = 99,
                  test_fun = test_parameters,
                  test_args = list(par = "y~x"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = !testthat::is_testing())

system.time(
a_out <- x_from_power(out,
                     x = "n",
                     progress = TRUE,
                     final_nrep = 100,
                     save_sim_all = FALSE,
                     x_interval = c(300, 1000),
                     algorithm = "probabilistic_bisection",
                     what = "point",
                     goal = "close_enough")
)

pba_diagnosis(a_out)

})

set_x_range_by_x <- function(object,
                        x,
                        pop_es_name,
                        target_power = .80,
                        k = 4,
                        x_max = switch(x, n = 1000, es = .7),
                        x_min = switch(x, n = 10, es = .0),
                        object_by_org = NULL,
                        what = NULL,
                        goal = NULL,
                        tol = NULL,
                        ci_level = .95) {
  if (x == "n") {
    out <- set_n_range_by_x(
                        object = object,
                        target_power = target_power,
                        k = k,
                        n_max = x_max,
                        object_by_org = object_by_org,
                        what = what,
                        goal = goal,
                        tol = tol,
                        ci_level = ci_level
                      )
    return(out)
  }
  if (x == "es") {
    out <- set_es_range_by_x(
                        object = object,
                        pop_es_name = pop_es_name,
                        target_power = target_power,
                        k = k,
                        es_max = x_max,
                        es_min = x_min,
                        object_by_org = object_by_org,
                        what = what,
                        goal = goal,
                        tol = tol,
                        ci_level = ci_level
                      )
    return(out)
  }
}

set_x_range <- function(object,
                        x,
                        pop_es_name,
                        target_power = .80,
                        k = 4,
                        x_max = switch(x, n = 1000, es = .7),
                        x_min = switch(x, n = 10, es = .0)) {
  if (x == "n") {
    out <- set_n_range(object = object,
                       target_power = target_power,
                       k = k,
                       n_max = x_max)
    return(out)
  }
  if (x == "es") {
    out <- set_es_range(object = object,
                        pop_es_name = pop_es_name,
                        target_power = target_power,
                        k = k,
                        es_max = x_max,
                        es_min = x_min)
    return(out)
  }
}

estimate_x <- function(power_n_fit,
                       target_power = .80,
                       interval = c(50, 2000),
                       extendInt = "no") {
  f <- function(n) {
    # stats::predict(power_n_fit,
    #                newdata = list(n = n)) - target_power
    stats::predict(power_n_fit,
                   newdata = list(x = n)) - target_power
  }
  n_target <- tryCatch(stats::uniroot(f,
                             interval = interval,
                             extendInt = extendInt),
                       error = function(e) e)
  if (inherits(n_target, "error")) {
    # Return NA if error occurred. E.g.,
    # - Root not in the interval.
    return(NA)
  }
  n_target <- round(n_target$root)
  # Negative sample size should be handled by the
  # calling function, e.g,, estimate_n_range().
  return(n_target)
}

#' @noRd
# Output:
# - k values of x with levels of power predicted by power_x_fit
#   The width determined by tolerance.
# If k == 1 and tolerance == 0,
# - yield the x with the predicted target_power.
# WARNING: If k == 1 but tolerance != 0, unexpected results may occur.
estimate_x_range <- function(power_x_fit,
                             x,
                             target_power = .80,
                             k = 5,
                             tolerance = .20,
                             power_min = .01,
                             power_max = .99,
                             interval = switch(x,
                                               n = c(50, 2000),
                                               es = c(0, .7)),
                             extendInt = NULL,
                             x_to_exclude = NULL) {
  x <- match.arg(x,
                 c("n", "es"))
  out <- switch(x,
                n = estimate_n_range(power_n_fit = power_x_fit,
                                     target_power = target_power,
                                     k = k,
                                     tolerance = tolerance,
                                     power_min = power_min,
                                     power_max = power_max,
                                     interval = interval,
                                     extendInt = extendInt,
                                     n_to_exclude = x_to_exclude),
                es = estimate_es_range(power_es_fit = power_x_fit,
                                       target_power = target_power,
                                       k = k,
                                       tolerance = tolerance,
                                       power_min = power_min,
                                       power_max = power_max,
                                       interval = interval,
                                       extendInt = extendInt,
                                       es_to_exclude = x_to_exclude))
  out
}

check_x <- function(ns,
                    interval,
                    n_to_exclude,
                    extendInt) {
  i <- rep(FALSE, length(ns))
  if (isFALSE(extendInt %in% c("yes", "upX"))) {
    i[ns > interval[2]] <- TRUE
  }
  if (isFALSE(extendInt %in% c("yes", "downX"))) {
    i[ns < interval[1]] <- TRUE
  }

  # Ns used are considered invalid
  i[ns %in% n_to_exclude] <- TRUE

  # Duplicated are considered invalid
  i[duplicated(ns)] <- TRUE

  i[is.na(ns)] <- TRUE

  i
}

nrep_from_power <- function(power_j,
                            target_power,
                            tolerance,
                            nrep_min,
                            nrep_max) {
  a <- abs(power_j - target_power)
  a <- pmin(a, tolerance)
  b <- 1 - a / tolerance
  d <- nrep_max - nrep_min
  out <- round(nrep_min + b * d)
  return(out)
}

#' @noRd
# Helper for printing
catwrap <- function(x,
                    width = 0.9 * getOption("width"),
                    indent = 0,
                    exdent = 0,
                    prefix = "",
                    simplify = TRUE,
                    initial = prefix,
                    sep = "\n",
                    fill = FALSE,
                    labels = NULL,
                    append = FALSE) {
  out <- strwrap(x,
                 width = width,
                 indent = indent,
                 exdent = exdent,
                 prefix = prefix,
                 simplify = simplify,
                 initial = initial)
  cat(out,
      sep = sep,
      fill = fill,
      labels = labels,
      append = append)
}


rejection_rates_add_ci <- function(object,
                                   level = .95,
                                   add_reject = TRUE,
                                   add_se = TRUE) {
  # No need for additional arguments because they are handled by rejection_rates()
  # level is handled by this function, not by rejection_rates()
  if (!is.data.frame(object)) {
    if (inherits(object, "power4test_by_es")) {
      df1 <- rejection_rates(object,
                             all_columns = TRUE)
    }
    if (inherits(object, "power4test_by_n")) {
      df1 <- rejection_rates(object,
                             all_columns = TRUE)
    }
  } else {
    # Assume it is already an output with rejection rates
    df1 <- object
  }
  # It works on any data frame with these two columns:
  # - `sig` or `reject`
  # - `nvalid`
  # It adds these columns:
  # - `reject` (if add_reject is TRUE)
  # - `reject_se`
  # - `reject_ci_lo`
  # - `reject_ci_hi`

  tmp <- match(c("sig", "reject"), colnames(df1))
  if (all(is.na(tmp))) {
    stop("CI cannot be computed because 'reject' or 'sig' column not found.")
  }
  tmp <- tmp[!is.na(tmp)][1]
  reject <- df1[, tmp, drop = TRUE]
  if (add_reject && isFALSE("reject" %in% colnames(df1))) {
    df1$reject <- df1$sig
  }
  df1$reject_se <- sqrt(reject * (1 - reject) / df1$nvalid)
  ci_i <- reject_ci(
            nreject = round(reject * df1$nvalid),
            nvalid = df1$nvalid,
            level = level,
            method = getOption("power4mome.ci_method", default = "wilson"))
  df1$reject_ci_lo <- as.vector(ci_i[, 1])
  df1$reject_ci_hi <- as.vector(ci_i[, 2])
  if (!add_se) {
    df1$reject_se <- NULL
  }
  df1
}

#' @noRd
# No longer needed because power_curve object
# has a predict method
predict_fit <- function(object,
                        newdata) {
  if (inherits(object, "glm")) {
    out <- stats::predict(object = object,
                          newdata = newdata,
                          type = "response")
  } else {
    out <- stats::predict(object = object,
                          newdata = newdata)
  }
  return(out)
}

#' @noRd
# Check whether any value of x tried already
# meet the ci_hit criterion
find_ci_hit <- function(object,
                        ci_level = .95,
                        target_power = .80,
                        final_nrep = 400,
                        closest_ok = FALSE,
                        if_ties = c("min", "max")) {
  if_ties <- match.arg(if_ties)
  # If no hit, return NULL
  # If hit, always return one number
  # If closest_ok, accept a trial with closest power level
  # No need to handle other arguments. They should be
  #   retrieved from the object by rejection_rates()
  by_x_ci <- rejection_rates_add_ci(object,
                                    level = ci_level)
  i0 <- (by_x_ci$reject_ci_lo < target_power) &
        (by_x_ci$reject_ci_hi > target_power)

  if (isFALSE(any(i0))) {
    if (closest_ok) {
      tmp <- which.min(abs(by_x_ci$reject - target_power))
      i0 <- rep(FALSE, nrow(by_x_ci))
      i0[tmp] <- TRUE
    } else {
      return(NULL)
    }
  }
  if (sum(i0) > 1) {
    # Find the value with CI hitting the target power
    # and has the smallest SE.
    i1 <- rank(by_x_ci$reject_se)
    i1[!i0] <- NA
    # Do not consider those with nrep < final_nrep
    i1[by_x_ci$nrep < final_nrep] <- NA
    # If ties, the smallest value will be used
    if (all(is.na(i1))) {
      return(NULL)
    }
    i2 <- switch(if_ties,
                 min = which(i1 == min(i1[i0], na.rm = TRUE))[1],
                 max = which(i1 == max(i1[i0], na.rm = TRUE))[1])
  } else {
    # Still check nrep
    # To ignore nrep, set nrep to 0.
    if (by_x_ci$nrep[i0] < final_nrep) {
      return(NULL)
    }
    i2 <- which(i0)
  }
  return(i2)
}

#' @noRd
# Find the solution
# based on goal and what
find_solution <- function(object,
                          target_power = .80,
                          ci_level = .95,
                          what = c("point", "ub", "lb"),
                          tol = 1e-2,
                          goal = c("ci_hit", "close_enough"),
                          final_nrep = 400,
                          closest_ok = FALSE,
                          if_ties = c("min", "max"),
                          weight_by = c("nrep", "ci_width", "se", "none"),
                          debug = FALSE) {

  what <- match.arg(what)
  goal <- match.arg(goal)
  weight_by <- match.arg(weight_by)

  out <- switch(goal,
                ci_hit = find_ci_hit(
                           object = object,
                           ci_level = ci_level,
                           target_power = target_power,
                           final_nrep = final_nrep,
                           closest_ok = closest_ok,
                           if_ties = if_ties),
                close_enough = find_close_enough(
                           object = object,
                           ci_level = ci_level,
                           target_power = target_power,
                           final_nrep = final_nrep,
                           tol = tol,
                           what = what,
                           closest_ok = closest_ok,
                           if_ties = if_ties,
                           weight_by = weight_by,
                           debug = debug))
  out
}

#' @noRd
# Find the solution
# based on goal and what
find_close_enough <- function(
  object,
  target_power = .80,
  ci_level = .95,
  what = c("point", "ub", "lb"),
  tol = 1e-2,
  final_nrep = 400,
  closest_ok = FALSE,
  if_ties = c("min", "max"),
  weight_by = c("nrep", "ci_width", "se", "none"),
  debug = FALSE) {

  # If no solution, return NULL
  # If solution, always return one number
  # If closest_ok, accept the closest trial
  what <- match.arg(what)
  if_ties <- match.arg(if_ties)
  weight_by <- match.arg(weight_by)

  # No need for other arguments because they are stored in object
  by_x_ci <- rejection_rates_add_ci(object,
                                    level = ci_level,
                                    add_se = TRUE)
  var_all <- switch(
                weight_by,
                se = by_x_ci$reject_se ^ 2,
                ci_width = (by_x_ci$reject_ci_hi - by_x_ci$reject_ci_lo),
                none = rep(1, nrow(by_x_ci)),
                nrep = max(by_x_ci$nvalid) / by_x_ci$nvalid
              )
  r_all <- switch(what,
                  point = by_x_ci$reject,
                  ub = by_x_ci$reject_ci_hi,
                  lb = by_x_ci$reject_ci_lo)
  r_all0 <- r_all - target_power
  r_all1 <- abs(r_all0)
  i0 <- r_all1 < tol

  if (isFALSE(any(i0))) {
    if (closest_ok) {
      tmp <- which.min(r_all1 * var_all)
      i0 <- rep(FALSE, nrow(by_x_ci))
      i0[tmp] <- TRUE
    } else {
      return(NULL)
    }
  }
  if (sum(i0) > 1) {
    # For close_enough,
    # distance should take precedence,
    # and so ties will not be considered.
    i1 <- rank(r_all1 * var_all)
    i1[!i0] <- NA
    # Do not consider those with nrep < final_nrep
    i1[by_x_ci$nrep < final_nrep] <- NA
    # If ties, the smallest value will be used
    if (all(is.na(i1))) {
      return(NULL)
    }
    i2 <- switch(if_ties,
                 min = which(i1 == min(i1[i0], na.rm = TRUE))[1],
                 max = which(i1 == max(i1[i0], na.rm = TRUE))[1])

  } else {
    # Still check nrep
    # To ignore nrep, set nrep to 0.
    if (by_x_ci$nrep[i0] < final_nrep) {
      return(NULL)
    }
    i2 <- which(i0)
  }
  return(i2)
}

#' @noRd
# Check whether the x_from_power object
# has x and (optionally) pop_es_name identical
# to those requested.
check_x_from_power_as_input <- function(object,
                                        x,
                                        pop_es_name,
                                        final_nrep,
                                        ci_level,
                                        rejection_rates_args) {
  if (!identical(x, object$x)) {
    stop("object's x is ", object$x, " but ",
         "requested x is ", x)
  }
  if (x == "es") {
    if (!identical(pop_es_name, object$pop_es_name)) {
      stop("object's pop_es_name is ", object$pop_es_name, " but ",
          "requested pop_es_name is ", pop_es_name)
    }
  }
  if (object$arg$final_nrep != final_nrep) {
    stop("object's final_nrep (",
         object$arg$final_nrep,
         ") is different from the requested final_nrep (",
         final_nrep,
         ").")
  }
  if (object$ci_level != ci_level) {
    stop("object's ci_level (",
         object$ci_level,
         ") is different from the requested ci_level (",
         ci_level,
         ").")
  }
  # Do not check merge_all_tests because
  # it should be determined internally.
  for (x in names(object$rejection_rates_args)) {
    if (x == "merge_all_tests") next
    if (object$rejection_rates_args[[x]] != rejection_rates_args[[x]]) {
      stop("object's rejection_rates_args:",
           x,
           " (",
           object$rejection_rates_args[[x]],
           ")",
           " is different from the new value (",
           rejection_rates_args[[x]],
           ").")
    }
  }

  return(TRUE)
}

#' @noRd
# Get the vector of x values already tried
get_x_tried <- function(object,
                        x) {
  # No need for other arguments because only `n` or `es` is used
  tmp <- rejection_rates_add_ci(object)
  out <- switch(x,
                n = tmp$n,
                es = tmp$es)
  out
}

#' @noRd
# Check whether a value of x has already been tried
in_x_tried <- function(test_x,
                       object,
                       x,
                       multiple_matches = FALSE) {
  # If yes, return the index
  # Otherwise, return NA
  x_tried <- get_x_tried(object = object,
                         x = x)
  if (multiple_matches) {
    out <- which(x_tried %in% test_x)
    if (length(out) == 0) {
      out <- NA
    }
  } else {
    out <- match(test_x, x_tried)
  }
  out
}

#' @noRd
force_new_x <- function(
                    x0,
                    x_tried,
                    x_interval,
                    x_type = c("es", "n"),
                    step = switch(x_type,
                                  n = 1,
                                  es = .01),
                    post_process = switch(x_type,
                                          es = function(x) x,
                                          n = ceiling),
                    k = 20
                  ) {
  k0 <- 0
  xi <- x0
  step_all <- as.vector(rbind(seq_len(k), -seq_len(k)))
  while ((k0 <= k * 2) &&
         ((xi %in% x_tried) ||
          (xi < min(x_interval)) ||
          (xi > max(x_interval)))) {
    k0 <- k0 + 1
    xi <- x0 + step_all[k0] * step
    xi <- post_process(xi)
  }
  xi
}

#' @noRd
# Determine the probable range of valid values
# for a parameter
fix_es_interval <- function(object,
                            x,
                            pop_es_name,
                            x_interval,
                            progress = TRUE,
                            step = .10,
                            es_min = -.90,
                            es_max = .90) {
  if ((x == "es") &&
      is.null(x_interval)) {
    if (progress) {
      cat("\n--- Interval for x (es) ---\n\n")
      cat("Determining the valid interval of values for '",
          pop_es_name,
          "' ...\n",
          sep = "")
    }
    es_tmp <- pop_es(object,
                     pop_es_name = pop_es_name)
    range_tmp <- tryCatch(check_valid_es_values(object,
                                                pop_es_name = pop_es_name,
                                                step = step,
                                                es_min = es_min,
                                                es_max = es_max),
                    error = function(e) e)
    if ((inherits(range_tmp, "error")) ||
        (all(is.na(range_tmp)))) {
      if (es_tmp < 0) {
        x_interval <- c(-.95, 0)
      } else {
        x_interval <- c(0, .95)
      }
      if (progress) {
        cat("Failed to find the valid range.\n")
        cat("This range will be used:",
            paste0(formatC(x_interval[1], digits = 3, format = "f"),
                  " to ",
                  formatC(x_interval[2], digits = 3, format = "f")),
            "\n")
        cat("Set 'x_interval' manually if necessary.\n")
      }
    } else {
      x_interval <- range_tmp
      if ((es_tmp < min(x_interval)) ||
          (es_tmp > max(x_interval))) {
        # TODO:
        # - Do we need this check? This should rarely happen.
      }
      # If es0 is in the interval, then
      if (es_tmp <= 0) {
        x_interval[x_interval >= 0] <- 0
      } else {
        x_interval[x_interval <= 0] <- 0
      }
      if (progress) {
        cat("The probable valid range, of the same sign of object's value, is:",
            paste0(formatC(x_interval[1], digits = 3, format = "f"),
                  " to ",
                  formatC(x_interval[2], digits = 3, format = "f")),
            "\n")
      }
    }
  }
  return(x_interval)
}

#' @noRd
# Find the x-intersection,
# as in Regula Falsi
x_from_y <- function(x1,
                     y1,
                     x2,
                     y2,
                     target = 0) {
  (target - y1) * (x2 - x1) / (y2 - y1) + x1
}

#' @noRd
# Can be used for most algorithms
# Input:
# - f_i: The power to be tested
# - target: The target power
# - nrep: The number of replications
# - final_nrep: The required nrep. Always not a solution if nrep != final_nrep
# - ci_level: The level of confidence of the CI
# - which:
#    - point: Check against the target power
#    - ub: Check the upper bound of the CI against the target power
#    - lb: Check the lower bound of the CI against the target power
#    Ignored if goal == "ci_hit"
# - tol:
#    - The tolerance used in "close_enough"
# - goal:
#    - ci_hit: f_i is a solution if its CI include target_power
#    - close_enough: f_i is as solution if
#        abs(what - target) < tol
check_solution <- function(f_i,
                           target_power,
                           nrep,
                           final_nrep,
                           ci_level = .95,
                           what = c("point", "ub", "lb"),
                           tol = 1e-2,
                           goal = c("ci_hit", "close_enough")) {
  if (nrep != final_nrep) {
    return(FALSE)
  }
  goal <- match.arg(goal)
  se_i <- sqrt(f_i * (1 - f_i) / nrep)
  ci_i <- reject_ci(
            nreject = round(f_i * nrep),
            nvalid = nrep,
            level = ci_level,
            method = getOption("power4mome.ci_method", default = "wilson"))
  cilb <- as.vector(ci_i[, 1])
  ciub <- as.vector(ci_i[, 2])
  if (goal == "ci_hit") {
    # Ignore what
    if ((cilb < target_power) && (ciub > target_power)) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  }
  chk_point <- switch(what,
                      point = f_i,
                      ub = ciub,
                      lb = cilb)
  out <- abs(chk_point - target_power) < tol
  out
}

#' @noRd
# Find the root by the Muller's method
# For reference. Not used directly.
muller <- function(
  f,
  ...,
  start,
  maxiter = 100,
  tol = 1e-5
) {
  x_all <- rep(NA, maxiter)
  y_all <- rep(NA, maxiter)
  if (length(start) != 3) {
    stop("'start' must be a vector of three numbers")
  }
  i <- 1
  start <- sort(start)
  xm2 <- start[1]
  xm1 <- start[2]
  x0i <- start[3]
  while (i <= maxiter) {
    x_all[i] <- x0i
    y_all[i] <- f(x0i, ...)
    xp1 <- root_muller_i(
      f = f,
      xm2 = xm2,
      xm1 = xm1,
      x0i = x0i,
      ...
    )
    fp1 <- f(xp1, ...)
    cat("----\n")
    cat("i =", i, "\n")
    cat("xp1 =", xp1, "\n")
    cat("fp1 =", fp1, "\n")
    if (abs(fp1) < tol) {
      break
    }
    xm2 <- xm1
    xm1 <- x0i
    x0i <- xp1
    i <- i + 1
  }
  list(x = x_all[seq_len(i)],
       y = y_all[seq_len(i)])
}

#' @noRd
# Find the next value based on the Muller's method
root_muller_i <- function(
  f,
  xm2,
  xm1,
  x0i,
  ym2 = NULL,
  ym1 = NULL,
  y0i = NULL,
  ...
) {
  if (is.null(ym2)) {
    ym2 <- f(xm2, ...)
  }
  if (is.null(ym1)) {
    ym1 <- f(xm1, ...)
  }
  if (is.null(y0i)) {
    y0i <- f(x0i, ...)
  }
  A <- ((xm1 - x0i) * (ym2 - y0i) - (xm2 - x0i) * (ym1 - y0i)) /
       ((x0i - xm1) * (xm1 - xm2) * (xm2 - x0i))
  B <- ((xm2 - x0i)^2 * (ym1 - y0i) - (xm1 - x0i)^2 * (ym2 - y0i)) /
       ((x0i - xm1) * (xm1 - xm2) * (xm2 - x0i))
  C <- y0i

  xp1a <- suppressWarnings(
            x0i + (-2 * C) /
            (B - sqrt(B^2 - 4 * A * C))
          )
  xp1b <- suppressWarnings(
            x0i + (-2 * C) /
            (B + sqrt(B^2 - 4 * A * C))
          )
  if (is.nan(xp1a)) {
    xp1a <- NA
  }
  if (is.nan(xp1b)) {
    xp1b <- NA
  }

  xp1 <- max(xp1a, xp1b, na.rm = TRUE)
  xp1
}

#' @noRd
check_solution_in_by_x <- function(
  object,
  target_power,
  final_nrep,
  ci_level = .95,
  what = c("point", "ub", "lb"),
  tol = 1e-2,
  goal = c("ci_hit", "close_enough")
) {
  what <- match.arg(what)
  goal <- match.arg(goal)
  if (inherits(object, "rejection_rates_df")) {
    reject_df <- object
  } else {
    # No need for other arguments. Only reject and nrep are used.
    reject_df <- rejection_rates(object,
                                ci_level = ci_level,
                                all_columns = TRUE,
                                se = TRUE)
  }
  reject_all <- reject_df$reject
  nrep_all <- reject_df$nrep
  chK_all <- mapply(
               check_solution,
               f_i = reject_all,
               nrep = nrep_all,
               MoreArgs = list(target_power = target_power,
                               final_nrep = final_nrep,
                               ci_level = ci_level,
                               what = what,
                               tol = tol,
                               goal = goal)
             )
  chK_all
}


#' @noRd

check_changes <- function(
    x_history,
    delta_tol = .01,
    last_k = 3) {
  x <- x_history[!is.na(x_history)]
  p <- length(x)
  if (p < last_k) return(TRUE)
  x_test <- rev(x_history[!is.na(x_history)])
  x_range <- range(x_test[1:last_k],
                   na.rm = TRUE)
  if (abs(x_range[2] - x_range[1]) < delta_tol) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

#' @noRd

check_rate <- function(
    x_history,
    delta_slope_tol = -.1,
    last_k = 3) {
  x <- x_history[!is.na(x_history)]
  p <- length(x)
  if (p < last_k) return(TRUE)
  x <- rev(rev(x)[1:last_k])
  trend <- stats::lm.fit(
          y = matrix(abs(x), ncol = 1),
          x = cbind(1, matrix(1:last_k, ncol = 1)))$coefficients["x2"]
  if (trend > delta_slope_tol) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

#' @noRd
# - Return the adjusted power
#   based on what and goal
target_power_adjusted <- function(
  target_power = .80,
  goal = c("ci_hit", "close_enough"),
  what = c("point", "ub", "lb"),
  tolerance = .02,
  nrep = 100,
  level = .95
) {
  if (goal == "close_enough") {
    if (what == "point") {
      return(target_power)
    }
    if (what %in% c("ub", "lb")) {
      # Find the power with ub close enough to target
      b <- switch(what,
                  ub = "cihi",
                  lb = "cilo")
      f <- function(x, adj) {
        reject_ci_wilson(nreject = x,
                         nvalid = nrep,
                         level = level)[, b] -
        target_power + adj
      }
      tmp1 <- stats::uniroot(
              f,
              adj = tolerance,
              interval = c(0, nrep)
            )
      out1 <- tmp1$root / nrep
      tmp2 <- stats::uniroot(
              f,
              adj = -tolerance,
              interval = c(0, nrep)
            )
      out2 <- tmp2$root / nrep
      return(mean(c(out1, out2)))
    }
  }
  if (goal == "ci_hit") {
    return(target_power)
  }
}

#' @noRd
# Input:
# - Rejection rates table
# - Target power (can be the adjusted power)
# Output:
# - The x-value by `x_from_y()`
x_from_y_rejection_rates <- function(
  reject_df,
  target_power,
  x,
  weight_by = c("nrep", "se")
) {

  weight_by <- match.arg(weight_by)

  y_diff <- reject_df$reject - target_power
  y_dist <- abs(reject_df$reject - target_power)

  reject_df$wt <- switch(
                    weight_by,
                    nrep = reject_df$nrep,
                    se = reject_df$reject_se
                  )

  # Closest and above
  y_above <- which(y_diff > 0)
  x_above_i <- y_above[which.min(y_dist[y_above] *
                reject_df$wt[y_above]^2)]
  # Closest and below
  y_below <- which(y_diff < 0)
  x_below_i <- y_below[which.min(y_dist[y_below] *
                reject_df$wt[y_below]^2)]

  x_out_above_i <- switch(x,
                          n = reject_df$n[x_above_i],
                          es = reject_df$es[x_above_i])
  x_out_below_i <- switch(x,
                          n = reject_df$n[x_below_i],
                          es = reject_df$es[x_below_i])
  x_reject_above_i <- reject_df$reject[x_above_i]
  x_reject_below_i <- reject_df$reject[x_below_i]

  x_between_i <- x_from_y(x1 = x_out_below_i,
                          y1 = x_reject_below_i,
                          x2 = x_out_above_i,
                          y2 = x_reject_above_i,
                          target = target_power)
  x_between_i
}

#' @noRd
# Modify rejection_rates_args in an by_x objects
set_rejection_rates_args_by_x <- function(
  object,
  rejection_rates_args
) {
  for (i in seq_along(object)) {
    tmp <-  attr(object[[i]], "args")
    tmp2 <- tmp$rejection_rates_args
    tmp2 <- utils::modifyList(
                  tmp2,
                  rejection_rates_args,
                  keep.null = TRUE
                )
    # merge_all_tests is always TRUE
    # Ignored by rejection_rates() if there is only one test
    tmp2$merge_all_tests <- TRUE
    tmp$rejection_rates_args <- tmp2
    attr(object[[i]], "args") <- tmp
  }
  object
}

#' @noRd
# Determine the tolerance based on the algorithms
set_tolerance <- function(
  algorithm,
  ...
) {
  fun <- match.fun(paste0("set_tolerance_", algorithm))
  do.call(fun,
          list(...))
}

#' @noRd
set_tolerance_bisection <- function(...) {
  .02
}

#' @noRd
set_tolerance_power_curve <- function(...) {
  .02
}

#' @noRd
set_tolerance_probabilistic_bisection <- function(...) {
  args <- list(...)

  proxy_power <- tryCatch(target_power_adjusted(
                    target_power = args$target_power,
                    goal = args$goal,
                    what = args$what,
                    tolerance = 0,
                    nrep = args$final_nrep,
                    level = args$ci_level
                  ),
                  error = function(e) e)
  if (inherits(proxy_power, "error")) {
    stop("The goal is not possible given the final_nrep and target power. Please try other values.")
  }

  tmp0 <- ifelse(
              inherits(proxy_power, "error"),
              yes = args$target_power,
              no = proxy_power
            )
  tmp1 <- reject_ci_wilson(
            nreject = ceiling(tmp0 * args$final_nrep),
            nvalid = args$final_nrep,
            level = args$ci_level
          )
  out <- min(abs(tmp1[1, ] - tmp0)) * .90
  out
}

#' @noRd
# Determine the tolerance based on the algorithms
set_max_trials <- function(
  algorithm,
  ...
) {
  fun <- match.fun(paste0("set_max_trials_", algorithm))
  do.call(fun,
          list(...))
}

#' @noRd
set_max_trials_bisection <- function(...) {
  10
}

#' @noRd
set_max_trials_power_curve <- function(...) {
  10
}

#' @noRd
set_max_trials_probabilistic_bisection <- function(...) {
  100
}

#' @noRd

do_FUN <- function(X,
                   FUN,
                   ...,
                   iseed = NULL,
                   parallel = TRUE,
                   progress = TRUE,
                   ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                   cl = NULL) {
  stop_cluster <- TRUE
  if (inherits(cl, "cluster")) {

    # ==== A cluster is supplied. Force parallel ====

    # If cl is supplied, never stop it.

    parallel <- TRUE
    ncores <- length(cl)
    stop_cluster <- FALSE
  }
  if (!isTRUE(ncores > 0)) {
    parallel <- FALSE
  }
  if (parallel) {

    # ==== Make a cluster ====

    if (is.null(cl)) {
      cl <- parallel::makeCluster(ncores)
    }

    # Export all functions defined in the global environment

    global_names <- ls(envir = .GlobalEnv)
    global_fs <- sapply(
                  global_names,
                  \(x) is.function(get(x))
                )
    if (any(global_fs)) {
      to_export <- global_names[global_fs]
      parallel::clusterExport(cl, to_export)
    }

    # Reproduce the search paths
    pkgs <- .packages()
    pkgs <- rev(pkgs)
    parallel::clusterExport(cl, "pkgs", envir = environment())
    parallel::clusterEvalQ(cl, {
                    sapply(pkgs,
                    function(x) {
                      library(x,
                             character.only = TRUE)
                    })
                  })

    if (!is.null(iseed)) {
      parallel::clusterSetRNGStream(cl = cl,
                                    iseed = iseed)
    }
    on.exit(if (stop_cluster) {try(parallel::stopCluster(cl), silent = TRUE)})
  } else {

    # ==== Serial ====

    cl <- NULL
    if (!is.null(iseed)) set.seed(iseed)
  }
  if (progress) {
    # Can be used for both serial and parallel
    out <- pbapply::pblapply(X = X,
                             FUN = FUN,
                             ...,
                             cl = cl)
  } else {
    if (parallel) {

      use_lb <- getOption("power4mome.use_lb", FALSE)

      if (use_lb) {
        # ==== load balancing ====
        out <- parallel::parLapplyLB(cl = cl,
                                     X = X,
                                     fun = FUN,
                                     ...)
      } else {
        # ==== No load balancing ====
        out <- parallel::parLapply(cl = cl,
                                   X = X,
                                   fun = FUN,
                                   ...)
      }
      # out <- parallel::parLapplyLB(cl = cl,
      #                              X = X,
      #                              fun = FUN,
      #                              ...)
    } else {
      out <- lapply(X = X,
                    FUN = FUN,
                    ...)
    }
  }
  return(out)
}

#' @title Test Several Indirect Effects
#'
#' @description Test several indirect effects
#' for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing an
#' indirect effect, by setting it
#' to the `test_fun` argument.
#'
#' It uses [manymome::many_indirect_effects()]
#' to do the test. It can be used on
#' models fitted by [lavaan::sem()]
#' or fitted by a sequence of calls
#' to [stats::lm()], although only
#' nonparametric bootstrap confidence
#' interval is supported for models
#' fitted by regression using
#' [stats::lm()].
#'
#' It also supports testing the total
#' indirect effect. Just set `omnibus`
#' to `"total"`.
#'
#' @return
#' In its normal usage, it returns
#' a data frame with the
#' following columns:
#'
#' - `est`: The estimated
#'  indirect effect for each path.
#'
#' - `cilo` and `cihi`: The
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively,
#'  for each indirect effect
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' - `test_label`: A column of labels
#'  generated to label the indirect
#'  effects.
#'
#' If `omnibus` is `"all_sig"` or
#' `"at_least_one"sig"`, then
#' the data frame has only one row,
#' and the columns `"est"`, `"cilo"`,
#' and `"cihi"` are `NA`. The column
#' `sig` is determined by whether
#' all paths are significant (`"all_sig"`)
#' or whether at least one path is
#' significant (`"at_least_one_sig"`).
#'
#'
#' @inheritParams test_indirect_effect
#'
#' @param m Must be a list of character
#' vectors. Each character vector stores the
#' name(s) of mediator(s) along a path.
#' The path
#' moves from the first mediator in the
#' vector to the last mediator in the
#' vector. If `NULL`, the stored paths
#' will be used, which are all the
#' indirect paths in the model between
#' `x` and `y`, by default.
#'
#'
#' @param ... Additional arguments to
#' be passed to [manymome::many_indirect_effects()].
#'
#' @param omnibus If `"no"`, the default,
#' then the test results for all paths
#' are stored. If `"all_sig"`, then
#' only one row of test is stored, and
#' the test is declared significant if
#' *all* paths are significant. If
#' `"at_least_one_sig"`, then only
#' one row of test is stored, and the
#' test is declared significant if
#' at least one of the paths is
#' significant. If `"at_least_k_sig"`,
#' then only one row of test is stored,
#' and the test is declared significant
#' if at least `k` of the paths is
#' significant, `k` determined by the
#' argument `at_least_k`. If `"total"`,
#' then the total indirect effect will
#' be tested.
#'
#' @param at_least_k The minimum number
#' of paths required to be significant
#' for the omnibus test to be considered
#' significant. Used when
#' `omnibus` is `"at_least_k_sig"`.
#'
#' @param p_adjust_method The method to be
#' passed to [p.adjust()] to adjust the
#' *p*-values when testing the effects.
#' Default is `"none"` and the *p*-values
#' will not be adjusted. The unadjusted
#' *p*-values will be stored in the column
#' `pvalue_org`.
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_med <-
#' "
#' m1 ~ x
#' m2 ~ x
#' y ~ m1 + m2 + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_med_es <-
#' "
#' y ~ m1: s
#' m1 ~ x: m
#' y ~ m2: s
#' m2 ~ x: l
#' y ~ x: n
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 100,
#'                        R = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test in each replication
#'
#' test_ind <- power4test(object = sim_only,
#'                        test_fun = test_k_indirect_effects,
#'                        test_args = list(x = "x",
#'                                         y = "y",
#'                                         mc_ci = TRUE))
#' print(test_ind,
#'       test_long = TRUE)
#'
#' # Set omnibus = "all_sig" to declare
#' # significant only if all paths are
#' # significant
#'
#' test_ind_all_sig <- power4test(
#'                        object = sim_only,
#'                        test_fun = test_k_indirect_effects,
#'                        test_args = list(x = "x",
#'                                         y = "y",
#'                                         mc_ci = TRUE,
#'                                         omnibus = "all_sig"))
#' print(test_ind_all_sig,
#'       test_long = TRUE)
#'
#' @export

test_k_indirect_effects <- function(
                            fit = fit,
                            x = NULL,
                            m = NULL,
                            y = NULL,
                            mc_ci = TRUE,
                            mc_out = NULL,
                            boot_ci = FALSE,
                            boot_out = NULL,
                            check_post_check = TRUE,
                            test_method = NULL,
                            ...,
                            omnibus = c("no", "all_sig", "at_least_one_sig", "at_least_k_sig", "total"),
                            at_least_k = 1,
                            p_adjust_method = "none",
                            fit_name = "fit",
                            get_map_names = FALSE,
                            get_test_name = FALSE
                          ) {

  # ==== Enable pvalue? ====

  args <- list(...)
  if (!is.null(mc_out)) {
    R <- length(mc_out)
  } else if (!is.null(boot_out)) {
    R <- length(boot_out)
  } else {
    R <- NULL
  }
  R <- args$R %||% formals(manymome::indirect_effect)$R
  ci_level <- args$level %||% formals(manymome::indirect_effect)$level
  R_bz_ok <- isTRUE(R %in% R_extrapolate(alpha = 1 - ci_level))
  bz_not_FALSE <- !isFALSE(options("power4mome.bz"))
  test_method_NULL <- is.null(test_method)
  if (R_bz_ok &&
      bz_not_FALSE &&
      test_method_NULL) {
    test_method <- "pvalue"
  } else {
    test_method <- match.arg(test_method,
                             c("ci", "pvalue"))
  }

  internal_options <- list()
  if (test_method == "pvalue") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = TRUE,
                                               pvalue_min_size = -Inf))
  }
  if (test_method == "ci") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = FALSE))
  }
  omnibus <- match.arg(omnibus)
  if (fit_name != "fit") {
    mc_name <- paste0(fit_name, "_mc_out")
    boot_name <- paste0(fit_name, "_boot_out")
  } else {
    mc_name <- "mc_out"
    boot_name <- "boot_out"
  }
  map_names <- c(fit = fit_name,
                 mc_out = mc_name,
                 boot_out = boot_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp <- paste0(x, "-...->", y)
    args <- as.list(match.call())
    tmp2 <- character(0)
    if (isTRUE(args$standardized_x) && !isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' standardized)")
    }
    if (!isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('y' standardized)")
    }
    if (isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' and 'y' standardized)")
    }
    return(paste0("test_indirects: ", tmp, collapse = ""))
  }
  if (boot_ci) mc_ci <- FALSE
  if (inherits(fit, "lavaan")) {
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
  } else {
    fit_ok <- TRUE
  }
  if (fit_ok) {
    all_paths <- try(fit@external$fit_external$all_paths, silent = TRUE)
    if (inherits(all_paths, "try-error") ||
        isTRUE(is.null(all_paths))) {
      all_paths <- try(attr(fit, "fit_external")$all_paths, silent = TRUE)
      if (inherits(all_paths, "try-error") ||
          isTRUE(is.null(all_paths))) {
        stop("Paths not found Please test each path separately.")
      }
    }
    if (is.null(m)) {
      i_xy <- sapply(
                all_paths,
                function(xx) {
                  (xx$x == x) && (xx$y == y)
                },
                simplify = TRUE
              )
      if (!any(i_xy)) {
        stop("No indirect path between ", x, " and ", y, ".")
      }
      all_paths0 <- all_paths[i_xy]
    } else {
      i_xy <- sapply(
                all_paths,
                function(xx) {
                  i0 <- (xx$x == x) && (xx$y == y)
                  i1 <- FALSE
                  for (mm in m) {
                    mm <- unname(mm)
                    mm_i <- unname(xx$m)
                    if (identical(mm, mm_i)) {
                      i1 <- TRUE
                      break
                    }
                  }
                  i0 && i1
                },
                simplify = TRUE
              )
      if (!any(i_xy)) {
        stop("No requested paths found in the model.")
      }
      all_paths0 <- all_paths[i_xy]
    }
    class(all_paths0) <- class(all_paths)
    out <- tryCatch(manymome::many_indirect_effects(
                                   paths = all_paths0,
                                   fit = fit,
                                   mc_ci = mc_ci,
                                   mc_out = mc_out,
                                   boot_ci = boot_ci,
                                   boot_out = boot_out,
                                   progress = FALSE,
                                   ...,
                                   internal_options = internal_options),
                  error = function(e) e)
  } else {
    out <- NA
  }
  if (inherits(out, "error") ||
      identical(out, NA)) {
    out2 <- data.frame(
              est = as.numeric(NA),
              cilo = as.numeric(NA),
              cihi = as.numeric(NA),
              sig = as.numeric(NA),
              pvalue = as.numeric(NA)
            )
    return(out2)
  }
  if (omnibus == "total") {
    out_ind <- out
    out_total <- manymome::total_indirect_effect(out_ind)
    tmp1 <- paste0(c(x, y), collapse = "-...->")
    tmp1 <- paste(tmp1, "(Total)")
    tmp <- stats::confint(out_total,
                          level = ci_level)
    out1 <- data.frame(
              ind = unname(stats::coef(out_total)),
              CI.lo = tmp[1, 1],
              CI.hi = tmp[1, 2],
              pvalue = out_total$boot_p %||% (out_total$mc_p %||% NA)
            )
    rownames(out1) <- tmp1
    out <- list(out)
    names(out) <- tmp1
    out1 <- cbind(test_label = rownames(out1),
                  out1)
    rownames(out1) <- NULL
  } else {
    out1 <- manymome::indirect_effects_from_list(
                                out,
                                add_sig = FALSE,
                                pvalue = TRUE,
                                se = FALSE
                              )
    if (p_adjust_method != "none") {
      out1$pvalue_org <- out1$pvalue
      out1$pvalue <- stats::p.adjust(
                          out1$pvalue_org,
                          method = p_adjust_method
                        )
    }
    out1 <- cbind(test_label = rownames(out1),
                  out1)
    rownames(out1) <- NULL
  }
  out1_names <- colnames(out1)
  out1_names <- gsub("CI.lo", "cilo", out1_names)
  out1_names <- gsub("CI.hi", "cihi", out1_names)
  out1_names <- gsub("ind", "est", out1_names)
  out1_names <- gsub("std", "est", out1_names)
  colnames(out1) <- out1_names
  if (test_method == "ci") {
    sig <- ifelse((out1$cilo > 0) |
                  (out1$cihi < 0),
                  yes = 1,
                  no = 0)
    out1$sig <- sig
  }
  R_case <- ""
  bz_alpha_ok <- isTRUE(all.equal(1 - out[[1]]$level,
                                  getOption("power4mome.bz.alpha",
                                                      default = .05)))
  if (test_method == "pvalue") {
    out1$sig <- ifelse(
                out1$pvalue < (1 -  out[[1]]$level),
                yes = 1,
                no = 0
              )
    R <- sapply(out,
                \(x) max(length(x$boot_indirect),
                         length(x$mc_indirect)))
    R_case <- bz_case(R[1])
    nlt0 <- sapply(out,
                \(x) max(sum(as.numeric(x$boot_indirect < 0)),
                         sum(as.numeric(x$mc_indirect < 0))))
    out1$R <- unname(R)
    out1$nlt0 <- unname(nlt0)
    out1$alpha <- unname(1 -  out[[1]]$level)
    if (bz_alpha_ok) {
      tmp <- lapply(out,
              function(x) {
                boot_est <- x$boot_indirect %||% x$mc_indirect
                boot_sig <- bz_sig_partition(
                              boot_est,
                              alpha = 1 - x$level
                            )
                boot_sig
              })
      tmp <- do.call(rbind,
                    unname(tmp))
      out1 <- cbind(out1,
                    tmp)
    }
  }
  if (omnibus %in% c("no", "total")) {
    attr(out1, "test_label") <- "test_label"
    return(out1)
  } else {
    out2 <- out1[1, ]
    tmp <- paste0(c(x, y), collapse = "-...->")
    tmp <- paste0(tmp, switch(omnibus,
                              all_sig = " (All sig)",
                              at_least_one_sig = " (1+ sig)",
                              at_least_k_sig = paste0(" (",
                                                      at_least_k,
                                                      "+ sig)"),
                              total = " (Total)"))
    out2[1, "test_label"] <- tmp
    out2[, c("est", "cilo", "cihi")] <- as.numeric(NA)
    tmp <- switch(omnibus,
                  all_sig = as.numeric(isTRUE(all(out1$sig == 1))),
                  at_least_one_sig = as.numeric(isTRUE(any(out1$sig == 1))),
                  at_least_k_sig = as.numeric(isTRUE(sum(out1$sig == 1) >= at_least_k)))
    out2$sig <- tmp
    if ((R_case == "one") &&
        (omnibus == "all_sig") &&
        bz_alpha_ok) {
      out1_bz <- add_bz_i(out1)
      out1_bz <- out1_bz[, !(colnames(out1_bz) %in% colnames(out2))]
      out1_bz2 <- apply(
                      out1_bz,
                      MARGIN = 2,
                      min
                    )
      out1_bz2 <- rbind(out1_bz2)
      rownames(out1_bz2) <- NULL
      out2 <- cbind(out2, out1_bz2)
    }
    if ((R_case == "cum") &&
        (omnibus == "all_sig") &&
        bz_alpha_ok) {
      out1_bz <- out1
      out1_bz <- out1_bz[, grepl("bz_", colnames(out1_bz))]
      out1_bz2 <- apply(
                      out1_bz,
                      MARGIN = 2,
                      \(x) as.numeric(all(x == 1))
                    )
      out1_bz2 <- rbind(out1_bz2)
      rownames(out1_bz2) <- NULL
      out2[, colnames(out1_bz2)] <- out1_bz2
    }
    if (any(is.na(out2$sig))) {
      out2$sig <- as.numeric(NA)
    }
    attr(out2, "test_label") <- "test_label"
    return(out2)
  }
}

#' @title Parse YAML-Stye Values For 'pop_es'
#'
#' @description Convert a YAML string
#' to a vector or list of `pop_es`
#' specification.
#'
#' @details
#' The function [pop_es_yaml()]
#' allows users to specify population
#' values of a model using one single
#' string, as in 'lavaan' model syntax.
#'
#' @inheritSection ptable_pop Specify 'pop_es' Using a Multiline String
#'
#'
#' @return
#' Either a named vector (for a
#' single-group model) or a list of
#' named vector (for a multigroup model)
#' of the population values of the
#' parameters (the effect sizes).
#'
#' @param text The multiline string to be parsed
#' to a specification of population
#' values.
#'
#' @seealso [ptable_pop()],
#' [power4test()], and other functions
#' that have the `pop_es` argument.
#'
#' @examples
#'
#' mod_es <- c("y ~ m" = "l",
#'             "m ~ x" = "m",
#'             "y ~ x" = "nil")
#'
#' mod_es_yaml <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: nil
#' "
#'
#' pop_es_yaml(mod_es_yaml)
#'
#'
#' @export
pop_es_yaml <- function(text) {
  # TODO:
  # - May add other features in the future.

  # Used to handle 'n' in YAML, which
  # is converted to `FALSE`.
  # Should not be needed because `nil`
  # should be used.
  my_hr <- list(seq = function(x) {
                  out <- as.character(x)
                  out <- gsub("FALSE",
                              "n",
                              out,
                              fixed = TRUE)
                  out
                },
                "bool.fix" = function(x) {
                  # Does not work for now.
                  # But it should not be needed anyway
                  ifelse(x,
                        yes = x,
                        no = "n")
                })

  out <- yaml::read_yaml(text = text,
                         handlers = my_hr)
  length_max <- sapply(out,
                       length)
  if (all(length_max == 1)) {
    out <- unlist(out)
  }
  out
}

#' @noRd
pop_es_yaml_check <- function(x) {
  chk1 <- isTRUE(length(x) == 1)
  chk2 <- isTRUE(is.character(x))
  chk3 <- isTRUE(is.null(names(x)))
  if (chk1 && chk2 && chk3) {
    out <- pop_es_yaml(x)
  } else {
    out <- x
  }
  return(out)
}

library(testthat)
suppressMessages(library(lavaan))

test_that("ptable_pop: indirect", {

# Simple mediation model

mod <-
"
m ~ x
y ~ m + x
"

mod_es1 <- c(".beta." = "m",
             ".ind.(x->m->y)" = "li")

mod_es2 <-
"
.beta.: m
.ind.(x->m->y): li
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

# Parallel mediation model
# Correlated errors

mod <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
m1 ~~ m2
"

mod_es1 <- c(".beta." = "s",
             ".ind.(y ~ m1 ~ x)" = "si",
             ".ind.(x -> m2 -> y)" = "mi",
             "m1 ~~ m2" = "l")

mod_es2 <-
"
.beta.: s
.ind.(y ~ m1 ~ x): si
.ind.(x -> m2 -> y): mi
m1 ~~ m2: l
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

# Parallel mediation model
# Correlated errors not specified in the model

# Moderated mediation model

mod <-
"
m ~ x + w + x:w
y ~ m + z + m:z + x + u + x:u
"

mod_es1 <- c(".beta." = "s",
             "m ~ x:w" = "l",
             ".ind.(x->m->y)" = "li")

mod_es2 <-
"
.beta.: s
m ~ x:w: l
.ind.(x->m->y): li
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

# Other models

mod <-
"
m1 ~ x + c1
m2 ~ m1 + x2 + c1
y ~  m2 + m1 + x + w + x:w + c1
"

mod_es1 <- c(".ind.(x->m1->m2)" = "li",
             "y ~ m2" = "l",
             "y ~ x" = "m",
             "y ~ w" = "s",
             "y ~ x:w" = "s",
             "x ~~ w" = "s")

mod_es2 <-
"
.ind.(x->m1->m2): li
y ~ m2: l
y ~ x: m
y ~ w: s
y ~ x:w: s
y ~~ w: s
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

# Moderated mediation multigroup model.
# Some are latent factors.

test_that("Multigroup models", {

mod <-
"
m1 ~ x
m2 ~ m1 + x
y ~  m2 + m1 + x + w + m2:w
"
mod_es1 <- list("m1 ~ x" = "m",
                "m2 ~ m1" = "s",
                "y ~ m2:w" = c("-m", "n"))

mod_es2 <-
"
m1 ~ x: m
m2 ~ m1: s
y ~ m2:w: [-m, nil]
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models: .ind.", {

mod <-
"
m1 ~ x + c1
m2 ~ m1 + x2 + c1
y ~  m2 + m1 + x + w + x:w + c1
"
mod_es1 <- list(".ind.(y~m2~m1~x)" = "si",
                "y ~ x" = c("m", "l", "n"),
                "y ~ w" = "s",
                "y ~ x:w" = "s",
                "x ~~ w" = c("s", "m", "l"))

mod_es2 <-
"
.ind.(x->m1->m2->y): si
y ~ x: [m, l, nil]
y ~ w: s
y ~ x:w: s
x ~~ w: [s, m, l]
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models", {

mod <-
"
m ~ x
y ~  m + x
"

mod_es1 <- list("m ~ x" = "-m",
                "y ~ m" = "l",
                "y ~ x" = c("m", "l", "n"))

mod_es2 <-
"
m ~ x: -m
y ~ m: l
y ~ x: [m, l, nil]
"

pop_es_yaml(mod_es2)

chk1 <- ptable_pop(mod, mod_es1)
chk2 <- ptable_pop(mod, mod_es2)

expect_equal(chk1$start,
             chk2$start,
             tolerance = 1e-1)

})

#' @noRd
# Input:
# - Parameter table
# - Original model syntax
# Output:
# -
pt_with_int <- function(ptable,
                        model_original) {
  to_add <- cov_to_add(ptable)
  # TODO:
  # - When `x:w ~~ y:z` is available in
  #   lavaan, return the model syntax directly.
  model_fixed <- paste(
                  c(model_original,
                    to_add,
                    "\n"),
                  collapse = "\n")
  model_fixed_2 <- gsub(
                     ":",
                     "__xx__",
                     model_fixed,
                     fixed = TRUE)
  fit_fixed <- fake_fit_for_int(
                  ptable = ptable,
                  model_int = model_fixed_2)
  pt_fixed <- lavaan::parameterTable(fit_fixed$fit)
  pt_fixed$lhs <- gsub("__xx__",
                       ":",
                       pt_fixed$lhs,
                       fixed = TRUE)
  pt_fixed$rhs <- gsub("__xx__",
                       ":",
                       pt_fixed$rhs,
                       fixed = TRUE)
  pt_fixed <- merge_start(pt_source = ptable,
                          pt_target = pt_fixed)
  attr(pt_fixed, "model_fixed") <- model_fixed
  attr(pt_fixed, "model") <- model_original
  pt_fixed
}

#' @noRd
# Input:
# - A parameter table or a lavaan object
# Output:
# - A vector of covariances to be added
cov_to_add <- function(object) {
  # Also works for a multigroup model
  if (inherits(object, "lavaan")) {
    pt <- lavaan::parameterTable(object)
  } else {
    pt <- object
  }
  ov_x <- lavaan::lavNames(pt, "ov.x")
  ov_int <- lavaan::lavNames(pt, "ov.interaction")
  ov_nox <- lavaan::lavNames(pt, "ov.nox")
  ov_nox <- setdiff(ov_nox, ov_int)
  cov_add <- list()
  for (xx in ov_int) {
    ww <- strsplit(xx, ":", fixed = TRUE)[[1]]
    tmp <- utils::combn(ww, m = 2, simplify = TRUE)
    cov_add <- c(cov_add,
                mapply(function(lhs, rhs) {
                  data.frame(lhs = lhs,
                             op = "~~",
                             rhs = rhs)
                },
                lhs = c(ww, tmp[1, ]),
                rhs = c(rep(xx, length(ww)),
                        tmp[2, ]),
                SIMPLIFY = FALSE,
                USE.NAMES = FALSE))
  }
  cov_add <- do.call(rbind,
                     cov_add)
  i_w <- cov_add$lhs %in% ov_x
  if (any(i_w)) {
    cov_add2 <- list()
    for (xx in unique(cov_add$rhs[i_w])) {
      cov_add2 <- c(cov_add2,
                  mapply(function(lhs, rhs) {
                    data.frame(lhs = lhs,
                               op = "~~",
                               rhs = rhs)
                  },
                  lhs = ov_x,
                  rhs = xx,
                  SIMPLIFY = FALSE,
                  USE.NAMES = FALSE))
    }
    cov_add <- rbind(cov_add,
                     do.call(rbind, cov_add2))
  }
  cov_add <- unique(cov_add)
  pt_cov <- pt[(pt$op == "~~") &
               (pt$free > 0) &
               (pt$lhs != pt$rhs) &
               ((pt$lhs %in% ov_x) & (pt$rhs %in% ov_x)), ]
  i1 <- grepl(":", pt_cov$lhs)
  tmp1 <- pt_cov[i1, "lhs"]
  pt_cov[i1, "lhs"] <- pt_cov[i1, "rhs"]
  pt_cov[i1, "rhs"] <- tmp1
  pt_cov1 <- pt_cov[, c("lhs", "op", "rhs")]
  pt_cov_all <- rbind(cov_add,
                      pt_cov1)
  pt_cov_all <- unique(pt_cov_all)
  i <- pt_cov_all$lhs != pt_cov_all$rhs
  pt_cov_all <- pt_cov_all[i, ]
  # Remove duplicated covariances
  for (i in seq_len(nrow(pt_cov_all))) {
    tmp_i <- unlist(pt_cov_all[i, c("lhs", "rhs")])
    tmp2 <- order(tmp_i)
    pt_cov_all[i, c("lhs", "rhs")] <- unname(tmp_i[tmp2])
  }
  pt_cov_all <- pt_cov_all[!duplicated(pt_cov_all), , drop = FALSE]
  pt_cov_all <- apply(pt_cov_all,
                      MARGIN = 1,
                      paste,
                      collapse = " ")
  pt_cov_all <- unname(pt_cov_all)
  pt_cov_all
}

#' @noRd
# Input:
# - A parameter table.
# - A model syntax.
# Output:
# - A list:
#   - A fitted object
#   - A list of fake covariance matrices
#   - A list of fake mean vectors
fake_fit_for_int <- function(
                        ptable,
                        model_int
                      ) {
  # Multigroup models supported
  ngroups <- max(ptable$group)
  gpnames <- paste0("gp", seq_len(ngroups))
  fit0 <- lavaan::sem(model_int,
                      do.fit = FALSE)
  vnames <- lavaan::lavNames(fit0,
                             type = "ov")
  p <- length(vnames)
  d1 <- diag(p)
  colnames(d1) <- rownames(d1) <- vnames
  m1 <- vector(mode = "numeric",
               length = p)
  names(m1) <- vnames
  dat_cov <- lapply(seq_len(ngroups),
                    function(x) d1)
  if (ngroups > 1) {
    dat_means <- lapply(seq_len(ngroups),
                      function(x) m1)
  } else {
    dat_means <- NULL
  }
  fit1 <- lavaan::sem(model_int,
                      sample.cov = dat_cov,
                      sample.mean = dat_means,
                      sample.nobs = rep(10000, ngroups),
                      do.fit = FALSE,
                      group.label = gpnames,
                      meanstructure = (ngroups > 1),
                      fixed.x = FALSE)
  list(fit = fit1,
       cov = dat_cov,
       mean = dat_means)
}

#' @noRd
# Input:
# - A parameter table.
# Output:
# - A parameter table, with covariance terms
#   ordered.
order_cov <- function(pt) {
  for (i in seq_len(nrow(pt))) {
    if (pt[i, "op", drop = TRUE] != "~~") next
    tmp_i <- unlist(pt[i, c("lhs", "rhs")])
    tmp2 <- order(tmp_i)
    pt[i, c("lhs", "rhs")] <- unname(tmp_i[tmp2])
  }
  pt
}

#' @noRd
# Input:
# - The source parameter table with starting values to be imported
# - The target parameter table.
# Output:
# - The target parameter table with starting values updated.
merge_start <- function(pt_source,
                        pt_target) {
  if (is.null(pt_source$group)) {
    ngroups <- 1
  } else {
    ngroups <- max(pt_source$group)
  }
  pt_source <- order_cov(pt_source)
  pt_target <- order_cov(pt_target)
  pt_source$tmplabel <- lavaan::lav_partable_labels(pt_source)
  pt_target$tmplabel <- lavaan::lav_partable_labels(pt_target)
  pt_target$tmpid <- seq_len(nrow(pt_target))
  col0 <- c("lhs", "op", "rhs", "tmplabel")
  tmp <- merge(pt_target[, c(col0, "tmpid")],
               pt_source[, c(col0, "start")],
               all.x = TRUE,
               sort = FALSE)
  tmp <- tmp[order(tmp$tmpid), ]
  i <- !is.na(tmp$start)
  pt_target[i, "start"] <- tmp[i, "start"]
  pt_target$tmplabel <- NULL
  pt_target$tmpid <- NULL
  pt_target
}

#' @noRd
# Check whether at least one mediator
# is involved in moderation.
# Input:
# - Model syntax or parameter table
# Output:
# - A character vector of mediator(s)
#   involved in moderation.
m_moderated <- function(object,
                        ngroups = 1) {
  if (is.character(object)) {
    pt0 <- lavaan::lavParseModelString(
                    object,
                    as.data.frame. = TRUE)
    attr(pt0, "constraints") <- NULL
    attr(pt0, "modifiers") <- NULL
    pt0 <- as.data.frame(lavaan::lav_partable_complete(pt0))
    pt <- pt0
  }
  int_term <- union(
                lavaan::lavNames(pt, "ov.interaction"),
                lavaan::lavNames(pt, "lv.interaction")
              )
  int_term_comp <- strsplit(
                      int_term,
                      split = ":"
                    )
  int_term_comp <- unique(unlist(int_term_comp))
  if (length(int_term) == 0) {
    return(character(0))
  }
  ovlv_y <- union(
              lavaan::lavNames(pt, "ov.nox"),
              lavaan::lavNames(pt, "lv.nox")
            )
  ovlv_y <- setdiff(ovlv_y, int_term)
  m_in_int <- intersect(ovlv_y, int_term_comp)
  return(m_in_int)
}

skip_on_cran()

library(testthat)
suppressMessages(library(lavaan))

test_that("add_cov", {

model <-
"
m ~ x + z + x:z
y ~ m + x + w + m:w
"

model_es <-
"
m ~ x:z: s
m ~ x: m
y ~ w: s
y ~ m:w: l
y ~ m: m
y ~ x: s
"

model_es_mg <-
"
m ~ x:z: [s, l]
m ~ x: m
y ~ w: s
y ~ m:w: l
y ~ m: m
y ~ x: s
"

tmp <- m_moderated(model)

expect_equal(tmp,
             "m")

out <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es,
            n = 100,
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

out_mg <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es_mg,
            n = 100,
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())



sim_out <- out$sim_all[[1]]
names(sim_out)
model_original <- sim_out$model_original
fit <- sim_out$fit0

pt <- ptable_pop(
        model = model,
        pop_es = model_es,
        add_cov_for_moderation = FALSE
      )

pt_mg <- ptable_pop(
        model = model,
        pop_es = model_es_mg,
        add_cov_for_moderation = FALSE
      )

pt_fixed <- pt_with_int(
              ptable = pt,
              model = model
            )
pt_fixed_mg <- pt_with_int(
              ptable = pt_mg,
              model = model
            )

pt_fixed2 <- ptable_pop(
        model = model,
        pop_es = model_es
      )

pt_fixed_mg2 <- ptable_pop(
        model = model,
        pop_es = model_es_mg
      )

sim_out <- out$sim_all[[1]]
sim_out_mg <- out_mg$sim_all[[1]]

tmpnames <- setdiff(lavNames(pt, "ov"), lavNames(pt, "ov.interaction"))
dat <- sim_out$mm_lm_dat_out[, tmpnames]

fit2 <- sem(pt_fixed,
            dat,
            fixed.x = FALSE)
fit3 <- sem(pt_fixed,
            dat + 2,
            fixed.x = FALSE)
fit4 <- sem(pt_fixed2,
            dat + 2,
            fixed.x = FALSE)
expect_equal(fitMeasures(fit2, "chisq"),
             fitMeasures(fit3, "chisq"))
expect_equal(fitMeasures(fit2, "chisq"),
             fitMeasures(fit4, "chisq"))

fit2b <- sem(model,
             dat,
             fixed.x = FALSE)
fit3b <- sem(model,
             dat + 2,
             fixed.x = FALSE)
fit4b <- sem(pt,
             dat + 2,
             fixed.x = FALSE)
expect_false(fitMeasures(fit2b, "chisq") == fitMeasures(fit3b, "chisq"))
expect_false(fitMeasures(fit2b, "chisq") == fitMeasures(fit4b, "chisq"))

tmpnames <- setdiff(lavNames(pt_mg, "ov"), lavNames(pt_mg, "ov.interaction"))
dat_mg <- sim_out_mg$mm_lm_dat_out[, c(tmpnames, "group")]
dat_mg_shifted <- dat_mg
dat_mg_shifted[, tmpnames] <- dat_mg_shifted[, tmpnames] + 2

fit2_mg <- sem(pt_fixed_mg,
               dat_mg,
               group = "group",
               fixed.x = FALSE)
fit3_mg <- sem(pt_fixed_mg,
               dat_mg_shifted,
               group = "group",
               fixed.x = FALSE)
fit4_mg <- sem(pt_fixed_mg2,
               dat_mg_shifted,
               group = "group",
               fixed.x = FALSE)
expect_equal(fitMeasures(fit2_mg, "chisq"),
             fitMeasures(fit3_mg, "chisq"))
expect_equal(fitMeasures(fit2_mg, "chisq"),
             fitMeasures(fit4_mg, "chisq"))

fit2b_mg <- sem(model,
               dat_mg,
               group = "group",
               fixed.x = FALSE)
fit3b_mg <- sem(pt_fixed_mg,
               dat_mg_shifted,
               group = "group",
               fixed.x = FALSE)
fit2c_mg <- sem(model,
               dat_mg,
               group = "group",
               fixed.x = FALSE,
               meanstructure = FALSE)
fit4b_mg <- sem(pt_mg,
               dat_mg_shifted,
               group = "group",
               fixed.x = FALSE,
               meanstructure = FALSE)
expect_false(fitMeasures(fit2b_mg, "chisq") == fitMeasures(fit3b_mg, "chisq"))
expect_false(fitMeasures(fit2c_mg, "chisq") == fitMeasures(fit4b_mg, "chisq"))

fix_cov_order <- function(pt) {
  for (i in seq_len(nrow(pt))) {
    if (pt$op[i] == "~~") {
      tmp <- sort(c(pt[i, "lhs"], pt[i, "rhs"]))
      pt[i, "lhs"] <- tmp[1]
      pt[i, "rhs"] <- tmp[2]
    }
  }
  pt
}

fit_extra <- sim_out$extra$fit
pt_extra <- parameterTable(fit_extra)

pt_extra <- fix_cov_order(pt_extra)
pt_fixed <- fix_cov_order(pt_fixed)

pt_extra$tmplabel <- lav_partable_labels(pt_extra)
pt_fixed$tmplabel <- lav_partable_labels(pt_fixed)
expect_equal(sort(pt_extra$tmplabel),
             sort(pt_fixed$tmplabel))

fit_extra_mg <- sim_out_mg$extra$fit
pt_extra_mg <- parameterTable(fit_extra_mg)

pt_extra_mg <- fix_cov_order(pt_extra_mg)
pt_fixed_mg <- fix_cov_order(pt_fixed_mg)

pt_extra_mg$tmplabel <- lav_partable_labels(pt_extra_mg)
pt_fixed_mg$tmplabel <- lav_partable_labels(pt_fixed_mg)
expect_equal(sort(pt_extra_mg$tmplabel),
             sort(pt_fixed_mg$tmplabel))

})

#' @noRd
header_str <- function(x,
                       hw = .8,
                       sym = "=",
                       min = 2,
                       prefix = character(0),
                       suffix = character(0),
                       sep_i = " ") {
  full_w <- options("width")$width
  hw1 <- floor(full_w * hw)
  x_len <- nchar(x)
  a <- floor((hw1 - x_len - 2) / 2)
  a <- max(min, a)
  out <- paste0(prefix,
                strrep(sym, a),
                sep_i, x, sep_i,
                strrep(sym, a),
                suffix)
  out
}

#' @title Test a Moderated Mediation Effect
#'
#' @description Test a moderated
#' mediation effect for a `power4test`
#' object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing a
#' moderated mediation effect, by
#' setting it to the `test_fun`
#' argument.
#'
#' It uses [manymome::index_of_mome()]
#' to do the test. It can be used on
#' models fitted by [lavaan::sem()]
#' or fitted by a sequence of calls
#' to [stats::lm()], although only
#' nonparametric bootstrap confidence
#' interval is supported for models
#' fitted by regression using
#' [stats::lm()].
#'
#' @return
#' In its normal usage, it returns
#' a named numeric vector with the
#' following elements:
#'
#' - `est`: The mean of the estimated
#'  indirect effect across datasets.
#'
#' - `cilo` and `cihi`: The means of the
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively.
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' @inheritParams test_indirect_effect
#'
#' @param fit The fit object, to be
#' passed to [manymome::index_of_mome()].
#'
#' @param w The name of the moderator.
#'
#' @param ... Additional arguments to
#' be passed to [manymome::index_of_mome()].
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x + w + x:w
#' y ~ m
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' m ~ x: n
#' y ~ x: m
#' m ~ w: l
#' m ~ x:w: l
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        R = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test in each replication
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_index_of_mome,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         w = "w",
#'                                         mc_ci = TRUE))
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#' @export
test_index_of_mome <- function(fit = fit,
                      x = NULL,
                      m = NULL,
                      y = NULL,
                      w = NULL,
                      mc_ci = TRUE,
                      mc_out = NULL,
                      boot_ci = FALSE,
                      boot_out = NULL,
                      check_post_check = TRUE,
                      test_method = NULL,
                      ...,
                      fit_name = "fit",
                      get_map_names = FALSE,
                      get_test_name = FALSE) {

  # ==== Enable pvalue? ====

  args <- list(...)
  if (!is.null(mc_out)) {
    R <- length(mc_out)
  } else if (!is.null(boot_out)) {
    R <- length(boot_out)
  } else {
    R <- NULL
  }
  R <- args$R %||% formals(manymome::index_of_mome)$R
  ci_level <- args$level %||% formals(manymome::index_of_mome)$level
  R_bz_ok <- isTRUE(R %in% R_extrapolate(alpha = 1 - ci_level))
  bz_not_FALSE <- !isFALSE(options("power4mome.bz"))
  test_method_NULL <- is.null(test_method)
  if (R_bz_ok &&
      bz_not_FALSE &&
      test_method_NULL) {
    test_method <- "pvalue"
  } else {
    test_method <- match.arg(test_method,
                             c("ci", "pvalue"))
  }

  internal_options <- list()
  if (test_method == "pvalue") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = TRUE,
                                               pvalue_min_size = -Inf))
  }
  if (test_method == "ci") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = FALSE))
  }
  if (fit_name != "fit") {
    mc_name <- paste0(fit_name, "_mc_out")
    boot_name <- paste0(fit_name, "_boot_out")
  } else {
    mc_name <- "mc_out"
    boot_name <- "boot_out"
  }
  map_names <- c(fit = fit_name,
                 mc_out = mc_name,
                 boot_out = boot_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp <- paste0(c(x, m, y),
                  collapse = "->")
    tmp <- paste0(tmp,
                  ", moderated by ",
                  w)
    return(paste0("test_index_of_mome: ", tmp, collapse = ""))
  }
  if (boot_ci) mc_ci <- FALSE
  if (inherits(fit, "lavaan")) {
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
  } else {
    fit_ok <- TRUE
  }
  if (fit_ok) {
    out <- tryCatch(manymome::index_of_mome(
                                 x = x,
                                 y = y,
                                 m = m,
                                 w = w,
                                 fit = fit,
                                 mc_ci = mc_ci,
                                 mc_out = mc_out,
                                 boot_ci = boot_ci,
                                 boot_out = boot_out,
                                 progress = FALSE,
                                 ...,
                                 internal_options = internal_options),
                   error = function(e) e)
  } else {
    out <- NA
  }
  if (inherits(out, "error") ||
      identical(out, NA)) {
    out2 <- c(est = as.numeric(NA),
              cilo = as.numeric(NA),
              cihi = as.numeric(NA),
              sig = as.numeric(NA),
              pvalue = as.numeric(NA))
    return(out2)
  }
  if (test_method == "ci") {
    ci0 <- stats::confint(out)
    out1 <- ifelse((ci0[1, 1] > 0) || (ci0[1, 2] < 0),
                    yes = 1,
                    no = 0)
  }
  if (test_method == "pvalue") {
    ci0 <- matrix(
              as.numeric(NA),
              nrow = 1,
              ncol = 2
            )
    out1 <- ifelse(
                out$pvalue < (1 - out$level),
                yes = 1,
                no = 0
              )
    bz_alpha_ok <- isTRUE(all.equal(1 - out$level,
                                    getOption("power4mome.bz.alpha",
                                                        default = .05)))
    if (out$ci_type %in% c("mc", "boot")) {
      diff_name <- switch(out$ci_type,
                          mc = "mc_diff",
                          boot = "boot_diff")
      est_diff <- out[[diff_name]]
      if (bz_alpha_ok) {
        est_sig <- bz_sig_partition(
                      est_diff,
                      alpha = 1 - out$level
                    )
      } else {
        est_sig <- as.numeric()
      }
      R <- length(est_diff)
      nlt0 <- sum(as.numeric(est_diff < 0))
    } else {
      R <- as.numeric(NA)
      nlt0 <- as.numeric(NA)
      est_sig <- as.numeric()
    }
  }
  out2 <- c(est = unname(stats::coef(out)),
            cilo = ci0[1, 1],
            cihi = ci0[1, 2],
            sig = out1,
            pvalue = out$pvalue %||% as.numeric(NA))
  if (test_method == "pvalue") {
    # For Boos & Zhang (2000)
    out2 <- c(out2, R = R, nlt0 = nlt0,
              alpha = 1 - out$level,
              est_sig)
  }
  return(out2)
}

#' @title Process Data by Generating Missing Values
#'
#' @description For the `process_data`
#' argument. It introduces missing
#' values in the generated data.
#'
#' @details
#' This function is to be used in
#' the `process_data` argument of
#' [power4test()].
#'
#' The missing values are generated by
#' [mice::ampute()]. Please refer to
#' its help page, the vignette for
#' this function:
#' [Generate missing values with ampute](https://rianneschouten.github.io/mice_ampute/vignette/ampute.html),
#' and Schouten, Lugtig, and Vink (2018).
#'
#' In addition to `data`, only two
#' arguments are explicitly defined for
#' [missing_values()]: `prop` and `mech`.
#' The argument `prop` is defined to remind
#' users of the default value for this
#' argument. The argument `mech`, specifying
#' the missing data mechanism, is
#' defined because its default value
#' is different from that of [mice::ampute()].
#' For [missing_values()], the default value
#' is `"MCAR"`, missing completely at
#' random.
#'
#' Please refer to the help page of
#' [mice::ampute()] for other available
#' arguments.
#'
#' ## Note for `patterns` and other arguments with variable names
#'
#' If arguments such as `patterns` are
#' specified, the variable names need to be
#' those in the generated data. If
#' indicator scores are generated, the
#' names need to be the names of the indicators,
#' not the names of the latent variables.
#'
#' @return
#' It returns a data frame with the
#' missing values inserted.
#'
#' @param data A data frame.
#'
#' @param ... Optional arguments to be
#' passed to [mice::ampute()].
#'
#' @param prop The proportion of missingness.
#' Default is 0.5, about 50% of the cases
#' have missing data.
#'
#' @param mech The missing data mechanism.
#' Default is `"MCAR"` (missing completely at random).
#' Other possible values are `"MAR"`
#' (missing at random) and `"MNAR"`
#' (missing not at random). Please refer
#' to the help of [mice::ampute()] for
#' details.
#'
#' @seealso [power4test()]. See also
#' [mice::ampute()] for the amputation
#' method.
#'
#' @references
#' Schouten, R. M., Lugtig, P., & Vink, G.
#' (2018). Generating missing values for
#' simulation purposes: A multivariate
#' amputation procedure.
#' *Journal of Statistical Computation and Simulation, 88*(15),
#' 2909--2930. \doi{10.1080/00949655.2018.1491577}
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#' # Simulate the data
#'
#' out <- power4test(
#'          nrep = 2,
#'          model = mod,
#'          pop_es = mod_es,
#'          n = 200,
#'          process_data = list(fun = "missing_values",
#'                              args = list(prop = .75)),
#'          test_fun = test_parameters,
#'          test_args = list(op = "~"),
#'          parallel = FALSE,
#'          iseed = 1234)
#'
#' dat <- pool_sim_data(out)
#' head(dat, 50)
#'
#' @export
missing_values <- function(
  data,
  ...,
  prop = 0.5,
  mech = "MCAR"
) {
  args0 <- list(...)
  cnames <- colnames(data)
  if (!is.null(args0$patterns)) {
    args0$patterns <- args0$patterns[, cnames]
  }
  if (!is.null(args0$weights)) {
    args0$weights <- args0$weights[, cnames]
  }
  args <- utils::modifyList(
            args0,
            list(data = data,
                 prop = prop,
                 mech = mech,
                 run = TRUE))
  out <- do.call(mice::ampute,
                 args)
  out$amp
}

#' @title Process Data by Computing Scale Scores
#'
#' @description For the `process_data`
#' argument. To compute scale scores
#' from indicators and replace the
#' indicators scores by computed
#' scale scores.
#'
#' @details
#' This function is to be used in
#' the `process_data` argument of
#' [power4test()].
#'
#' It retrieves the attribute
#' `"number_of_indicators"`,
#' stored by [power4test()], to identify
#' factors with indicators, computes
#' the scale scores based on `method`,
#' and replace the indicators by the
#' scale scores.
#'
#' All subsequent steps, such as the
#' test functions, will see only the
#' scale scores, or original scores if
#' a variable has no indicator. The
#' model will also be fitted on the
#' scale scores, not on the indicators.
#'
#' It can be used to estimate power
#' for analyzing the scale scores,
#' taking into account the measurement
#' error due to imperfect reliability.
#'
#' @return
#' It returns a data frame with the
#' scale scores computed.
#'
#' @param data A data frame with the
#' indicator scores. It must has an
#' attribute `number_of_indicators`.
#' The same argument used by
#' [power4test()]. This attribute is
#' used to identify the factor names
#' and their indicators.
#'
#' @param method The method to be used
#' to compute the scale scores. Can
#' be `"mean"` or `"sum"`. The default
#' `na.rm = FALSE` will be used.
#' Therefore, `data` must not have
#' missing data.
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Specify the numbers of indicators and reliability coefficients
#'
#' k <- c(y = 3,
#'        m = 4,
#'        x = 5)
#' rel <- c(y = .70,
#'          m = .70,
#'          x = .70)
#'
#' # Simulate the data
#'
#' out <- power4test(
#'          nrep = 2,
#'          model = mod,
#'          pop_es = mod_es,
#'          n = 200,
#'          number_of_indicators = k,
#'          reliability = rel,
#'          process_data = list(fun = "scale_scores"),
#'          test_fun = test_parameters,
#'          test_args = list(op = "~"),
#'          parallel = FALSE,
#'          iseed = 1234)
#'
#' dat <- pool_sim_data(out)
#' head(dat)
#'
#' @export
scale_scores <- function(data,
                         method = c("mean", "sum")) {
  method <- match.arg(method)
  score_function <- switch(
                      method,
                      mean = mean,
                      sum = sum)
  k_ind <- attr(data, "number_of_indicators")
  # ==== Any factors with indicators? ====
  if (isTRUE(is.null(k_ind)) ||
      (all(k_ind == 1))) {
    return(data)
  }
  f_names <- names(k_ind)
  f <- function(x, xk) {
    paste0(x, seq_len(xk))
  }
  ind_names <- mapply(
                FUN = f,
                x = f_names,
                xk = k_ind,
                SIMPLIFY = FALSE
              )
  for (xi in f_names) {
    xi_ind <- ind_names[[xi]]
    xi_scores <- apply(
                  data[, xi_ind, drop = FALSE],
                  MARGIN = 1,
                  FUN = score_function
                )
    data[, xi_ind] <- NULL
    data[, xi] <- xi_scores
  }
  return(data)
}

#' @title Random Variable From a Uniform Distribution
#'
#' @description Generate random numbers
#' from a uniform distribution, with
#' user-specified population mean and
#' standard deviation.
#'
#' @details
#' First, the user specifies
#' the parameters, min and max, and the
#' desired population mean and standard
#' deviation. Then the random numbers
#' will be generated and rescaled with
#' the desired population mean and
#' standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param min min for runif.
#'
#' @param max max for runif.
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @return
#' A vector of the generated random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- runif_rs(n = 5000,
#'               min = 2,
#'               max = 4,
#'               pmean = 3,
#'               psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

runif_rs <- function(n = 10,
                     min = 0,
                     max = 1,
                     pmean = 0,
                     psd = 1) {
    #arg0 <- match.call()
    min0 <- min
    max0 <- max
    bmean <- (min0 + max0) / 2
    bsd   <- (max0 - min0) / sqrt(12)
    x <- stats::runif(n, min0, max0)
    x <- pmean + psd * (x - bmean) / bsd
    x
  }

#' @title Test an Indirect Effect
#'
#' @description Test an indirect effect
#' for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing an
#' indirect effect, by setting it
#' to the `test_fun` argument.
#'
#' It uses [manymome::indirect_effect()]
#' to do the test. It can be used on
#' models fitted by [lavaan::sem()]
#' or fitted by a sequence of calls
#' to [stats::lm()], although only
#' nonparametric bootstrap confidence
#' interval is supported for models
#' fitted by regression using
#' [stats::lm()].
#'
#' @return
#' In its normal usage, it returns
#' a named numeric vector with the
#' following elements:
#'
#' - `est`: The mean of the estimated
#'  indirect effect across datasets.
#'
#' - `cilo` and `cihi`: The means of the
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively.
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' @param fit The fit object, to be
#' passed to [manymome::indirect_effect()].
#'
#' @param x The name of the `x`-variable,
#' the predictor.
#'
#' @param m A character vector of the
#' name(s) of mediator(s). The path
#' moves from the first mediator in the
#' vector to the last mediator in the
#' vector. Can be `NULL` and the path
#' is a direct path without mediator.
#'
#' @param y The name of the `y`-variable,
#' the outcome variable.
#'
#' @param mc_ci Logical. If `TRUE`, the
#' default, Monte Carlo confidence
#' intervals will be formed. This argument
#' and `boot_ci` cannot be both `TRUE`.
#'
#' @param mc_out The pre-generated
#' Monte Carlo estimates generated by
#' [manymome::do_mc], stored in
#' a `power4test` object. Users should
#' not set this argument and should let
#' [power4test()] to set it automatically.
#'
#' @param boot_ci Logical. If `TRUE`,
#' the default, nonparametric bootstrap
#' confidence intervals will be formed.
#' This argument
#' and `mc_ci` cannot be both `TRUE`.
#'
#' @param boot_out The pre-generated
#' bootstrap estimates generated by
#' [manymome::do_boot], stored in
#' a `power4test` object. Users should
#' not set this argument and should let
#' [power4test()] to set it automatically.
#'
#' @param check_post_check Logical. If
#' `TRUE`, the default, and the model
#' is fitted by `lavaan`, the test
#' will be conducted only if the model
#' passes the `post.check` conducted
#' by [lavaan::lavInspect()] (with
#' `what = "post.check"`).
#'
#' @param ... Additional arguments to
#' be passed to [manymome::indirect_effect()].
#'
#' @param fit_name The name of the
#' model fit object to be extracted.
#' Default is `"fit"`. Used only when
#' more than one model is fitted in
#' each replication. This should be
#' the name of the model on which the
#' test is to be conducted.
#'
#' @param get_map_names Logical. Used
#' by [power4test()] to determine how
#' to extract stored information and
#' assign them to this function. Users
#' should not use this argument.
#'
#' @param get_test_name Logical. Used
#' by [power4test()] to get the default
#' name of this test. Users should not
#' use this argument.
#'
#' @param test_method The method to do
#' the test. If `"ci"`, then the
#' confidence interval (e.g., bootstrap
#' confidence interval) will be used to
#' do the test. If
#' `"pvalue"`, then asymmetric *p*-value
#' by Asparouhov & Muthén (2021) will
#' be used to do the test, and the
#' confidence interval will not be
#' computed. If `NULL`, its value will
#' be set to `"pvalue"` if the number of
#' simulated/bootstrap samples (`R`) is
#' a value that is supported by the
#' method by Boos and Zhang (2000), and
#' set to `"ci"` otherwise.
#'
#' @references
#' Asparouhov, A., & Muthén, B. (2021). Bootstrap p-value computation.
#' Retrieved from https://www.statmodel.com/download/FAQ-Bootstrap%20-%20Pvalue.pdf
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_med_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 100,
#'                        R = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test in each replication
#'
#' test_ind <- power4test(object = sim_only,
#'                        test_fun = test_indirect_effect,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         mc_ci = TRUE))
#' print(test_ind,
#'       test_long = TRUE)
#'
#' @export

test_indirect_effect <- function(fit = fit,
                                 x = NULL,
                                 m = NULL,
                                 y = NULL,
                                 mc_ci = TRUE,
                                 mc_out = NULL,
                                 boot_ci = FALSE,
                                 boot_out = NULL,
                                 check_post_check = TRUE,
                                 test_method = NULL,
                                 ...,
                                 fit_name = "fit",
                                 get_map_names = FALSE,
                                 get_test_name = FALSE) {

  # ==== Enable pvalue? ====

  args <- list(...)
  if (!is.null(mc_out)) {
    R <- length(mc_out)
  } else if (!is.null(boot_out)) {
    R <- length(boot_out)
  } else {
    R <- NULL
  }
  ci_level <- args$level %||% formals(manymome::indirect_effect)$level
  R_bz_ok <- isTRUE(R %in% R_extrapolate(alpha = 1 - ci_level))
  bz_not_FALSE <- !isFALSE(options("power4mome.bz"))
  test_method_NULL <- is.null(test_method)
  if (R_bz_ok &&
      bz_not_FALSE &&
      test_method_NULL) {
    test_method <- "pvalue"
  } else {
    test_method <- match.arg(test_method,
                             c("ci", "pvalue"))
  }
  internal_options <- list()
  if (test_method == "pvalue") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = TRUE,
                                               pvalue_min_size = -Inf))
  }
  if (test_method == "ci") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = FALSE))
  }
  if (fit_name != "fit") {
    mc_name <- paste0(fit_name, "_mc_out")
    boot_name <- paste0(fit_name, "_boot_out")
  } else {
    mc_name <- "mc_out"
    boot_name <- "boot_out"
  }
  map_names <- c(fit = fit_name,
                 mc_out = mc_name,
                 boot_out = boot_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp <- paste0(c(x, m, y),
                  collapse = "->")
    args <- as.list(match.call())
    tmp2 <- character(0)
    if (isTRUE(args$standardized_x) && !isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' standardized)")
    }
    if (!isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('y' standardized)")
    }
    if (isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' and 'y' standardized)")
    }
    return(paste0("test_indirect: ", tmp, collapse = ""))
  }
  if (boot_ci) mc_ci <- FALSE
  if (inherits(fit, "lavaan")) {
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
  } else {
    fit_ok <- TRUE
  }
  if (fit_ok) {
    out <- tryCatch(manymome::indirect_effect(
                                   x = x,
                                   y = y,
                                   m = m,
                                   fit = fit,
                                   mc_ci = mc_ci,
                                   mc_out = mc_out,
                                   boot_ci = boot_ci,
                                   boot_out = boot_out,
                                   progress = FALSE,
                                   ...,
                                   internal_options = internal_options),
                  error = function(e) e)
  } else {
    out <- NA
  }
  if (inherits(out, "error") ||
      identical(out, NA)) {
    out2 <- c(est = as.numeric(NA),
              cilo = as.numeric(NA),
              cihi = as.numeric(NA),
              sig = as.numeric(NA),
              pvalue = as.numeric(NA))
    return(out2)
  }
  bz_alpha_ok <- isTRUE(all.equal(1 - out$level,
                                  getOption("power4mome.bz.alpha",
                                                      default = .05)))
  if (test_method == "ci") {
    ci0 <- stats::confint(out)
    out1 <- ifelse((ci0[1, 1] > 0) || (ci0[1, 2] < 0),
                    yes = 1,
                    no = 0)
  }
  if (test_method == "pvalue") {
    ci0 <- matrix(
              as.numeric(NA),
              nrow = 1,
              ncol = 2
            )
    if (bz_alpha_ok) {
      boot_est <- out$boot_indirect %||% out$mc_indirect
      boot_sig <- bz_sig_partition(
                    boot_est,
                    alpha = 1 - out$level
                  )
    } else {
      boot_sig <- NULL
    }
    if (isTRUE(boot_ci)) {
      out1a <- out$boot_p %||% as.numeric(NA)
      R <- length(out$boot_indirect)
      nlt0 <- sum(as.numeric(out$boot_indirect < 0))
    } else if (isTRUE(mc_ci)) {
      out1a <- out$mc_p %||% as.numeric(NA)
      R <- length(out$mc_indirect)
      nlt0 <- sum(as.numeric(out$mc_indirect < 0))
    } else {
      out1a <- as.numeric(NA)
      R <- as.numeric(NA)
      nlt0 <- as.numeric(NA)
    }
    out1 <- ifelse(
                out1a < (1 - out$level),
                yes = 1,
                no = 0
              )
  }
  out2 <- c(est = unname(stats::coef(out)),
            cilo = ci0[1, 1],
            cihi = ci0[1, 2],
            sig = out1,
            pvalue = out$mc_p %||% (out$boot_p %||% as.numeric(NA)))
  if (test_method == "pvalue") {
    # For Boos & Zhang (2000)
    out2 <- c(out2, R = R, nlt0 = nlt0,
              alpha = 1 - out$level,
              boot_sig)
  }
  return(out2)
}

#' @title Predict Method for a 'power_curve' Object
#'
#' @description Compute the predicted
#' values in a model fitted by
#' [power_curve()].
#'
#' @details
#' The `predict` method of `power_curve`
#' objects works in two modes.
#'
#' If new
#' data is not supplied (through
#' `newdata`), it retrieves the stored
#' results and calls the corresponding
#' methods to compute the predicted
#' values, which are the predicted
#' rejection rates (power levels if
#' the null hypothesis is false,
#' e.g., the population effect size is
#' equal to zero).
#'
#' If new data is supplied, such as
#' a named list with a vector of
#' sample sizes, they will be used to
#' compute the predicted rejection
#' rates.
#'
#' @return
#' It returns a numeric vector of
#' the predicted rejection rates.
#'
#' @param object A `power_curve`
#' object.
#'
#' @param newdata A data frame with
#' a column named `x`. It can also
#' be a named list, with one element
#' named `x` and is a vector of the
#' values. If not supplied, values of
#' `x` stored in `object` will be used.
#'
#' @param ...  Additional arguments.
#' Passed to the corresponding
#' `predict` method.
#'
#' @seealso [power_curve()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the effect sizes (population parameter values)
#'
#' model_simple_med_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: s
#' "
#'
#' # Simulate datasets to check the model
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 50,
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234,
#'                        parallel = FALSE,
#'                        progress = FALSE)
#'
#' # By n: Do a test for different sample sizes
#'
#' out1 <- power4test_by_n(sim_only,
#'                         nrep = 5,
#'                         test_fun = test_parameters,
#'                         test_args = list(par = "y~x"),
#'                         n = c(25, 100, 200, 1000),
#'                         by_seed = 1234,
#'                         parallel = FALSE,
#'                         progress = FALSE)
#'
#' pout1 <- power_curve(out1)
#' pout1
#' predict(pout1,
#'         newdata = list(x = c(150, 250, 500)))
#'
#'
#' @export

predict.power_curve <- function(object,
                                newdata,
                                ...) {
  fit <- object$fit
  if (missing(newdata)) {
    newdata <- object$reject_df
  }
  if (inherits(fit, "glm")) {
    out <- stats::predict(object = fit,
                          newdata = newdata,
                          type = "response",
                          ...)
  } else {
    out <- stats::predict(object = fit,
                          newdata = newdata,
                          ...)
  }
  return(out)
}

#' @title Test All Moderation Effects
#'
#' @description Test all moderation
#' effects by testing all product
#' terms for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing all
#' product terms, by
#' setting it to the `test_fun`
#' argument.
#'
#' It is just a wrapper to
#' [test_parameters()]. It will
#' first identifies all product
#' terms (terms with `:` in the names),
#' and then call [test_parameters()],
#' with `pars` set to select the
#' regression coefficients of these
#' terms.
#'
#' @return
#' In its normal usage, it returns
#' the output returned by
#' [lavaan::parameterEstimates()]
#' or [lmhelprs::lm_list_to_partable()],
#' with the following modifications:
#'
#' - `est`: The parameter estimates,
#'  even if standardized estimates
#'  are requested (not `est.std`).
#'
#' - `cilo` and `cihi`: The
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively (not
#'  `ci.lower` and `ci.upper`).
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' - `test_label`: A column of labels
#'  generated by
#'  [lavaan::lav_partable_labels()],
#'  which are usually the labels used by
#'  `coef()` to label the parameters.
#'
#' @inheritParams test_parameters
#'
#' @seealso [power4test()],
#' [test_parameters()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x + w1 + x:w1
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' m ~ x: n
#' y ~ x: m
#' m ~ w1: n
#' m ~ x:w1: l
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 4,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test in each replication
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_moderation)
#'
#' print(test_out,
#'       test_long = TRUE)
#'
#' @export
test_moderation <- function(fit = fit,
                            standardized = FALSE,
                            check_post_check = TRUE,
                            ...,
                            p_adjust_method = "none",
                            fit_name = "fit",
                            get_map_names = FALSE,
                            get_test_name = FALSE) {
  map_names <- c(fit = fit_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp <- character(0)
    if (standardized) {
      return(paste("test_moderation: CIs (standardized)", tmp))
    } else {
      return(paste("test_moderation: CIs", tmp))
    }
  }

  if (inherits(fit, "lm_list")) {
    fit_type <- "lm_list"
  } else if (inherits(fit, "lavaan")) {
    fit_type <- "lavaan"
  } else {
    stop("fit is not a supported object.")
  }

  if (fit_type == "lm_list") {
    est0 <- lmhelprs::lm_list_to_partable(fit)
  } else {
    est0 <- lavaan::parameterEstimates(fit,
                                      se = FALSE,
                                      pvalue = FALSE,
                                      ci = FALSE)
  }
  est0$test_label <- lavaan::lav_partable_labels(est0)
  i <- est0$op %in% "~"
  j <- grepl(":", est0$test_label)
  k <- i & j
  if (length(k) == 0) {
    stop("No product terms found in the model.")
  }
  pars <- est0$test_label[k]
  out <- test_parameters(fit = fit,
                         standardized = standardized,
                         pars = pars,
                         p_adjust_method = p_adjust_method,
                         ...)
  return(out)
}

#' @title Simulate Datasets Based on a Model
#'
#' @description Get a model matrix and
#' effect size specification and
#' simulate a number of datasets,
#' along with other information.
#'
#' @details
#'
#' The function [sim_data()] generates
#' a list of datasets based on a population
#' model.
#'
#' # The role of `sim_data()`
#'
#' The function [sim_data()] is used by
#' the all-in-one function
#' [power4test()]. Users usually do not
#' call this function directly, though
#' developers can use this function to
#' develop other functions for power
#' analysis, or to build their own
#' workflows to do the power analysis.
#'
#' # Workflow
#'
#' The function [sim_data()] does two tasks:
#'
#' - Determine the actual population
#'  model with population values based
#'  on:
#'
#'    - A model syntax for the observed
#'      variables (for a path model)
#'      or the latent factors (for a
#'      latent variable model).
#'
#'    - A textual specification of the
#'      effect sizes of parameters.
#'
#'    - The number of indicators for
#'      each latent factor if the model
#'      is a latent variable model.
#'
#'    - The reliability of each latent
#'      factor as measured by the
#'      indicators if the model is a
#'      latent factor model.
#'
#' - Generate *nrep* simulated datasets
#'  from the population model.
#'
#' The simulated datasets can then be
#' used to fit a model, test
#' parameters, and estimate power.
#'
#' The output is usually used by
#' [fit_model()] to fit a target model,
#' by default the population model, to each
#' of the dataset.
#'
#' # Set 'number_of_indicators' and 'reliability'
#'
#' The arguments `number_of_indicators`
#' and `reliability` are used to
#' specify the number of indicators
#' (e.g., items) for each factor,
#' and the population reliability
#' coefficient of each factor,
#' if the variables in the model
#' syntax are latent variables.
#'
#' Optionally, `loading_difference` can
#' be used to generate indicators with
#' unequal standardized factor loadings,
#' and `reference` can be used to specify
#' the indicator with the medium,
#' strongest, or weakest standardized
#' factor loading as the first indicator,
#' which is used as the reference indicator
#' in `lavaan`.
#'
#' ## Single-Group Model
#'
#' If a variable in the model is to be
#' replaced by indicators in the generated
#' data, set
#' `number_of_indicators` to a named
#' numeric vector. The names are the
#' variables of variables with
#' indicators, as appeared in the
#' `model` syntax. The value of each
#' name is the number of indicators.
#'
#' The
#' argument `reliability` should then be
#' set a named numeric vector (or list,
#' see the section on multigroup models)
#' to specify the population reliability
#' coefficient ("omega") of each set of
#' indicators. The population standardized factor
#' loadings are then computed to ensure
#' that the population reliability
#' coefficient is of the target value.
#'
#' These are examples for a single group
#' model:
#'
#' \preformatted{number of indicator = c(m = 3, x = 4, y = 5)}
#'
#' The numbers of indicators for `m`,
#' `x`, and `y` are 3, 4, and 5,
#' respectively.
#'
#' \preformatted{reliability = c(m = .90, x = .80, y = .70)}
#'
#' The population reliability
#' coefficients of `m`, `x`, and `y` are
#' .90, .80, and .70, respectively.
#'
#' ## Multigroup Models
#'
#' Multigroup models are supported.
#' The number of groups is inferred
#' from `pop_es` (see the help page
#' of [ptable_pop()]), or directly
#' from `ptable`.
#'
#' For a multigroup model, the number
#' of indicators for each variable
#' must be the same across groups.
#'
#' However, the population reliability
#' coefficients can be different
#' across groups. For a multigroup model
#' of *k* groups,
#' with one or more population reliability
#' coefficients differ across groups,
#' the argument `reliability` should be
#' set to a named list. The names are
#' the variables to which the population
#' reliability coefficients are to be
#' set. The element for each name is
#' either a single value for the common
#' reliability coefficient, or a
#' numeric vector of the reliability
#' coefficient of each group.
#'
#' This is an example of `reliability`
#' for a model with 2 groups:
#'
#' \preformatted{reliability = list(x = .80, m = c(.70, .80))}
#'
#' The reliability coefficients of `x` are
#' .80 in all groups, while the
#' reliability coefficients of `m` are
#' .70 in one group and .80 in another.
#'
#' ## Equal Numbers of Indicators and/or Reliability Coefficients
#'
#' If all variables in the model has
#' the same number of indicators,
#' `number_of_indicators` can be set
#' to one single value.
#'
#' Similarly, if all sets of indicators
#' have the same population reliability
#' in all groups, `reliability` can also
#' be set to one single value.
#'
#' # Specify The Distributions of Exogenous Variables Or Error Terms Using 'x_fun'
#'
#' By default, variables and error
#' terms are generated
#' from a multivariate normal distribution.
#' If desired, users can supply the
#' function used to generate an exogenous
#' variable and error term by setting `x_fun` to
#' a named list.
#'
#' The names of the list are the variables
#' for which a user function will be used
#' to generate the data.
#'
#' Each element of the list must also
#' be a list. The first element of this
#' list, can be unnamed, is the
#' function to be used. If other arguments
#' need to be supplied, they should be
#' included as named elements of this list.
#'
#' For example:
#'
#' \preformatted{x_fun = list(x = list(power4mome::rexp_rs),
#'              w = list(power4mome::rbinary_rs,
#'                       p1 = .70)))}
#'
#' The variables `x` and `w` will be
#' generated by user-supplied functions.
#'
#' For `x`, the function is
#' `power4mome::rexp_rs`. No additional
#' argument when calling this function.
#'
#' For `w`, the function is
#' `power4mome::rbinary_rx`. The argument
#' `p1 = .70` will be passed to this
#' function when generating the values
#' of `w`.
#'
#' If a variable is an endogenous
#' variable (e.g., being predicted by
#' another variable in a model), then
#' `x_fun` is used to generate its
#' *error term*. Its implied population
#' distribution may still be different
#' from that generate by `x_fun` because
#' the distribution also depends on the
#' distribution of other variables
#' predicting it.
#'
#' These are requirements for the
#' user-functions:
#'
#' - They must return a numeric vector.
#'
#' - They mush has an argument `n` for
#'   the number of values.
#'
#' - The *population* mean and standard
#'   deviation of the generated values
#'   must be 0 and 1, respectively.
#'
#' The package `power4mome` has
#' helper functions for generating
#' values from some common nonnormal
#' distributions and then scaling them
#' to have population mean and standard
#' deviation equal to 0 and 1 (by default), respectively.
#' These are some of them:
#'
#' - [rbinary_rs()].
#'
#' - [rexp_rs()].
#'
#' - [rbeta_rs()].
#'
#' - [rlnorm_rs()].
#'
#' - [rpgnorm_rs()].
#'
#' To use `x_fun`, the variables must
#' have zero covariances with other
#' variables in the population. It is
#' possible to generate nonnormal
#' multivariate data but we believe this
#' is rarely needed when estimating
#' power *before* having the data.
#'
#' @inheritSection ptable_pop Specify the Population Model by 'model'
#'
#' @inheritSection ptable_pop Specify 'pop_es' Using Named Vectors
#'
#' @inheritSection ptable_pop Specify 'pop_es' Using a Multiline String
#'
#' @inheritSection ptable_pop Set the Values for Effect Size Labels ('es1' and 'es2')
#'
#' @seealso [power4test()]
#'
#'
#' @param nrep The number of replications
#' to generate the simulated datasets.
#' Default is 10.
#'
#' @param ptable The output of
#' [ptable_pop()], which is a
#' `ptable_pop` object, representing the
#' population model. If `NULL`, the
#' default, [ptable_pop()] will be
#' called to generate the `ptable_pop`
#' object, using arguments such as
#' `model` and `pop_es`.
#'
#' @param model The `lavaan` model
#' syntax of the population model.
#' Ignored if `ptable` is
#' specified. See [ptable_pop] on
#' how to specify this argument.
#'
#' @param pop_es The character to
#' specify population effect sizes.
#' See [ptable_pop] on
#' how to specify this argument.
#' Ignored if `ptable` is
#' specified.
#'
#' @param ... For [sim_data], parameters
#' to be passed to [ptable_pop()]. For
#' [print.sim_data()], these arguments
#' are ignored.
#'
#' @param n The sample size for each
#' dataset. Default is 100.
#'
#' @param iseed The seed for the random
#' number generator. Default is `NULL`
#' and the seed is not changed.
#'
#' @param number_of_indicators A named
#' vector to specify the number of
#' indicators for each factors. See
#' the help page on how to set this
#' argument. Default is `NULL` and all
#' variables in the model syntax are
#' observed variables.
#' See the help page on how
#' to use this argument.
#'
#' @param reliability A named vector
#' (for a single-group model) or a
#' named list of named vectors
#' (for a multigroup model)
#' to set the reliability coefficient
#' of each set of indicators. Default
#' is `NULL`.
#' See the help page on how
#' to use this argument.
#'
#' @param loading_difference A named vector
#' (for a single-group model) or a
#' named list of named vectors
#' (for a multigroup model)
#' to set the difference in factor
#' loadings between neighboring indicators
#' of each set of indicators. Default
#' is `NULL`, and all indicators of
#' a factor have
#' the same factor loadings. If specified,
#' must be specified for all factors
#' named in `reliability`, even for
#' those with all loadings equal.
#'
#' @param reference A named vector
#' (for a single-group model) or a
#' named list of named vectors
#' (for a multigroup model)
#' to indicate which indicator will
#' be the first indicator (and so
#' is the reference indicator, by default).
#' Default
#' is `NULL`, and for all factors,
#' the indicator with
#' the medium loading in a factor is
#' the first indicator. Has no effect
#' if loading difference is zero (and
#' so all indicators have the same
#' loadings). If specified,
#' must be specified for all factors
#' named in `reliability`, even for
#' those with all loadings equal. Accepted
#' values are `"medium"`, `"weakest"`,
#' and `"strongest"`.
#'
#' @param x_fun The function(s) used to
#' generate the exogenous variables or
#' error terms. If
#' not supplied, or set to `list()`, the
#' default, the variables are generated
#' from a multivariate normal
#' distribution. See the help page on how
#' to use this argument.
#'
#' @param e_fun The function(s) used to
#' generate the error terms of indicators,
#' if any. If
#' not supplied, or set to `list()`, the
#' default, the error terms of indicators
#' are generated
#' from a multivariate normal
#' distribution. Specify in the same
#' way as `x_fun`. Refer to the help
#' page on `x_fun` on how to use this
#' argument.
#'
#' @param process_data If not `NULL`, it
#' must be a named list with these
#' elements: `fun` (required), the function
#' to further process the simulated
#' data, such as generating missing data using
#' functions such as [mice::ampute()]; `args` (optional), a
#' named list of arguments to be passed
#' to `fun`, except the one for the
#' source data; `sim_data_name` (optional) the
#' name of the argument to receive the
#' simulated data (e.g., `"data"` for
#' [mice::ampute()]), default to
#' `"data"` if it is not set; `processed_data_name`
#' (optional), the name of the data frame
#' after being processed by `fun`,
#' such as the data frame
#' with missing data in the output of
#' `fun` (e.g., `"amp"` for [mice::ampute()]),
#' if omitted, the output of `fun` should
#' be the data frame with missing data.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to simulate
#' the datasets. Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' of data simulation will be displayed.
#' Default is `FALSE.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param n_ratio If the model is a
#' multigroup model, and `n` is a single
#' number, this should be a numeric
#' vector used to determine the sample
#' size for each group. For example,
#' for a two-group model, if `n` is 100
#' and `n_ratio` is `c(1, 0.5)`, then
#' the sample sizes for the two groups
#' are 100 and 50, respectively. If
#' equal to 1, then all groups have the
#' same sample size.
#'
#' @param cl A cluster, such as one created
#' by [parallel::makeCluster()]. If `NULL`,
#' a cluster will be created, but will be
#' stopped on exit. If set to an existing
#' cluster, it will not be stopped when
#' the function exits; users need to
#' stop it manually.
#'
#' @return
#' The function [sim_out()] returns
#' a list of the class `sim_data`,
#' with length `nrep`. Each element
#' is a `sim_data_i` object, with
#' the following major elements:
#'
#' - `ptable`: A `lavaan` parameter
#'  table of the model, with population
#'  values set in the column `start`.
#'  (It is the output of the
#'  function [ptable_pop()].)
#'
#' - `mm_out`: The population model
#'  represented by model matrices
#'  as in `lavaan`. (It is the output
#'  of the function
#'  [model_matrices_pop()].)
#'
#' - `mm_lm_out`: A list of regression
#'  model formula, one for each
#'  endogenous variable. (It is the
#'  output of the internal function
#'  `mm_lm()`.)
#'
#' - `mm_lm_dat_out`: A simulated dataset
#'  generated from the population model.
#'  (It is the output of the internal
#'  function `mm_lm_data()`).
#'
#' - `model_original`: The original model
#'  syntax (i.e., the argument `model`).
#'
#' - `model_final`: A modified model
#'  syntax if the model is a latent
#'  variable model. Indicators are added
#'  to the syntax.
#'
#' - `fit0`: The output of [lavaan::sem()]
#'  with `ptable` as the model and
#'  `do.fit` set to `FALSE`. Used for the
#'  easy retrieval of information
#'  about the model.
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "m ~ x
#'  y ~ m + x"
#'
#' # Specify the population values
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate the simulated datasets
#'
#' data_all <- sim_data(nrep = 5,
#'                      model = mod,
#'                      pop_es = es,
#'                      n = 100,
#'                      iseed = 1234)
#'
#' data_all
#'
#' @export
sim_data <- function(nrep = 10,
                     ptable = NULL,
                     model = NULL,
                     pop_es = NULL,
                     ...,
                     n = 100,
                     iseed = NULL,
                     number_of_indicators = NULL,
                     reliability = NULL,
                     loading_difference = NULL,
                     reference = NULL,
                     x_fun = list(),
                     e_fun = list(),
                     process_data = NULL,
                     parallel = FALSE,
                     progress = FALSE,
                     ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                     n_ratio = 1,
                     cl = NULL) {

  if (is.null(ptable)) {
    if (is.null(model) || is.null(pop_es)) {
      stop("Both model and pop_es must be set if ptable is not set.")
    }
    ptable <- ptable_pop(model = model,
                         pop_es = pop_es,
                         ...)
  }
  mm_out <- model_matrices_pop(ptable,
                               drop_list_single_group = FALSE)
  mm_lm_out <- mm_lm(mm_out,
                     drop_list_single_group = FALSE)

  fit_tmp <- try(lavaan::sem(
                    ptable,
                    do.fit = FALSE
                  ),
                silent = TRUE)
  if (inherits(fit_tmp, "try-error")) {
    all_paths <- list()
  } else {
    all_paths <- manymome::all_indirect_paths(
                      fit_tmp
                    )
  }

  out <- do_FUN(X = seq_len(nrep),
                FUN = sim_data_i,
                ptable = ptable,
                model = model,
                mm_out = mm_out,
                mm_lm_out = mm_lm_out,
                n = n,
                number_of_indicators = number_of_indicators,
                reliability = reliability,
                loading_difference = loading_difference,
                reference = reference,
                x_fun = x_fun,
                e_fun = e_fun,
                process_data = process_data,
                fit_external = list(all_paths = all_paths),
                iseed = iseed,
                parallel = parallel,
                progress = progress,
                ncores = ncores,
                n_ratio = n_ratio,
                cl = cl)
  class(out) <- c("sim_data", class(out))
  return(out)
}

#' @param digits The numbers of digits
#' displayed after the decimal.
#'
#' @param digits_descriptive The
#' number of digits displayed after
#' the decimal for the descriptive
#' statistics table.
#'
#' @param x The `sim_data` object
#' to be printed.
#'
#' @param data_long If `TRUE`, detailed
#' information will be printed.
#'
#' @param fit_to_all_args A named list
#' of arguments to be passed to
#' [lavaan::sem()] when the model is
#' fitted to a sample combined from
#' all samples stored.
#'
#' @param est_type The type of estimates
#' to be printed. Can be a character
#' vector of one to two elements. If
#' only `"standardized"`, then the
#' standardized estimates are printed.
#' If only `"unstandardized"`, then the
#' unstandardized estimates are printed.
#' If a vector like
#' `c("standardized", "unstandardized")`,
#' then both unstandardized and
#' standardized estimates are printed.
#'
#' @param variances Logical. Whether
#' variances and error variances are printed.
#' Default depends on `est_type`. If
#' `"unstandardized"` is in `est_type`,
#' then default is `TRUE` If
#' only `"standardized"` is in `est_type`,
#' then default is `FALSE`.
#'
#' @param pure_x,pure_y When Logical. When
#' printing indirect effects, whether
#' only "pure" x-variables (variables
#' not predicted by another other variables)
#' and/or "pure" y-variables (variables
#' that do not predict any other variables
#' other than indicators) will be included
#' in enumerating the paths.
#'
#' @return
#' The `print` method of `sim_data`
#' returns `x` invisibly. It is called for
#' its side effect.
#'
#' @rdname sim_data
#' @export
print.sim_data <- function(x,
                           digits = 3,
                           digits_descriptive = 2,
                           data_long = TRUE,
                           fit_to_all_args = list(),
                           est_type = "standardized",
                           variances = NULL,
                           pure_x = TRUE,
                           pure_y = TRUE,
                           ...) {

  est_type <- match.arg(est_type,
                        choices = c("standardized", "unstandardized"),
                        several.ok = TRUE)

  if (is.null(variances)) {
    if (identical("standardized", est_type)) {
      variances <- FALSE
    } else {
      variances <- TRUE
    }
  }

  # This line needed for printing boot_out and mc_out
  requireNamespace("manymome", quietly = TRUE)

  x_i <- x[[1]]
  ptable <- x_i$ptable
  fit0 <- x_i$fit0

  group_labels <- x_i$group_labels
  ngroups <- length(group_labels)
  if (is.null(ngroups)) ngroups <- 1

  ptable0 <- lavaan::parameterTable(fit0)
  ptable <- set_user_pop(ptable,
                         fit = fit0)
  ptable1 <- ptable[, c("lhs",
                        "op",
                        "rhs",
                        "group",
                        "block",
                        "exo",
                        "label",
                        "start")]
  colnames(ptable1)[which(colnames(ptable1) == "start")] <- "Population"
  if (max(ptable1$group) == 1) {
    ptable1$group <- NULL
  }

  class(ptable1) <- c("lavaan.parameterEstimates", class(ptable1))
  if (ngroups > 1) {
    attr(ptable1, "group.label") <- group_labels
  }

  model_original <- x_i$model_original

  cat(header_str("Model Information",
                 prefix = "\n",
                 suffix = "\n"))

  cat(header_str("Model on Factors/Variables",
                 hw = .4,
                 prefix = "\n",
                 suffix = "\n"))
  cat(model_original)

  model_final <- x_i$model_final
  cat(header_str("Model on Variables/Indicators",
                 hw = .4,
                 prefix = "\n",
                 suffix = "\n"))
  cat(model_final, sep = "\n")

  cat(header_str("Population Values",
                 hw = .4,
                 prefix = "",
                 suffix = "\n"))
  print(ptable1,
        digits = digits)

  # Multigroup models automatically supported
  has_w <- FALSE
  all_ind <- tryCatch(pop_indirect(x,
                                   pure_x = pure_x,
                                   pure_y = pure_y,
                                   progress = TRUE),
                       warning = function(w) w,
                       error = function(e) e)
  if (!inherits(all_ind, "error")) {
    if (length(all_ind) > 0) {
      cat(header_str("Population Conditional/Indirect Effect(s)",
                    hw = .4,
                    prefix = "",
                    suffix = "\n"))
      for (all_ind_i in all_ind) {
        print(all_ind_i,
              digits = digits)
      }
      cat("\n")
    }
  }

  k0 <- x_i$number_of_indicators

  rel0 <- x_i$reliability

  if (!is.null(rel0[[1]])) {
    rel <- do.call(rbind,
                  rel0)
    rel <- as.data.frame(rel)
    rel_n <- nrow(rel)
    rownames(rel) <- paste0("Group ", seq_len(rel_n))
    cat(header_str("Population Reliability",
                  hw = .4,
                  prefix = "",
                  suffix = "\n\n"))
    print(rel,
          row.names = isTRUE(rel_n > 1),
          digits = digits)
  }

  if (!is.null(k0[[1]]) &&
      !is.null(x_i$lambda)) {
    # TODO:
    # - Handle loading_difference != 0
    lambda_pop <- x_i$lambda
    tmpf <- function(x) {
      tmp <- lapply(
          x,
          \(y) paste0(formatC(y, digits = digits, format = "f"),
                      collapse = ",")
        )
      out0 <- do.call(
                rbind,
                tmp
              )
      colnames(out0) <- "Standardized Loadings"
      out0
    }
    if (ngroups > 1) {
      lambda_str <- lapply(
          lambda_pop,
          tmpf
        )
    } else {
      lambda_str <- tmpf(lambda_pop)
    }

    cat(header_str("Population Standardized Loadings",
                  hw = .4,
                  prefix = "\n",
                  suffix = "\n"))

    if (ngroups > 1) {
      for (tmpi in names(lambda_str)) {
        cat("\n", tmpi, ":\n")
        print(lambda_str[[tmpi]],
              quote = FALSE)
      }
    } else {
      cat("\n")
      print(lambda_str,
            quote = FALSE)
    }
    cat("\n")

    # # The following code print the wrong standardized loadings
    # # However, the computation is correct. Only the printout is wrong.
    # lambda0 <- mapply(function(xx, yy) {
    #                     out <- mapply(lambda_from_reliability,
    #                                   p = xx,
    #                                   omega = yy)
    #                     out
    #                   },
    #                   xx = k0,
    #                   yy = rel0,
    #                   SIMPLIFY = FALSE)
    # lambda <- do.call(rbind,
    #                   lambda0)
    # lambda <- as.data.frame(lambda)
    # lambda_n <- nrow(lambda)
    # rownames(lambda) <- paste0("Group ", seq_len(lambda_n))
    # cat(header_str("Population Standardized Loadings",
    #               hw = .4,
    #               prefix = "\n",
    #               suffix = "\n\n"))
    # print(lambda,
    #       row.names = isTRUE(lambda_n > 1),
    #       digits = digits)

  }

  # Summarize data

    nrep <- length(x)
    n <- nrow(x_i$mm_lm_dat_out)

    cat(header_str("Data Information",
                  prefix = "",
                  suffix = "\n\n"))

    cat("Number of Replications: ", nrep, "\n")
    cat("Sample Sizes: ", paste0(n, collapse = ", "), "\n")

  if (data_long) {
    all_data <- pool_sim_data(x)

    fit_to_all_args0 <- list(model = model_final,
                             data = all_data,
                             se = "none",
                             test = "none",
                             group = x_i$group_name,
                             fixed.x = FALSE)
    fit_to_all_args1 <- utils::modifyList(fit_to_all_args0,
                                          fit_to_all_args)
    fit_all <- do.call(lavaan::sem,
                       fit_to_all_args1)

    if (isTRUE(identical(est_type, "standardized"))) {
      # Standardized Only
      est_all <- lavaan::standardizedSolution(fit_all,
                                              se = FALSE,
                                              pvalue = FALSE,
                                              ci = FALSE,
                                              output = "text")
      if (!variances) {
        i <- est_all$lhs == est_all$rhs
        est_all <- est_all[!i, ]
      }
      tmp_est_hdr <- "Standardized Estimates"
    }

    if (isTRUE(identical(est_type, "unstandardized"))) {
      # Unstandardized Only
      est_all <- lavaan::parameterEstimates(fit_all,
                                            se = FALSE,
                                            pvalue = FALSE,
                                            ci = FALSE,
                                            standardized = FALSE,
                                            output = "text")
      if (!variances) {
        i <- est_all$lhs == est_all$rhs
        est_all <- est_all[!i, ]
      }
      tmp_est_hdr <- "Unstandardized Estimates"
    }

    if (isTRUE(all(c("unstandardized", "standardized") %in% est_type))) {
      # Both unstandardized and standardized
      est_all <- lavaan::parameterEstimates(fit_all,
                                            se = FALSE,
                                            pvalue = FALSE,
                                            ci = FALSE,
                                            standardized = TRUE,
                                            output = "text")
      if (!variances) {
        i <- est_all$lhs == est_all$rhs
        est_all <- est_all[!i, ]
      }
      tmp_est_hdr <- "Unstandardized and Standardized Estimates"
    }

    cat(header_str("Descriptive Statistics",
                  hw = .4,
                  prefix = "\n",
                  suffix = "\n\n"))
    if (ngroups > 1) {
      tmp <- which(colnames(all_data) == x_i$group_name)
      print(psych::describeBy(
                            all_data[, -tmp],
                            group = all_data[[x_i$group_name]],
                            range = FALSE),
            digits = digits_descriptive)
    } else {
      print(psych::describe(all_data,
                            range = FALSE),
            digits = digits_descriptive)
    }

    # ==== Response Frequencies ====

    response_count <- sapply(
                        all_data[seq_len(n * nrep), ],
                        \(x) length(unique(x))
                      )
    tmp1 <- response_count <= 10
    if (!is.null(x_i$group_name)) {
      tmp1[x_i$group_name] <- FALSE
    }

    if (any(tmp1)) {
      cat(header_str("Response Proportions",
                    hw = .4,
                    prefix = "\n",
                    suffix = "\n"))
      tmp2 <- unique(response_count[tmp1])
      if (ngroups > 1) {
        group <- all_data[[x_i$group_name]]
      } else {
        group <- rep("Single Group", nrow(all_data))
      }
      for (zz in tmp2) {
        tmp3 <- (response_count == zz)
        tmp5 <- by(all_data[, tmp3, drop = FALSE],
                   INDICES = group,
                   psych::responseFrequency,
                   simplify = FALSE)
        cat("\n")
        print(tmp5,
              digits = digits_descriptive)
      }
    }
    # Print missing data pattern

    mp <- tryCatch(miss_pattern(all_data),
                   error = function(e) e)
    if (!inherits(mp, "error")) {
      if (nrow(mp) != 1) {
        cat(header_str("Missing Data Pattern",
                          hw = .4,
                          prefix = "\n",
                          suffix = "\n\n"))
        cat("Missing data is present\n\n")
        print_miss_pattern(mp,
                           digits = max(0, digits = digits - 1))
      }
    }

    tmp <- paste("Parameter Estimates Based on All",
                nrep,
                "Samples Combined")
    cat(header_str(tmp,
                  prefix = "\n",
                  suffix = "\n\n"))

    cat("Total Sample Size:", n * nrep, "\n")

    cat(header_str(tmp_est_hdr,
                  hw = .4,
                  prefix = "\n",
                  suffix = "\n\n"))

    if (!variances) {
      cat("Variances and error variances omitted.\n")
    }

    print(est_all,
          nd = digits)
  } else {
    cat("\nCall print with 'data_long = TRUE' for further information.\n")
  }

  invisible(x)
}

#' @description
#' The function
#'
#' @param object Either a `sim_data`
#' object or a `power4test` object.
#' It extracts the simulated data
#' and return them, combined to one
#' single data frame or, if `as_list`
#' is `TRUE`, as a list of data
#' frames.
#'
#' @param as_list Logical. If `TRUE`,
#' the simulated datasets is returned as one
#' single data frame. If `FALSE`, they
#' are returned as a list of data
#' frames.
#'
#' @return
#' The function `pool_sim_data()` returns
#' either one data frame or a list
#' of data frames, depending on the
#' argument `as_list`
#'
#' @rdname sim_data
#' @export
pool_sim_data <- function(object,
                          as_list = FALSE) {
  if (!inherits(object, "power4test") &&
      !inherits(object, "sim_data")) {
    stop("pool_sim_data() only supports 'power4test' or 'sim_data' objects.")
  }
  if (inherits(object, "power4test")) {
    object <- object$sim_all
  }
  all_data <- lapply(object,
                     function(x) x$mm_lm_dat_out)
  if (!as_list) {
    all_data <- do.call(rbind,
                        all_data)
  }
  all_data
}

#' @noRd
set_user_pop <- function(ptable,
                         fit) {
  if (!(":=" %in% ptable$op)) {
    return(ptable)
  }
  user_fun <- fit@Model@def.function
  i <- ptable$free
  i <- i[i > 0]
  j <- match(i, ptable$free)
  pop <- ptable$start[j]
  user_pop <- user_fun(pop)
  k <- match(names(user_pop), ptable$lhs)
  ptable[k, "start"] <- user_pop
  ptable
}


#' @noRd
sim_data_i <- function(repid = 1,
                       n = 100,
                       model = NULL,
                       pop_es = NULL,
                       ptable = NULL,
                       mm_out = NULL,
                       mm_lm_out = NULL,
                       number_of_indicators = NULL,
                       reliability = NULL,
                       loading_difference = NULL,
                       reference = NULL,
                       x_fun = list(),
                       e_fun = list(),
                       process_data = NULL,
                       fit_external = NULL,
                       seed = NULL,
                       drop_list_single_group = TRUE,
                       merge_groups = TRUE,
                       n_ratio = 1) {
  if (!is.null(seed)) set.seed(seed)
  if (is.null(ptable)) {
    ptable <- ptable_pop(model = model,
                        pop_es = pop_es,
                        standardized = TRUE)
  } else {
    model <- attr(ptable, "model")
  }
  if (is.null(mm_out)) {
    mm_out <- model_matrices_pop(ptable,
                                drop_list_single_group = FALSE)
  }
  if (is.null(mm_lm_out)) {
  mm_lm_out <- mm_lm(mm_out,
                     drop_list_single_group = FALSE)
  }

  ngroups <- max(ptable$group)
  if (ngroups > 1) {

    # ==== Set sample sizes for multigroup models ====

    if (length(n) == 1) {

      # ==== Use n_ratio ====

      if ((length(n_ratio) != ngroups)) {
        if (length(n_ratio) == 1) {
          n_ratio <- rep(n_ratio, ngroups)
        } else {
          stop("The length of n_ratio is not equal to the number of groups.")
        }
      }
      n <- round(n_ratio * n)
    } else {

      # ==== n is a vector ====

      if (length(n) != ngroups) {
        stop("The length of n is not equal to the number of groups.")
      }
      # n is used directly
    }
  }

  vnames <- lavaan::lavNames(ptable,
                             type = "ov")
  p <- length(vnames)

  if ((length(number_of_indicators) == 1) &&
      is.numeric(number_of_indicators) &&
      is.null(names(number_of_indicators))) {
    number_of_indicators <- rep(number_of_indicators,
                                p)
    names(number_of_indicators) <- vnames
  }
  if (!is.list(number_of_indicators)) {
    number_of_indicators <- rep(list(number_of_indicators),
                                ngroups)
  }

  if ((length(reliability) == 1) &&
      is.numeric(reliability) &&
      is.null(names(reliability))) {
    reliability <- rep(reliability,
                                p)
    names(reliability) <- vnames
  }
  if (!is.list(reliability)) {
    reliability <- rep(list(reliability),
                            ngroups)
  } else {
    reliability <- split_par_es(reliability)
  }

  if ((length(loading_difference) == 1) &&
      is.numeric(loading_difference) &&
      is.null(names(loading_difference))) {
    loading_difference <- rep(loading_difference,
                             p)
    names(loading_difference) <- vnames
  }
  if (!is.list(loading_difference)) {
    loading_difference <- rep(list(loading_difference),
                            ngroups)
  } else {
    loading_difference <- split_par_es(loading_difference)
  }

  if ((length(reference) == 1) &&
      is.numeric(reference) &&
      is.null(names(reference))) {
    reference <- rep(reference,
                             p)
    names(reference) <- vnames
  }
  if (!is.list(reference)) {
    reference <- rep(list(reference),
                            ngroups)
  } else {
    reference <- split_par_es(reference)
  }
  mm_lm_dat_out <- mapply(mm_lm_data,
                          object = mm_lm_out,
                          n = n,
                          number_of_indicators = number_of_indicators,
                          reliability = reliability,
                          loading_difference = loading_difference,
                          reference = reference,
                          MoreArgs = list(keep_f_scores = FALSE,
                                          x_fun = x_fun,
                                          e_fun = e_fun,
                                          process_data = process_data),
                          SIMPLIFY = FALSE)
  # mm_lm_dat_out is always a list, even for
  # single-group models
  lambda_pop <- sapply(
                  mm_lm_dat_out,
                  attr,
                  which = "lambda",
                  simplify = FALSE
                )
  model_original <- model
  # add_indicator_syntax() already supports
  # a model syntax with "x:z ~~ y:w"

  # ==== Check scale scores ====

  tmp <- sapply(mm_lm_dat_out,
                FUN = colnames,
                simplify = FALSE)
  tmp <- unique(unlist(tmp))
  v_with_indicators <- setdiff(vnames, tmp)

  # ==== Add indicator syntax of for factors replaced by indicators ====

  if (length(v_with_indicators) > 0) {
    tmp1 <- number_of_indicators[[1]][v_with_indicators]
    tmp2 <- reliability[[1]][v_with_indicators]
    model <- add_indicator_syntax(model,
                                  number_of_indicators = tmp1,
                                  reliability = tmp2)
  }

  if (!is.null(attr(ptable, "model_fixed"))) {
    if (utils::packageVersion("lavaan") <= "0.6.19") {
      # lavaan 0.6.20+ should support "x:w ~~ y:z"
      # TODO:
      # - Decide which ptable to store the variables are
      #   latent variables.
      attr(model, "ptable") <- ptable
    }
  }
  tmp <- ptable
  tmp$est <- tmp$start
  # fixed.x set to FALSE such that covariances are also displayed
  fit0 <- lavaan::sem(tmp,
              do.fit = FALSE,
              fixed.x = FALSE)
  if (ngroups > 1) {
    group_labels <- names(mm_out)
    group_name <- "group"
  } else {
    group_labels <- NULL
    group_name <- NULL
  }
  if (drop_list_single_group && (ngroups == 1)) {
    mm_out <- mm_out[[1]]
    mm_lm_out <- mm_lm_out[[1]]
    mm_lm_dat_out <- mm_lm_dat_out[[1]]
    lambda_pop <- lambda_pop[[1]]
  }
  if (ngroups > 1) {
    for (i in group_labels) {
      mm_lm_dat_out[[i]]$group <- i
    }
  }
  if (merge_groups && (ngroups > 1)) {
    mm_lm_dat_out <- do.call(rbind,
                             mm_lm_dat_out)
    rownames(mm_lm_dat_out) <- NULL
  }
  out <- list(ptable = ptable,
              mm_out = mm_out,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out,
              model_original = model_original,
              model_final = model,
              fit0 = fit0,
              group_name = group_name,
              group_labels = group_labels,
              number_of_indicators = number_of_indicators,
              reliability = reliability,
              fit_external = fit_external,
              lambda = lambda_pop)
  class(out) <- c("sim_data_i", class(out))
  out
}

#' @noRd
pop_indirect <- function(x,
                         progress = TRUE,
                         pure_x = TRUE,
                         pure_y = TRUE) {
  x_i <- x[[1]]
  ptable0 <- lavaan::parameterTable(x_i$fit0)
  # Need to select a subset because
  # many_indirect_effects can be very slow
  # for a large sample
  dat_tmp <- x_i$mm_lm_dat_out
  n <- nrow(dat_tmp)
  i <- sample.int(n, 1000, replace = TRUE)
  dat_tmp <- dat_tmp[i, ]
  fit_to_all_args0 <- list(model = x_i$model_final,
                           data = dat_tmp,
                           se = "none",
                           test = "none",
                           group = x_i$group_name,
                           check.post = FALSE,
                           fixed.x = FALSE)
  fit_all <- suppressWarnings(do.call(lavaan::sem,
                      fit_to_all_args0))

  # TODO:
  # - Need a better way to find product terms
  p_terms <- lavaan::lavNames(ptable0, "ov.interaction")

  if (length(p_terms) == 0) {
    p_terms <- NULL
  }

  x_terms <- NULL
  y_terms <- NULL
  if (pure_x) {
    x_terms <- pure_x(x[[1]]$fit0)
  } else {
    x_terms <- all_x(x[[1]]$fit0)
  }
  if (pure_y) {
    y_terms <- pure_y(x[[1]]$fit0)
  } else {
    y_terms <- all_y(x[[1]]$fit0)
  }

  # Multigroup models automatically supported
  out <- list()
  if ((length(x_terms) > 0) &&
      (length(y_terms) > 0)) {
    all_paths <- manymome::all_indirect_paths(x[[1]]$fit0,
                                              exclude = p_terms,
                                              x = x_terms,
                                              y = y_terms)
    has_indirect <- isTRUE(length(all_paths) > 0)
    all_direct_paths <- get_direct(
                          x = x_terms,
                          y = y_terms,
                          ptable = ptable0
                        )
    ngroups <- lavaan::lavTech(x[[1]]$fit0, "ngroups")
    if (ngroups > 1) {

      # ==== Multigroup Models =====

      if (has_indirect) {
        # TODO:
        # - Within-group moderation not yet supported in
        #   multigroup model.
        all_ind <- tryCatch(manymome::many_indirect_effects(
                              all_paths,
                              fit = fit_all,
                              est = ptable0),
                            warning = function(e) e,
                            error = function(e) e)
        if (inherits(all_ind, "warning")) {
          if (grepl("moderator", all_ind$message)) {
            # TODO:
            # - Handle multigroup model with within-group moderation
            all_ind <- tryCatch(suppressWarnings(manymome::many_indirect_effects(
                                  all_paths,
                                  fit = fit_all,
                                  est = ptable0)),
                                error = function(e) e)
          }
        }
        if (!inherits(all_ind, "error")) {
          out <- c(out, list(all_ind))
        }
      }
    } else {

      # ==== Single-group Models =====

      all_paths_combined <- c(all_paths, all_direct_paths)
      all_w <- get_w_for_paths(all_paths_combined,
                              fit = fit_all)
      no_w_i <- sapply(all_w, function(xx) length(xx) == 0)
      has_w_i <- !no_w_i
      if (any(no_w_i)) {
        if (progress) {
          cat(paste("(Computing indirect effects for",
                    sum(no_w_i),
                    "paths ...)\n\n"))
        }
        all_ind <- manymome::many_indirect_effects(all_paths_combined[no_w_i],
                                                  fit = fit_all,
                                                  est = ptable0)
        out <- c(out, list(all_ind))
      }
      if (any(has_w_i)) {
        i0 <- which(has_w_i)
        if (progress) {
          cat(paste("(Computing conditional effects for",
                    sum(has_w_i),
                    "paths ...)\n\n"))
        }
        for (i in i0) {

          # Exclude paths moderated by mediators
          if (!all(all_w[[i]] %in% x_terms)) next

          fit_tmp <- fix_w_sd_and_mean(x_i,
                                       all_w[[i]])
          cond <- do.call(manymome::cond_indirect_effects,
                        c(all_paths_combined[[i]],
                          list(wlevels = all_w[[i]],
                                fit = fit_tmp,
                                est = ptable0)))
          out <- c(out, list(cond))
        }
      }
    }

  }
  out
}

#' @noRd
fix_w_sd_and_mean <- function(x_i,
                              w) {
  # Can only be used for single-group model
  dat_tmp <- x_i$mm_lm_dat_out
  n <- nrow(dat_tmp)
  i <- sample.int(n, 1000, replace = TRUE)
  dat_tmp <- dat_tmp[i, ]
  ptable0 <- lavaan::parameterTable(x_i$fit0)
  for (w_i in w) {
    w_tmp <- scale(dat_tmp[, w_i])[, 1]
    w_var <- ptable0[(ptable0$lhs == w_i) &
                   (ptable0$rhs == w_i) &
                   (ptable0$op == "~~"), "est"]
    i1 <- (ptable0$lhs == w_i) &
          (ptable0$op == "~1")
    if (any(i1)) {
      w_m <- ptable0[i1, "est"]
    } else {
      w_m <- 0
    }
    w_tmp <- w_tmp * sqrt(w_var) + w_m
    dat_tmp[, w_i] <- w_tmp
  }
  v_ind <- grepl(":", colnames(dat_tmp), fixed = TRUE)
  dat_tmp <- dat_tmp[, !v_ind]
  fit_to_all_args0 <- list(model = x_i$model_final,
                          data = dat_tmp,
                          se = "none",
                          test = "none",
                          group = x_i$group_name,
                          check.post = FALSE,
                          fixed.x = FALSE)
  fit_all <- suppressWarnings(do.call(lavaan::sem,
                      fit_to_all_args0))
  fit_all
}

library(testthat)
suppressMessages(library(lavaan))

test_that("fit_model: Update with new data", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

data_i1 <- sim_data(nrep = 2,
                    model = model_simple_med,
                    pop_es = model_simple_med_es,
                    n = 100,
                    iseed = 1234,
                    progress = !is_testing())
data_i2 <- sim_data(nrep = 2,
                    model = model_simple_med,
                    pop_es = model_simple_med_es,
                    n = 120,
                    iseed = 5678,
                    progress = !is_testing())

fit_i1 <- fit_model(data_i1)
fit_i2 <- fit_model(data_i2)
fit_i1_from_2 <- fit_model(data_i2,
                           fit_out = fit_i1)
expect_equal(coef(fit_i2[[2]]),
             coef(fit_i1_from_2[[2]]))

fit_i1_from_2_new_model <- fit_model(data_i2,
                                     model = c("m ~ x", "y ~ m"),
                                     fit_out = fit_i1,
                                     progress = !is_testing())
fit_i2_new_model <- fit_model(data_i2,
                              model = c("m ~ x", "y ~ m"),
                              progress = !is_testing())
expect_equal(coef(fit_i2_new_model[[2]]),
             coef(fit_i1_from_2_new_model[[2]]))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

out1 <- power4test(nrep = 2,
                   model = model_simple_med,
                   pop_es = model_simple_med_es,
                   n = 199,
                   do_the_test = FALSE,
                   iseed = 1234,
                   parallel = FALSE,
                   progress = !is_testing())

out2 <- power4test(nrep = 2,
                   model = model_simple_med,
                   pop_es = model_simple_med_es,
                   n = 199,
                   number_of_indicators = k,
                   reliability = rel,
                   do_the_test = FALSE,
                   iseed = 1234,
                   parallel = FALSE,
                   progress = !is_testing())

chk1a <- pool_sim_data(out1)
chk1b <- pool_sim_data(out1$sim_all)
expect_identical(chk1a,
                 chk1b)

chk2a <- pool_sim_data(out2, as_list = TRUE)
chk2b <- pool_sim_data(out2$sim_all, as_list = TRUE)
expect_identical(chk2a,
                 chk2b)

expect_true(is.data.frame(chk1a))
expect_true(is.list(chk2a) && !is.data.frame(chk2a))

})

library(testthat)

test_that("missing_values", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

# Default

sim_only1 <- power4test(
                  nrep = 1,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 12000,
                  process_data = list(fun = "missing_values"),
                  fit_model_args = list(estimator = "ML"),
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)
tmp <- pool_sim_data(sim_only1)
expect_setequal(colnames(tmp),
                c("x", "y", "m"))
chk <- mice::md.pattern(tmp, plot = FALSE)
chk2 <- as.numeric(rownames(chk))[1:4] / nrow(tmp)

expect_equal(chk2[1],
             .50,
             tolerance = 1e-1)

# Additional arguments 1

my_patterns <- matrix(1, nrow = 2, ncol = 3)
colnames(my_patterns) <- c("x", "y", "m")
my_patterns[1, "x"] <- 0
my_patterns[2, "y"] <- 0
my_patterns

sim_only2 <- power4test(
                  nrep = 1,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 12000,
                  process_data = list(
                    fun = "missing_values",
                    args = list(patterns = my_patterns)),
                  fit_model_args = list(estimator = "ML"),
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)
tmp <- pool_sim_data(sim_only2)
expect_setequal(colnames(tmp),
                c("x", "y", "m"))
chk <- mice::md.pattern(tmp, plot = FALSE)
chk2 <- as.numeric(rownames(chk))[1:4] / nrow(tmp)

expect_equal(chk2[1],
             .50,
             tolerance = 1e-1)
expect_all_equal(as.numeric(chk[1:3, "m"]),
                 1)


# Additional arguments 1

my_patterns <- matrix(1, nrow = 2, ncol = 3)
colnames(my_patterns) <- c("m", "x", "y")
my_patterns[1, "x"] <- 0
my_patterns[2, "y"] <- 0
my_patterns

sim_only3 <- power4test(
                  nrep = 1,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 10000,
                  process_data = list(
                    fun = "missing_values",
                    args = list(patterns = my_patterns,
                                mech = "MAR",
                                prop = .25)),
                  fit_model_args = list(estimator = "ML"),
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)
tmp <- pool_sim_data(sim_only3)
expect_setequal(colnames(tmp),
                c("x", "y", "m"))
chk <- mice::md.pattern(tmp, plot = FALSE)
chk2 <- as.numeric(rownames(chk))[1:4] / nrow(tmp)

expect_equal(chk2[1],
             .75,
             tolerance = 1e-1)
expect_all_equal(as.numeric(chk[1:3, "m"]),
                 1)

# With indicators

k <- c(y = 3,
       m = 4,
       x = 5)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only1 <- power4test(
                  nrep = 1,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  number_of_indicators = k,
                  reliability = rel,
                  n = 12000,
                  process_data = list(fun = "missing_values"),
                  fit_model_args = list(estimator = "ML"),
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)
tmp <- pool_sim_data(sim_only1)
chk <- mice::md.pattern(tmp, plot = FALSE)
chk2 <- as.numeric(rownames(chk))[1:4] / nrow(tmp)

expect_equal(chk2[1],
             .50,
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("fit_model: Error and warning catching", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 8,
       m = 8,
       x = 8)
rel <- c(y = .30,
         m = .30,
         x = .30)

data_i <- sim_data_i(model = model_simple_med,
                     pop_es = model_simple_med_es,
                     n = 50,
                     number_of_indicators = k,
                     reliability = rel,
                     seed = 1234)

expect_no_warning(fit_i <- fit_model_i(data_i))
expect_warning(lavInspect(fit_i, "post.check"))

expect_no_warning(data_all <- sim_data(nrep = 2,
                                       model = model_simple_med,
                                       pop_es = model_simple_med_es,
                                       n = 50,
                                       number_of_indicators = k,
                                       reliability = rel,
                                       iseed = 2345,
                                       parallel = FALSE,
                                       progress = !is_testing()))
expect_no_warning(fit_all <- fit_model(data_all,
                                       control = list(iter.max = 50)))
suppressWarnings(tmp <- sapply(fit_all,
                               lavInspect,
                               what = "post.check"))
expect_true(any(!tmp))
suppressWarnings(tmp <- sapply(fit_all,
                               lavInspect,
                               what = "converged"))
expect_true(any(!tmp))

expect_no_warning(data_all <- sim_data(nrep = 3,
                                       model = model_simple_med,
                                       pop_es = model_simple_med_es,
                                       n = 200,
                                       number_of_indicators = k,
                                       reliability = rel,
                                       iseed = 2345,
                                       parallel = FALSE,
                                       progress = !is_testing()))
data_all[[2]]$model_final <- data_all[[2]]$model_original
expect_no_warning(fit_all <- fit_model(data_all,
                                       control = list(iter.max = 50)))
expect_true(inherits(fit_all[[2]], "error"))
})

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test: Print Indirect", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)


# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 100,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_no_error(capture.output(print(power_all_sim_only)))

tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]])["x -> m -> y"],
             .50 * .30,
             ignore_attr = TRUE)

model_simple_med <-
"
m ~ x
y ~ m + x
"

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 100,
                                 fit_model_args = list(fit_function = "lm"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_no_error(capture.output(print(power_all_sim_only)))

tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]])["x -> m -> y"],
             .50 * .30,
             ignore_attr = TRUE)

# Case 2

model_simple_med <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + m1 + x
"

model_simple_med_es <-
"
.ind.(x->m1->m2->y): si
y ~ m1: s
y ~ x: m
"

k <- c(y = 3,
       m1 = 3,
       m2 = 4,
       x = 3)
rel <- c(y = .80,
         m1 = .70,
         m2 = .80,
         x = .70)

# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 500,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "x -> m1 -> m2")
tmp <- pop_indirect(power_all_sim_only$sim_all,
                    pure_x = TRUE,
                    pure_y = TRUE,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]])[1],
             .141,
             ignore_attr = TRUE)


power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 500,
                                 fit_model_args = list(fit_function = "lm"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "x -> m1 -> m2")
tmp <- pop_indirect(power_all_sim_only$sim_all,
                    pure_x = TRUE,
                    pure_y = TRUE,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]])[1],
             .141,
             ignore_attr = TRUE)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models", {

model2 <-
"
m ~ x + w + x:w
y ~  m + x
"

model2_es <- list("m ~ x" = "-m",
                  "y ~ m" = "l",
                  "y ~ x" = c("m", "l", "n"))

dats <- sim_data(nrep = 9,
                 model = model2,
                 pop_es = model2_es,
                 n = c(100, 200, 300),
                 iseed = 1234,
                 progress = !is_testing())
tmp <- capture.output(print(dats))
expect_true(all(!grepl("function", tmp)))
})

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test: Print Indirect", {

mod <-
"
m ~ x + w + x:w
y ~ m + x
"

mod_es <-
"
.ind.(x->m->y): mi
m ~ x:w: l
y ~ x: m
"

# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = mod,
                                 pop_es = mod_es,
                                 n = 100,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "w -> m -> y")
suppressWarnings(tmp <- pop_indirect(power_all_sim_only$sim_all,
                                     progress = !is_testing()))
# expect_equal(coef(tmp[[1]])["x -> m -> y"],
#              .361,
#              ignore_attr = TRUE)
tmp2 <- sapply(tmp, function(xx) class(xx)[1])
expect_true("cond_indirect_effects" %in% tmp2)

power_all_sim_only <- power4test(nrep = 10,
                                 model = mod,
                                 pop_es = mod_es,
                                 n = 100,
                                 fit_model_args = list(fit_function = "lm"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "w -> m -> y")
suppressWarnings(tmp <- pop_indirect(power_all_sim_only$sim_all,
                                     progress = !is_testing()))
# expect_equal(coef(tmp[[1]])["x -> m -> y"],
#              .361,
#              ignore_attr = TRUE)
tmp2 <- sapply(tmp, function(xx) class(xx)[1])
expect_true("cond_indirect_effects" %in% tmp2)

# Case 2

mod <-
"
y ~ x + w + x:w + c1
"

mod_es <-
"
y ~ x:w: s
y ~ x: m
"

# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = mod,
                                 pop_es = mod_es,
                                 n = 500,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_true(any(grepl("Conditional",
                 capture.output(print(power_all_sim_only)))))
tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_true(length(tmp) > 0)



power_all_sim_only <- power4test(nrep = 10,
                                 model = mod,
                                 pop_es = mod_es,
                                 n = 500,
                                 fit_model_args = list(fit_function = "lm"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_true(any(grepl("Conditional",
                 capture.output(print(power_all_sim_only)))))
tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_true(length(tmp) > 0)

# Multigroup

# Case 1

mod <-
"
m ~ x
y ~ m + x + w + x:w
"

mod_es<-
"
m ~ x: [m, l]
y ~ m: [s, m]
y ~ x: [-m, nil]
"

# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = mod,
                                 pop_es = mod_es,
                                 n = 100,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

if (!is_testing()) {
print(power_all_sim_only)
}

expect_output(print(power_all_sim_only),
              "Group1.x -> m -> y")
tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]]),
             c(.10 * .30, .30 * .50),
             ignore_attr = TRUE)

})

library(testthat)
suppressMessages(library(lavaan))
library(manymome)

skip_if_not_installed("lmhelprs")

test_that("fit_model: User fit function", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

data_i <- sim_data_i(model = model_simple_med,
                     pop_es = model_simple_med_es,
                     n = 500,
                     seed = 1234)

fit_i <- fit_model_i(data_i,
                     model = c("m ~ x", "y ~ m"),
                     fit_function = "lm")

expect_s3_class(fit_i,
                "lm_list_lmhelprs")
lm_m <- lm(m ~ x,
           data_i$mm_lm_dat_out)
lm_y <- lm(y ~ m,
           data_i$mm_lm_dat_out)
expect_equal(coef(fit_i[[2]]),
             coef(lm_y))

ind <- indirect_effect(fit = fit_i,
                       x = "x",
                       m = "m",
                       y = "y")
expect_equal(unname(coef(ind)),
             unname(coef(lm_m)["x"] * coef(lm_y)["m"]))

})

library(testthat)
suppressMessages(library(lavaan))

# All-In-One

test_that("power4test: mg", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

# Generate the data

out1 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = c(100, 50),
          test_fun = test_parameters,
          test_args = list(par = c("m~x", "m~x.g2")),
          progress = !is_testing(),
          iseed = 1234
        )

out2 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = 100,
          test_fun = test_parameters,
          test_args = list(par = c("m~x", "m~x.g2")),
          progress = !is_testing(),
          iseed = 1234,
          n_ratio = c(1, .5)
        )

dat1 <- pool_sim_data(out1)
dat2 <- pool_sim_data(out2)

expect_equal(table(dat1$group),
             table(dat2$group))

out3 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = 100,
          test_fun = test_parameters,
          test_args = list(par = c("m~x", "m~x.g2")),
          progress = !is_testing(),
          iseed = 1234,
          n_ratio = 1
        )

dat3 <- pool_sim_data(out3)
expect_true(length(unique(table(dat3$group))) == 1)

expect_error(power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = 100,
          test_fun = test_parameters,
          test_args = list(par = c("m~x", "m~x.g2")),
          progress = !is_testing(),
          iseed = 1234,
          n_ratio = c(1, 2, 3)
        ),
        "n_ratio")

out4 <- power4test(
          nrep = 2,
          model = model_simple_med,
          pop_es = model_simple_med_es,
          n = 100,
          test_fun = test_parameters,
          test_args = list(par = c("m~x", "m~x.g2")),
          progress = !is_testing(),
          iseed = 1234,
          n_ratio = c(.5, 1)
        )

dat4 <- pool_sim_data(out4)

expect_true(all.equal(as.vector(table(dat4$group)),
                      c(100, 200)))

outby <- power4test_by_n(
            out2,
            n = c(150, 200),
            by_seed = 4567
          )

expect_equal(as.vector(table(pool_sim_data(outby[[1]])$group)),
             c(300, 150))
expect_equal(as.vector(table(pool_sim_data(outby[[2]])$group)),
             c(400, 200))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("fit_model: User models", {

# Simple Mediation Model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

data_i <- sim_data_i(model = model_simple_med,
                     pop_es = model_simple_med_es,
                     n = 500,
                     seed = 1234)

fit_i <- fit_model_i(data_i,
                     model = c("m ~ x", "y ~ m"))
expect_equal(fitMeasures(fit_i, "df"),
             1,
             ignore_attr = TRUE)
})

skip_on_cran()

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out and do_test", {

mod <-
"
y ~ x + w1 + x:w1
"


es <-
"
y ~ x:w1: m
y ~ x: s
y ~ w1: m
"

data_all <- sim_data(nrep = 2,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)
fit_all <- fit_model(data_all)
sim_all <- sim_out(data_all = data_all,
                   fit = fit_all)
tmp <- capture.output(print(sim_all))
expect_true(any(grepl("[w1]", tmp, fixed = TRUE)))
expect_true(any(grepl("[x]", tmp, fixed = TRUE)))

})

library(testthat)

test_that("scale-scores: indirect effect", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only1 <- power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "scale_scores"),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)

test_ind1 <- power4test(
                  object = sim_only1,
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  progress = !is_testing())
test_ind1

tmp <- pool_sim_data(sim_only1)
expect_setequal(colnames(tmp),
                c("x", "y", "m"))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out: descriptive for multigroup model", {

mod <-
"
m ~ x
y ~ m + x
"

es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

data_all <- sim_data(nrep = 2,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)

expect_output(print(data_all),
              "Descriptive statistics by group",
              fixed = TRUE)


})

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_data: one lv", {

mod <-
"
m ~ x
y ~ m + x
"

pop_es <-
"
m ~ x: s
y ~ m: s
y ~ x: s
"

data_all <- sim_data(nrep = 1,
                     model = mod,
                     pop_es = pop_es,
                     n = 100,
                     number_of_indicators = c(m = 3),
                     reliability = c(m = .70),
                     progress = !is_testing(),
                     iseed = 1234)

chk <- c("y", "x", "m1", "m2", "m3")
expect_setequal(chk,
                colnames(data_all[[1]]$mm_lm_dat_out))

})

library(testthat)

test_that("ordinal_variables", {

expect_true(is.list(cut_patterns()))
expect_true(length(cut_patterns("s3")) == 2)
expect_setequal(cut_patterns("ma4"),
                -cut_patterns("-ma4"))

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only1 <- power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "ordinal_variables",
                                      args = list(cut_patterns = c(x = "s3"),
                                                  cuts = list(m = c(-2, 0, 2)))),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)

expect_output(print(sim_only1, data_long = TRUE),
              " Response Proportions ")

tmp <- pool_sim_data(sim_only1)

expect_equal(length(table(tmp$m1)), 4)
expect_equal(length(table(tmp$x1)), 3)


sim_only1 <- power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "ordinal_variables",
                                      args = list(cut_patterns = c(x = "-ma3"))),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)

tmp <- pool_sim_data(sim_only1)

expect_equal(length(table(tmp$x1)), 3)

sim_only1 <- power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "ordinal_variables",
                                      args = list(cuts = list(x = c(-2, 1, 2)))),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)

tmp <- pool_sim_data(sim_only1)

expect_equal(length(table(tmp$x1)), 4)

sim_only1 <- power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "ordinal_variables"),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)

tmp <- pool_sim_data(sim_only1)

expect_true(length(table(tmp$x1)) > 10)


expect_error(power4test(
                  nrep = 2,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  number_of_indicators = k,
                  reliability = rel,
                  process_data = list(fun = "ordinal_variables",
                                      args = list(cut_patterns = c(x = "test"))),
                  fit_model_args = list(estimator = "ML"),
                  R = 100,
                  do_the_test = FALSE,
                  progress = !is_testing(),
                  parallel = FALSE,
                  iseed = 1234)
)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_data: lambda vary", {

mod <-
"
m ~ x
y ~ m + x
"

pop_es <-
"
m ~ x: s
y ~ m: s
y ~ x: s
"

data_all <- sim_data(nrep = 1,
                     model = mod,
                     pop_es = pop_es,
                     n = 10000,
                     number_of_indicators = c(m = 3),
                     reliability = c(m = .70),
                     loading_difference = c(m = .10),
                     reference = c(m = "strongest"),
                     progress = !is_testing(),
                     parallel = FALSE,
                     iseed = 1234)

chk <- c("y", "x", "m1", "m2", "m3")
expect_setequal(chk,
                colnames(data_all[[1]]$mm_lm_dat_out))

tmp <- pool_sim_data(data_all)
mod_m <- "m =~ m1 + m2 + m3"
fit <- cfa(mod_m, tmp)
std <- standardizedSolution(fit)[1:3, "est.std"]
expect_true(which.max(std) == 1)

data_all <- sim_data(nrep = 1,
                     model = mod,
                     pop_es = pop_es,
                     n = 10000,
                     number_of_indicators = c(m = 3, x = 5),
                     reliability = c(x = .8, m = .70),
                     loading_difference = c(x = .05, m = .10),
                     reference = c(m = "medium", x = "weakest"),
                     progress = !is_testing(),
                     iseed = 1234,
                     parallel = FALSE)

chk <- c("y", "x1", "x2", "x4", "x3", "x5", "m1", "m2", "m3")
expect_setequal(chk,
                colnames(data_all[[1]]$mm_lm_dat_out))

tmp <- pool_sim_data(data_all)

mod_m <- "m =~ m1 + m2 + m3"
fit <- cfa(mod_m, tmp)
std <- standardizedSolution(fit)[1:3, "est.std"]
expect_false(which.max(std) == 1)
expect_false(which.min(std) == 1)

mod_x <- "x =~ x1 + x2 + x3 + x4 + x5"
fit <- cfa(mod_x, tmp)
std <- standardizedSolution(fit)[1:5, "est.std"]
expect_true(which.min(std) == 1)

expect_equal(data_all[[1]]$lambda$x,
             lambda_from_reliability(p = 5, omega = .80, d = .05, ref = "weakest"))
expect_equal(data_all[[1]]$lambda$m,
             lambda_from_reliability(p = 3, omega = .70, d = .10, ref = "medium"))

})

test_that("sim_data: lambda vary: multigroup", {

mod <-
"
m ~ x
y ~ m + x
"

pop_es <-
"
m ~ x:
 - s
 - m
y ~ m: s
y ~ x: s
"

data_all <- sim_data(nrep = 1,
                     model = mod,
                     pop_es = pop_es,
                     n = 10000,
                     number_of_indicators = c(m = 3),
                     reliability = c(m = .70),
                     loading_difference = c(m = .10),
                     reference = c(m = "strongest"),
                     progress = !is_testing(),
                     parallel = FALSE,
                     iseed = 1234)

expect_equal(data_all[[1]]$lambda$Group1$m,
             lambda_from_reliability(p = 3, omega = .70, d = .10, ref = "strongest"))

})

skip_on_cran()

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out and do_test", {

mod <-
"
m1 ~ x + w1 + x:w1
m2 ~ x
m3 ~ x + w4 + x:w4
y ~ m1 + x + w2 + w3 + m1:w2 + m2 + m1:w3 + m3 + w5 + x:w5
"


es <-
"
m3 ~ x:w4: s
m1 ~ x: m
m1 ~ w1: s
m1 ~ x:w1: l
y ~ m1:w2: m
y ~ m1:w3: s
y ~ m1: l
y ~ x: s
y ~ x:w5: m
"

data_all <- sim_data(nrep = 2,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)
fit_all <- fit_model(data_all)
sim_all <- sim_out(data_all = data_all,
                   fit = fit_all)
tmp <- capture.output(print(sim_all))
expect_true(any(grepl("x, w2, w3", tmp, fixed = TRUE)))
expect_true(any(grepl("w1, w2, w3", tmp, fixed = TRUE)))
expect_true(any(grepl("[w4]", tmp, fixed = TRUE)))
expect_true(any(grepl("[w5]", tmp, fixed = TRUE)))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out and do_test", {

mod <-
"m ~ x
 y ~ m + x"
es <-
c("y ~ m" = "m",
  "m ~ x" = "m",
  "y ~ x" = "n")
data_all <- sim_data(nrep = 3,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)

expect_output(print(data_all), "Variances and error variances")
expect_output(print(data_all, variances = TRUE), "Variances:")
expect_output(print(data_all, est_type = "unstandardized"), "Unstandardized")
expect_output(print(data_all, est_type = c("standardized", "unstandardized")), "Unstandardized")


})

# A long test

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test: Print Indirect", {

# Multigroup

# Case 1

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <-
"
m ~ x: [m, l]
y ~ m: [s, m]
y ~ x: [-m, nil]
"

k <- c(y = 3,
       m = 3,
       x = 3)

rel <- c(y = .70,
         m = .70,
         x = .70)


# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 100,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "Group1.x -> m -> y")

tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]]),
             c(.10 * .30, .50 * .30),
             ignore_attr = TRUE)

# Case 2

model_simple_med <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + m1 + x
"

model_simple_med_es <-
"
.ind.(x->m1->m2->y): [mi, li]
y ~ m1: s
y ~ x: [s, nil]
"

k <- c(y = 3,
       m1 = 3,
       m2 = 4,
       x = 3)
rel <- c(y = .80,
         m1 = .70,
         m2 = .80,
         x = .70)

# Generate the data

power_all_sim_only <- power4test(nrep = 2,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 500,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only),
              "Group1.x -> m1 -> y")

tmp <- pop_indirect(power_all_sim_only$sim_all,
                    progress = !is_testing())
expect_equal(coef(tmp[[1]])["Group2.x -> m1 -> m2 -> y"],
             .51,
             ignore_attr = TRUE)
expect_equal(coef(tmp[[1]])["Group1.x -> m1 -> m2 -> y"],
             .361,
             ignore_attr = TRUE)

})

#' @noRd
as_test_data_frame_i <- function(
  x,
  test_name
) {
  # Convert a test results vector to a one-row test results data frame
  if ("test_results" %in% names(x)) {
    # x should be the test results
    x <- x$test_results
  }
  if (!is.null(dim(x))) {
    # It is not a vector. Assume it is a data frame and return it as-is.
    # Skip checking x, for efficiency.
    return(x)
  }
  out <- data.frame(
            test_label = test_name,
            rbind(x)
          )
  attr(out, "test_label") <- "test_label"
  out
}

#' @noRd
as_test_data_frame <- function(
  test_output
) {
  # Convert a list of test vectors to a list of one-row test data frame
  test_name <- attr(
                  test_output,
                  "test_name"
                )
  for (i in seq_along(test_output)) {
    test_output[[i]]$test_results <-
      as_test_data_frame_i(test_output[[i]]$test_results,
                           test_name = test_name)
  }
  test_output
}

#' @noRd
as_test_data_frame_all_tests <- function(
  object
) {
  if (inherits(object, "power4test")) {
    object <- object$test_all
  }
  # object: The test_all element of a `power4test` object
  for (i in seq_along(object)) {
    object[[i]] <- as_test_data_frame(object[[i]])
  }
  object
}

#' @noRd

rbind_diff_cols <- function(x) {
  # Bind rows with different colnames
  test_names <- names(x)
  all_cnames <- lapply(x,
                       \(x) colnames(x))
  all_cnames <- unique(unlist(all_cnames,
                       use.names = FALSE))
  out <- lapply(x,
            \(xx) {
              tmp <- setdiff(all_cnames, colnames(xx))
              if (length(tmp) == 0) {
                return(xx)
              }
              xx[, tmp] <- NA
              xx
            })
  out <- mapply(function(xx, yy) {
                  out <- cbind(test = xx,
                               yy)
                  rownames(out) <- NULL
                  out
                },
                xx = test_names,
                yy = out,
                SIMPLIFY = FALSE)
  do.call(rbind, unname(out))
}

#' @noRd
collapse_all_tests <- function(
  object,
  keep = c("est", "cilo", "cihi", "sig", "pvalue"),
  alpha = .05,
  p_adjust_method = "none"
) {
  # p_adjust_method: To be passed to p.adjust()
  # alpha: Used if p_adjust_method != "none",
  #        Update `sig`.
  # Get test_all
  # Collapse all tests into one test
  # NOTE:
  # test_labels no longer relevant
  # because the column names are
  # not guaranteed to be the same
  # across tests.
  # This functions should be used only
  # with collapse across tests.

  if (inherits(object, "power4test")) {
    object <- object$test_all
  }
  test_all <- as_test_data_frame_all_tests(object)
  nrep <- length(test_all[[1]])

  out <- lapply(seq_len(nrep),
            function(i) {
              test_results_all_i <- lapply(test_all,
                                  \(x) x[[i]]$test_results)
              out1 <- rbind_diff_cols(test_results_all_i)
              tmp <- rep("--", nrow(out1))
              i <- out1$test_label == out1$test
              out1$test_label[i] <- tmp[i]
              # Some old tests may not have pvalues
              keep <- intersect(keep, colnames(out1))
              out1 <- out1[, c("test", "test_label", keep), drop = FALSE]
              if ((p_adjust_method != "none") &&
                  ("pvalue" %in% colnames(out1))) {
                if (!all(is.na(out1$pvalue))) {
                  out1$pvalue_org <- out1$pvalue
                  out1$pvalue <- stats::p.adjust(
                                        out1$pvalue_org,
                                        method = p_adjust_method
                                      )
                  out1$sig <- as.numeric(out1$pvalue < alpha)
                }
              }
              out1 <- list(test_results = out1)
              attr(out1$test_results, "test_label") <- "test_label"
              out1
            })
  attr(out, "test_name") <- "All tests merged into one"
  attr(out, "tests") <- names(object)
  out
}

library(testthat)
suppressMessages(library(lavaan))

test_that("summarize tests: merge tests", {

model <-
"
m ~ x + z + x:z
y ~ m + x
"

model_es <-
"
m ~ x:z: s
m ~ x: m
y ~ m: m
y ~ x: s
"

out <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es,
            n = 100,
            test_fun = test_parameters,
            test_args = list(pars = "m~x"),
            ci_type = "mc",
            R = 100,
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out)

out2 <- power4test(
            out,
            test_fun = test_parameters,
            test_args = list(pars = c("m~z", "y~m")),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out2)

out3 <- power4test(
            out2,
            test_fun = test_index_of_mome,
            test_args = list(x = "x",
                             y = "y",
                             m = "m",
                             w = "z"),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

rejection_rates(out3)

out3$test_all[[3]][[1]]
out3$test_all[[2]][[1]]

attributes(out3$test_all[[3]])

attributes(out3$test_all[[2]][[1]]$test_results)

as_test_data_frame_i(out3$test_all[[3]][[1]],
                     test_name = attr(out3$test_all[[3]], "test_name"))
as_test_data_frame_i(out3$test_all[[3]][[1]]$test_results,
                     test_name = attr(out3$test_all[[3]], "test_name"))

expect_identical(as_test_data_frame(out3$test_all[[1]]),
                 out3$test_all[[1]])

chk <- as_test_data_frame(out3$test_all[[3]])
expect_false(identical(
                  chk,
                  out3$test_all[[3]])
              )
expect_s3_class(chk[[1]]$test_results, "data.frame")

tmp <- as_test_data_frame_all_tests(out3)

if (!is_testing()) {
print(out2$test_all, test_long = TRUE)
print(out3$test_all, test_long = TRUE)
print(tmp, test_long = TRUE)
}

tmp2 <- collapse_all_tests(out3)

summarize_one_test_data_frame(tmp2,
                              collapse = "all_sig")
summarize_one_test_data_frame(out3$test_all[[2]])
summarize_one_test_data_frame(out3$test_all[[2]],
                              collapse = "all_sig")

tmp <- summarize_tests(out3,
                       merge_all_tests = TRUE)
expect_output(print(tmp),
              "merged")
expect_true(length(tmp) == 1)

tmp <- summarize_tests(out3,
                       collapse = "all_sig",
                       merge_all_tests = TRUE)
expect_true(nrow(tmp[[1]]$mean) == 1)

expect_output(print(tmp),
              "merged")

tmp <- rejection_rates(out3,
                merge_all_tests = TRUE)

expect_output(print(tmp),
              "merged")

})

#' @title Set Population Values
#'
#' @description Return a parameter table
#' with the population values of selected
#' parameters.
#'
#' @details
#'
#' These are the input and output values
#' for `set_pop()`.
#'
#' Input:
#'
#' - par_es: The effect sizes for selected coefficients
#'
#' - es1: Effect sizes for normal coefficients
#'
#' - es2: Effect sizes for moderation (product terms)
#'
#' Output:
#'
#' - A lhs-op-rhs data frame with population values (pop)
#'
#' @return
#' A data frame with these columns: `lhs`,
#' `op`, `rhs`, `pop`, and `es`. The first
#' three columns correspond to those in
#' a `lavaan` parameter table. The column
#' `pop` stores the population values.
#' The column `es` stores the original
#' labels, for reference.
#'
#' @param par_es A named character
#' vector. The names are `lavaan` for
#' the seleted parameters. For example,
#' `m ~ x` denotes the path from `x` to
#' `m`. Can specify more than one
#' parameters. For example, `y ~ m + x`
#' denotes the two paths from `m` and
#' `x` to `y`. The value is the label
#' for the effect size: `s` for small,
#' `m` for medium, and `l` for large.
#' There are two possible values, one
#' set, `es1`, for correlations and
#' regression coefficients, the other
#' set, `es2`, for standardized
#' moderation effect, the coefficients
#' of a product term.
#'
#' @param es1 A named vector to set the
#' values for each label of the effect
#' size of correlations and regression
#' paths.
#' Default is `c("n" = .00, "nil" = .00, "s" = .10, "m" = .30, "l" = .50)`.
#'
#' @param es2 A named vector to set the
#' values for each label of the effect
#' size of product term.
#' Default is `c("n" = .00, "nil" = .00, "s" = .05, "m" = .10, "l" = .15)`.
#'
#' @examples
#' \donttest{
#' }
#'
#' @noRd
set_pop <- function(par_es,
                    es1 = c("n" = .00,
                            "nil" = .00,
                            "s" = .10,
                            "m" = .30,
                            "l" = .50,
                            "si" = .141,
                            "mi" = .361,
                            "li" = .510,
                            "sm" = .20,
                            "ml" = .40),
                    es2 = c("n" = .00,
                            "nil" = .00,
                            "s" = .05,
                            "m" = .10,
                            "l" = .15,
                            "sm" = .075,
                            "ml" = .125),
                    es_ind = c("si",
                               "mi",
                               "li")) {
  es10 <- es_long(es1)
  es20 <- es_long(es2)
  es10_env <- list2env(as.list(es10))
  es20_env <- list2env(as.list(es20))
  to_set <- lavaan::lavParseModelString(names(par_es),
                                        as.data.frame. = TRUE)
  to_set$pop <- NA
  for (x in seq_along(par_es)) {
    is_inter <- isTRUE(grepl(":", to_set$rhs[x], fixed = TRUE))
    if (is_inter) {
      y <- try(eval(parse(text = par_es[x]),
                     envir = es20_env),
               silent = TRUE)
      if (inherits(y, "try-error") ||
          !is.numeric(y)) {
        to_set[x, "pop"] <- NA
      } else {
        to_set[x, "pop"] <- y
      }
    } else {
      y <- try(eval(parse(text = par_es[x]),
                    envir = es10_env),
              silent = TRUE)
      if (inherits(y, "try-error") ||
          !is.numeric(y)) {
        # Not needed.
        # Kept for backward compatibility
        # Check if it is a component
        x_i <- strsplit(par_es[x],
                        "_",
                        fixed = TRUE)[[1]]
        x_i_num <- suppressWarnings(as.numeric(x_i))
        if (is.numeric(x_i_num) &&
            all(!is.na(x_i_num))) {
          es_num <- x_i_num[1] ^ (1 / x_i_num[2])
          to_set[x, "pop"] <- es_num
        }
      } else {
        to_set[x, "pop"] <- y
      }
    }
  }
  to_set$es <- par_es
  to_set[, c("lhs", "op", "rhs", "pop", "es")]
}

#' @noRd
# Input:
# - A names vector of effect sizes
# Output:
# - A expanded named vector with names for negative effects.
#   E.g., "m" to "-m".

es_long <- function(es) {
  es10 <- stats::setNames(es,
                          gsub(".",
                               "",
                               names(es),
                               fixed = TRUE))
  es10n <- stats::setNames(-1 * es,
                           paste0("-",
                                  names(es10)))
  est1x <- c(es10, es10n)
  est1x
}


#' @noRd
# Input:
# - The effect sizes for selected coefficients
# Output:
# - The expanded named vectors

fix_par_es <- function(par_es,
                       model) {
  par_es_org <- par_es
  i <- match(c(".beta.", ".cov."), names(par_es))
  i_ind <- which(grepl("^.ind.", names(par_es)))
  i <- c(i, i_ind)
  par_es_def <- par_es[i]
  par_es_def <- par_es_def[!is.na(par_es_def)]
  all_beta_es <- character(0)
  all_cov_es <- character(0)
  all_ind_es <- character(0)
  if (!all(is.na(i))) {
    par_es <- par_es[-i[!is.na(i)]]
    ptable <- lavaan::parTable(lavaan::sem(model = model,
                                           do.fit = FALSE))
    if (".beta." %in% names(par_es_def)) {
      all_beta <- ptable[ptable$op == "~", c("lhs", "op", "rhs")]
      all_beta <- apply(all_beta, 1, paste, collapse = " ")
      all_beta_es <- rep(par_es_def[".beta."], length(all_beta))
      names(all_beta_es) <- all_beta
    }
    if (".cov." %in% names(par_es_def)) {
      all_cov <- ptable[ptable$op == "~~", ]
      all_cov <- all_cov[all_cov$lhs != all_cov$rhs, ]
      all_cov <- all_cov[all_cov$exo == 1, ]
      tmp <- lavaan::lavNames(ptable, "ov.interaction")
      if (length(tmp) > 0) {
        all_cov <- all_cov[!((all_cov$lhs == tmp) |
                            (all_cov$rhs == tmp)), ]
      }
      all_cov <- all_cov[, c("lhs", "op", "rhs")]
      all_cov <- apply(all_cov, 1, paste, collapse = " ")
      all_cov_es <- rep(par_es_def[".cov."], length(all_cov))
      names(all_cov_es) <- all_cov
    }
    if (length(i_ind) > 0) {
      # Expand to component paths
      i2 <- which(grepl("^.ind.", names(par_es_def)))
      par_es_ind <- par_es_def[i2]
      any_negative <- grepl("^-", trimws(par_es_ind))
      if (any(any_negative)) {
        tmp1 <- par_es_ind[any_negative][1]
        tmp2 <- paste0(names(tmp1), " set to ", tmp1, ".")
        stop("Cannot set the value of .ind.() to negative value. E.g.,",
             tmp2)
      }
      ind_comp <- sapply(names(par_es_ind),
                         expand_to_components,
                         simplify = FALSE,
                         USE.NAMES = TRUE)
      if (any(duplicated(unname(unlist(ind_comp))))) {
        stop("The paths in '.ind.' cannot overlap.")
      }
      tmpfct <- function(x, y) {
        k <- length(y)
        out <- rep(x, k)
        # k indicates the number of component paths
        out <- paste0("(", out, ")^(1 / ", k, ")")
        names(out) <- y
        out
      }
      all_ind_es <- mapply(tmpfct,
                           x = par_es_ind,
                           y = ind_comp,
                           USE.NAMES = FALSE,
                           SIMPLIFY = FALSE)
      all_ind_es <- unlist(all_ind_es)
    }
  }
  out <- character(0)
  for (i in seq_along(par_es)) {
    x_name <- names(par_es[i])
    tmp1 <- lavaan::lavParseModelString(x_name,
                                        as.data.frame. = TRUE)
    tmp2 <- paste(tmp1$lhs,
                  tmp1$op,
                  tmp1$rhs)
    tmp3 <- rep(par_es[i], length(tmp2))
    names(tmp3) <- tmp2
    out <- c(out, tmp3)
  }
  # Add all_beta_es
  all_def <- c(all_beta_es, all_cov_es)
  tmp <- setdiff(names(all_def), names(out))
  out <- c(out, all_def[tmp])

  # .ind. override other specification
  if (length(all_ind_es) > 0) {
    tmp <- setdiff(names(out), names(all_ind_es))
    out <- c(all_ind_es, out[tmp])
  }

  out
}

#' @noRd
# Copied from another project by the author
# lambda_from_reliability <- function(p = 3,
#                                     omega = .70) {
#   lambda <- sqrt(omega / (p - omega * (p - 1)))
#   lambda
# }

#' @noRd
# Copied from another project by the author
lambda_from_reliability <- function(p = 3,
                                    omega = .70,
                                    d = 0,
                                    ref = NULL) {
  if (is.null(ref)) {
    ref <- "medium"
  } else {
    ref <- match.arg(ref, c("medium", "weakest", "strongest"))
  }
  # Find the sequence of lambdas that are:
  # - of equal successive distance, and
  # - have the target omega
  # Work for generating a sequence of identical loadings
  d <- abs(d)
  dp <- sum((seq_len(p) - (p + 1) / 2)^2)
  lambda0 <- (dp * d^2 - p) * omega / ((omega * p - omega - p) * p)
  lambda0 <- sqrt(lambda0)
  out <- expand_lambda(lambda0 = lambda0,
                       p = p,
                       d = d)
  i <- floor(mean(seq_len(p)))
  outmed <- c(out[i], out[-i])
  out <- switch(ref,
                medium = outmed,
                strongest = rev(out),
                weakest = out)
  out
}

#' @noRd
# Copied from another project by the author
expand_lambda <- function(lambda0,
                          p,
                          d) {
  # Create a sequence of lambdas
  i <- seq_len(p)
  out <- sapply(i,
                function(x) {
                  lambda0 + (x - (p + 1) / 2) * d
                })
  out
}

#' @noRd
# Create a list of par_es for groups
split_par_es <- function(object) {
  ngroups <- max(sapply(object,
                        length,
                        USE.NAMES = FALSE))
  for (xx in names(object)) {
    i <- length(object[[xx]])
    if (i == 1) {
      object[[xx]] <- rep(object[[xx]], ngroups)
    } else if (i != ngroups) {
      stop("Each element of pop_es must either one element of equal number of elements.")
    }
  }
  i <- seq_len(ngroups)
  out <- lapply(i, function(x) {
              sapply(object,
                     function(y) y[x],
                     simplify = TRUE)
            })
  return(out)
}

#' @noRd
# Input:
# - E.g.,
#   - ".ind.(x1-> m1->m2->y)"
# Output:
# - c("m1~x1", "m2~m1", "y~m2")
expand_to_components <- function(x,
                                 pattern = c(l2r = "->",
                                             r2l = "~")) {
  x0 <- trimws(gsub("^.ind.", "", x))
  x1 <- sub("\\(", "", x0)
  x1 <- sub("\\)$", "", x1)
  x1 <- trimws(x1)
  style <- ""
  tmp <- sapply(pattern,
                \(x) grepl(x, x1, fixed = TRUE))
  if (sum(tmp) != 1) {
    tmp2 <- paste0(pattern, collapse = ", ")
    stop("Only one style (out of these: ",
         tmp2,
         ") can be used.")
  }
  pattern_used <- pattern[tmp]
  style <- names(pattern)[tmp]

  x2 <- strsplit(x1,
                 pattern_used,
                 fixed = TRUE)[[1]]
  x2 <- sapply(x2, trimws)
  x2a <- x2[-length(x2)]
  x2b <- x2[-1]
  out <- switch(style,
                l2r = paste0(x2b,
                             " ~ ",
                             x2a),
                r2l = paste0(x2a,
                             " ~ ",
                             x2b))
  out
}

#' @noRd
# Find the maximum number of component paths
max_num_comp <- function(x,
                         num_min = 2) {
  tmp <- regexpr("_([[:alnum:]]+)$", x)
  tmp2 <- ifelse(tmp > 0,
                 substring(x,
                           tmp + 1),
                 "")
  out <- as.numeric(tmp2)
  out <- out[!is.na(out)]
  if (length(out) == 0) {
    return(num_min)
  }
  out <- max(out, na.rm = TRUE)
  out
}


#' @noRd
# Expand indirect labels.
# E.g., si to si.1, si.2, si.3.
# Output
# - A numeric vector
expand_ind_labels <- function(x,
                              es_ind = c("si",
                                         "mi",
                                         "li"),
                              num_comp = 2) {
  es_ind_x <- x[es_ind]
  es_ind_x <- es_ind_x[!is.na(es_ind_x)]
  if (length(es_ind_x) == 0) {
    return(x)
  }
  es_others <- x[setdiff(names(x), es_ind)]
  es_ind1 <- union(es_ind, names(es_ind_x))
  tmpfct <- function(xx,
                     num_comp) {
    sapply(seq_len(num_comp),
           function(yy) {
             z1 <- paste0(xx, "_", yy)
             z2 <- es_ind_x[xx]^(1 / yy)
             stats::setNames(z2, z1)
           },
           USE.NAMES = FALSE)
  }
  out <- lapply(es_ind1,
                tmpfct,
                num_comp = num_comp)
  out <- unlist(out)
  c(es_others, out)
}

#' @noRd
check_valid_es_values <- function(object,
                                  pop_es_name,
                                  es_min = -.95,
                                  es_max =  .95,
                                  step = 0.05,
                                  R0 = 2,
                                  n_std = 5000,
                                  n = 1000,
                                  verbose = FALSE,
                                  ...) {
  if (!(es_min < es_max)) {
    stop("'es_min must be less than 'es_max'.")
  }
  if (length(seq(es_min, es_max, by = step)) <= 1) {
    stop("'step' results in only one value between 'es_min' and 'es_max")
  }
  work_mode <- ""
  if (inherits(object, "power4test")) {
    work_mode <- "power4test"
  } else if (inherits(object, "ptable_pop")) {
    work_mode <- "ptable_pop"
  } else {
    stop("object not supported")
  }
  if (work_mode == "ptable_pop") {
    if (is.null(attr(object, "standardized"))) {
      stop("The parameter table is not generated by power4test() or pop_table().")
    }
    attr(object, "n_std") <- n_std
  }
  if (work_mode == "power4test") {
    R <- attr(object, "args")$R
    if (!is.null(R)) {
      R <- R0
    }
  }

  # Assume there is only one region of valid values
  # TODO:
  # - Relax this assumption?

  es_to_test <- seq(es_min, es_max, by = step)
  es_ok <- rep(FALSE, length(es_to_test))
  names(es_ok) <- sprintf("%4.3f", es_to_test)
  for (i in seq_along(es_to_test)) {
    pop_es_i <- as.character(es_to_test[i])
    names(pop_es_i) <- pop_es_name
    if (verbose) dput(pop_es_i)
    chk <- switch(work_mode,
                  ptable_pop = tryCatch(suppressWarnings(update_ptable_pop(object,
                                                         pop_es_i)),
                                        error = function(e) e),
                  power4test = tryCatch(suppressWarnings(power4test(object,
                                                                    nrep = 1,
                                                                    n = n,
                                                                    pop_es = pop_es_i,
                                                                    R = R,
                                                                    n_std = n_std,
                                                                    do_the_test = FALSE,
                                                                    parallel = FALSE,
                                                                    progress = FALSE,
                                                                    ...)),
                                        error = function(e) e))
    if (!inherits(chk, "error")) {
      es_ok[i] <- TRUE
    } else {
      if (verbose) print(chk)
    }
  }
  if (!any(es_ok)) {
    # No valid values
    return(c(NA, NA))
  }

  if (verbose) print(es_ok)
  es_range <- range(es_to_test[es_ok])
  es_range
}

skip_on_cran()

library(testthat)

test_that("Check valid es values", {

verbose <- FALSE

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       progress = !is_testing(),
                       parallel = FALSE)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE),
                       progress = FALSE,
                       parallel = FALSE)

chk1 <- check_valid_es_values(test_out,
                              pop_es_name = "y ~ x",
                              verbose = verbose)
chk2 <- check_valid_es_values(test_out$sim_all[[1]]$ptable,
                              pop_es_name = "y ~ x",
                              verbose = verbose)
expect_equal(chk1,
             c(-.95, .7))
expect_equal(chk1,
             chk2)

chk1 <- check_valid_es_values(test_out,
                              pop_es_name = "y ~ m",
                              verbose = verbose)
chk2 <- check_valid_es_values(test_out$sim_all[[1]]$ptable,
                              pop_es_name = "y ~ m",
                              verbose = verbose)
expect_equal(chk1,
             c(-.95, .95))
expect_equal(chk1,
             chk2)

chk1 <- check_valid_es_values(test_out,
                              pop_es_name = "m ~ x",
                              verbose = verbose)
chk2 <- check_valid_es_values(test_out$sim_all[[1]]$ptable,
                              pop_es_name = "m ~ x",
                              verbose = verbose)
expect_equal(chk1,
             c(-.95, .95))
expect_equal(chk2,
             chk1)

# Case 2

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "-m",
                         "y ~ x" = "-s")

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 1234,
                       progress = !is_testing(),
                       parallel = FALSE)

chk1 <- check_valid_es_values(sim_only,
                              pop_es_name = "y ~ x",
                              verbose = verbose)
chk2 <- check_valid_es_values(sim_only$sim_all[[1]]$ptable,
                              pop_es_name = "y ~ x",
                              verbose = verbose)
expect_equal(chk1,
             c(-.7, .95))
expect_equal(chk1,
             chk2)

chk1 <- check_valid_es_values(sim_only,
                              pop_es_name = "y ~ m",
                              verbose = verbose)
chk2 <- check_valid_es_values(sim_only$sim_all[[1]]$ptable,
                              pop_es_name = "y ~ m",
                              verbose = verbose)
expect_equal(chk1,
             c(-.95, .95))
expect_equal(chk2,
             chk1)

chk1 <- check_valid_es_values(sim_only,
                              pop_es_name = "m ~ x",
                              verbose = verbose)
chk2 <- check_valid_es_values(sim_only$sim_all[[1]]$ptable,
                              pop_es_name = "m ~ x",
                              verbose = verbose)
expect_equal(chk1,
             c(-.95, .95))
expect_equal(chk2,
             chk1)

# Check error

expect_error(check_valid_es_values(sim_only,
                                   pop_es_name = "m ~ x",
                                   es_min = .5,
                                   es_max = -.5,
                                   verbose = verbose))
expect_error(check_valid_es_values(sim_only,
                                   pop_es_name = "m ~ x",
                                   es_min = .5,
                                   es_max = .5,
                                   verbose = verbose))
expect_equal(check_valid_es_values(sim_only,
                                   pop_es_name = "m ~ x",
                                   es_min = 2,
                                   es_max = 3,
                                   verbose = verbose),
             c(NA, NA))

})

# Test

test_that("set_pop", {
model_es1 <- c("m1 ~ x" = "-m",
               "m2 ~ m1" = "s",
               "y ~ m2" = "l",
               "y ~ x" = "l",
               "y ~ w" = "s",
               "y ~ x:w" = "s",
               "x ~~ w" = "s")

model_es2 <- c("m1 ~ x" = "-m",
               "m2 ~ m1" = "s",
               "y ~ m2 + x" = "l",
               "y ~ w" = "s",
               "y ~ x:w" = "s",
               "w ~~ x" = "s")

ptable_es1 <- set_pop(model_es1)
ptable_es2 <- set_pop(fix_par_es(model_es2))
expect_identical(ptable_es1,
                 ptable_es2)

es1_temp <- c("s" = .10,
              "m" = .30,
              "l" = .50)
es1_tmpn <- es1_temp * -1
names(es1_tmpn) <- paste0("-", names(es1_tmpn))
es1_all <- c(es1_temp, es1_tmpn)
es2_temp <- c("s" = .05,
              "m" = .10,
              "l" = .15)
es2_tmpn <- es2_temp * -1
names(es2_tmpn) <- paste0("-", names(es2_tmpn))
es2_all <- c(es2_temp, es2_tmpn)
expect_equal(ptable_es1$pop[6],
             es2_all[ptable_es1$es][6],
             ignore_attr = TRUE)

pop_es <- c("m ~ x" = "log(.20)",
            "y ~ m" = ".20",
            "y ~ x" = "s * m")
tmp <- set_pop(pop_es)
expect_equal(tmp$pop,
             c(log(.20), .20, .10 * .30),
             ignore_attr = TRUE)

pop_es <- c("m ~ x" = "sqrt(.40)",
            "y ~ m:x" = "-s",
            "y ~ x" = "s + .10")
tmp <- set_pop(pop_es)
expect_equal(tmp$pop,
             c(sqrt(.40), -.05, .10 + .10),
             ignore_attr = TRUE)

mod <-
"
m ~ x
y ~ m + x
"

mod_es1 <- c(".beta." = "m",
             ".ind.(x->m->y)" = "li")

pop_es <- fix_par_es(mod_es1,
                      model = mod)

tmp <- set_pop(pop_es)
expect_equal(tmp$pop,
             c(sqrt(.510), sqrt(.510), .30),
             ignore_attr = TRUE)

mod <-
"
m ~ x
y ~ m + x
"

mod_es1 <- c(".beta." = "m",
             ".ind.(x->m->y)" = ".10^2")

pop_es <- fix_par_es(mod_es1,
                      model = mod)

tmp <- set_pop(pop_es)
expect_equal(tmp$pop,
             c(.10, .10, .30),
             ignore_attr = TRUE)

})

# Test

test_that("fix_par_es: indirect", {

x1 <- ".ind.(x1-> m1->m2->y)"
x2 <- ".ind. (   x1   -> m1 -> m2-> y)"
x3 <- ".ind. (y ~ m2 ~ m1~x1  )  "

tmp1 <- expand_to_components(x1)
tmp2 <- expand_to_components(x2)
tmp3 <- expand_to_components(x3)

expect_equal(tmp1,
             tmp2)
expect_equal(sort(tmp1),
             sort(tmp3))


model <-
"
m1 ~ x1
m2 ~ m1
m3 ~ x2
y ~ m2 + m3 + x1
y2 ~ m3 + x2
"

model_es <- c("m1 ~ x1" = "-m",
              "m2 ~ m1" = "s",
              "y ~ m3" = "l",
              ".beta." = "m",
              ".cov." = "-s",
              ".ind.(x1-> m1->m2->y)" = "mi",
              ".ind.(y~m3~ x2)" = "li")

tmp1 <- fix_par_es(model_es,
                   model)
tmp1_chk <- c(`m1 ~ x1` = "(mi)^(1 / 3)", `m2 ~ m1` = "(mi)^(1 / 3)", `y ~ m2` = "(mi)^(1 / 3)",
`y ~ m3` = "(li)^(1 / 2)", `m3 ~ x2` = "(li)^(1 / 2)", `y ~ x1` = "m", `y2 ~ m3` = "m",
`y2 ~ x2` = "m", `x1 ~~ x2` = "-s")

expect_equal(sort(tmp1),
             sort(tmp1_chk))

# max_num_comp() is no longer needed
# # Get maximum numbers of components

# expect_equal(max_num_comp(tmp1),
#              3)
# expect_equal(max_num_comp(tmp1[6:9]),
#              2)

# Set the values

es1_tmp <- c("n" = .00,
             "nil" = .00,
             "s" = .10,
             "m" = .30,
             "l" = .50,
             "si" = .141,
             "mi" = .361,
             "li" = .510)
es_long(es1_tmp)

expand_ind_labels(es1_tmp, num_comp = 2)
expand_ind_labels(es1_tmp, num_comp = 10)

es_tmp1 <- set_pop(tmp1,
                   es1 = c("n" = .00,
                           "nil" = .00,
                           "s" = .10,
                           "m" = .30,
                           "l" = .50,
                           "si" = .141,
                           "mi" = .361,
                           "li" = .510),
                   es2 = c("n" = .00,
                           "nil" = .00,
                           "s" = .05,
                           "m" = .10,
                           "l" = .15))

es_tmp1$lavlabel <- lavaan::lav_partable_labels(es_tmp1)

expect_equal(es_tmp1[es_tmp1$lavlabel == "m1~x1", "pop"],
             es1_tmp["mi"]^(1 / 3),
             ignore_attr = TRUE)
expect_equal(es_tmp1[es_tmp1$lavlabel == "m3~x2", "pop"],
             es1_tmp["li"]^(1 / 2),
             ignore_attr = TRUE)

})

#' @noRd
# Input:
# - A list of model matrices
# - The corresponding parameter table.
# Output:
# - The parameter table with start values updated
start_from_mm <- function(ptable,
                          mm) {
  ptable_out <- ptable
  fit <- lavaan::sem(ptable,
                     do.fit = FALSE)
  mm0 <- lavaan::lavInspect(fit,
                            "partable",
                            drop.list.single.group = FALSE)
  for (k0 in seq_along(mm0)) {
    for (k in seq_along(mm0[[k0]])) {
      mmi <- mm0[[k0]][[k]]
      mmi1 <- mm[[k0]][[k]]
      i <- which(mmi > 0)
      j <- mmi[i]
      mmic <- col(mmi)
      mmir <- row(mmi)
      for (x in seq_along(j)) {
        ptable_out[j[x], "start"] <- mmi1[mmir[i[x]], mmic[i[x]]]
      }
    }
  }
  ptable_out
}

#' @noRd
# Input:
# - mm_lm object
# Output:
# - Generated data
mm_lm_data <- function(object,
                       n,
                       number_of_indicators = NULL,
                       reliability = NULL,
                       loading_difference = NULL,
                       reference = NULL,
                       keep_f_scores = FALSE,
                       x_fun = list(),
                       e_fun = list(),
                       process_data = NULL) {
  lm_y <- object$lm_y
  psi <- object$psi
  all_vars <- colnames(psi)
  i <- grepl("[:]", all_vars)
  p <- length(lm_y)
  q <- ncol(psi)
  x_raw <- gen_all_x(psi = psi,
                     n = n,
                     all_x = all_vars,
                     x_fun = x_fun)
  dat_all <- matrix(NA,
                    nrow = n,
                    ncol = length(all_vars))
  colnames(dat_all) <- all_vars
  dat_all[, colnames(x_raw)] <- x_raw
  if (!is.null(lm_y)) {
    # Update product terms
    for (y in names(lm_y)) {
      lm_y_i <- lm_y[[y]]
      beta_i <- lm_y_i$beta
      e_i <- dat_all[, y]
      y_i <- dat_all[, colnames(beta_i), drop = FALSE] %*% t(beta_i) + e_i
      # Replace the errors by the generated values
      dat_all[, y] <- as.vector(y_i)
      dat_all <- update_p_terms(dat_all)
    }
  }
  if (!is.null(number_of_indicators)) {
    dat_all <- add_indicator_scores(dat_all,
                                    ps = number_of_indicators,
                                    rels = reliability,
                                    e_fun = e_fun,
                                    keep_f_scores = keep_f_scores,
                                    ds = loading_difference,
                                    refs = reference)
  }
  lambda_pop <- attr(dat_all, "lambda")
  dat_all <- as.data.frame(dat_all)
  # In case a processor needs lambda
  attr(dat_all, "lambda") <- lambda_pop
  if (is.list(process_data)) {
    process_data_fun <- match.fun(process_data$fun)
    attr(dat_all, "number_of_indicators") <- number_of_indicators
    tmp <- list(dat_all)
    names(tmp) <- process_data$sim_data_name %||% "data"
    if (is.null(process_data$args)) {
      process_data$args <- list()
    }
    process_data_args <- utils::modifyList(process_data$args,
                                          tmp)
    dat_all_amp <- do.call(process_data_fun,
                           process_data_args)
    m_name <- process_data$processed_data_name
    if (!is.null(m_name)) {
      dat_all <- dat_all_amp[[m_name]]
    } else {
      dat_all <- dat_all_amp
    }
  }
  # In case a processor removes the attributes
  attr(dat_all, "lambda") <- lambda_pop
  return(dat_all)
}

#' @noRd
# Input:
# - Original model syntax
# - number_of_indicators
# - reliability
# Note:
# - To be used internally. Assume that the argument values are valid.
# - Assume the prefix of indicators are equal to the factor names.
#   - E.g., x1, x2, ... for the factor x.
# Output:
# - Modified model syntax with the measurement part
add_indicator_syntax <- function(model,
                                 number_of_indicators = NULL,
                                 reliability = NULL) {
  f_names <- names(number_of_indicators)
  reliability <- reliability[f_names]
  if (!is.null(number_of_indicators)) {
    for (i in seq_along(number_of_indicators)) {
      i_name <- names(number_of_indicators)[i]
      model <- c(model,
                paste0(i_name, " =~ ",
                       paste0(i_name,
                              seq_len(number_of_indicators[i]),
                              collapse = " + ")))
    }
  }
  return(model)
}


#' @noRd
# Input:
# - A matrix of scores (e.g., from mm_lm_data()).
# - ps: A named vector of the number of indicators.
# - rels: A named vector of reliability coefficients.
# - ds: A named vector of 'd's, differences in lambdas.
# - keep_f_scores: If TRUE, factor scores are ratained.
# Output:
# - A matrix with indicator scores added.
add_indicator_scores <- function(x,
                                 ps,
                                 rels,
                                 refs = NULL,
                                 keep_f_scores = FALSE,
                                 e_fun = NULL,
                                 ds = NULL) {
  if (!setequal(names(ps), names(rels))) {
    stop("'ps' and 'rels' do not match in names.")
  }
  f_names <- names(ps)
  if (length(setdiff(f_names, colnames(x))) != 0) {
    stop("Some factors not found in the data.")
  }
  rels <- rels[f_names]

  ds0 <- rels
  ds0[] <- 0
  if (!is.null(ds)) {
    ds0[f_names] <- ds[f_names]
  }

  refs0 <- vector("character", length(rels))
  names(refs0) <- names(rels)
  refs0[] <- "medium"
  if (!is.null(refs)) {
    refs0[f_names] <- refs[f_names]
  }
  f_scores <- sapply(f_names,
                     function(xx) x[, xx, drop = TRUE],
                     simplify = FALSE)
  if (is.list(e_fun) && length(e_fun) > 0) {
    e_fun_names <- names(e_fun)
    e_fun1 <- sapply(f_names,
                     function(x, e_fun) {
                       if (x %in% e_fun_names) {
                         return(e_fun[[x]])
                       } else {
                         return(list())
                       }
                     },
                     e_fun = e_fun,
                     simplify = FALSE,
                     USE.NAMES = TRUE)
  } else {
    e_fun1 <- vector("list", length(f_names))
    names(e_fun1) <- f_names
  }
  prefixes <- f_names
  out0 <- mapply(gen_indicator_scores,
                 f_score = f_scores,
                 p = ps,
                 omega = rels,
                 prefix = prefixes,
                 e_fun = e_fun1,
                 d = ds0,
                 ref = refs0,
                 SIMPLIFY = FALSE)
  lambda_pop <- sapply(
                  out0,
                  attr,
                  which = "lambda",
                  simplify = FALSE
                )
  out1 <- do.call(cbind,
                  out0)
  out2 <- cbind(x, out1)
  if (!keep_f_scores) {
    # TODO:
    # - Should also remove product terms
    i <- match(f_names, colnames(x))
    out2 <- out2[, -i, drop = FALSE]
  }
  attr(out2, "lambda") <- lambda_pop
  return(out2)
}

#' @noRd
# Input:
# - f_score: A vector or column vector of factor scores.
# - p: The number of indicators.
# - omega: The reliability.
# - prefix: The prefix for naming the indicators.
# - d: The difference in lambdas between neighbouring items
# - ref: Which indicator will be the first indicator
# Ouput:
# - A matrix of indicator scores
gen_indicator_scores <- function(f_score,
                                 p,
                                 omega,
                                 prefix = "x",
                                 e_fun = list(),
                                 d = 0,
                                 ref = "medium") {
  # e_fun is of this form
  # list(rexp_rs, ....)
  f_score <- matrix(as.vector(f_score),
                    ncol = 1)
  n <- nrow(f_score)
  lambda0 <- lambda_from_reliability(p = p,
                                    omega = omega,
                                    d = d,
                                    ref = ref)
  # lambda and e_sd may differ across items.
  lambda1 <- matrix(lambda0,
                    nrow = 1,
                    ncol = p)
  e_sd <- sqrt(1 - lambda0^2)
  if (length(e_fun) > 0) {
    ee_fun <- e_fun[[1]]
    ee_fun <- match.fun(ee_fun)
    ee_args <- e_fun[-1]
    ee_args <- utils::modifyList(ee_args,
                                 list(n = n * p))
    e <- do.call(ee_fun,
                 ee_args)
    e <- matrix(e,
                nrow = n,
                ncol = p)
  } else {
    e <- matrix(stats::rnorm(n * p,
                             mean = 0,
                             sd = 1),
                nrow = n,
                ncol = p)
  }
  e <- e %*% diag(e_sd)
  x <- f_score %*% lambda1 + e
  colnames(x) <- paste0(prefix, seq(from = 1,
                                    to = p))
  attr(x, "lambda") <- lambda0
  return(x)
}

#' @noRd
# Check product terms
# Input:
# - Output of model_matrices_pop()
# Output:
# - Logical. TRUE if the model has one or more product terms
has_p_terms <- function(object) {
  vnames <- colnames(object$beta)
  return(any(grepl(":", vnames)))
}

#' @noRd
# Input:
# - Output of model_matrices_pop()
# Output:
# - Psi matrix if all variables standardized
#
psi_std <- function(object,
                    n_std = 100000) {
  std_by_monte_carlo <- FALSE
  if (has_p_terms(object)) {
    out <- psi_std_monte_carlo(object = object,
                               n_std = n_std)
    std_by_monte_carlo <- TRUE
  } else {
    out <- tryCatch(psi_std_analytic(object = object),
                    error = function(e) e)
    if (inherits(out, "error")) {
      out <- psi_std_monte_carlo(object = object,
                                 n_std = n_std)
      std_by_monte_carlo <- TRUE
    }
  }
  if (any(diag(out) < 0)) {
    tmp <- colnames(out)[(diag(out) < 0)]
    msg <- paste0("Negative model or implied variance(s) for ",
                  paste0(tmp, collapse = ", "),
                  ". ",
                  "Please check the model.")
    stop(msg)
  }
  if (any(diag(out) > 1)) {
    # TODO:
    # - Should skip all derived terms, such as x^2.
    # - Find a better way to identify derived terms.

    # Skip product terms
    tmp0 <- colnames(out)
    tmp1 <- sapply(strsplit(tmp0, ":", fixed = TRUE),
                   length)
    tmp2 <- diag(out)[(tmp1 == 1)]
    if (any(tmp2 > 1)) {
      tmp <- colnames(out)[(tmp2 >= 1)]
      msg <- paste0("Model or implied variance(s) for ",
                    paste0(tmp, collapse = ", "),
                    " greater than 1 when standardized. ",
                    "Please check the model.")
      stop(msg)
    }
  }
  attr(out, "std_by_monte_carlo") <- std_by_monte_carlo
  return(out)
}

#' @noRd
# Input:
# - Output of model_matrices_pop()
# Output:
# - Psi matrix if all variables standardized
# Note:
# - Determine by Monte Carlo
# - Can have product terms
psi_std_analytic <- function(object) {
  vnames <- colnames(object$beta)
  p <- length(vnames)
  object_lm <- mm_lm(object)
  lm_y <- object_lm$lm_y
  mm_tmp <- object
  for (yy in names(lm_y)) {
    sigma <- implied_sigma(mm_tmp)
    mm_tmp$psi[yy, yy] <- 2 - sigma[yy, yy]
  }
  # Check
  if (all.equal(rep(1, p),
                diag(implied_sigma(mm_tmp)),
                tolerance = 1e-4,
                check.attributes = FALSE,
                check.class = FALSE)) {
    return(mm_tmp$psi)
  } else {
    stop("Analytical standardization failed. ",
         "Please set std_force_monte_carlo to TRUE ",
         "to standardize by simulation.")
  }
}

#' @noRd
endo_beta <- function(object) {
  beta <- object$beta
  vnames <- colnames(beta)
  exo <- apply(beta,
               MARGIN = 1,
               function(xx) all(xx == 0))
  vnames[!exo]
}

#' @noRd
# Input:
# - Output of model_matrices_pop()
# Output:
# - Psi matrix if all variables standardized
# Note:
# - Determine by Monte Carlo
# - Can have product terms
psi_std_monte_carlo <- function(object,
                                n_std = 100000,
                                x_fun = list()) {
  object <- mm_lm(object)
  lm_y <- object$lm_y
  psi <- object$psi
  all_vars <- colnames(psi)
  i <- grepl("[:]", all_vars)
  p <- length(lm_y)
  q <- ncol(psi)
  x_raw <- gen_all_x(psi = psi,
                     n = n_std,
                     all_x = all_vars,
                     x_fun = list())
  dat_all <- matrix(NA,
                    nrow = n_std,
                    ncol = length(all_vars))
  colnames(dat_all) <- all_vars
  dat_all[, colnames(x_raw)] <- x_raw
  for (y in names(lm_y)) {
    lm_y_i <- lm_y[[y]]
    beta_i <- lm_y_i$beta
    y_hat <- dat_all[, colnames(beta_i), drop = FALSE] %*% t(beta_i)
    y_hat_var <- stats::var(y_hat[, 1])
    y_e_var <- 1 - y_hat_var
    # Update psi
    psi[y, ] <- psi[y, ] * sqrt(y_e_var)
    psi[, y] <- psi[, y] * sqrt(y_e_var)
    # Rescale the errors
    y_e <- dat_all[, y, drop = TRUE] * sqrt(y_e_var) / stats::sd(dat_all[, y])
    # Replace the errors by the generated values
    # Update product terms
    dat_all <- update_p_terms(dat_all)
    dat_all[, y] <- y_e + y_hat
  }
  return(psi)
}


#' @noRd
# Generate all x's
gen_all_x <- function(psi,
                      n,
                      all_x,
                      x_fun = list()) {
  x_raw <- gen_pure_x(psi = psi,
                      n = n,
                      x_fun = x_fun)
  x_raw <- add_p_terms(x_raw,
                       all_x = all_x)
  x_raw
}

#' @noRd
# Compute the product term
update_p_terms <- function(x) {
  out <- x
  for (xx in colnames(out)) {
    if (!grepl("[:]", xx)) next
    a <- strsplit(xx, ":")[[1]]
    b <- apply(out[, a],
               MARGIN = 1,
               prod)
    out[, xx] <- b
  }
  out
}

#' @noRd
# Compute the product term
add_p_terms <- function(x,
                        all_x) {
  n <- nrow(x)
  p1 <- length(all_x)
  x0 <- colnames(x)
  out <- matrix(NA,
                ncol = p1,
                nrow = n)
  colnames(out) <- all_x
  out[, x0] <- x
  for (xx in colnames(out)) {
    if (!grepl("[:]", xx)) next
    a <- strsplit(xx, ":")[[1]]
    b <- apply(out[, a],
                MARGIN = 1,
                prod)
    out[, xx] <- b
  }
  out
}

#' @noRd
gen_pure_x <- function(psi,
                       n,
                       x_fun = list()) {
  p <- ncol(psi)
  psi_q <- chol(psi)
  x0 <- matrix(stats::rnorm(n * p),
               nrow = n,
               ncol = p)
  x <- x0 %*% psi_q
  colnames(x) <- colnames(psi)

  # Generate data using x_fun functions
  if (length(x_fun) > 0) {
    xnames <- colnames(x)
    x_fun_names <- names(x_fun)
    if (!all(x_fun_names %in% xnames)) {
      tmp <- setdiff(x_fun_names,
                     xnames)
      msg <- paste0(paste0(tmp, collapse = ", "),
                    " in x_fun but not in the model.")
      stop(msg)
    }
    for (xx in x_fun_names) {
      psi_xx <- psi[xx, ]
      psi_xx <- psi_xx[-which(xnames == xx)]
      if (!isTRUE(all.equal(psi_xx,
                            rep(0, length(psi_xx)),
                            tolerance = 1e-6,
                            check.attributes = FALSE))) {
        tmp <- paste0(xx, " must have zero covariance with other variables",
                      " to be eligible for using x_fun.")
        stop(tmp)
      }
    }
    tmpfct <- function(xx,
                       n) {
      xx_fun <- xx[[1]]
      xx_fun <- match.fun(xx_fun)
      xx <- xx[-1]
      xx <- utils::modifyList(xx,
                              list(n = n))
      out <- do.call(xx_fun,
                     xx)
      out
    }
    x_fun_out <- sapply(x_fun,
                        tmpfct,
                        n = n,
                        simplify = FALSE,
                        USE.NAMES = TRUE)
    for (xx in x_fun_names) {
      x[, xx] <- x_fun_out[[xx]]
    }
  }

  x
}

#' @noRd
implied_sigma <- function(mm) {
  beta <- mm$beta
  psi <- mm$psi
  p <- ncol(psi)
  sigma <- solve(diag(p) - beta) %*% psi %*% t(solve(diag(p) - beta))
  return(sigma)
}

#' @noRd

miss_pattern <- function(data) {
  # It is an internal function
  # Assume data is either a data frame or a matrix
  n <- nrow(data)
  p <- ncol(data)
  if (sum(stats::complete.cases(data)) == n) {
    # Complete data
    out <- matrix(1,
                  ncol = p,
                  nrow = 1)
    out[1, ] <- 1
    colnames(out) <- colnames(data)
    rownames(out) <- nrow(data)
    nvalid <- rep(n, p)
    names(nvalid) <- colnames(data)
    attr(out, "nvalid") <- nvalid
    return(out)
  }
  data_na <- !is.na(data)
  pattern <- apply(data_na,
                   MARGIN = 1,
                   \(x) paste0(as.numeric(x), collapse = ""))
  pattern_n0 <- table(pattern)
  pattern_n <- as.numeric(pattern_n0)
  names(pattern_n) <- names(pattern_n0)
  pattern_n <- pattern_n[order(names(pattern_n), decreasing = TRUE)]
  n_pattern <- length(pattern_n)
  tmpfct <- function(x) {
    as.numeric(strsplit(x, "")[[1]])
  }
  out <- do.call(rbind,
                 lapply(names(pattern_n),
                        tmpfct))
  colnames(out) <- colnames(data)
  rownames(out) <- pattern_n

  nvalid <- colSums(diag(pattern_n) %*% out)
  names(nvalid) <- colnames(data)
  attr(out, "nvalid") <- nvalid
  out
}

#' @noRd
print_miss_pattern <- function(mp,
                               digits = 2) {
  mp_out <- mp
  mp_count <- as.numeric(rownames(mp))
  n <- sum(as.numeric(rownames(mp)))
  mp_prop <- mp_count / n
  mp_prop_str <- formatC(mp_prop,
                          digits = digits + 2,
                          format = "f")
  mp_prop_str <- gsub("0.", ".", mp_prop_str,
                      fixed = TRUE)
  mp_nvalid <- attr(mp, "nvalid")
  mp_prop_valid_str <- formatC(mp_nvalid / n,
                                digits = digits,
                                format = "f")
  mp_prop_valid_str <- gsub("0.", ".", mp_prop_valid_str,
                            fixed = TRUE)
  mp_out[] <- ifelse(as.character(mp) == 1,
                      yes = "O",
                      no = "-")
  mp_p <- rowSums(mp)
  mp_out <- cbind(`P Prop` = mp_prop_str,
                  mp_out,
                  `# V` = mp_p)
  mp_out <- rbind(mp_out,
                  c("V Prop", mp_prop_valid_str, ""))
  rownames(mp_out) <- NULL
  print(as.data.frame(mp_out),
        quote = FALSE,
        row.names = FALSE,
        right = TRUE)
  cat("\nNote:\n")
  catwrap("- 'O': A variable has data in a pattern.",
          exdent = 2)
  catwrap("- '-': A variable has missing data in a pattern.",
          exdent = 2)
  catwrap("- P Prop: Proportion of each missing pattern.",
          exdent = 2)
  catwrap("- # V: Number of non-missing variable(s) in a pattern.",
          exdent = 2)
  catwrap("- V Prop: Proportion of non-missing data of each variables.",
          exdent = 2)
  cat("\n")
}

#' @noRd
pure_x <- function(fit) {
  tmp1 <- union(
            lavaan::lavNames(fit, "ov.x"),
            lavaan::lavNames(fit, "lv.x")
          )
  tmp2 <- setdiff(
            lavaan::lavNames(fit, "eqs.x"),
            lavaan::lavNames(fit, "eqs.y")
          )
  tmp3 <- union(lavaan::lavNames(fit, "ov.ind"),
                lavaan::lavNames(fit, "ov.interaction"))
  tmp3 <- union(tmp3,
                lavaan::lavNames(fit, "lv.interaction"))
  out <- union(tmp1, tmp2)
  out <- setdiff(out, tmp3)
  out
}

#' @noRd
pure_y <- function(fit) {
  tmp1 <- union(
            lavaan::lavNames(fit, "ov.y"),
            lavaan::lavNames(fit, "lv.y")
          )
  tmp2 <- setdiff(
            lavaan::lavNames(fit, "eqs.y"),
            lavaan::lavNames(fit, "eqs.x")
          )
  tmp3 <- union(lavaan::lavNames(fit, "ov.ind"),
                lavaan::lavNames(fit, "ov.interaction"))
  tmp3 <- union(tmp3,
                lavaan::lavNames(fit, "lv.interaction"))
  out <- union(tmp1, tmp2)
  out <- setdiff(out, tmp3)
  out
}

#' @noRd
all_x <- function(fit) {
  tmp1 <- union(
            lavaan::lavNames(fit, "ov.x"),
            lavaan::lavNames(fit, "lv.x")
          )
  tmp3 <- union(lavaan::lavNames(fit, "ov.ind"),
                lavaan::lavNames(fit, "ov.interaction"))
  tmp3 <- union(tmp3,
                lavaan::lavNames(fit, "lv.interaction"))
  out <- setdiff(tmp1, tmp3)
  out
}

#' @noRd
all_y <- function(fit) {
  tmp1 <- union(
            lavaan::lavNames(fit, "ov.y"),
            lavaan::lavNames(fit, "lv.y")
          )
  tmp3 <- union(lavaan::lavNames(fit, "ov.ind"),
                lavaan::lavNames(fit, "ov.interaction"))
  tmp3 <- union(tmp3,
                lavaan::lavNames(fit, "lv.interaction"))
  out <- setdiff(tmp1, tmp3)
  out
}

#' @noRd
get_w <- function(
            x,
            m,
            y,
            fit) {
  ind_prod <- manymome::indirect_i(
                x = x,
                m = m,
                y = y,
                fit = fit,
                get_prods_only = TRUE
              )
  out <- sapply(
          ind_prod,
          function(xx) {
            if (identical(xx, NA)) {
              return(as.character(NA))
            } else {
              if (is.null(xx$w)) {
                return(as.character(NA))
              } else {
                return(xx$w)
              }
            }
          })
  out <- unlist(out)
  out <- out[!is.na(out)]
  out
}

#' @noRd
get_w_for_paths <- function(all_paths,
                            fit) {
  f <- function(xx) {
    do.call(get_w,
          c(xx, list(fit = fit)))
  }
  out <- sapply(all_paths,
                f,
                simplify = FALSE,
                USE.NAMES = TRUE)
  out
}

#' @noRd
get_direct <- function(x,
                       y,
                       ptable) {
  out <- list()
  for (x_i in x) {
    for (y_i in y) {
      i <- (ptable$lhs == y_i) &
           (ptable$rhs == x_i) &
           (ptable$op == "~")
      if (any(i)) {
        out_i <- list(list(x = x_i,
                           m = NULL,
                           y = y_i))
        names(out_i) <- paste0(x_i, " -> ", y_i)
        out <- c(out, out_i)
      }
    }
  }
  out
}

library(testthat)

test_that("miss_pattern", {

set.seed(1234)
n <- 200
p <- 8
prop <- .50
dat <- matrix(rnorm(n * p),
              n, p)
dat[sample.int(n * p, ceiling(n * prop))] <- NA
# dat[sample.int(n, ceiling(n * (1 - prop))), ] <- 1
dat <- as.data.frame(dat)

out <- miss_pattern(dat)
chk0 <- !is.na(dat)
expect_equal(attr(out, "nvalid"),
             colSums(chk0))
expect_equal(as.numeric(rownames(out)[1]),
             sum(complete.cases(dat)))

# Complete data
set.seed(1234)
n <- 123
p <- 8
dat <- matrix(rnorm(n * p),
              n, p)
dat <- as.data.frame(dat)

out <- miss_pattern(dat)
expect_equal(nrow(out), 1)
expect_equal(colnames(out),
             colnames(dat))
expect_equal(unique(attr(out, "nvalid")),
             n)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Generate nonnormal data", {

mod <-
"
m ~ x + w + x:w
y ~ m + x
"

set.seed(1234)
n <- 100000
pop_es <- c(".beta." = "m")

ptable <- ptable_pop(model = mod,
                     pop_es = pop_es)
mm_out <- model_matrices_pop(ptable)
mm_lm_out <- mm_lm(mm_out)

set.seed(1234)
mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                            n = 10000,
                            x_fun = list(x = list(rexp_rs),
                                         w = list(rbinary_rs,
                                                  p1 = .70)))
chk <- mm_lm_dat_out
head(chk)
expect_equal(length(table(chk[, "w"])),
             2)
expect_equal(chk[1:10, "x"] * chk[1:10, "w"],
             chk[1:10, "x:w"])


psi <- mm_lm_out$psi
all_vars <- colnames(psi)
x_raw <- gen_all_x(psi = psi,
                   n = 10000,
                   all_x = all_vars,
                   x_fun = list(x = list(rexp_rs),
                                 w = list(rbinary_rs,
                                          p1 = .70)))
chk <- x_raw
head(chk)
expect_equal(length(table(chk[, "w"])),
             2)
expect_equal(chk[1:10, "x"] * chk[1:10, "w"],
             chk[1:10, "x:w"])

x_raw <- gen_pure_x(psi = psi,
                    n = n,
                    x_fun = list(x = list(rexp_rs),
                                 w = list(rbinary_rs,
                                          p1 = .70)))
expect_equal(length(table(x_raw[, "w"])),
             2)

sim_out <- sim_data_i(ptable = ptable,
                      n = n,
                      seed = 1234,
                      x_fun = list(x = list(rexp_rs),
                                 w = list(rbinary_rs,
                                          p1 = .70)))
chk <- sim_out$mm_lm_dat_out
head(chk)
expect_equal(length(table(chk[, "w"])),
             2)
expect_equal(chk[1:10, "x"] * chk[1:10, "w"],
             chk[1:10, "x:w"])

sim_out <- sim_data(nrep = 2,
                    model = mod,
                    pop_es = pop_es,
                    n = n,
                    iseed = 1234,
                    x_fun = list(x = list(rexp_rs),
                               w = list(rbinary_rs,
                                        p1 = .70)),
                    progress = !is_testing())
chk <- sim_out[[2]]$mm_lm_dat_out
head(chk)
expect_equal(length(table(chk[, "w"])),
             2)
expect_equal(chk[1:10, "x"] * chk[1:10, "w"],
             chk[1:10, "x:w"])
})

library(testthat)
suppressMessages(library(lavaan))

omega_test <- function(data) {
  p <- ncol(data)
  mod <- paste("f1 =~",
              paste0("item", seq_len(p), collapse = " + "))
  fit <- lavaan::cfa(mod, data)
  lambda <- c(1, coef(fit)[1:(p - 1)])
  f_var <- lavaan::lavInspect(fit, "cov.lv")[1, 1]
  total_var <- sum(lavaan::lavInspect(fit, "sampstats")$cov)
  omega <- (sum(lambda)^2) * f_var / total_var
  omega
}

test_that("gen_indicator_scores", {

set.seed(1234)
n <- 50000
f <- rnorm(n)
p0 <- 7
omega0 <- .65
dat <- gen_indicator_scores(f_score = f,
                            p = p0,
                            omega = omega0,
                            prefix = "item")
dat <- as.data.frame(dat)
expect_equal(omega_test(dat),
             omega0,
             tolerance = 1e-2)
expect_equal(apply(dat, 2, sd),
             rep(1, p0),
             ignore_attr = TRUE,
             tolerance = 1e-2)
expect_equal(colMeans(dat),
             rep(0, p0),
             ignore_attr = TRUE,
             tolerance = 1e-2)

set.seed(1234)
n <- 50000
f <- rnorm(n)
p0 <- 3
omega0 <- .70
dat <- gen_indicator_scores(f_score = f,
                            p = p0,
                            omega = omega0,
                            prefix = "item")
dat <- as.data.frame(dat)
expect_equal(omega_test(dat),
             omega0,
             tolerance = 1e-2)

set.seed(1234)
n <- 50000
f <- rnorm(n)
p0 <- 3
omega0 <- .20
dat <- gen_indicator_scores(f_score = f,
                            p = p0,
                            omega = omega0,
                            prefix = "item")
dat <- as.data.frame(dat)
expect_equal(omega_test(dat),
             omega0,
             tolerance = 1e-2)

})

library(testthat)
suppressMessages(library(lavaan))

check_gen_dat <- function(model,
                          pop_es,
                          n = 10000,
                          seed = NULL,
                          number_of_indicators = NULL,
                          reliability = NULL,
                          keep_f_scores = FALSE) {
  set.seed(seed)
  ptable <- ptable_pop(model,
                       pop_es = pop_es,
                       standardized = TRUE)
  mm_out <- model_matrices_pop(ptable)
  mm_lm_out <- mm_lm(mm_out)
  mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                              n = n,
                              number_of_indicators = number_of_indicators,
                              reliability = reliability,
                              keep_f_scores = keep_f_scores)
  model <- add_indicator_syntax(model,
                                number_of_indicators = number_of_indicators,
                                reliability = reliability)
  fit <- lavaan::sem(model,
             data = mm_lm_dat_out,
             fixed.x = FALSE)
  tmp <- ptable
  tmp$est <- tmp$start
  fit0 <- lavaan::sem(tmp,
              do.fit = FALSE,
              fixed.x = FALSE)
  out <- list(ptable = ptable,
              mm_out = mm_out,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out,
              fit = fit,
              fit0 = fit0)
  out
}

test_check_out <- function(check_out) {
  expect_equal(diag(implied_sigma(check_out$mm_out)),
              rep(1, ncol(check_out$mm_lm_dat_out)),
              tolerance = 1e-1,
              ignore_attr = TRUE)
  # expect_equal(check_out$mm_out$psi,
  #             check_out$mm_out_std$psi)
  expect_equal(diag(cov(check_out$mm_lm_dat_out)),
              rep(1, ncol(check_out$mm_lm_dat_out)),
              tolerance = 1e-1,
              ignore_attr = TRUE)
  expect_equal(coef(check_out$fit),
              coef(check_out$fit0),
              tolerance = 1e-1)
  # expect_equal(check_out$mm_out_std$psi,
  #             check_out$mm_out_std2$psi,
  #             tolerance = 1e-1)
  # expect_equal(check_out$ptable_std$start,
  #             check_out$ptable_std2$start,
  #             tolerance = 1e-1)
}

test_that("ptable_pop: cfa", {

# Correlation model

model_cfa <-
"
f1 ~~ f2 + f3
f2 ~~ f3
"

model_cfa_es <- c(".cov." = "m",
                  "f1 ~~ f2" = "l",
                  "f2 ~~ f3" = "m")

check_out <- check_gen_dat(model_cfa,
                           model_cfa_es,
                           seed = 1234)

expect_equal(check_out$ptable$start[1:3],
             c(.50, .00, .30))

# test_check_out(check_out)

# CFA

model_cfa <-
"
f1 ~~ f2 + f3
f2 ~~ f3
"

model_cfa_es <- c(".cov." = "m",
                  "f1 ~~ f2" = "l",
                  "f2 ~~ f3" = "m")

check_out <- check_gen_dat(model_cfa,
                           model_cfa_es,
                           seed = 1234,
                           number_of_indicators = c(f1 = 4,
                                                    f3 = 3,
                                                    f2 = 5),
                           reliability = c(f2 = .70,
                                           f1 = .80,
                                           f3 = .60))

expect_equal(standardizedSolution(check_out$fit)$est.std[1:3],
             c(.50, .00, .30),
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

check_gen_dat <- function(model,
                          pop_es,
                          n = 10000,
                          seed = NULL,
                          number_of_indicators = NULL,
                          reliability = NULL,
                          keep_f_scores = FALSE) {
  set.seed(seed)
  ptable <- ptable_pop(model,
                       pop_es = pop_es,
                       standardized = TRUE)
  mm_out <- model_matrices_pop(ptable)
  # mm_out_original <- mm_out
  # mm_out$psi <- psi_std(mm_out)
  # ptable_std <- start_from_mm(ptable,
                              # mm_out)
  # ptable_std2 <- ptable_pop(model,
  #                           pop_es = pop_es,
  #                           standardized = TRUE)
  # mm_out_std <- model_matrices_pop(ptable_std)
  # mm_out_std2 <- model_matrices_pop(model,
  #                                   pop_es = pop_es,
  #                                   standardized = TRUE)
  mm_lm_out <- mm_lm(mm_out)
  mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                              n = n,
                              number_of_indicators = number_of_indicators,
                              reliability = reliability,
                              keep_f_scores = keep_f_scores)
  fit <- lavaan::sem(model,
             data = mm_lm_dat_out,
             fixed.x = FALSE)
  tmp <- ptable
  tmp$est <- tmp$start
  fit0 <- lavaan::sem(tmp,
              do.fit = FALSE,
              fixed.x = FALSE)
  out <- list(ptable = ptable,
              # mm_out_original = mm_out_original,
              mm_out = mm_out,
              # mm_out_std = mm_out_std,
              # ptable_std = ptable_std,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out,
              # ptable_std2 = ptable_std2,
              # mm_out_std2 = mm_out_std2,
              fit = fit,
              fit0 = fit0)
  out
}

test_check_out <- function(check_out) {
  expect_equal(diag(implied_sigma(check_out$mm_out)),
              rep(1, ncol(check_out$mm_lm_dat_out)),
              tolerance = 1e-1,
              ignore_attr = TRUE)
  # expect_equal(check_out$mm_out$psi,
  #             check_out$mm_out_std$psi)
  expect_equal(diag(cov(check_out$mm_lm_dat_out)),
              rep(1, ncol(check_out$mm_lm_dat_out)),
              tolerance = 1e-1,
              ignore_attr = TRUE)
  pt <- parameterTable(check_out$fit)
  pt0 <- parameterTable(check_out$fit0)
  pt <- order_cov(pt)
  pt0 <- order_cov(pt0)
  pt$tmp <- lav_partable_labels(pt)
  pt0$tmp <- lav_partable_labels(pt0)
  tmp2 <- intersect(pt$tmp, pt0$tmp)
  expect_equal(pt$est[match(tmp2, pt$tmp)],
               pt0$est[match(tmp2, pt0$tmp)],
               tolerance = 1e-1)
  # expect_equal(coef(check_out$fit),
  #             coef(check_out$fit0)[],
  #             tolerance = 1e-1)
  # expect_equal(check_out$mm_out_std$psi,
  #             check_out$mm_out_std2$psi,
  #             tolerance = 1e-1)
  # expect_equal(check_out$ptable_std$start,
  #             check_out$ptable_std2$start,
  #             tolerance = 1e-1)
}

test_that("ptable_pop", {

# Simple mediation model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = "m",
                         "m ~ x" = "l",
                         "y ~ x" = "n")

check_out <- check_gen_dat(model_simple_med,
                           model_simple_med_es,
                           seed = 1234)

expect_equal(check_out$ptable$start[1:3],
             c(.50, .30, .00))

test_check_out(check_out)

# Parallel mediation model
# Correlated errors

model_med_parallel <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
m1 ~~ m2
"

model_med_parallel_es <- c(".beta." = "s",
                           "y ~ x" = "-m",
                           "m1 ~ x" = "l",
                           "m2 ~ x" = "s",
                           "y ~ m1" = "m",
                           "m1 ~~ m2" = "l")

check_out <- check_gen_dat(model_med_parallel,
                           model_med_parallel_es,
                           seed = 1234)

test_check_out(check_out)

# psych::describe(check_out$mm_lm_dat_out)

# Parallel mediation model
# Correlated errors not specified in the model

model_med_parallel <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
"

model_med_parallel_es <- c(".beta." = "s",
                           "y ~ x" = "-m",
                           "m1 ~~ m2" = "l")

# expect_warning(check_out <- check_gen_dat(model_med_parallel,
#                                           model_med_parallel_es,
#                                           seed = 1234))
check_out <- check_gen_dat(model_med_parallel,
                                          model_med_parallel_es,
                                          seed = 1234)
expect_equal(check_out$ptable$start[1:5],
             c(.10, .10, .10, .10, -.30))

test_check_out(check_out)

# psych::describe(check_out$mm_lm_dat_out)

# Moderation only

model_mod <-
"
y ~ x + w + x:w + c1
"

model_mod_es <- c(".beta." = "s",
                  "y ~ w" = "-m",
                  "x ~~ w" = "l")

# expect_warning(check_out <- check_gen_dat(model_mod,
#                                           model_mod_es,
#                                           seed = 1234))
check_out <- check_gen_dat(model_mod,
                           model_mod_es,
                           seed = 1234)
expect_equal(check_out$ptable$start[1:4],
             c(.10, -.30, .05, .10))
expect_equal(check_out$ptable$start[7],
             .50)

test_check_out(check_out)

dat <- check_out$mm_lm_dat_out
datz <- as.data.frame(scale(dat))

lm_out <- lm(y ~ x*w, datz)
expect_equal(coef(lm_out)[-1],
             c(.10, -.30, .05),
             tolerance = 1e-1,
             ignore_attr = TRUE)

# psych::describe(check_out$mm_lm_dat_out)

# Moderated mediation model

model_mod_med <-
"
m ~ x + w + x:w
y ~ m + z + m:z + x + u + x:u
"

model_mod_med_es <- c(".beta." = "s",
                      "y ~ x" = "-m",
                      "m ~ x:w" = "l",
                      "x ~~ w + z + u" = "s",
                      "w ~~ z + u" = "l")

check_out <- check_gen_dat(model_mod_med,
                           model_mod_med_es,
                           seed = 1234)
check_out$ptable$coeflabel <- lav_partable_labels(check_out$ptable)
expect_equal(check_out$ptable$start[1:9],
             c(.10, .10, .15, .10, .10, .05, -.30, .10, .05))
i <- match(c("w~~z", "u~~w"), check_out$ptable$coeflabel)
expect_equal(check_out$ptable$start[i],
             c(.50, .50))
i <- match(c("w~~x", "u~~x"), check_out$ptable$coeflabel)
expect_equal(check_out$ptable$start[i],
             c(.10, .10))

test_check_out(check_out)

dat <- check_out$mm_lm_dat_out
datz <- as.data.frame(scale(dat))
y ~ m + z + m:z + x + u + x:u
lm_out <- lm(y ~ m*z + x*u, datz)
expect_equal(coef(lm_out)[-1],
             c(.10, .10, -.30, .10, .05, .05),
             tolerance = 1e-1,
             ignore_attr = TRUE)

# psych::describe(check_out$mm_lm_dat_out)

# Other models

model2 <-
"
m1 ~ x + c1
m2 ~ m1 + x2 + c1
y ~  m2 + m1 + x + w + x:w + c1
"

model2_es <- c("m1 ~ x" = "-m",
               "m2 ~ m1" = "s",
               "y ~ m2" = "l",
               "y ~ x" = "m",
               "y ~ w" = "s",
               "y ~ x:w" = "s",
               "x ~~ w" = "s")

check_out <- check_gen_dat(model2,
                           model2_es,
                           seed = 1234)

test_check_out(check_out)

# psych::describe(check_out$mm_lm_dat_out)

# Simple mediation model
# Effect sizes set numerically

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = ".21",
                         "m ~ x" = ".31",
                         "y ~ x" = "n")

check_out <- check_gen_dat(model_simple_med,
                           model_simple_med_es,
                           seed = 1234)

expect_equal(check_out$ptable[1, "start"],
             .31)
expect_equal(check_out$ptable[2, "start"],
             .21)

})

#' @title Power By Sample Sizes
#'
#' @description Estimate power for a
#' set of sample sizes.
#'
#' @details
#' The function [power4test_by_n()]
#' regenerates
#' datasets for a set of sample sizes
#' and does the stored tests in each of
#' them.
#'
#' Optionally, it can also be run
#' on a object with no stored tests.
#' In this case, additional arguments
#' must be set to instruct [power4test()]
#' on the tests to be conducted.
#'
#' It is usually used to examine the
#' power over a set of sample sizes.
#'
#' @return
#' The function
#' [power4test_by_n()] returns a
#' `power4test_by_n` object, which is a
#' list of `power4test` objects, one for
#' each sample size.
#'
#' @param object A `power4test` object,
#' or a `power4test_by_n` object.
#' If it is a `power4test_by_n` object,
#' the first element, which is a
#' `power4test` object, will be used
#' as the value of this argument.
#'
#' @param n A numeric vector of sample
#' sizes for which the simulation will
#' be conducted.
#'
#' @param progress Logical. Whether
#' progress of the simulation will be
#' displayed.
#'
#' @param ... For [power4test_by_n()],
#' they are arguments to be passed
#' to [power4test()]. For [c.power4test_by_n()],
#' they are [power4test_by_n()] outputs
#' to be combined together.
#' For the `print` method of the output
#' of [power4test_by_n()], they are
#' arguments to be passed to the
#' `print` method of the output of
#' [power4test()] ([print.power4test()]).
#'
#' @param by_seed If set to a number,
#' it will be used to generate the
#' seeds for each call to [power4test()].
#' If `NULL`, the default, then seeds
#' will still be randomly generated but
#' the results cannot be easily reproduced.
#'
#' @param by_nrep If set to a number,
#' it will be used to generate the
#' number of replications (`nrep`) for
#' each call to [power4test()].
#' If set to a numeric vector of the
#' same length as `n`, then these are
#' the `nrep` values for each of the
#' calls, allowing for different numbers
#' of replications for the sample sizes.
#' If `NULL`, the default, then the
#' original `nrep` will be used.
#' This argument is used by
#' [x_from_power()] for efficiency, and
#' is rarely used when calling
#' this function directly.
#'
#' @param save_sim_all If `FALSE`,
#' the dataset in the
#' `power4test` objects are not saved,
#' to reduce
#' the size of the output. Default
#' is `TRUE`.
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_med_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = model_simple_med,
#'                        pop_es = model_simple_med_es,
#'                        n = 100,
#'                        R = 40,
#'                        ci_type = "boot",
#'                        fit_model_args = list(fit_function = "lm"),
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_indirect_effect,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         boot_ci = TRUE,
#'                                         mc_ci = FALSE))
#'
#' out <- power4test_by_n(test_out,
#'                        n = c(100, 110, 120))
#' out_reject <- rejection_rates(out)
#' out_reject
#'
#' @export
power4test_by_n <- function(object,
                            n = NULL,
                            progress = TRUE,
                            ...,
                            by_seed = NULL,
                            by_nrep = NULL,
                            save_sim_all = TRUE) {
  if (!inherits(object, "power4test") &&
      !inherits(object, "power4test_by_n")) {
    stop("Only support 'power4test' or 'power4test_by_n' objects.")
  }
  # Use the first object if the object is a power4test_by_n object:
  if (inherits(object, "power4test_by_n")) {
    object <- object[[1]]
  }
  if (!is.null(object$sim_all[[1]]$group_name)) {
    # Support multigroup models only if `n` has one value
    if (length(attr(object, "args")$n) != 1) {
      stop("For multigroup models, n_ratio must be used to specify sample sizes.")
    }
  }
  out <- list()

  # # Disable this for now
  # # Store the original object
  # n_org <- attr(object, "args")$n
  # out[[1]] <- object

  if (length(by_seed) == length(n)) {
    seeds <- by_seed
  } else {
    if (!is.null(by_seed)) {
      set.seed(by_seed)
    }
    seeds <- sample.int(99999999,
                        size = length(n))
  }

  if (length(by_nrep) == length(n)) {
    new_nrep <- by_nrep
  } else {
    if (!is.null(by_nrep)) {
      new_nrep <- rep(by_nrep, length(n))
    } else {
      new_nrep <- NULL
    }
  }
  for (i in seq_along(n)) {
    x <- n[i]
    if (progress) {
      cat("\nUpdating the simulation for sample size:",
          paste(x, collapse = ", "),
          "\n")
    }
    if (is.null(new_nrep)) {
      tmp_out <- tryCatch(power4test(
                              object = object,
                              n = x,
                              progress = progress,
                              iseed = seeds[i],
                              ...
                            ),
                            error = function(e) e)
    } else {
      tmp_out <- tryCatch(power4test(
                              object = object,
                              n = x,
                              nrep = new_nrep[i],
                              progress = progress,
                              iseed = seeds[i],
                              ...
                            ),
                            error = function(e) e)
    }

    if (inherits(tmp_out, "error")) {

      if (progress) {
        msg <- "Error in estimating power for this sample size:"
        msg <- c(msg, paste0("- ", x))
        msg <- c(msg, "The error message:")
        msg <- c(msg, paste0("- ", tmp_out$message))
        msg <- c(msg, "This sample sizes is skipped.")
        cat(c("", msg, ""), sep = "\n")
      }

      next
    }

    if (!save_sim_all) {
      # Keep the first element
      tmp_out$sim_all <- tmp_out$sim_all[1]
    }
    out[[as.character(x)]] <- tmp_out
  }
  class(out) <- c("power4test_by_n", class(out))
  out
}

#' @rdname power4test_by_n
#'
#' @param sort When combining the
#' objects, whether they will be sorted
#' by sample sizes. Default is `TRUE`.
#'
#' @param skip_checking_models Whether
#' the check of the data generation model
#' will be checked. Default is `TRUE`.
#' Should be set to `FALSE` only when
#' users are certain the they are based
#' on the same model, or when the model
#' is not saved (e.g., `save_sim_all`
#' set to `FALSE` when the objects
#' were generated).
#' This argument is used by
#' [x_from_power()] for efficiency, and
#' is rarely used when calling
#' the `c` method directly.
#'
#' @param tolerance_if_std_by_monte_carlo
#' The tolerance to be used by [all.equal()]
#' when checking population values. If
#' standardization is conducted using
#' error variances estimated by
#' Monte Carlo simulation, then they may
#' not be exactly the same across replications.
#' This is the tolerance used to compare
#' population values, to allow for
#' minor variation due to simulation.
#'
#' @return
#' The method [c.power4test_by_n()] returns
#' a `power4test_by_n` object with
#' all the elements (tests for different
#' sample sizes) combined.
#'
#' @details
#' The `c` method of `power4test_by_n`
#' objects
#' is used to combine tests from different
#' runs of [power4test_by_n()].
#' @export
c.power4test_by_n <- function(...,
                              sort = TRUE,
                              skip_checking_models = FALSE,
                              tolerance_if_std_by_monte_carlo = getOption("power4mome.tolerance_if_std_by_monte_carlo",
                                                                          default = .01)) {

  # Check whether they have the same ptable
  tmp <- list(...)
  if ((length(tmp) > 1) && !skip_checking_models) {
    ptables <- lapply(tmp,
                      \(x) {x[[1]]$sim_all[[1]]$ptable})
    std_by_monte_carlo <- sapply(
                            ptables,
                            \(x) isTRUE(attr(x, "std_by_monte_carlo"))
                          )
    std_by_monte_carlo <- any(std_by_monte_carlo)
    # Check structure
    j_start <- match("start", colnames(ptables[[1]]))
    for (i in seq_along(ptables)[-1]) {
      chk <- identical(ptables[[i]][, -j_start, drop = FALSE],
                       ptables[[i - 1]][, -j_start, drop = FALSE])
      if (!chk) {
        stop("Not all objects are based on the same model.")
      }
    }
    std_by_monte_carlo <- any(std_by_monte_carlo)
    # Check population values,
    #
    for (i in seq_along(ptables)[-1]) {
      chk <- isTRUE(all.equal(
                        ptables[[i]][, j_start, drop = TRUE],
                        ptables[[i - 1]][, j_start, drop = TRUE],
                        tolerance = ifelse(isTRUE(std_by_monte_carlo),
                                           yes = tolerance_if_std_by_monte_carlo,
                                           no = sqrt(.Machine$double.eps))
                      ))
      if (!chk) {
        stop("Not all population values are based on the same model.")
      }
    }
  }
  out <- NextMethod()
  out["sort"] <- NULL
  out["skip_checking_models"] <- NULL

  class(out) <- c("power4test_by_n", class(out))
  if (!sort) {
    return(out)
  }
  ns <- as.numeric(names(out))
  i <- order(ns)
  out <- out[i]
  class(out) <- c("power4test_by_n", class(out))
  return(out)
}

#' @noRd
is_std_by_monte_carlo <- function(object) {

}

#' @param object_by_n A `power4test_by_n`
#' object, which is an output of
#' [power4test_by_n()].
#'
#' @param all_columns If `TRUE`, all
#' columns stored by a test will be
#' printed. Default is `FALSE` and
#' only essential columns related to
#' power will be printed.
#'
#' @return
#' The function
#' returns a data frame which is
#' similar to the output of
#' [rejection_rates()], with a
#' column `n` added for the sample size
#' for each test.
#'
#' @description
#' The function
#' is used to extract the rejection
#' rates form a `power4test_by_n`
#' object, with sample sizes added to
#' the output.
#'
#' @noRd
rejection_rates_by_n <- function(object_by_n,
                                 all_columns = FALSE,
                                 nrep_if_diff = TRUE,
                                 ...) {
  tmpfct <- function(x, n) {
    out_i <- rejection_rates(x,
                             all_columns = all_columns,
                             keep_nrep = nrep_if_diff,
                             ...)
    out_i <- data.frame(n = n,
                        out_i)
    out_i
  }
  out <- mapply(tmpfct,
                x = object_by_n,
                n = as.numeric(names(object_by_n)),
                SIMPLIFY = FALSE)
  # Keep common columns
  cnames <- lapply(out,
                   \(x) tryCatch(colnames(x),
                                 error = function(e) names(x)))
  cnames1 <- Reduce(intersect,
                    cnames)
  out <- lapply(out,
                \(x) tryCatch(x[, cnames1],
                              error = function(e) x[cnames1]))
  out <- do.call(rbind,
                 out)
  if (nrep_if_diff &&
      !all_columns) {
    nrep_range <- try(suppressWarnings(diff(range(out$nrep, na.rm = TRUE))), silent = TRUE)
    if (isFALSE(nrep_range > 1)) {
      out$nrep <- NULL
    }
  }
  rownames(out) <- NULL
  out
}

#' @rdname power4test_by_n
#'
#' @param original_object The object
#' to be converted to a `power4test_by_n`
#' object.
#'
#' @return
#' The function [as.power4test_by_n()] returns
#' a `power4test_by_n` object converted
#' from the input object.
#'
#' @details
#' The function [as.power4test_by_n()]
#' is used to convert a `power4test`
#' object to a `power4test_by_n`
#' object, if it is not already one.
#' Useful when concatenating
#' `power4test` objects with
#' `power4test_by_n` objects.
#' @export
as.power4test_by_n <- function(original_object) {
  if (inherits(original_object, "power4test")) {
    x0 <- attr(original_object, "args")$n
    original_object <- list(original_object)
    class(original_object) <- c("power4test_by_n", class(original_object))
    names(original_object) <- as.character(x0)
    return(original_object)
  }
  if (inherits(original_object, "power4test_by_n")) {
    return(original_object)
  }
  stop("original_object not of a supported class.")
}

#' @rdname power4test_by_n
#'
#' @param x The object
#' to be printed.
#'
#' @param print_all If `TRUE`, all
#' elements in `x`, that is, the results
#' of all sample sizes examined, will
#' be printed. If `FALSE`, then only
#' those of the first sample size
#' will be printed.
#'
#' @return
#' The `print`-method of `power4test_by_n`
#' objects returns the object invisibly.
#' It is called for its side-effect.
#'
#' @export
print.power4test_by_n <- function(
    x,
    print_all = FALSE,
    ...
  ) {
  n_tried <- names(x)
  catwrap(
      c(paste("The sample(s) examined:",
              paste0(sort(n_tried),
                     collapse = ", ")))
      )

  if (print_all) {
    cat("\nThe output for each `power4test` analysis:\n")
    ii <- seq_along(x)
  } else {
    cat("\nThe output of the first element:\n")
    ii <- 1
  }

  for (i in ii) {
    n_i <- attr(x[[i]], "args")$n
    xx <- x[[i]]
    class(xx) <- "power4test"
    cat(header_str("",
                  hw = .84,
                  sym = "-",
                  prefix = "\n",
                  suffix = "\n",
                  sep_i = ""))
    cat(header_str(paste0("Sample Size: ", n_i),
                  hw = .8,
                  sym = "-",
                  suffix = "\n"))
    cat(header_str("",
                  hw = .84,
                  sym = "-",
                  suffix = "\n",
                  sep_i = ""))
    print(xx,
          ...)
  }
  if (!print_all) {
    cat("\n")
    catwrap(
        "Print with 'print_all = TRUE' to print all elements."
        )
  }
  invisible(x)
}

skip_on_cran()

library(testthat)

test_that("Power by n", {

mod <-
"
m ~ x
y ~ m + x
"
mod_es <-
"
y ~ x: s
.ind.x->m->y: .09
"
out <- power4test(nrep = 5,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  R = 1000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  progress = !is_testing(),
                  parallel = FALSE)
tmp <- power4test_by_n(out,
                       n = c(10, 50, -10),
                       by_nrep = 10,
                       progress = !is_testing())
expect_true(length(tmp) == 2)
expect_equal(names(tmp), c("10", "50"))
})

library(testthat)
suppressMessages(library(lavaan))

test_that("c.power4test_by_n: moderation", {

model <-
"
m ~ x + w + x:w
y ~ m + x
"

model_es <-
"
m ~ x: m
m ~ w: s
m ~ x:w: l
y ~ m: l
y ~ x: s
"

out <- power4test(nrep = 2,
                  model = model,
                  pop_es = model_es,
                  n = 50000,
                  progress = !is_testing(),
                  iseed = 1234)

test_out <- power4test(object = out,
                       test_fun = test_moderation)

test_n_out1 <- power4test_by_n(
                  test_out,
                  n = c(100, 110),
                  progress = !is_testing())
test_n_out2 <- power4test_by_n(
                  test_out,
                  n = c(200, 210),
                  progress = !is_testing())
expect_no_error(c(test_n_out1, test_n_out2))

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("Power by n: Concatenate", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 3,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       parallel = FALSE,
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

tmp <- as.power4test_by_n(sim_only)
expect_true(is.list(tmp))
expect_true(length(tmp) == 1)
expect_true(inherits(tmp, "power4test_by_n"))

out1 <- power4test_by_n(test_out,
                        n = c(100, 110),
                        by_seed = 1234,
                        progress = !is_testing())
out2 <- power4test_by_n(test_out,
                        n = c(130, 120),
                        by_seed = 1234,
                        progress = !is_testing())
out <- c(out1, out2)
out_reject <- rejection_rates(out)

tmp <- sapply(out,
              \(x) {attr(x, "args")$iseed})
out_chk <- power4test_by_n(test_out,
                           n = c(100, 110, 120, 130),
                           by_seed = tmp,
                           progress = !is_testing())
out_reject_chk <- rejection_rates(out_chk)

expect_identical(out_reject,
                 out_reject_chk)

# Test sort = FALSE

out <- c(out1, out2, sort = FALSE)
out_reject <- rejection_rates(out)
tmp <- sapply(out,
              \(x) {attr(x, "args")$iseed})
out_chk <- power4test_by_n(test_out,
                           n = c(100, 110, 130, 120),
                           by_seed = tmp,
                           progress = !is_testing())
out_reject_chk <- rejection_rates(out_chk)

expect_identical(out_reject,
                 out_reject_chk)

model_simple_med2 <-
"
m ~ x
y ~ m
"
model_simple_med_es2 <- c("y ~ m" = "l",
                          "m ~ x" = "m")
sim_only2 <- power4test(nrep = 3,
                        model = model_simple_med2,
                        pop_es = model_simple_med_es2,
                        n = 100,
                        R = 50,
                        ci_type = "boot",
                        fit_model_args = list(fit_function = "lm"),
                        do_the_test = FALSE,
                        progress = !is_testing(),
                        parallel = FALSE,
                        iseed = 1234)
test_out2 <- power4test(object = sim_only2,
                        test_fun = test_indirect_effect,
                        test_args = list(x = "x",
                                         m = "m",
                                         y = "y",
                                         boot_ci = TRUE,
                                         mc_ci = FALSE))
out3 <- power4test_by_n(test_out2,
                        n = c(105, 115),
                        progress = !is_testing())
expect_error(c(out3, out1))

})

#' @title Random Variable From an Exponential Distribution
#'
#' @description Generate random numbers
#' from an exponential distribution,
#' rescaled to have user-specified
#' population mean and standard
#' deviation.
#'
#' @details
#' First, specify the parameter,
#' `rate`, and the
#' desired population mean and standard
#' deviation. The random numbers, drawn
#' from an exponential distribution by
#' [stats::rexp()], will then be
#' rescaled with the desired population
#' mean and standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param rate `rate` for [stats::rexp()].
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @param rev If TRUE, the distribution
#' is revered to generate a negatively
#' skewed distribution. Default is
#' FALSE.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rexp_rs(n = 5000,
#'              rate = 4,
#'              pmean = 3,
#'              psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

rexp_rs <- function(n = 10,
                    rate = 1,
                    pmean = 0,
                    psd = 1,
                    rev = FALSE) {
  bmean <- 1 / rate
  bsd   <- 1 / rate
  x <- stats::rexp(n, rate)
  if (!rev) {
    x <- pmean + psd * (x - bmean) / bsd
  } else {
    x <- pmean - psd * (x - bmean) / bsd
  }
  x
}

#' @title Create a 'sim_out' Object
#'
#' @description Combine the outputs
#' of [sim_data()], [fit_model()],
#' and optionally [gen_mc()] and/or
#' [gen_boot()] to one
#' single object.
#'
#' @details
#' It merges into one object the output
#' of [sim_data()], which is a list of
#' `nrep` simulated datasets,
#' [fit_model()], which is a list of the
#' [lavaan::sem()] output for the `nrep`
#' datasets, and optionally the output
#' of [gen_mc()] or [gen_boot()], which is a list of the
#' `R` sets of Monte Carlo or bootstrap estimates
#' based on the results of
#' [fit_model()]. The list has `nrep`
#' elements, each element with the data,
#' the model fit
#' results, and optionally the Monte
#' Carlo estimates matched.
#'
#' This object can then be used for testing
#' effects of interests, which are
#' further processed to estimate the
#' power of this test.
#'
#' The function [sim_out()] is used by
#' the all-in-one function
#' [power4test()]. Users usually do not
#' call this function directly, though
#' developers can use this function to
#' develop other functions for power
#' analysis, or to build their own
#' workflows to do the power analysis.
#'
#' @seealso [power4test()]
#'
#' @return
#' The function [sim_out()] returns a `sim_out` object, which
#' is a list of length equal to the
#' length of `data_all`. Each element
#' of the list is a `sim_data` object
#' with the element `extra` added to
#' it. Other named elements will be
#' added under this name. For example.
#' the output of [fit_model()]
#' for this replication can be added
#' to `fit`, under `extra`. See
#' the description of the argument
#' `...` for details.
#'
#' @param data_all The output of
#' [sim_data()].
#'
#' @param ... Named arguments of
#' objects to be added to each
#' replication under the element
#' `extra`. For example, if set to
#' `fit = fit_all`, where `fit_all`
#' is the output of [fit_model()],
#' then `data_all[[1]]$extra$fit`
#' will be set to the first output
#' in `fit_all`.
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "m ~ x
#'  y ~ m + x"
#'
#' # Specify the population values
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate the simulated datasets
#'
#' dats <- sim_data(nrep = 5,
#'                  model = mod,
#'                  pop_es = es,
#'                  n = 100,
#'                  iseed = 1234)
#'
#' # Fit the population model to each dataset
#'
#' fits <- fit_model(dats)
#'
#' # Combine the results to one object
#'
#' sim_out_all <- sim_out(data_all = dats,
#'                        fit = fits)
#' sim_out_all
#'
#' # Verify that the elements of fits are set to extra$fit
#'
#' library(lavaan)
#' parameterEstimates(fits[[1]])
#' parameterEstimates(sim_out_all[[1]]$extra$fit)
#' parameterEstimates(fits[[2]])
#' parameterEstimates(sim_out_all[[2]]$extra$fit)
#'
#' @export
#'
sim_out <- function(data_all,
                    ...) {
  nrep <- length(data_all)
  args <- list(...)
  add_new_extra <- FALSE
  if (length(args) > 0) {
    add_new_extra <- TRUE
    args_by_rep <- split_by_rep(args)
  } else {
    # TODO:
    # - Need to decide whether NA is a
    #   good option. It will not be
    #   added anyway.
    args_by_rep <- rep(NA,
                       nrep)
  }
  tmpfct <- function(x,
                     new_extra = NULL,
                     add_new_extra = FALSE) {
    # Always work, args or not
    if (!add_new_extra) {
      return(x)
    }
    if (!is.null(x$extra)) {
      x$extra <- append(x$extra,
                        new_extra)
      x
    } else {
      x$extra <- new_extra
    }
    x
  }
  out0 <- mapply(tmpfct,
                  x = data_all,
                  new_extra = args_by_rep,
                  MoreArgs = list(add_new_extra = add_new_extra),
                  SIMPLIFY = FALSE)
  class(out0) <- c("sim_out", "sim_data",
                    setdiff(class(out0), c("sim_out", "sim_data")))
  return(out0)
}

#' @param digits The numbers of digits
#' displayed after the decimal.
#'
#' @param digits_descriptive The
#' number of digits displayed after
#' the decimal for the descriptive
#' statistics table.
#'
#' @param fit_to_all_args A named list
#' of arguments to be passed to
#' [lavaan::sem()] when the model is
#' fitted to a sample combined from
#' all samples stored.
#'
#' @param x The `sim_out` object
#' to be printed.
#'
#' @return
#' The `print` method of `sim_out`
#' returns `x` invisibly. Called for
#' its side effect.
#'
#' @rdname sim_out
#' @export
print.sim_out <- function(x,
                          digits = 3,
                          digits_descriptive = 2,
                          fit_to_all_args = list(),
                          ...) {
  NextMethod()
  print_extra(x)
  invisible(x)
}

#' @noRd

print_extra <- function(x) {
  extra <- x[[1]]$extra
  if (is.null(extra)) {
    return(NULL)
  }
  tmp <- sapply(extra,
                function(x) identical(x, NA))
  if (!any(tmp)) {
    return(NULL)
  }
  extra <- extra[!tmp]
  extra_names <- names(extra)
  cat(header_str("Extra Element(s) Found",
                 prefix = "\n",
                 suffix = "\n\n"))
  cat(paste0("- ",
             extra_names),
      sep = "\n")

  cat(header_str("Element(s) of the First Dataset",
                 hw = .5,
                 prefix = "\n",
                 suffix = "\n"))

  for (xx in extra_names) {
    tmp <- paste0("<", xx, ">")
    cat(header_str(tmp,
                   hw = .4,
                   prefix = "\n",
                   suffix = "\n\n"))
    print(extra[[xx]])
  }
  return(NULL)
}

#' @noRd
# Process optional arguments
split_by_rep <- function(args) {
  arg_names <- names(args)
  if (is.null(arg_names)) {
    stop("The additional arguments must be named.")
  }
  nrep <- sapply(args,
                 length)
  if (!all(nrep == max(nrep))) {
    tmp <- paste(arg_names, collapse = ", ")
    stop("The numbers of replications in <",
         tmp,
         "> must be the same.")
  }
  tmpfct <- function(...,
                     arg_names) {
    list(...)
  }
  out <- do.call(mapply,
                 c(list(FUN = tmpfct),
                   args,
                   list(SIMPLIFY = FALSE)))
  out
}

#' @title Test a Conditional Indirect Effect
#'
#' @description Test a conditional
#' indirect effect
#' for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing a
#' conditional
#' indirect effect, by setting it
#' to the `test_fun` argument.
#'
#' It uses [manymome::cond_indirect()]
#' to do the test. It can be used on
#' models fitted by [lavaan::sem()]
#' or fitted by a sequence of calls
#' to [stats::lm()], although only
#' nonparametric bootstrap confidence
#' interval is supported for models
#' fitted by regression using
#' [stats::lm()].
#'
#' It can also be used to test
#' a conditional effect on a direct path
#' with no mediator. Just omit `m` when
#' calling the function.
#'
#' @return
#' In its normal usage, it returns
#' a named numeric vector with the
#' following elements:
#'
#' - `est`: The mean of the estimated
#'  indirect effect across datasets.
#'
#' - `cilo` and `cihi`: The means of the
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively.
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' @inheritParams test_indirect_effect
#'
#' @param fit The fit object, to be
#' passed to [manymome::cond_indirect()].
#'
# @param x <- Inherited
#'
# @param m <- Inherited
#'
# @param y <- Inherited
#'
#' @param wvalues A numeric vector of
#' named elements. The names are the
#' variable names of the moderators,
#' and the values are the values to
#' which the moderators will be set to.
#' Default is `NULL`.
#'
# @param mc_ci <- Inherited
#'
# @param mc_out <- Inherited
#'
# @param boot_ci <- Inherited
#'
# @param boot_out <- Inherited
#'
#' @param ... Additional arguments to
#' be passed to [manymome::cond_indirect()].
#'
# @param fit_name <- Inherited
#'
# @param get_map_names <- Inherited
#'
# @param get_test_name <- Inherited
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_mod <-
#' "
#' m ~ x + w + x:w
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_mod_es <-
#' "
#' y ~ m: l
#' y ~ x: n
#' m ~ x: m
#' m ~ w: n
#' m ~ x:w: l
#' "
#'
#' # Simulate the data
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = model_simple_mod,
#'                        pop_es = model_simple_mod_es,
#'                        n = 100,
#'                        R = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the test in each replication
#'
#' test_ind <- power4test(object = sim_only,
#'                        test_fun = test_cond_indirect,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         wvalues = c(w = 1),
#'                                         mc_ci = TRUE))
#' print(test_ind,
#'       test_long = TRUE)
#'
#' @export

test_cond_indirect <- function(fit = fit,
                               x = NULL,
                               m = NULL,
                               y = NULL,
                               wvalues = NULL,
                               mc_ci = TRUE,
                               mc_out = NULL,
                               boot_ci = FALSE,
                               boot_out = NULL,
                               check_post_check = TRUE,
                               test_method = NULL,
                               ...,
                               fit_name = "fit",
                               get_map_names = FALSE,
                               get_test_name = FALSE) {

  # ==== Enable pvalue? ====

  args <- list(...)
  if (!is.null(mc_out)) {
    R <- length(mc_out)
  } else if (!is.null(boot_out)) {
    R <- length(boot_out)
  } else {
    R <- NULL
  }
  R <- args$R %||% formals(manymome::cond_indirect)$R
  ci_level <- args$level %||% formals(manymome::cond_indirect)$level
  R_bz_ok <- isTRUE(R %in% R_extrapolate(alpha = 1 - ci_level))
  bz_not_FALSE <- !isFALSE(options("power4mome.bz"))
  test_method_NULL <- is.null(test_method)
  if (R_bz_ok &&
      bz_not_FALSE &&
      test_method_NULL) {
    test_method <- "pvalue"
  } else {
    test_method <- match.arg(test_method,
                             c("ci", "pvalue"))
  }

  internal_options <- list()
  if (test_method == "pvalue") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = TRUE,
                                               pvalue_min_size = -Inf))
  }
  if (test_method == "ci") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = FALSE))
  }
  if (fit_name != "fit") {
    mc_name <- paste0(fit_name, "_mc_out")
    boot_name <- paste0(fit_name, "_boot_out")
  } else {
    mc_name <- "mc_out"
    boot_name <- "boot_out"
  }
  map_names <- c(fit = fit_name,
                 mc_out = mc_name,
                 boot_out = boot_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp_w <- cond_name(wvalues,
                       ...)
    tmp <- paste0(c(x, m, y),
                  collapse = "->")
    tmp <- paste0(tmp,
                  " (",
                  tmp_w,
                  ")")
    args <- as.list(match.call())
    tmp2 <- character(0)
    if (isTRUE(args$standardized_x) && !isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' standardized)")
    }
    if (!isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('y' standardized)")
    }
    if (isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' and 'y' standardized)")
    }
    return(paste0("test_cond_indirect: ", tmp, collapse = ""))
  }
  if (boot_ci) mc_ci <- FALSE
  if (inherits(fit, "lavaan")) {
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
    ngroups <- lavaan::lavInspect(fit, "ngroups")
  } else {
    fit_ok <- TRUE
    ngroups <- 1
  }
  if (fit_ok) {
    out <- tryCatch(manymome::cond_indirect(
                                 x = x,
                                 y = y,
                                 m = m,
                                 wvalues = wvalues,
                                 fit = fit,
                                 mc_ci = mc_ci,
                                 mc_out = mc_out,
                                 boot_ci = boot_ci,
                                 boot_out = boot_out,
                                 progress = FALSE,
                                 ...),
                    error = function(e) e)
  } else {
    out <- NA
  }
  if (inherits(out, "error") ||
      identical(out, NA)) {
    out2 <- c(est = as.numeric(NA),
              cilo = as.numeric(NA),
              cihi = as.numeric(NA),
              sig = as.numeric(NA),
              pvalue = as.numeric(NA))
    return(out2)
  }
  if (test_method == "ci") {
    ci0 <- stats::confint(out)
    out1 <- ifelse((ci0[1, 1] > 0) || (ci0[1, 2] < 0),
                    yes = 1,
                    no = 0)
  }
  if (test_method == "pvalue") {
    ci0 <- matrix(
              as.numeric(NA),
              nrow = 1,
              ncol = 2
            )
    bz_alpha_ok <- isTRUE(all.equal(1 - out$level,
                                    getOption("power4mome.bz.alpha",
                                                        default = .05)))
    if (bz_alpha_ok) {
      boot_est <- out$boot_indirect %||% out$mc_indirect
      boot_sig <- bz_sig_partition(
                    boot_est,
                    alpha = 1 - out$level
                  )
    } else {
      boot_sig <- NULL
    }
    if (isTRUE(boot_ci)) {
      out1a <- out$boot_p %||% as.numeric(NA)
      R <- length(out$boot_indirect)
      nlt0 <- sum(as.numeric(out$boot_indirect < 0))
    } else if (isTRUE(mc_ci)) {
      out1a <- out$mc_p %||% as.numeric(NA)
      R <- length(out$mc_indirect)
      nlt0 <- sum(as.numeric(out$mc_indirect < 0))
    } else {
      out1a <- as.numeric(NA)
      R <- as.numeric(NA)
      nlt0 <- as.numeric(NA)
    }
    out1 <- ifelse(
                out1a < (1 - out$level),
                yes = 1,
                no = 0
              )
  }
  out2 <- c(est = unname(stats::coef(out)),
            cilo = ci0[1, 1],
            cihi = ci0[1, 2],
            sig = out1,
            pvalue = out$mc_p %||% (out$boot_p %||% as.numeric(NA)))
  if (test_method == "pvalue") {
    # For Boos & Zhang (2000)
    out2 <- c(out2, R = R, nlt0 = nlt0,
              alpha = 1 - out$level,
              boot_sig)
  }
  return(out2)
}

#' @noRd

cond_name <- function(wvalues,
                      ...) {
  if (!is.null(wvalues)) {
    w1 <- names(wvalues)
    out <- paste0(w1, " = ", wvalues)
    out <- paste(out,
                 collapse = "; ")
  } else {
    out <- character(0)
  }
  args <- list(...)
  if (!is.null(args$group)) {
    out2 <- args$group
  } else {
    out2 <- character(0)
  }
  out3 <- paste0(out, out2, collapse = ";")
  return(out3)
}

# Functions used in test

#' @noRd
test_summary <- function(object) {
  if (inherits(object, "power4test")) {
    object <- object$test_all
  }
  out <- sapply(object,
                test_summary_i,
                simplify = FALSE)
  out
}

#' @noRd
test_summary_i <- function(object) {
  test_i <- object[[1]]$test_results
  if (is.vector(test_i)) {
    # A vector
    out <- test_summary_i_vector_one_test(object)
    return(out)
  }
  if (isTRUE(length(dim(test_i)) == 2)) {
    # A table, likely a data frame
    out <- test_summary_i_data_frame(object)
    return(out)
  }
  stop("test_results not supported. Something's wrong.")
}

#' @noRd
test_summary_i_vector_one_test <- function(object) {
  test_results_all <- sapply(object,
                             function(xx) xx$test_results)
  test_results_all <- as.data.frame(t(test_results_all),
                                    check.names = FALSE)
  colMeans(test_results_all, na.rm = TRUE)
}

#' @noRd
test_summary_i_data_frame <- function(object) {
  test_i <- object[[1]]$test_results
  test_label <- attr(test_i, "test_label")
  i <- sapply(test_i,
              is.numeric)
  out0 <- sapply(test_i[, "test_label", drop = TRUE],
                 function(xx) {
                   t(sapply(object,
                            function(yy) {
                              tmp <- yy$test_results
                              unlist(tmp[tmp[, test_label] == xx, i])
                            },
                            simplify = TRUE))
                 },
                 simplify = FALSE)
  out1 <- t(sapply(out0,
                   colMeans,
                   na.rm = TRUE))
  out2 <- test_i
  out2[, i] <- out1
  class(out2) <- class(test_i)
  out2
}

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test: With update", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

test_par <- function(object,
                     par,
                     alpha = .05) {
  est <- lavaan::parameterEstimates(object)
  if (par %in% est$label) {
    i <- (est$lhs == par) &
         (est$label == par)
    i <- which(i)
  } else {
    par1 <- lavaan::lavParseModelString(par,
                                        as.data.frame. = TRUE)
    i <- (est$lhs == par1$lhs) &
        (est$op == par1$op) &
        (est$rhs == par1$rhs)
    i <- which(i)
  }
  out <- c(est = est[i, "est"],
           cilo = est[i, "ci.lower"],
           cihi = est[i, "ci.upper"],
           sig = as.numeric(est[i, "pvalue"] < alpha))
  out
}

par_results <- function(object) {
  object
}

# Generate the data

power_all_sim_only <- power4test(nrep = 3,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 50,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 progress = !is_testing(),
                                 parallel = FALSE,
                                 iseed = 1234)

# Update the model and regenerate the data

power_all_sim_only_new_es <- power4test(object = power_all_sim_only,
                                        pop_es = c("y ~ m" = "s"))

expect_equal(power_all_sim_only_new_es$sim_all[[3]]$ptable[2, "start"],
             .10)
expect_equal(power_all_sim_only$sim_all[[3]]$ptable[2, "start"],
             .50)

# Update the model and regenerate the data

power_all_sim_only_new_n <- power4test(object = power_all_sim_only,
                                       n = 55)
expect_equal(nrow(power_all_sim_only_new_n$sim_all[[3]]$mm_lm_dat_out),
             55)
expect_equal(nrow(power_all_sim_only$sim_all[[3]]$mm_lm_dat_out),
             50)

# Update the model and n and regenerate the data

power_all_sim_only_new_n_pop_es <- power4test(object = power_all_sim_only,
                                              n = 60,
                                              pop_es = c("y ~ m" = "n"))
expect_equal(power_all_sim_only_new_n_pop_es$sim_all[[3]]$ptable[2, "start"],
             .00)
expect_equal(nrow(power_all_sim_only_new_n_pop_es$sim_all[[3]]$mm_lm_dat_out),
             60)

# Indirect effect

ind_results <- function(out) {
  ci0 <- stats::confint(out)
  out1 <- ifelse((ci0[1, 1] > 0) || (ci0[1, 2] < 0),
                  yes = 1,
                  no = 0)
  out2 <- c(est = unname(coef(out)),
            cilo = ci0[1, 1],
            cihi = ci0[1, 2],
            sig = out1)
  return(out2)
}

# Do the test
# - Need only the arguments for the test.

power_all_test_only <- power4test(object = power_all_sim_only,
                                  test_fun = test_par,
                                  test_args = list(par = "y~m"),
                                  map_names = c(object = "fit"))

# power_all_test_only <- power4test(object = power_all_sim_only,
#                                   test_fun = manymome::indirect_effect,
#                                   test_args = list(x = "x",
#                                                    m = "m",
#                                                    y = "y",
#                                                    mc_ci = TRUE),
#                                   map_names = c(fit = "fit",
#                                                 mc_out = "mc_out"),
#                                   results_fun = ind_results)

expect_true(identical(power_all_test_only$sim_all,
                      power_all_sim_only$sim_all))

(chk1 <- test_summary(power_all_test_only))

chk2 <- mean(sapply(power_all_test_only$sim_all,
                    function(x) {coef(x$extra$fit)["b"]}))
expect_equal(chk1[[1]]["est"],
             chk2,
             ignore_attr = TRUE)


# Update the model, regenerate the data, and redo the test

power_all_test_only_new_es <- power4test(object = power_all_test_only,
                                         pop_es = c("y ~ m" = "s"))
expect_false(identical(power_all_test_only$sim_all,
                      power_all_test_only_new_es$sim_all))
expect_equal(power_all_test_only$sim_all[[3]]$ptable[2, "start"],
             .50)
expect_equal(power_all_test_only_new_es$sim_all[[3]]$ptable[2, "start"],
             .10)

expect_false(identical(test_summary(power_all_test_only_new_es),
                       test_summary(power_all_test_only)))

(chk1 <- test_summary(power_all_test_only_new_es))

chk2 <- mean(sapply(power_all_test_only_new_es$sim_all,
                    function(x) {coef(x$extra$fit)["b"]}))
expect_equal(chk1[[1]]["est"],
             chk2,
             ignore_attr = TRUE)

# Update the sample size, regenerate the data, and redo the test

power_all_test_only_new_n <- power4test(object = power_all_test_only,
                                        n = 200)
expect_false(identical(test_summary(power_all_test_only_new_n),
                       test_summary(power_all_test_only)))

# Update the sample size, regenerate the data, and redo the test

power_all_test_only_new_n_pop_es <- power4test(object = power_all_test_only,
                                               pop_es = c("y ~ m" = "s"),
                                               n = 200)
expect_false(identical(test_summary(power_all_test_only_new_n_pop_es),
                       test_summary(power_all_test_only)))

# Add a test

power_all_test_add_an_effect <- power4test(object = power_all_test_only,
                                           test_fun = test_par,
                                           test_args = list(par = "y~x"),
                                           map_names = c(object = "fit"),
                                           test_name = "y~x")

expect_true(length(test_summary(power_all_test_add_an_effect)) == 2)

# Update the sample size, regenerate the data, and redo two tests

power_all_test_add_an_effect_new_n <- power4test(object = power_all_test_add_an_effect,
                                                 pop_es = c("y ~ m" = "s"),
                                                 n = 100)

chk1 <- test_summary(power_all_test_add_an_effect)
chk2 <- test_summary(power_all_test_add_an_effect_new_n)

expect_false(identical(chk1,
                       chk2))

(chk1 <- test_summary(power_all_test_add_an_effect_new_n))

chk2 <- mean(sapply(power_all_test_add_an_effect_new_n$sim_all,
                    function(x) {coef(x$extra$fit)["y~x"]}))
expect_equal(chk1[[2]]["est"],
             chk2,
             ignore_attr = TRUE)

# Update the nrep

power_all_test_only_new_nrep <- power4test(object = power_all_test_only,
                                           nrep = 5)
expect_equal(length(power_all_test_only_new_nrep$sim_all),
             5)
expect_equal(length(power_all_test_only_new_nrep$test_all[[1]]),
             5)
expect_equal(length(power_all_test_only$sim_all),
             3)
expect_equal(length(power_all_test_only$test_all[[1]]),
             3)

power_all_sim_only_5 <- power4test(nrep = 5,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 50,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 progress = !is_testing())

power_all_test_only_5 <- power4test(object = power_all_sim_only_5,
                                  test_fun = test_par,
                                  test_args = list(par = "y~m"),
                                  map_names = c(object = "fit"))

expect_identical(power_all_sim_only_5$sim_all[[1]]$mm_lm_dat_out[1:10, ],
                 power_all_test_only_new_nrep$sim_all[[1]]$mm_lm_dat_out[1:10, ])

expect_identical(power_all_test_only_5$test_all[[1]][[5]]$test_results,
                 power_all_test_only_new_nrep$test_all[[1]][[5]]$test_results)

})

library(testthat)

test_that("indirect effects: p.adjust", {

mod <-
"
m1 ~ a1*x
m2 ~ a2*x
m3 ~ a3*x
m4 ~ a4*x
y ~ b1*m1 + b2*m2 + b3*m3 + b4*m4 + x
a1b1 := a1 * b1
a2b2 := a2 * b2
"

mod_es <- c("y ~ m1" = "l",
            "m1 ~ x" = "m",
            "y ~ m2" = "s",
            "m2 ~ x" = "l",
            "y ~ m3" = "n",
            "m3 ~ x" = "n",
            "y ~ m4" = "n",
            "m4 ~ x" = "n",
            "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       fit_model_args = list(estimator = "ML"),
                       R = 100,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        p_adjust_method = "BH"),
                       iseed = 2345)
(chk <- test_summary(test_ind))

p_org <- test_ind$test_all[[1]][[1]]$test_results$pvalue_org
p_adj <- test_ind$test_all[[1]][[1]]$test_results$pvalue

expect_equal(p.adjust(p_org, method = "BH"),
             p_adj)

})

library(testthat)

test_that("All parameters: lm", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

test_out$test_all[[1]][[2]]

(chk <- test_summary(test_out))
names(chk)

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(pars = "y~m"))

(chk <- test_summary(test_out))
names(chk)
expect_true(length(chk) == 2)

fits <- lapply(sim_only$sim_all,
               function(x) x$extra$fit)
chk_outs <- sapply(fits,
                   function(x) {
                     confint(x[[1]])[2, 1]
                   })
expect_equal(chk[[1]]$cilo[1],
             mean(chk_outs))

test_out_summary <- summarize_tests(test_out)

})

library(testthat)

test_that("All parameters", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

test_out$test_all[[1]][[2]]

(chk <- test_summary(test_out))
names(chk)

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(standardized = TRUE))

(chk <- test_summary(test_out))
names(chk)
expect_true(length(chk) == 2)

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(op = "~",
                                        pars = "a"))

(chk <- test_summary(test_out))
names(chk)
expect_true(length(chk) == 3)

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(op = ":="))

(chk <- test_summary(test_out))
names(chk)
expect_true(length(chk) == 4)

fits <- lapply(sim_only$sim_all,
               function(x) x$extra$fit)
chk_outs <- sapply(fits,
                   function(x) {
                     est <- lavaan::parameterEstimates(x,
                                                       remove.nonfree = TRUE)
                     (est$ci.lower > 0) | (est$ci.upper < 0)
                   })
chk_sigs <- rowMeans(chk_outs,
                     na.rm = TRUE)
expect_identical(chk[[1]]$sig,
                 chk_sigs)

chk_outs <- sapply(fits,
                   function(x) {
                     lavaan::standardizedSolution(x)$est.std
                   })
chk_est <- rowMeans(chk_outs,
                    na.rm = TRUE)
expect_identical(chk[[2]]$est,
                 chk_est)

expect_true(all(chk[[3]]$op %in% "~"))

expect_identical(chk[[4]]$op,
                 ":=")

test_out_summary <- summarize_tests(test_out)

})

library(testthat)

test_that("cond indirect effect", {

mod <-
"
m ~ x + w1 + x:w1
y ~ m + w2 + m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w2 = 1, w1 = 0),
                                        mc_ci = TRUE))

(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

test_indb <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w2 = 1, w1 = 0),
                                        mc_ci = TRUE,
                                        test_method = "pvalue"))

(chkb <- test_summary(test_indb))
expect_equal(chk[[1]]["sig"],
             chkb[[1]]["sig"])


test_ind2 <- power4test(object = test_ind,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "x",
                                         y = "m",
                                         wvalues = c(w1 = 0),
                                         mc_ci = TRUE))

(chk <- test_summary(test_ind2))
expect_true(length(chk) == 2)

test_ind2b <- power4test(object = test_indb,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "x",
                                         y = "m",
                                         wvalues = c(w1 = 0),
                                         mc_ci = TRUE,
                                         test_method = "pvalue"))
(chkb <- test_summary(test_ind2b))
expect_equal(sapply(chk, \(x) x["sig"]),
             sapply(chkb, \(x) x["sig"]))

test_ind3 <- power4test(object = test_ind2,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "m",
                                         y = "y",
                                         wvalues = c(w2 = 0),
                                         mc_ci = TRUE))

(chk <- test_summary(test_ind3))
expect_true(length(chk) == 3)

test_ind3b <- power4test(object = test_ind2b,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "m",
                                         y = "y",
                                         wvalues = c(w2 = 0),
                                         mc_ci = TRUE,
                                         test_method = "pvalue"))
(chkb <- test_summary(test_ind3b))
expect_equal(sapply(chk, \(x) x["sig"]),
             sapply(chkb, \(x) x["sig"]))

# test_ind4 <- suppressWarnings(power4test(object = test_ind3,
#                                          test_fun = test_indirect_effect,
#                                          test_args = list(x = "x",
#                                                            m = "m",
#                                                            y = "y",
#                                                            mc_ci = TRUE,
#                                                            standardized_x = TRUE,
#                                                            standardized_y = TRUE)))

# (chk <- test_summary(test_ind4))
# expect_true(length(chk) == 4)

# test_ind4_summary <- summarize_tests(test_ind4$test_all)

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("cond indirect effects", {

mod <-
"
m ~ x + w1 + x:w1
y ~ m + w2 + m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect_effects,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wlevels = c("w2", "w1"),
                                        boot_ci = TRUE))

(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

test_ind2 <- power4test(object = test_ind,
                        test_fun = test_cond_indirect_effects,
                        test_args = list(x = "x",
                                         y = "m",
                                         wlevels = c("w1"),
                                         boot_ci = TRUE))

(chk <- test_summary(test_ind2))
expect_true(length(chk) == 2)

test_ind3 <- power4test(object = test_ind2,
                        test_fun = test_cond_indirect_effects,
                        test_args = list(x = "m",
                                         y = "y",
                                         wlevels = "w2",
                                         boot_ci = TRUE))

(chk <- test_summary(test_ind3))
expect_true(length(chk) == 3)

# test_ind4 <- suppressWarnings(power4test(object = test_ind3,
#                                          test_fun = test_indirect_effect,
#                                          test_args = list(x = "x",
#                                                            m = "m",
#                                                            y = "y",
#                                                            mc_ci = TRUE,
#                                                            standardized_x = TRUE,
#                                                            standardized_y = TRUE)))

# (chk <- test_summary(test_ind4))
# expect_true(length(chk) == 4)

# test_ind4_summary <- summarize_tests(test_ind4$test_all)

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("cond indirect effect", {

mod <-
"
m ~ x + w1 + x:w1
y ~ m + w2 + m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w2 = 1, w1 = 0),
                                        boot_ci = TRUE))

(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

test_ind2 <- power4test(object = test_ind,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "x",
                                         y = "m",
                                         wvalues = c(w1 = 0),
                                         boot_ci = TRUE))

(chk <- test_summary(test_ind2))
expect_true(length(chk) == 2)

test_ind3 <- power4test(object = test_ind2,
                        test_fun = test_cond_indirect,
                        test_args = list(x = "m",
                                         y = "y",
                                         wvalues = c(w2 = 0),
                                         boot_ci = TRUE))

(chk <- test_summary(test_ind3))
expect_true(length(chk) == 3)

# test_ind4 <- suppressWarnings(power4test(object = test_ind3,
#                                          test_fun = test_indirect_effect,
#                                          test_args = list(x = "x",
#                                                            m = "m",
#                                                            y = "y",
#                                                            mc_ci = TRUE,
#                                                            standardized_x = TRUE,
#                                                            standardized_y = TRUE)))

# (chk <- test_summary(test_ind4))
# expect_true(length(chk) == 4)

# test_ind4_summary <- summarize_tests(test_ind4$test_all)

})

library(testthat)

skip_if_not_installed("lmhelprs")

test_that("indirect effect by lm", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE),
                       progress = !is_testing())

(chk <- test_summary(test_ind))

fits <- lapply(sim_only$sim_all,
               function(x) x$extra$fit)
boot_out <- lapply(sim_only$sim_all,
                   function(x) x$extra$boot_out)
chk_outs <- mapply(manymome::indirect_effect,
                   fit = fits,
                   boot_out = boot_out,
                   MoreArgs = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   boot_ci = TRUE),
                  SIMPLIFY = FALSE)
chk_cis <- lapply(chk_outs,
                  confint)
chk_cis <- do.call(rbind, chk_cis)
chk_sigs <- (chk_cis[, 1] > 0) | (chk_cis[, 2] < 0)

expect_equal(chk[[1]]["sig"],
             mean(chk_sigs),
             ignore_attr = TRUE)
expect_equal(chk[[1]][c("cilo", "cihi")],
             colMeans(chk_cis),
             ignore_attr = TRUE)

})

library(testthat)
suppressMessages(library(lavaan))

# All-In-One

test_that("power4test", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

test_par <- function(object,
                     par,
                     alpha = .05) {
  est <- lavaan::parameterEstimates(object)
  if (par %in% est$label) {
    i <- (est$lhs == par) &
         (est$label == par)
    i <- which(i)
  } else {
    par1 <- lavaan::lavParseModelString(par,
                                        as.data.frame. = TRUE)
    i <- (est$lhs == par1$lhs) &
        (est$op == par1$op) &
        (est$rhs == par1$rhs)
    i <- which(i)
  }
  out <- c(est = est[i, "est"],
           cilo = est[i, "ci.lower"],
           cihi = est[i, "ci.upper"],
           sig = as.numeric(est[i, "pvalue"] < alpha))
  out
}

par_results <- function(object) {
  object
}

# Generate the data

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 100,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 R = 50,
                                 do_the_test = FALSE,
                                 progress = !is_testing(),
                                 iseed = 1234)

# Indirect effect

ind_results <- function(out) {
  ci0 <- stats::confint(out)
  out1 <- ifelse((ci0[1, 1] > 0) || (ci0[1, 2] < 0),
                  yes = 1,
                  no = 0)
  out2 <- c(est = unname(coef(out)),
            cilo = ci0[1, 1],
            cihi = ci0[1, 2],
            sig = out1)
  return(out2)
}

# Do the test
# - Need only the arguments for the test.

power_all_test_only <- power4test(object = power_all_sim_only,
                                  test_fun = manymome::indirect_effect,
                                  test_args = list(x = "x",
                                                   m = "m",
                                                   y = "y",
                                                   mc_ci = TRUE),
                                  map_names = c(fit = "fit",
                                                mc_out = "mc_out"),
                                  results_fun = ind_results)

summary_all <- test_summary(power_all_test_only)
summary_all

fits <- lapply(power_all_sim_only$sim_all,
               function(x) x$extra$fit)
mc_outs <- lapply(power_all_sim_only$sim_all,
                  function(x) x$extra$mc_out)
ind_outs <- mapply(manymome::indirect_effect,
                   fit = fits,
                   mc_out = mc_outs,
                   MoreArgs = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  SIMPLIFY = FALSE)
ind_cis <- t(sapply(ind_outs,
                    confint))
ind_sigs <- (ind_cis[, 1] > 0) | (ind_cis[, 2] < 0)

expect_equal(summary_all$`manymome::indirect_effect`["sig"],
             mean(ind_sigs),
             ignore_attr = TRUE)
expect_equal(summary_all$`manymome::indirect_effect`[c("cilo", "cihi")],
             colMeans(ind_cis),
             ignore_attr = TRUE)

# Do the test and add it to the power4test object
# - Need only the arguments for the test.

power_all_test_only2 <- power4test(object = power_all_test_only,
                                   test_fun = manymome::indirect_effect,
                                   test_args = list(x = "x",
                                                    y = "y",
                                                    mc_ci = TRUE),
                                   map_names = c(fit = "fit",
                                                 mc_out = "mc_out"),
                                   test_name = "Direct Effect",
                                   test_note = "y ~ x",
                                   results_fun = ind_results)

summary_all <- test_summary(power_all_test_only2)
summary_all

ind_outs <- mapply(manymome::indirect_effect,
                   fit = fits,
                   mc_out = mc_outs,
                   MoreArgs = list(x = "x",
                                   y = "y",
                                   mc_ci = TRUE),
                  SIMPLIFY = FALSE)
ind_cis <- t(sapply(ind_outs,
                    confint))
ind_sigs <- (ind_cis[, 1] > 0) | (ind_cis[, 2] < 0)

expect_equal(summary_all$`Direct Effect`["sig"],
             mean(ind_sigs),
             ignore_attr = TRUE)
expect_equal(summary_all$`Direct Effect`[c("cilo", "cihi")],
             colMeans(ind_cis),
             ignore_attr = TRUE)

# Do the test and replace the previous tests in the power4test object
# - Need only the arguments for the test.

power_all_test_only3 <- power4test(object = power_all_test_only2,
                                   test_fun = manymome::indirect_effect,
                                   test_args = list(x = "x",
                                                    m = "m",
                                                    y = "y",
                                                    mc_ci = TRUE),
                                   test_note = "Updated indirect effect",
                                   map_names = c(fit = "fit",
                                                 mc_out = "mc_out"),
                                   results_fun = ind_results)

expect_equal(length(power_all_test_only3$test_all),
             2)

summary_all <- test_summary(power_all_test_only3)
summary_all

fits <- lapply(power_all_sim_only$sim_all,
               function(x) x$extra$fit)
mc_outs <- lapply(power_all_sim_only$sim_all,
                  function(x) x$extra$mc_out)
ind_outs <- mapply(manymome::indirect_effect,
                   fit = fits,
                   mc_out = mc_outs,
                   MoreArgs = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  SIMPLIFY = FALSE)
ind_cis <- t(sapply(ind_outs,
                    confint))
ind_sigs <- (ind_cis[, 1] > 0) | (ind_cis[, 2] < 0)

expect_equal(summary_all$`manymome::indirect_effect`["sig"],
             mean(ind_sigs),
             ignore_attr = TRUE)
expect_equal(summary_all$`manymome::indirect_effect`[c("cilo", "cihi")],
             colMeans(ind_cis),
             ignore_attr = TRUE)

})

skip_on_cran()

library(testthat)

test_that("indirect effect: lavaan", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       R = 100,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_ind <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        mc_ci = TRUE),
                       progress = !is_testing())

(chk <- test_summary(test_ind))
expect_true(length(chk) == 1)

# # This test is sensitive the RNG. Need to revise
# test_indb <- power4test(object = sim_only,
#                        test_fun = test_indirect_effect,
#                        test_args = list(x = "x",
#                                         m = "m",
#                                         y = "y",
#                                         mc_ci = TRUE,
#                                         test_method = "pvalue"),
#                        progress = !is_testing())
# expect_equal(test_summary(test_ind)[[1]]["sig"],
#              test_summary(test_indb)[[1]]["sig"])

test_ind2 <- power4test(object = test_ind,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        y = "m",
                                        mc_ci = TRUE),
                       progress = !is_testing())

(chk <- test_summary(test_ind2))
expect_true(length(chk) == 2)

# # This test is sensitive the RNG. Need to revise
# test_ind2b <- power4test(object = test_ind,
#                        test_fun = test_indirect_effect,
#                        test_args = list(x = "x",
#                                         y = "m",
#                                         mc_ci = TRUE,
#                                         test_method = "pvalue"),
#                        progress = !is_testing())
# expect_equal(test_summary(test_ind2)[[1]]["sig"],
#              test_summary(test_ind2b)[[1]]["sig"])

test_ind3 <- power4test(object = test_ind2,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "m",
                                        y = "y",
                                        mc_ci = TRUE),
                       progress = !is_testing())

(chk <- test_summary(test_ind3))
expect_true(length(chk) == 3)

# # This test is sensitive the RNG. Need to revise
# test_ind3b <- power4test(object = test_ind2,
#                        test_fun = test_indirect_effect,
#                        test_args = list(x = "m",
#                                         y = "y",
#                                         mc_ci = TRUE,
#                                         test_method = "pvalue"),
#                        progress = !is_testing())
# expect_equal(test_summary(test_ind3)[[1]]["sig"],
#              test_summary(test_ind3b)[[1]]["sig"])

})

library(testthat)

test_that("indirect effects", {

mod <-
"
m1 ~ a1*x
m2 ~ a2*x
y ~ b1*m1 + b2*m2 + x
a1b1 := a1 * b1
a2b2 := a2 * b2
"

mod_es <- c("y ~ m1" = "l",
            "m1 ~ x" = "m",
            "y ~ m2" = "s",
            "m2 ~ x" = "l",
            "y ~ x" = "n")
k <- c(y = 3,
       m1 = 3,
       m2 = 3,
       x = 3)
rel <- c(y = .70,
         m1 = .70,
         m2 = .70,
         x = .70)

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       R = 100,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_tot <- power4test(object = sim_only,
                       progress = !is_testing(),
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        omnibus = "total"))

(chk <- test_summary(test_tot))
expect_true(length(chk) == 1)

test_ind <- power4test(object = sim_only,
                       progress = !is_testing(),
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE))

chk1 <- test_tot$test_all[[1]][[1]]$test_results$est
chk2 <- test_ind$test_all[[1]][[1]]$test_results$est

expect_equal(chk1,
             sum(chk2))

})

library(testthat)

test_that("Test moderation: lm: Multiline", {

mod <-
"
m ~ x + w1 +
 x:w1
y ~ m +
  w2 +
     m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_moderation)

(chk <- test_summary(test_out))

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(pars = c("m~x:w1",
                                                 "y~m:w2")))
(chk <- test_summary(test_out))

expect_equal(chk[[1]]$sig,
             chk[[2]]$sig)

# Standardized

expect_error(test_out <- power4test(object = sim_only,
                       test_fun = test_moderation,
                       test_args = list(standardized = TRUE),
                       progress = !is_testing()))

})

library(testthat)

test_that("Test moderation: lm", {

mod <-
"
m ~ x + w1 + x:w1
y ~ m + w2 + m:w2 + x
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_moderation)

(chk <- test_summary(test_out))

test_out <- power4test(object = test_out,
                       test_fun = test_parameters,
                       test_args = list(pars = c("m~x:w1",
                                                 "y~m:w2")))
(chk <- test_summary(test_out))

expect_equal(chk[[1]]$sig,
             chk[[2]]$sig)

# Standardized

expect_error(test_out <- power4test(object = sim_only,
                       test_fun = test_moderation,
                       test_args = list(standardized = TRUE),
                       progress = !is_testing()))

})

library(testthat)

test_that("index of mome", {

mod <-
"
m ~ x + w + x:w
y ~ m
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w" = "n",
            "m ~ x:w" = "l")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       parallel = FALSE,
                       iseed = 1234)

test_out1 <- power4test(object = sim_only,
                         test_fun = test_index_of_mome,
                         test_args = list(x = "x",
                                          m = "m",
                                          y = "y",
                                          w = "w",
                                          mc_ci = TRUE))

(chk <- test_summary(test_out1))


test_out1b <- power4test(object = sim_only,
                         test_fun = test_index_of_mome,
                         test_args = list(x = "x",
                                          m = "m",
                                          y = "y",
                                          w = "w",
                                          mc_ci = TRUE,
                                          test_method = "pvalue"))

(chkb <- test_summary(test_out1b))
expect_equal(chk[[1]]["sig"],
             chkb[[1]]["sig"])

fits <- lapply(sim_only$sim_all,
               function(x) x$extra$fit)
mc_outs <- lapply(sim_only$sim_all,
                  function(x) x$extra$mc_out)
chk_outs <- mapply(manymome::index_of_mome,
                   fit = fits,
                   mc_out = mc_outs,
                   MoreArgs = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   w = "w",
                                   mc_ci = TRUE),
                  SIMPLIFY = FALSE)
chk_cis <- lapply(chk_outs,
                  confint)
chk_cis <- do.call(rbind, chk_cis)
chk_sigs <- (chk_cis[, 1] > 0) | (chk_cis[, 2] < 0)

expect_equal(chk[[1]]["sig"],
             mean(chk_sigs),
             ignore_attr = TRUE)
expect_equal(chk[[1]][c("cilo", "cihi")],
             colMeans(chk_cis),
             ignore_attr = TRUE)

expect_no_error(capture.output(print(test_out1)))

})

#' @title Summarize Test Results
#'
#' @description Extract and summarize
#' test results.
#'
#' @details
#' The function [summarize_tests()]
#' is used to extract
#' information from each test stored
#' in a `power4test` object.
#'
#' The method `print.test_out_list()` is
#' used to print the content of a list
#' of test stored in a `power4test`
#' object, with the option to print
#' just the names of tests.
#'
#' # The role of `summarize_tests()` and related functions
#'
#' The function [summarize_tests()] and
#' related print methods are used by
#' the all-in-one function
#' [power4test()] and its `summary`
#' method. Users usually do not
#' call them directly, though
#' developers can use this function to
#' develop other functions for power
#' analysis, or to build their own
#' workflows to do the power analysis.
#'
#' @return
#' The function [summarize_tests()] returns
#' a list of the class `test_summary_list`.
#' Each element contains a summary of a
#' test stored. The elements are of
#' the class `test_summary`, with
#' these elements:
#'
#' - `test_attributes`: The stored
#'  information of a test, for printing.
#'
#' - `nrep`: The number of datasets
#'  (replications).
#'
#' - `mean`: The means of numeric
#'  information. For significance
#'  tests, these are the rejection
#'  rates.
#'
#' - `nvalid`: The number of non-`NA`
#'  replications used to compute each
#'  mean.
#'
#' @param object A `power4test` object
#' or the element `test_all` in
#' a `power4test` object.
#'
#' @param collapse Whether a single
#' decision (significant vs. not significant)
#' is made across all tests for a test
#' that consists of several tests
#' (e.g., the tests of several parameters).
#' If `"none"`, tests will be summarized
#' individually. If `"all_sig"`, then
#' the set of tests is considered significant
#' if all individual tests are significant.
#' If `"at_least_one_sig"`, then the set of
#' tests is considered significant if
#' at least one of the tests is significant.
#' If `"at_least_k_sig"`, then the set of
#' tests is considered significant if
#' at least `k` tests are significant,
#' `k` set by the argument `at_least_k`.
#'
#' @param at_least_k Used by `collapse`,
#' the number of tests required to be
#' significant for the set of tests to
#' be considered significant.
#'
#' @param merge_all_tests If `TRUE`, all
#' the tests in each replication will be
#' merged into one test.
#'
#' @param p_adjust_method The method to be
#' passed to [p.adjust()] to adjust the
#' *p*-values when testing the effects.
#' Default is `"none"` and the *p*-values
#' will not be adjusted. The unadjusted
#' *p*-values will be stored in the column
#' `pvalue_org`. Ignored if some tests
#' do not have *p*-values stored.
#' NOTE: Use this only if all tests
#' can be  conducted using *p*-values.

#'
#' @param alpha The level of significance
#' to use when using `p_adjust_method`.
#' The significance results (the column
#' `sig`) will be updated using the
#' adjusted *p*-values. Used only if
#' `p_adjust_method` is not `"none"`.
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Simulated datasets
#'
#' sim_only <- power4test(nrep = 2,
#'                        model = mod,
#'                        pop_es = es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Test the parameters in each dataset
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters)
#'
#' # Print the summary
#'
#' summarize_tests(test_out)
#'
#' @export
summarize_tests <- function(object,
                            collapse = c("none",
                                         "all_sig",
                                         "at_least_one_sig",
                                         "at_least_k_sig"),
                            at_least_k = 1,
                            merge_all_tests = FALSE,
                            p_adjust_method = "none",
                            alpha = .05) {
  collapse <- match.arg(collapse)
  if (inherits(object, "power4test")) {
    object <- object$test_all
  }
  if (length(object) == 1) {
    # - merge_all_tests set to FALSE if there is only one test
    merge_all_tests <- FALSE
  }
  if (merge_all_tests) {
    object <- list(all_tests_merged = collapse_all_tests(
                      object,
                      alpha = alpha,
                      p_adjust_method = p_adjust_method
                    ))
  }
  out <- sapply(object,
                summarize_test_i,
                collapse = collapse,
                at_least_k = at_least_k,
                simplify = FALSE)
  class(out) <- c("test_summary_list", class(out))
  out
}

#' @rdname summarize_tests
#'
#' @param x The object to be printed.
#'
#' @param digits The numbers of digits
#' after the decimal when printing
#' numeric results.
#'
#' @param ... Optional arguments.
#' Not used.
#'
#' @return
#' The `print` methods returns `x` invisibly.
#' They are called for their side
#' effects.
#'
#' @export
print.test_summary_list <- function(x,
                                    digits = 3,
                                    ...) {
  for (xx in x) {
    print(xx,
          digits = digits,
          ...)
  }
  invisible(x)
}

#' @rdname summarize_tests
#'
#' @export
print.test_summary <- function(x,
                               digits = 2,
                               ...) {
  x0 <- unclass(x)
  nvalid <- x0$nvalid
  test_attr <- x0$test_attributes
  test_name <- test_attr$test_name
  cat(header_str(paste0("<", test_name, ">"),
                 hw = .7,
                 prefix = "\n",
                 suffix = "\n\n"))
  if (is.vector(x0$mean)) {
    results_type <- "vector"
  } else {
    results_type <- "data.frame"
  }
  is_merged_tests <- isTRUE(!is.null(test_attr$tests))
  if (results_type == "vector") {
    nvalid_sig <- unname(nvalid["sig"])
    propvalid_sig <- unname(nvalid_sig / x0$nrep)
    cat("Mean(s) across replication:\n")
    out0 <- formatC(x0$mean,
                    digits = digits,
                    format = "f")
    print(as.data.frame(rbind(out0),
                        check.names = FALSE),
          row.names = FALSE)
    cat("\n")
    cat("- The value 'sig' is the rejection rate.\n")
    cat("- If the null hypothesis is false, this is the power.\n")
    cat("- Number of valid replications for rejection rate:",
        nvalid_sig, "\n")
    cat("- Proportion of valid replications for rejection rate:",
        formatC(propvalid_sig, digits = digits, format = "f"),
        "\n")
  }
  if (results_type == "data.frame") {
    nvalid_sig <- unname(nvalid[, "sig", drop = TRUE])
    propvalid_sig <- unname(nvalid_sig / x0$nrep)
    one_nvalid <- all(nvalid_sig == max(nvalid_sig))
    cat("Mean(s) across replication:\n")
    df_means <- x0$mean
    if (!one_nvalid) {
      df_means$p_sig <- propvalid_sig
    }
    df_means <- format_num_cols(df_means,
                                digits = digits)
    print(df_means)
    cat("\n")
    cat("- The column 'sig' shows the rejection rates.\n")
    cat("- If the null hypothesis is false, the rate is the power.\n")
    if (one_nvalid) {
      cat("- Number of valid replications for rejection rate(s):",
          max(nvalid_sig), "\n")
      cat("- Proportion of valid replications for rejection rate(s):",
          formatC(max(propvalid_sig), digits = digits, format = "f"),
          "\n")
    } else {
      cat(strwrap(paste0("- The column 'p_sig' shows proportions of ",
                        "valid replications for rejection rates."),
                  exdent = 2), sep = "\n")
    }
    if (is_merged_tests) {
      cat(strwrap(paste0("- Test(s) merged:"),
                  exdent = 2), sep = "\n")
      cat(paste(" -", test_attr$tests),
          sep = "\n")
    }
  }
  invisible(x)
}

#' @rdname summarize_tests
#'
#' @param test_long If `TRUE`, a detailed
#' report will be printed.
#'
#' @export
print.test_out_list <- function(x,
                                digits = 3,
                                test_long = TRUE,
                                ...) {
  # TODO:
  # - Retrieve and print test notes.
  if (!test_long) {
    test_names <- names(x)
    cat(header_str("Test(s) Conducted",
                  hw = .8,
                  prefix = "\n",
                  suffix = "\n\n"))
    cat(paste0("- ", test_names),
        sep = "\n")
    cat("\n")
    cat("Call print() and set 'test_long = TRUE' for a detailed report.\n")
  } else {
    out <- summarize_tests(x)
    print(out,
          digits = digits,
          ...)
  }
  invisible(x)
}

#' @noRd
format_num_cols <- function(object,
                            digits = digits) {
  tmp <- as.data.frame(object,
                       check.names = FALSE)
  tmp2 <- lapply(tmp,
                  function(x,
                            digits) {
                    if (is.numeric(x)) {
                      x <- formatC(x,
                                    digits = digits,
                                    format = "f")
                      return(x)
                    } else {
                      return(x)
                    }
                  },
                  digits = digits)
  tmp2 <- as.data.frame(tmp2,
                        check.names = FALSE)
  class(tmp2) <- class(object)
  tmp2
}

#' @noRd
summarize_test_i <- function(x,
                             collapse = collapse,
                             at_least_k = at_least_k) {
  test_i <- x[[1]]$test_results
  if (is.vector(test_i)) {
    # A vector
    out <- summarize_one_test_vector(x)
    return(out)
  }
  if (isTRUE(length(dim(test_i)) == 2)) {
    # A table, likely a data frame
    out <- summarize_one_test_data_frame(x,
                                         collapse = collapse,
                                         at_least_k = at_least_k)
    return(out)
  }
  stop("test_results not supported. Something's wrong.")
}

#' @noRd
summarize_one_test_vector <- function(x) {
  test_results_all <- lapply(x,
                             function(xx) xx$test_results)
  has_R <- !is.null(test_results_all[[1]]["R"])
  if (has_R) {
    R <- unname(test_results_all[[1]]["R"])
    Rext <- R_extrapolate()
    R_case <- bz_case(R)
    bz_alpha_ok <- isTRUE(all.equal(unname(test_results_all[[1]]["alpha"]),
                                    getOption("power4mome.bz.alpha",
                                                        default = .05)))
    do_bz <- (R_case != "") &&
             getOption("power4mome.bz", default = TRUE) &&
             bz_alpha_ok
  } else {
    R <- NULL
    do_bz <- FALSE
    R_case <- ""
  }
  if (do_bz) {
    if (R_case == "one") {
      Rk <- Rext[seq(1, which(Rext == R) - 1)]
      test_results_all <- lapply(
              test_results_all,
              add_rr_ext,
              R = R,
              Rk = Rk
            )
    }
  }
  nrep <- length(test_results_all)
  i_na <- sapply(
            test_results_all,
            \(x) all(is.na(x))
          )
  test_results_all_not_na <- test_results_all[!i_na]
  test_results_all_not_na <- do.call(rbind,
                              test_results_all_not_na)
  test_results_all_not_na <- as.data.frame(test_results_all_not_na,
                                    check.names = FALSE)
  if (any(i_na)) {
    # Fill NA rows
    tmp1 <- sum(i_na)
    tmp2 <- test_results_all_not_na[rep(1, tmp1), , drop = FALSE]
    tmp2[] <- NA
    tmp3 <- rbind(test_results_all_not_na,
                  tmp2)
    rownames(tmp3) <- NULL
    test_results_all <- tmp3
  } else {
    test_results_all <- test_results_all_not_na
  }
  test_means <- colMeans(test_results_all, na.rm = TRUE)
  if (do_bz) {
    tmp <- bz_rr(test_means)
    test_means["sig"] <- as.numeric(tmp)
    bz_model <- attr(
                  tmp,
                  "bz_model"
                )
  } else {
    bz_model <- NULL
  }
  test_not_na <- apply(test_results_all,
                       2,
                       function(x) {sum(!is.na(x))})
  out1 <- list(test_attributes = attributes(x),
               nrep = nrep,
               mean = test_means,
               nvalid = test_not_na)
  # May add other attributes in the future
  attr(out1,
      "extra") <- list(bz_extrapolated = do_bz,
                       R_case = R_case,
                       bz_model = bz_model)
  class(out1) <- c("test_summary", class(out1))
  out1
}

#' @noRd
summarize_one_test_data_frame <- function(x,
                                          collapse = c("none",
                                                       "all_sig",
                                                       "at_least_one_sig",
                                                       "at_least_k_sig"),
                                          at_least_k = 1) {
  collapse <- match.arg(collapse)
  nrep <- length(x)
  test_i <- x[[1]]$test_results
  test_label <- attr(test_i, "test_label")
  i <- sapply(test_i,
              is.numeric)
  i_names <- colnames(test_i)[i]
  # Identify failed iteration
  no_test_label <- sapply(
          x,
          \(xx) !(test_label %in% colnames(xx$test_results))
         )
  x_ok <- x[!no_test_label]
  out0 <- sapply(test_i[, "test_label", drop = TRUE],
                 function(xx) {
                   t(sapply(x_ok,
                            function(yy) {
                              tmp <- yy$test_results
                              unlist(tmp[tmp[, test_label] == xx, i])
                            },
                            simplify = TRUE))
                 },
                 simplify = FALSE)

  if (any(no_test_label)) {
    # Fill NA rows
    for (i in seq_along(out0)) {
      j <- sum(no_test_label)
      tmp <- out0[[i]]
      tmp2 <- tmp[rep(1, j), ]
      tmp2[] <- NA
      tmp3 <- rbind(tmp, tmp2)
      out0[[i]] <- tmp3
    }
  }

  do_bz <- FALSE
  has_R <- FALSE
  R_case <- ""
  bz_model <- NULL

  has_R <- "R" %in% colnames(out0[[1]])
  if (has_R) {
    R <- sapply(out0,
                \(x) x[, "R"],
                simplify = FALSE)
    R <- unname(unlist(R))
    if (all(R != R[1])) {
      # Not all R equal. Do not do Boos-Zhang
      do_bz <- FALSE
    } else {
      R <- R[1]
      Rext <- R_extrapolate()
      R_case <- bz_case(R)
      bz_alpha_ok <- isTRUE(all.equal(tryCatch(unname(out0[[1]][, "alpha"])[1],
                                               error = function(e) NA),
                                      getOption("power4mome.bz.alpha",
                                                          default = .05)))
      do_bz <- (R_case != "") &&
                getOption("power4mome.bz", default = TRUE) &&
                bz_alpha_ok
    }
  } else {
    R <- NULL
    do_bz <- FALSE
    R_case <- ""
  }

  if ((length(out0) == 1) ||
      (collapse == "none")) {
    if (do_bz) {
      if (R_case == "one") {
        for (j1 in seq_along(out0)) {
          tmp0 <- out0[[j1]]
          # bz_* may have been added by the test function.
          # Add bz_* only if not available.
          if (!(paste0("bz_", R) %in% colnames(tmp0))) {
            tmp2 <- add_bz_i(tmp0)
            out0[[j1]] <- tmp2
          }
        }
      }
    }
    out1 <- t(sapply(out0,
                    colMeans,
                    na.rm = TRUE))
    if (do_bz) {
      bz_model <- as.list(seq_len(nrow(out1)))
      for (j1 in seq_len(nrow(out1))) {
        tmp <- bz_rr(out1[j1, , drop = TRUE])
        bz_model[[j1]] <- attr(tmp,
                               "bz_model")
        out1[j1, "sig"] <- tmp
      }
    } else {
      bz_model <- NULL
    }
    test_not_na <- t(sapply(out0,
                      function(x) {
                        apply(x,
                              2,
                              function(xx) {sum(!is.na(xx))})
                      }))
    test_means <- test_i
    test_means[, i_names] <- out1[, i_names, drop = FALSE]
  } else {
    out1a <- out0[[1]]
    out1a[] <- as.numeric(NA)
    sig0 <- sapply(out0,
                    function(xx) xx[, "sig", drop = TRUE],
                    simplify = TRUE)
    if (collapse == "all_sig") {
      # Boos-Zhang method "all_sig" only
      if (isTRUE(R_case %in% c("one", "cum"))) {
        out1_bz1 <- merge_for_collapse(out0)
        if (!any(grepl("bz_", colnames(out1_bz1[[1]])))) {
          # Add bz_* if not present
          out1_bz1 <- lapply(
                        out1_bz1,
                        function(x) {
                          a <- add_bz_i(x)
                          i <- grepl("bz_", colnames(a))
                          a[, i]
                        }
                      )
        } else {
          out1_bz1 <- lapply(
                        out1_bz1,
                        function(x) {
                          i <- grepl("bz_", colnames(x))
                          x[, i]
                        }
                      )
        }
        # Always have bz_*
        out1_bz2 <- lapply(
                      out1_bz1,
                      \(x) apply(x,
                                 MARGIN = 2,
                                 min))
        out1_bz2 <- do.call(rbind,
                            out1_bz2)
        i <- colnames(out1_bz2)
        if (all(i %in% colnames(out1a))) {
          out1a[, i] <- out1_bz2
        } else {
          out1a <- cbind(out1a, out1_bz2)
        }
      } else {
        sig1 <- apply(sig0,
                      MARGIN = 1,
                      function(xx) as.numeric(all(xx > 0)),
                      simplify = TRUE)
        out1a[, "sig"] <- sig1
      }
    }
    if (collapse == "at_least_one_sig") {
      sig1 <- apply(sig0,
                    MARGIN = 1,
                    function(xx) as.numeric(any(xx > 0)),
                    simplify = TRUE)
      out1a[, "sig"] <- sig1
    }
    if (collapse == "at_least_k_sig") {
      sig1 <- apply(sig0,
                    MARGIN = 1,
                    function(xx) as.numeric(isTRUE(sum(xx > 0, na.rm = TRUE) >= at_least_k)),
                    simplify = TRUE)
      out1a[, "sig"] <- sig1
    }
    out1b <- colMeans(out1a, na.rm = TRUE)
    out1 <- out1a[1, , drop = FALSE]
    out1[] <- out1b
    if (do_bz) {
      bz_model <- as.list(seq_len(nrow(out1)))
      for (j1 in seq_len(nrow(out1))) {
        tmp <- bz_rr(out1[j1, , drop = TRUE])
        bz_model[[j1]] <- attr(tmp,
                               "bz_model")
        out1[j1, "sig"] <- tmp
      }
    } else {
      bz_model <- NULL
    }
    test_not_na <- out1a[1, , drop = FALSE]
    test_not_na[] <- sum(apply(sig0,
                       MARGIN = 1,
                       function(xx) as.numeric(all(!is.na(xx)))))
    test_means <- test_i[1 , ]
    test_means[] <- NA
    if (collapse == "at_least_k_sig") {
      tmp <- paste0("at_least_", at_least_k, "_sig")
    } else {
      tmp <- collapse
    }
    test_means[1, "test_label"] <- tmp
    test_means[, i_names] <- out1[, i_names, drop = FALSE]
  }
  class(test_means) <- class(test_i)
  out1 <- list(test_attributes = attributes(x),
               nrep = nrep,
               mean = test_means,
               nvalid = test_not_na)
  # May add other attributes in the future
  attr(out1,
      "extra") <- list(bz_extrapolated = do_bz,
                       R_case = R_case,
                       bz_model = bz_model)
  class(out1) <- c("test_summary", class(out1))
  out1
}

#' @title Plot The Results of 'x_from_power'
#'
#' @description It plots the results
#' of 'x_from_power', such as the
#' estimated power against sample size.
#'
#' @details
#' The `plot` method of `x_from_power`
#' objects currently plots the relation
#' between estimated power and
#' the values examined by [x_from_power()].
#' Other elements
#' can be requested (see the argument
#' `what`), and they can be customized
#' individually.
#'
#' @return
#' The `plot`-method of `x_from_power`
#' returns `x` invisibly.
#' It is called for its side effect.
#'
#' @param x An `x_from_power` object,
#' the output of [x_from_power()].
#'
#' @param what A character vector of
#' what to include in the
#' plot. Possible values are
#' `"ci"` (confidence intervals
#' for the estimated value of the
#' predictor),
#' `"power_curve"` (the crude power
#' curve, if available), `"final_x"`
#' (a vertical line for
#' the value of the predictor with estimated
#' power close enough to the target
#' power by confidence interval),
#' `"final_power"` (a horizontal line
#' for the estimated power of the
#' final value of the predictor),
#' `"target_power"` (a horizontal
#' line for the target power),
#' and `"sig_area"` (the area significantly
#' higher or lower than the target
#' power, if `goal` is `"close_enough"`
#' and `what` is `"lb"` or `"ub"`).
#' By default, all these elements will
#' be plotted.
#'
#' @param text_what A character vector
#' of what numbers to be added as
#' labels. Possible values are
#' `"final_x"` (the value of the
#' predictor with
#' estimated power close enough to
#' the target power by confidence interval)
#' `"final_power"` (the estimated
#' power of the final value of the
#' predictor), and
#' `"sig_area"` (labeling the area significantly
#' higher or lower than the target
#' power, if `goal` is `"close_enough"`
#' and `what` is `"lb"` or `"ub"`).
#' By default, all these labels will
#' be added.
#'
#' @param digits The number of digits
#' after the decimal that will be used
#' when adding numbers.
#'
#' @param main The title of the plot.
#' If `NULL`, it will be generated
#' automatically.
#'
#' @param xlab,ylab The labels for the
#' horizontal and vertical axes,
#' respectively.
#'
#' @param ci_level The level of
#' confidence of the confidence intervals,
#' if requested. Default is `.95`, denoting
#' 95%.
#'
#' @param pars_ci A named list of
#' arguments to be passed to [arrows()]
#' to customize the drawing of the
#' confidence intervals.
#'
#' @param pars_power_curve A named list of
#' arguments to be passed to [points()]
#' to customize the drawing of the
#' power curve.
#'
#' @param pars_ci_final_x A named list of
#' arguments to be passed to [arrows()]
#' to customize the drawing of the
#' confidence interval of the final
#' value of the predictor.
#'
#' @param pars_target_power A named list
#' of arguments to be passed to [abline()]
#' when drawing the horizontal line
#' for the target power.
#'
#' @param pars_final_x A
#' named list of arguments to be passed
#' to [abline()] when drawing the
#' vertical line for the final value
#' of the predictor.
#'
#' @param pars_final_power A
#' named list of arguments to be passed
#' to [abline()] when drawing the
#' horizontal line for the estimated
#' power at the final value of the
#' predictor.
#'
#' @param pars_text_final_x A
#' named list of arguments to be passed
#' to [text()] when adding the
#' label for the final value of the
#' predictor.
#'
#' @param pars_text_final_power A
#' named list of arguments to be passed
#' to [text()] when adding the
#' label for the estimated power
#' of final value of the predictor.
#'
#' @param pars_sig_area A named list
#' of arguments to be passed to
#' [rect()] when shading the area
#' significantly higher or lower than
#' the target power.
#'
#' @param pars_text_sig_area A named list
#' of arguments to be passed to
#' [text()] when labelling the area
#' significantly higher or lower than
#' the target power.
#'
#' @param prop_of_trials The proportion
#' of trials to be included in the plot.
#' If `NULL`, it will be determined
#' based on the algorithm used.
#'
#' @param min_trials_for_prop The minimum
#' number of trials for `prop_of_trials`
#' to be used.
#' If `NULL`, it will be determined
#' based on the algorithm used.
#'
#' @param override_for_pba If `TRUE`,
#' the default, the values of
#' some arguments will
#' be overriden internally if the
#' algorithm used is `"probabilistic_bisection"`,
#' to make the plot suitable for
#' this algorithm. For example, the
#' power curve and the confidence intervals
#' for trials other than the solution
#' will not be plotted, even if requested.
#' If `FALSE`, then argument values
#' will be not be changed internally.
#'
#' @param ... Optional arguments.
#' Passed to [plot()] when drawing
#' the estimated power against the
#' predictor.
#'
#' @seealso [x_from_power()]
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' # Generate the datasets
#'
#' sim_only <- power4test(nrep = 10,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do a test
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters,
#'                        test_args = list(pars = "m~x"))
#'
#' # Determine the sample size with a power of .80 (default)
#'
#' power_vs_n <- x_from_power(test_out,
#'                            x = "n",
#'                            progress = TRUE,
#'                            target_power = .80,
#'                            final_nrep = 10,
#'                            max_trials = 1,
#'                            seed = 2345)
#' plot(power_vs_n)
#'
#' @importFrom graphics abline arrows par points text title rect
#' @importFrom grDevices adjustcolor

#'
#' @export
plot.x_from_power <- function(x,
                              what = c("ci", "power_curve", "final_x", "final_power", "target_power",
                                       switch(x$x, n = "sig_area", es = NULL)),
                              text_what = c("final_x", "final_power",
                                            switch(x$x, n = "sig_area", es = NULL)),
                              digits = 3,
                              main = NULL,
                              xlab = NULL,
                              ylab = "Estimated Power",
                              ci_level = .95,
                              pars_ci = list(),
                              pars_power_curve = list(),
                              pars_ci_final_x = list(lwd = 3,
                                                     length = .2,
                                                     col = "blue"),
                              pars_target_power = list(lty = "dashed",
                                                       lwd = 2,
                                                       col = "black"),
                              pars_final_x = list(lty = "dotted"),
                              pars_final_power = list(lty = "dotted", col = "blue"),
                              pars_text_final_x = list(y = 0, pos = 3, cex = 1),
                              pars_text_final_power = list(pos = 3, cex = 1),
                              pars_sig_area = list(col = adjustcolor("lightblue",
                                                   alpha.f = .1)),
                              pars_text_sig_area = list(cex = 1),
                              prop_of_trials = NULL,
                              min_trials_for_prop = NULL,
                              override_for_pba = TRUE,
                              ...) {

  if (is.null(main)) {
    if (x$algorithm == "probabilistic_bisection") {
      is_pba <- TRUE
      main <- paste0("Search History ",
                     "(Target Power: ",
                     formatC(x$target_power, digits = digits, format = "f"),
                     ")")
   } else {
      is_pba <- FALSE
      override_for_pba <- FALSE
      main <- paste0("Power Curve ",
                     "(Target Power: ",
                     formatC(x$target_power, digits = digits, format = "f"),
                     ")")
   }
  }

  if (is_pba &&
      override_for_pba) {
    prop_of_trials <- prop_of_trials %||% .95
    min_trials_for_prop <- min_trials_for_prop %||% 20
  }

  if (is_pba &&
      override_for_pba) {
    what <- setdiff(what, c("power_curve"))
  }

  args <- list(...)

  what <- match.arg(what, several.ok = TRUE)
  text_what <- match.arg(text_what, several.ok = TRUE)
  predictor <- x$x

  # Set xlab

  if (is.null(xlab)) {
    if (predictor == "n") {
      xlab <- "Sample Size"
    }
    if (predictor == "es") {
      xlab <- paste0("Parameter: ",
                     x$pop_es_name)
    }
  }

  # Was a solution found?

  solution_found <- isFALSE(identical(NA, x$x_final))

  # === Set the xlim? ===

  if (is.null(args$xlim) &&
      !is.null(prop_of_trials)) {
    base_xlim <- NULL
    tmp <- as.data.frame(rejection_rates(x$power4test_trials))
    tmp1 <- tmp[, predictor]
    if (min_trials_for_prop <= nrow(tmp)) {
      # # tmp should have already been sorted
      # tmp <- tmp[order(tmp[, predictor, drop = TRUE]), ]
      tmp2 <- stats::ecdf(tmp1)
      i0 <- (1 - prop_of_trials) / 2
      i1 <- 1 - i0 * 2
      if (solution_found) {
        ix <- x$x_final
        p_final <- tmp2(ix)
        if ((p_final < i0) ||
            (p_final > i1)) {
        }
      }
      base_xlim <- stats::quantile(
                      tmp2,
                      unname(c(i0, i1))
                    )
    }
  } else {
    base_xlim <- args$xlim
  }

  if (is.null(args$lwd) &&
      override_for_pba &&
      is_pba) {
    base_lwd <- .5
    base_lty <- "dotted"
  } else {
    base_lwd <- args$lwd
    base_lty <- NULL
  }

  # === Draw the base plot: Power vs. Predictor

  # It is intended *not* to use plot.power_curve().
  # It is possible that the fit failed.

  args_tmp <- args
  args_tmp$xlim <- NULL
  do.call(plot_power_x,
          c(list(object = x$power4test_trials,
               predictor = predictor,
               main = main,
               xlab = xlab,
               ylab = ylab,
               xlim = base_xlim,
               lwd = base_lwd,
               lty = base_lty),
            args_tmp))

  # === Add CIs?

  if ("ci" %in% what) {

    if (solution_found && ("final_x" %in% what)) {
      # A solution was found and final_x line is to be drawn.
      # Draw the final_x CI separately

      # Draw the CI for the final x
      tmp <- x$power4test_trials[x$i_final]
      class(tmp) <- class(x$power4test_trials)
      tmp_args <- utils::modifyList(pars_ci_final_x,
                                    list(object = tmp,
                                         predictor = predictor))
      do.call(plot_power_x_ci,
              tmp_args)
      # Drop final_n from the CI lists
      tmp_for_ci <- x$power4test_trials[-x$i_final]
      class(tmp_for_ci) <- class(x$power4test_trials)
    } else {
      # Keep all CIs in the same way
      tmp_for_ci <- x$power4test_trials
    }

    # Draw the other CIs
    if (!override_for_pba ||
        !is_pba) {
      tmp_args <- utils::modifyList(pars_ci,
                                    list(object = tmp_for_ci,
                                        predictor = predictor))
      do.call(plot_power_x_ci,
              tmp_args)
    }
  }

  # === Draw the power curve?

  # It is intentional *not* to use the power curve plot method.

  if ("power_curve" %in% what) {
    tmp_args <- utils::modifyList(pars_power_curve,
                                  list(object = x$power4test_trials,
                                       predictor = predictor,
                                       power_x_fit = x$power_curve))
    do.call(plot_power_curve_x,
            tmp_args)
  }

  # === Draw a horizontal line for the target power?

  if ("target_power" %in% what) {
    tmp_args <- utils::modifyList(pars_target_power,
                                  list(h = x$target_power))
    do.call(abline,
            tmp_args)
  }

  if (solution_found) {

    # === Draw a vertical line for the final N?

    if ("final_x" %in% what) {
      tmp_args <- utils::modifyList(pars_final_x,
                                    list(v = x$x_final))
      do.call(abline,
              tmp_args)
    }

    # === Draw a horizontal line for the final power?

    if ("final_power" %in% what) {
      tmp_args <- utils::modifyList(pars_final_power,
                                    list(h = x$power_final))
      do.call(abline,
              tmp_args)
    }

    # === Add a label for the final x?

    if ("final_x" %in% text_what) {
      x_final_str <- formatC(x$x_final,
                             digits = switch(predictor,
                                             n = 0,
                                             es = digits),
                             format = "f")
      tmp_args <- utils::modifyList(pars_text_final_x,
                                    list(x = x$x_final,
                                         labels = x_final_str))
      do.call(text,
              tmp_args)
    }

    if (("sig_area" %in% what) &&
        (x$goal == "close_enough")) {
      if (x$what == "lb") {
        tmp_args <- utils::modifyList(pars_sig_area,
                                      list(
                                        xleft = x$x_final,
                                        xright = par("usr")[2],
                                        ybottom = par("usr")[3],
                                        ytop = par("usr")[4],
                                        border = NULL)
                                      )
        do.call(rect,
                tmp_args)
      }
      if (x$what == "ub") {
        tmp_args <- utils::modifyList(pars_sig_area,
                                      list(
                                        xleft = par("usr")[1],
                                        xright = x$x_final,
                                        ybottom = par("usr")[3],
                                        ytop = par("usr")[4],
                                        border = NULL)
                                      )
        do.call(rect,
                tmp_args)
      }
    }

    if (("sig_area" %in% text_what) &&
        (x$goal == "close_enough")) {
      if (x$what == "lb") {
        tmp_str <- paste0("Power sig. >\n",
                           formatC(x$target_power,
                                   digits,
                                   format = "f"))
        tmp_args <- utils::modifyList(pars_text_sig_area,
                                      list(x = mean(c(x$x_final,
                                                      par("usr")[2])),
                                           y = mean(c(x$target_power,
                                                      par("usr")[3])),
                                           labels = tmp_str))
        do.call(text,
                tmp_args)
      }
      if (x$what == "ub") {
        tmp_str <- paste0("Power sig. <\n",
                           formatC(x$target_power,
                                   digits,
                                   format = "f"))
        tmp_args <- utils::modifyList(pars_text_sig_area,
                                      list(x = mean(c(x$x_final,
                                                      par("usr")[1])),
                                           y = mean(c(x$target_power,
                                                      par("usr")[3])),
                                           labels = tmp_str))
        do.call(text,
                tmp_args)
      }
    }

    # === Add a label for the final power?

    if ("final_power" %in% text_what) {
      tmp <- par("usr")
      tmp_args <- utils::modifyList(pars_text_final_power,
                                    list(y = x$power_final,
                                         x = tmp[1] + (tmp[2] - tmp[1]) * .05,
                                         labels = formatC(x$power_final,
                                                          digits = digits,
                                                          format = "f")))
      if (!is.null(pars_text_final_power$x)) {
        tmp_args <- utils::modifyList(tmp_args,
                                      list(x = pars_text_final_power$x))
      }
      do.call(text,
              tmp_args)
    }

  }

  invisible(x)
}

#' @noRd
plot_power_x <- function(object,
                         predictor,
                         type = "l",
                         ylim = c(0, 1),
                         ci_level = .95,
                         ...) {
  # No need for other arguments. They should be stored in object.
  reject0 <- rejection_rates_add_ci(object,
                                    level = ci_level)
  x <- switch(predictor,
              n = "n",
              es = "es")
  reject0$x <- reject0[[x]]
  reject0$power <- reject0$reject
  plot(power ~ x,
       data = reject0,
       type = type,
       ylim = ylim,
       ...)
}

#' @noRd
plot_power_x_ci <- function(object,
                            predictor,
                            length = .1,
                            angle = 90,
                            code = 3,
                            col = "grey50",
                            ci_level = .95,
                            ...) {
  # No need for other arguments. They should be stored in object.
  reject0 <- rejection_rates_add_ci(object,
                                    level = ci_level)
  x <- switch(predictor,
              n = "n",
              es = "es")
  reject0$x <- reject0[[x]]
  reject0$power <- reject0$reject

  # Some CIs may be of zero width
  i <- !(reject0$reject_ci_lo == reject0$reject_ci_hi)
  if (any(i)) {
    arrows(x0 = reject0$x[i],
           y0 = reject0$reject_ci_lo[i],
           x1 = reject0$x[i],
           y1 = reject0$reject_ci_hi[i],
           length = length,
           angle = angle,
           code = code,
           col = col,
           ...)
  }
}

#' @noRd

#' @noRd

plot_power_curve_x <- function(object,
                               predictor,
                               power_x_fit,
                               type = "l",
                               lwd = 2,
                               col = "red",
                               length_of_new_x = 20,
                               ci_level = .95,
                               ...) {
  # power_x_fit is a power_curve object
  # No need for other arguments. They should be stored in object.
  reject0 <- rejection_rates_add_ci(object,
                                    level = ci_level)
  x <- switch(predictor,
              n = "n",
              es = "es")
  reject0$x <- reject0[[x]]
  reject0$power <- reject0$reject
  x_new <- seq(min(reject0$x),
               max(reject0$x),
               length.out = length_of_new_x)
  if (inherits(power_x_fit$fit, "nls") || inherits(power_x_fit$fit, "lm")) {
    y_new <- stats::predict(power_x_fit,
                            newdata = list(x = x_new))
    points(x = x_new,
           y = y_new,
           type = type,
           lwd = lwd,
           col = col,
           ...)
  }
}

#' @rdname plot.x_from_power
#'
#' @param pars_text_final_x_lower,pars_text_final_x_upper
#' If two values of the predictor are to
#' be printed, these are the named list
#' of the arguments to be passed to [text()]
#' when adding the labels for these two
#' values.
#'
#' @details
#' The `plot`-method for
#' `n_region_from_power` objects is
#' a modified version of the `plot`-method
#' for `x_from_power`. It plots the
#' results of two runs of [n_from_power()]
#' in one plot. It is otherwise similar
#' to the `plot`-method for `x_from_power`.
#'
#' @return
#' The `plot`-method of `n_region_from_power`
#' returns `x` invisibly.
#' It is called for its side effect.
#'
#' @export
plot.n_region_from_power <- function(x,
                              what = c("ci", "power_curve", "final_x", "final_power", "target_power",
                                       "sig_area"),
                              text_what = c("final_x", "final_power",
                                            "sig_area"),
                              digits = 3,
                              main = paste0("Power Curve ",
                                            "(Target Power: ",
                                            formatC(x$below$target_power, digits = digits, format = "f"),
                                            ")"),
                              xlab = NULL,
                              ylab = "Estimated Power",
                              ci_level = .95,
                              pars_ci = list(),
                              pars_power_curve = list(),
                              pars_ci_final_x = list(lwd = 2,
                                                     length = .2,
                                                     col = "blue"),
                              pars_target_power = list(lty = "dashed",
                                                       lwd = 2,
                                                       col = "black"),
                              pars_final_x = list(lty = "dotted"),
                              pars_final_power = list(lty = "dotted", col = "blue"),
                              pars_text_final_x = list(pos = 3, cex = 1),
                              pars_text_final_x_lower = pars_text_final_x,
                              pars_text_final_x_upper = pars_text_final_x,
                              pars_text_final_power = list(cex = 1),
                              pars_sig_area = list(col = adjustcolor("lightblue",
                                                   alpha.f = .1)),
                              pars_text_sig_area = list(cex = 1),
                              ...) {

  what <- match.arg(what, several.ok = TRUE)
  text_what <- match.arg(text_what, several.ok = TRUE)
  predictor <- "n"

  # Set xlab

  xlab <- "Sample Size"

  # Was a solution found?
  solution_found_below <- isFALSE(identical(NA, x$below$x_final))
  solution_found_above <- isFALSE(identical(NA, x$above$x_final))
  solution_found <- solution_found_below || solution_found_above

  # === Draw the base plot: Power vs. Predictor

  # It is intended *not* to use plot.power_curve().
  # It is possible that the fit failed.

  # No need for other arguments. They should be stored in object.
  a1 <- rejection_rates(
              x$below$power4test_trials,
              all_columns = TRUE
            )
  a2 <- rejection_rates(
              x$above$power4test_trials,
              all_columns = TRUE
            )
  a <- rbind(a1, a2)
  i <- order(a$n)
  a <- a[i, ]
  a <- a[!duplicated(a$n), ]

  n_min <- min(a$n)
  n_max <- max(a$n)
  n_width <- n_max - n_min

  x_min <- n_min
  if (solution_found_below) {
    tmp <- (x$below$x_final - n_min) / n_width
    if (tmp < .20) {
      tmp2 <- (x$below$x_final - .20 * n_max) /
              (1 - .20)
      x_min <- floor(tmp2)
    }
  }

  x_max <- n_max
  if (solution_found_above) {
    tmp <- (x$above$x_final - n_min) / n_width
    if (tmp > .80) {
      tmp2 <- (x$above$x_final - n_min + .80 * n_min) / .80
      x_max <- ceiling(tmp2)
    }
  }

  args0 <- list(...)
  args1 <- utils::modifyList(
               args0,
               list(object = a,
                    predictor = predictor,
                    main = main,
                    xlab = xlab,
                    ylab = ylab)
              )
  if (is.null(args0$xlim)) {
    args1 <- utils::modifyList(args1,
                               list(xlim = c(x_min, x_max)))
  }

  do.call(plot_power_x,
          args1)

  # === Add CIs?

  if ("ci" %in% what) {

    a_for_ci <- a

    if (solution_found &&
        ("final_x" %in% what)) {

      # A solution was found and final_x line is to be drawn.
      # Draw the final_x CI separately

      # Draw the CI for the final x

      if (solution_found_below) {
        tmp <- x$below$power4test_trials[x$below$i_final]
        class(tmp) <- class(x$below$power4test_trials)
        tmp_args <- utils::modifyList(pars_ci_final_x,
                                      list(object = tmp,
                                          predictor = predictor))
        do.call(plot_power_x_ci,
                tmp_args)

        a_for_ci <- a_for_ci[a_for_ci$n != x$below$x_final, ]

      }

      if (solution_found_above) {
        tmp <- x$above$power4test_trials[x$above$i_final]
        class(tmp) <- class(x$above$power4test_trials)
        tmp_args <- utils::modifyList(pars_ci_final_x,
                                      list(object = tmp,
                                          predictor = predictor))
        do.call(plot_power_x_ci,
                tmp_args)

        a_for_ci <- a_for_ci[a_for_ci$n != x$above$x_final, ]

      }

    }

    # Draw the other CIs
    tmp_args <- utils::modifyList(pars_ci,
                                  list(object = a_for_ci,
                                       predictor = predictor))
    do.call(plot_power_x_ci,
            tmp_args)
  }

  # === Draw the power curve?

  # It is intentional *not* to use the power curve plot method.

  if ("power_curve" %in% what) {
    tmp1 <- x$below$power4test_trials
    tmp2 <- x$above$power4test_trials
    i <- setdiff(names(tmp2), names(tmp1))
    if (length(i) > 0) {
      # No need to check because the models must be the same
      tmp_trials <- c(tmp1, tmp2[i], skip_checking_models = TRUE)
    } else {
      tmp_trials <- tmp1
    }
    fit <- power_curve(tmp_trials)
    tmp_args <- utils::modifyList(pars_power_curve,
                                  list(object = a,
                                       predictor = "n",
                                       power_x_fit = fit))
    do.call(plot_power_curve_x,
            tmp_args)
  }

  # === Draw a horizontal line for the target power?

  if ("target_power" %in% what) {
    tmp_args <- utils::modifyList(pars_target_power,
                                  list(h = x$below$target_power))
    do.call(abline,
            tmp_args)
  }

  if (solution_found) {

    # === Draw a vertical line for the final N?

    if ("final_x" %in% what) {
      if (solution_found_below) {
        tmp_args <- utils::modifyList(pars_final_x,
                                      list(v = x$below$x_final))
        do.call(abline,
                tmp_args)
      }
      if (solution_found_above) {
        tmp_args <- utils::modifyList(pars_final_x,
                                      list(v = x$above$x_final))
        do.call(abline,
                tmp_args)
      }
    }

    # === Draw a horizontal line for the final power?

    if ("final_power" %in% what) {
      if (solution_found_below) {
        tmp_args <- utils::modifyList(pars_final_power,
                                      list(h = x$below$power_final))
        do.call(abline,
                tmp_args)
      }
      if (solution_found_above) {
        tmp_args <- utils::modifyList(pars_final_power,
                                      list(h = x$above$power_final))
        do.call(abline,
                tmp_args)
      }
    }

    # === Add a label for the final x?

    if ("final_x" %in% text_what) {
      if (solution_found_below) {
        x_final_str <- formatC(x$below$x_final,
                              digits = switch(predictor,
                                              n = 0,
                                              es = digits),
                              format = "f")
        tmp_args <- utils::modifyList(list(y = par("usr")[3] + .05 * (par("usr")[4] - par("usr")[3])),
                                      pars_text_final_x_lower)
        tmp_args <- utils::modifyList(tmp_args,
                                      list(x = x$below$x_final,
                                          labels = x_final_str))
        do.call(text,
                tmp_args)
      }
      if (solution_found_above) {
        x_final_str <- formatC(x$above$x_final,
                              digits = switch(predictor,
                                              n = 0,
                                              es = digits),
                              format = "f")
        tmp_args <- utils::modifyList(list(y = par("usr")[3] + .10 * (par("usr")[4] - par("usr")[3])),
                                      pars_text_final_x_upper)
        tmp_args <- utils::modifyList(tmp_args,
                                      list(x = x$above$x_final,
                                          labels = x_final_str))
        do.call(text,
                tmp_args)
      }
    }

    if ("sig_area" %in% what) {
      if (solution_found_below) {
        tmp_args <- utils::modifyList(pars_sig_area,
                                      list(
                                        xleft = par("usr")[1],
                                        xright = x$below$x_final,
                                        ybottom = par("usr")[3],
                                        ytop = par("usr")[4],
                                        border = NULL)
                                      )
        do.call(rect,
                tmp_args)
      }
      if (solution_found_above) {
        tmp_args <- utils::modifyList(pars_sig_area,
                                      list(
                                        xleft = x$above$x_final,
                                        xright = par("usr")[2],
                                        ybottom = par("usr")[3],
                                        ytop = par("usr")[4],
                                        border = NULL)
                                      )
        do.call(rect,
                tmp_args)
      }
    }

    if ("sig_area" %in% text_what) {
      if (solution_found_below) {
        tmp_str <- paste0("Power sig. <\n",
                           formatC(x$below$target_power,
                                   digits,
                                   format = "f"))
        tmp_args <- utils::modifyList(pars_text_sig_area,
                                      list(x = mean(c(x$below$x_final,
                                                      par("usr")[1])),
                                           y = mean(c(x$below$target_power,
                                                      par("usr")[3])),
                                           labels = tmp_str))
        do.call(text,
                tmp_args)
      }
      if (solution_found_above) {
        tmp_str <- paste0("Power sig. >\n",
                           formatC(x$above$target_power,
                                   digits,
                                   format = "f"))
        tmp_args <- utils::modifyList(pars_text_sig_area,
                                      list(x = mean(c(x$above$x_final,
                                                      par("usr")[2])),
                                           y = mean(c(x$above$target_power,
                                                      par("usr")[3])),
                                           labels = tmp_str))
        do.call(text,
                tmp_args)
      }
    }

    # === Add a label for the final power?

    if ("final_power" %in% text_what) {
      if (solution_found_below) {
        tmp <- par("usr")
        tmp_args <- utils::modifyList(pars_text_final_power,
                                      list(y = x$below$power_final,
                                          x = tmp[1] + (tmp[2] - tmp[1]) * .05,
                                          pos = 1,
                                          labels = formatC(x$below$power_final,
                                                            digits = digits,
                                                            format = "f")))
        if (!is.null(pars_text_final_power$x)) {
          tmp_args <- utils::modifyList(tmp_args,
                                        list(x = pars_text_final_power$x))
        }
        do.call(text,
                tmp_args)
      }
      if (solution_found_above) {
        tmp <- par("usr")
        tmp_args <- utils::modifyList(pars_text_final_power,
                                      list(y = x$above$power_final,
                                          x = tmp[1] + (tmp[2] - tmp[1]) * .05,
                                          pos = 3,
                                          labels = formatC(x$above$power_final,
                                                            digits = digits,
                                                            format = "f")))
        if (!is.null(pars_text_final_power$x)) {
          tmp_args <- utils::modifyList(tmp_args,
                                        list(x = pars_text_final_power$x))
        }
        do.call(text,
                tmp_args)
      }
    }

  }

  invisible(x)
}

#' @noRd
# Most arguments should have default values
# Probabilistic Bisection
alg_prob_bisection <- function(
    object,
    x,
    pop_es_name,
    ...,
    target_power = .80,
    x_min = switch(
              x,
              n = c(50),
              es = c(.00)
            ),
    x_max = switch(
              x,
              n = c(2000),
              es = c(.70)
            ),
    progress = TRUE,
    progress_type = c("cli", "cat"),
    x_include_interval = FALSE,
    x_interval = switch(x,
                    n = c(50, 2000),
                    es = c(.00, .70)
                  ),
    simulation_progress = FALSE,
    save_sim_all = FALSE,
    is_by_x = FALSE,
    object_by_org,
    final_nrep = 2000,
    final_R = NULL,
    ci_level = .95,
    extendInt = c("no", "yes", "downX", "upX"),
    max_trials = 10,
    R = final_R,
    digits = 3,
    lower_hard = switch(x, n = 10, es = 0),
    upper_hard = switch(x, n = 10000, es = .999),
    extend_maxiter = 5,
    what = c("point", "ub", "lb"),
    goal = c("close_enough", "ci_hit"),
    tol = NULL,
    variants = list()
) {

  # Termination Criteria
  # - Number of trials: max_trials
  # - The range of x in the last_k trials is within delta_tol
  # - The range of f in the last_k trials is within delta_tol_f
  # - The width of the dominant highest-density region
  # TODO:
  # - Add time allowed?

  # Solution found
  # - Close enough (use tol, determined by final_nrep)
  # - CI hits (use ci, determined by final_nrep)

  progress_type <- match.arg(progress_type)

  # ==== Default to Boos-Zhang ====

  if (is.null(variants$bz)) {
    bz_old <- options(power4mome.bz = TRUE)
    on.exit(options(bz_old))
  }

  # ==== Pre-search setup ====

  a_out <- power_algorithm_prob_bisection_pre_i(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    x_max = x_max,
    x_min = x_min,
    progress = progress,
    progress_type = progress_type,
    x_include_interval = x_include_interval,
    x_interval = x_interval,
    simulation_progress = simulation_progress,
    save_sim_all = save_sim_all,
    is_by_x = is_by_x,
    object_by_org = object_by_org,
    final_nrep = final_nrep,
    final_R = final_R,
    what = what,
    goal = goal,
    tol = tol,
    ci_level = ci_level
  )

  # ==== Process output ====

  x_interval_updated <- a_out$x_interval_updated
  by_x_1 <- a_out$by_x_1
  fit_1 <- a_out$fit_1

  # TODO:
  # - Check whether it works
  if (is_by_x) {
    tmp <- setdiff(names(object_by_org), names(by_x_1))
    by_x_1 <- c(by_x_1,
                object_by_org[tmp],
                skip_checking_models = TRUE)
  }

  rm(a_out)

  lower_hard <- min(x_interval)
  upper_hard <- max(x_interval)

  # ==== Start the search ====

  a_out <- power_algorithm_prob_bisection(
    object = object,
    x = x,
    pop_es_name = pop_es_name,
    target_power = target_power,
    ci_level = ci_level,
    x_interval = x_interval_updated,
    extendInt = extendInt,
    progress = progress,
    progress_type = progress_type,
    simulation_progress = simulation_progress,
    max_trials = max_trials,
    final_nrep = final_nrep,
    final_R = final_R,
    R = R,
    save_sim_all = save_sim_all,
    by_x_1 = by_x_1,
    fit_1 = fit_1,
    is_by_x = is_by_x,
    digits = digits,
    lower_hard = lower_hard,
    upper_hard = upper_hard,
    what = what,
    goal = goal,
    tol = tol,
    power_model = variants$power_model,
    power_curve_start = variants$power_curve_start,
    lower_bound = variants$lower_bound,
    upper_bound = variants$upper_bound,
    nls_control = variants$nls_control %||% list(),
    nls_args = variants$nls_args %||% list(),
    variants = variants
  )

  # ==== Return the output ====

  a_out
}

#' @noRd
power_algorithm_prob_bisection <- function(
                                      object,
                                      x,
                                      pop_es_name = NULL,
                                      ...,
                                      target_power = .80,
                                      ci_level = .95,
                                      x_interval = switch(x, n = c(50, 2000),
                                                             es = c(.00, .70)),
                                      extendInt = c("no", "yes", "downX", "upX"),
                                      progress = TRUE,
                                      progress_type = c("cli", "cat"),
                                      simulation_progress = FALSE,
                                      max_trials = 100,
                                      final_nrep = 400,
                                      final_R = NULL,
                                      R = NULL,
                                      power_model = NULL,
                                      power_curve_start = NULL,
                                      lower_bound = NULL,
                                      upper_bound = NULL,
                                      nls_control = list(),
                                      nls_args = list(),
                                      save_sim_all = FALSE,
                                      start = mean(x_interval),
                                      by_x_1 = NULL,
                                      fit_1 = NULL,
                                      is_by_x = FALSE,
                                      digits = 3,
                                      lower_hard = switch(x, n = 10, es = 0),
                                      upper_hard = switch(x, n = 10000, es = .999),
                                      extend_maxiter = 3,
                                      what = c("point", "ub", "lb"),
                                      goal = c("close_enough", "ci_hit"),
                                      tol = NULL,
                                      last_k = 5,
                                      variants = list()) {
  extendInt <- match.arg(extendInt)
  what <- match.arg(what)
  goal <- match.arg(goal)
  progress_type <- match.arg(progress_type)
  status <- NULL
  changes_ok <- TRUE
  final_R <- final_R %||% R

  # ==== Initialize histories holders ====

  f_history <- vector("numeric", max_trials + 1)
  f_history[] <- NA
  x_history <- vector("numeric", max_trials + 1)
  x_history[] <- NA
  x_interval_history <- matrix(NA,
                               nrow = max_trials + 1,
                               ncol = 2)
  colnames(x_interval_history) <- c("lower", "upper")
  f_interval_history <- matrix(NA,
                               nrow = max_trials + 1,
                               ncol = 2)
  colnames(f_interval_history) <- c("lower", "upper")
  reject_history <- vector("numeric", max_trials + 1)
  reject_history[] <- NA

  reject_by_power_curve_history <- vector("numeric", max_trials + 1)
  reject_by_power_curve_history[] <- NA

  dfun_history <- vector("list", max_trials + 1)

  p_c_history <- vector("numeric", max_trials + 1)
  p_c_history[] <- NA

  z_history <- vector("numeric", max_trials + 1)
  z_history[] <- NA

  hdr_x_history <- vector("list", max_trials + 1)
  hdr_power_history <- vector("list", max_trials + 1)

  final_check_history <- vector("logical", max_trials + 1)
  final_check_history[] <- FALSE

  seed_history <- vector("numeric", max_trials + 1)
  seed_history[] <- NA

  time_passed_history <- vector("numeric", max_trials + 1)
  time_passed_history[] <- NA

  i <- NA

  # what: The value to be examined.
  # goal:
  # - ci_hit: Only relevant for what == "point"
  # - close_enough: Can be used for all what.

  x_type <- x

  # ==== Default for variants ====

  variants0 <- list(delta_tol = NULL,
                    last_k = 5,
                    delta_tol_f = NULL,
                    use_power_curve_assist = TRUE,
                    use_power_curve_min_points = 2,
                    power_curve_args = list(),
                    use_power_curve_hybrid = TRUE,
                    total_nrep = 5000,
                    initial_nrep = NULL,
                    nrep_step = 0,
                    step_up_factor = 4,
                    step_up_factor_f = 4,
                    trial_nrep = NULL,
                    npoints = 5000,
                    p = .51,
                    use_estimated_p = TRUE,
                    adjust_p_c = .90,
                    hdr_prob = NULL,
                    hdr_power_tol = NULL,
                    hdr_prob_ci_level_ratio = .90,
                    perturbation_times = 2,
                    max_final_checks = 3,
                    final_check_cooldown = NULL,
                    dfun_integer = switch(x,
                                          n = TRUE,
                                          NULL),
                    rollback = TRUE,
                    bz = TRUE,
                    min_interval_width = c(n = 100,
                                           es = .20))
  variants <- utils::modifyList(variants0,
                                variants)
  if (is.null(variants$hdr_prob)) {
    variants$hdr_prob <- ci_level * variants$hdr_prob_ci_level_ratio
  }


  if (is.null(variants$final_check_cooldown)) {
    variants$final_check_cooldown <- variants$last_k
  }

  # ==== Default to Boos-Zhang ====

  if (variants$bz) {
    bz_old <- options(power4mome.bz = TRUE)
    on.exit(options(bz_old))
  }

  # Default for hdr_power_tol to be set later

  # ==== Set trial_nrep, initial_nrep, and nrep_step

  if (is.null(variants$trial_nrep)) {
    variants$trial_nrep <- ceiling(variants$total_nrep / max_trials)
  }
  if (is.null(variants$initial_nrep)) {
    variants$initial_nrep <- variants$trial_nrep
  } else if (variants$initial_nrep > variants$trial_nrep) {
    variants$initial_nrep <- variants$trial_nrep
  }
  if (is.null(variants$nrep_step)) {
    variants$nrep_step <- variants$initial_nrep
  }

  # ==== Set proxy_power for power_curve_assist ====

  proxy_power <- tryCatch(target_power_adjusted(
                    target_power = target_power,
                    goal = goal,
                    what = what,
                    tolerance = 0,
                    nrep = final_nrep,
                    level = ci_level
                  ),
                  error = function(e) e)
  if (inherits(proxy_power, "error")) {
    stop("The goal is not possible given the final_nrep and target power. Please try other values.")
    proxy_power <- NA
  }

  # ==== Emulate ci_hit by close_enough ====

  if (goal == "ci_hit") {
    goal <- "close_enough"
    what <- "point"
    # Force tol to be determined by final_nrep
    tol <- NULL
  }

  # ==== Set default for tol ====

  if (is.null(tol)) {
    tmp0 <- ifelse(
                inherits(proxy_power, "error"),
                yes = target_power,
                no = proxy_power
              )
    tmp1 <- reject_ci_wilson(
              nreject = ceiling(tmp0 * final_nrep),
              nvalid = final_nrep,
              level = ci_level
            )
    tmp3 <- min(abs(tmp1[1, ] - tmp0)) * .90
    tol <- tmp3
  }

  # ==== Set default for hdr_power_tol ====

  if (is.null(variants$hdr_power_tol)) {
    tmp0 <- ifelse(
                inherits(proxy_power, "error"),
                yes = target_power,
                no = proxy_power
              )
    tmp1 <- reject_ci_wilson(
              nreject = ceiling(tmp0 * final_nrep),
              nvalid = final_nrep,
              level = ci_level
            )
    variants$hdr_power_tol <- diff(tmp1[1, ])
  }

  # ==== Set default for delta_tol ====

  if (is.null(variants$delta_tol)) {
    delta_tol <- switch(
                    x,
                    n = 2,
                    es = .002
                  )
  } else {
    delta_tol <- variants$delta_tol
  }

  last_k <- variants$last_k

  # ==== Set default for delta_tol_f ====

  if (is.null(variants$delta_tol_f)) {
    if (!is.na(proxy_power)) {
      tmp1 <- reject_ci_wilson(
                nreject = ceiling(tmp0 * final_nrep),
                nvalid = final_nrep,
                level = .95
              )
      delta_tol_f <- diff(range(tmp1[1, ])) / 2
    } else {
      delta_tol_f <- .001
    }
  } else {
    delta_tol_f <- variants$delta_tol_f
  }

  # Create the objective function

  # Tbe output is:
  # - a scalar,
  # - with the attribute "output" storing the full by_* output.
  # - The scalar depends on 'what'

  # ==== Generate the objective function ====

  if (!is.na(proxy_power)) {
    # If proxy power can be computed,
    # use it as the target for close_enough
    f_power <- proxy_power
    f_what <- "point"
    f_goal <- "close_enough"
  } else {
    # If proxy power cannot be computed,
    # use the confidence bound as the goal
    f_power <- target_power
    f_what <- what
    f_goal <- "close_enough"
  }

  if (((variants$initial_nrep * f_power) %% 1) == 0) {
    variants$initial_nrep <- variants$initial_nrep + 1
  }

  f <- gen_objective(object = object,
                     x = x,
                     pop_es_name = pop_es_name,
                     target_power = f_power,
                     ci_level = ci_level,
                     progress = progress,
                     digits = digits,
                     nrep = variants$initial_nrep,
                     R = R,
                     what = f_what,
                     simulation_progress = simulation_progress,
                     save_sim_all = save_sim_all,
                     store_output = TRUE,
                     target_nrep = final_nrep,
                     progress_type = progress_type)

  # ==== Get the initial interval ====

  # Find f.lower and f.upper

  # Arguments for rejection rates should be retrieved from the object
  by_x_ci <- rejection_rates_add_ci(by_x_1,
                                    level = ci_level)
  x_tried <- get_x_tried(by_x_ci,
                         x = x)
  reject_tried <- by_x_ci$reject
  lower <- min(x_interval)
  upper <- max(x_interval)

  if (progress) {
    print_interval(lower = lower,
                   upper = upper,
                   digits = digits,
                   x_type = x_type,
                   prefix = "Initial interval:")
  }

  # Compute f.lower and f.upper if not available

  # f.lower

  # ==== Compute f for the lower bound  ====

  tmp <- in_x_tried(lower,
                    object = by_x_ci,
                    x = x,
                    multiple_matches = TRUE)

  # Ensure the nrep is the same
  # Do not extract those with nrep == final_nrep
  # This should be the job of x_from_power().

  if (!any(is.na(tmp))) {
    tmp_nrep <- by_x_ci$nrep[tmp]
    tmp_nrep_i <- tmp_nrep %in% variants$initial_nrep
    if (any(tmp_nrep_i)) {
      tmp <- tmp[which(tmp_nrep_i)[1]]
    } else {
      tmp <- NA
    }
  }

  lower_in_by_x_1 <- tmp

  if (!is.na(tmp)) {

    # ==== Extract from by_x_1  ====

    output_tmp <- by_x_1[tmp]
    class(output_tmp) <- class(by_x_1)
    f.lower <- f(x_i = lower,
                 power_i = reject_tried[tmp],
                 progress = FALSE)
    attr(f.lower, "output") <- output_tmp
  } else {
    if (progress) {
      if (progress_type == "cat") {
        cat("\nDo the simulation for the lower bound:\n")
      }
      if (progress_type == "cli") {
        # TODO:
        # - Use cli step
        cat("Do the simulation for the lower bound ...\n")
      }
    }

    # ==== Estimate f for the lower bound  ====

    f.lower <- f(x_i = lower,
                 x = x,
                 pop_es_name = pop_es_name,
                 target_power = f_power,
                 ci_level = ci_level,
                 progress = ifelse(progress_type == "cat", progress, FALSE),
                 progress_type = "cat",
                 digits = digits,
                 nrep = variants$initial_nrep,
                 R = R,
                 what = f_what,
                 simulation_progress = simulation_progress,
                 save_sim_all = save_sim_all,
                 store_output = TRUE,
                 target_nrep = final_nrep)
    by_x_1 <- c(by_x_1, attr(f.lower, "output"),
                skip_checking_models = TRUE)
  }

  # f.upper

  # ==== Compute f for the upper bound  ====

  tmp <- in_x_tried(upper,
                    object = by_x_ci,
                    x = x,
                    multiple_matches = TRUE)

  # Ensure the nrep is the same
  # Do not extract those with nrep == final_nrep
  # This should be the job of x_from_power().

  if (!any(is.na(tmp))) {
    tmp_nrep <- by_x_ci$nrep[tmp]
    tmp_nrep_i <- tmp_nrep %in% variants$initial_nrep
    if (any(tmp_nrep_i)) {
      tmp <- tmp[which(tmp_nrep_i)[1]]
    } else {
      tmp <- NA
    }
  }

  upper_in_by_x_1 <- tmp

  if (!is.na(tmp)) {

    # ==== Extract from by_x_1  ====

    output_tmp <- by_x_1[tmp]
    class(output_tmp) <- class(by_x_1)
    f.upper <- f(x_i = upper,
                 power_i = reject_tried[tmp],
                 progress = FALSE)
    attr(f.upper, "output") <- output_tmp
  } else {
    if (progress) {
      if (progress_type == "cat") {
        cat("\nDo the simulation for the upper bound:\n")
      }
      if (progress_type == "cli") {
        # TODO:
        # - Use cli step
        cat("Do the simulation for the upper bound ...\n")
      }
    }

    # ==== Estimate f for the upper bound  ====

    f.upper <- f(x_i = upper,
                 x = x,
                 pop_es_name = pop_es_name,
                 target_power = f_power,
                 ci_level = ci_level,
                 progress = ifelse(progress_type == "cat", progress, FALSE),
                 progress_type = "cat",
                 digits = digits,
                 nrep = variants$initial_nrep,
                 R = R,
                 what = f_what,
                 simulation_progress = simulation_progress,
                 save_sim_all = save_sim_all,
                 store_output = TRUE,
                 target_nrep = final_nrep)
    by_x_1 <- c(by_x_1, attr(f.upper, "output"),
                skip_checking_models = TRUE)
  }

  if (progress) {
    if (progress_type == "cat") {
      cat("\n")
      print_interval(lower = lower,
                     upper = upper,
                     digits = digits,
                     x_type = x_type,
                     prefix = "Initial interval:")
      cat("\n")
      cat("- Rejection Rates:\n")
      tmp <- rejection_rates(by_x_1,
                            level = ci_level)
      print(tmp, annotation = FALSE)
      cat("\n")
    }
  }

  do_search <- TRUE

  # ==== Is one of the bound a solution?  ====

  # Check whether lower or upper is already a solution

  # This should nearly never happen because initial_nrep
  # should be less than final_nrep.
  # Keep the code, just in case.

  output_lower <- attr(f.lower, "output")
  # Arguments for rejection rates should be retrieved from the object
  # No need for other arguments. Only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_lower,
            all_columns = TRUE
          )
  reject_lower <- tmp$reject
  nrep_lower <- tmp$nrep

  output_upper <- attr(f.upper, "output")
  # Arguments for rejection rates should be retrieved from the object
  # No need for other arguments. Only `reject` and `nrep` are used.
  tmp <- rejection_rates(
            output_upper,
            all_columns = TRUE
          )
  reject_upper <- tmp$reject
  nrep_upper <- tmp$nrep

  tmp <- check_solution_bounds(
            f.lower = f.lower,
            f.upper = f.upper,
            target_power = target_power,
            final_nrep = final_nrep,
            ci_level = ci_level,
            what = what,
            goal = goal,
            tol = tol
          )

  ok_lower <- tmp["ok_lower"]
  ok_upper <- tmp["ok_upper"]

  if (ok_lower || ok_upper) {

    # ==== Solution in the interval. Skip the search ====

    # One of them is a solution. No need to check the interval,
    # even if the interval is below or above the root.

    do_search <- FALSE

  } else {

    # ==== Extend the interval if necessary ====

    # Fix the interval
    # The original interval is returned if it is OK

    if ((sign(f.lower) != sign(f.upper)) &&
        (abs(lower - upper) >= 0)) {
      # ==== The interval is valid ====
      extend_progress <- FALSE
    } else {
      # ==== Extend the interval ====
      extend_progress <- progress
    }

    interval_updated <- extend_interval(f = f,
                                        x = x,
                                        pop_es_name = pop_es_name,
                                        target_power = f_power,
                                        ci_level = ci_level,
                                        nrep = variants$initial_nrep,
                                        target_nrep = final_nrep,
                                        R = R,
                                        what = f_what,
                                        goal = f_goal,
                                        tol = -Inf,
                                        simulation_progress = simulation_progress,
                                        save_sim_all = save_sim_all,
                                        progress = extend_progress,
                                        x_type = x_type,
                                        by_x_1 = by_x_1,
                                        lower = lower,
                                        upper = upper,
                                        f.lower = f.lower,
                                        f.upper = f.upper,
                                        lower_hard = lower_hard,
                                        upper_hard = upper_hard,
                                        extendInt = extendInt,
                                        extend_maxiter = extend_maxiter,
                                        trace = as.numeric(extend_progress),
                                        digits = digits,
                                        store_output = TRUE,
                                        overshoot = switch(x,
                                                           n = 1,
                                                           es = .10),
                                        variants = variants,
                                        proxy_power = proxy_power,
                                        extend_only = TRUE
                                        )
    by_x_1 <- interval_updated$by_x_1 %||% by_x_1
    ci_hit <- FALSE
    solution_found <- FALSE

    lower <- interval_updated$lower
    upper <- interval_updated$upper

    # ==== Update bounds' outputs, if necessary ====

    # The outputs of lower and upper are always in by_x_1

    # Store the output back into f.lower, if not yet stored

    f.lower <- interval_updated$f.lower
    if (is.null(attr(f.lower, "output"))) {
      if (!is.na(lower_in_by_x_1)) {
        # Need to do this because a value may have been attempted more than once
        tmp <- lower_in_by_x_1
      } else {
        tmp <- in_x_tried(test_x = lower,
                          object = by_x_1,
                          x = x)
      }
      tmp <- by_x_1[tmp]
      class(tmp) <- class(by_x_1)
      attr(f.lower, "output") <- tmp
    }

    # Store the output back into f.upper, if not yet stored

    f.upper <- interval_updated$f.upper
    if (is.null(attr(f.upper, "output"))) {
      if (!is.na(upper_in_by_x_1)) {
        # Need to do this because a value may have been attempted more than once
        tmp <- upper_in_by_x_1
      } else {
        tmp <- in_x_tried(test_x = lower,
                          object = by_x_1,
                          x = x)
      }
      tmp <- by_x_1[tmp]
      class(tmp) <- class(by_x_1)
      attr(f.upper, "output") <- tmp
    }

    # ==== Fix the start value, if necessary ====

    # Fix the start value if it is outside the new interval

    # Not used in PBA

    if ((start <= lower) || (start >= upper)) {
      start <- mean(c(lower, upper))
    }

    # ==== Store output(s) to by_x_1, if necessary ====

    # Store the output of f.lower to by_x_1, if not yet stored

    tmp <- in_x_tried(test_x = lower,
                      object = by_x_1,
                      x = x)
    if (is.na(tmp)) {
      # No need to check for multiple matches,
      # because this is done only if no match.
      by_x_1 <- c(by_x_1, attr(interval_updated$f.lower, "output"),
                  skip_checking_models = TRUE)
    }

    # Store the output of f.upper to by_x_1, if not yet stored

    tmp <- in_x_tried(test_x = upper,
                      object = by_x_1,
                      x = x)
    if (is.na(tmp)) {
      # No need to check for multiple matches,
      # because this is done only if no match.
      by_x_1 <- c(by_x_1, attr(interval_updated$f.upper, "output"),
                  skip_checking_models = TRUE)
    }

    # ==== Does the updated interval has a solution?  ====

    # Check whether the updated lower or upper is already a solution

    # This should nearly never happen because initial_nrep
    # should be less than final_nrep.
    # Keep the code, just in case.

    output_lower <- attr(f.lower, "output")
    # Arguments for rejection rates should be retrieved from the object
    # No need for other arguments. Only `reject` and `nrep` are used.
    tmp <- rejection_rates(
             output_lower,
             all_columns = TRUE
           )
    reject_lower <- tmp$reject
    nrep_lower <- tmp$nrep

    output_upper <- attr(f.upper, "output")
    # Arguments for rejection rates should be retrieved from the object
    # No need for other arguments. Only `reject` and `nrep` are used.
    tmp <- rejection_rates(
             output_upper,
             all_columns = TRUE
           )
    reject_upper <- tmp$reject
    nrep_upper <- tmp$nrep

    tmp <- check_solution_bounds(
              f.lower = f.lower,
              f.upper = f.upper,
              target_power = target_power,
              final_nrep = final_nrep,
              ci_level = ci_level,
              what = what,
              goal = goal,
              tol = tol
            )

    ok_lower <- tmp["ok_lower"]
    ok_upper <- tmp["ok_upper"]

    if ((interval_updated$extend_status > 0) &&
        (!ok_lower && !ok_upper)) {

      # ==== No solution and interval invalid. Skip the search ====

      # Interval not OK and no bounds are the solution

      if (progress) {
        cat("\n")
        cat(names(interval_updated$extend_status), ".\n")
        cat("... and none of the bounds are solution.\n")
        cat("Try another interval.\n\n")

      }

      # Should quit

      do_search <- FALSE

    }

  }

  # ==== Is one of the bounds a solution? ====

  # This should nearly never happen because initial_nrep
  # should be less than final_nrep.
  # Keep the code, just in case.

  if (ok_lower || ok_upper) {

    # ==== Yes. Skip the search ====

    if (progress) {
      cat("One of the bounds in the interval is already a solution.\n\n")
    }

    do_search <- FALSE
    ci_hit <- switch(goal,
                     ci_hit = TRUE,
                     close_enough = NA)
    solution_found <- TRUE

    status <- bisection_status_message(0, status)

    # The lower bound take precedence

    if (ok_lower) {
      x_i <- lower
      out_i <- f.lower
      output_i <- output_lower
      reject_i <- reject_lower
    } else {
      x_i <- upper
      out_i <- f.upper
      output_i <- output_upper
      reject_i <- reject_upper
    }
  }

  # ==== Do the search ====

  if (do_search) {

    # ==== Set the prior ====

    f.lower_i <- f.lower
    f.upper_i <- f.upper
    lower_i <- lower
    upper_i <- upper
    status <- 0
    perturbation_count <- variants$perturbation_times
    final_checks_available <- variants$max_final_checks

    dfun_i <- gen_dfun(
                  interval = c(lower_i, upper_i),
                  npoints = variants$npoints,
                  integer = variants$dfun_integer
                )

    p <- variants$p

    x_i <- q_dfun(
                  dfun_i,
                  prob = .50
                )

    if (x_type == "n") {
      x_i <- ceiling(x_i)
    }

    # ==== Start the loop ====

    i <- 1
    nrep_i <- variants$initial_nrep
    nreps_total <- nrep_i
    step_up_factor <- variants$step_up_factor
    step_up_factor_f <- variants$step_up_factor_f
    time_start <- Sys.time()
    pb_id <- NA
    terminate_hdr_power_tol <- FALSE
    terminate_change_ok_x <- FALSE
    terminate_change_ok_f <- FALSE
    nrep_i_old <- NA
    do_final_check <- FALSE
    ok <- FALSE
    final_check_cooldown_i <- 0
    time_passed_i_total <- NA
    nrep_final_check_total <- 0
    last_k_str <- character(0)
    last_k_f_str <- character(0)

    # Always do a final check, even if max_trials is reached.
    while (((i <= max_trials) &&
            (nreps_total < variants$total_nrep)) ||
           do_final_check) {

      # ==== Progress: Set up cli ====

      x_i_str <- switch(
                    x,
                    n = paste0("|n:", as.character(x_i)),
                    es = paste0("|es:",
                                formatC(
                                  x_i,
                                  digits = digits,
                                  format = "f"
                                ))
                  )

      if (progress &&
          (progress_type == "cli") &&
          is.na(pb_id)) {
        # The progress message for the loop
        cat("Search Progress Note:",
            "- #: Iteration number",
            "- nrep: The number of replications",
            "- n/es: The value to try",
            "- TE: Time elapsed",
            paste0("- ETA: Estimated time to do all ",
                   variants$total_nrep,
                  " replications and one final check"),
            "- Rep: The number of replications used out of the total number of replications",
            paste0("- Dx: The range of changes of x in the last ", last_k, " iterations"),
            paste0("- Df: The range of changes of f in the last ", last_k, " iterations"),
            paste0("- PP: The region(s) of probable power"),
            sep = "\n")
        changes_last_k <- character(0)
        changes_last_k_f <- character(0)
        time_passed <- character(0)
        eta <- character(0)
        hdr_f_str <- character(0)
        pb_id <- cli::cli_progress_message(
            " #:{i}|nrep:{nrep_i}{x_i_str}{time_passed}{eta}|Rep:{nreps_total}/{variants$total_nrep}{changes_last_k}{changes_last_k_f}{hdr_f_str}"
          )
      }

      # ==== Progress: Show iteration progress ====

      time_passed0 <- difftime(
                        Sys.time(),
                        time_start
                      )
      if (progress) {
        if (progress_type == "cat") {
          cat("\nIteration #", i, "\n")
          cat("\nTotal replications accumulated: ", nreps_total)
          cat("\nNumber of replications for this iteration: ", nrep_i)
          cat("\n")
        }
        if (progress_type == "cli") {
          time_passed <- paste0("|TE:",
                                format(time_passed0, digits = 3),
                                "")
          if (i >= 2) {
            tmp <- (variants$total_nrep - nreps_total + final_nrep) *
                   time_passed_i_total / (nreps_total + nrep_final_check_total)
            tmp2 <- tmp
            units(tmp2) <- "secs"
            if (as.numeric(tmp2) > 60 * 60) {
              units(tmp) <- "hours"
            } else if (as.numeric(tmp2) > 60) {
              units(tmp) <- "mins"
            }
            eta <- paste0("|ETA:",
                          format(tmp, digits = 3),
                          "")
          }
          cli::cli_progress_update(id = pb_id,
                                   force = TRUE)
        }
      }

      # ==== Compute f(x) ====

      time_start_i <- Sys.time()

      tmp <- in_x_tried_advanced(
                test_x = x_i,
                object = by_x_1,
                x = x_type,
                nrep = nrep_i,
                seed_exclude = seed_history)

      if (is.na(tmp)) {
        # ==== New output ====
        if (do_final_check) {
          tmp_R <- final_R
          tmp_sim_progress <- progress
        } else {
          tmp_R <- R
          tmp_sim_progress <- simulation_progress
        }
        out_i <- f(
                  x_i = x_i,
                  x = x,
                  pop_es_name = pop_es_name,
                  target_power = f_power,
                  ci_level = ci_level,
                  progress = progress,
                  progress_type = progress_type,
                  digits = digits,
                  nrep = nrep_i,
                  R = tmp_R,
                  what = f_what,
                  simulation_progress = tmp_sim_progress,
                  save_sim_all = save_sim_all,
                  store_output = TRUE,
                  target_nrep = final_nrep,
                  pb_id = pb_id
                )
        use_existing_out <- FALSE
      } else {
        # ==== Retrieve existing output ===
        output_tmp <- by_x_1[tmp]
        class(output_tmp) <- class(by_x_1)
        out_i <- f(
                  x_i = upper,
                  power_i = rejection_rates(output_tmp)$reject,
                  progress = FALSE
                )
        attr(out_i, "output") <- output_tmp
        use_existing_out <- TRUE
      }

      time_passed_i <- difftime(
                        Sys.time(),
                        time_start_i
                      )

      time_passed_history[i] <- time_passed_i

      if (is.na(time_passed_i_total)) {
        time_passed_i_total <- time_passed_i
      } else {
        time_passed_i_total <- time_passed_i_total + time_passed_i
      }

      output_i <- attr(out_i, "output")

      if (!use_existing_out) {
        by_x_1 <- c(by_x_1, output_i,
                    skip_checking_models = TRUE)
      }

      seed_history[i] <- attr(output_i[[1]], "args")$iseed
      # Arguments for rejection rates should be retrieved from the object
      # No need for other arguments. Only `reject` is used.
      reject_i <- rejection_rates(output_i)$reject

      if (progress) {
        if (progress_type == "cat") {
          cat("- Rejection Rates:\n")
          # Arguments for rejection rates should be retrieved from the object
          tmp <- rejection_rates(by_x_1,
                                 level = ci_level)
          print(tmp, annotation = FALSE)
          cat("\n")
        }
      }

      # ==== Histories ====

      f_history[i] <- as.numeric(out_i)
      x_history[i] <- x_i
      reject_history[i] <- reject_i

      # ==== Is x a solution? =====

      if (do_final_check) {
        final_check_history[i] <- TRUE
        ok <- check_solution(
                f_i = reject_i,
                target_power = target_power,
                nrep = final_nrep,
                final_nrep = final_nrep,
                ci_level = ci_level,
                what = what,
                goal = goal,
                tol = tol
              )
      }

      # Some steps will be conducted even if this iteration
      # is a final check, because x_i may not be a solution.

      # ==== Check and display changes in the last_k iterations ====

      # For PBA, changes_ok FALSE is a termination criterion,
      # not a problem.

      changes_ok <- check_changes(
              x_history = x_history,
              delta_tol = delta_tol,
              last_k = last_k
            )
      changes_ok_f <- check_changes(
              x_history = f_history,
              delta_tol = delta_tol_f,
              last_k = last_k
            )

      if ((i >= last_k) && !ok) {
        tmp <- x_history[(i - last_k + 1):i]
        tmp <- diff(range(tmp))
        last_k_str <- switch(x_type,
                          n = tmp,
                          es = formatC(tmp,
                                       digits = digits,
                                       format = "f"))
        tmp <- f_history[(i - last_k + 1):i]
        tmp <- diff(range(tmp))
        last_k_f_str <- formatC(
                            tmp,
                            digits = digits,
                            format = "f"
                          )
      }

      if (progress && (i >= last_k) && !ok) {
        if (progress_type == "cat") {
          cat("The range of changes in the last",
              last_k,
              "iteration:",
              last_k_str,
              "\n")
        }
        if (progress_type == "cli") {
          changes_last_k <- paste0(
                                  "|",
                                  "Dx:",
                                  last_k_str)
          changes_last_k_f <- paste0(
                                  "|",
                                  "Df:",
                                  last_k_f_str)
          cli::cli_progress_update(
              id = pb_id,
              force = TRUE
            )
        }
      }

      # ==== Where the solution may be ====

      if (sign(out_i) == sign(f.lower_i)) {

        # ==== Solution on the right ====

        z_i <- 1

      } else {

        # ==== Solution on the left ====

        z_i <- -1

      }

      # ==== Update the density function ====

      if (variants$use_estimated_p) {
        p_c_i <- tryCatch(p_c(
                    target_power = target_power,
                    power_i = reject_i,
                    goal = goal,
                    what = what,
                    trial_nrep = nrep_i,
                    level = ci_level,
                    final_nrep = final_nrep
                  ), error = function(e) e)
        if (!inherits(p_c_i, "error")) {
          p_c_i <- max(variants$adjust_p_c * p_c_i,
                      variants$p)
        } else {
          p_c_i <- variants$p
        }
        p_c_history[i] <- p_c_i
        dfun_i <- update_dfun(
                    dfun = dfun_i,
                    x_i = x_i,
                    p = p_c_i,
                    z_i = z_i
                  )
      } else {
        p_c_history[i] <- p
        dfun_i <- update_dfun(
                    dfun = dfun_i,
                    x_i = x_i,
                    p = p,
                    z_i = z_i
                  )
      }

      dfun_history[[i]] <- dfun_i
      z_history[i] <- z_i

      # ==== Update x_i? ====

      if (do_final_check) {

        # ==== A final check ====

        if (ok) {
          # No need to update. x_i is a solution.
          # This is the final iteration.
        } else {
          # Need to continue the search
          x_i <- q_dfun(
                    dfun = dfun_i,
                    prob = .50
                  )
        }

      } else {

        # ==== A PBA iteration ====

        x_i <- q_dfun(
                  dfun = dfun_i,
                  prob = .50
                )
      }

      if (x_type == "n") {
        x_i <- ceiling(x_i)
      }

      crlo <- q_dfun(
                  dfun_i,
                  prob = .10
                )
      crhi <- q_dfun(
                  dfun_i,
                  prob = .90
                )

      # ==== Histories ====

      x_interval_history[i, ] <- c(crlo, crhi)

      # No need to reuse results if x has tried.
      # A new draw gives new information.

      # ==== Termination: f interval ====

      cat_newline <- FALSE

      # Do this check even if doing a final check

      hdr_x <- hdr(
                  dfun = dfun_i,
                  prob = variants$hdr_prob
                )
      # Search is unstable if only a narrow range is used.
      # Therefore, use the full range of x.
      fit_tmp <- tryCatch(do.call(
                          power_curve,
                          c(list(object = by_x_1),
                            variants$power_curve_args)
                        ),
                        error = function(e) e)
      if (inherits(fit_tmp, "power_curve")) {
        hdr_power <- lapply(
                        hdr_x,
                        \(x) {
                          stats::predict(
                              fit_tmp,
                              newdata = list(x = x)
                            )
                        }
                      )
        reject_by_power_curve_history[i] <-
                          stats::predict(
                              fit_tmp,
                              newdata = list(x = x_i)
                            )
      } else {
        hdr_power <- lapply(
                        hdr_x,
                        \(x) c(NA, NA)
                      )
        reject_by_power_curve_history[i] <- NA
      }
      hdr_x_history[[i]] <- hdr_x
      hdr_power_history[[i]] <- hdr_power

      if (length(hdr_power) > 1) {
        hdr_power_pb <- sapply(
                            hdr_x,
                            attr,
                            which = "pb",
                            USE.NAMES = FALSE
                          )
        hdr_power_pb <- hdr_power_pb / sum(hdr_power_pb)
        hdr_power_dominant <- hdr_power_pb > .80
        hdr_power_dominant_i <- which(hdr_power_dominant)
      } else if (length(hdr_power) == 1) {
        hdr_power_dominant <- TRUE
        hdr_power_dominant_i <- 1
      } else {
        hdr_power_dominant <- FALSE
      }
      terminate_hdr_power_tol <- FALSE
      if (any(hdr_power_dominant) &&
          isFALSE(all(is.na(hdr_power[[hdr_power_dominant_i]]))) &&
          (diff(hdr_power[[hdr_power_dominant_i]]) <= variants$hdr_power_tol) &&
          (proxy_power >= hdr_power[[hdr_power_dominant_i]][1]) &&
          (proxy_power <= hdr_power[[hdr_power_dominant_i]][2])) {
        terminate_hdr_power_tol <- TRUE
      }

      if (progress &&
          (progress_type == "cli")) {
          if (any(hdr_power_dominant)) {
            tmp0 <- hdr_power[hdr_power_dominant_i]
          } else {
            tmp0 <- hdr_power
          }
          tmp1 <- sapply(
                    tmp0,
                    \(x) sprintf("[%1$4.3f,%2$4.3f]", x[1], x[2])
                  )
          tmp2 <- paste0(tmp1, collapse = ";")
          hdr_f_str <- paste0("|PP:", tmp2)
          cli::cli_progress_update(id = pb_id, force = TRUE)
      }

      # ==== Termination: last_k changes ====

      terminate_change_ok_x <- FALSE
      terminate_change_ok_f <- FALSE

      # Do this check even if doing a final check
      # Not much change
      # - A final check but not a solution? -> Perturbation
      # - A final check and solution found? -> change_ok ignored
      # - A PBA iteration -> Do a final check next
      if (do_final_check) {
        if (!ok) {
          if (!changes_ok || !changes_ok_f) {
            if (perturbation_count > 0) {
              perturbation_count <- perturbation_count - 1
              x_i <- x_i + delta_tol * 50 * z_i
              x_i <- min(max(lower_i, x_i), upper_i)
              x_i <- switch(
                        x_type,
                        n = ceiling(x_i),
                        es = x_i
                      )
            } else {
              status <- bisection_status_message(3, status)
              terminate_change_ok_x <- !changes_ok
              terminate_change_ok_f <- !changes_ok_f
            }
          }
        }
      } else {
        terminate_change_ok_x <- !changes_ok
        terminate_change_ok_f <- !changes_ok_f
      }
      # if ((!changes_ok || !changes_ok_f) &&
      #     (perturbation_count == 0)) {

      #   status <- bisection_status_message(3, status)

      #   terminate_change_ok_x <- !changes_ok
      #   terminate_change_ok_f <- !changes_ok_f

      # } else if ((!changes_ok || !changes_ok_f) &&
      #            !ok &&
      #            do_final_check) {
      #   # If not a solution, then the next iteration is a PBA iteration
      #   perturbation_count <- perturbation_count - 1
      #   x_i <- x_i + delta_tol * 50 * z_i
      #   x_i <- min(max(lower_i, x_i), upper_i)
      #   # x_i <- q_dfun(dfun_i,
      #   #               prob = runif(1, -.10, .10) +
      #   #                      sample(c(.30, .70), size = 1))
      #   x_i <- switch(
      #             x_type,
      #             n = ceiling(x_i),
      #             es = x_i
      #           )
      # }

      # ==== Next i: Set nrep ====

      any_terminate <-  terminate_hdr_power_tol ||
                        terminate_change_ok_x ||
                        terminate_change_ok_f

      if (do_final_check) {

        # ==== A final_check iteration ====

        nrep_final_check_total <- nrep_final_check_total + final_nrep

        if (ok) {

          # ==== Solution found ====

          if (progress) {
            if (progress_type == "cli") {
              cli::cli_process_done(id = pb_id)
            }
            cat("\n** Search ended **: Solution found.")
          }

        } else {

          # ==== Solution not found ====

          # The next iteration is a PBA iteration

          last_final_check <- (final_checks_available == 0)
          do_final_check <- FALSE
          nrep_i <- nrep_i_old
          final_check_cooldown_i <- variants$final_check_cooldown

          if (progress) {
            # if (progress_type == "cli") {
            #   cli::cli_progress_update(id = pb_id,
            #                            force = TRUE)
            # }
            if (last_final_check) {
              cat("\nSolution not found in Iteration ",
                  i,
                  ".",
                  sep = "")
            } else {
              cat("\nSolution not found in Iteration ",
                  i,
                  ". Continue the search.",
                  sep = "")
            }
            cat_newline <- TRUE
          }

        }

      } else {

        # ==== A PBA iteration ====

        if (any_terminate &&
            any(final_check_history) &&
            (final_check_cooldown_i == 0)) {

          # ==== Rollback if x_i too close to the previous candidate ====

          x_tried_solution <- x_history[final_check_history]
          if ((abs(min(x_i - x_tried_solution)) < tol / 4) &&
              variants$rollback) {

            # ==== Rollback ====

            i1 <- (abs(x_i - x_history) > tol / 4) & !final_check_history
            i3 <- (apply(x_interval_history, MARGIN = 1, diff) / diff(x_interval)) > .10
            i2 <- i1 & i3
            if (all(is.na(i2))) {
              i2 <- 1
            } else {
              i2 <- sample(which(i2), size = 1)
            }

            # Start again from i2

            if (progress) {
              if (progress_type == "cli") {
                cli::cli_progress_update(id = pb_id,
                                         force = TRUE)
              }
              cat("\nCandidate solution too close to the previous candidate(s).\n",
                  "Restart from the state in Iteration ", i2, " again.",
                  sep = "")
              cat_newline <- TRUE
            }

            # Reset the cooldown

            any_terminate <- FALSE
            final_check_cooldown_i <- variants$final_check_cooldown

            # Use failed final check(s) because they are informative

            i3 <- which(final_check_history)

            x_i <- x_history[i2]
            dfun_i <- dfun_history[[i2]]

            for (ii in i3) {
              dfun_i <- update_dfun(
                          dfun = dfun_i,
                          x_i = x_history[ii],
                          p = p_c_history[ii],
                          z_i = z_history[ii]
                        )
            }
            x_i <- q_dfun(
                      dfun = dfun_i,
                      prob = .50
                    )
            if (x_type == "n") {
              x_i <- ceiling(x_i)
            }
          }
        }


        if (any_terminate) {

          # ==== At least one criterion met ====

          if (final_checks_available > 0) {

            if (final_check_cooldown_i == 0) {

              # ==== Do a final check next ====

              do_final_check <- TRUE
              final_checks_available <- final_checks_available - 1

              nrep_i_old <- nrep_i
              nrep_i <- final_nrep

            } else {

              # ==== In cooldown stage ====

            }

          } else {

            # ==== Terminate anyway ====

          }

        } else {

          # ==== Increase nrep_i? ====

          if (!is.na(nrep_i_old)) {
            nrep_i <- nrep_i_old
          }

          step_up <- !check_changes(
                  x_history = x_history,
                  delta_tol = delta_tol * step_up_factor,
                  last_k = last_k
                )
          # delta_tol_f is dynamically determined.
          # No need for step_up_factor_f
          step_up_factor_f <- min(step_up_factor, .5)
          step_up_f <- !check_changes(
                  x_history = f_history,
                  delta_tol = delta_tol_f * step_up_factor_f,
                  last_k = last_k
                )
          if (step_up || step_up_f) {
            tmp <- variants$initial_nrep +
                  variants$step_up_factor * variants$nrep_step
            nrep_i <- min(final_nrep,
                          nrep_i + variants$nrep_step,
                          tmp)
            if (step_up) {
              step_up_factor <- max(0, step_up_factor - 1)
            }
          }

        }

        # Reduce the counter in all cases

        final_check_cooldown_i <- max(final_check_cooldown_i - 1, 0)

        ## ==== Update nreps_total ====

        if (!do_final_check) {
          nreps_total <- nreps_total + nrep_i
        }



      }

      # ==== Terminate? ====

      if (ok) {

        # ==== Solution found ====

        solution_found <- TRUE

        break

      } else if (any_terminate &&
                 (final_checks_available == 0) &&
                 !do_final_check) {

        # ==== No more final checks ====

        if (progress) {
          if (progress_type == "cli") {
            cli::cli_process_done(id = pb_id)
          }
          if (!changes_ok) {
            cat("\n** Search ended **: The range of changes of x in the last ",
                last_k,
                " iterations is less than ",
                delta_tol,
                ".",
                sep = "")
          }
          if (!changes_ok_f) {
            cat("\n** Search ended **: The range of changes of f in the last ",
                last_k,
                " iterations is less than ",
                delta_tol_f,
                ".",
                sep = "")
          }
          if (terminate_hdr_power_tol) {
            cat("\n** Search ended **: The width of the dominant region of probable power is ",
                formatC(variants$hdr_power_tol,
                        digits = 4,
                        format = "f"),
                " or less.",
                sep = "")
          }
        }

        if (cat_newline) {
          cat("\n")
        }

        break

      }

      # ==== Progress: Before the next iteration ====

      if (do_final_check) {
        if (progress) {
          # if (progress_type == "cli") {
          #   cli::cli_progress_update(id = pb_id,
          #                            force = TRUE)
          # }
          tmp <- variants$max_final_checks - final_checks_available
          if (terminate_hdr_power_tol) {
            cat("\nDo Final Check ",
                tmp,
                ": The width of the dominant region of probable power is ",
                formatC(variants$hdr_power_tol,
                        digits = 4,
                        format = "f"),
                " or less.",
                sep = "")
            cat_newline <- TRUE
          }
          if (terminate_change_ok_x) {
            cat("\nDo Final Check ",
                tmp,
                ": The range of changes of x in the last ",
                last_k,
                " iterations is less than ",
                delta_tol,
                ".",
                sep = "")
            cat_newline <- TRUE
          }
          if (terminate_change_ok_f) {
            cat("\nDo Final Check ",
                tmp,
                ": The range of changes of f in the last ",
                last_k,
                " iterations is less than ",
                delta_tol_f,
                ".",
                sep = "")
            cat_newline <- TRUE
          }
        }
      }

      if (cat_newline) {
        cat("\n")
      }

      # ==== Next i ====

      i <- i + 1

    }

    # ==== End the loop ====

    # Actual time passed
    time_passed0 <- difftime(
                  Sys.time(),
                  time_start
                )

    if (progress_type == "cli") {
      tryCatch(cli::cli_process_done(id = pb_id),
               error = function(e) e)
    }

    if (i == max_trials) {
      if (progress) {
        cat("\n** Search ended **: Maximum number of trials reached.")
      }
    }

    if ((nreps_total >= variants$total_nrep) &&
        (i != max_trials)) {
      if (progress) {
        cat("\n** Search ended **: Total number of replications reached.")
      }
    }

    if (progress) {
      cat("\nSummary:\n")
      cat("- Number of iterations:", i, "\n")
      cat("- Number of replications:", nreps_total, "\n")
      cat("- Time elapsed:", format(time_passed0, digits = 4), "\n")
      if (length(last_k_str) > 0) {
        cat("- The range of changes x in the last",
            last_k,
            "iteration:",
            last_k_str,
            "\n")
      }
      if (length(last_k_f_str)) {
        cat("- The range of changes f in the last",
            last_k,
            "iteration:",
            last_k_f_str,
            "\n")
      }
    }

  } else {

    # No iteration

  }

  # Prepare the output

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.

  # ** x_tried **
  # The sample sizes or population values
  # examined.

  # ** x_out **
  # The output in the last trial.

  # ** power_out **
  # The power in the last trial.

  # ** nrep_out **
  # The nrep in the last trial.

  # ** ci_out **
  # The ci in the last trial.

  # ** by_x_out **
  # The by_* output in the last trial

  # ** i2 **
  # The row number of the last trial
  # in by_x_1.

  # ** solution_found **
  # TRUE if an acceptable solution is found

  # ** dfun_out **
  # The density function

  # ==== Prepare the output ====

  if (!do_search &&
      (!ok_lower && !ok_upper)) {

    # ==== No solution. Set _out to NA ====

    status <- bisection_status_message(1, status)

    # No solution for whatever reason

    x_out <- NA
    power_out <- NA
    nrep_out <- NA
    ci_out <- NA
    by_x_out <- NA
    i2 <- NA
    dfun_out <- NA

  } else {

    # ==== Solution found. Store it ====

    # Store the last results,
    # regardless of solution

    out_i <- as.numeric(out_i)
    by_ci_i <- rejection_rates_add_ci(output_i,
                                      level = ci_level)
    ci_i <- unlist(by_ci_i[1, c("reject_ci_lo", "reject_ci_hi")])
    x_tried <- get_x_tried(by_x_1,
                           x = x)
    # An x may have been tried more than once.
    # Need to match by seed.
    # It may not be a solution. Therefore, no need to match by nrep.
    x_tried_seed <- seed_history[i]
    seed_tried <- sapply(
                    by_x_1,
                    \(x) attr(x, "args")$iseed
                  )
    i2a <- which(seed_tried %in% x_tried_seed)
    i2b <- which(x_tried %in% x_i)
    i2 <- intersect(i2a, i2b)
    by_x_ci <- rejection_rates_add_ci(by_x_1,
                                      level = ci_level)
    x_out <- x_i
    power_out <- unlist(by_ci_i[1, "reject"])
    nrep_out <- unlist(by_ci_i[1, "nrep"])
    ci_out <- ci_i
    by_x_out <- by_ci_i

    dfun_out <- dfun_i

  }

  # Available regardless of do_search

  # ==== Update the power curve ====

  fit_1 <- power_curve(by_x_1,
                      formula = power_model,
                      start = power_curve_start,
                      lower_bound = lower_bound,
                      upper_bound = upper_bound,
                      nls_control = nls_control,
                      nls_args = nls_args,
                      verbose = progress)
  x_tried <- get_x_tried(by_x_1,
                         x = x)

  if (progress) {
    if (solution_found) {
      cat("Solution found.\n")
    } else {
      cat("Solution not found.\n")
    }
  }

  # ==== Return the output ====

  # The final x_i are always returned if available
  # ci_hit is used to decide whether
  # final x_i is a solution.

  # Not used in probabilistic bisection
  f_interval_history[] <- NA

  i_tmp <- !is.na(x_history)

  out <- list(by_x_1 = by_x_1,
              fit_1 = fit_1,
              ci_hit = ci_hit,
              x_tried = x_tried,
              x_out = x_out,
              power_out = power_out,
              nrep_out = nrep_out,
              ci_out = ci_out,
              by_x_out = by_x_out,
              i2 = i2,
              solution_found = solution_found,
              status = status,
              iteration = i,
              x_history = x_history[i_tmp],
              x_interval_history = x_interval_history[i_tmp, ],
              f_interval_history = f_interval_history[i_tmp, ],
              reject_history = reject_history[i_tmp],
              reject_by_power_curve_history = reject_by_power_curve_history[i_tmp],
              f_history = f_history[i_tmp],
              dfun_history = dfun_history[i_tmp],
              p_c_history = p_c_history[i_tmp],
              hdr_x_history = hdr_x_history[i_tmp],
              hdr_power_history = hdr_power_history[i_tmp],
              final_check_history = final_check_history[i_tmp],
              seed_history = seed_history[i_tmp],
              time_passed_history = time_passed_history[i_tmp],
              tol = tol,
              delta_tol = delta_tol,
              last_k = last_k,
              what = what,
              goal = goal,
              ci_level = ci_level,
              dfun_out = dfun_out,
              proxy_power = proxy_power,
              f_power = f_power,
              f_what = f_what,
              f_goal = f_goal,
              hdr_power_tol = variants$hdr_power_tol)
  out

}


#' @noRd
power_algorithm_prob_bisection_pre_i <- function(object,
                                            x,
                                            pop_es_name,
                                            target_power,
                                            x_max,
                                            x_min,
                                            progress,
                                            progress_type = c("cat", "cli"),
                                            x_include_interval = FALSE,
                                            x_interval,
                                            simulation_progress,
                                            save_sim_all,
                                            is_by_x,
                                            object_by_org,
                                            final_nrep,
                                            final_R,
                                            what,
                                            goal,
                                            tol,
                                            ci_level
                                            ) {

  progress_type <- match.arg(progress_type)

  # PBA does not extend the range dynamically (for now).
  # Moreover, the initial nrep is usually very small.
  # Therefore, use x_interval as-is. Do not try a narrower interval.

  x_i <- x_interval

  # # ==== Initial values ====

  # # Inherited from bisection. Not used for now.

  # # This method only needs an initial interval
  # if (inherits(object_by_org, "power4test_by_n") ||
  #     inherits(object_by_org, "power4test_by_es")) {
  #   x_i <- set_x_range_by_x(
  #                     object,
  #                     x = x,
  #                     pop_es_name = pop_es_name,
  #                     target_power = target_power,
  #                     k = 2,
  #                     x_max = x_max,
  #                     x_min = x_min,
  #                     object_by_org = object_by_org,
  #                     what = what,
  #                     goal = goal,
  #                     tol = tol,
  #                     ci_level = ci_level)
  # } else {
  #   x_i <- set_x_range(object,
  #                     x = x,
  #                     pop_es_name = pop_es_name,
  #                     target_power = target_power,
  #                     k = 2,
  #                     x_max = x_max,
  #                     x_min = x_min)
  # }

  # # For bisection, no need to exclude the value in the input objects
  # # Exclude the value in the input object

  # # ==== Set x_interval ====

  # if (x_include_interval) {
  #   # For bisection, should exclude them initially,
  #   # and let extend_interval() to do the job
  #   x_i <- sort(c(x_interval, x_i))
  # }
  # x_i <- sort(unique(x_i))
  # if (length(x_i) == 1) {
  #   x_interval_min <- min(x_interval)
  #   x_interval_max <- max(x_interval)
  #   if ((x_i < x_interval_min) ||
  #       (x_i > x_interval_max)) {
  #     x_i <- x_interval
  #   } else {
  #     tmp <- mean(x_interval)
  #     if (x_i < tmp) {
  #       x_i <- c(x_i, x_interval_max)
  #     } else if (x_i > tmp) {
  #       x_i <- c(x_i, x_interval_min)
  #     } else {
  #       x_i <- c(mean(c(x_i, x_interval_min)),
  #                mean(c(x_i, x_interval_max)))
  #       x_i <- range(x_i)
  #     }
  #   }
  # } else {
  #   x_i <- range(x_i)
  # }

  # ==== Update by_x (by_x_i) ====

  by_x_i <- switch(x,
          n = as.power4test_by_n(object),
          es = as.power4test_by_es(object,
                                    pop_es_name = pop_es_name)
        )

  # ** by_x_1 **
  # The collection of all values tried and their results
  # to be updated when new value is tried.
  # Used after the end of the loop.

  # ** fit_1 **
  # The latest power curve
  # To be updated whenever by_x_1 is updated.
  # Used after the end of the loop.
  # Not used in PBA

  # ==== Return the output ====

  out <- list(by_x_1 = by_x_i,
              fit_1 = NULL,
              x_interval_updated = x_i)

  return(out)
}

#' @noRd
gen_dfun <- function(interval = c(50, 2000),
                     npoints = NULL,
                     integer = NULL
                    ) {
  if (is.null(integer)) {
    # ==== Determine whether treating x as an integer ====
    tmp <- seq(from = interval[1],
               to = interval[2])
    if (length(tmp) == diff(interval) + 1) {
      integer <- TRUE
      npoints <- length(tmp)
    } else {
      integer <- FALSE
    }
  }
  if (integer) {
    x <- seq(from = interval[1],
             to = interval[2])
    out <- cbind(x = x,
                 prob = 1 / length(x))
  } else {
    x0 <- seq(from = interval[1],
             to = interval[2],
             length = npoints + 1)
    x <- x0[-length(x0)] + diff(x0) / 2
    out <- cbind(x = x,
                 prob = 1 / (length(x)),
                 xlo = x0[-length(x0)],
                 xhi = x0[-1])
  }
  attr(out, "integer") <- integer
  out
}

#' @noRd
dfun_prob <- function(
  dfun,
  x
) {
  i <- which.min(abs(dfun[, "x", drop = TRUE] - x))
  unname(dfun[i, "prob", drop = TRUE])
}

#' @noRd
# Not used
update_dfun_prob <- function(
  dfun,
  x,
  new_prob
) {
  i <- which.min(abs(dfun[, "x", drop = TRUE] - x))
  dfun[i, "prob"] <- new_prob
  dfun
}

#' @noRd
normalize_dfun <- function(
  dfun
) {
  dfun[, "prob"] <- dfun[, "prob"] / sum(dfun[, "prob"])
  dfun
}

#' @noRd
# Update the density function
update_dfun <- function(
  dfun,
  x_i,
  p,
  z_i
) {
  if (isTRUE(attr(dfun, "integer"))) {
    out <- update_dfun_integer(
              dfun = dfun,
              x_i = x_i,
              p = p,
              z_i = z_i)
  } else {
    out <- update_dfun_continuous(
              dfun = dfun,
              x_i = x_i,
              p = p,
              z_i = z_i)
  }
  out
}

#' @noRd
# Update the density function
update_dfun_integer <- function(
  dfun,
  x_i,
  p,
  z_i
) {
  # j <- which.min(abs(dfun[, "x"] - x_i))
  j <- match(ceiling(x_i), dfun[, "x"])
  cd_i <- cumsum(dfun[, "prob"])[j]
  gamma_i <- (1 - cd_i) * p +
             cd_i * (1 - p)
  if (z_i > 0) {
    d_new <- dfun[, "prob"] / gamma_i
    p_i <- ifelse(
              dfun[, "x"] >= x_i,
              yes = p,
              no = (1 - p)
            )
    d_new <- d_new * p_i
  } else {
    d_new <- dfun[, "prob"] / (1 - gamma_i)
    p_i <- ifelse(
              dfun[, "x"] >= x_i,
              yes = (1 - p),
              no = p
            )
    d_new <- d_new * p_i
  }
  dfun_new <- dfun
  dfun_new[, "prob"] <- d_new
  dfun_new <- normalize_dfun(dfun_new)
  attr(dfun_new, "integer") <- attr(dfun, "integer")
  dfun_new
}

#' @noRd
# Update the density function
update_dfun_continuous <- function(
  dfun,
  x_i,
  p,
  z_i
) {
  # j <- which.min(abs(dfun[, "x"] - x_i))
  j <- findInterval(x_i, c(dfun[1, "xlo"], dfun[, "xhi"]))
  cd_i <- cumsum(dfun[, "prob"])[j]
  gamma_i <- (1 - cd_i) * p +
             cd_i * (1 - p)
  if (z_i > 0) {
    d_new <- dfun[, "prob"] / gamma_i
    p_i <- ifelse(
              seq_along(d_new) >= j,
              yes = p,
              no = (1 - p)
            )
    d_new <- d_new * p_i
  } else {
    d_new <- dfun[, "prob"] / (1 - gamma_i)
    p_i <- ifelse(
              seq_along(d_new) >= j,
              yes = (1 - p),
              no = p
            )
    d_new <- d_new * p_i
  }
  dfun_new <- dfun
  dfun_new[, "prob"] <- d_new
  dfun_new <- normalize_dfun(dfun_new)
  attr(dfun_new, "integer") <- attr(dfun, "integer")
  dfun_new
}

#' @noRd
q_dfun <- function(
  dfun,
  prob = .50
) {
  if (attr(dfun, "integer")) {
    out <- q_dfun_integer(
              dfun = dfun,
              prob = prob
            )
  } else {
    out <- q_dfun_continuous(
              dfun = dfun,
              prob = prob
            )
  }
  out
}

#' @noRd
q_dfun_integer <- function(
  dfun,
  prob = .50
) {
  qfun <- cumsum(dfun[, "prob"])
  # i <- length(tmp[tmp > 0])
  # i <- which.min(abs(tmp))
  i <- findInterval(prob,
                    c(0, qfun),
                    rightmost.closed = TRUE)
  # No need for interpolation in integer mode
  out <- dfun[i, "x"]
  unname(out)
}

#' @noRd
q_dfun_continuous <- function(
  dfun,
  prob = .50
) {
  qfun <- cumsum(dfun[, "prob"])
  qfun0 <- c(0, qfun[-length(qfun)])
  i <- findInterval(prob,
                    c(0, qfun),
                    rightmost.closed = TRUE)
  # Do interpolation
  yd <- (prob - qfun0[i]) / (qfun[i] - qfun0[i])
  x0 <- dfun[i, "xlo"]
  x1 <- dfun[i, "xhi"]
  out <- x0 + yd * (x1 - x0)
  unname(out)
}

#' @noRd
gen_nreps <- function(
  total_nrep = 10000,
  max_trials = 100,
  initial_nrep = 50
) {
  # TODO:
  # - Handle incompatible settings
  a <- total_nrep - initial_nrep * max_trials
  b <- max_trials - 1
  d <- seq(1, b)
  e <- a / sum(d)
  out <- c(initial_nrep, initial_nrep + d * e)
  out <- round(out)
}

skip("Internal")

# Examine the effect of the sequence

library(testthat)

test_that("probabilistic bisection: Sequence", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
m ~ x: s
y ~ m: m
y ~ x: s
"

####### n

out <- power4test(nrep = 20,
                  model = mod,
                  pop_es = mod_es,
                  n = 100,
                  fit_model_args = list(fit_function = "lm"),
                  test_fun = test_parameters,
                  test_args = list(par = "m~x"),
                  parallel = TRUE,
                  iseed = 1234)

by_x_1 <- power4test_by_n(out,
                          n = 90)

# Close enough

set.seed(12345)
a_out <- power_algorithm_prob_bisection(
                                  object = out,
                                  x = "n",
                                  by_x_1 = by_x_1,
                                  final_nrep = 2000,
                                  max_trials = 100,
                                  variants = list(total_nrep = 100 * 50))
x_history <- a_out$x_history
f_history <- a_out$f_history
dfun_history <- a_out$dfun_history[seq_along(f_history)]

i <- seq_along(x_history)
i <- sample.int(length(x_history))
x_tmp <- x_history[i]
f_tmp <- f_history[i]
p <- .60
all_equal_history <- vector("logical", length(i))
dfun_i <- gen_dfun(
              interval = c(50, 2000),
              npoints = (2000 - 50 + 1)
            )
for (i in seq_along(f_tmp)) {
  if (sign(f_tmp[i]) == -1) {
    z_i <- 1
  } else {
    z_i <- -1
  }

  dfun_i <- update_dfun(
              dfun = dfun_i,
              x_i = x_tmp[i],
              p = p,
              z_i = z_i
            )
  plot(dfun_i, type = "l", col = "blue")
  plot(dfun_history[[i]], type = "l", col = "red")
  all_equal_history[i] <- isTRUE(all.equal(
                              dfun_i,
                              dfun_history[[i]])
                            )
}
all_equal_history
all.equal(dfun_i,
          dfun_history[[i]])
ylim <- range(c(dfun_i[, "prob"],
                dfun_history[[i]][, "prob"]))
plot(dfun_i,
     type = "l",
     col = "blue",
     ylim = ylim)
points(dfun_history[[i]],
       type = "l",
       col = "red")
plot(dfun_i[, "prob"],
     dfun_history[[i]][, "prob"])

# Confirmed that the final posterior distribution
# does not depend on the order of the iterations.

dfun_out <- a_out$dfun_out

dfun_history_last <- dfun_history[[length(dfun_history)]]
all.equal(dfun_out,
          dfun_history_last)
ylim <- range(c(dfun_out[, "prob"],
                dfun_history_last[, "prob"]))
plot(dfun_out,
     type = "l",
     col = "blue",
     ylim = ylim)
points(dfun_history_last,
       type = "l",
       col = "red")


})

#' @title Process Data by Creating Ordinal Variables
#'
#' @description For the `process_data`
#' argument. It converts continuous
#' indicator variable to ordinal variables.
#'
#' @details
#' This function is to be used in
#' the `process_data` argument of
#' [power4test()].
#'
#' It is used for converting the
#' continuous indicator scores generated
#' to ordinal scores (with values 1, 2, 3, etc.).
#'
#' For example, if the cut points (thresholds) are
#' -2 and 2, then sores will be converted
#' to three categories: (-Inf to -2],
#' (-2 to 2], and (2 to Inf]. The intervals
#' are closed on the right. That is,
#' a score of -2 is in the interval
#' (-Inf to -2], not in the interval
#' (-2 to 2).
#'
#' The conversion is implemented by [cut()].
#'
#' There are two ways to specify the
#' conversion. The first one is to use
#' some built-in thresholds, based on
#' Savalei and Rhemtulla (2013), Table 1.
#' Call [cut_patterns()] with no argument
#' to list all the built-in patterns
#' and their names.
#'
#' Alternatively, users can specify the
#' thresholds directly through the argument
#' `cuts`.
#'
#' Currently, all indicators of a latent
#' variable must be converted in the same
#' way.
#'
#' @return
#' The function [ordinal_variables()]
#' returns a data frame with the
#' the converted scores.
#'
#' @param data A data frame.
#'
#' @param cut_patterns A named vector.
#' The names are the names of the latent
#' variables for which indicator scores
#' will be converted. Each value must be
#' the name of one of the built-in
#' patterns (call `cut_patterns()` to
#' list the patterns and their names).
#' Can be used with `cuts` but a latent
#' variable should appear only either
#' in `cut_patterns` or `cuts`, not both.
#'
#' @param cuts A named list.
#' The names are the names of the latent
#' variables for which indicator scores
#' will be converted.
#' Each element is a vector of the
#' thresholds for the conversion. `-Inf`
#' and `Inf` will be automatically
#' included during the conversion.
#' Can be used with `cut_patterns` but a latent
#' variable should appear only either
#' in `cut_patterns` or `cuts`, not both.
#'
#' @seealso [power4test()]. See also
#' [cut()] for the implementation.
#'
#' @references
#' Savalei, V., & Rhemtulla, M. (2013).
#' The performance of robust test
#' statistics with categorical data.
#' *British Journal of Mathematical and Statistical Psychology*,
#' *66*(2), 201--223.
#' \doi{10.1111/j.2044-8317.2012.02049.x}
#'
#' @examples
#'
#' # Specify the model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' y ~ m: l
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Specify the numbers of indicators and reliability coefficients
#'
#' k <- c(y = 3,
#'        m = 4,
#'        x = 5)
#' rel <- c(y = .70,
#'          m = .70,
#'          x = .70)
#'
#' # Simulate the data
#'
#' out <- power4test(
#'          nrep = 2,
#'          model = mod,
#'          pop_es = mod_es,
#'          n = 200,
#'          number_of_indicators = k,
#'          reliability = rel,
#'          process_data = list(fun = "ordinal_variables",
#'                              args = list(cut_patterns = c(x = "s3"),
#'                                          cuts = list(m = c(-2, 0, 2)))),
#'          test_fun = test_parameters,
#'          test_args = list(op = "~"),
#'          parallel = FALSE,
#'          iseed = 1234)
#'
#' dat <- pool_sim_data(out)
#' head(dat, 50)
#'
#' # The built-in patterns
#'
#' cut_patterns()
#'
#' @export
ordinal_variables <- function(
  data,
  cut_patterns = NULL,
  cuts = NULL
) {
  k_ind <- attr(data, "number_of_indicators")
  # ==== Any factors with indicators? ====
  if (isTRUE(is.null(k_ind)) ||
      (all(k_ind == 1))) {
    return(data)
  }
  f_names <- names(k_ind)
  f <- function(x, xk) {
    paste0(x, seq_len(xk))
  }
  ind_names <- mapply(
                FUN = f,
                x = f_names,
                xk = k_ind,
                SIMPLIFY = FALSE
              )
  for (xi in f_names) {
    if (xi %in% names(cut_patterns)) {
      xi_ind <- ind_names[[xi]]
      cut_patterns_i <- cut_patterns[xi]
      xi_scores <- apply(
                    data[, xi_ind, drop = FALSE],
                    MARGIN = 2,
                    FUN = cut_pattern_scores,
                    cut_pattern = cut_patterns_i
                  )
      data[, xi_ind] <- xi_scores
    } else if (xi %in% names(cuts)) {
      xi_ind <- ind_names[[xi]]
      cuts_i <- cuts[[xi]]
      xi_scores <- apply(
                    data[, xi_ind, drop = FALSE],
                    MARGIN = 2,
                    FUN = \(x, breaks) {
                      as.numeric(cut(x, breaks = breaks))
                    },
                    breaks = c(-Inf, cuts_i, Inf)
                  )
      data[, xi_ind] <- xi_scores
    }
  }
  return(data)
}

#' @param which A name of a built-in pattern.
#'
#' @return
#' The function [cut_patterns()]
#' returns a named list of the built-in
#' patterns if `which = NULL`, and
#' a numeric vector of the thresholds
#' of a built-in pattern if `which` is
#' set to one of the names of the built-in
#' patterns.
#'
#' @rdname ordinal_variables
#' @export
cut_patterns <- function(
  which = NULL
) {
  # Savalei & Rhemtulla (2013) Table 1
  pattern_list <- list(
    s2 = c(0),
    s3 = c(-.83, .83),
    s4 = c(-1.25, 0, 1.25),
    s5 = c(-1.50, -.50, .50, 1.50),
    s6 = c(-1.60, -.83, 0, .83, 1.60),
    s7 = c(-1.79, -1.07, -.36, .36, 1.07, 1.79),
    ma2 = c(.36),
    ma3 = c(-.50, .76),
    ma4 = c(-.31, .79, 1.66),
    ma5 = c(-.70, .39, 1.16, 2.05),
    ma6 = c(-1.05, .08, .81, 1.44, 2.33),
    ma7 = c(-1.43, -.43, .38, .94, 1.44, 2.54),
    ea2 = c(1.04),
    ea3 = c(.58, 1.13),
    ea4 = c(.28, .71, 1.23),
    ea5 = c(.05, .44, .84, 1.34),
    ea6 = c(-.13, .25, .61, .99, 1.48),
    ea7 = c(-.25, .13, .47, .81, 1.18, 1.64),
    `-ma2` = -rev(c(.36)),
    `-ma3` = -rev(c(-.50, .76)),
    `-ma4` = -rev(c(-.31, .79, 1.66)),
    `-ma5` = -rev(c(-.70, .39, 1.16, 2.05)),
    `-ma6` = -rev(c(-1.05, .08, .81, 1.44, 2.33)),
    `-ma7` = -rev(c(-1.43, -.43, .38, .94, 1.44, 2.54)),
    `-ea2` = -rev(c(1.04)),
    `-ea3` = -rev(c(.58, 1.13)),
    `-ea4` = -rev(c(.28, .71, 1.23)),
    `-ea5` = -rev(c(.05, .44, .84, 1.34)),
    `-ea6` = -rev(c(-.13, .25, .61, .99, 1.48)),
    `-ea7` = -rev(c(-.25, .13, .47, .81, 1.18, 1.64))
  )
  if (is.null(which)) {
    return(pattern_list)
  } else {
    if (which %in% names(pattern_list)) {
      return(pattern_list[[which]])
    } else {
      stop("The pattern ",
           sQuote(which),
           " is not supported. Call cut_patterns() for supported patterns.")
    }
  }
}

#' @noRd
cut_pattern_scores <- function(
  x,
  cut_pattern
) {
  out <- cut(
            x,
            breaks = c(-Inf,
                       cut_patterns(cut_pattern),
                       Inf))
  as.numeric(out)
}

#' @title Random Variable From a Lognormal Distribution
#'
#' @description Generate random numbers
#' from a lognormal distribution,
#' rescaled to have user-specified
#' population mean and standard
#' deviation.
#'
#' @details
#' First, specify the parameter,
#' `mui` and `sigma`, and the
#' desired population mean and standard
#' deviation. The random numbers, drawn
#' from a lognormal distribution by
#' [stats::rlnorm()], will then be
#' rescaled with the desired population
#' mean and standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param mui The parameter `mui` to
#' be used by [stats::rlnorm()].
#'
#' @param sigma The parameter `sigma` to
#' be used by [stats::rlnorm()].
#'
#' @param rev If TRUE, the distribution
#' is revered to generate a negatively
#' skewed distribution. Default is
#' FALSE.
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard deviation.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rlnorm_rs(n = 5000,
#'                mui = 0,
#'                sigma = 1,
#'                pmean = 0,
#'                psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

rlnorm_rs <- function(n = 10,
                      mui = 0,
                      sigma = 1,
                      rev = FALSE,
                      pmean = 0,
                      psd = 1) {
  sigmasq <- sigma^2
  bmean <- exp(mui + sigmasq / 2)
  bsd   <- sqrt((exp(sigmasq) - 1) * exp(2 * mui + sigmasq))
  x <- stats::rlnorm(n, mui, sigma)
  if (!rev) {
    x <- pmean + psd * (x - bmean) / bsd
  } else {
    x <- pmean - psd * (x - bmean) / bsd
  }
  x
}

#' @title Summarize 'x_from_power' Results
#'
#' @description The summary method of
#' the output of [x_from_power()].
#'
#' @details
#' The `summary` method simply prepares the
#' results of [x_from_power()]
#' to be printed in details.
#'
#' @return
#' The `summary` method for
#' `x_from_power` objects returns an
#' object of the class
#' `summary.x_from_power`, which is
#' simply the output of [x_from_power()],
#' with a `print` method dedicated for
#' detailed summary. Please refer
#' to [x_from_power()] for the contents.
#'
#' @param object An `x_from_power`-class
#' object, such as the output of
#' [x_from_power()], or an object of the
#' class `n_region_from_power`, such as
#' the output of
#' [n_region_from_power()].
#'
#' @param ... Additional arguments.
#' Not used for now.
#'
#' @seealso [x_from_power()],
#' [n_region_from_power()]
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' mod_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' # Generate the datasets
#'
#' sim_only <- power4test(nrep = 5,
#'                        model = mod,
#'                        pop_es = mod_es,
#'                        n = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 2345)
#'
#' # Do a test
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_parameters,
#'                        test_args = list(pars = "m~x"))
#'
#' # Determine the sample size with a power of .80 (default)
#'
#' power_vs_n <- x_from_power(test_out,
#'                            x = "n",
#'                            progress = TRUE,
#'                            target_power = .80,
#'                            final_nrep = 5,
#'                            max_trials = 1,
#'                            seed = 1234)
#' summary(power_vs_n)
#'
#' @export
summary.x_from_power <- function(object,
                                 ...) {
  class(object) <- "summary.x_from_power"
  return(object)
}

#' @rdname summary.x_from_power
#' @export
summary.n_region_from_power <- function(
                                object,
                                ...) {
  class(object) <- "summary.n_region_from_power"
  return(object)
}

#' @rdname summary.x_from_power
#'
#' @param x The output of
#' [summary.x_from_power()], the
#' `summary` method of
#' an `x_from_power` object,
#' which is the output of
#' [x_from_power()], or the output
#' of [summary.n_region_from_power()],
#' the `summary` method of an
#' `n_region_from_power` object (the
#' output of [n_region_from_power()]).
#'
#' @param digits The number of digits
#' after the decimal when printing
#' the results.
#'
#' @return
#' The `print`-method of `summary.x_from_power`
#' objects returns the object `x`
#' invisibly.
#' It is called for its side effect.
#'
#' @export
print.summary.x_from_power <- function(x,
                                       digits = 3,
                                       ...) {

  cat("\n====== x_from_power Results ======\n\n")
  my_call <- x$call
  cat("Call:\n")
  if (!is.symbol(my_call$object)) {
    my_call$object <- as.symbol("<hidden>")
  }
  if (!is.symbol(my_call[[1]])) {
    my_call[[1]] <- as.symbol("x_from_power")
  }
  print(my_call)
  solution_found <- !is.na(x$x_final)

  predictor <- x$x
  cat("\n")
  cat("Predictor (x):",
      switch(predictor,
             n = "Sample Size",
             es = "Effect Size"),
      "\n")
  if (predictor == "es") {
    cat("Parameter Name (pop_es_name):",
        x$pop_es_name,
        "\n")
  }

  target_power_str <- formatC(x$target_power, digits = digits, format = "f")
  cat("\n- Target Power:",
      target_power_str,
      "\n")

  if (x$goal == "close_enough") {
    tmp1 <- switch(x$what,
                   point = "estimated power",
                   lb = "lower confidence bound",
                   ub = "upper confidence bound")
    tmp2 <- paste0(
        "- Goal: ",
        "Find 'x' with estimated ",
        tmp1,
        " close enough to the target power.")
    catwrap(tmp2,
            exdent = 2)
  }

  if (x$goal == "ci_hit") {
    tmp2 <- paste0(
        "- Goal: ",
        "Find 'x' with the confidence interval of ",
        "the estimated power enclosing the ",
        "target power.")
    catwrap(tmp2,
            exdent = 2)
  }

  cat("\n=== Major Results ===\n\n")
  if (solution_found) {
    x_final_str <- formatC(x$x_final,
                           digits = switch(predictor,
                                           n = 0,
                                           es = digits),
                           format = "f")
    tmp <- switch(x$x,
                  n = "(Sample Size)",
                  es = paste0("(", x$pop_es_name, ")"))
    cat("- Final Value ", tmp, ": ", x_final_str, "\n\n", sep = "")
    cat("- Final Estimated Power:",
        formatC(x$power_final, digits = digits, format = "f"),
        "\n")
    cat("- Confidence Interval: [",
        paste0(formatC(x$ci_final, digits = digits, format = "f"), collapse = "; "),
        "]\n", sep = "")
    cat("- Level of confidence: ",
        formatC(x$ci_level*100, digits = max(0, digits - 2), format = "f"), "%", "\n", sep = "")
    cat("- Based on", x$nrep_final, "replications.\n")
  } else {
    cat("- Solution not found.\n")
    catwrap(paste(c("- None of the values examined",
                    "in the interval meet the target power."),
                    collapse = " "),
            exdent = 2)
    if (isFALSE(identical(NA, x$x_estimated))) {
      x_estimated_str <- formatC(x$x_estimated,
                                 digits = switch(predictor,
                                                 n = 0,
                                                 es = digits),
                                 format = "f")
      cat("- The crude estimate of required value is ",
          x_estimated_str,
          ".\n", sep = "")
      cat("- Note: Estimated by the power curve.\n")
    }
    catwrap(paste(c("- Try changing the settings, such as",
                    "expanding the range of values",
                    "by setting 'x_interval' to one that",
                    "includes the crude estimate, if",
                    "available, increasing 'max_trials',",
                    "or increasing 'tolerance' if the goal",
                    "is 'close_enough'."),
                    collapse = " "),
            exdent = 2)
  }

  cat("\n=== Technical Information ===\n\n")
  cat("- Algorithm:",
      x$algorithm,
      "\n")
  if (x$goal == "close_enough") {
    cat("- Tolerance for 'close enough':",
      "Within",
      formatC(x$technical$tol,
              digits + 2,
              format = "f"),
      "of",
      target_power_str,
      "\n")
  }
  tmp0 <- switch(x$x,
                 n = ceiling(x$x_tried),
                 es = x$x_tried)
  tmp <- formatC(range(tmp0),
                  digits = switch(x$x,
                                  n = 0,
                                  es = digits),
                  format = "f")
  cat("- The range of values explored:",
      paste(tmp, collapse = " to "), "\n")
  cat("- Time spent in the search:",
      format(x$time_spent, digits = 4),
      "\n")
  if (isFALSE(identical(NA, x$power_curve))) {
    tmp <- class(x$power_curve$fit)[1]
    power_curve_name <- switch(tmp,
                               nls = "Nonlinear Regression Model",
                               glm = "Logistic Regression",
                               lm = "Linear Regression")
    cat("- The final crude model for the power-predictor relation:\n")
    cat("\nModel Type:",
        power_curve_name,
        "\n\n")
    tmp1 <- x$power_curve
    # Use the print method of power_curve objects
    print(tmp1)
    cat("\n")
  }
  cat("- Detailed Results:\n\n")
  print(x$rejection_rates,
        digits = 3)
  cat("\n")

  invisible(x)
}

#' @rdname summary.x_from_power
#'
#' @return
#' The `print`-method of `summary.n_region_from_power`
#' objects returns the object `x`
#' invisibly.
#' It is called for its side effect.
#'
#' @export
print.summary.n_region_from_power <- function(
                                      x,
                                      digits = 3,
                                      ...) {
  cat("\n======<< Summary for the Lower Region >>======\n\n")
  print(summary(x$below),
        digits = digits,
        ...)
  cat("\n\n======<< Summary for the Upper Region >>======\n\n")
  print(summary(x$above),
        digits = digits,
        ...)
  invisible(x)
}

#' @title Random Binary Variable
#'
#' @description Generate random numbers
#' from a distribution of 0 or 1,
#' rescaled to have user-specified
#' population mean and standard
#' deviation.
#'
#' @details
#' First, specify probability of 1
#' (`p1`), and the desired population
#' mean and standard deviation. The
#' random numbers, drawn from a
#' distribution of 0 (`1 - p1`
#' probability) and 1 (`p1`
#' probability), will then be rescaled
#' with the desired population mean and
#' standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param p1 The probability of being
#' 1, before rescaling.
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rbinary_rs(n = 5000,
#'                 p1 = .5,
#'                 pmean = 3,
#'                 psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#'
#' @export

rbinary_rs <- function(n = 10,
                    p1 = .50,
                    pmean = 0,
                    psd = 1) {
  bmean <- p1
  bsd   <- sqrt(p1 * (1 - p1))
  x <- sample.int(2,
                  size = n,
                  replace = TRUE,
                  prob = c(1 - p1, p1)) - 1
  x <- pmean + psd * (x - bmean) / bsd
  x
}

#' @title Estimate the Power of a Test
#'
#' @description An all-in-one function
#' that receives a model specification,
#' generates datasets, fits a model, does
#' the target test, and returns the test
#' results.
#'
#' @details
#' The function [power4test()] is an
#' all-in-one function for
#' estimating the power of a test for
#' a model, given the sample size
#' and effect sizes (population values
#' of model parameters).
#'
#' @details
#'
#' # Workflow
#'
#' This is the workflow:
#'
#' - If `object` is an output
#'  of the output
#'  of a previous call to
#'  [power4test()] with `do_the_test`
#'  set to `FALSE` and so has only the
#'  model and the simulated data,
#'  the following steps
#'  will be skipped and go directly to
#'  doing the test.
#'
#'    - Call [sim_data()] to determine
#'      the population model and
#'      generate the datasets, using
#'      arguments such as `model` and
#'      `pop_es`.
#'
#'    - Call [fit_model()] to fit a
#'      model to each of the datasets,
#'      which is the population model
#'      by default.
#'
#'    - If `R` is not `NULL` and
#'      `ci_type = "mc"`, call
#'      [gen_mc()] to generate Monte
#'      Carlo estimates using
#'      [manymome::do_mc()]. The estimates
#'      can be used by supported functions
#'      such as [test_indirect_effect()].
#
#'    - If `R` is not `NULL` and
#'      `ci_type = "boot"`, call
#'      [gen_boot()] to generate
#'      bootstrap estimates using
#'      [manymome::do_boot()]. The estimates
#'      can be used by supported functions
#'      such as [test_indirect_effect()].
#'
#'    - Merge the results into a
#'      `sim_out` object by calling
#'      [sim_out()].
#'
#'    - If `do_the_test` is `FALSE`,
#'      skip the remaining steps and
#'      return a `power4test` object,
#'      which contains only the data
#'      generated and optionally the
#'      Monte Carlo or bootstrap
#'      estimates.
#'
#' - If `do_the_test` is `TRUE`, do
#'   the test.
#'
#'    - [do_test()] will be called to do
#'      the test in the fit output of
#'      each dataset.
#'
#' - Return a `power4test` object which
#'   include the output of `sim_out`
#'   and, if `do_the_test` is `TRUE`,
#'   the output of [do_test()].
#'
#' This function is to be used when
#' users are interested only in the
#' power of one or several tests on a
#' particular aspect of the model, such
#' as a parameter, given a specific
#' effect sizes and sample sizes.
#'
#' Detailed description on major
#' arguments can be found in sections
#' below.
#'
#' NOTE: The technical internal workflow of
#' of [power4test()] can be found in
#' this page: <https://sfcheung.github.io/power4mome/articles/power4test_workflow.html>.
#'
#' # Updating a Condition
#'
#' The function [power4test()] can also be used to
#' update a condition when only some
#' selected aspects is to be changed.
#'
#' For example,
#' instead of calling this function with
#' all the arguments set just to change
#' the sample size, it can be called
#' by supplying an existing
#' `power4test` object and set only
#' `n` to a new sample size. The data
#' and the tests will be updated
#' automatically. See the examples for
#' an illustration.
#'
#' # Adding Another Test
#'
#' The function [power4test()] can also be used to
#' add a test to the output from a
#' previous call to [power4test()].
#'
#' For example, after simulating the
#' datasets and doing one test,
#' the output can be set to `object`
#' of [power4test()], and set only
#' `test_fun` and, optionally,
#' `test_fun_args` to do one more test
#' on the generated datasets. The output
#' will be the original
#' object with the results of the new
#' test added. See the examples for
#' an illustration.
#'
#' # Model Fitting Arguments ('fit_model_args')
#'
#' For power analysis, usually, the
#' population model (`model`) is to be
#' fitted, and there is no need to
#' set `fit_model_args`.
#'
#' If power analysis is to be conducted
#' for fitting a model that is not the
#' population model, of if non-default
#' settings are desired when fitting
#' a model, then the argument `fit_model_args`
#' needed to be set to customize the
#' call to [fit_model()].
#'
#' For example,
#' users may want to examine the power
#' of a test when a misspecified model
#' is fitted, or the power of a test
#' when MLR is used as the estimator
#' when calling [lavaan::sem()].
#'
#' See the help page of [fit_model()]
#' for some examples.
#'
#' @inheritSection ptable_pop Specify the Population Model by 'model'
#'
#' @inheritSection ptable_pop Specify 'pop_es' Using Named Vectors
#'
#' @inheritSection ptable_pop Specify 'pop_es' Using a Multiline String
#'
#' @inheritSection ptable_pop Set the Values for Effect Size Labels ('es1' and 'es2')
#'
#' @inheritSection sim_data Set 'number_of_indicators' and 'reliability'
#'
#' @inheritSection sim_data Specify The Distributions of Exogenous Variables Or Error Terms Using 'x_fun'
#'
#' @inheritSection do_test Major Test-Related Arguments
#'
# It is a known issue that inherited
# section cannot be placed before @details
# https://github.com/r-lib/roxygen2/issues/900
#
#
# This advanced feature to be introduced later
# #' ## Multiple Models
# #'
# #' More than one model can be fitted to
# #' each replication. This is done
# #' by setting `fit_model_args` to
# #' a named list. The names are the names
# #' used to identify the models, and
# #' each element is a list of named
# #' list of arguments for a model.
# #'
# #' For example:
# #'
# #' \preformatted{
# #' fit_model_args = list(fit = list(),
# #'                       fit2 = list(model = mod2),
# #'                       fit3 = list(model = mod3))
# #' }
# #'
# #' Three models will be fitted. The
# #' first model is the model ued to
# #' generate the data, named `"fit"`.
# #' The second and first models are
# #' named `"fit2"` and `"fit3"`,
# #' respectively, with `"fit2"` fitted
# #' with `model = mod2` and `"fit3"`
# #' fitted with `model = mod3`.
# #'
# #' If Monte Carlos or bootstrap estimates
# #' are to be generated, they will be
# #' generated for each model, using the
# #' values for their arguments.
#'
#' @return
#' An object of the class `power4test`,
#' which is a list with two elements:
#'
#' - `sim_all`: The output of [sim_out()].
#'
#' - `test_all`: A named list of the
#'    output of [do_test()]. The names
#'    are the values of `test_name`.
#'    This list can have more than one
#'    test because a call to
#'    [power4test()] can add new tests
#'    to a `power4test` object.
#'
#' @inheritParams do_test
#'
#' @inheritParams fit_model
#'
#' @inheritParams ptable_pop
#'
#' @inheritParams sim_data
#'
#' @param object Optional. If set to a
#' `power4test` object, it will be
#' updated using the value(s) in `n`,
#' `pop_es`, and/or `nrep` if they changed,
#' or a new test will be conducted and
#' added to `objet`. See the help page
#' for details. Default is `NULL`.
#'
#' @param nrep The number of replications
#' to generate the simulated datasets.
#' Default is `NULL`. Must be set when
#' called to create a `power4test`
#' object.
#'
#' @param ptable The output of
#' [ptable_pop()], which is a
#' `ptable_pop` object, representing the
#' population model. If `NULL`, the
#' default, [ptable_pop()] will be
#' called to generate the `ptable_pop`
#' object using `model` and `pop_es`.
#'
#' @param model The `lavaan` model
#' syntax of the population model,
#' to be used by [ptable_pop()].
#' See 'Details' of
#' on how to specify the model.
#' Ignored if `ptable` is
#' specified.
#'
#' @param pop_es The character vector or
#' multiline string to
#' specify population effect sizes
#' (population values of parameters). See
#' the help page on how to specify this
#' argument.
#' Ignored if `ptable` is
#' specified.
#'
# @param standardized <- Inherited
#'
#' @param n The sample size for each
#' dataset. Default is 100.
#'
# @param iseed <- Inherited
#'
# @param number_of_indicators <- Inherited
#'
# @param reliability <- Inherited
#'
# @param loading_difference <- Inherited
#'
# @param reference <- Inherited
#'
# @param x_fun <- Inherited
#'
# @param e_fun <- Inherited
#
# @param process_data <- Inherited
#'
#' @param fit_model_args A list of the
#' arguments to be passed to [fit_model()]
#' when fitting the
#' model.
#' Should be a named list
#' with names being the names of the
#' arguments.
#'
#' @param R The number of replications
#' to generate the Monte Carlo or
#' bootstrapping estimates
#' for each fit output. No Monte Carlo
#' nor bootstrapping
#' estimates will be generated if `R`
#' is set to `NULL`.
#'
#' @param ci_type The type of
#' simulation-based confidence
#' intervals to use. Can be either
#' `"mc"` for Monte Carlo method
#' (the default) or `"boot"` for
#' nonparametric bootstrapping method.
#' Relevant for test functions that
#' make use of estimates generate by
#' [gen_boot()] or [gen_mc()], such as
#' [test_indirect_effect()].
#'
#' @param gen_mc_args A list of
#' arguments to be passed to
#' [manymome::do_mc()] when generating
#' the Monte Carlo estimates.
#' Should be a named argument
#' with names being the names of the
#' arguments. Used only if
#' `ci_type` is `"mc".`
#'
#' @param gen_boot_args A list of
#' arguments to be passed to
#' [manymome::do_boot()] when generating
#' the bootstrap estimates.
#' Should be a named argument
#' with names being the names of the
#' arguments. Used only if
#' `ci_type` is `"boot".
#'
#' @param test_name String. The name
#' of the test. Default is `NULL`,
#' and the name will be created from
#' `test_fun`. Note that if `sim_out`
#' is a `power4test` object and already
#' has a test of this name stored, it
#' will be replaced by the new results.
#'
#' @param test_note String. An optional
#' note for the test, stored in the
#' attribute `test_note` of the output
#' of [do_test()]. Default is `NULL`.
#'
#' @param do_the_test If `TRUE`,
#' [do_test()] will be called to do the
#' test specified by `test_fun` on the
#' fit output of each dataset.
#'
#' @param sim_all If set to either a
#' `sim_out` object (the output of
#' [sim_out()] or a `power4test` object
#' (the output of [power4test()]), the
#' stored datasets and fit outputs will
#' be used for doing the test. Setting
#' `object` to the output of [power4test()]
#' is now the preferred method, but this
#' argument is kept for backward
#' compatibility.
#'
#' @param iseed The seed for the random
#' number generator. Default is `NULL`
#' and the seed is not changed. This
#' seed will be set only once, when
#' calling [sim_data()].
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used when calling
#' other functions, if appropriate.
#' Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' of each step will be displayed.
#' Default is `TRUE`.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param rejection_rates_args Default
#' argument values to be used when
#' [rejection_rates()] is called. Can
#' be overriden when calling
#' [rejection_rates()].
#'
# @param n_ratio <- Inherited
#'
# @param es1 <- Inherited
#'
# @param es2 <- Inherited
#'
# @param n_std <- Inherited
#'
# @param std_force_monte_carlo <- Inherited
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_med <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_med_es <-
#' "
#' m ~ x: m
#' y ~ m: l
#' y ~ x: n
#' "
#'
#' # Set nrep to a large number in real analysis, such as 400
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' out <- power4test(nrep = 10,
#'                   model = model_simple_med,
#'                   pop_es = model_simple_med_es,
#'                   n = 100,
#'                   test_fun = test_parameters,
#'                   test_args = list(pars = "m~x"),
#'                   iseed = 1234,
#'                   parallel = FALSE,
#'                   progress = TRUE)
#'
#' print(out,
#'       test_long = TRUE)
#'
#' # Change the sample size
#'
#' out1 <- power4test(out,
#'                    n = 200,
#'                    iseed = 1234,
#'                    parallel = FALSE,
#'                    progress = TRUE)
#'
#' print(out1,
#'       test_long = TRUE)
#'
#' # Add one more test
#'
#' out2 <- power4test(out,
#'                    test_fun = test_parameters,
#'                    test_args = list(pars = "y~x"),
#'                    parallel = FALSE,
#'                    progress = TRUE)
#'
#' print(out2,
#'       test_long = TRUE)
#'
#' @export

power4test <- function(object = NULL,
                       nrep = NULL,
                       ptable = NULL,
                       model = NULL,
                       pop_es = NULL,
                       standardized = TRUE,
                       n = NULL,
                       number_of_indicators = NULL,
                       reliability = NULL,
                       loading_difference = NULL,
                       reference = NULL,
                       x_fun = list(),
                       e_fun = list(),
                       process_data = NULL,
                       fit_model_args = list(),
                       R = NULL,
                       ci_type = "mc",
                       gen_mc_args = list(),
                       gen_boot_args = list(),
                       test_fun = NULL,
                       test_args = list(),
                       map_names = c(fit = "fit"),
                       results_fun = NULL,
                       results_args = list(),
                       test_name = NULL,
                       test_note = NULL,
                       do_the_test = TRUE,
                       sim_all = NULL,
                       iseed = NULL,
                       parallel = FALSE,
                       progress = TRUE,
                       ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                       es1 = c("n" = .00,
                               "nil" = .00,
                               "s" = .10,
                               "m" = .30,
                               "l" = .50,
                               "si" = .141,
                               "mi" = .361,
                               "li" = .510,
                               "sm" = .20,
                               "ml" = .40),
                       es2 = c("n" = .00,
                               "nil" = .00,
                               "s" = .05,
                               "m" = .10,
                               "l" = .15,
                               "sm" = .075,
                               "ml" = .125),
                       es_ind = c("si",
                                  "mi",
                                  "li"),
                       n_std = 100000,
                       std_force_monte_carlo = FALSE,
                       rejection_rates_args = list(collapse = "none",
                                                   at_least_k = 1,
                                                   merge_all_tests = FALSE,
                                                   p_adjust_method = "none",
                                                   alpha = .05),
                       n_ratio = 1) {

  # TOOD:
  # - Should allow only limited changes
  #   when updating a power4test object.

  ci_type <- match.arg(ci_type,
                       c("mc", "boot"))

  update_test <- FALSE

  update_power4test <- FALSE
  update_data <- TRUE
  if (!is.null(object)) {
    if (!inherits(object, "power4test")) {
      stop("object is not a power4test object.")
    }
    update_power4test <- TRUE
  }

  if (!update_power4test && is.null(nrep)) {
    stop("'nrep' must be set unless updating a power4test object.")
  }

  if (!update_power4test) {
    # Store the evaluated arguments
    args <- formals(power4test)
    args <- lapply(args,
                   eval,
                   envir = parent.frame())
  } else {
    args <- attr(object, "args")
  }
  my_call <- match.call()
  call_args <- as.list(my_call)[-1]
  call_args <- lapply(call_args,
                      eval,
                      envir = parent.frame())
  args <- utils::modifyList(args,
                            as.list(call_args))
  args$object <- NULL
  # args available in all cases.
  # It should be used whenever possible,
  # unless we explicitly need the value in this call.

  # ==== If parallel, make a cluster ====

  if (!isTRUE(args$ncores > 0)) {
    args$parallel <- FALSE
  }
  if (args$parallel) {

    # ==== Make a cluster ====

    cl <- parallel::makeCluster(ncores)
    on.exit(try(parallel::stopCluster(cl), silent = TRUE))

  } else {

    # ==== Serial ====

    cl <- NULL

  }

  if (update_power4test && !is.null(pop_es)) {
    # Population effect size changed
    # Data must be updated
    update_data <- TRUE
    # Update ptable
    old_ptable <- object$sim_all[[1]]$ptable
    new_ptable <- update_ptable_pop(old_ptable,
                                    new_pop_es = pop_es)
    ptable <- new_ptable
    # Use the updated population model
    args$ptable <- ptable
    args$model <- NULL
    args$pop_es <- NULL
  } else {
    # Population model not updated
    # Still update data if any of the following has changed
    if (is.null(n) &&
        is.null(number_of_indicators) &&
        is.null(reliability) &&
        is.null(loading_difference) &&
        is.null(reference) &&
        identical(x_fun, list()) &&
        identical(e_fun, list()) &&
        is.null(nrep)) {
      update_data <- FALSE
    }
  }

  if (update_data) {
    # Determine the arguments to use
    if (update_power4test) {
      # Regenerate the data
      # Mandatory use of stored arguments.
      # Changes in this call already incorporated into args.
      if (args$progress) {
        cat("Re-simulate the data:\n")
      }
      sim_data_args <- list(nrep = args$nrep,
                            ptable = ptable,
                            model = args$model,
                            pop_es = args$pop_es,
                            es1 = es1,
                            es2 = es2,
                            es_ind = es_ind,
                            standardized = standardized,
                            n_std = n_std,
                            std_force_monte_carlo = std_force_monte_carlo,
                            n = args$n,
                            number_of_indicators = args$number_of_indicators,
                            reliability = args$reliability,
                            loading_difference = args$loading_difference,
                            reference = args$reference,
                            x_fun = args$x_fun,
                            e_fun = args$e_fun,
                            process_data = args$process_data,
                            iseed = args$iseed,
                            parallel = args$parallel,
                            progress = args$progress,
                            ncores = args$ncores,
                            n_ratio = args$n_ratio,
                            cl = cl)

      fit_model_args <- args$fit_model_args
    } else {
      # A fresh run to generate the data
      if (args$progress) {
        cat("Simulate the data:\n")
      }
      sim_data_args <- list(nrep = args$nrep,
                            ptable = args$ptable,
                            model = args$model,
                            pop_es = args$pop_es,
                            es1 = args$es1,
                            es2 = args$es2,
                            es_ind = args$es_ind,
                            standardized = standardized,
                            n_std = n_std,
                            std_force_monte_carlo = std_force_monte_carlo,
                            n = args$n,
                            number_of_indicators = args$number_of_indicators,
                            reliability = args$reliability,
                            loading_difference = args$loading_difference,
                            reference = args$reference,
                            x_fun = args$x_fun,
                            e_fun = args$e_fun,
                            process_data = args$process_data,
                            iseed = args$iseed,
                            parallel = args$parallel,
                            progress = args$progress,
                            ncores = args$ncores,
                            n_ratio = args$n_ratio,
                            cl = cl)
    }
    data_all <- do.call(sim_data,
                        sim_data_args)

    fit_model_args <- fix_fit_model_args(fit_model_args)

    fit_args0 <- sapply(fit_model_args,
                        utils::modifyList,
                        val = list(parallel = args$parallel,
                                   progress = args$progress,
                                   ncores = args$ncores,
                                   cl = cl),
                        simplify = FALSE)

    if (args$progress) {
      cat("Fit the model(s):\n")
    }
    # fit_all is always a named list
    fit_all <- sapply(fit_args0,
                      function(x,
                               data_all) {
                        do.call(fit_model,
                                c(list(data_all = data_all),
                                  x))
                      },
                      data_all = data_all,
                      simplify = FALSE)
    # fit_all <- do.call(fit_model,
    #                    fit_args0)
    if (!is.null(args$R) && (args$ci_type == "mc")) {
      # - It's OK to reuse iseed because the processes are different.
      mc_args0 <- utils::modifyList(args$gen_mc_args,
                                    list(R = args$R,
                                        parallel = args$parallel,
                                        progress = args$progress,
                                        ncores = args$ncores,
                                        iseed = args$iseed,
                                        cl = cl))
      if (args$progress) {
        cat("Generate Monte Carlo estimates:\n")
      }
      # mc_all is always a named list
      mc_all <- sapply(fit_all,
                        function(x) {
                          do.call(gen_mc,
                                  c(list(fit_all = x),
                                    mc_args0))
                        },
                        simplify = FALSE)
      # mc_all <- do.call(gen_mc,
      #                   mc_args0)
    } else {
      mc_all <- rep(NA, length(fit_all[[1]]))
      mc_all <- sapply(names(fit_all),
                       function(x) mc_all,
                       simplify = FALSE)
    }
    # The first mc_all is always named "mc_out"
    tmp <- paste0(names(fit_all),
                  "_mc_out")
    tmp[1] <- "mc_out"
    names(mc_all) <- tmp

    if (!is.null(args$R) && (args$ci_type == "boot")) {
      # - It's OK to reuse the seed because the processes are different.
      boot_args0 <- utils::modifyList(args$gen_boot_args,
                                      list(R = args$R,
                                           parallel = args$parallel,
                                           progress = args$progress,
                                           ncores = args$ncores,
                                           iseed = args$iseed,
                                           cl = cl))
      if (args$progress) {
        cat("Generate bootstrap estimates:\n")
      }
      # boot_all is always a named list
      boot_all <- sapply(fit_all,
                        function(x) {
                          do.call(gen_boot,
                                  c(list(fit_all = x),
                                    boot_args0))
                        },
                        simplify = FALSE)
      # boot_all <- do.call(gen_boot,
      #                     boot_args0)
    } else {
      boot_all <- rep(NA, length(fit_all[[1]]))
      boot_all <- sapply(names(fit_all),
                         function(x) boot_all,
                         simplify = FALSE)
    }
    # The first boot_all is always named "boot_out"
    tmp <- paste0(names(fit_all),
                  "_boot_out")
    tmp[1] <- "boot_out"
    names(boot_all) <- tmp

    sim_all <- do.call(sim_out,
                       c(list(data_all = data_all),
                         fit_all,
                         mc_all,
                         boot_all))

  } else {
    sim_all <- object$sim_all
  }

  if (is.null(object$test_all) &&
      is.null(test_fun)) {
    do_the_test <- FALSE
  }

  if (update_data && !is.null(object$test_all)) {
    # Data updated and test exits.
    # Mandatory update of test to make the object
    # internally consistent.
    update_test <- TRUE
  }
  test_all <- NULL

  if (do_the_test && !update_test) {
    if (is.null(test_name)) {
      test_name <- deparse(substitute(test_fun))
      test_args_tmp <- utils::modifyList(test_args,
                                         list(get_test_name = TRUE))
      test_name0 <- tryCatch(do.call(test_fun,
                                     test_args_tmp),
                             error = function(e) e)
      if (!inherits(test_name0, "error")) {
        if (is.character(test_name0) &&
            length(test_name0) == 1) {
          test_name <- test_name0
        }
      }
    }
    if (args$progress) {
      cat("Do the test:",
          test_name,
          "\n")
    }
    tmp_args <- test_args
    tmp_args$get_map_names <- TRUE
    map_names0 <- tryCatch(do.call(test_fun, tmp_args),
                           error = function(e) e)
    if (!inherits(map_names, "error")) {
      if (is.character(map_names0) && !is.null(names(map_names0))) {
        map_names_user <- map_names
        map_names <- map_names0
        map_names[names(map_names_user)] <- map_names_user
      }
    }
    test_all <- do_test(sim_all,
                        test_fun = test_fun,
                        test_args = test_args,
                        map_names = map_names,
                        results_fun = results_fun,
                        results_args = results_args,
                        parallel = args$parallel,
                        progress = args$progress,
                        ncores = args$ncores,
                        cl = cl)
    attr(test_all, "test_note") <- test_note
    attr(test_all, "test_name") <- test_name
    test_all <- list(test_all)
    names(test_all) <- test_name
  }

  if (update_test) {
    if (args$progress) {
      cat("Update the test(s):\n")
    }
    # Only update tests. Ignore test arguments
    # Clear all argument values about test
    tmp <- formals(power4test)
    args$test_fun <- tmp$test_fun
    args$test_args <- tmp$test_args
    args$map_names <- tmp$map_names
    args$results_fun <- args$results_fun
    args$results_args <- args$results_args
    args$test_name <- args$test_name
    args$test_note <- args$test_note
    test_all <- sapply(object$test_all,
                       update_test_i,
                       sim_all = sim_all,
                       parallel = args$parallel,
                       progress = args$progress,
                       ncores = args$ncores,
                       cl = cl,
                       simplify = FALSE,
                       USE.NAMES = TRUE)

  }

  if (!update_power4test) {
    out <- list(sim_all = sim_all,
                test_all = test_all)
    attr(out, "args") <- args
    class(out) <- c("power4test", class(out))
  } else {
    attr(object, "args") <- args
    if (update_data) {
      object$sim_all <- sim_all
    }
    if (!is.null(test_all)) {
      if (update_test) {
        object$test_all <- test_all
      } else {
        object$test_all[[test_name]] <- test_all[[1]]
      }
    }
    out <- object
  }
  if (!is.null(out$test_all)) {
    if (!inherits(out$test_all, "test_out_list")) {
      class(out$test_all) <- c("test_out_list", class(out$test_all))
    }
  }
  out
}

#' @param digits The numbers of digits
#' displayed after the decimal.
#'
#' @param digits_descriptive The
#' number of digits displayed after
#' the decimal for the descriptive
#' statistics table.
#'
#' @param x The object
#' to be printed.
#'
#' @param what A string vector of
#' what to print, `"data"` for
#' simulated data and `"test"` for
#' stored test(s). Default is
#' `c("data", "test")`.
#'
#' @param data_long If `TRUE`, detailed
#' results will be printed when printing
#' the simulated data.
#'
#' @param test_long If `TRUE`, detailed
#' results will be printed when printing
#' test(s).
#'
#' @param fit_to_all_args A named list
#' of arguments to be passed to
#' [lavaan::sem()] when the model is
#' fitted to a sample combined from
#' all samples stored.
#'
#' @param ... Optional arguments to
#' be passed to other print methods
#'
#' @return
#' The `print` method of `power4test`
#' returns `x` invisibly. Called for
#' its side effect.
#'
#' @rdname power4test
#' @export
print.power4test <- function(x,
                             what = c("data", "test"),
                             digits = 3,
                             digits_descriptive = 2,
                             data_long = FALSE,
                             test_long = TRUE,
                             fit_to_all_args = list(),
                             ...) {
  what <- match.arg(what, several.ok = TRUE)
  if ("data" %in% what) {
    print(x$sim_all,
          data_long = data_long,
          digits = digits,
          digits_descriptive = digits_descriptive,
          fit_to_all_args = fit_to_all_args,
          ...)
  }
  if (("test" %in% what) &&
      !is.null(x$test_all)) {
    print(x$test_all,
          test_long = test_long,
          digits = digits,
          digits_descriptive = digits_descriptive,
          ...)
  }
  invisible(x)
}

#' @noRd
update_test_i <- function(test_i,
                          sim_all,
                          parallel = FALSE,
                          progress = FALSE,
                          ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                          cl = NULL) {
  test_fun <- attr(test_i, "test_fun")
  test_args <- attr(test_i, "test_args")
  map_names <- attr(test_i, "map_names")
  results_fun <- attr(test_i, "results_fun")
  results_args <- attr(test_i, "results_args")
  test_name <- attr(test_i, "test_name")
  if (progress) {
      cat("Update",
          test_name,
          ":\n")
  }
  test_note <- attr(test_i, "test_note")
  test_new <- do_test(sim_all,
                      test_fun = test_fun,
                      test_args = test_args,
                      map_names = map_names,
                      results_fun = results_fun,
                      results_args = results_args,
                      parallel = parallel,
                      progress = progress,
                      ncores = ncores,
                      cl = cl)
  attr(test_new, "test_note") <- test_note
  attr(test_new, "test_name") <- test_name

  test_new
}

#' @noRd
fix_fit_model_args <- function(fit_model_args) {
  # Conditions
  # Scenarios:
  # - [DONE] A list with no element.
  #   - list()
  #     - One fit model, and no additional arguments
  # - [DONE] A list with one element, named "fit"
  #   - list(fit = list())
  #   - list(fit = list(missing = "fiml", estimator = "MLR"))
  #     - One fit model, and these are the arguments
  #     - The first name must be "fit"
  # - [DONE] A list with more than one arguments. The first one is not named "fit"
  #   - list(args = list())
  #     - One fit model, and the value of args is a list.
  #   - list(missing = "fiml")
  #     - One fit model, and this is the arguments
  #   - list(missing = "fiml", estimator = "MLR")
  #     - One fit model, and these are the arguments
  # - [DONE] A list with more than one argument. The first one is named "fit"
  #   - list(fit = list(),
  #          fitx = list(missing = "fiml.x", estimator = "ML"))
  #     - Two or more fit models.
  #     - The first name must be "fit"
  #   - list(fit = list(missing = "fiml", estimator = "MLR"),
  #          fitx = list(missing = "fiml.x", estimator = "ML"))
  #     - Two or more fit models.
  #     - The first name must be "fit"
  if (length(fit_model_args) == 0) {
    # Zero length list
    fit_model_args <- list(fit = fit_model_args)
    return(fit_model_args)
  }
  list_names <- names(fit_model_args)
  if (any(list_names == "")) {
    # Do not support auto-naming. Too ambiguous.
    stop("All elements of 'fit_model_args' must be named.")
  }

  if (list_names[1] != "fit") {
    # if (length(fit_model_args) == 1) {
    #   return(fit_model_args)
    # }
    # This won't throw an error:
    # list(fit1 = list(missing = "fiml"),
    #      fitx = list(missing = "fiml.x",
    #                  estimator = "ML")))
    # But leave it this way for now. Users should not do this,
    # but there is no robust way to differentiat this from:
    # list(arg1 = list(),
    #      arg2 = list())
    fit_model_args <- list(fit = fit_model_args)
    return(fit_model_args)
  }

  # Two or more models
  fit_k <- all(sapply(fit_model_args,
                      is.list))
  if (!fit_k) {
    stop("The elements for all models must be lists.")
  } else {
    if (is.null(names(fit_model_args))) {
      # No longer support auto-naming.
      # Too ambiguous
      # This condition won't happen anyway,
      # because all top elements must be named.
      # Auto names
      names(fit_model_args) <- paste0("fit", seq_along(fit_model_args))
      names(fit_model_args)[1] <- "fit"
    } else {
      if (any(names(fit_model_args) %in% "")) {
        stop("All models in 'fit_model_args' must be named")
      }
      if (names(fit_model_args)[1] != "fit") {
        stop("The first model must be named 'fit'.")
      }
    }
    # Two or more named models, all named.
    return(fit_model_args)
  }

}

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test", {

simple_amp <- function(complete_data,
                       prop = .10) {
  out0 <- complete_data
  if ((prop > 0) && (prop < 1)) {
    n <- nrow(complete_data)
    p <- ncol(complete_data)
    nstart <- n * p
    n0 <- sample.int(nstart, size = ceiling(nstart * prop))
    m0 <- matrix(FALSE,
                 nrow = n,
                 ncol = p)
    m0[n0] <- TRUE
    out0[m0] <- NA
  }
  list(sth_missing = out0)
}

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 200,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 process_data = list(fun = simple_amp,
                                                     sim_data_name = "complete_data",
                                                     processed_data_name = "sth_missing",
                                                     args = list(prop = .10)),
                                 R = 50,
                                 do_the_test = FALSE,
                                 parallel = FALSE,
                                 progress = !is_testing(),
                                 iseed = 1234)
chk <- power_all_sim_only$sim_all[[1]]$mm_lm_dat_out
expect_equal(sum(is.na(chk)) / (nrow(chk) * ncol(chk)),
             .1)

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 200,
                                 process_data = list(fun = simple_amp,
                                                     sim_data_name = "complete_data",
                                                     processed_data_name = "sth_missing",
                                                     args = list(prop = .20)),
                                 R = 50,
                                 do_the_test = FALSE,
                                 parallel = FALSE,
                                 progress = !is_testing(),
                                 iseed = 1234)
chk <- power_all_sim_only$sim_all[[1]]$mm_lm_dat_out
expect_equal(sum(is.na(chk)) / (nrow(chk) * ncol(chk)),
             .2)

})

skip("A long test with parallel processing. Test interactively")

test_that("power4test: update respect parallel", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

# Generate the data

tmp1 <- system.time(
out <- power4test(
        nrep = 10,
        model = model_simple_med,
        pop_es = model_simple_med_es,
        n = 50,
        test_fun = test_parameters,
        test_args = list(op = ":="),
        parallel = TRUE,
        ncores = 2,
        progress = !is_testing(),
        iseed = 1234)
)
tmp1
expect_true(tmp1["elapsed"] > 2 * tmp1["user.self"])

tmp2 <- system.time(
out2 <- power4test(
          out,
          test_fun = test_indirect_effect,
          test_args = list(x = "x",
                       m = "m",
                       y = "y"),
          progress = !is_testing()
        )
)
expect_true(tmp2["elapsed"] > 2 * tmp2["user.self"])

})

library(testthat)

test_that("Model summary", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + cp*x
ab := a * b
total := a * b + cp
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "-s")
k <- c(y = 3,
       m = 4,
       x = 5)
rel <- c(y = .70,
         m = .80,
         x = .90)

sim_only <- power4test(nrep = 10,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 200,
                       R = 10,
                       number_of_indicators = k,
                       reliability = rel,
                       fit_model_args = list(estimator = "ML"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

sim_all <- sim_only$sim_all

expect_no_error(capture.output(print(sim_all)))

# Multigroup

model_simple_med2 <-
"
m ~ c(a1, a2)*x
y ~ c(b1, b2)*m + x
ab1 := a1 * b1
ab2 := a2 * b2
"

model_simple_med_es2 <- list("y ~ m" = "l",
                             "m ~ x" = "m",
                             "y ~ x" = c("-s", "m"))

rel2 <- list(y = .70,
             m = c(.80, .70),
             x = .90)

sim_gp02 <- power4test(nrep = 2,
                       model = model_simple_med2,
                       pop_es = model_simple_med_es2,
                       n = c(100, 200),
                       R = 5,
                       ci_type = "boot",
                       number_of_indicators = k,
                       reliability = rel2,
                       fit_model_args = list(estimator = "ML"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

expect_no_error(capture.output(print(sim_gp02$sim_all)))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("power4test", {

simple_amp <- function(complete_data,
                       prop = .10) {
  out0 <- complete_data
  if ((prop > 0) && (prop < 1)) {
    n <- nrow(complete_data)
    p <- ncol(complete_data)
    nstart <- n * p
    n0 <- sample.int(nstart, size = ceiling(nstart * prop))
    m0 <- matrix(FALSE,
                 nrow = n,
                 ncol = p)
    m0[n0] <- TRUE
    out0[m0] <- NA
  }
  list(sth_missing = out0)
}

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .70,
         m = .70,
         x = .70)

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 200,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 process_data = list(fun = simple_amp,
                                                     sim_data_name = "complete_data",
                                                     processed_data_name = "sth_missing",
                                                     args = list(prop = .10)),
                                 R = 50,
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only, data_long = TRUE),
              "Missing data")

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 200,
                                 process_data = list(fun = simple_amp,
                                                     sim_data_name = "complete_data",
                                                     processed_data_name = "sth_missing",
                                                     args = list(prop = .20)),
                                 R = 50,
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_output(print(power_all_sim_only, data_long = TRUE),
              "Missing data")

power_all_sim_only <- power4test(nrep = 10,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 n = 200,
                                 R = 50,
                                 do_the_test = FALSE,
                                 iseed = 1234,
                                 parallel = FALSE,
                                 progress = !is_testing())

expect_false(any(grepl("Missing Data",
                 capture.output(print(power_all_sim_only, data_long = TRUE)))))


})

library(testthat)

test_that("fix_fit_model_args", {
  # Conditions
  # Scenarios:
  # - [DONE] A list with no element.
  #   - list()
  #     - One fit model, and no additional arguments
  # - [DONE] A list with one element, named "fit"
  #   - list(fit = list())
  #   - list(fit = list(missing = "fiml", estimator = "MLR"))
  #     - One fit model, and these are the arguments
  #     - The first name must be "fit"
  # - [DONE] A list with more than one arguments. The first one is not named "fit"
  #   - list(args = list())
  #     - One fit model, and the value of args is a list.
  #   - list(missing = "fiml")
  #     - One fit model, and this is the arguments
  #   - list(missing = "fiml", estimator = "MLR")
  #     - One fit model, and these are the arguments
  # - [DONE] A list with more than one argument. The first one is named "fit"
  #   - list(fit = list(),
  #          fitx = list(missing = "fiml.x", estimator = "ML"))
  #     - Two or more fit models.
  #     - The first name must be "fit"
  #   - list(fit = list(missing = "fiml", estimator = "MLR"),
  #          fitx = list(missing = "fiml.x", estimator = "ML"))
  #     - Two or more fit models.
  #     - The first name must be "fit"
expect_identical(fix_fit_model_args(list()),
                 list(fit = list()))
expect_identical(fix_fit_model_args(list(fit = list())),
                 list(fit = list()))
expect_identical(fix_fit_model_args(list(fit = list(missing = "fiml"))),
                 list(fit = list(missing = "fiml")))
expect_identical(fix_fit_model_args(list(fit = list(missing = "fiml",
                                                    estimator = "MLR"))),
                 list(fit = list(missing = "fiml",
                                 estimator = "MLR")))
expect_identical(fix_fit_model_args(list(fit = list(missing = "fiml",
                                                    tmp = list(1, 2, 3)))),
                 list(fit = list(missing = "fiml",
                                 tmp = list(1, 2, 3))))
expect_identical(fix_fit_model_args(list(fit = list(tmp = list(1, 2, 3)))),
                 list(fit = list(tmp = list(1, 2, 3))))
expect_identical(fix_fit_model_args(list(args = list())),
                 list(fit = list(args = list())))
expect_identical(fix_fit_model_args(list(missing = "fiml")),
                 list(fit = list(missing = "fiml")))
expect_identical(fix_fit_model_args(list(missing = "fiml",
                                         estimator = "MLR")),
                 list(fit = list(missing = "fiml",
                                 estimator = "MLR")))
expect_identical(fix_fit_model_args(list(fit = list(),
                                         fitx = list(missing = "fiml.x",
                                                     estimator = "ML"))),
                 list(fit = list(),
                      fitx = list(missing = "fiml.x",
                                  estimator = "ML")))
expect_identical(fix_fit_model_args(list(fit = list(missing = "fiml"),
                                         fitx = list(missing = "fiml.x",
                                                     estimator = "ML"))),
                 list(fit = list(missing = "fiml"),
                      fitx = list(missing = "fiml.x",
                                  estimator = "ML")))
# This won't throw an error:
# list(fit1 = list(missing = "fiml"),
#      fitx = list(missing = "fiml.x",
#                  estimator = "ML")))
# But leave it this way for now. Users should not do this,
# but there is no robust way to differentiat this from:
# list(arg1 = list(),
#      arg2 = list())
# expect_error(fix_fit_model_args(list(fit1 = list(missing = "fiml"),
#                                      fitx = list(missing = "fiml.x",
#                                                  estimator = "ML"))))
})

library(testthat)
suppressMessages(library(lavaan))

# All-In-One

test_that("power4test: Test es", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")
k <- c(y = 3,
       m = 3,
       x = 3)
rel <- c(y = .80,
         m = .80,
         x = .80)

# Generate the data

power_all_sim_only <- power4test(nrep = 2,
                                 model = model_simple_med,
                                 pop_es = model_simple_med_es,
                                 es1 = c("n" = .30,
                                         "nil" = .30,
                                         "s" = .15,
                                         "m" = .20,
                                         "l" = .25),
                                 n = 10000,
                                 number_of_indicators = k,
                                 reliability = rel,
                                 fit_model_args = list(estimator = "ML"),
                                 do_the_test = FALSE,
                                 progress = !is_testing(),
                                 parallel = FALSE,
                                 iseed = 1234)
tmp <- fit_model_i(power_all_sim_only$sim_all[[1]])
std <- standardizedSolution(tmp, se = FALSE)
expect_equal(std$est.std[c(2, 1, 3)],
             c(.25, .20, .30),
             tolerance = 1e-1)

})

library(testthat)
suppressMessages(library(lavaan))

# All-In-One

test_that("power4test: k models", {

model_simple_med <-
"
m ~ a*x
y ~ b*m + x
ab := a * b
"

model_simple_med_es <- c("y ~ m" = "l",
                         "m ~ x" = "m",
                         "y ~ x" = "n")

model_simple_med_complete <-
"
m ~ a*x
y ~ b*m
ab := a * b
"

model_simple_med_test <-
"
m ~ a*x
y ~ 0*m
"


# Generate the data

power_all_sim_only_k1 <- power4test(nrep = 3,
                                    model = model_simple_med,
                                    pop_es = model_simple_med_es,
                                    n = 100,
                                    do_the_test = FALSE,
                                    progress = !is_testing(),
                                    parallel = FALSE,
                                    iseed = 1234)

power_all_sim_only_k2 <- power4test(nrep = 3,
                                    model = model_simple_med,
                                    pop_es = model_simple_med_es,
                                    n = 100,
                                    fit_model_args = list(fit = list(),
                                                          fit2 = list(model = model_simple_med_complete),
                                                          fit3 = list(model = model_simple_med_test) ),
                                    do_the_test = FALSE,
                                    progress = !is_testing(),
                                    parallel = FALSE,
                                    iseed = 1234)

power_all_test_only <- power4test(object = power_all_sim_only_k2,
                                  test_fun = test_parameters,
                                  test_args = list(pars = "b"),
                                  map_names = c(fit = "fit"),
                                  test_name = "Model 1")
power_all_test_only <- power4test(object = power_all_test_only,
                                  test_fun = test_parameters,
                                  test_args = list(pars = "b",
                                                   fit_name = "fit2"),
                                  map_names = c(fit = "fit2"),
                                  test_name = "Model 2")
power_all_test_only <- power4test(object = power_all_test_only,
                                  test_fun = test_parameters,
                                  test_args = list(pars = "a",
                                                   fit_name = "fit3"),
                                  map_names = c(fit = "fit3"),
                                  test_name = "Model 3")

if (!is_testing()) {
print(power_all_test_only$test_all,
      test_long = TRUE)
}

chk1 <- power_all_test_only$test_all$`Model 1`[[1]]$test_results$est
chk2 <- coef(power_all_test_only$sim_all[[1]]$extra$fit)["b"]
expect_equal(chk1,
             chk2,
             ignore_attr = TRUE)

chk1 <- power_all_test_only$test_all$`Model 2`[[1]]$test_results$est
chk2 <- coef(power_all_test_only$sim_all[[1]]$extra$fit2)["b"]
expect_equal(chk1,
             chk2,
             ignore_attr = TRUE)

chk1 <- power_all_test_only$test_all$`Model 3`[[1]]$test_results$est
chk2 <- coef(power_all_test_only$sim_all[[1]]$extra$fit3)["a"]
expect_equal(chk1,
             chk2,
             ignore_attr = TRUE)

})

library(testthat)

test_that("All parameters: lm: p-adjust", {

model_simple_med <-
"
m1 ~ x
m2 ~ x
m3 ~ x
y ~ m1 + m2 + m3 + x
"

model_simple_med_es <- c("y ~ m1" = "l",
                         "m1 ~ x" = "m",
                         "y ~ x" = "n")

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(p_adjust_method = "BH"))

p_org <- test_out$test_all[[1]][[2]]$test_results$pvalue_org
p_adj <- test_out$test_all[[1]][[2]]$test_results$pvalue

expect_equal(
  p.adjust(p_org, method = "BH"),
  p_adj
)

# Only some parameters are selected

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(p_adjust_method = "BH",
                                        par = c("m1~x", "m2~x", "y~m1", "y~x")))

p_org <- test_out$test_all[[1]][[2]]$test_results$pvalue_org
p_adj <- test_out$test_all[[1]][[2]]$test_results$pvalue

expect_equal(
  p.adjust(p_org, method = "BH"),
  p_adj
)

})

skip_on_cran()

library(testthat)

test_that("iseed in parallel", {

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es1 <- c("y ~ m" = "l",
                          "m ~ x" = "m",
                          "y ~ x" = "n")

sim_only1 <- power4test(nrep = 2,
                        model = model_simple_med,
                        pop_es = model_simple_med_es1,
                        n = 100,
                        R = 50,
                        ci_type = "mc",
                        fit_model_args = list(fit_function = "lavaan"),
                        do_the_test = FALSE,
                        progress = !is_testing(),
                        parallel = TRUE,
                        ncores = 2,
                        iseed = 1234)

sim_only2 <- power4test(nrep = 2,
                        model = model_simple_med,
                        pop_es = model_simple_med_es1,
                        n = 100,
                        R = 50,
                        ci_type = "mc",
                        fit_model_args = list(fit_function = "lavaan"),
                        do_the_test = FALSE,
                        progress = !is_testing(),
                        parallel = TRUE,
                        ncores = 2,
                        iseed = 1234)

expect_equal(sim_only1$sim_all[[1]]$extra$mc_out[[1]]$est,
             sim_only2$sim_all[[1]]$extra$mc_out[[1]]$est)

sim_only1 <- power4test(nrep = 2,
                        model = model_simple_med,
                        pop_es = model_simple_med_es1,
                        n = 100,
                        R = 50,
                        ci_type = "boot",
                        fit_model_args = list(fit_function = "lm"),
                        do_the_test = FALSE,
                        progress = !is_testing(),
                        parallel = TRUE,
                        ncores = 2,
                        iseed = 1234)

sim_only2 <- power4test(nrep = 2,
                        model = model_simple_med,
                        pop_es = model_simple_med_es1,
                        n = 100,
                        R = 50,
                        ci_type = "boot",
                        fit_model_args = list(fit_function = "lm"),
                        do_the_test = FALSE,
                        progress = !is_testing(),
                        parallel = TRUE,
                        ncores = 2,
                        iseed = 1234)

expect_equal(sim_only1$sim_all[[1]]$extra$boot_out[[1]]$est,
             sim_only2$sim_all[[1]]$extra$boot_out[[1]]$est)

})

library(testthat)

test_that("Test moderation: lm", {

mod <-
"
m ~ x + w1 + x:w1 + w3 + x:w3
y ~ m + w2 + m:w2 + x + w4 + m:w4
"

mod_es <- c("m ~ x" = "n",
            "y ~ x" = "m",
            "m ~ w1" = "n",
            "m ~ x:w1" = "l",
            "y ~ m:w2" = "-s")

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       progress = !is_testing(),
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_moderation,
                       test_args = list(p_adjust_method = "BH"))

p_org <- test_out$test_all[[1]][[1]]$test_results$pvalue_org
p_adj <- test_out$test_all[[1]][[1]]$test_results$pvalue

expect_equal(
  p.adjust(p_org, method = "BH"),
  p_adj
)

})

skip("A long test with parallel processing. Test interactively.")

library(testthat)

test_that("do_FUN with global environment functions", {

mod <-
"
m ~ x
y ~ m + x
"

mod_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

test_process <- function(data) {
  data$x <- data$x * 10
  data
}

# Simulate the data
out <- power4test(
         nrep = 2,
         model = mod,
         pop_es = mod_es,
         n = 5000,
         process_data = list(fun = "test_process"),
         test_fun = test_parameters,
         test_args = list(op = "~"),
         parallel = TRUE,
         iseed = 1234)

expect_equal(sd((out$sim_all[[1]]$mm_lm_dat_out)$x),
             10,
             tolerance = 1e-1)

})

skip_on_cran()

library(testthat)
suppressMessages(library(lavaan))

test_that("add_cov", {

model <-
"
m ~ x + z + x:z
y ~ m + x + w + m:w
"

model_es <-
"
m ~ x:z: s
m ~ x: m
y ~ w: s
y ~ m:w: l
y ~ m: m
y ~ x: s
"

model_es_mg <-
"
m ~ x:z: [s, l]
m ~ x: m
y ~ w: s
y ~ m:w: l
y ~ m: m
y ~ x: s
"

out <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es,
            n = 100,
            number_of_indicators = c(y = 4,
                                     x = 3,
                                     z = 5),
            reliability = c(x = .70,
                            z = .80,
                            y = .60),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

out_mg <- power4test(
            nrep = 2,
            model = model,
            pop_es = model_es_mg,
            n = 100,
            number_of_indicators = c(y = 4,
                                     w = 3,
                                     x = 5),
            reliability = c(x = .70,
                            w = .80,
                            y = .60),
            iseed = 1234,
            parallel = FALSE,
            progress = !is_testing())

sim_out <- out$sim_all[[1]]
names(sim_out)
model_original <- sim_out$model_original
model_final <- sim_out$model_final
expect_true(all(c("y2", "x3", "z5")
                %in%
                colnames(sim_out$mm_lm_dat_out)))
if (utils::packageVersion("lavaan") <= "0.6.19") {
  expect_true(is.data.frame(attr(model_final, "ptable")))
}

sim_out_mg <- out_mg$sim_all[[1]]
model_original_mg <- sim_out_mg$model_original
model_final_mg <- sim_out_mg$model_final
expect_true(all(c("w2", "x2", "y4")
                %in%
                colnames(sim_out_mg$mm_lm_dat_out)))
if (utils::packageVersion("lavaan") <= "0.6.19") {
  expect_true(is.data.frame(attr(model_final_mg, "ptable")))
}

})

#' @title Fit a Model to a List of Datasets
#'
#' @description Get the output of
#' [sim_data()] and fit a model to each
#' of the stored datasets.
#'
#' @details
#' By default, the function [fit_model()]
#'
#' - extracts the model
#' stored in the output of [sim_data()],
#'
#' - fits the model to each dataset
#' simulated using `fit_function`,
#' default to `"lavaan"` and
#' [lavaan::sem()] will be called,
#'
#' - and returns the results.
#'
#' If the datasets
#' were generated from a multigroup
#' model when calling [sim_data()],
#' a multigroup model is fitted.
#'
#' # The role of `fit_model()`
#'
#' This function is used by the
#' all-in-one function [power4test()].
#' Users usually do not call this
#' function directly, though
#' developers can use this function to
#' customize the model fitting step in
#' power analysis.
#'
#' @seealso See [power4test()] for
#' the all-in-one function that uses
#' this function, and [sim_data()]
#' for the function generating datasets
#' for this function.
#'
#' @return
#' An object of the class `fit_out`,
#' which is a list of the output of
#' `fit_function` ([lavaan::sem()]
#' by default). If an error occurred
#' when fitting the model to a dataset,
#' then this element will be the error
#' message from the fit function.
#'
#' @param data_all The output
#' of [sim_data()], or a `sim_data`
#' class object.
#'
#' @param model The model to be fitted.
#' If `NULL`, the default, the model
#' stored in `data_all`, which should
#' be the data generation model,
#' will be used.
#'
#' @param fit_function The function to
#' be used to fit the model. Can also
#' be a string: `"lavaan"` (the default) for
#' [lavaan::sem()], and `"lm"` or `many_lm`
#' for [lmhelprs::many_lm()].
#' Other functions can also be used
#' if necessary.
#'
#' @param arg_data_name The name of the
#' argument of `fit_function` expecting
#' the dataset. Default is `"data"`.
#'
#' @param arg_model_name The name of
#' the argument of `fit_function`
#' expecting the model definition.
#' Default is `"model"`.
#'
#' @param arg_group_name The name of
#' the argument of `fit_function`
#' expecting the name of the group
#' variable. Used only for multigroup
#' models. Default is `"group"`.
#'
#' @param ... Optional arguments to be
#' passed to `fit_function` when
#' fitting the model.
#'
#' @param fit_out If set to a `fit_out`
#' object (the output of [fit_model()]),
#' then all missing arguments will be
#' retrieved from `fit_out`. That is,
#' users can use `fit_model(data_all = new_data, fit_out = old_out)`
#' to re-fit a model originally fitted
#' in `old_out` on a new list of dataset
#' (`new_data`). No need to include
#' all other arguments.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to fit the
#' models. Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' of model fitting will be displayed.
#' Default is `FALSE.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param cl A cluster, such as one created
#' by [parallel::makeCluster()]. If `NULL`,
#' a cluster will be created, but will be
#' stopped on exit. If set to an existing
#' cluster, it will not be stopped when
#' the function exits; users need to
#' stop it manually.
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "m ~ x
#'  y ~ m + x"
#'
#' # Specify the effect sizes (population parameter values)
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate several simulated datasets
#'
#' data_all <- sim_data(nrep = 5,
#'                      model = mod,
#'                      pop_es = es,
#'                      n = 100,
#'                      iseed = 1234)
#'
#' # Fit the population model to each datasets
#'
#' fit_all <- fit_model(data_all)
#' fit_all[[1]]
#'
#' # Fit the population model using the MLR estimator
#'
#' fit_all_mlr <- fit_model(data_all,
#'                          estimator = "MLR")
#' fit_all_mlr[[1]]
#'
#' # Fit a model different from the population model,
#' # with the MLR estimator
#'
#' mod2 <-
#' "m ~ x
#'  y ~ m"
#'
#' fit_all_mlr2 <- fit_model(data_all,
#'                           mod2,
#'                           estimator = "MLR")
#' fit_all_mlr2[[1]]
#'
#' @export
fit_model <- function(data_all = NULL,
                      model = NULL,
                      fit_function = "lavaan",
                      arg_data_name = "data",
                      arg_model_name = "model",
                      arg_group_name = "group",
                      ...,
                      fit_out = NULL,
                      parallel = FALSE,
                      progress = FALSE,
                      ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                      cl = NULL) {
  # Store the arguments such the
  # it can be updated with new data.
  # It is intentional not to store the call.

  # This check can be removed after lmhelprs is updated on CRAN.
  # The version requirement should be added to DESCRIPTION later.
  tmp <- deparse(substitute(fit_function))
  if (isTRUE(grepl("lmhelprs", tmp)) || identical("lm", tmp)) {
    lmhelprs_supported <- (utils::packageVersion("lmhelprs") >= "0.4.2")
    if (isFALSE(lmhelprs_supported)) {
      stop("lmhelprs 0.4.2 or later is required. ",
          "If not available from CRAN, can be installed from ",
          "GitHub: remotes::install_github('sfcheung/lmhelprs')")
    }
  }

  update_fit <- FALSE
  if (!is.null(fit_out)) {
    if (inherits(fit_out, "fit_out")) {
      update_fit <- TRUE
    } else {
      stop("'fit_out' is not a supported object.")
    }
  }
  if (!update_fit) {
    # Store the evaluated arguments
    args <- formals(fit_model)
    args$`...` <- NULL
    args <- lapply(args,
                   eval,
                   envir = parent.frame())
  } else {
    args <- attr(fit_out, "args")
  }
  my_call <- match.call()
  call_args <- as.list(my_call)[-1]
  i <- which(names(call_args) == "data_all")
  if (length(i) != 0) {
    call_args <- call_args[-i]
  }
  call_args <- lapply(call_args,
                      eval,
                      envir = parent.frame())
  args <- utils::modifyList(args,
                            as.list(call_args))
  args$fit_out <- NULL
  args$data_all <- NULL
  # args available in all cases.
  # It should be used whenever possible,
  # unless we explicitly need the value in this call.
  # Parallel *will* be used, though
  # specified by args.
  out <- do.call(do_FUN,
                 c(list(X = data_all,
                        FUN = fit_model_i),
                   args))
  class(out) <- c("fit_out", class(out))
  attr(out, "args") <- args
  return(out)
}


#' @noRd
fit_model_i <- function(data_i,
                        model = NULL,
                        fit_function = "lavaan",
                        arg_data_name = "data",
                        arg_model_name = "model",
                        arg_group_name = "group",
                        ...) {
  if (is.character(fit_function)) {
    fit_function_org <- fit_function
    fit_function <- switch(fit_function,
                           lavaan = lavaan::sem,
                           lm = lmhelprs::many_lm,
                           many_lm = lmhelprs::many_lm,
                           fit_function)

  } else {
    fit_function_org <- character(0)
  }
  # Anomalies should be checked in
  # subsequent steps, not during fitting
  # the model to many datasets.
  # TODO:
  # - Update the following lines to
  #   support using a parameter table.
  if (is.null(model)) {
    model_to_fit <- data_i$model_final
  } else {
    model_to_fit <- model
  }
  tmp <- attr(model_to_fit, "ptable")
  if (!is.null(tmp) &&
      is.null(data_i$number_of_indicators[[1]]) &&
      identical(getNamespaceName(environment(fit_function)),
                c(name = "lavaan"))) {
    # Do this only for lavaan
    # TODO:
    # - Support moderated mediation model with indicators.
    model_to_fit <- tmp
  }

  # Fix the model if lm() is used to fitting the model
  if (fit_function_org %in% c("lm", "many_lm")) {
    model_to_fit <- fix_many_lm_model(model_to_fit)
  }

  # For single-group models,
  # data_i$group_name would be NULL.
  fit_args0 <- list()
  fit_args0[[arg_model_name]] <- model_to_fit
  fit_args0[[arg_data_name]] <- data_i$mm_lm_dat_out
  fit_args0[[arg_group_name]] = data_i$group_name
  fit_args <- utils::modifyList(list(...),
                                fit_args0)
  fit <- tryCatch(suppressWarnings(do.call(fit_function,
                                           fit_args)),
                  error = function(e) e)
  if (inherits(fit, "lavaan")) {
    fit@external$fit_external <- data_i$fit_external
  } else {
    try({attr(fit, "fit_external") <- data_i$fit_external},
        silent = TRUE)
  }
  return(fit)
}

#' @title Helpers for the Boos-and-Zhang (2000) Method
#'
#' @description Helpers for the method
#' by Boos and Zhang (2000) for estimating
#' rejection rate for resampling-based
#' tests.
#'
#' @details Boos and Zhang (2000) proposed
#' a method to estimate the rejection
#' rate (power, if the null hypothesis is false)
#' for methods based on resampling, such
#' as nonparametric bootstrapping. This
#' method is used by some functions
#' in `power4mome`, such as [x_from_power()].
#'
#' This method is implemented internally.
#' Some helper functions regarding this
#' method is exported for users.
#'
#' @references
#' Boos, D. D., & Zhang, J. (2000).
#' Monte Carlo evaluation of
#' resampling-based hypothesis tests.
#' *Journal of the American Statistical Association*,
#' *95*(450), 486--492.
#' \doi{10.1080/01621459.2000.10474226}
#'
#' @seealso [x_from_power()]
#'
#' @name bz_helpers
NULL

#' @details
#' The function [Rs_bz_supported()] returns
#' the number of bootstrap samples (for bootstrapping)
#' or simulated samples (for Monte Carlo),
#' both called resamples below for brevity,
#' usually specified by the argument `R`,
#' that can be
#' used for the Boos-Zhang-2000 method,
#' given the desired two-tailed level of significance,
#' (.05 by default).
#' If possible, setting the number of resamples.
#' (e.g., setting `R` when calling [x_from_power()])
#' will automatically enable the Boos-Zhang-2000
#' method (unless explicitly turned off
#' by setting the option `"power4mome.bz"`
#' to `FALSE` by `options("power4mome.bz") <- FALSE`),
#' substantially reducing the processing
#' time. For now, only two-tailed tests are supported.
#'
#' @return
#' The function [Rs_bz_supported()] returns
#' a numeric vector of the numbers of
#' resamples supported.
#'
#' @param alpha The level of significance,
#' two-tailed.
#'
#' @param Rmax The maximum number of resamples
#' to be returned. Default is 359.
#' Though it is possible to use 1999
#' resamples or even more with Boos-Zhang-2000,
#' using such a large number of resamples defeats
#' the goal to reduce processing time.
#'
#'
#' @examples
#' # === Rs_bz_supported ===
#'
#' # alpha = .05
#' Rs_bz_supported()
#'
#' # alpha = .01
#' Rs_bz_supported(alpha = .01)
#'
#' @rdname bz_helpers
#' @export
Rs_bz_supported <- function(
  alpha = .05,
  Rmax = getOption(
            "power4mome.bz_Rmax",
            default = 359)
) {
  R_extrapolate(alpha = alpha,
                Rmax = Rmax)
}

#' @details
#' Given a target maximum number of
#' resamples and a level of significance,
#' the function [R_for_bz()] returns
#' largest number of resamples that can be
#' used for the Boos-Zhang-2000 method.
#' This function can be used for
#' arguments such as `R` in [x_from_power()]
#' to automatically find the largest
#' value supported by the Boos-Zhang-2000
#' method.
#'
#' @param R_target The target maximum
#' number of resamples.
#'
#' @return
#' The function [R_for_bz()] returns
#' a scalar.
#'
#' @examples
#' # === R_for_bz ===
#'
#' R_for_bz(200)
#' R_for_bz(500)
#'
#' @rdname bz_helpers
#' @export
R_for_bz <- function(
  R_target,
  alpha = .05
) {
  Rs <- R_extrapolate(
          alpha = alpha,
          Rmax = R_target
        )
  if (length(Rs) > 0) {
    return(max(Rs))
  } else {
    stop("No supported values")
  }
}

#' @title Random Variable From a Generalized Normal Distribution
#'
#' @description Generate random numbers
#' from generalized normal distribution,
#' rescaled to have user-specified
#' population mean and standard
#' deviation.
#'
#' @details
#' First, specify the parameter `p`
#' and the desired population
#' mean and standard deviation. The
#' random numbers, drawn from a
#' generalized normal distribution by
#' [pgnorm::rpgnorm()], will then be
#' rescaled with the desired population
#' mean and standard.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param p The parameter of the
#' distribution. Must be a positive
#' non-zero number. Default is 2,
#' resulting in a normal distribution.
#' Higher than 2 results in negative
#' excess kurtosis. Lower than 2 results
#' in positive excess kurtosis.
#'
#' @param pmean Population mean.
#'
#' @param psd Population standard
#' deviation.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rpgnorm_rs(n = 5000,
#'                 p = 2,
#'                 pmean = 0,
#'                 psd = 1)
#' mean(x)
#' sd(x)
#' hist(x)
#' x_kurt <- function(p) {gamma(5/p)*gamma(1/p)/(gamma(3/p)^2) - 3}
#'
#' p <- 6
#' x <- rpgnorm_rs(n = 50000,
#'                 p = p,
#'                 pmean = 0,
#'                 psd = 1)
#' mean(x)
#' sd(x)
#' x_kurt(p)
#' qqnorm(x); qqline(x)
#'
#' p <- 1
#' x <- rpgnorm_rs(n = 50000,
#'                 p = p,
#'                 pmean = 0,
#'                 psd = 1)
#' mean(x)
#' sd(x)
#' x_kurt(p)
#' qqnorm(x); qqline(x)
#' @export

rpgnorm_rs <- function(n = 10,
                       p = 2,
                       pmean = 0,
                       psd = 1) {
    t_sd <- sqrt(gamma(3 / p) / gamma(1 / p)) * (p ^ (1 / p))
    pmean + psd * pgnorm::rpgnorm(n, p) / t_sd
  }

#' @title Random Variable From a Beta Distribution (User Range)
#'
#' @description Generate random numbers
#' from a beta distribution, rescaled to
#' have user-specified population mean
#' and standard deviation, and within a
#' specific range.
#'
#' @details
#' First, specify the two parameters,
#' `shape1` and `shape2`, and the
#' desired population mean and standard
#' deviation. The random numbers, drawn
#' from a beta distribution by
#' [stats::rbeta()] will then be
#' rescaled to the desired population range.
#'
#' @param n The number of random numbers
#' to generate.
#'
#' @param bmean The population mean.
#'
#' @param bsd The population standard
#' deviation. If `bsd` is zero or
#' negative, all random numbers will be
#' equal to `bmean`.
#'
#' @param blow The lower bound of the
#' target range.
#'
#' @param bhigh The upper bound of the
#' target range.
#'
#' @return
#' A vector of the generated
#' random numbers.
#'
#' @examples
#' set.seed(90870962)
#' x <- rbeta_rs2(n = 5000,
#'                bmean = .80,
#'                bsd = .10,
#'                blow = .00,
#'                bhigh = .95)
#' mean(x)
#' sd(x)
#' hist(x)
#' y <- rbeta_rs2(n = 5000,
#'                bmean = 4,
#'                bsd = 3,
#'                blow = -10,
#'                bhigh = 10)
#' mean(y)
#' sd(y)
#' hist(y)
#'
#' @export

rbeta_rs2 <- function(n = 10,
                      bmean,
                      bsd,
                      blow = 0,
                      bhigh = 1) {
  if (bsd > 0) {
    brange <- bhigh - blow
    mean1 <- (bmean - blow) / brange
    var1 <- (bsd^2) / (brange^2)
    d <- (1 - mean1) / mean1
    myalpha <- mean1 * (mean1 * (1 - mean1) / var1 - 1)
    mybeta <- myalpha * d
    results <- stats::rbeta(n,
                            myalpha,
                            mybeta) * brange + blow
  }

  if (bsd <= 0) {
    results <- rep(bmean, n)
  }

  results
}

#' @noRd
# Find the approximate chance to have
# a sample power "beyond" the target power
# if the population power is power_i,
# given the "sample size" (trial_nrep).
p_c0 <- function(
  target_power,
  power_i,
  trial_nrep
) {
  if (target_power == power_i) {
    return(.50)
  }
  f <- function(level) {
    a <- reject_ci_wilson(
            nreject = trial_nrep * power_i,
            nvalid = trial_nrep,
            level = level)[, ifelse(power_i < target_power, "cihi", "cilo")]
    a - target_power
  }
  tmp1 <- stats::uniroot(
          f,
          interval = c(0.0000001, .999999)
        )
  level_out <- tmp1$root
  p_c <- 1 - (1 - level_out) / 2
  p_c
}

#' @noRd
# Find the approximate chance to have
# a sample power "beyond" the target power
# if the population power is power_i,
# given the "sample size" (trial_nrep).
p_c <- function(
  target_power,
  power_i,
  goal,
  what,
  trial_nrep,
  tolerance = .001,
  level = .95,
  final_nrep
) {
  if (what == "point") {
    out <- p_c0(
        target_power = target_power,
        power_i = power_i,
        trial_nrep = trial_nrep
      )
    return(out)
  }
  target_power_b <- target_power_adjusted(
      target_power = target_power,
      goal = goal,
      what = what,
      tolerance = tolerance,
      nrep = final_nrep,
      level = level
    )
  out <- p_c0(
      target_power = target_power_b,
      power_i = power_i,
      trial_nrep = trial_nrep
    )
  out
}

#' @noRd
# Form a highest-density region(s)
hdr <- function(
  dfun,
  prob = .80
) {
  x <- dfun[, "x", drop = TRUE]
  pbi <- dfun[, "prob", drop = TRUE]
  po <- order(pbi)
  for (i in po) {
    pbi <- pbi - pbi[i]
    if (sum(pbi) <= prob) break
  }
  a <- pbi > 0
  # TODO:
  # - There should be an easier way to do this
  d0 <- vector("numeric", length = length(a))
  for (i in seq_along(a)) {
    if (i == 1) {
      d0[i] <- 1
    } else {
      if (isTRUE(xor(a[i], a[i - 1]))) {
        d0[i] <- d0[i - 1] + 1
      } else {
        d0[i] <- d0[i - 1]
      }
    }
  }
  h1 <- tapply(
          x,
          INDEX = d0,
          FUN = range
        )
  h2 <- tapply(
          a,
          INDEX = d0,
          FUN = \(x) isTRUE(x[1])
        )
  h3 <- tapply(
          pbi,
          INDEX = d0,
          FUN = sum
        )
  out <- unname(h1[h2])
  out <- .mapply(
            \(x, y) {
              attr(x, "pb") <- y
              x
            },
            list(x = out,
                y = h3[h2]),
            MoreArgs = NULL
          )
}

#' @noRd
# Check whether a value of x has already been tried
in_x_tried_advanced <- function(
                      test_x,
                      object,
                      x,
                      nrep = NULL,
                      seed_exclude = NULL) {
  # If yes, return the index
  # Otherwise, return NA
  x_tried <- get_x_tried(object = object,
                         x = x)
  reject_ci <- rejection_rates_add_ci(
                    object,
                    add_se = FALSE,
                    add_reject = FALSE)
  nrep_tried <- reject_ci$nrep
  seed_tried <- sapply(object,
                  \(x) attr(x, "args")$iseed)

  i1 <- which(x_tried %in% test_x)

  if (length(i1) == 0) {
    return(NA)
  }

  # ==== Exclude xs used (seeds in seed_exclude)

  if (!is.null(seed_exclude)) {
    seed_exclude <- seed_exclude[!is.na(seed_exclude)]
    if (length(seed_exclude) > 0) {
      i2 <- which(!(seed_tried %in% seed_exclude))
      i1 <- intersect(i1, i2)
    }
  }

  # ==== Keep only xs with the required nrep ====

  if (!is.null(nrep)) {
    i3 <- which(nrep_tried %in% nrep)
    i1 <- intersect(i1, i3)
  }

  # If multiple matches, select the first.
  # This is OK because it will have the
  # required nrep but has not been used.

  if (length(i1) > 0) {
    return(i1[1])
  } else {
    return(NA)
  }

  return(NA)
}

library(testthat)

test_that("p_c", {

tmp1 <- p_c0(
    target_power = .80,
    power_i = .70,
    trial_nrep = 50
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp2 <- reject_ci_wilson(
          nreject = .70 * 50,
          nvalid = 50,
          level = tmp1
        )
expect_all_equal(tmp2[, "cihi"],
                 .80)

tmp1 <- p_c0(
    target_power = .80,
    power_i = .85,
    trial_nrep = 100
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp2 <- reject_ci_wilson(
          nreject = .85 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_all_equal(tmp2[, "cilo"],
                 .80)

tmp1 <- p_c(
    target_power = .80,
    power_i = .85,
    trial_nrep = 100,
    goal = "close_enough",
    what = "point"
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp2 <- reject_ci_wilson(
          nreject = .85 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_all_equal(tmp2[, "cilo"],
                 .80)

# ==== ub =====

tmp1 <- p_c(
    target_power = .80,
    power_i = .75,
    trial_nrep = 100,
    goal = "close_enough",
    what = "ub",
    final_nrep = 2000
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp3 <- target_power_adjusted(
    target_power = .80,
    goal = "close_enough",
    what = "ub",
    nrep = 2000
  )

tmp2 <- reject_ci_wilson(
          nreject = .75 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_equal(
          unname(tmp2[, "cihi"]),
          tmp3,
          tolerance = 1e-2)


tmp1 <- p_c(
    target_power = .80,
    power_i = .79,
    trial_nrep = 100,
    goal = "close_enough",
    what = "ub",
    final_nrep = 2000
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp3 <- target_power_adjusted(
    target_power = .80,
    goal = "close_enough",
    what = "ub",
    nrep = 2000
  )

tmp2 <- reject_ci_wilson(
          nreject = .79 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_equal(
          unname(tmp2[, "cilo"]),
          tmp3,
          tolerance = 1e-2)

# ==== lb =====

tmp1 <- p_c(
    target_power = .80,
    power_i = .85,
    trial_nrep = 100,
    goal = "close_enough",
    what = "lb",
    final_nrep = 2000
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp3 <- target_power_adjusted(
    target_power = .80,
    goal = "close_enough",
    what = "lb",
    nrep = 2000
  )

tmp2 <- reject_ci_wilson(
          nreject = .85 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_equal(
          unname(tmp2[, "cilo"]),
          tmp3,
          tolerance = 1e-2)


tmp1 <- p_c(
    target_power = .80,
    power_i = .70,
    trial_nrep = 100,
    goal = "close_enough",
    what = "lb",
    final_nrep = 2000
  )
tmp1 <- 1 - (1 - tmp1) * 2

tmp3 <- target_power_adjusted(
    target_power = .80,
    goal = "close_enough",
    what = "lb",
    nrep = 2000
  )

tmp2 <- reject_ci_wilson(
          nreject = .70 * 100,
          nvalid = 100,
          level = tmp1
        )
expect_equal(
          unname(tmp2[, "cihi"]),
          tmp3,
          tolerance = 1e-2)


})

# ==== HDR ====

test_that("HDR", {

# set.seed(1234)
# x <- rnorm(10000, mean = 500, sd = 10)
# x <- round(x)
# x <- table(x)
# plot(x)

x <- 10:2000
d1 <- dnorm(x, mean = 500, sd = 100)
d2 <- dnorm(x, mean = 1000, sd = 20)
range(d1)
range(d2)
d <- d1 + d2
range(d)
d <- d / sum(d)
range(d)
dfun1 <- cbind(x = x,
               prob = d)
# plot(dfun1)

x <- 10:2000
d1 <- dnorm(x, mean = 500, sd = 100)
range(d1)
d <- d1 / sum(d1)
range(d)
dfun2 <- cbind(x = x,
               prob = d)
# plot(dfun2)

out1 <- hdr(dfun1, prob = .8)
out2 <- hdr(dfun1, prob = .7)

expect_true(length(out1) == 2)
expect_true(length(out2) == 2)

expect_true(diff(out1[[1]]) > diff(out2[[1]]))
expect_true(diff(out1[[2]]) > diff(out2[[2]]))

out1 <- hdr(dfun2, prob = .8)
out2 <- hdr(dfun2, prob = .7)

expect_true(length(out1) == 1)
expect_true(length(out2) == 1)

expect_true(diff(out1[[1]]) > diff(out2[[1]]))


})

#' @title Generate Bootstrap Estimates
#'
#' @description Get a list of the output
#' of [lavaan::sem()] or [lmhelprs::many_lm()]
#' and generate
#' bootstrap estimates of model
#' parameters.
#'
#' @details
#' The function [gen_boot()]
#' simply calls [manymome::do_boot()]
#' on each output of
#' [lavaan::sem()] or [lmhelprs::many_lm()]
#' in `fit_all`. The
#' simulated
#' estimates can then be used to test
#' effects such as indirect effects,
#' usually by functions from the
#' `manymome` package, such as
#' [manymome::indirect_effect()].
#'
#' # The role of `gen_boot()`
#'
#' This function is used by the
#' all-in-one function [power4test()].
#' Users usually do not call this
#' function directly, though
#' developers can use this function to
#' customize the workflow of the
#' power analysis.
#'
#' @seealso See [power4test()] for
#' the all-in-one function that uses
#' this function.
#'
#' @param fit_all The output of
#' [fit_model()] or an object of the
#' class `fit_out`.
#'
#' @param R The number of replications
#' to generate the bootstrap estimates
#' for each fit output.
#'
#' @param ... Optional arguments to be
#' passed to [manymome::do_boot()] when
#' generating the bootstrap estimates.
#'
#' @param iseed The seed for the random
#' number generator. Default is `NULL`
#' and the seed is not changed.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to generate
#' bootstrap estimates for the
#' fit outputs. Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' will be displayed.
#' Default is `FALSE.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param compute_implied_stats Whether
#' implied statistics are computed
#' in each bootstrap samples. Usually
#' not needed and so default to `FALSE`
#'
#' @param cl A cluster, such as one created
#' by [parallel::makeCluster()]. If `NULL`,
#' a cluster will be created, but will be
#' stopped on exit. If set to an existing
#' cluster, it will not be stopped when
#' the function exits; users need to
#' stop it manually.
#'
#' @return
#' A `boot_list` object, which is a list
#' of the output of [manymome::do_boot()].
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "m ~ x
#'  y ~ m + x"
#'
#' # Specify the effect sizes (population parameter values)
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate several simulated datasets
#'
#' data_all <- sim_data(nrep = 2,
#'                      model = mod,
#'                      pop_es = es,
#'                      n = 50,
#'                      iseed = 1234)
#'
#' # Fit the population model to each datasets
#'
#' fit_all <- fit_model(data_all)
#'
#' # Generate bootstrap estimates for each replication
#'
#' boot_all <- gen_boot(fit_all,
#'                      R = 10,
#'                      iseed = 4567)
#' boot_all
#'
#'
#' @export
#'
gen_boot <- function(fit_all,
                     R = 100,
                     ...,
                     iseed = NULL,
                     parallel = FALSE,
                     progress = FALSE,
                     ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                     compute_implied_stats = FALSE,
                     cl = NULL) {
  # This check can be removed after manymome is updated on CRAN.
  # The version requirement should be added to DESCRIPTION later.
  manymome_implied <- (utils::packageVersion("manymome") >= "0.2.7.1")
  if (isFALSE(manymome_implied)) {
    stop("manymome 0.2.7.1 or later is required. ",
         "If not available from CRAN, can be installed from ",
         "GitHub: remotes::install_github('sfcheung/manymome')")
  }
  out <- do_FUN(X = fit_all,
                FUN = gen_boot_i,
                R = R,
                ...,
                compute_implied_stats = compute_implied_stats,
                iseed = iseed,
                parallel = parallel,
                progress = progress,
                ncores = ncores,
                cl = cl)
  class(out) <- c("boot_list", class(out))
  return(out)
}

#' @noRd
gen_boot_i <- function(fit_i,
                       R = 100,
                       seed = NULL,
                       compute_implied_stats = FALSE,
                       ...) {
  boot_out <- tryCatch(manymome::do_boot(fit = fit_i,
                                         R = R,
                                         ...,
                                         compute_implied_stats = compute_implied_stats,
                                         parallel = FALSE,
                                         progress = FALSE),
                      error = function(e) e)
  boot_out
}

# An algorithm function should named alg_*
# These are the required arguments
# - object
# - x
# - pop_es_names
# - ... # To force other arguments to be named
# - target_power
# - x_max
# - x_min
# - progress
# - x_interval
# - x_include_interval
# - simulation_progress
# - save_sim_all
# - is_by_x
# - object_by_org
# - final_nrep
# - final_R
# - ci_level
# - extendInt
# - max_trials
# - R

# === power_curve ===
# - xs_per_trial
# - nrep0
# - R0
# - power_model
# - start
# - lower_bound
# - upper_bound
# - nls_control
# - nls_args
# - nrep_steps
# - final_R
# - final_xs_per_trial
# - pre_i_xs
# - pre_i_nrep
# - pre_i_R
# - power_min
# - power_max
# - power_tolerance_in_interval
# - power_tolerance_in_final
# - delta_tol

# === bisection ===
# - digits
# - lower_hard
# - upper_hard
# - extend_maxiter
# - what
# - goal
# - tol
# - variants

# These are the required elements in the output
# - by_x_1
# - fit_1
# - x_tried
# - x_out
# - power_out
# - nrep_out
# - ci_out
# - by_x_out
# - i2
# - ci_hit
# - solution_found
# - what
# - goal

# ** by_x_1 **
# The collection of all values tried and their results
# to be updated when new value is tried.
# Used after the end of the loop.

# ** fit_1 **
# The latest power curve
# To be updated whenever by_x_1 is updated.
# Used after the end of the loop.

# ** x_tried **
# The sample sizes or population values
# examined.

# ** x_out **
# The output in the last trial.

# ** power_out **
# The power in the last trial.

# ** nrep_out **
# The nrep in the last trial.

# ** ci_out **
# The ci in the last trial.

# ** by_x_out **
# The by_* output in the last trial

# ** i2 **
# The row number of the last trial
# in by_x_1.

# ** ci_hit **
# TRUE if the CI in by_x_1 contains the target_power

# ** solution_found **
# TRUE if an acceptable solution is found

# These are the required elements in the output for goal = close_enough
# - tol

# ** tol **
# The tolerance used when goal = close_enough.
# This is required to check whether a value is a solution.

# Functions for the Boos-Zhang method

# Functions affected by this methods
# - summarize_one_test_vector()
# - rejection_rates_i_vector()
# - test_indirect_effect()
# - test_index_of_mome()
# - test_cond_indirect()
#
# - summarize_one_test_data_frame()
# - rejection_rates_i_data_frame()
# - test_k_indirect_effects()
# - test_cond_indirect_effects()

#' @noRd
merge_for_collapse <- function(out0) {
  outnames <- colnames(out0[[1]])
  a <- lapply(out0,
              \(x) split(x, seq_len(nrow(x))))
  f <- function(...,
                cnames = NULL) {
    args <- list(...)
    out <- do.call(
            rbind,
            args)
    colnames(out) <- cnames
    out
  }
  b <- do.call(mapply,
               c(list(FUN = f),
                 a,
                 list(MoreArgs = list(cnames = outnames),
                      SIMPLIFY = FALSE)))
}

#' @noRd
add_bz_i <- function(outi) {
  # Receive a table or a vector
  # Add bz_pvalues from R and nlt0
  out_type <- ifelse(is.null(dim(outi)),
                     yes = "vector",
                     no = "table")
  if (out_type == "vector") {
    outz1 <- rbind(outi)
  } else {
    outz1 <- outi
    if (!inherits(outz1, "data.frame")) {
      outz1 <- as.data.frame(outz1)
    }
  }
  outz1 <- split(outz1,
                 seq_len(nrow(outz1)),
                 drop = FALSE)
  R <- sapply(outz1,
              \(x) x[,"R"])
  if (length(unique(R)) != 1) {
    return(outi)
  } else {
    R <- unique(R)
  }
  R_case <- sapply(R,
                    bz_case)
  if (!all(R_case == "one")) {
    return(outi)
  }
  Rext <- R_extrapolate()
  R_case <- R_case[1]
  Rk <- Rext[seq(1, which(Rext == R) - 1)]
  for (j1 in seq_along(outz1)) {
    # Need to keep the colnames
    tmp <- unlist(outz1[[j1]][, c("sig", "R", "nlt0")])
    tmp2 <- add_rr_ext(
                tmp,
                R = R,
                Rk = Rk
              )
    tmp2 <- tmp2[-which(names(tmp2) %in% c("sig", "R", "nlt0"))]
    tmp3 <- cbind(outz1[[j1]], rbind(tmp2))
    outz1[[j1]] <- tmp3
  }
  if (out_type == "vector") {
    # TODO:
    # - Check
    return((unlist(outz1[[1]])))
  } else {
    out <- do.call(rbind,
                   outz1)
    if (inherits(outi, "matrix")) {
      out <- as.matrix(out)
    }
    rownames(out) <- rownames(outi)
    return(out)
  }
}

#' @noRd
bz_sig_partition <- function(boot_est,
                             alpha = .05) {
  # If R denotes a set of Rs,
  # compute the rejection rate for each partition and
  # return a vector of these rates
  R <- length(boot_est)
  R_case <- bz_case(R)
  if (R_case == "cum") {
    Rs <- R_indices(R)
    boot_ps <- sapply(Rs,
                  \(x) est2p(boot_est[x],
                             min_size = -Inf)
                )
    boot_sig <- as.numeric(boot_ps < alpha)
    names(boot_sig) <- paste0("bz_", names(boot_ps))
  } else {
    boot_sig <- numeric()
  }
  boot_sig
}

#' @noRd
R_indices <- function(R) {
  # Generate a list of indices
  # for a set of Rs
  Rext <- R_extrapolate()
  Rext_cm <- cumsum(Rext)
  i <- which(R == Rext_cm)
  if (length(i) == 0) stop("R not supported.")
  Rs <- Rext[seq(from = 1, to = i)]
  ik <- split(seq_len(R),
              rep(seq_along(Rs),
              times = Rs))
  names(ik) <- as.character(Rs)
  ik
}

#' @noRd
bz_case <- function(R) {
  # Check whether the R is one of those
  # for one-R extrapolation ("one"), or
  # for a set of Rs ("cum")
  # Default is ""
  Rext <- R_extrapolate()
  Rext_cm <- cumsum(Rext)
  out <- ""
  if (R %in% Rext[-1]) {
    out <- "one"
  }
  if ((R %in% Rext_cm[-1])) {
    out <- "cum"
  }
  out
}

#' @noRd
bz_rr <- function(out) {
  # Do simple linear regression to
  # estimate the rejection rate when R == Inf
  # Input:
  # - A vector with "bz_*" elements,
  #   "*" the number of resamples.
  tmp <- names(out)
  tmp <- tmp[grepl("bz_",
                   tmp,
                   fixed = TRUE)]
  rr_mean <- out[tmp]
  tmp2 <- gsub("bz_",
               "",
               tmp,
               fixed = TRUE)
  Rk <- as.numeric(tmp2)
  lm_out <- stats::lm(rr_mean ~ I(1 / Rk))
  rr_inf <- unname(stats::coef(lm_out)["(Intercept)"])
  rr_inf <- min(max(rr_inf, 0), 1)
  attr(rr_inf, "bz_model") <- lm_out
  rr_inf
}

#' @noRd
add_rr_ext <- function(
                outi,
                R,
                Rk) {
  # Add extrapolated values to a vector
  # R is the number of resamples
  # Rk is the values of R to extrapolate the
  # rejection rates
  out <- outi
  nlt0 <- unname(out["nlt0"])
  for (Ri in Rk) {
    tmp1 <- rr_extrapolated(
                x = R,
                nlt0 = nlt0,
                k = Ri
              )
    tmp2 <- paste0("bz_",
                   as.character(Ri))
    names(tmp1) <- tmp2
    out <- c(out,
             tmp1)
  }
  tmp1 <- unname(out["sig"])
  tmp2 <- paste0("bz_",
                 as.character(R))
  names(tmp1) <- tmp2
  out <- c(out,
            tmp1)
  out
}

#' @noRd
rr_extrapolated <- function(
      x,
      nlt0,
      k,
      alpha = .05
    ) {
  # Compute the rejection rates for
  # a value of R (k), given
  # - the number of estimates, or
  # - the number of resamples
  if (length(x) == 0) {
    # TODO:
    # - This section does not work,
    #   but it is not used.
    # x is the bootstrap estimates
    x <- x[!is.na(x)]
    R <- length(x)
    i <- as.numeric(x < 0)
    cm <- sum(i)
  } else {
    # Assume x is R
    R <- x
    cm <- nlt0
  }
  cn <- R - cm
  alpha2 <- alpha / 2
  p1 <- stats::phyper(
            floor(k * alpha2),
            m = cm,
            n = cn,
            k = k)
  p2 <- stats::phyper(
            ceiling(k * (1 - alpha2)),
            m = cm,
            n = cn,
            k = k,
            lower.tail = FALSE)
  p <- p1 + p2
  p
}

#' @noRd
R_extrapolate <- function(
                  alpha = .05,
                  Rmax = getOption(
                            "power4mome.bz_Rmax",
                            default = 359)
) {
  # Find supported values of R
  # alpha is two-tailed alpha
  alpha2 <- alpha / 2
  tmp1 <- seq(1, Rmax)
  tmp2 <- (tmp1 + 1) * alpha2
  tmp1[round(tmp2) == round(tmp2, 8)]
}

#' @title Do a Test on Each Replication
#'
#' @description Do a test on each
#' replication in the output of
#' [sim_out()].
#'
#' @details
#' The function [do_test()] does
#' an arbitrary test in each
#' replication using the function set to
#' `test_fun`.
#'
#' # The role of `do_test()`
#'
#' The function [do_test()] is used by
#' the all-in-one function
#' [power4test()]. Users usually do not
#' call this function directly, though
#' developers can use this function to
#' develop other functions for power
#' analysis, or to build their own
#' workflows to do the power analysis.
#'
#' # Major Test-Related Arguments
#'
#' ## The test function (test_fun)
#'
#' The function set to `test_fun`,
#' the *test function*, usually
#' should work
#' on the output of [lavaan::sem()],
#' [lmhelprs::many_lm()], or
#' [stats::lm()], but can also be a
#' function that works on the output
#' of the function set to `fit_function`
#' when calling [fit_model()] or
#' [power4test()] (see `fit_model_args`).
#'
#' The function has two default
#' requirements.
#'
#' First, it has an argument `fit`, to
#' be set to the output of
#' [lavaan::sem()] or another output
#' stored in the element `extra$fit` of
#' a replication in the `sim_all`
#' object. (This requirement can be
#' relaxed; see the section on `map_names`.)
#'
#' That is, the function definition
#' should be of this from: `FUN(fit,
#' ...)`. This is the form of all
#' `test_*` functions in `power4mome`.
#'
#' If other arguments are to be passed
#' to the test function, they can be set
#' to `test_args` as a named list.
#'
#' Second, the test function must
#' returns an output that (a) can be
#' processed by the results function
#' (see below), or (b) is of
#' the required format for the output of
#' a results function (see the next
#' section). If it already returns
#' an output of the required format,
#' then there is no need to set the
#' results function.
#'
#' ## The results function (results_fun)
#'
#' The test results will be extracted
#' from the output of `test_fun` by the
#' function set to `results_fun`,
#' the *results function*. If
#' the `test_fun` already returns an
#' output of the expected format
#' (see below), then set `results_fun`
#' to `NULL`, the default. The output
#' of `test_fun` will be used for
#' estimating power.
#'
#' The function set to `results_fun`
#' must accept the output of `test_fun`,
#' as the first argument, and return a
#' named list (which can be a data frame)
#' or a named vector with some
#' of the following
#' elements:
#'
#' - `est`: Optional. The estimate of a
#'  parameter, if applicable.
#'
#' - `se`: Optional. The standard error
#'  of the estimate, if applicable.
#'
#' - `cilo`: Optional. The lower limit of the
#'  confidence interval, if applicable.
#'
#' - `cihi`: Optional. The upper limit of the
#'  confidence interval, if applicable.
#'
#' - `sig`: Required. If `1`, the test is
#'  significant. If `0`, the test is not
#'  significant. If the test cannot be
#'  done for any reason, it should be
#'  `NA`.
#'
#' The results can then be used to
#' estimate the power or Type I
#' error of the test.
#'
#' For example, if
#' the null hypothesis is false, then
#' the proportion of significant, that
#' is, the mean of the values of `sig`
#' across replications, is the power.
#'
#' ## Built-in test functions
#'
#' The package `power4mome` has some ready-to-use
#' test functions:
#'
#' - [test_indirect_effect()]
#'
#' - [test_cond_indirect()]
#'
#' - [test_cond_indirect_effects()]
#'
#' - [test_moderation()]
#'
#' - [test_index_of_mome()]
#'
#' - [test_parameters()]
#'
#' Please refer to their help pages
#' for examples.
#'
#' ## The argument map_names
#'
#' This argument is for developers using
#' a test function that has a different
#' name for the argument of the fit
#' object (`"fit"`, the default).
#'
#' If `test_fun` is set to a function
#' that works on an output of, say,
#' [lavaan::sem()] but the argument name
#' for the output is not `fit`, the
#' mapping can be changed by
#' `map_names`.
#'
#' For example,
#' [lavaan::parameterEstimates()]
#' receives an output of [lavaan::sem()]
#' and reports the test results of model
#' parameters. However, the argument
#' name for the `lavaan` output is
#' `object.` To instruct [do_test()] to
#' do the test correctly when setting
#' `test_fun` to
#' `lavaan::parameterEstimates`, add
#' `map_names = c(object = "fit")`. The
#' element `fit` in a replication will
#' then set to the argument `object` of
#' [lavaan::parameterEstimates()].
#'
#' # The background for having the `results_fun` argument
#'
#' In the early development of
#' `power4mome`, `test_fun` is designed
#' to accept existing functions from
#' other packages, such as
#' [manymome::indirect_effect()]. Their
#' outputs are not of the required
#' format for power analysis, and so
#' results functions are needed to
#' process their outputs. In the current
#' version of `power4mome`, some
#' ready-to-use test functions, usually
#' wrappers of these existing functions
#' from other packages, have been
#' developed, and they no longer need
#' results functions to process the
#' output. The argument `results_fun` is
#' kept for backward compatibility and
#' advanced users to use test functions
#' from other packages.
#'
#' @seealso See [power4test()] for
#' the all-in-one function that uses
#' this function. See [test_indirect_effect()],
#' [test_cond_indirect()],
#' [test_cond_indirect_effects()],
#' [test_moderation()],
#' [test_index_of_mome()], and
#' [test_parameters()] for examples of
#' test functions.
#'
#' @return
#' An object of the class `test_out`,
#' which is a list of length equal to
#' `sim_all`, one element for each
#' replication. Each element of the list
#' has two elements:
#'
#' - `test`: The output of the function
#'   set to `test_fun`.
#'
#' - `test_results`: The output of
#'   the function set to `results_fun`.
#'
#' @param sim_all The output of
#' [sim_out()].
#'
#' @param test_fun A function to do the
#' test. See 'Details' for the requirement
#' of this function. There are some
#' built-in test functions in `power4mome`,
#' described in 'Details'.
#'
#' @param test_args A list of arguments
#' to be passed to the `test_fun`
#' function. Default is `list()`.
#'
#' @param map_names A named character
#' vector specifying how the content of
#' the element `extra` in
#' each replication of `sim_all` map
#' to the argument of `test_fun`.
#' Default is `c(fit = "fit")`,
#' indicating that the element `fit`
#' in the element `extra` is set to
#' the argument `fit` of `test_fun`.
#' That is, for the first replication,
#' `fit = sim_out[[1]]$extra$fit` when
#' calling `test_fun`.
#'
#' @param results_fun The function to be
#' used to extract the test results.
#' See `Details` for the requirements
#' of this function. Default is `NULL`,
#' assuming that the output of
#' `test_fun` can be used directly.
#'
#' @param results_args A list of
#' arguments to be passed to the
#' `results_fun` function. Default is
#' `list()`.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to do the
#' tests. Default is `FALSE`.
#'
#' @param progress If `TRUE`, the progress
#' of tests will be displayed.
#' Default is `FALSE.
#'
#' @param ncores The number of CPU
#' cores to use if parallel processing
#' is used.
#'
#' @param cl A cluster, such as one created
#' by [parallel::makeCluster()]. If `NULL`,
#' a cluster will be created, but will be
#' stopped on exit. If set to an existing
#' cluster, it will not be stopped when
#' the function exits; users need to
#' stop it manually.
#'
#' @examples
#'
#' # Specify the population model
#'
#' mod <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' # Specify the effect sizes (population parameter values)
#'
#' es <-
#' "
#' y ~ m: m
#' m ~ x: m
#' y ~ x: n
#' "
#'
#' # Generate several simulated datasets
#'
#' data_all <- sim_data(nrep = 5,
#'                      model = mod,
#'                      pop_es = es,
#'                      n = 100,
#'                      iseed = 1234)
#'
#' # Fit the population model to each datasets
#'
#' fit_all <- fit_model(data_all)
#'
#' # Generate Monte Carlo estimates for forming Monte Carlo confidence intervals
#'
#' mc_all <- gen_mc(fit_all,
#'                  R = 50,
#'                  iseed = 4567)
#'
#' # Combine the results to a 'sim_all' object
#' sim_all <- sim_out(data_all = data_all,
#'                    fit = fit_all,
#'                    mc_out = mc_all)
#'
#' # Test the indirect effect in each replication
#' # Set `parallel` to TRUE for faster, usually much faster, analysis
#' # Set `progress` to TRUE to display the progress of the analysis
#'
#' test_all <- do_test(sim_all,
#'                     test_fun = test_indirect_effect,
#'                     test_args = list(x = "x",
#'                                      m = "m",
#'                                      y = "y",
#'                                      mc_ci = TRUE),
#'                     parallel = FALSE,
#'                     progress = FALSE)
#'
#' # The result for each dataset
#' lapply(test_all, function(x) x$test_results)
#'
#'
#' @export
#'
do_test <- function(sim_all,
                    test_fun,
                    test_args = list(),
                    map_names = c(fit = "fit"),
                    results_fun = NULL,
                    results_args = list(),
                    parallel = FALSE,
                    progress = FALSE,
                    ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
                    cl = NULL) {
  out <- do_FUN(X = sim_all,
                FUN = do_test_i,
                test_fun = test_fun,
                test_args = test_args,
                map_names = map_names,
                results_fun = results_fun,
                results_args = results_args,
                parallel = parallel,
                progress = progress,
                ncores = ncores,
                cl = cl)

  attr(out, "test_fun") <- test_fun
  attr(out, "test_args") <- test_args
  attr(out, "map_names") <- map_names
  attr(out, "results_fun") <- results_fun
  attr(out, "results_args") <- results_args

  class(out) <- c("test_out", class(out))
  return(out)
}

#' @noRd
do_test_i <- function(out_i,
                      test_fun,
                      test_args = list(0),
                      map_names = c(fit = "fit"),
                      results_fun = NULL,
                      results_args = list()) {
  args_map <- sapply(map_names,
                     function(x,
                              out_i) {
                       out_i$extra[[x]]
                     },
                     out_i = out_i,
                     simplify = FALSE,
                     USE.NAMES = FALSE)
  args1 <- c(args_map,
             test_args)
  # # Disable error catching
  # # Error *should* be raised
  # out0 <- tryCatch(
  #           do.call(
  #             test_fun,
  #             args1
  #           ),
  #           error = function(e) e)
  # if (inherits(out0, "error")) {
  #   return(list(test = NULL,
  #               test_results = NULL))
  # }
  out0 <- do.call(
            test_fun,
            args1
          )
  if (is.function(results_fun)) {
    # # Disable error catching
    # # Error *should* be raised
    # out1 <- tryCatch(
    #           do.call(
    #             results_fun,
    #             c(list(out0), results_args)
    #           ),
    #           error = function(e) NULL)
    out1 <- do.call(
                results_fun,
                c(list(out0), results_args)
              )
  } else {
    out1 <- out0
  }

  return(list(test = out0,
              test_results = out1))
}

library(testthat)
suppressMessages(library(lavaan))

test_that("sim_out and do_test", {

mod <-
"m ~ x
 y ~ m + x"
es <-
c("y ~ m" = "m",
  "m ~ x" = "m",
  "y ~ x" = "n")
data_all <- sim_data(nrep = 3,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     progress = !is_testing(),
                     iseed = 1234)
fit_all <- fit_model(data_all)
# mc_all <- gen_mc(fit_all,
#                  R = 100,
#                  iseed = 4567)
sim_all <- sim_out(data_all = data_all,
                   fit = fit_all)

# Check fit_external

tmp <- sim_all[[3]]$fit_external$all_paths
expect_equal(unname(unlist(manymome::all_paths_to_df(tmp))),
             c("x", "y", "m"))
tmp <- fit_all[[3]]@external$fit_external$all_paths
expect_equal(unname(unlist(manymome::all_paths_to_df(tmp))),
             c("x", "y", "m"))

fit_all_lm <- fit_model(data_all, fit_function = "lm")
tmp <- attr(fit_all_lm[[3]], "fit_external")$all_paths
expect_equal(unname(unlist(manymome::all_paths_to_df(tmp))),
             c("x", "y", "m"))

est_test <- function(fit,
                     par_name = NULL,
                     alpha = .05,
                     out_type = c("vector", "list")) {
  out_type <- match.arg(out_type)
  est <- lavaan::parameterEstimates(fit,
                                    ci = TRUE)
  est$lavlabel <- lavaan::lav_partable_labels(est)
  i <- which(est$lavlabel == par_name)
  out <- switch(out_type,
          vector = c(est = est[i, "est"],
                     se = est[i, "se"],
                     cilo = est[i, "ci.lower"],
                     cihi = est[i, "ci.upper"],
                     sig = ifelse(est[i, "pvalue"] < alpha, 1, 0)),
          list = list(est = est[i, "est"],
                      se = est[i, "se"],
                      cilo = est[i, "ci.lower"],
                      cihi = est[i, "ci.upper"],
                      sig = ifelse(est[i, "pvalue"] < alpha, 1, 0)))
  out
}

test_i <- do_test_i(sim_all[[1]],
                    test_fun = est_test,
                    test_args = list(par_name = "y~x"))
expect_equal(test_i$test_results["se"],
             parameterEstimates(fit_all[[1]])[3, "se"],
             ignore_attr = TRUE)

test_all <- do_test(sim_all,
                    test_fun = est_test,
                    test_args = list(par_name = "y~x"),
                    parallel = FALSE,
                    progress = !is_testing())

expect_equal(test_all[[3]]$test_results["se"],
             parameterEstimates(fit_all[[3]])[3, "se"],
             ignore_attr = TRUE)

test_all <- do_test(sim_all,
                    test_fun = est_test,
                    test_args = list(par_name = "y~x",
                                     out_type = "list"),
                    parallel = FALSE,
                    progress = !is_testing())

expect_equal(test_all[[3]]$test_results$se,
             parameterEstimates(fit_all[[3]])[3, "se"],
             ignore_attr = TRUE)

est_results <- function(est,
                        par_name = NULL,
                        alpha = .05,
                        out_type = c("vector", "list")) {
  out_type <- match.arg(out_type)
  est$lavlabel <- lavaan::lav_partable_labels(est)
  i <- which(est$lavlabel == par_name)
  out <- switch(out_type,
          vector = c(est = est[i, "est"],
                     se = est[i, "se"],
                     cilo = est[i, "ci.lower"],
                     cihi = est[i, "ci.upper"],
                     sig = ifelse(est[i, "pvalue"] < alpha, 1, 0)),
          list = list(est = est[i, "est"],
                      se = est[i, "se"],
                      cilo = est[i, "ci.lower"],
                      cihi = est[i, "ci.upper"],
                      sig = ifelse(est[i, "pvalue"] < alpha, 1, 0)))
  out
}

test_all <- do_test(sim_all,
                    test_fun = lavaan::parameterEstimates,
                    map_names = c(object = "fit"),
                    results_fun = est_results,
                    results_args = list(par_name = "y~x"),
                    parallel = FALSE,
                    progress = !is_testing())
expect_equal(test_all[[3]]$test_results["se"],
             parameterEstimates(fit_all[[3]])[3, "se"],
             ignore_attr = TRUE)

test_all <- do_test(sim_all,
                    test_fun = lavaan::parameterEstimates,
                    map_names = c(object = "fit"),
                    results_fun = est_results,
                    results_args = list(par_name = "y~x",
                                        out_type = "list"),
                    parallel = FALSE,
                    progress = !is_testing())
expect_equal(test_all[[3]]$test_results$se,
             parameterEstimates(fit_all[[3]])[3, "se"],
             ignore_attr = TRUE)


})

#' @title Generate the Population Model
#'
#' @description Generate the complete
#' population model using the model
#' syntax and user-specified effect
#' sizes (population parameter values).
#'
#' @details
#' The function [ptable_pop()] generates a `lavaan`
#' parameter table that can be used
#' to generate data based on the population
#' values of model parameters.
#'
#' # The role of `ptable_pop()`
#'
#' The function [ptable_pop()] is used by
#' the all-in-one function
#' [power4test()]. Users usually do not
#' call this function directly, though
#' developers can use this function to
#' develop other functions for power
#' analysis, or to build their own
#' workflows to do the power analysis.
#'
#' # Specify the Population Model by 'model'
#'
#' ## Single-Group Model
#'
#' For a single-group model, `model`
#' should be a `lavaan` model syntax
#' string of the *form* of the model.
#' The population values of the model
#' parameters are to be determined by
#' `pop_es`.
#'
#' If the model has latent factors,
#' the syntax in `model` should specify
#' only the *structural model* for the
#' *latent factors*. There is no
#' need to specify the measurement
#' part. Other functions will generate
#' the measurement part on top of this
#' model.
#'
#' For example, this is a simple mediation
#' model:
#'
#' \preformatted{"m ~ x
#'  y ~ m + x"}
#'
#' Whether `m`, `x`, and `y` denote
#' observed variables or latent factors
#' are determined by other functions,
#' such as [power4test()].
#'
#' ## Multigroup Model
#'
#' Because the model is the population
#' model, equality constraints are
#' irrelevant and the model syntax
#' specifies only the *form* of the
#' model. Therefore, `model` is
#' specified as in the case of single
#' group models.
#'
#' # Specify 'pop_es' Using Named Vectors
#'
#' The argument `pop_es` is for specifying
#' the population values of model
#' parameters. This section describes
#' how to do this using named vectors.
#'
#' ## Single-Group Model
#'
#' If `pop_es` is specified by a named
#' vector, it must follow the convention
#' below.
#'
#' - The names of the vectors are
#'   `lavaan` names for the selected
#'   parameters. For example, `m ~ x`
#'   denotes the path from `x` to `m`.
#'
#' - Alternatively, the names can be
#'   either `".beta."` or `".cov."`.
#'   Use `".beta."` to set the default
#'   values for all regression coefficients.
#'   Use `".cov."` to set the default
#'   values for all correlations of
#'   exogenous variables (e.g., predictors).
#'
#' - The names can also be of this form:
#'   `".ind.(<path>)"`, whether `<path>`
#'   denote path in the model. For
#'   example, `".ind.(x->m->y)"` denotes
#'   the path from `x` through `m` to
#'   `y`. Alternatively, the `lavaan`
#'   symbol `~` can also be used:
#'   `".ind.(y~m~x)"`. This form is used
#'   to set the indirect effect (standardized,
#'   by default) along this path. The
#'   value for this name will override
#'   other settings.
#'
#' - If using `lavaan` names, can
#'   specify more than one parameter
#'   using `+`. For example, `y ~ m + x`
#'   denotes the two paths from `m` and
#'   `x` to `y`.
#'
#' - The value of each element can be
#'   the label for the effect size: `n`
#'   for nil, `s` for small, `m` for
#'   medium, and `l` for large. The
#'   value for each label is determined
#'   by `es1` and `es2`. See the section
#'   on specifying these two arguments.
#'
#' - The value of `pop_es` can also be
#'   set to a value, but it must be
#'   quoted as a string, such as `"y ~
#'   x" = ".31"`.
#'
#' This is an example:
#'
#' \preformatted{c(".beta." = "s",
#'   "m1 ~ x" = "-m",
#'   "m2 ~ m1" = "l",
#'   "y ~ x:w" = "s")}
#'
#' In this example,
#'
#' - All regression coefficients are
#'  set to small (`s`) by default, unless
#'  specified otherwise.
#'
#' - The path from `x` to `m1` is
#'  set to medium and negative (`-m`).
#'
#' - The path from `m1` to `m2` is set
#'  to large (`l`).
#'
#' - The coefficient of the product
#'  term `x:w` when predicting `y` is
#'  set to small (`s`).
#'
#' ### Indirect Effect
#'
#' When setting an indirect effect to
#' a symbol (default: `"si"`, `"mi"`,
#' `"li"`, with `"i"` added to differentiate
#' them from the labels for a direct path),
#' the corresponding value is used to
#' determine the population values of
#' *all* component paths along the pathway.
#' All the values are assumed to be equal.
#' Therefore, `".ind.(x->m->y)" = ".20"`
#' is equivalent to setting `m ~ x`
#' and `y ~ m` to the square root of
#' .20, such that the corresponding
#' indirect effect is equal to the
#' designated value.
#'
#' This behavior, though restricted,
#' is for quick manipulation of the
#' indirect effect. If different values
#' along a pathway, set the value for
#' each path directly.
#'
#' Only nonnegative value is supported.
#' Therefore, `".ind.(x->m->y)" = "-si"`
#' and `".ind.(x->m->y)" = "-.20"` will
#' throw an error.
#'
#' ## Multigroup Model
#'
#' The argument `pop_es` also supports multigroup
#' models.
#'
#' For `pop_es`, instead of
#' named vectors, named *list* of
#' named vectors should be used.
#'
#' - The names are the parameters, or
#'   keywords such as `.beta.` and
#'   `.cov.`, like specifying the
#'   population values for a single
#'   group model.
#'
#' - The elements are character vectors.
#'   If it has only one element (e.g.,
#'   a single string), then it is the
#'   the population value for all groups.
#'   If it has more than one element
#'   (e.g., a vector of three strings),
#'   then they are the population values
#'   of the groups. For a model of *k*
#'   groups, each vector must have
#'   either *k* elements or one element.
#'
#' This is an example:
#'
#' \preformatted{list("m ~ x" = "m",
#'      "y ~ m" = c("s", "m", "l"))}
#'
#' In this model, the population value
#' of the path `m ~ x` is medium (`m`) for
#' all groups, while the population
#' values for the path `y ~ m` are
#' small (`s`), medium (`m`), and large (`l`),
#' respectively.
#'
#' # Specify 'pop_es' Using a Multiline String
#'
#' When setting the argument `pop_es`,
#' instead of using a named vector
#' or named list for
#' `pop_es`, the population values of
#' model parameters can also be
#' specified using a multiline string,
#' as illustrated below, to be parsed
#' by [pop_es_yaml()].
#'
#' ## Single-Group Model
#'
#' This is an example of the multiline string
#' for a single-group model:
#'
#' \preformatted{y ~ m: l
#' m ~ x: m
#' y ~ x: nil}
#'
#' The string must follow this format:
#'
#' - Each line starts with `tag:`.
#'
#'   - `tag` can be the name of a
#'      parameter, in `lavaan` model
#'      syntax format.
#'
#'     - For example, `m ~ x`
#'       denotes the path from `x` to `m`.
#'
#'   - A tag in `lavaan` model syntax can
#'     specify more than one parameter
#'     using `+`.
#'
#'     - For example, `y ~ m + x`
#'       denotes the two paths from `m` and
#'       `x` to `y`.
#'
#'   - Alternatively, the `tag` can be
#'   either `.beta.` or `.cov.`.
#'
#'     - Use `.beta.` to set the default
#'       values for all regression coefficients.
#'
#'     - Use `.cov.` to set the default
#'       values for all correlations of
#'       exogenous variables (e.g., predictors).
#'
#' - After each tag is the value of the
#'   population value:
#'
#'     -`nil` for nil (zero),
#'
#'     - `s` for small,
#'
#'     - `m` for medium, and
#'
#'     - `l` for large.
#'
#'     - `si`, `mi`, and `li` for
#'       small, medium, and large a
#'       standardized indirect effect,
#'       respectively.
#'
#'   Note: `n` *cannot* be used in this mode.
#'
#'   The
#'   value for each label is determined
#'   by `es1` and `es2` as described
#'   in [ptable_pop()].
#'
#'   - The value can also be
#'     set to a numeric value, such as
#'     `.30` or `-.30`.
#'
#' This is another example:
#'
#' \preformatted{.beta.: s
#' y ~ m: l}
#'
#' In this example, all regression
#' coefficients are `small`, while
#' the path from `m` to `y` is large.
#'
#' ## Multigroup Model
#'
#' This is an example of the string
#' for a multigroup model:
#'
#' \preformatted{y ~ m: l
#' m ~ x:
#'   - nil
#'   - s
#' y ~ x: nil}
#'
#' The format is similar to that for
#' a single-group model. If a parameter
#' has the same value for all groups,
#' then the line can be specified
#' as in the case of a single-group
#' model: `tag: value`.
#'
#' If a parameter has different
#' values across groups, then it must
#' be in this format:
#'
#' - A line starts with the tag, followed
#'   by two or more lines. Each line
#'   starts with a hyphen `-` and the
#'   value for a group.
#'
#' For example:
#'
#' \preformatted{m ~ x:
#'   - nil
#'   - s}
#'
#' This denotes that the model has
#' two groups. The values of the path
#' from `x` to `m` for the two
#' groups are 0 (`nil`) and
#' small (`s`), respectively.
#'
#' Another equivalent way to specify
#' the values are using `[]`, on
#' the same line of a tag.
#'
#' For example:
#'
#' \preformatted{m ~ x: [nil, s]}
#'
#' The number of groups is inferred
#' from the number of values for
#' a parameter. Therefore, if a tag
#' has more than one value, each tag
#' must has the same number of value,
#' or only one value.
#'
#' The tag `.beta.` and `.cov.` can
#' also be used for multigroup models.
#'
#' ## Which Approach To Use
#'
#' Note that using named vectors or
#' named lists is more reliable. However,
#' using a multiline string is
#' more user-friendly. If this method
#' failed, please use named vectors or
#' named list instead.
#'
#' ## Technical Details
#'
#' The multiline string is parsed by [yaml::read_yaml()].
#' Therefore, the format requirement
#' is actually that of YAML. Users
#' knowledgeable of YAML can use other
#' equivalent way to specify the string.
#'
#' # Set the Values for Effect Size Labels ('es1' and 'es2')
#'
#' The vector `es1` is for correlations,
#' regression coefficients, and
#' indirect effect, and the
#' vector `es2` is for for standardized
#' moderation effect, the coefficients
#' of a product term. These labels
#' are to be used in interpreting
#' the specification in `pop_es`.
#'
#' @seealso [power4test()], and
#' [pop_es_yaml()] on an alternative
#' way to specify population values.
#'
#' @return
#' The function [ptable_pop()] returns
#' a `lavaan` parameter table of the
#' model, with the column `start` set to the
#' population values.
#'
#' @param model String. The model defined
#' by `lavaan` model syntax. See 'Details'.
#'
#' @param pop_es It can be a data frame
#' with these columns: `lhs`, `op`,
#' `rhs`, and `pop`. The first three
#' columns correspond to those in a
#' `lavaan` parameter table. The column
#' `pop` stores the population values.
#' The column `es` stores the original
#' labels, for reference. It can also be
#' a named character vector (named list
#' for multigroup models) or a multiline string,
#' which are
#' the preferred approaches. See the
#' help page
#' on how to specify this vector.
#'
#' @param es1 Set the
#' values for each label of the effect
#' size (population value) for correlations and regression
#' paths.
#' Used only if `pop_es` is a named
#' vector or a multiline string.
#' See the help page on how to specify
#' this argument.
#'
#' @param es2 Set the
#' values for each label of the effect
#' size (population value) for product term.
#' Used only if `pop_es` is a named
#' vector or a multiline string.
#' See the help page on how to specify
#' this argument.
#'
#' @param es_ind The names of labels
#' denoting the effect size of an
#' indirect effect. They will be
#' used to determine the population
#' values of the component paths along
#' an indirect path.
#'
#' @param standardized Logical. If
#' `TRUE`, the default, variances and
#' error variances are scaled to ensure
#' the population variances of the
#' endogenous variables are close to
#' one, and hence the effect sizes
#' (population values) are
#' standardized effect sizes if the
#' variances of the continuous exogenous
#' variables are also equal to one.
#'
#' @param n_std The sample size used to
#' determine the error variances by
#' simulation when `std_force_monte_carlo`
#' is `TRUE`.
#'
#' @param std_force_monte_carlo Logical.
#' If `FALSE`, the default,
#' standardization is done analytically
#' if the model has no product terms,
#' and by simulation if the model has
#' product terms. That is, error variances
#' required to ensure implied variances equal
#' to one are determined by simulation.
#' If `TRUE`, simulation
#' will be used whether the model has
#' product terms or not. Always fall
#' back to simulation if
#' analytical standardization failed.
#'
#' @param add_cov_for_moderation
#' Logical. If `TRUE`, the default, for
#' a model in which one or product terms
#' for moderation involve one or more
#' mediator, covariances between their
#' error terms and the product terms
#' will be added automatically. If these
#' covariances are not added, the model
#' may not be invariant to linear
#' transformation of some variables in
#' the model.
#'
#' @examples
#'
#' # Specify the model
#'
#' model1 <-
#' "
#' m1 ~ x + c1
#' m2 ~ m1 + x2 + c1
#' y ~  m2 + m1 + x + w + x:w + c1
#' "
#'
#' # Specify the population values
#'
#' model1_es <- c("m1 ~ x" = "-m",
#'                "m2 ~ m1" = "s",
#'                "y ~ m2" = "l",
#'                "y ~ x" = "m",
#'                "y ~ w" = "s",
#'                "y ~ x:w" = "s",
#'                "x ~~ w" = "s")
#'
#' ptable_final1 <- ptable_pop(model1,
#'                             pop_es = model1_es)
#' ptable_final1
#'
#' # Use a multiline string, illustrated by a simpler model
#'
#' model2 <-
#' "
#' m ~ x
#' y ~ m + x
#' "
#'
#' model2_es_a <- c("m ~ x" = "s",
#'                  "y ~ m" = "m",
#'                  "y ~ x" = "nil")
#'
#' model2_es_b <-
#' "
#' m ~ x: s
#' y ~ m: m
#' y ~ x: nil
#' "
#'
#' ptable_model2_a <- ptable_pop(model2,
#'                               pop_es = model2_es_a)
#' ptable_model2_b <- ptable_pop(model2,
#'                               pop_es = model2_es_b)
#'
#' ptable_model2_a
#' ptable_model2_b
#'
#' identical(ptable_model2_a,
#'           ptable_model2_b)
#'
#' @export
# Input:
# - model: The model syntax
# - pop_es: The parameter table of parameters with population values OR
#           A named vector of the effect size labels for selected parameters
# Output:
# - The parameter table with population values
ptable_pop <- function(model = NULL,
                       pop_es = NULL,
                       es1 = c("n" = .00,
                               "nil" = .00,
                               "s" = .10,
                               "m" = .30,
                               "l" = .50,
                               "si" = .141,
                               "mi" = .361,
                               "li" = .510,
                               "sm" = .20,
                               "ml" = .40
                               ),
                       es2 = c("n" = .00,
                               "nil" = .00,
                               "s" = .05,
                               "m" = .10,
                               "l" = .15,
                               "sm" = .075,
                               "ml" = .125),
                       es_ind = c("si",
                                  "mi",
                                  "li"),
                       standardized = TRUE,
                       n_std = 100000,
                       std_force_monte_carlo = FALSE,
                       add_cov_for_moderation = TRUE) {
  if (is.null(model) || is.null(pop_es)) {
    stop("Both model and pop_es must be set.")
  }
  # pop_es a YAML string? If yes, convert it
  pop_es <- pop_es_yaml_check(pop_es)

  par_pop <- pop_es2par_pop(pop_es = pop_es,
                            es1 = es1,
                            es2 = es2,
                            model = model,
                            to_one_table = TRUE,
                            es_ind = es_ind)
  ngroups <- max(par_pop$group)
  # Single group ptable
  if (ngroups > 1) {
    # Handle labels in the syntax
    tmp <- lavaan::lavParseModelString(model,
                                       as.data.frame. = TRUE)
    tmp$label <- ""
    tmp$mod.idx <- 0
    attr(tmp, "modifiers") <- list()
    attr(tmp, "constraints") <- list()
    fit0 <- lavaan::sem(tmp,
                        do.fit = FALSE,
                        fixed.x = FALSE)
    ptable0 <- lavaan::parTable(fit0)
  } else {
    fit0 <- lavaan::sem(model,
                        do.fit = FALSE,
                        fixed.x = FALSE)
    ptable0 <- lavaan::parTable(fit0)
  }

  # Use fake data to create the target parameter table
  if (ngroups > 1) {
    ptable0_ng <- ptable0
    gpnames <- paste0("gp", seq_len(ngroups))
    vnames <- lavaan::lavNames(ptable0,
                               type = "ov")
    p <- length(vnames)
    d1 <- diag(p)
    colnames(d1) <- rownames(d1) <- vnames
    dat_cov <- lapply(seq_len(ngroups),
                      function(x) d1)
    fit0 <- lavaan::sem(model,
                        sample.cov = dat_cov,
                        sample.nobs = rep(10000, ngroups),
                        do.fit = FALSE,
                        group.label = gpnames,
                        fixed.x = FALSE)
    ptable0 <- lavaan::parTable(fit0)

    # Fix starting values from MG
    ptable0$tmp1 <- paste0(ptable0$lhs,
                           ptable0$op,
                           ptable0$rhs)
    ptable0_ng$tmp1 <- paste0(ptable0_ng$lhs,
                              ptable0_ng$op,
                              ptable0_ng$rhs)
    ptable0_ng$start_ng <- ptable0_ng$start
    ptable0_ng$est_ng <- ptable0_ng$est
    pt_tmp <- merge(ptable0,
                    ptable0_ng[, c("tmp1", "start_ng", "est_ng")],
                    all.x = TRUE,
                    all.y = FALSE,
                    sort = FALSE)
    i <- !is.na(pt_tmp$start_ng)
    pt_tmp$start[i] <- pt_tmp$start_ng[i]
    i <- !is.na(pt_tmp$est_ng)
    pt_tmp$est[i] <- pt_tmp$est_ng[i]
    ptable0 <- pt_tmp[, colnames(ptable0)]
    ptable0 <- ptable0[order(ptable0$id), ]
    ptable0$tmp1 <- NULL
    rownames(ptable0) <- NULL
  }

  # Check par_pop
  vnames <- lavaan::lavNames(ptable0,
                             type = "ov")
  vnamee_par_pop <- lavaan::lavNames(par_pop,
                                     type = "ov")
  chk_names <- setdiff(vnamee_par_pop,
                       vnames)
  if (length(chk_names) > 0) {
    tmp <- paste0(chk_names,
                  collapse = ", ")
    msg <- paste(tmp,
                 "is/are in pop_es but not in the model,",
                 "and parameter(s) involved is/are ignored.")
    warning(msg)
  }

  par_pop2 <- merge(par_pop,
                    ptable0[, c("lhs", "op", "rhs", "id")],
                    all.x = TRUE,
                    all.y = FALSE)
  if (any(is.na(par_pop2$id))) {
    tmp1 <- is.na(par_pop2$id)
    tmp2 <- apply(par_pop2[tmp1, c("lhs", "op", "rhs"), drop = FALSE],
                  MARGIN = 1,
                  paste0,
                  collapse = " ")
    tmp2 <- paste(tmp2, collapse = ",")
    # warning("One or more parameters in 'pop_es' is not in the model: ",
    #         tmp2)
  }
  par_pop2 <- par_pop2[, c("id", "lhs", "op", "rhs", "group", "pop")]

  ptable1 <- merge(ptable0,
                   par_pop2,
                   all.x = TRUE)
  ptable1[!is.na(ptable1$pop), "start"] <- ptable1[!is.na(ptable1$pop), "pop"]
  ptable1$pop <- NULL
  # TODO:
  # - Check equality constraints
  attr(ptable1, "model") <- model
  if (standardized) {
    mm <- model_matrices_pop(ptable1,
                             drop_list_single_group = FALSE)
    if (ncol(mm[[1]]$psi) != 0) {
      for (i in seq_along(mm)) {
        mm[[i]]$psi <- psi_std(mm[[i]],
                               n_std = n_std)
      }
    } else {
      # CFA model or correlation only
      # Make sure theta is a covariance matrix,
      # in case variances accidentally set.
      for (i in seq_along(mm)) {
        mm[[i]]$theta <- stats::cov2cor(mm[[i]]$theta)
      }
    }
    std_by_monte_carlo <- attr(mm[[i]]$psi, "std_by_monte_carlo")
    if (is.null(std_by_monte_carlo)) {
      # For backward compatibility
      std_by_monte_carlo <- NA
    }
    ptable1 <- start_from_mm(ptable1,
                             mm)
    attr(ptable1, "model") <- model
    attr(ptable1, "std_by_monte_carlo") <- std_by_monte_carlo
  }
  if ((length(m_moderated(model, ngroups = ngroups)) > 0) &&
      add_cov_for_moderation) {
    # Models with mediators involved in moderation
    ptable_fixed <- pt_with_int(
                      ptable = ptable1,
                      model_original = model
                    )
    ptable1 <- ptable_fixed
    if (utils::packageVersion("lavaan") > "0.6.19") {
      # lavaan 0.6.20+ should support "x:w ~~ y:z"
      attr(ptable1, "model") <- attr(ptable1, "model_fixed")
    }
  }

  # It is intentional not saving the call and
  # saving the argument values.
  # For the ease of updating.
  attr(ptable1, "pop_es") <- pop_es
  attr(ptable1, "es1") <- es1
  attr(ptable1, "es2") <- es2
  attr(ptable1, "es_ind") <- es_ind
  # par_pop is one single table with a group column
  attr(ptable1, "par_pop") <- par_pop
  attr(ptable1, "n_std") <- n_std
  attr(ptable1, "standardized") <- standardized
  attr(ptable1, "std_force_monte_carlo") <- std_force_monte_carlo

  class(ptable1) <- c("ptable_pop", class(ptable1))
  ptable1
}

#' @noRd
# Should only update pop_es
# Input:
# - A ptable object
# - new_pop_es: pop_es for parameters to be updated
# Output:
# - A ptable based on updated pop_es
update_ptable_pop <- function(object,
                              new_pop_es) {
  if (!inherits(object, "ptable_pop")) {
    stop("Can only update a ptable_pop object.")
  }
  es1 <- attr(object, "es1")
  es2 <- attr(object, "es2")
  es_ind <- attr(object, "es_ind")
  model <- attr(object, "model")
  old_par_pop <- attr(object, "par_pop")
  if (is.data.frame(old_par_pop)) {
    old_par_pop <- split_par_pop(old_par_pop)
  }
  ngroups <- length(old_par_pop)
  new_par_pop <- pop_es2par_pop(new_pop_es,
                                es1 = es1,
                                es2 = es2,
                                model = model,
                                es_ind = es_ind,
                                ngroups = ngroups)
  updated_par_pop <- update_par_pop(add = new_par_pop,
                                    par_pop = old_par_pop)
  updated_par_pop <- par_pop_to_one_table(updated_par_pop)
  out <- ptable_pop(model = model,
                    pop_es = updated_par_pop,
                    es1 = es1,
                    es2 = es2,
                    standardized = attr(object, "standardized"),
                    n_std = attr(object, "n_std"),
                    std_force_monte_carlo = attr(object, "std_force_monte_carlo"))
  out
}

#' @noRd
# Input:
# - par_pop
# Output:
# - A list of par_pops, one for each
#   group, even if the number of groups
#   is 1.
split_par_pop <- function(par_pop) {
  # Already a list of table(s)?
  if (is.list(par_pop)) {
    if (is.data.frame(par_pop[[1]])) {
      return(par_pop)
    }
  }
  out <- par_pop
  out$group <- NULL
  out1 <- split(out,
                f = par_pop$group)
  names(out1) <- NULL
  out1
}

#' @details
#'
#' # The role of `model_matrices_pop()`
#'
#' The function [model_matrices_pop()]
#' generates models matrices with
#' population values, used by [ptable_pop()].
#' Users usually do not
#' call this function directly, though
#' developers can use this build their own
#' workflows to generate the data.
#'
#' @return
#' The function [model_matrices_pop()]
#' returns a `lavaan` LISREL-style model
#' matrices (like the output of
#' [lavaan::lavInspect()] with `what`
#' set to `"free"`), with matrix elements
#' set to the population values. If
#' `x` is the model syntax, it will be
#' stored in the attributes `model`.
#' If the model is a multigroup model
#' with *k* groups (*k* greater than 1),
#' then it returns a list of *k* lists
#' of `lavaan` LISREL-style model
#' matrices unless `drop_list_single_group`
#' is `TRUE`.
#'
#' @param x It can be a 'lavaan' model
#' syntax, to be passed to [ptable_pop()],
#' or a parameter table with the column
#' `start` set to the population values,
#' such as the output of [ptable_pop()].
#'
#' @param ... If `x` is a model syntax,
#' these are arguments to be passed to
#' [ptable_pop()].
#'
#' @param drop_list_single_group If
#' `TRUE` and the number of groups is
#' equal to one, the output will be
#' a list of matrices of one group
#' only. Default if `TRUE`.
#'
#' @examples
#' # model_matrices_pop
#'
#' model_matrices_pop(ptable_final1)
#'
#' model_matrices_pop(model1,
#'                    pop_es = model1_es)
#'
#' @rdname ptable_pop
#' @export

model_matrices_pop <- function(x,
                               ...,
                               drop_list_single_group = TRUE) {
  if (is.character((x))) {
    ptable <- ptable_pop(model = x,
                         ...)
  } else {
    ptable <- x
  }
  fit1 <- lavaan::sem(ptable,
                      do.fit = FALSE)
  ngroups <- lavaan::lavInspect(fit1,
                                "ngroups")
  mm <- lavaan::lavInspect(fit1,
                           "partable",
                           drop.list.single.group = FALSE)
  mm2 <- sapply(mm,
                set_start,
                ptable = ptable,
                simplify = FALSE)
  attr(mm2, "header") <- NULL
  if (is.character((x))) {
    attr(mm2, "model") <- x
    for (i in seq_along(mm2)) {
      attr(mm2[[i]], "model") <- x
    }
  } else {
    attr(mm2, "model") <- attr(x, "model")
    for (i in seq_along(mm2)) {
      attr(mm2[[i]], "model") <- attr(x, "model")
    }
  }
  if (drop_list_single_group && (ngroups == 1)) {
    mm2 <- mm2[[1]]
  }
  mm2
}

#' @noRd
# Update par_pop
# Input:
# - par_pop
# - pop_es
# - es1
# - es2
# - model
# Output:
# - par_pop
update_par_pop <- function(add,
                           par_pop) {
  # par_pop should be a list of tables
  out <- par_pop
  ngroups <- length(out)
  for (i in seq_len(ngroups)) {
    # Can ignore group because
    # each group has its own table
    if (is.null(add[[i]])) {
      # No change for the i-th group
      next
    }
    tmp <- merge(out[[i]],
                 add[[i]],
                 by = c("lhs", "op", "rhs"),
                 all.x = TRUE,
                 all.y = FALSE,
                 suffix = c("", ".add"))
    i2 <- !is.na(tmp$pop.add)
    i3 <- !is.na(tmp$es.add)
    tmp$pop[i2] <- tmp$pop.add[i2]
    tmp$es[i3] <- tmp$es.add[i3]
    tmp$pop.add <- NULL
    tmp$es.add <- NULL
    out[[i]] <- tmp
  }
  # The output is always a list
  out
}

#' @noRd
# Input:
# - pop_es
# - Other arguments needed.
# Output:
# - par_pop. Can be a list or a table.
pop_es2par_pop <- function(pop_es,
                           es1,
                           es2,
                           model,
                           to_one_table = FALSE,
                           es_ind,
                           ngroups = 1) {
  # Always process par_pop as a list until existing
  if (is.character(pop_es) || is.numeric(pop_es)) {

    if (ngroups > 1) {

      # ==== Multigroup model ====

      if (length(pop_es) > 1) {
        stop("For multigroup models, only support changing one parameter.")
      }

      gp_id0 <- get_gp_id(
                    pop_es = pop_es,
                    ngroups = ngroups
                  )
      gp_id <- gp_id0$gp_id
      names(pop_es) <- gp_id0$new_name
      par_pop0 <- vector(
                      mode = "list",
                      length = ngroups
                    )
      pop_es <- fix_par_es(pop_es,
                           model = model)
      par_pop <- set_pop(pop_es,
                         es1 = es1,
                         es2 = es2,
                         es_ind = es_ind)
      par_pop0[[gp_id]] <- par_pop
      par_pop <- par_pop0

    } else {

      # ==== Single-group model ====

      pop_es <- fix_par_es(pop_es,
                           model = model)
      par_pop <- set_pop(pop_es,
                         es1 = es1,
                         es2 = es2,
                         es_ind = es_ind)
      par_pop <- list(par_pop)
    }
  } else if (is.list(pop_es)) {
    if (is.data.frame(pop_es)) {
      # A one-table par_pop
      # Convert to a list
      par_pop <- split_par_pop(pop_es)
    } else if (is.data.frame(pop_es[[1]])) {
      # Already a list of table(s)
      par_pop <- pop_es
    } else {
      # A list of character vectors
      pop_es <- split_par_es(pop_es)
      pop_es <- lapply(pop_es,
                      FUN = fix_par_es,
                      model = model)
      par_pop <- lapply(pop_es,
                        set_pop,
                        es1 = es1,
                        es2 = es2,
                        es_ind = es_ind)
    }
  }
  par_pop <- lapply(par_pop,
                    dup_cov)
  if (to_one_table) {
    par_pop <- par_pop_to_one_table(par_pop)
  }
  par_pop
}

#' @noRd
# Input:
# - par_pop
# Output:
# - One single table
par_pop_to_one_table <- function(par_pop) {
  # Already one table?
  if (is.data.frame(par_pop)) {
    return(par_pop)
  }
  ngroups <- length(par_pop)
  for (i in seq_along(par_pop)) {
    par_pop[[i]]$group <- i
  }
  par_pop <- do.call(rbind,
                     par_pop)
  rownames(par_pop) <- NULL
  par_pop
}

#' @noRd
# Set the starting values in the model matrix
# Input:
# - The list of model matrices in lavaan
# Output:
# - The parameter table with starting values (in "start")

set_start <- function(mm,
                      ptable) {
  for (k in seq_along(mm)) {
    mmi <- mm[[k]]
    i <- which(mmi > 0)
    j <- mmi[i]
    mmic <- col(mmi)
    mmir <- row(mmi)
    for (x in seq_along(j)) {
      mmi[mmir[i[x]], mmic[i[x]]] <- ptable[j[x], "start"]
    }
    mm[[k]] <- mmi
  }
  mm
}

#' @noRd
# Input:
# - Lavaan model matrices
# Output:
# - A list of:
#   - regression models
#   - covariance matrix of x-variables
mm_lm <- function(mm,
                  drop_list_single_group = TRUE) {
  # TODO:
  # - Find a more robust way to check
  #   the number of groups.
  if (is.null(attr(mm[[1]], "model"))) {
    out <- mm_lm_i(mm)
    if (!drop_list_single_group) {
      out <- list(out)
    }
  } else {
    out <- lapply(mm,
                  mm_lm_i)
  }
  return(out)
}

#' @noRd
mm_lm_i <- function(mm) {
  # TODO:
  # - Check whether the transpose of nox-beta is in echelon form.
  # - Handle models with no y-variables (e.g., CFA).
  model <- attr(mm, "model")
  if (is.null(model)) {
    stop("Model syntax not found")
  }
  # Handle labels
  tmp <- lavaan::lavParseModelString(model,
                                     as.data.frame. = TRUE)
  tmp$label <- ""
  tmp$mod.idx <- 0
  attr(tmp, "modifiers") <- list()
  attr(tmp, "constraints") <- list()
  fit1 <- lavaan::sem(tmp,
                      do.fit = FALSE)
  mm1 <- lavaan::lavInspect(fit1,
                            "partable")
  beta1 <- mm1$beta
  psi1 <- mm1$psi
  if (is.null(beta1)) {
    # No regression paths. A CFA model.
    # Set psi to theta, the covariance
    # matrix of factor scores to be
    # generated.
    out <- list(lm_y = NULL,
                psi_pure_x = NULL,
                psi = mm$theta)
    return(out)
  }
  ix <- apply(beta1,
              MARGIN = 1,
              function(x) all(x == 0))
  x_names <- names(ix)[ix]
  y_names <- setdiff(colnames(beta1), x_names)
  beta <- mm$beta
  psi <- mm$psi
  gen_lm <- function(y) {
    beta1_i <- beta1[y, , drop = FALSE]
    beta_i <- beta[y, beta1_i > 0, drop = FALSE]
    psi_i <- psi[beta1_i > 0, beta1_i > 0, drop = FALSE]
    list(beta = beta_i,
         psi = psi_i)
  }
  lm_y <- sapply(y_names,
                 gen_lm,
                 simplify = FALSE)
  psi_pure_x <- psi[x_names, x_names, drop = FALSE]
  out <- list(lm_y = lm_y,
              psi_pure_x = psi_pure_x,
              psi = psi)
  return(out)
}

#' @noRd
# Fix covariances in a parameter table
dup_cov <- function(ptable) {
  i <- ptable$op == "~~"
  if (any(i)) {
    j <- match(c("rhs", "op", "lhs"),
               colnames(ptable))
    j <- c(j,
           sort(setdiff(seq_len(ncol(ptable)), j)))
    ptable_rev <- ptable[i, j]
    colnames(ptable_rev) <- colnames(ptable)
    ptable <- rbind(ptable,
                    ptable_rev)
    rownames(ptable) <- NULL
  }
  ptable
}

#' @noRd
# Find group ID from the name
get_gp_id <- function(pop_es,
                  ngroups) {
  gp_id0 <- strsplit(
                names(pop_es),
                split = ".",
                fixed = TRUE)[[1]]
  if (length(gp_id0) == 1) {
    # Assume to be group 1
    gp_id <- 1
    new_name <- names(pop_es)
  } else {
    gp_id1 <- gp_id0[length(gp_id0)]
    gp_id <- as.numeric(gsub("g", "", gp_id1, fixed = TRUE))
    if (!(gp_id %in% seq_len(ngroups))) {
      stop("Group ID not valid: ", gp_id1)
    }
    new_name <- gsub(paste0(".", gp_id1), "",
                     names(pop_es),
                     fixed = TRUE)
  }
  return(list(gp_id = gp_id,
              new_name = new_name))
}

library(testthat)
suppressMessages(library(lavaan))

test_that("ptable_pop: Nonpositive definite", {

# Simple mediation model

mod <-
"
m ~ x
y ~ m + x
"

pop_es <- c("m ~ x" = ".90",
            "y ~ m" = ".90",
            "y ~ x" = ".90")

expect_error(ptable_pop(model = mod,
                        pop_es = pop_es))

})

library(testthat)
suppressMessages(library(lavaan))

check_gen_dat <- function(model,
                          pop_es,
                          n = 10000,
                          seed = NULL,
                          number_of_indicators = NULL,
                          reliability = NULL,
                          keep_f_scores = FALSE) {
  set.seed(seed)
  ptable <- ptable_pop(model,
                       pop_es = pop_es,
                       standardized = TRUE)
  mm_out <- model_matrices_pop(ptable)
  mm_lm_out <- mm_lm(mm_out)
  mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                              n = n,
                              number_of_indicators = number_of_indicators,
                              reliability = reliability,
                              keep_f_scores = keep_f_scores)
  model <- add_indicator_syntax(model,
                                number_of_indicators = number_of_indicators,
                                reliability = reliability)
  fit <- lavaan::sem(model,
             data = mm_lm_dat_out,
             fixed.x = FALSE)
  tmp <- ptable
  tmp$est <- tmp$start
  fit0 <- lavaan::sem(tmp,
              do.fit = FALSE,
              fixed.x = FALSE)
  out <- list(ptable = ptable,
              mm_out = mm_out,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out,
              fit = fit,
              fit0 = fit0)
  out
}

test_check_out <- function(check_out) {
  std <- lavaan::standardizedSolution(check_out$fit,
                                      se = FALSE)
  expect_equal(std$est.std[1:3],
               coef(check_out$fit0)[1:3],
               ignore_attr = TRUE,
               tolerance = 1e-1)
}

test_that("ptable_pop: With indicator scores", {

# Simple mediation model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = "m",
                         "m ~ x" = "l",
                         "y ~ x" = "n")

check_out <- check_gen_dat(model_simple_med,
                           model_simple_med_es,
                           seed = 1234,
                           number_of_indicators = c(y = 4,
                                                    x = 3,
                                                    m = 5),
                           reliability = c(x = .70,
                                           m = .80,
                                           y = .60))

test_check_out(check_out)

# Parallel mediation model
# Correlated errors

model_med_parallel <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
m1 ~~ m2
"

model_med_parallel_es <- c(".beta." = "s",
                           "y ~ x" = "-m",
                           "m1 ~ x" = "l",
                           "m2 ~ x" = "s",
                           "y ~ m1" = "m",
                           "m1 ~~ m2" = "l")

check_out <- check_gen_dat(model_med_parallel,
                           model_med_parallel_es,
                           seed = 1234,
                           number_of_indicators = c(y = 4,
                                                    x = 3,
                                                    m2 = 5),
                           reliability = c(x = .70,
                                           m2 = .80,
                                           y = .60))

test_check_out(check_out)

# Parallel mediation model
# Correlated errors not specified in the model

model_med_parallel <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
"

model_med_parallel_es <- c(".beta." = "s",
                           "y ~ x" = "-m",
                           "m1 ~~ m2" = "l")

check_out <- check_gen_dat(model_med_parallel,
                                          model_med_parallel_es,
                                          seed = 1234,
                           number_of_indicators = c(y = 4,
                                                    m1 = 2,
                                                    m2 = 5),
                           reliability = c(m1 = .70,
                                           m2 = .80,
                                           y = .60))
test_check_out(check_out)

# Moderated mediation model

model_mod_med <-
"
m ~ x + w
y ~ m + z + m:z + x + u
"

model_mod_med_es <- c(".beta." = "s",
                      "y ~ x" = "-m",
                      "x ~~ w + z + u" = "s",
                      "w ~~ z + u" = "l")

check_out <- check_gen_dat(model_mod_med,
                           model_mod_med_es,
                           seed = 1234,
                           number_of_indicators = c(x = 4,
                                                    w = 2,
                                                    y = 5),
                           reliability = c(x = .70,
                                           w = .80,
                                           y = .60))

test_check_out(check_out)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models", {

model2 <-
"
m ~ x
y ~  m + x
"

model2_es <- list("m ~ x" = "-m",
                  "y ~ m" = "l",
                  "y ~ x" = c("m", "l", "n"))
set.seed(1234)
out <- ptable_pop(model = model2,
                  pop_es = model2_es,
                  n_std = 5e4)
mm_out <- model_matrices_pop(out)
mm_lm_out <- mm_lm(mm_out)

model2_es_ng <- list("m ~ x" = "-m",
                     "y ~ m" = "l",
                     "y ~ x" = c("n"))
out_ng <- ptable_pop(model = model2,
                     pop_es = model2_es_ng)
mm_out_ng <- model_matrices_pop(out_ng)

tmp1 <- out[out$group == 3, ]
tmp2 <- out_ng
tmp1$ptlabel <- lav_partable_labels(tmp1)
tmp2$ptlabel <- lav_partable_labels(tmp2)

expect_equal(tmp1[tmp1$ptlabel == "m~~m", "start"],
             tmp2[tmp2$ptlabel == "m~~m", "start"])

expect_identical(mm_out[[3]],
                 mm_out_ng)

data_i <- sim_data_i(model = model2,
                     pop_es = model2_es,
                     n = 200,
                     number_of_indicators = 4,
                     reliability = .80,
                     seed = 1234)

fit_i <- fit_model_i(data_i)

dats <- sim_data(nrep = 9,
                 model = model2,
                 pop_es = model2_es,
                 n = c(100, 200, 300),
                 number_of_indicators = c(x = 4, y = 4, m = 4),
                 reliability = list(x = .80,
                                    y = c(.80, .80, .80),
                                    m = .80),
                 iseed = 1234,
                 progress = !is_testing())
expect_equal(coef(dats[[1]]$fit0)["m~x"],
             -.30,
             ignore_attr = TRUE)
expect_equal(nrow(dats[[1]]$mm_lm_dat_out),
             100 + 200 + 300)
expect_true(all(paste0("x", 1:4) %in%
                colnames(dats[[1]]$mm_lm_dat_out)))
expect_true(all(paste0("y", 1:4) %in%
                colnames(dats[[1]]$mm_lm_dat_out)))
expect_true(all(paste0("m", 1:4) %in%
                colnames(dats[[1]]$mm_lm_dat_out)))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models", {

model2 <-
"
m1 ~ x
m2 ~ x
y ~  m1 + m2 + x
"

model2_es <- list(".beta." = "-s",
                  "m1 + m2 ~ x" = "-m",
                  "y ~ m1" = "l",
                  "y ~ m2" = c("m", "l"))
set.seed(1234)
out <- ptable_pop(model = model2,
                  pop_es = model2_es,
                  n_std = 5e4)
mm_out <- model_matrices_pop(out)
mm_lm_out <- mm_lm(mm_out)

model2_es_ng <- c("y ~ x" = "-s",
                  "m1 + m2 ~ x" = "-m",
                  "y ~ m1" = "l",
                  "y ~ m2" = "m")
out_ng <- ptable_pop(model = model2,
                     pop_es = model2_es_ng)
mm_out_ng <- model_matrices_pop(out_ng)

tmp1 <- out[out$group == 1, ]
tmp2 <- out_ng
tmp1$ptlabel <- lav_partable_labels(tmp1)
tmp2$ptlabel <- lav_partable_labels(tmp2)

expect_equal(tmp1[tmp1$ptlabel == "y~~m2", "start"],
             tmp2[tmp2$ptlabel == "y~~m2", "start"])

expect_equal(tmp1[tmp1$ptlabel == "y~~x", "start"],
             tmp2[tmp2$ptlabel == "y~~x", "start"])

expect_identical(mm_out[[1]],
                 mm_out_ng)

data_i <- sim_data_i(model = model2,
                     pop_es = model2_es,
                     n = 200,
                     number_of_indicators = c(m1 = 3, m2 = 4),
                     reliability = c(m2 = .60, m1 = .70),
                     seed = 1234)

fit_i <- fit_model_i(data_i)

dats <- sim_data(nrep = 9,
                 model = model2,
                 pop_es = model2_es,
                 n = c(300, 100),
                 number_of_indicators = c(m1 = 3, m2 = 4),
                 reliability = c(m2 = .60, m1 = .70),
                 iseed = 1234,
                 progress = !is_testing())
expect_equal(coef(dats[[1]]$fit0)["y~m2"],
             .30,
             ignore_attr = TRUE)
expect_equal(nrow(dats[[1]]$mm_lm_dat_out),
             300 + 100)
expect_true(all(paste0("m1", 1:3) %in%
                colnames(dats[[1]]$mm_lm_dat_out)))
expect_true(all(paste0("m2", 1:4) %in%
                colnames(dats[[1]]$mm_lm_dat_out)))
})

library(testthat)
suppressMessages(library(lavaan))

test_that("update par_pop", {

# Single-group

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = "m",
                         "m ~ x" = "l",
                         "y ~ x" = "n")

par_pop <- pop_es2par_pop(model_simple_med_es,
                          es1 = eval(formals(ptable_pop)$es1),
                          es2 = eval(formals(ptable_pop)$es2),
                          es_ind = eval(formals(ptable_pop)$es_ind),
                          model = model_simple_med)

par_pop

par_pop_add <- pop_es2par_pop(c("m ~ x" = "s"),
                              es1 = eval(formals(ptable_pop)$es1),
                              es2 = eval(formals(ptable_pop)$es2),
                              es_ind = eval(formals(ptable_pop)$es_ind),
                              model = model_simple_med)

par_pop2 <- update_par_pop(add = par_pop_add,
                           par_pop = par_pop)

expect_equal(par_pop2[[1]][1, "es"],
             "s")

ptable <- ptable_pop(model = model_simple_med,
                     pop_es = model_simple_med_es)

ptable2 <- update_ptable_pop(ptable,
                             new_pop_es = c("m ~ x" = "s"))

expect_equal(ptable2[1, "start"],
             .10)

expect_equal(ptable[1, "start"],
             .50)

# Multigroup

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- list(".beta." = "m",
                            "m ~ x" = c("n", "s"),
                            "y ~ x" = "n")

par_pop <- pop_es2par_pop(model_simple_med_es,
                          es1 = eval(formals(ptable_pop)$es1),
                          es2 = eval(formals(ptable_pop)$es2),
                          es_ind = eval(formals(ptable_pop)$es_ind),
                          model = model_simple_med)

par_pop

par_pop_add <- pop_es2par_pop(list("m ~ x" = c("l", "m")),
                              es1 = eval(formals(ptable_pop)$es1),
                              es2 = eval(formals(ptable_pop)$es2),
                              es_ind = eval(formals(ptable_pop)$es_ind),
                              model = model_simple_med)

par_pop2 <- update_par_pop(add = par_pop_add,
                           par_pop = par_pop)

expect_equal(par_pop2[[1]][1, "es"],
             "l")
expect_equal(par_pop2[[2]][1, "es"],
             "m")

ptable <- ptable_pop(model = model_simple_med,
                     pop_es = model_simple_med_es)

ptable2 <- update_ptable_pop(ptable,
                             new_pop_es = list("m ~ x" = c("s", "l")))

expect_equal(ptable2[1, "start"],
             .10)
expect_equal(ptable2[7, "start"],
             .50)

expect_equal(ptable[1, "start"],
             .00)
expect_equal(ptable[7, "start"],
             .10)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("ptable_pop: Wrong pop_es", {

# Simple mediation model

mod <-
"
m ~ x
y ~ m + x
"

pop_es <- c("m ~ x2" = "m",
            "y ~ m" = "s",
            "y ~ x" = "s")

expect_warning(ptable_pop(model = mod,
                          pop_es = pop_es))
})

library(testthat)
suppressMessages(library(lavaan))


check_gen_dat <- function(model,
                          pop_es,
                          n = 10000,
                          seed = NULL) {
  set.seed(seed)
  ptable <- ptable_pop(model,
                       pop_es = pop_es,
                       standardized = TRUE)
  mm_out <- model_matrices_pop(ptable)
  mm_lm_out <- mm_lm(mm_out)
  mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                              n = n)
  out <- list(ptable = ptable,
              mm_out = mm_out,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out)
  out
}


test_that("add_indicator_scores", {

# Simple mediation model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = "m",
                         "m ~ x" = "l",
                         "y ~ x" = "n")

check_out <- check_gen_dat(model_simple_med,
                           model_simple_med_es,
                           seed = 1234,
                           n = 50000)
dat <- as.matrix(check_out$mm_lm_dat_out)
head(dat)

set.seed(1357)
dat0 <- add_indicator_scores(dat,
                             ps = c(x = 4, m = 7),
                             rels = c(x = .80, m = .70),
                             ds = c(m = 0.05, x = 0.10),
                             refs = c(m = "medium", x = "strongest"))
mod <-
"
x =~ x1 + x2 + x3 + x4
m =~ m1 + m2 + m3 + m4 + m5 + m6 + m7
"
fit <- cfa(mod, data = dat0)
std <- standardizedSolution(fit)

expect_equal(sort(std[1:4, "est.std"]) /
             sort(lambda_from_reliability(p = 4, omega = .80, d = .10)),
             rep(1, 4),
             tolerance = 1e-2)

expect_equal(sort(std[1:7 + 4, "est.std"]) /
             sort(lambda_from_reliability(p = 7, omega = .70, d = .05)),
             rep(1, 7),
             tolerance = 1e-2)

# lambda_from_reliability(p = 4, omega = .80, d = .10, ref = "strongest")
expect_true(which.max(std[1:4, "est.std"]) == 1)

# lambda_from_reliability(p = 7, omega = .70, d = .05, ref = "medium")
expect_true(which.max(std[1:7 + 4, "est.std"]) != 1)
expect_true(which.min(std[1:7 + 4, "est.std"]) != 1)

set.seed(1357)
dat0 <- add_indicator_scores(dat,
                             ps = c(x = 4, m = 7),
                             rels = c(x = .80, m = .70),
                             ds = c(m = 0.05, x = 0.10),
                             refs = c(m = "weakest", x = "medium"))
mod <-
"
x =~ x1 + x2 + x3 + x4
m =~ m1 + m2 + m3 + m4 + m5 + m6 + m7
"
fit <- cfa(mod, data = dat0)
std <- standardizedSolution(fit)

expect_equal(sort(std[1:4, "est.std"]) /
             sort(lambda_from_reliability(p = 4, omega = .80, d = .10)),
             rep(1, 4),
             tolerance = 1e-2)

expect_equal(sort(std[1:7 + 4, "est.std"]) /
             sort(lambda_from_reliability(p = 7, omega = .70, d = .05)),
             rep(1, 7),
             tolerance = 1e-2)

# lambda_from_reliability(p = 4, omega = .80, d = .10, ref = "medium")
expect_true(which.max(std[1:4, "est.std"]) != 1)
expect_true(which.min(std[1:4, "est.std"]) != 1)

# lambda_from_reliability(p = 7, omega = .70, d = .05, ref = "weakest")
expect_true(which.min(std[1:7 + 4, "est.std"]) == 1)

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Generate nonnormal indicator", {

mod <-
"
m ~ x
y ~ m + x
"

set.seed(1234)
n <- 100000
pop_es <- c(".beta." = "m")

ptable <- ptable_pop(model = mod,
                     pop_es = pop_es)
mm_out <- model_matrices_pop(ptable)
mm_lm_out <- mm_lm(mm_out)

set.seed(1234)
mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                            n = 10000,
                            number_of_indicators = c(x = 4, y = 4, m = 4),
                            reliability = c(x = .20, y = .60, m = .70),
                            e_fun = list(x = list(rexp_rs),
                                         m = list(rbinary_rs,
                                                  p1 = .50)))
chk <- mm_lm_dat_out
head(chk)
tmp <- mm_lm_dat_out[, "x2"]
expect_true(mean(tmp) > median(tmp))

})

library(testthat)
suppressMessages(library(lavaan))


check_gen_dat <- function(model,
                          pop_es,
                          n = 10000,
                          seed = NULL) {
  set.seed(seed)
  ptable <- ptable_pop(model,
                       pop_es = pop_es,
                       standardized = TRUE)
  mm_out <- model_matrices_pop(ptable)
  mm_lm_out <- mm_lm(mm_out)
  mm_lm_dat_out <- mm_lm_data(mm_lm_out,
                              n = n)
  out <- list(ptable = ptable,
              mm_out = mm_out,
              mm_lm_out = mm_lm_out,
              mm_lm_dat_out = mm_lm_dat_out)
  out
}


test_that("add_indicator_scores", {

# Simple mediation model

model_simple_med <-
"
m ~ x
y ~ m + x
"

model_simple_med_es <- c(".beta." = "m",
                         "m ~ x" = "l",
                         "y ~ x" = "n")

check_out <- check_gen_dat(model_simple_med,
                           model_simple_med_es,
                           seed = 1234)
dat <- as.matrix(check_out$mm_lm_dat_out)
head(dat)
dat0 <- add_indicator_scores(dat,
                             ps = c(x = 4, m = 2),
                             rels = c(x = .30, m = .70))
expect_setequal(colnames(dat0),
                c("y", "x1", "x2", "x3", "x4", "m1", "m2"))
dat0 <- add_indicator_scores(dat,
                             ps = c(x = 4, m = 2),
                             rels = c(x = .30, m = .70),
                             keep_f_scores = TRUE)
expect_setequal(colnames(dat0),
                c("m", "y", "x", "x1", "x2", "x3", "x4", "m1", "m2"))
dat0 <- add_indicator_scores(dat,
                             ps = c(x = 4, m = 2, y = 10),
                             rels = c(x = .30, m = .70, y = .90))
expect_false(any(c("x", "m", "y") %in% colnames(dat0)))

expect_error(add_indicator_scores(dat,
                                  ps = c(x = 4, m = 2, y2 = 10),
                                  rels = c(x = .30, m = .70, y = .90)))
expect_error(add_indicator_scores(dat,
                                  ps = c(x = 4, m = 2),
                                  rels = c(x = .30, m = .70, y = .90)))
expect_error(add_indicator_scores(dat,
                                  ps = c(x = 4, m = 2, z = 10),
                                  rels = c(x = .30, m = .70, z = .90)))

})

library(testthat)
suppressMessages(library(lavaan))

# Moderated mediation multigroup model.
# Some are latent factors.

test_that("Multigroup models", {

model1 <-
"
m1 ~ x
m2 ~ m1 + x
y ~  m2 + m1 + x + w + m2:w
"
model1_es <- list("m1 ~ x" = "m",
                  "m2 ~ m1" = "s",
                  "y ~ m2:w" = c("-m", "n"))
set.seed(1234)
out <- ptable_pop(model = model1,
                  pop_es = model1_es,
                  n_std = 5e4)
mm_out <- model_matrices_pop(out)
mm_lm_out <- mm_lm(mm_out)

model1_es_ng <- c("m1 ~ x" = "m",
                  "m2 ~ m1" = "s",
                  "y ~ m2:w" = "-m")
out_ng <- ptable_pop(model = model1,
                     pop_es = model1_es_ng)
mm_out_ng <- model_matrices_pop(out_ng)

tmp1 <- out[out$group == 1, ]
tmp2 <- out_ng
tmp1$ptlabel <- lav_partable_labels(tmp1)
tmp2$ptlabel <- lav_partable_labels(tmp2)

expect_equal(tmp1[tmp1$ptlabel == "y~~m2:w", "start"],
             tmp2[tmp2$ptlabel == "y~~m2:w", "start"],
             tolerance = 1e-3)

tmp <- mm_out[[1]]
tmp$nu <- NULL
tmp$alpha <- NULL
expect_identical(tmp,
                 mm_out_ng,
                 tolerance = 1e-3)

data_i <- sim_data_i(model = model1,
                     pop_es = model1_es,
                     n = 200,
                     number_of_indicators = c(m1 = 3, x = 4),
                     reliability = list(x = .60,
                                        m1 = c(.80, .70)),
                     seed = 1234)
fit_i <- fit_model_i(data_i)
summary(fit_i)

data_i2 <- sim_data_i(model = model1,
                      ptable = out,
                      mm_out = mm_out,
                      mm_lm_out = mm_lm_out,
                      n = 200,
                      number_of_indicators = c(m1 = 3, x = 4),
                      reliability = list(x = .60,
                                         m1 = c(.80, .70)),
                      seed = 1234)
fit_i2 <- fit_model_i(data_i)

expect_equal(fitMeasures(fit_i, "chisq"),
             fitMeasures(fit_i2, "chisq"))

})

library(testthat)
suppressMessages(library(lavaan))

test_that("Multigroup models", {

model1 <-
"
m1 ~ x + c1
m2 ~ m1 + x2 + c1
y ~  m2 + m1 + x + w + x:w + c1
"
model1_es <- list("m1 ~ x" = "-m",
                  "m2 ~ m1" = "s",
                  "y ~ m2" = "l",
                  "y ~ x" = c("m", "l", "n"),
                  "y ~ w" = "s",
                  "y ~ x:w" = "s",
                  "x ~~ w" = c("s", "m", "l"))
set.seed(1234)
out <- ptable_pop(model = model1,
                  pop_es = model1_es,
                  n_std = 5e4)
mm_out <- model_matrices_pop(out)
mm_lm_out <- mm_lm(mm_out)

model1_es_ng <- list("m1 ~ x" = "-m",
                     "m2 ~ m1" = "s",
                     "y ~ m2" = "l",
                     "y ~ x" = c("n"),
                     "y ~ w" = "s",
                     "y ~ x:w" = "s",
                     "x ~~ w" = c("l"))
out_ng <- ptable_pop(model = model1,
                     pop_es = model1_es_ng)
mm_out_ng <- model_matrices_pop(out_ng)

tmp1 <- out[out$group == 3, ]
tmp2 <- out_ng
tmp1$ptlabel <- lav_partable_labels(tmp1)
tmp2$ptlabel <- lav_partable_labels(tmp2)

expect_equal(tmp1[tmp1$ptlabel == "m2~~m2", "start"],
             tmp2[tmp2$ptlabel == "m2~~m2", "start"],
             tolerance = 1e-3)

expect_identical(mm_out[[3]],
                 mm_out_ng,
                 tolerance = 1e-3)

data_i <- sim_data_i(model = model1,
                     pop_es = model1_es,
                     n = 200,
                     seed = 1234)
fit_i <- fit_model_i(data_i)
summary(fit_i)

data_i2 <- sim_data_i(model = model1,
                      ptable = out,
                      mm_out = mm_out,
                      mm_lm_out = mm_lm_out,
                      n = 200,
                      seed = 1234)
fit_i2 <- fit_model_i(data_i)

expect_equal(fitMeasures(fit_i, "chisq"),
             fitMeasures(fit_i2, "chisq"))

})

#' @title Test Several Conditional Indirect Effects
#'
#' @description Test several conditional
#' indirect effects
#' for a `power4test` object.
#'
#' @details
#' This function is to be used in
#' [power4test()] for testing several
#' conditional
#' indirect effects, by setting it
#' to the `test_fun` argument.
#'
#' It uses [manymome::cond_indirect_effects()]
#' to do the test. It can be used on
#' models fitted by [lavaan::sem()]
#' or fitted by a sequence of calls
#' to [stats::lm()], although only
#' nonparametric bootstrap confidence
#' interval is supported for models
#' fitted by regression using
#' [stats::lm()].
#'
#' It can also be used to test
#' conditional effects on a direct path
#' with no mediator. Just omit `m` when
#' calling the function.
#'
#' @return
#' In its normal usage, it returns
#' the output returned by
#' [manymome::cond_indirect_effects()],
#' with the following modifications:
#'
#' - `est`: The estimated
#'  conditional indirect effects.
#'
#' - `cilo` and `cihi`: The
#'  lower and upper limits of the
#'  confidence interval (95% by
#'  default), respectively.
#'
#' - `sig`: Whether a test by confidence
#'  interval is significant (`1`) or
#'  not significant (`0`).
#'
#' - `test_label`: A column of labels
#'  generated to label the conditional
#'  effects.
#'
#' @inheritParams test_indirect_effect
#'
#' @param fit The fit object, to be
#' passed to [manymome::cond_indirect_effects()].
#'
# @param x <- Inherited
#'
# @param m <- Inherited
#'
# @param y <- Inherited
#'
#' @param wlevels The output of
#' [manymome::merge_mod_levels()], or
#' the moderator(s) to be passed to
#' [manymome::mod_levels_list()].
#' If all the moderators can be
#' represented by one variable, that is,
#' each moderator is (a) a numeric
#' variable, (b) a dichotomous
#' categorical variable, or (c) a
#' factor or string variable used in
#' [stats::lm()] in fit, then it is a
#' vector of the names of the moderators
#' as appeared in the data frame. If at
#' least one of the moderators is a
#' categorical variable represented by
#' more than one variable, such as
#' user-created dummy variables used in
#' [lavaan::sem()], then it must be a
#' list of the names of the moderators,
#' with such moderators represented by
#' a vector of names. For example:
#' `list("w1", c("gpgp2", "gpgp3")`,
#' the first moderator `w1` and the
#' second moderator a three-category
#' variable represented by `gpgp2` and
#' `gpgp3`. See the help page of
#' [manymome::cond_indirect_effects()]
#' for further details.
#'
# @param mc_ci <- Inherited
#'
# @param mc_out <- Inherited
#'
# @param boot_ci <- Inherited
#'
# @param boot_out <- Inherited
#'
#' @param compare_groups If the model
#' is a multigroup model, compute and
#' test group differences for all pairwise
#' combinations of the groups. Ignored
#' if the model is a single-group model.
#'
#' @param ... Additional arguments to
#' be passed to [manymome::cond_indirect_effects()].
#'
# @param fit_name <- Inherited
#'
# @param get_map_names <- Inherited
#'
# @param get_test_name <- Inherited
#'
#' @seealso [power4test()]
#'
#' @examples
#'
#' # Specify the model
#'
#' model_simple_mod <-
#' "
#' m ~ x + w + x:w
#' y ~ m + x
#' "
#'
#' # Specify the population values
#'
#' model_simple_mod_es <-
#' "
#' y ~ m: l
#' y ~ x: n
#' m ~ x: m
#' m ~ w: n
#' m ~ x:w: l
#' "
#'
#' # Simulate the data
#'
#' # Set nrep to a larger value in real analysis, such as 400
#' sim_only <- power4test(nrep = 5,
#'                        model = model_simple_mod,
#'                        pop_es = model_simple_mod_es,
#'                        n = 100,
#'                        R = 100,
#'                        do_the_test = FALSE,
#'                        iseed = 1234)
#'
#' # Do the tests in each replication
#'
#' test_out <- power4test(object = sim_only,
#'                        test_fun = test_cond_indirect_effects,
#'                        test_args = list(x = "x",
#'                                         m = "m",
#'                                         y = "y",
#'                                         wlevels = c("w"),
#'                                         mc_ci = TRUE))
#' print(test_out,
#'       test_long = TRUE)
#'
#' @export

test_cond_indirect_effects <- function(fit = fit,
                                       x = NULL,
                                       m = NULL,
                                       y = NULL,
                                       wlevels = NULL,
                                       mc_ci = TRUE,
                                       mc_out = NULL,
                                       boot_ci = FALSE,
                                       boot_out = NULL,
                                       check_post_check = TRUE,
                                       test_method = NULL,
                                       compare_groups = FALSE,
                                       ...,
                                       fit_name = "fit",
                                       get_map_names = FALSE,
                                       get_test_name = FALSE) {

  # ==== Enable pvalue? ====

  args <- list(...)
  if (!is.null(mc_out)) {
    R <- length(mc_out)
  } else if (!is.null(boot_out)) {
    R <- length(boot_out)
  } else {
    R <- NULL
  }
  R <- args$R %||% formals(manymome::cond_indirect_effects)$R
  ci_level <- args$level %||% formals(manymome::cond_indirect)$level
  R_bz_ok <- isTRUE(R %in% R_extrapolate(alpha = 1 - ci_level))
  bz_not_FALSE <- !isFALSE(options("power4mome.bz"))
  test_method_NULL <- is.null(test_method)
  if (R_bz_ok &&
      bz_not_FALSE &&
      test_method_NULL) {
    test_method <- "pvalue"
  } else {
    test_method <- match.arg(test_method,
                             c("ci", "pvalue"))
  }

  internal_options <- list()
  if (test_method == "pvalue") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = TRUE,
                                               pvalue_min_size = -Inf))
  }
  if (test_method == "ci") {
    internal_options <- utils::modifyList(internal_options,
                                          list(skip_ci = FALSE))
  }
  if (fit_name != "fit") {
    mc_name <- paste0(fit_name, "_mc_out")
    boot_name <- paste0(fit_name, "_boot_out")
  } else {
    mc_name <- "mc_out"
    boot_name <- "boot_out"
  }
  map_names <- c(fit = fit_name,
                 mc_out = mc_name,
                 boot_out = boot_name)
  if (get_map_names) {
    return(map_names)
  }
  if (get_test_name) {
    tmp <- paste0(c(x, m, y),
                  collapse = "->")
    args <- as.list(match.call())
    tmp2 <- character(0)
    if (isTRUE(args$standardized_x) && !isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' standardized)")
    }
    if (!isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('y' standardized)")
    }
    if (isTRUE(args$standardized_x) && isTRUE(args$standardized_y)) {
      tmp <- paste0(tmp, " ('x' and 'y' standardized)")
    }
    return(paste0("test_cond_indirect_effects: ", tmp, collapse = ""))
  }
  if (boot_ci) mc_ci <- FALSE
  if (inherits(fit, "lavaan")) {
    ngroups <- lavaan::lavInspect(fit, "ngroups")
    fit_ok <- lavaan::lavInspect(fit, "converged") &&
              (suppressWarnings(lavaan::lavInspect(fit, "post.check") ||
               !check_post_check))
  } else {
    ngroups <- 1
    fit_ok <- TRUE
  }

  # ==== Compute conditional indirect effects ====

  if (fit_ok) {
    if (is.null(wlevels)) {

      # ==== wlevels is NULL: Multigroup? ====

      if (ngroups == 1) {
        stop("wlevels not supplied but the model is not a multigroup model.")
      }
      out <- tryCatch(manymome::cond_indirect_effects(
                                          x = x,
                                          y = y,
                                          m = m,
                                          fit = fit,
                                          mc_ci = mc_ci,
                                          mc_out = mc_out,
                                          boot_ci = boot_ci,
                                          boot_out = boot_out,
                                          progress = FALSE,
                                          ...,
                                          internal_options = internal_options),
                    error = function(e) e)
    } else {

      # ==== wlevels is not NULL ====

      out <- tryCatch(manymome::cond_indirect_effects(
                                          x = x,
                                          y = y,
                                          m = m,
                                          wlevels = wlevels,
                                          fit = fit,
                                          mc_ci = mc_ci,
                                          mc_out = mc_out,
                                          boot_ci = boot_ci,
                                          boot_out = boot_out,
                                          progress = FALSE,
                                          ...,
                                          internal_options = internal_options),
                    error = function(e) e)
    }
  } else {
    out <- NA
  }
  if (inherits(out, "error") ||
      identical(out, NA)) {
    out2 <- data.frame(
              est = as.numeric(NA),
              cilo = as.numeric(NA),
              cihi = as.numeric(NA),
              sig = as.numeric(NA),
              pvalue = as.numeric(NA)
            )
    return(out2)
  }
  # TODO:
  # - Update to support multigroup model with moderation in the future

  # ==== Prepare the output ====

  if (ngroups > 1) {

    # ==== Multigroup model ====

    if (compare_groups) {

      # ==== Compare groups ====

      level <- attr(out, "full_output")[[1]]$level
      out_org <- out
      out <- all_cond_diff(
                  out_org,
                  level = level
                )
    }
    tmp <- out$Group
    tmp2 <- paste0(c(x, m, y),
                   collapse = "->")
    test_label <- paste(tmp2, "|", tmp)
    out2 <- as.data.frame(out,
                          check.names = FALSE)
    out2 <- cbind(test_label = test_label,
                  out2)
  } else {

    # ==== Single-group model ====

    tmp <- rownames(attr(out, "wlevels"))
    tmp2 <- paste0(c(x, m, y),
                   collapse = "->")
    test_label <- paste(tmp2, "|", tmp)
    out2 <- as.data.frame(out,
                          check.names = FALSE)
    out2 <- cbind(test_label = test_label,
                  out2)
  }
  # Add pvalues
  tmpfct <- function(x) {
    x$mc_p %||% (x$boot_p %||% as.numeric(NA))
  }
  out_p <- sapply(attr(out, "full_output"),
                  tmpfct,
                  USE.NAMES = FALSE)
  i <- match("Sig", colnames(out2))
  out2 <- cbind(out2[, 1:i],
                pvalue = out_p,
                out2[, -c(1:i)])

  tmp <- colnames(out2)
  if ("std" %in% tmp) {
    tmp <- gsub("ind", "est_raw", tmp, fixed = TRUE)
    tmp <- gsub("std", "est", tmp, fixed = TRUE)
  } else {
    tmp <- gsub("ind", "est", tmp, fixed = TRUE)
  }
  tmp <- gsub("CI.lo", "cilo", tmp, fixed = TRUE)
  tmp <- gsub("CI.hi", "cihi", tmp, fixed = TRUE)
  tmp <- gsub("Sig", "sig", tmp, fixed = TRUE)
  colnames(out2) <- tmp
  if (test_method == "ci") {
    out1 <- ifelse((out2$cilo > 0) | (out2$cihi < 0),
                  yes = 1,
                  no = 0)
    out2$sig <- out1
  }
  if (test_method == "pvalue") {
    out_all <- attr(out,
                    "full_output")
    ci_type <- NA
    if (!is.null(out_all[[1]]$mc_indirect)) {
      ci_type <- "mc"
    }
    if (!is.null(out_all[[1]]$boot_indirect)) {
      ci_type <- "boot"
    }
    p_name <- switch(ci_type,
                     mc = "mc_p",
                     boot = "boot_p")
    p_all <- sapply(out_all,
                    \(x) unname(x[[p_name]]),
                    USE.NAMES = FALSE)
    out2$pvalue <- p_all
    out2$sig <- ifelse(
                out2$pvalue < (1 -  out_all[[1]]$level),
                yes = 1,
                no = 0
              )
    est_name <- switch(ci_type,
                       mc = "mc_indirect",
                       boot = "boot_indirect")
    est_all <- lapply(out_all,
                      \(x) x[[est_name]])
    bz_alpha_ok <- isTRUE(all.equal(1 - out_all[[1]]$level,
                                    getOption("power4mome.bz.alpha",
                                                        default = .05)))
    R <- sapply(est_all,
                length)
    nlt0 <- sapply(est_all,
                   \(x) sum(as.numeric(x < 0)))
    out2$R <- R
    out2$nlt0 <- nlt0
    out2$alpha <- 1 - out_all[[1]]$level
    if (bz_alpha_ok) {
      tmp <- lapply(out_all,
              function(x) {
                boot_est <- x$boot_indirect %||% x$mc_indirect
                boot_sig <- bz_sig_partition(
                              boot_est,
                              alpha = 1 - x$level
                            )
                boot_sig
              })
      tmp <- do.call(rbind,
                    unname(tmp))
      out2 <- cbind(out2,
                    tmp)
    }
  }
  rownames(out2) <- NULL
  attr(out2, "test_label") <- "test_label"
  return(out2)
}

#' @noRd
# Input:
# - A cond_indirect_effects object
# Output:
# - A data frame of pairwise comparison
all_cond_diff <- function(
  out,
  level
) {
  k <- nrow(out)
  all_ij <- utils::combn(
              seq_len(k),
              m = 2,
              simplify = FALSE
            )
  outa <- lapply(
              all_ij,
              \(xx) {
                ij_cond_diff(
                  out,
                  from = xx[1],
                  to = xx[2],
                  level = level
                )
              }
            )
  full_output <- lapply(
                    outa,
                    \(x) attr(x, "full_output")
                  )
  outz <- do.call(
            rbind,
            outa
          )
  attr(outz, "full_output") <- full_output
  outz
}

#' @noRd
# Input:
# - A cond_indirect_effects object
# Output:
# - A vector with results from a cond_indirect_diff object
ij_cond_diff <- function(
  out,
  from,
  to,
  level
) {
  outa <- manymome::cond_indirect_diff(
              out,
              from = from,
              to = to,
              level = level
            )
  # Need these
  # - est: index
  # - ci: cilo, cihi
  # - pvalue
  # - sig
  # - test_label
  ind <- outa$index
  CI.lo <- outa$ci[1]
  CI.hi <- outa$ci[2]
  if (!identical(outa$mc_diff, NA)) {
    boot_diff <- outa$mc_diff
  } else if (!identical(outa$boot_diff, NA)) {
    boot_diff <- outa$boot_diff
  } else {
    boot_diff <- NULL
  }
  pvalue <- est2p(
              boot_diff,
              min_size = -Inf
            )
  Sig <- ifelse((CI.lo > 0) | (CI.hi < 0),
                yes = "Sig",
                "")
  label <- paste0(
              out$Group[to],
              " - ",
              out$Group[from]
            )
  fake_out <- list()
  if (outa$ci_type == "mc") {
    fake_out$mc_indirect <- outa$mc_diff
    fake_out$mc_p <- pvalue
  } else if (outa$ci_type == "boot") {
    fake_out$boot_indirect <- outa$boot_diff
    fake_out$boot_p <- pvalue
  }
  fake_out$level <- level
  outb <- data.frame(
              Group = label,
              ind = ind,
              CI.lo = CI.lo,
              CI.hi = CI.hi,
              Sig = Sig,
              check.names = FALSE
            )
  row.names(outb) <- NULL
  attr(outb, "full_output") <- fake_out
  outb
}

---
title: "Power Analysis for Moderation, Mediation, and Moderated Mediation"
date: "2026-03-10"
output:
  rmarkdown::html_vignette:
    options:
      toc:
        depth: 2
      number_sections: true
vignette: >
  %\VignetteIndexEntry{Power Analysis for Moderation, Mediation, and Moderated Mediation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
csl: apa.csl
---



# Introduction

This article is a brief illustration of how
to use `power4test()`
from the package
[power4mome](https://sfcheung.github.io/power4mome/)
to do power analysis of mediation,
moderation, and moderated mediation
in a model to be fitted by structural
equation modeling using `lavaan`.

# Prerequisite

Basic knowledge about fitting models
by `lavaan` is required. Readers are
also expected to have basic knowledge
of mediation, moderation, and/or
moderated mediation.

# Scope

This is a brief illustration. More
complicated scenarios and other features
of `power4mome` will be described in
other vignettes.

# Package

This introduction only needs the following
package:


``` r
library(power4mome)
```

# Workflow

Two functions are sufficient for estimating
power given a model, population
values, sample size, and the test to be
used. This is the workflow:

1.  Specify the model syntax for the
    population model, in `lavaan`
    style, and set the population values of
    the model parameters.

3.  Call `power4test()` to examine the setup
    and the datasets generated. Repeat
    this and previous steps until the model is
    specified correctly.

4.  Call `power4test()` again, with the
    test to do specified.

5.  Call `rejection_rates()` to
    compute the power.

# Mediation

Let's consider a simple mediation
model. We would like to estimate the
power of testing a mediation effect
by Monte Carlo confidence interval.

## Specify the Population Model

This is the model syntax


``` r
mod <-
"
m ~ x
y ~ m + x
"
```

Note that, even if we are going to
test mediation, moderation, or
moderated mediation effects, we do not
need to add any labels to this model.
This will be taken care of by the
test functions, through the use of
the package
[`manymome`](https://sfcheung.github.io/manymome/)
[@cheung_manymome_2024].


## Specify The Population Values {#es_convention}

There are two approaches to do this:

- Using named vectors or lists.

- Using a multiline string similar to `lavaan`
model syntax.

The second approach is
demonstrated below.

Suppose we want to estimate the power
when:

- The path from `x` to `m` are "large"
  in strength.

- The path from `m` to `y` are "medium"
  in strength.

- The path from `x` to `m` are "small"
  in strength.

By default, `power4mome` uses this convention
for regression path and correlation:^[
Users can specify the values directly
if necessary.]

- Small: .10 (or -.10)

- Medium: .30 (or -.30)

- Large: .50 (or -.50)

For a product term in moderation, this is
the convention:

- Small: .05 (or -.05)

- Medium: .10 (or -.10)

- Large: .15 (or -.15)

All these values are for the standardized
solution (the correlations and
so-called "betas").

The following string denotes the desired
values:


``` r
mod_es <-
"
m ~ x: l
y ~ m: m
y ~ x: s
"
```

Each line starts with a *tag*, which
is the parameter presented in `lavaan`
syntax. The tag ends with a colon, `:`.

After the colon is *population value*, which
can be:

- A string denoting the value. By default:

    - `s`: Small. (`-s` for small and negative.)

    - `m`: Medium. (`-m` for medium and negative.)

    - `l`: Large. (`-l` for large and negative.)

    - `nil`: Zero.

All other regression coefficients and covariances,
if not specified in this string, are set to zero.

## Call `power4test()` to Check the Model


``` r
out <- power4test(nrep = 2,
                  model = mod,
                  pop_es = mod_es,
                  n = 50000,
                  iseed = 1234)
```

These are the arguments used:

- `nrep`: The number of replications.
  In this stage, a small number can
  be used. It is more important to
  have a large sample size than to have
  many replications.

- `model`: The model syntax.

- `pop_es`: The string setting the
  population values.

- `n`: The sample size in each replications.
  In this stage, just for checking the
  model and the data generation, this
  number can be set to a large value unless
  the model is slow to fit when the
  sample size is large.

- `iseed`: If supplied, it is used to
  set the seed for the random number
  generator. It is advised to always set
  this to an arbitrary integer, to make
  the results reproducible.^[The functions
  used are `parallel::clusterSetRNGStream()`
  for parallel processing, and `set.seed()`
  for serial processing.]

The population values can be shown by
printing this object:


``` r
out
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> == Model on Variables/Indicators ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   m ~                        
#>     x                 0.500  
#>   y ~                        
#>     m                 0.300  
#>     x                 0.100  
#> 
#> Variances:
#>                    Population
#>    .m                 0.750  
#>    .y                 0.870  
#>     x                 1.000  
#> 
#> (Computing indirect effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Indirect Effect(s) ==
#> 
#>               ind
#> x -> m -> y 0.150
#> x -> y      0.100
#> 
#>  - The 'ind' column shows the indirect effect(s).
#>  
#> ======================= Data Information =======================
#> 
#> Number of Replications:  2 
#> Sample Sizes:  50000 
#> 
#> Call print with 'data_long = TRUE' for further information.
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         5
#> 
#>   Number of observations                         50000
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 0.000
#>   Degrees of freedom                                 0
```

By default, the population model
will be fitted to each dataset, hence
the section `<fit>`. Perfect fit is
expected if the population model is
a saturated model.

We check the section `Population Values`
to see whether the values are what we
expected.

- In this example, the population values for the regression
paths are what we specified.

If they are different from what we expect,
check the string for `pop_es` to see whether
we set the population values correctly.

If necessary, we can check the data
generation by adding `data_long = TRUE`
when printing the output:


``` r
print(out,
      data_long = TRUE)
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> == Model on Variables/Indicators ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   m ~                        
#>     x                 0.500  
#>   y ~                        
#>     m                 0.300  
#>     x                 0.100  
#> 
#> Variances:
#>                    Population
#>    .m                 0.750  
#>    .y                 0.870  
#>     x                 1.000  
#> 
#> (Computing indirect effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Indirect Effect(s) ==
#> 
#>               ind
#> x -> m -> y 0.150
#> x -> y      0.100
#> 
#>  - The 'ind' column shows the indirect effect(s).
#>  
#> ======================= Data Information =======================
#> 
#> Number of Replications:  2 
#> Sample Sizes:  50000 
#> 
#> ==== Descriptive Statistics ====
#> 
#>   vars     n mean sd skew kurtosis se
#> m    1 1e+05 0.00  1 0.01     0.03  0
#> y    2 1e+05 0.01  1 0.01     0.00  0
#> x    3 1e+05 0.00  1 0.01     0.01  0
#> 
#> ===== Parameter Estimates Based on All 2 Samples Combined =====
#> 
#> Total Sample Size: 100000 
#> 
#> ==== Standardized Estimates ====
#> 
#> Variances and error variances omitted.
#> 
#> Regressions:
#>                     est.std
#>   m ~                      
#>     x                 0.500
#>   y ~                      
#>     m                 0.295
#>     x                 0.102
#> 
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         5
#> 
#>   Number of observations                         50000
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 0.000
#>   Degrees of freedom                                 0
```

The section `Descriptive Statistics`, generated
by `psych::describe()`, shows basic
descriptive statistics for the observed
variables. As expected, they have means
close to zero and standard deviations
close to one, because the datasets were
generated using the standardized model.

The section `Parameter Estimates Based on`
shows the parameter estimates when the
*population model* is fitted to *all the
datasets combined*. When the total
sample size is large, these estimates
should be close to the population values.

The results show that we have specified
the population model correctly. We
can proceed to specify the test and
estimate the power.

## Call `power4test()` to Do the Target Test {#med_power}

We can now do the simulation to estimate
power. A large number of datasets (e.g., 500)
of the target sample size are to be generated,
and then the target test will be conducted
in each of these datasets.

Suppose we would like to estimate the
power of using Monte Carlo confidence intervals
to test the indirect effect from `x` to
`y` through `m`, when sample size is 50.
This is the call:


``` r
out <- power4test(nrep = 400,
                  model = mod,
                  pop_es = mod_es,
                  n = 50,
                  R = 2000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "x",
                                   m = "m",
                                   y = "y",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
```

These are the new arguments used:

- `R`: The number of replications used
  to generate the Monte Carlo simulated
  estimates, 2000 in this example. In
  real studies, this number should be
  10000 or even 20000 for Monte Carlo
  confidence intervals. However, 2000
  is sufficient because the goal is to
  estimate power by generating many
  intervals, rather
  than to have one single stable interval.

- `ci_type`: The method used to
  generate estimates. Support both Monte
  Carlo (`"mc"`) and nonparametric
  bootstrapping (`"boot"`).^[They
  are implemented by `manymome::do_mc()`
  and `manymome::do_boot()`, respectively.]
  Although bootstrapping is usually used
  to test an indirect effect, it is very
  slow to do `R` bootstrapping in `nrep`
  datasets (the model will be fitted
  `R * nrep` times). Therefore, it is
  preferable to use Monte Carlo confidence
  intervals to do
  the initial estimation.

- `test_fun`: The function to be used
  to do the test for each replication.
  Any function following a specific
  requirement can be used, and
  `power4mome` comes with several
  built-in functions for some common tests.
  The function `test_indirect_effect()`
  is used to test an indirect effect
  in the model.

- `test_args`: A named list of arguments
  to be supplied to `test_fun`.
  For `test_indirect_effect()`, it is
  a named list specifying the predictor
  (`x`), the mediator(s) (`m`), and
  the outcome (`y`). A path with any
  number of mediators can be supported.
  Please refer to the help page of
  `test_indirect_effect()`.^[The test
  is implemented by `manymome::indirect()`.]

- `parallel`: If the test to be conducted
  is slow, which is the case for
  tests done by Monte Carlo or nonparametric
  bootstrapping confidence intervals,
  it is advised to enable parallel
  processing by setting `parallel` to
  `TRUE`.^[The number of cores is determined
  automatically but can be set directly
  by the `ncores` argument.]

For `nrep = 400`, the 95% confidence
limits
for a power of .80 are about .04 below
and above .80. This should be precise
enough for determining whether a
sample size has sufficient power.

This is the default printout:


``` r
out
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> == Model on Variables/Indicators ==
#> 
#> m ~ x
#> y ~ m + x
#> 
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   m ~                        
#>     x                 0.500  
#>   y ~                        
#>     m                 0.300  
#>     x                 0.100  
#> 
#> Variances:
#>                    Population
#>    .m                 0.750  
#>    .y                 0.870  
#>     x                 1.000  
#> 
#> (Computing indirect effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Indirect Effect(s) ==
#> 
#>               ind
#> x -> m -> y 0.150
#> x -> y      0.100
#> 
#>  - The 'ind' column shows the indirect effect(s).
#>  
#> ======================= Data Information =======================
#> 
#> Number of Replications:  400 
#> Sample Sizes:  50 
#> 
#> Call print with 'data_long = TRUE' for further information.
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> - mc_out
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         5
#> 
#>   Number of observations                            50
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 0.000
#>   Degrees of freedom                                 0
#> 
#> =========== <mc_out> ===========
#> 
#> 
#> == A 'mc_out' class object ==
#> 
#> Number of Monte Carlo replications: 2000 
#> 
#> 
#> =============== <test_indirect: x->m->y> ===============
#> 
#> Mean(s) across replication:
#>    est   cilo  cihi   sig pvalue
#>  0.152 -0.009 0.347 0.468  0.166
#> 
#> - The value 'sig' is the rejection rate.
#> - If the null hypothesis is false, this is the power.
#> - Number of valid replications for rejection rate: 400 
#> - Proportion of valid replications for rejection rate: 1.000
```

The summary of the test:


```
#> =============== <test_indirect: x->m->y> ===============
#> 
#> Mean(s) across replication:
#>    est   cilo  cihi   sig pvalue
#>  0.152 -0.009 0.347 0.468  0.166
#> 
#> - The value 'sig' is the rejection rate.
#> - If the null hypothesis is false, this is the power.
#> - Number of valid replications for rejection rate: 400 
#> - Proportion of valid replications for rejection rate: 1.000
```

The mean of the estimates across all the
replications is 0.152, close to
the population value.

## Compute the Power

The power estimate is simply the proportion
of significant results, the *rejection*
*rate*, because the
null hypothesis is false. In addition to
using `test_long = TRUE` in `print()`,
the rejection
rate can also be retrieved by
`rejection_rates()`.


``` r
out_power <- rejection_rates(out)
out_power
#> [test]: test_indirect: x->m->y 
#> [test_label]: Test 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.152 1.000  0.468  0.419  0.516
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



In the example above, the estimated
power of the test of the indirect
effect, conducted by Monte Carlo
confidence intervals, is
0.468, under the column
`reject`.

`p.v` is
the proportion of valid results
across replications. `1.000` means that
the test conducted normally in all
replications.

By default, the 95% confidence interval
of the rejection rate (power)
based on normal approximation is
also printed, under the column
`r.cilo` and `r.cihi`. In this example,
the 95% confidence interval is
[0.419; 0.516].

# Moderation

Let's consider a moderation model,
with some control variables.

## Specify the Population Model and Values {#pop_es_xw}


``` r
mod2 <-
"
y ~ x + w + x:w + control
"
```

This model has only moderation, with
the predictor `x` and the moderator
`w`. The product term is included in
the `lavaan` style, `x:w`.

It is
unrealistic to specific the population
values for all control variables. Therefore,
we can just add a *proxy*, `control`
to represent the *set of control variables*
that may be included.

This is the syntax for the population
values:


``` r
mod2_es <-
"
.beta.: s
x ~~ control: s
y ~ control: s
y ~ x:w: l
"
```

This example introduces one useful
tag, `.beta.` For a model with many
paths, it is inconvenient to specify
all of them manually. The tag `.beta.`
specifies the *default value* for
*all regression paths not specified explicitly*,
which is small (.10)
in this example. If a path is explicitly
included (such as `y ~ control` and
`y ~ x:w`), the manually specified value
will be used instead of `.beta.`.

This example also illustrates that we
can set the population values for
*correlations* (covariances in the
standardized solution). Control variables
are included usually because they may
correlate with the predictors. Therefore,
in this example, it is hypothesized
that there is a small correlation between
`x` and the proxy control variable (`x ~~ control: s`).

Last, recall from [this section](#es_convention)
that the convention for product term
values is different: `l` denotes
.15 for product terms.

## Call `power4test()` to Check the Model

We check the model first:


``` r
out2 <- power4test(nrep = 2,
                   model = mod2,
                   pop_es = mod2_es,
                   n = 50000,
                   iseed = 1234)
```


``` r
print(out2,
      data_long = TRUE)
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> y ~ x + w + x:w + control
#> 
#> == Model on Variables/Indicators ==
#> 
#> y ~ x + w + x:w + control
#> 
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   y ~                        
#>     x                 0.100  
#>     w                 0.100  
#>     x:w               0.150  
#>     control           0.100  
#> 
#> Covariances:
#>                    Population
#>   x ~~                       
#>     w                 0.000  
#>     x:w               0.000  
#>     control           0.100  
#>   w ~~                       
#>     x:w               0.000  
#>     control           0.000  
#>   x:w ~~                     
#>     control           0.000  
#> 
#> Variances:
#>                    Population
#>    .y                 0.945  
#>     x                 1.000  
#>     w                 1.000  
#>     x:w               1.000  
#>     control           1.000  
#> 
#> (Computing indirect effects for 1 paths ...)
#> 
#> (Computing conditional effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Effect(s) ==
#> 
#>                ind
#> control -> y 0.100
#> 
#>  - The 'ind' column shows the effect(s).
#>  
#> == Conditional effects ==
#> 
#>  Path: x -> y
#>  Conditional on moderator(s): w
#>  Moderator(s) represented by: w
#> 
#>       [w] (w)    ind
#> 1 M+1.0SD   1  0.250
#> 2 Mean      0  0.100
#> 3 M-1.0SD  -1 -0.050
#> 
#>  - The 'ind' column shows the conditional effects.
#>  
#> 
#> == Conditional effects ==
#> 
#>  Path: w -> y
#>  Conditional on moderator(s): x
#>  Moderator(s) represented by: x
#> 
#>       [x] (x)    ind
#> 1 M+1.0SD   1  0.250
#> 2 Mean      0  0.100
#> 3 M-1.0SD  -1 -0.050
#> 
#>  - The 'ind' column shows the conditional effects.
#>  
#> 
#> ======================= Data Information =======================
#> 
#> Number of Replications:  2 
#> Sample Sizes:  50000 
#> 
#> ==== Descriptive Statistics ====
#> 
#>         vars     n mean sd  skew kurtosis se
#> y          1 1e+05 0.00  1  0.02     0.01  0
#> x          2 1e+05 0.00  1  0.01     0.01  0
#> w          3 1e+05 0.00  1  0.00    -0.02  0
#> x:w        4 1e+05 0.00  1 -0.03     6.01  0
#> control    5 1e+05 0.01  1  0.00     0.00  0
#> 
#> ===== Parameter Estimates Based on All 2 Samples Combined =====
#> 
#> Total Sample Size: 100000 
#> 
#> ==== Standardized Estimates ====
#> 
#> Variances and error variances omitted.
#> 
#> Regressions:
#>                     est.std
#>   y ~                      
#>     x                 0.097
#>     w                 0.104
#>     x:w               0.152
#>     control           0.099
#> 
#> Covariances:
#>                     est.std
#>   x ~~                     
#>     w                -0.003
#>     x:w               0.002
#>     control           0.101
#>   w ~~                     
#>     x:w              -0.004
#>     control           0.001
#>   x:w ~~                   
#>     control           0.004
#> 
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         5
#> 
#>   Number of observations                         50000
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 0.000
#>   Degrees of freedom                                 0
```



The population values for the regression
paths are what we specified, and the
estimates based on 5 &times; 10<sup>4</sup>
by 2 or
100000
support that the dataset were generated
correctly.

NOTE: If a product term is involved,
and the component terms (`x` and `w` in this
example) are correlated, the population
standard deviation of this product term
may not be equal to one [@bohrnstedt_exact_1969].
Therefore, the
model can be specified correctly even if
the standard deviations of product terms
in the section `Descriptive Statistics`
are not close to one.

## Call `power4test()` to Test The Moderation Effect

We can now do the simulation to estimate
power. In this simple model, the test
is just a test of the product term,
`x:w`. This model can be fitted by
linear regression using `lm()`. Let's
estimate the power when the sample size
is 50 and the model is fitted by `lm()`:


``` r
out2 <- power4test(nrep = 400,
                   model = mod2,
                   pop_es = mod2_es,
                   n = 100,
                   fit_model_args = list(fit_function = "lm"),
                   test_fun = test_moderation,
                   iseed = 1234,
                   parallel = TRUE)
```

These are the new arguments used:

- `fit_model_args`: This named list
  stores additional arguments for `fit_model()`.
  By default, `lavaan::sem()` is used.
  To fit the model by linear regression
  using `lm()`, add `fit_function = "lm"`
  to the list.^[See the help page of
  `fit_model()` on other available
  arguments.]

- `test_fun`: It is set to
  `test_moderation`, provided by `power4mome`.
  This function automatically identifies
  all product terms in a model and
  test them. The test used depends
  on method used to fit the model.
  If `lm()` is used, then the usual
  *t* test is used.^[The test name
  has `CIs` in it but this is
  equivalent to using the *t* test
  when the model is fitted by `lm()`.]

## Compute the Power

We can ues `rejection_rates()`
again to estimate the power:


``` r
out2_power <- rejection_rates(out2)
out2_power
#> [test]: test_moderation: CIs  
#> [test_label]: y~x:w 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.158 1.000  0.347  0.302  0.395
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



The estimated
power of the test of the product term,
`x:w`, is
0.347, with 95% confidence
interval [0.302; 0.395].

# Moderated mediation

Let's consider a moderated mediation
model.

## Specify the Population Model and Values


``` r
mod3 <-
"
m ~ x + w + x:w
y ~ m + x
"
```

This model is a mediation model with
the *a*-path, `m ~ x`, moderated by
`w`. As explained before, there is
no need to use any label nor define
and parameters. This will be handled
by the test function to be used.

This is the syntax for the population
values:


``` r
mod3_es <-
"
.beta.: s
m ~ x: m
y ~ m: m
m ~ x:w: s
"
```

Please refer to [the previous section](#pop_es_xw)
on setting up this syntax.

## Call `power4test()` to Check the Model

We check the model first:


``` r
out3 <- power4test(nrep = 2,
                   model = mod3,
                   pop_es = mod3_es,
                   n = 50000,
                   iseed = 1234)
```


``` r
print(out3,
      data_long = TRUE)
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> m ~ x + w + x:w
#> y ~ m + x
#> 
#> == Model on Variables/Indicators ==
#> 
#> m ~ x + w + x:w
#> y ~ m + x
#> 
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   m ~                        
#>     x                 0.300  
#>     w                 0.100  
#>     x:w               0.050  
#>   y ~                        
#>     m                 0.300  
#>     x                 0.100  
#> 
#> Covariances:
#>                    Population
#>   x ~~                       
#>     w                 0.000  
#>     x:w               0.000  
#>   w ~~                       
#>     x:w               0.000  
#> 
#> Variances:
#>                    Population
#>    .m                 0.898  
#>    .y                 0.881  
#>     x                 1.000  
#>     w                 1.000  
#>     x:w               1.000  
#> 
#> (Computing indirect effects for 1 paths ...)
#> 
#> (Computing conditional effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Effect(s) ==
#> 
#>          ind
#> x -> y 0.100
#> 
#>  - The 'ind' column shows the effect(s).
#>  
#> == Conditional indirect effects ==
#> 
#>  Path: x -> m -> y
#>  Conditional on moderator(s): w
#>  Moderator(s) represented by: w
#> 
#>       [w] (w)   ind   m~x   y~m
#> 1 M+1.0SD   1 0.105 0.350 0.300
#> 2 Mean      0 0.090 0.300 0.300
#> 3 M-1.0SD  -1 0.075 0.250 0.300
#> 
#>  - The 'ind' column shows the conditional indirect effects.
#>  - 'm~x','y~m' is/are the path coefficient(s) along the path conditional
#>    on the moderator(s).
#> 
#> 
#> == Conditional indirect effects ==
#> 
#>  Path: w -> m -> y
#>  Conditional on moderator(s): x
#>  Moderator(s) represented by: x
#> 
#>       [x] (x)   ind   m~w   y~m
#> 1 M+1.0SD   1 0.045 0.150 0.300
#> 2 Mean      0 0.030 0.100 0.300
#> 3 M-1.0SD  -1 0.015 0.050 0.300
#> 
#>  - The 'ind' column shows the conditional indirect effects.
#>  - 'm~w','y~m' is/are the path coefficient(s) along the path conditional
#>    on the moderator(s).
#> 
#> 
#> ======================= Data Information =======================
#> 
#> Number of Replications:  2 
#> Sample Sizes:  50000 
#> 
#> ==== Descriptive Statistics ====
#> 
#>     vars     n mean sd skew kurtosis se
#> m      1 1e+05    0  1 0.03     0.03  0
#> y      2 1e+05    0  1 0.01    -0.01  0
#> x      3 1e+05    0  1 0.00    -0.02  0
#> w      4 1e+05    0  1 0.00     0.01  0
#> x:w    5 1e+05    0  1 0.04     5.92  0
#> 
#> ===== Parameter Estimates Based on All 2 Samples Combined =====
#> 
#> Total Sample Size: 100000 
#> 
#> ==== Standardized Estimates ====
#> 
#> Variances and error variances omitted.
#> 
#> Regressions:
#>                     est.std
#>   m ~                      
#>     x                 0.303
#>     w                 0.099
#>     x:w               0.052
#>   y ~                      
#>     m                 0.299
#>     x                 0.098
#> 
#> Covariances:
#>                     est.std
#>   x ~~                     
#>     w                 0.003
#>     x:w              -0.001
#>   w ~~                     
#>     x:w               0.008
#> 
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         7
#> 
#>   Number of observations                         50000
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 0.007
#>   Degrees of freedom                                 2
#>   P-value (Chi-square)                           0.997
```



The population values and the
estimates based on 5 &times; 10<sup>4</sup>
by 2 or
100000
are what we expect.

## Call `power4test()` to Test The Moderated Mediation Effect

To estimate the power of a
moderated mediation effect, we can
test the *index of moderated mediation*
[@hayes_index_2015]. In this example,
it is the product of the coefficient
`m ~ x:w` and the coefficient `y ~ m`.
This can be done by the test function
`test_index_of_mome()`, provided
by `power4mome`. Again, Monte
Carlo confidence interval is used.

Let's estimate
the power when sample size is 100.


``` r
out3 <- power4test(nrep = 400,
                   model = mod3,
                   pop_es = mod3_es,
                   n = 100,
                   R = 2000,
                   ci_type = "mc",
                   test_fun = test_index_of_mome,
                   test_args = list(x = "x",
                                    m = "m",
                                    y = "y",
                                    w = "w",
                                    mc_ci = TRUE),
                   iseed = 1234,
                   parallel = TRUE)
```

The call is similar to the one used in
[testing mediation](#med_power).

This is the new argument used:

- `test_fun`: It is set to
  `test_index_of_mome()` in this example.
  This function is similar to
  `test_indirect_effect()`, with one
  more argument, `w`, for the moderator.
  Although this example has only one mediator,
  it support any number of mediators
  along a path.^[The test is implemented
  by `manymome::test_index_of_mome()`.]

## Compute the Power

We can ues `rejection_rates()`
again to estimate the power:


``` r
out3_power <- rejection_rates(out3)
out3_power
#> [test]: test_index_of_mome: x->m->y, moderated by w 
#> [test_label]: Test 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.016 1.000  0.055  0.037  0.082
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



The estimated
power of the test of moderated mediation
effect, conducted by a test of the
index of moderated mediation, is
0.055,
95% confidence interval [0.037; 0.082].

Unlike the previous example on moderation
tested by regression, estimating the power
of Monte Carlo confidence intervals is
substantially slower. However,
this is necessary because Monte Carlo
or nonparametric bootstrapping
confidence interval is the test usually
used in moderated mediation (and mediation).

# Repeating a Simulation With A Different Sample Size {#new_n}

The function `power4test()` also supports
*redoing* an analysis using a new value
for the sample size (or population effect sizes
set to `pop_es`). Simply

- set the output of `power4test` as the first
argument, and

- set the *new value* for `n`.

For example, we can repeat the
simulation for the test of moderation [above](#pop_es_xw),
but for a sample size of 200. We
simply call `power4test()` again,
set the previous output (`out2` in
the example for moderation) as the
first argument, and set `n` to a new
value (200 in this example):


``` r
out2_new_n <- power4test(out2,
                         n = 200)
out2_new_n
```



This is the estimated power when
the sample size is 200.


``` r
out2_new_n_reject <- rejection_rates(out2_new_n)
out2_new_n_reject
#> [test]: test_moderation: CIs  
#> [test_label]: y~x:w 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.148 1.000  0.527  0.479  0.576
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



The estimated power is 0.527,
95% confidence interval
[0.479; 0.576],
when the sample size is
200.

This technique can be
repeated to find the required sample
size for a target power, and can be
used for all the other scenarios covered
above, such as mediation and
moderated mediation.

# Find the Sample Size With The Desired Power

There are several more efficient ways to
find the sample size with the desired power.

## Using `n_region_from_power()`

The function `n_region_from_power()`
can be used to find the *region* of
sample sizes likely to have the desired
power. If the default settings are
to be used, then it can be called
directly on the output of `power4test()`:


``` r
out2_region <- n_region_from_power(out2,
                                   seed = 2345)
```

This is the recommended way for sample
size planning, when there is no predetermined
range of sample sizes.

See
[the templates](https://sfcheung.github.io/power4mome/articles/)
for examples on using `n_region_from_power()`
for common models.

## Using `power4test_by_n()`

The function
`power4test_by_n()` can be used To
estimate the power for a sequence of
sample sizes. For example, we
can estimate the power in the moderation
model above for these sample sizes:
250, 300, 350, 400.


``` r
out2_several_ns <- power4test_by_n(out2,
                                   n = c(250, 300, 350, 400),
                                   by_seed = 4567)
```

The first argument is the output
of `power4test()` for an arbitrary sample size.

The argument `n` is a numeric vector of
sample sizes to examine.

The argument `by_seed`, if set to an integer,
try to make the results reproducible.

The call will take some times to run because
it is equivalent to calling `power4test()`
once for each sample size.

The rejection rates for each sample size
can be retrieved by `rejection_rates()`
too:


``` r
rejection_rates(out2_several_ns)
#> [test]: test_moderation: CIs  
#> [test_label]: y~x:w 
#>     n   est   p.v reject r.cilo r.cihi
#> 1 250 0.149 1.000  0.660  0.612  0.705
#> 2 300 0.150 1.000  0.733  0.687  0.774
#> 3 350 0.151 1.000  0.810  0.769  0.845
#> 4 400 0.150 1.000  0.850  0.812  0.882
#> Notes:
#> - n: The sample size in a trial.
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```

The results show that, to have a power of
about .800 to detect the moderation effect,
a sample size of about 350 is needed.

This approach is used when the range of
sample sizes has already been decided and
the levels of power are needed to determine
the final sample size.

Please refer to the
help page of `power4test_by_n()`
for other examples.

## Using `x_from_power()`

The function `x_from_power()`
can be used to systematically search
within an interval the sample size with
the target power. This takes longer to
run but, instead of manually trying
different sample size, this function
do the search automatically.

This approach can be used when the goal
is to find the probable minimum or
maximum sample size with the desired
level of power. The first approach,
using `n_region_from_power()`, simply
uses this approach twice to find the region
of sample sizes.

See
[this article](https://sfcheung.github.io/power4mome/articles/x_from_power_for_n.html)
for an illustration of how to use
`x_from_power()`.

# Other Advanced Features

This brief illustration only covers the
basic features of `power4mome`. These
are other advanced features to be
covered in other articles:

- There is no inherent restriction on the
  form of the model. Typical models that
  can be specified in `lavaan` model
  syntax can be the population model,
  although there may be special models
  in which `power4test` does not yet support.

- The population model can be a model
  with latent factors and indicators.
  Nevertheless, users can specify only
  the relation among the factors.
  There is no need to include indicators
  in the model syntax, and also no need
  to manually specify the factor loadings.
  The number of indicators for each
  factor and the factor loadings are
  set by the argument `number_of_indicators`
  and `reliability` (see the help page
  of `sim_data()` on how to set them).
  The model syntax used to fit to the
  data will automatically include the
  indicators. An introduction can be
  found in `vignette("power4test_latent_mediation")`.
  Examples can be found in
  [templates](https://sfcheung.github.io/power4mome/articles/)
  for models with latent variables.

- Though not illustrated above,
  estimating the power of tests
  conducted by nonparametric
  bootstrapping is supported, although
  it will take longer to run.

- Although this package focuses on
  moderation, mediation, and
  moderated mediation, in principle,
  the power of any test can be estimated,
  as long as a test function for
  `test_fun` is available. Some other
  functions are provided with `power4mome`
  (e.g., `test_parameters()` for testing
  all free model parameters). See the
  help page of `do_test()` on how to
  write a function to do a test not
  available in `power4mome`.

- When estimating power, usually the
  population model is fitted to the
  data. However, it is possible to
  fit any other model to the generated
  data. This can be done by using
  the argument `fit_model_args` to
  set the argument `model` of
  `fit_model()`.

- Preliminary support for
  multigroup model is available.
  See the help pages of `ptable_pop()`
  and `pop_es_yaml()` on how to
  specify the population value syntax.
  Functions will be added for tests
  relevant to multigroup models (e.g.,
  testing the between-group difference
  in an indirect effect).

- Although we illustrated only rerunning
  an analysis with a new sample size (`n`),
  it is also possible to rerun an analysis
  using a new population value for a
  parameter. This can be done by using
  the previous output of `power4test()`
  as the first argument, and setting
  only `pop_es` to a named vector:

```r
out2_new_xw <- power4test(out2,
                          pop_es = c("y ~ x:w" = ".30"))
```

- Basic support for generating
  nonnormal variables, including
  dichotomous variables is available. See the
  argument `x_fun` of `power4test()`
  for details.

# Limitations

- Monte Carlo confidence interval is
  not supported for models fitted by
  `lm()` (regression). To estimate
  power of testing mediation or
  moderated mediation effects
  in models fitted by `lm()`,
  `ci_type = "boot"` is needed.

# References

---
title: "Power Analysis for Latent Variable Mediation"
date: "2026-03-10"
output:
  rmarkdown::html_vignette:
    options:
      toc:
        depth: 2
      number_sections: true
vignette: >
  %\VignetteIndexEntry{Power Analysis for Latent Variable Mediation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
csl: apa.csl
---



# Introduction

This article is a brief illustration of how
to use `power4test()`
from the package
[power4mome](https://sfcheung.github.io/power4mome/)
to do power analysis for the mediation
effect among latent factors in a
model to be fitted by structural
equation model modeling using `lavaan`.

# Prerequisite

Basic knowledge about fitting models
by `lavaan` is required. Readers are
also expected to have basic knowledge
of mediation and structural equation
modeling.

# Scope

To make this vignette self-contained,
some sections from
`vignette("power4mome")` are repeated
here.

To do power analysis for a mediation
effect in a path model with no latent
factors, please refer to
`vignette("power4mome")`.

# Package

This introduction only needs the following
package:


``` r
library(power4mome)
```

# Workflow

Two functions are sufficient for estimating
power given a model, population
values, sample size, and the test to be
used. This is the workflow:

1.  Specify the model syntax for the
    population model, in `lavaan`
    style, and set the population values of
    the model parameters.

3.  Call `power4test()` to examine the setup
    and the datasets generated. Repeat
    this and previous steps until the model is
    specified correctly.

4.  Call `power4test()` again, with the
    test to do specified.

5.  Call `rejection_rates()` to
    compute the power.

# Mediation

Let's consider a simple mediation
model with three factors. We would like
to estimate the
power of testing the mediation effect
by Monte Carlo confidence interval.

## Specify the Population Model

For a model with latent factor,
we only need to specify the model syntax
for the *factors*. No need to include
the measurement part and the indicators.

This is the model syntax:


``` r
mod <-
"
fm ~ fx
fy ~ fm + fx
"
```

The latent variables are `fx`, `fm`,
and `fy`. There is an indirect path
from `fx` to `fy`, through `fm`.

Note that, even if we are going to
test a mediation effect, we do not
need to add any labels to this model.
This will be taken care of by the
test functions, through the use of
the package
[`manymome`](https://sfcheung.github.io/manymome/)
[@cheung_manymome_2024].

## Specify The Population Values {#es_convention}

There are two approaches to do this:

- Using named vectors or lists.

- Using a multiline string similar to `lavaan`
model syntax.

The second approach is
demonstrated below.

Suppose we want to estimate the power
when:

- The path from `fx` to `fm` are "large"
  in strength.

- The path from `fm` to `fy` are "large"
  in strength.

- The path from `fx` to `fm` are "small"
  in strength.

By default, `power4mome` use this convention
for regression path and correlation:^[
Users can specify the values directly
if necessary.]

- Small: .10 (or -.10)

- Medium: .30 (or -.30)

- Large: .50 (or -.50)

All these values are for the standardized
solution (the correlations and
so-called "betas").

The following string denotes the desired
values:


``` r
mod_es <-
"
fm ~ fx: l
fy ~ fm: l
fy ~ fx: s
"
```

Each line starts with a *tag*, which
is the parameter presented in `lavaan`
syntax. The tag ends with a colon, `:`.

After the colon is *population value*, which
can be:

- A string denoting the value. By default:

    - `s`: Small. (`-s` for small and negative.)

    - `m`: Medium. (`-m` for medium and negative.)

    - `l`: Large. (`-l` for large and negative.)

    - `nil`: Zero.

All other regression coefficients and covariances,
if not specified in this string, are set to zero.

## Specify the Measurement Part {#measurement}

Power analysis is usually conducted before
data collection. We rarely know in advance
the factor loadings of all items. For the
purpose of power analysis, which is not
intended to be conducted with the knowledge
of all factor loadings, we believe that,
instead of specifying all the loadings,
it is sufficient to specify two values for
each factor:

- The number of indicators.

- The population reliability.

This is the approach used in `power4mome`.

For each factor, the population standardized
factor loadings for each indicator will
be derived automatically from the hypothesized
(or expected) population reliability and the number
of indicators, assuming that all indicators
have equal loadings.

Although the equal-loading assumption is unrealistic,
in *a priori* power analysis, it is difficult,
if not impossible, to specify the pattern
of factor loadings. This level of details is also
not necessary because the power estimated
is merely used to guide the planning of
data collection, instead of estimating the
"true" power after data is collected.

Two arguments will be used to set
the number of indicators and the reliability.

- `number_of_indicators`: This should be
  a *named* vector of the number of
  indicators for each factor. The names
  are the names of the factors as appeared
  in the model syntax, and the values
  are the number of indicators.

- `reliability`: This should be
  a *named* vector of the reliability
  for each factor. The names
  are the names of the factors as appeared
  in the model syntax, and the values
  are population reliability.

For example, suppose we will use the
following vectors:


``` r
k <- c(fm = 3,
       fx = 4,
       fy = 5)
```


``` r
mod_rel <- c(fy = .70,
             fm = .60,
             fx = .50)
```

The numbers of indicators for
`fx`, `fm`, and `fy` are 4, 3, and 5,
respectively.

The population reliability coefficients
for `fx`, `fm`, and `fy` are .50,
.60, and .70, respectively. In real
research, reliability as low as .50
can be problematic. We chose these values
merely for illustration

The orders are intentionally arbitrary,
to demonstrate the order does not matter.
The names will be used to interpret the
numbers correctly.

## Call `power4test()` to Check the Model

We are all set and can call `power4test()`
to check the model:


``` r
out <- power4test(nrep = 2,
                  model = mod,
                  pop_es = mod_es,
                  number_of_indicators = k,
                  reliability = mod_rel,
                  n = 50000,
                  iseed = 1234)
```

These are the arguments used:

- `nrep`: The number of replications.
  In this stage, a small number can
  be used. It is more important to
  have a large sample size than to have
  many replications.

- `model`: The model syntax.

- `pop_es`: The string setting the
  population values.

- `number_of_indicators`: A named vector
  of the number of indicators for each
  factor, described in the previous section.

- `reliability`: A named vector
  of the population reliability for each
  factor, described in the previous section.

- `n`: The sample size in each replications.
  In this stage, just for checking the
  model and the data generation, this
  number can be set to a large one unless
  the model is slow to fit when the
  sample size is large.

- `iseed`: If supplied, it is used to
  set the seed for the random number
  generator. It is advised to always set
  this to an arbitrary integer, to make
  the results reproducible.^[The functions
  used are `parallel::clusterSetRNGStream()`
  for parallel processing, and `set.seed()`
  for serial processing.]

The population values can be shown by
printing this object:


``` r
out
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> fm ~ fx
#> fy ~ fm + fx
#> 
#> == Model on Variables/Indicators ==
#> 
#> fm ~ fx
#> fy ~ fm + fx
#> 
#> fm =~ fm1 + fm2 + fm3
#> fy =~ fy1 + fy2 + fy3 + fy4 + fy5
#> fx =~ fx1 + fx2 + fx3 + fx4
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   fm ~                       
#>     fx                0.500  
#>   fy ~                       
#>     fm                0.500  
#>     fx                0.100  
#> 
#> Variances:
#>                    Population
#>    .fm                0.750  
#>    .fy                0.690  
#>     fx                1.000  
#> 
#> (Computing indirect effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Indirect Effect(s) ==
#> 
#>                  ind
#> fx -> fm -> fy 0.250
#> fx -> fy       0.100
#> 
#>  - The 'ind' column shows the indirect effect(s).
#>  
#> ==== Population Reliability ====
#> 
#>   fy  fm  fx
#>  0.7 0.6 0.5
#> 
#> == Population Standardized Loadings ==
#> 
#>    Standardized Loadings        
#> fm 0.577,0.577,0.577            
#> fx 0.447,0.447,0.447,0.447      
#> fy 0.564,0.564,0.564,0.564,0.564
#> 
#> ======================= Data Information =======================
#> 
#> Number of Replications:  2 
#> Sample Sizes:  50000 
#> 
#> Call print with 'data_long = TRUE' for further information.
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 44 iterations
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                        27
#> 
#>   Number of observations                         50000
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                38.042
#>   Degrees of freedom                                51
#>   P-value (Chi-square)                           0.910
```



By the default, the population model
will be fitted to each dataset, hence
the section `<fit>`. The fit is not
"perfect" because the model is not a
saturated model. However, the *p*-value
is high and not significant (because, with the population model
fitted, the chance of significant is close to .05).

Although we specify only the structure
for the latent factors, we can see the
automatically generated measurement part
syntax in the section `Model on Variables/Indicators`:


```
#> fm =~ fm1 + fm2 + fm3
#> fy =~ fy1 + fy2 + fy3 + fy4 + fy5
#> fx =~ fx1 + fx2 + fx3 + fx4
```

It confirmed that we specified the
measurement part correctly.

We then check the section `Population Values`
to see whether the values are what we
expected.

- In this example, the population values for the regression
paths are what we specified.

If they are different from what we expect,
check the string for `pop_es` to see whether
we set the population values correctly.

- The section `Variances:` shows the
  variances or *error* variances of
  the latent factors. Because `fm`
  and `fy` are endogenous factors,
  the values presented, next to `.fm`
  and `.fy`, are error variances, and that's
  why they are not 1, unlike `fx`.

The section `Population Reliability`
shows the population reliability coefficients:


```
#> ==== Population Reliability ====
#> 
#>   fy  fm  fx
#>  0.7 0.6 0.5
```

They are the values we set to `reliability`.

The section `Population Standardized Loadings`
shows the standardized factor loadings
for each factor. Only one value for each
factor because the loadings are assumed to
be the same for all items:


```
#> == Population Standardized Loadings ==
#> 
#>    Standardized Loadings        
#> fm 0.577,0.577,0.577            
#> fx 0.447,0.447,0.447,0.447      
#> fy 0.564,0.564,0.564,0.564,0.564
```

If necessary, we can check the data
generation by adding `data_long = TRUE`
when printing the output:

```r
print(out,
      data_long = TRUE)
```



Two new sections will be printed.
This is the first one:


```
#> ==== Descriptive Statistics ====
#> 
#>     vars     n mean   sd  skew kurtosis se
#> fm1    1 1e+05 0.00 1.00  0.01     0.00  0
#> fm2    2 1e+05 0.01 1.00  0.01    -0.01  0
#> fm3    3 1e+05 0.00 1.01  0.00    -0.02  0
#> fx1    4 1e+05 0.00 1.00  0.00     0.00  0
#> fx2    5 1e+05 0.00 1.00 -0.01     0.00  0
#> fx3    6 1e+05 0.00 1.00 -0.01    -0.02  0
#> fx4    7 1e+05 0.00 1.00  0.00     0.00  0
#> fy1    8 1e+05 0.00 1.00  0.01     0.01  0
#> fy2    9 1e+05 0.00 1.00  0.01     0.00  0
#> fy3   10 1e+05 0.00 1.00  0.00     0.01  0
#> fy4   11 1e+05 0.01 1.00  0.01     0.00  0
#> fy5   12 1e+05 0.00 1.00  0.02    -0.01  0
```

The section `Descriptive Statistics`, generated
by `psych::describe()`, shows basic
descriptive statistics for indicators.
As expected, they have means
close to zero and standard deviations
close to one, because the datasets were
generated using the standardized model.

This is the second one:


```
#> fy5   12 1e+05 0.00 1.00  0.02    -0.01  0
#> 
#> ===== Parameter Estimates Based on All 2 Samples Combined =====
#> 
#> Total Sample Size: 100000 
#> 
#> ==== Standardized Estimates ====
#> 
#> Variances and error variances omitted.
#> 
#> Latent Variables:
#>                     est.std
#>   fm =~                    
#>     fm1               0.578
#>     fm2               0.579
#>     fm3               0.578
#>   fy =~                    
#>     fy1               0.559
#>     fy2               0.560
#>     fy3               0.563
#>     fy4               0.565
#>     fy5               0.569
#>   fx =~                    
#>     fx1               0.445
#>     fx2               0.443
#>     fx3               0.445
#>     fx4               0.445
#> 
#> Regressions:
#>                     est.std
#>   fm ~                     
#>     fx                0.504
#>   fy ~                     
#>     fm                0.505
#>     fx                0.097
```

The section `Parameter Estimates Based on`
shows the parameter estimates when the
*population model* is fitted to *all the
datasets combined*. When the total
sample size is large, these estimates
should be close to the population values.

The results show that we have specified
the population model correctly. We
can proceed to specify the test and
estimate the power.

## Call `power4test()` to Do the Target Test {#med_power}

We can now do the simulation to estimate
power. A large number of datasets (e.g., 400)
of the target sample size are to be generated,
and then the target test will be conducted
in each of these datasets.

Suppose we would like to estimate the
power of using Monte Carlo confidence interval
to test the indirect effect from `fx` to
`fy` through `fm`, when sample size is 150.
This is the call:


``` r
out <- power4test(nrep = 400,
                  model = mod,
                  pop_es = mod_es,
                  number_of_indicators = k,
                  reliability = mod_rel,
                  n = 150,
                  R = 2000,
                  ci_type = "mc",
                  test_fun = test_indirect_effect,
                  test_args = list(x = "fx",
                                   m = "fm",
                                   y = "fy",
                                   mc_ci = TRUE),
                  iseed = 1234,
                  parallel = TRUE)
```

These are the new arguments used:

- `R`: The number of replications used
  to generate the Monte Carlo simulated
  estimates, 2000 in this example. In
  real studies, this number should be
  10000 or even 20000 for Monte Carlo
  confidence intervals. However, 2000
  is sufficient because the goal is to
  estimate power by generating many
  intervals, rather
  than to have one single stable interval.

- `ci_type`: The method used to
  generate estimates. Support both Monte
  Carlo (`"mc"`) and nonparametric
  bootstrapping (`"boot"`).^[They
  are implemented by `manymome::do_mc()`
  and `manymome::do_boot()`, respectively.]
  Although bootstrapping is usually used
  to test an indirect effect, it is very
  slow to do `R` bootstrapping in `nrep`
  datasets (the model will be fitted
  `R * nrep` times). Therefore, it is
  preferable to use Monte Carlo to do
  the initial estimation.

- `test_fun`: The function to be used
  to do the test for each replication.
  Any function following a specific
  requirement can be used, and
  `power4mome` comes with several
  built-in functions for some common tests.
  The function `test_indirect_effect()`
  is used to test an indirect effect
  in the model.

- `test_args`: A named list of arguments
  to be supplied to `test_fun`.
  For `test_indirect_effect()`, it is
  a named list specifying the predictor
  (`x`), the mediator(s) (`m`), and
  the outcome (`y`). A path with any
  number of mediators can be supported.
  Please refer to the help page of
  `test_indirect_effect()`.^[The test
  is implemented by `manymome::indirect()`.]

- `parallel`: If the test to be conducted
  is slow, which is the case for
  tests done by Monte Carlo or nonparametric
  bootstrapping confidence interval,
  it is advised to enable parallel
  processing by setting `parallel` to
  `TRUE`.^[The number of cores is determined
  automatically but can be set directly
  by the `ncores` argument.]

For `nrep = 400`, the 95% confidence
limits
for a power of .80 are about .04 below
and above .80. This should be precise
enough for determining whether a
sample size has sufficient power. If
a higher precision is desired, set
`nrep` to 1000 or 2000 for the sample
size to be used.

This is the default printout:


``` r
out
#> 
#> ====================== Model Information ======================
#> 
#> == Model on Factors/Variables ==
#> 
#> fm ~ fx
#> fy ~ fm + fx
#> 
#> == Model on Variables/Indicators ==
#> 
#> fm ~ fx
#> fy ~ fm + fx
#> 
#> fm =~ fm1 + fm2 + fm3
#> fy =~ fy1 + fy2 + fy3 + fy4 + fy5
#> fx =~ fx1 + fx2 + fx3 + fx4
#> ====== Population Values ======
#> 
#> Regressions:
#>                    Population
#>   fm ~                       
#>     fx                0.500  
#>   fy ~                       
#>     fm                0.500  
#>     fx                0.100  
#> 
#> Variances:
#>                    Population
#>    .fm                0.750  
#>    .fy                0.690  
#>     fx                1.000  
#> 
#> (Computing indirect effects for 2 paths ...)
#> 
#> == Population Conditional/Indirect Effect(s) ==
#> 
#> == Indirect Effect(s) ==
#> 
#>                  ind
#> fx -> fm -> fy 0.250
#> fx -> fy       0.100
#> 
#>  - The 'ind' column shows the indirect effect(s).
#>  
#> ==== Population Reliability ====
#> 
#>   fy  fm  fx
#>  0.7 0.6 0.5
#> 
#> == Population Standardized Loadings ==
#> 
#>    Standardized Loadings        
#> fm 0.577,0.577,0.577            
#> fx 0.447,0.447,0.447,0.447      
#> fy 0.564,0.564,0.564,0.564,0.564
#> 
#> ======================= Data Information =======================
#> 
#> Number of Replications:  400 
#> Sample Sizes:  150 
#> 
#> Call print with 'data_long = TRUE' for further information.
#> 
#> ==================== Extra Element(s) Found ====================
#> 
#> - fit
#> - mc_out
#> 
#> === Element(s) of the First Dataset ===
#> 
#> ============ <fit> ============
#> 
#> lavaan 0.6-21 ended normally after 38 iterations
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                        27
#> 
#>   Number of observations                           150
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                69.181
#>   Degrees of freedom                                51
#>   P-value (Chi-square)                           0.046
#> 
#> =========== <mc_out> ===========
#> 
#> 
#> == A 'mc_out' class object ==
#> 
#> Number of Monte Carlo replications: 2000 
#> 
#> 
#> ============= <test_indirect: fx->fm->fy> =============
#> 
#> Mean(s) across replication:
#>    est   cilo  cihi   sig pvalue
#>  0.336 -0.056 0.832 0.492  0.100
#> 
#> - The value 'sig' is the rejection rate.
#> - If the null hypothesis is false, this is the power.
#> - Number of valid replications for rejection rate: 400 
#> - Proportion of valid replications for rejection rate: 1.000
```

## Compute the Power

The power estimate is simply the proportion
of significant results, the *rejection*
*rate*, because the
null hypothesis is false. The rejection
rate can be retrieved by
`rejection_rates()`.


``` r
out_power <- rejection_rates(out)
out_power
#> [test]: test_indirect: fx->fm->fy 
#> [test_label]: Test 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.336 1.000  0.492  0.444  0.541
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



In the example above, the estimated
power of the test of the indirect
effect, conducted by Monte Carlo
confidence interval, is
0.492, under the column
`reject`.

`p.v` is
the proportion of valid results
across replications. `1.000` means that
the test conducted normally in all
replications.

By default, the 95% confidence interval
of the rejection rate (power)
based on normal approximation is
also printed, under the column
`r.cilo` and `r.cihi`. In this example,
the 95% confidence interval is
[0.444; 0.541].

# Repeat a Simulation With A Different Sample Size {#new_n}

The function `power4test()` also supports
*redoing* an analysis using a new value
for the sample size (or population effect sizes
set to `pop_es`). Simply

- set the output of `power4test` as the first
argument, and

- set the *new value* for `n`.

For example, we can repeat the
simulation for the test of indirect
effect,
but for a smaller sample size of 200. We
simply call `power4test()` again,
set the previous output (`out` in
the example above) as the
first argument, and set `n` to a new
value (200 in this example):


``` r
out_new_n <- power4test(out,
                        n = 200)
out_new_n
```



This is the estimated power when
the sample size is 200.


``` r
out_new_n_reject <- rejection_rates(out_new_n)
out_new_n_reject
#> [test]: test_indirect: fx->fm->fy 
#> [test_label]: Test 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.352 1.000  0.818  0.777  0.852
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



The estimated power is 0.818,
95% confidence interval
[0.777; 0.852],
when the sample size is
200.

This technique can be
repeated to find the required sample
size for a target power.

# Repeat a Simulation With Different Numbers of Indicators or Reliability

We can also redo an analysis using a new value
for reliability. For example, we may want to see
whether we can have a higher power if
we use more reliable scales.

As in the previous
example, we just call `power4test()` one the
original output, but set `reliability`
to a new vector.

Assume that we want to know the power
for the same scenario, but with all scales
having a population reliability of .80:


``` r
out_new_rel <- power4test(out,
                          reliability = c(fx = .80,
                                          fm = .80,
                                          fy = .80))
out_new_rel
```

This is the estimated power with
higher population reliability:


``` r
out_new_rel_reject <- rejection_rates(out_new_rel)
out_new_rel_reject
#> [test]: test_indirect: fx->fm->fy 
#> [test_label]: Test 
#>     est   p.v reject r.cilo r.cihi
#> 1 0.235 1.000  0.988  0.971  0.995
#> Notes:
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```



The estimated power is 0.988,
95% confidence interval
[0.971; 0.995], much higher than
the original scenario.

# Find the Sample Size With Desired Power

There are several more efficient ways to
find the sample size with the desired power.

## Using `n_region_from_power()`

This is the recommended way for sample
size planning, when there is no predetermined
range of sample sizes, and time is not
a concern in searching for the sample size
with the target level of power.

The function `n_region_from_power()`
can be used to find the *region* of
sample sizes likely to have the desired
power. If the default settings are
to be used, then it can be called
directly on the output of `power4test()`:


``` r
out2_region <- n_region_from_power(out2,
                                   seed = 2345)
```

See
[the templates](https://sfcheung.github.io/power4mome/articles/)
for examples on using `n_region_from_power()`
for common models.

## Using `power4test_by_n()`

This approach is used when the range of
sample sizes has already been decided and
the levels of power are needed to determine
the final sample size.

First, the function
`power4test_by_n()` can be used To
estimate the power for a sequence of
sample sizes. For example, suppose
we do not know that the power is about
.80 with a sample size of 200. We
can estimate the power in the mediation
model above for these sample sizes:
175, 200, 225, 250.


``` r
out_several_ns <- power4test_by_n(out,
                                  n = c(175, 200, 225, 250),
                                  by_seed = 4567)
```

The first argument is the output
of `power4test()` for an arbitrary sample size.

The argument `n` is a numeric vector of
sample sizes to examine. For each `n`,
`nrep` datasets will be generated. Although
there is no limit on the number of sample sizes
to try, it is recommended to restrict the number
of sample sizes to 5 or less.

The argument `by_seed`, if set to an integer,
tries to make the results reproducible.
Note that reproducibility is always possible
due to parallel processing. Use a larger
`nrep` to ensure stability of the results
if necessary.

The call will take some time to run because
it is equivalent to calling `power4test()`
once for each sample size.

The rejection rates for each sample size
can be retrieved by `rejection_rates()`
too:


``` r
rejection_rates(out_several_ns)
#> [test]: test_indirect: fx->fm->fy 
#> [test_label]: Test 
#>     n   est   p.v reject r.cilo r.cihi
#> 1 175 0.327 1.000  0.703  0.656  0.745
#> 2 200 0.329 1.000  0.833  0.793  0.866
#> 3 225 0.340 1.000  0.897  0.864  0.924
#> 4 250 0.330 1.000  0.907  0.875  0.932
#> Notes:
#> - n: The sample size in a trial.
#> - p.v: The proportion of valid replications.
#> - est: The mean of the estimates in a test across replications.
#> - reject: The proportion of 'significant' replications, that is, the
#>   rejection rate. If the null hypothesis is true, this is the Type I
#>   error rate. If the null hypothesis is false, this is the power.
#> - r.cilo,r.cihi: The confidence interval of the rejection rate, based
#>   on Wilson's (1927) method.
#> - Refer to the tests for the meanings of other columns.
```

The method `plot()` can also be used
to plot the results:


``` r
plot(out_several_ns)
```

<div class="figure" style="text-align: center">
<img src="power4test_latent_mediation_plot-1.png" alt="Power by Sample Size"  />
<p class="caption">Power by Sample Size</p>
</div>

The results show that, to have a power of
about .800 to detect the mediation effect,
a sample size of about 200 is needed.

Please refer to the
help page of `power4test_by_n()`
for other examples.

## Using `x_from_power()`

The function `x_from_power()`
can be used to systematically search
within an interval the sample size with
the target power. This takes longer to
run but, instead of manually trying
different sample size, this function
do the search automatically.

This approach can be used when the goal
is to find the probable minimum or
maximum sample size with the desired
level of power. The first approach,
using `n_region_from_power()`, simply
uses this approach twice to find the region
of sample sizes.

See
[this article](https://sfcheung.github.io/power4mome/articles/x_from_power_for_n.html)
for an illustration of how to use
`x_from_power()`.

# Other Scenarios

For other scenarios, such as moderation
and moderated mediation, please refer to
`vignette("power4mome")`.

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rpgnorm_rs.R
\name{rpgnorm_rs}
\alias{rpgnorm_rs}
\title{Random Variable From a Generalized Normal Distribution}
\usage{
rpgnorm_rs(n = 10, p = 2, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{p}{The parameter of the
distribution. Must be a positive
non-zero number. Default is 2,
resulting in a normal distribution.
Higher than 2 results in negative
excess kurtosis. Lower than 2 results
in positive excess kurtosis.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from generalized normal distribution,
rescaled to have user-specified
population mean and standard
deviation.
}
\details{
First, specify the parameter \code{p}
and the desired population
mean and standard deviation. The
random numbers, drawn from a
generalized normal distribution by
\code{\link[pgnorm:rpgnorm]{pgnorm::rpgnorm()}}, will then be
rescaled with the desired population
mean and standard.
}
\examples{
set.seed(90870962)
x <- rpgnorm_rs(n = 5000,
                p = 2,
                pmean = 0,
                psd = 1)
mean(x)
sd(x)
hist(x)
x_kurt <- function(p) {gamma(5/p)*gamma(1/p)/(gamma(3/p)^2) - 3}

p <- 6
x <- rpgnorm_rs(n = 50000,
                p = p,
                pmean = 0,
                psd = 1)
mean(x)
sd(x)
x_kurt(p)
qqnorm(x); qqline(x)

p <- 1
x <- rpgnorm_rs(n = 50000,
                p = p,
                pmean = 0,
                psd = 1)
mean(x)
sd(x)
x_kurt(p)
qqnorm(x); qqline(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_model.R
\name{fit_model}
\alias{fit_model}
\title{Fit a Model to a List of Datasets}
\usage{
fit_model(
  data_all = NULL,
  model = NULL,
  fit_function = "lavaan",
  arg_data_name = "data",
  arg_model_name = "model",
  arg_group_name = "group",
  ...,
  fit_out = NULL,
  parallel = FALSE,
  progress = FALSE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  cl = NULL
)
}
\arguments{
\item{data_all}{The output
of \code{\link[=sim_data]{sim_data()}}, or a \code{sim_data}
class object.}

\item{model}{The model to be fitted.
If \code{NULL}, the default, the model
stored in \code{data_all}, which should
be the data generation model,
will be used.}

\item{fit_function}{The function to
be used to fit the model. Can also
be a string: \code{"lavaan"} (the default) for
\code{\link[lavaan:sem]{lavaan::sem()}}, and \code{"lm"} or \code{many_lm}
for \code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}.
Other functions can also be used
if necessary.}

\item{arg_data_name}{The name of the
argument of \code{fit_function} expecting
the dataset. Default is \code{"data"}.}

\item{arg_model_name}{The name of
the argument of \code{fit_function}
expecting the model definition.
Default is \code{"model"}.}

\item{arg_group_name}{The name of
the argument of \code{fit_function}
expecting the name of the group
variable. Used only for multigroup
models. Default is \code{"group"}.}

\item{...}{Optional arguments to be
passed to \code{fit_function} when
fitting the model.}

\item{fit_out}{If set to a \code{fit_out}
object (the output of \code{\link[=fit_model]{fit_model()}}),
then all missing arguments will be
retrieved from \code{fit_out}. That is,
users can use \code{fit_model(data_all = new_data, fit_out = old_out)}
to re-fit a model originally fitted
in \code{old_out} on a new list of dataset
(\code{new_data}). No need to include
all other arguments.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to fit the
models. Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
of model fitting will be displayed.
Default is `FALSE.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{cl}{A cluster, such as one created
by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}. If \code{NULL},
a cluster will be created, but will be
stopped on exit. If set to an existing
cluster, it will not be stopped when
the function exits; users need to
stop it manually.}
}
\value{
An object of the class \code{fit_out},
which is a list of the output of
\code{fit_function} (\code{\link[lavaan:sem]{lavaan::sem()}}
by default). If an error occurred
when fitting the model to a dataset,
then this element will be the error
message from the fit function.
}
\description{
Get the output of
\code{\link[=sim_data]{sim_data()}} and fit a model to each
of the stored datasets.
}
\details{
By default, the function \code{\link[=fit_model]{fit_model()}}
\itemize{
\item extracts the model
stored in the output of \code{\link[=sim_data]{sim_data()}},
\item fits the model to each dataset
simulated using \code{fit_function},
default to \code{"lavaan"} and
\code{\link[lavaan:sem]{lavaan::sem()}} will be called,
\item and returns the results.
}

If the datasets
were generated from a multigroup
model when calling \code{\link[=sim_data]{sim_data()}},
a multigroup model is fitted.
}
\section{The role of \code{fit_model()}}{
This function is used by the
all-in-one function \code{\link[=power4test]{power4test()}}.
Users usually do not call this
function directly, though
developers can use this function to
customize the model fitting step in
power analysis.
}

\examples{

# Specify the population model

mod <-
"m ~ x
 y ~ m + x"

# Specify the effect sizes (population parameter values)

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate several simulated datasets

data_all <- sim_data(nrep = 5,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     iseed = 1234)

# Fit the population model to each datasets

fit_all <- fit_model(data_all)
fit_all[[1]]

# Fit the population model using the MLR estimator

fit_all_mlr <- fit_model(data_all,
                         estimator = "MLR")
fit_all_mlr[[1]]

# Fit a model different from the population model,
# with the MLR estimator

mod2 <-
"m ~ x
 y ~ m"

fit_all_mlr2 <- fit_model(data_all,
                          mod2,
                          estimator = "MLR")
fit_all_mlr2[[1]]

}
\seealso{
See \code{\link[=power4test]{power4test()}} for
the all-in-one function that uses
this function, and \code{\link[=sim_data]{sim_data()}}
for the function generating datasets
for this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ptable_pop.R
\name{ptable_pop}
\alias{ptable_pop}
\alias{model_matrices_pop}
\title{Generate the Population Model}
\usage{
ptable_pop(
  model = NULL,
  pop_es = NULL,
  es1 = c(n = 0, nil = 0, s = 0.1, m = 0.3, l = 0.5, si = 0.141, mi = 0.361, li = 0.51,
    sm = 0.2, ml = 0.4),
  es2 = c(n = 0, nil = 0, s = 0.05, m = 0.1, l = 0.15, sm = 0.075, ml = 0.125),
  es_ind = c("si", "mi", "li"),
  standardized = TRUE,
  n_std = 1e+05,
  std_force_monte_carlo = FALSE,
  add_cov_for_moderation = TRUE
)

model_matrices_pop(x, ..., drop_list_single_group = TRUE)
}
\arguments{
\item{model}{String. The model defined
by \code{lavaan} model syntax. See 'Details'.}

\item{pop_es}{It can be a data frame
with these columns: \code{lhs}, \code{op},
\code{rhs}, and \code{pop}. The first three
columns correspond to those in a
\code{lavaan} parameter table. The column
\code{pop} stores the population values.
The column \code{es} stores the original
labels, for reference. It can also be
a named character vector (named list
for multigroup models) or a multiline string,
which are
the preferred approaches. See the
help page
on how to specify this vector.}

\item{es1}{Set the
values for each label of the effect
size (population value) for correlations and regression
paths.
Used only if \code{pop_es} is a named
vector or a multiline string.
See the help page on how to specify
this argument.}

\item{es2}{Set the
values for each label of the effect
size (population value) for product term.
Used only if \code{pop_es} is a named
vector or a multiline string.
See the help page on how to specify
this argument.}

\item{es_ind}{The names of labels
denoting the effect size of an
indirect effect. They will be
used to determine the population
values of the component paths along
an indirect path.}

\item{standardized}{Logical. If
\code{TRUE}, the default, variances and
error variances are scaled to ensure
the population variances of the
endogenous variables are close to
one, and hence the effect sizes
(population values) are
standardized effect sizes if the
variances of the continuous exogenous
variables are also equal to one.}

\item{n_std}{The sample size used to
determine the error variances by
simulation when \code{std_force_monte_carlo}
is \code{TRUE}.}

\item{std_force_monte_carlo}{Logical.
If \code{FALSE}, the default,
standardization is done analytically
if the model has no product terms,
and by simulation if the model has
product terms. That is, error variances
required to ensure implied variances equal
to one are determined by simulation.
If \code{TRUE}, simulation
will be used whether the model has
product terms or not. Always fall
back to simulation if
analytical standardization failed.}

\item{add_cov_for_moderation}{Logical. If \code{TRUE}, the default, for
a model in which one or product terms
for moderation involve one or more
mediator, covariances between their
error terms and the product terms
will be added automatically. If these
covariances are not added, the model
may not be invariant to linear
transformation of some variables in
the model.}

\item{x}{It can be a 'lavaan' model
syntax, to be passed to \code{\link[=ptable_pop]{ptable_pop()}},
or a parameter table with the column
\code{start} set to the population values,
such as the output of \code{\link[=ptable_pop]{ptable_pop()}}.}

\item{...}{If \code{x} is a model syntax,
these are arguments to be passed to
\code{\link[=ptable_pop]{ptable_pop()}}.}

\item{drop_list_single_group}{If
\code{TRUE} and the number of groups is
equal to one, the output will be
a list of matrices of one group
only. Default if \code{TRUE}.}
}
\value{
The function \code{\link[=ptable_pop]{ptable_pop()}} returns
a \code{lavaan} parameter table of the
model, with the column \code{start} set to the
population values.

The function \code{\link[=model_matrices_pop]{model_matrices_pop()}}
returns a \code{lavaan} LISREL-style model
matrices (like the output of
\code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} with \code{what}
set to \code{"free"}), with matrix elements
set to the population values. If
\code{x} is the model syntax, it will be
stored in the attributes \code{model}.
If the model is a multigroup model
with \emph{k} groups (\emph{k} greater than 1),
then it returns a list of \emph{k} lists
of \code{lavaan} LISREL-style model
matrices unless \code{drop_list_single_group}
is \code{TRUE}.
}
\description{
Generate the complete
population model using the model
syntax and user-specified effect
sizes (population parameter values).
}
\details{
The function \code{\link[=ptable_pop]{ptable_pop()}} generates a \code{lavaan}
parameter table that can be used
to generate data based on the population
values of model parameters.
}
\section{The role of \code{ptable_pop()}}{
The function \code{\link[=ptable_pop]{ptable_pop()}} is used by
the all-in-one function
\code{\link[=power4test]{power4test()}}. Users usually do not
call this function directly, though
developers can use this function to
develop other functions for power
analysis, or to build their own
workflows to do the power analysis.
}

\section{Specify the Population Model by 'model'}{
\subsection{Single-Group Model}{

For a single-group model, \code{model}
should be a \code{lavaan} model syntax
string of the \emph{form} of the model.
The population values of the model
parameters are to be determined by
\code{pop_es}.

If the model has latent factors,
the syntax in \code{model} should specify
only the \emph{structural model} for the
\emph{latent factors}. There is no
need to specify the measurement
part. Other functions will generate
the measurement part on top of this
model.

For example, this is a simple mediation
model:

\preformatted{"m ~ x
 y ~ m + x"}

Whether \code{m}, \code{x}, and \code{y} denote
observed variables or latent factors
are determined by other functions,
such as \code{\link[=power4test]{power4test()}}.
}

\subsection{Multigroup Model}{

Because the model is the population
model, equality constraints are
irrelevant and the model syntax
specifies only the \emph{form} of the
model. Therefore, \code{model} is
specified as in the case of single
group models.
}
}

\section{Specify 'pop_es' Using Named Vectors}{
The argument \code{pop_es} is for specifying
the population values of model
parameters. This section describes
how to do this using named vectors.
\subsection{Single-Group Model}{

If \code{pop_es} is specified by a named
vector, it must follow the convention
below.
\itemize{
\item The names of the vectors are
\code{lavaan} names for the selected
parameters. For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
\item Alternatively, the names can be
either \code{".beta."} or \code{".cov."}.
Use \code{".beta."} to set the default
values for all regression coefficients.
Use \code{".cov."} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
\item The names can also be of this form:
\code{".ind.(<path>)"}, whether \verb{<path>}
denote path in the model. For
example, \code{".ind.(x->m->y)"} denotes
the path from \code{x} through \code{m} to
\code{y}. Alternatively, the \code{lavaan}
symbol \code{~} can also be used:
\code{".ind.(y~m~x)"}. This form is used
to set the indirect effect (standardized,
by default) along this path. The
value for this name will override
other settings.
\item If using \code{lavaan} names, can
specify more than one parameter
using \code{+}. For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
\item The value of each element can be
the label for the effect size: \code{n}
for nil, \code{s} for small, \code{m} for
medium, and \code{l} for large. The
value for each label is determined
by \code{es1} and \code{es2}. See the section
on specifying these two arguments.
\item The value of \code{pop_es} can also be
set to a value, but it must be
quoted as a string, such as \code{"y ~ x" = ".31"}.
}

This is an example:

\preformatted{c(".beta." = "s",
  "m1 ~ x" = "-m",
  "m2 ~ m1" = "l",
  "y ~ x:w" = "s")}

In this example,
\itemize{
\item All regression coefficients are
set to small (\code{s}) by default, unless
specified otherwise.
\item The path from \code{x} to \code{m1} is
set to medium and negative (\code{-m}).
\item The path from \code{m1} to \code{m2} is set
to large (\code{l}).
\item The coefficient of the product
term \code{x:w} when predicting \code{y} is
set to small (\code{s}).
}
\subsection{Indirect Effect}{

When setting an indirect effect to
a symbol (default: \code{"si"}, \code{"mi"},
\code{"li"}, with \code{"i"} added to differentiate
them from the labels for a direct path),
the corresponding value is used to
determine the population values of
\emph{all} component paths along the pathway.
All the values are assumed to be equal.
Therefore, \code{".ind.(x->m->y)" = ".20"}
is equivalent to setting \code{m ~ x}
and \code{y ~ m} to the square root of
.20, such that the corresponding
indirect effect is equal to the
designated value.

This behavior, though restricted,
is for quick manipulation of the
indirect effect. If different values
along a pathway, set the value for
each path directly.

Only nonnegative value is supported.
Therefore, \code{".ind.(x->m->y)" = "-si"}
and \code{".ind.(x->m->y)" = "-.20"} will
throw an error.
}

}

\subsection{Multigroup Model}{

The argument \code{pop_es} also supports multigroup
models.

For \code{pop_es}, instead of
named vectors, named \emph{list} of
named vectors should be used.
\itemize{
\item The names are the parameters, or
keywords such as \code{.beta.} and
\code{.cov.}, like specifying the
population values for a single
group model.
\item The elements are character vectors.
If it has only one element (e.g.,
a single string), then it is the
the population value for all groups.
If it has more than one element
(e.g., a vector of three strings),
then they are the population values
of the groups. For a model of \emph{k}
groups, each vector must have
either \emph{k} elements or one element.
}

This is an example:

\preformatted{list("m ~ x" = "m",
     "y ~ m" = c("s", "m", "l"))}

In this model, the population value
of the path \code{m ~ x} is medium (\code{m}) for
all groups, while the population
values for the path \code{y ~ m} are
small (\code{s}), medium (\code{m}), and large (\code{l}),
respectively.
}
}

\section{Specify 'pop_es' Using a Multiline String}{
When setting the argument \code{pop_es},
instead of using a named vector
or named list for
\code{pop_es}, the population values of
model parameters can also be
specified using a multiline string,
as illustrated below, to be parsed
by \code{\link[=pop_es_yaml]{pop_es_yaml()}}.
\subsection{Single-Group Model}{

This is an example of the multiline string
for a single-group model:

\preformatted{y ~ m: l
m ~ x: m
y ~ x: nil}

The string must follow this format:
\itemize{
\item Each line starts with \verb{tag:}.
\itemize{
\item \code{tag} can be the name of a
parameter, in \code{lavaan} model
syntax format.
\itemize{
\item For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
}
\item A tag in \code{lavaan} model syntax can
specify more than one parameter
using \code{+}.
\itemize{
\item For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
}
\item Alternatively, the \code{tag} can be
either \code{.beta.} or \code{.cov.}.
\itemize{
\item Use \code{.beta.} to set the default
values for all regression coefficients.
\item Use \code{.cov.} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
}
}
\item After each tag is the value of the
population value:

-\code{nil} for nil (zero),
\itemize{
\item \code{s} for small,
\item \code{m} for medium, and
\item \code{l} for large.
\item \code{si}, \code{mi}, and \code{li} for
small, medium, and large a
standardized indirect effect,
respectively.
}

Note: \code{n} \emph{cannot} be used in this mode.

The
value for each label is determined
by \code{es1} and \code{es2} as described
in \code{\link[=ptable_pop]{ptable_pop()}}.
\itemize{
\item The value can also be
set to a numeric value, such as
\code{.30} or \code{-.30}.
}
}

This is another example:

\preformatted{.beta.: s
y ~ m: l}

In this example, all regression
coefficients are \code{small}, while
the path from \code{m} to \code{y} is large.
}

\subsection{Multigroup Model}{

This is an example of the string
for a multigroup model:

\preformatted{y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil}

The format is similar to that for
a single-group model. If a parameter
has the same value for all groups,
then the line can be specified
as in the case of a single-group
model: \code{tag: value}.

If a parameter has different
values across groups, then it must
be in this format:
\itemize{
\item A line starts with the tag, followed
by two or more lines. Each line
starts with a hyphen \code{-} and the
value for a group.
}

For example:

\preformatted{m ~ x:
  - nil
  - s}

This denotes that the model has
two groups. The values of the path
from \code{x} to \code{m} for the two
groups are 0 (\code{nil}) and
small (\code{s}), respectively.

Another equivalent way to specify
the values are using \verb{[]}, on
the same line of a tag.

For example:

\preformatted{m ~ x: [nil, s]}

The number of groups is inferred
from the number of values for
a parameter. Therefore, if a tag
has more than one value, each tag
must has the same number of value,
or only one value.

The tag \code{.beta.} and \code{.cov.} can
also be used for multigroup models.
}

\subsection{Which Approach To Use}{

Note that using named vectors or
named lists is more reliable. However,
using a multiline string is
more user-friendly. If this method
failed, please use named vectors or
named list instead.
}

\subsection{Technical Details}{

The multiline string is parsed by \code{\link[yaml:read_yaml]{yaml::read_yaml()}}.
Therefore, the format requirement
is actually that of YAML. Users
knowledgeable of YAML can use other
equivalent way to specify the string.
}
}

\section{Set the Values for Effect Size Labels ('es1' and 'es2')}{
The vector \code{es1} is for correlations,
regression coefficients, and
indirect effect, and the
vector \code{es2} is for for standardized
moderation effect, the coefficients
of a product term. These labels
are to be used in interpreting
the specification in \code{pop_es}.
}

\section{The role of \code{model_matrices_pop()}}{
The function \code{\link[=model_matrices_pop]{model_matrices_pop()}}
generates models matrices with
population values, used by \code{\link[=ptable_pop]{ptable_pop()}}.
Users usually do not
call this function directly, though
developers can use this build their own
workflows to generate the data.
}

\examples{

# Specify the model

model1 <-
"
m1 ~ x + c1
m2 ~ m1 + x2 + c1
y ~  m2 + m1 + x + w + x:w + c1
"

# Specify the population values

model1_es <- c("m1 ~ x" = "-m",
               "m2 ~ m1" = "s",
               "y ~ m2" = "l",
               "y ~ x" = "m",
               "y ~ w" = "s",
               "y ~ x:w" = "s",
               "x ~~ w" = "s")

ptable_final1 <- ptable_pop(model1,
                            pop_es = model1_es)
ptable_final1

# Use a multiline string, illustrated by a simpler model

model2 <-
"
m ~ x
y ~ m + x
"

model2_es_a <- c("m ~ x" = "s",
                 "y ~ m" = "m",
                 "y ~ x" = "nil")

model2_es_b <-
"
m ~ x: s
y ~ m: m
y ~ x: nil
"

ptable_model2_a <- ptable_pop(model2,
                              pop_es = model2_es_a)
ptable_model2_b <- ptable_pop(model2,
                              pop_es = model2_es_b)

ptable_model2_a
ptable_model2_b

identical(ptable_model2_a,
          ptable_model2_b)

# model_matrices_pop

model_matrices_pop(ptable_final1)

model_matrices_pop(model1,
                   pop_es = model1_es)

}
\seealso{
\code{\link[=power4test]{power4test()}}, and
\code{\link[=pop_es_yaml]{pop_es_yaml()}} on an alternative
way to specify population values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_data_ampute.R
\name{missing_values}
\alias{missing_values}
\title{Process Data by Generating Missing Values}
\usage{
missing_values(data, ..., prop = 0.5, mech = "MCAR")
}
\arguments{
\item{data}{A data frame.}

\item{...}{Optional arguments to be
passed to \code{\link[mice:ampute]{mice::ampute()}}.}

\item{prop}{The proportion of missingness.
Default is 0.5, about 50\% of the cases
have missing data.}

\item{mech}{The missing data mechanism.
Default is \code{"MCAR"} (missing completely at random).
Other possible values are \code{"MAR"}
(missing at random) and \code{"MNAR"}
(missing not at random). Please refer
to the help of \code{\link[mice:ampute]{mice::ampute()}} for
details.}
}
\value{
It returns a data frame with the
missing values inserted.
}
\description{
For the \code{process_data}
argument. It introduces missing
values in the generated data.
}
\details{
This function is to be used in
the \code{process_data} argument of
\code{\link[=power4test]{power4test()}}.

The missing values are generated by
\code{\link[mice:ampute]{mice::ampute()}}. Please refer to
its help page, the vignette for
this function:
\href{https://rianneschouten.github.io/mice_ampute/vignette/ampute.html}{Generate missing values with ampute},
and Schouten, Lugtig, and Vink (2018).

In addition to \code{data}, only two
arguments are explicitly defined for
\code{\link[=missing_values]{missing_values()}}: \code{prop} and \code{mech}.
The argument \code{prop} is defined to remind
users of the default value for this
argument. The argument \code{mech}, specifying
the missing data mechanism, is
defined because its default value
is different from that of \code{\link[mice:ampute]{mice::ampute()}}.
For \code{\link[=missing_values]{missing_values()}}, the default value
is \code{"MCAR"}, missing completely at
random.

Please refer to the help page of
\code{\link[mice:ampute]{mice::ampute()}} for other available
arguments.
\subsection{Note for \code{patterns} and other arguments with variable names}{

If arguments such as \code{patterns} are
specified, the variable names need to be
those in the generated data. If
indicator scores are generated, the
names need to be the names of the indicators,
not the names of the latent variables.
}
}
\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"
# Simulate the data

out <- power4test(
         nrep = 2,
         model = mod,
         pop_es = mod_es,
         n = 200,
         process_data = list(fun = "missing_values",
                             args = list(prop = .75)),
         test_fun = test_parameters,
         test_args = list(op = "~"),
         parallel = FALSE,
         iseed = 1234)

dat <- pool_sim_data(out)
head(dat, 50)

}
\references{
Schouten, R. M., Lugtig, P., & Vink, G.
(2018). Generating missing values for
simulation purposes: A multivariate
amputation procedure.
\emph{Journal of Statistical Computation and Simulation, 88}(15),
2909--2930. \doi{10.1080/00949655.2018.1491577}
}
\seealso{
\code{\link[=power4test]{power4test()}}. See also
\code{\link[mice:ampute]{mice::ampute()}} for the amputation
method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power_curve.R
\name{power_curve}
\alias{power_curve}
\alias{print.power_curve}
\title{Power Curve}
\usage{
power_curve(
  object,
  formula = NULL,
  start = NULL,
  lower_bound = NULL,
  upper_bound = NULL,
  nls_args = list(),
  nls_control = list(),
  verbose = FALSE,
  models = c("nls", "logistic", "lm"),
  nls_options = list(min_points = 4, regions = c(0, 0.45, 0.75, 0.85, 1))
)

\method{print}{power_curve}(x, data_used = FALSE, digits = 3, right = FALSE, row.names = FALSE, ...)
}
\arguments{
\item{object}{An object of the class
\code{power4test_by_n} or \code{power4test_by_es},
which is the output of \code{\link[=power4test_by_n]{power4test_by_n()}}
or \code{\link[=power4test_by_es]{power4test_by_es()}}.}

\item{formula}{A formula of the model
for \code{\link[stats:nls]{stats::nls()}}. It can also be
a list of formulas, and the models
will be fitted successively by
\code{\link[stats:nls]{stats::nls()}}, with the first
model fitted successfully adopted.
The response variable in the formula must be named
\code{reject}, and the predictor named
\code{x}. Whether \code{x} represents \code{n} or
\code{es} depends on the class of \code{object}.
If \code{NULL}, the default, it will be
determined internally based on the type of
\code{object}.}

\item{start}{Either a named vector
of the start value(s) of parameter(s)
in \code{formula}, or a list of named
vectors of the starting value(s)
of the list of formula(s). If \code{NULL},
the default, they will be determined
internally.}

\item{lower_bound}{Either a named vector
of the lower bound(s) of parameter(s)
in \code{formula}, or a list of named
vectors of the lower bound(s)
for the list of formula(s). They will
be passed to \code{lower} of \code{\link[stats:nls]{stats::nls()}}.
If \code{NULL}, the default, it will be
determined internally based on the type of
\code{object}.}

\item{upper_bound}{Either a named vector
of the upper bound(s) of parameter(s)
in \code{formula}, or a list of named
vectors of the upper bound(s)
for the list of formula(s). They will
be passed to \code{upper} of \code{\link[stats:nls]{stats::nls()}}.
If \code{NULL}, the default, it will be
determined internally based on the type of
\code{object}.}

\item{nls_args}{A named list of
arguments to be used when calling
\code{\link[stats:nls]{stats::nls()}}. Used to override
internal default, such as the
algorithm (default is \code{"port"}).
Use this argument with cautions.}

\item{nls_control}{A named list of
arguments to be passed the \code{control}
argument of \code{\link[stats:nls]{stats::nls()}}. The values will
override internal default values,
and also override \code{nls_args}.
Use this argument with cautions.}

\item{verbose}{Logical. Whether
messages will be printed when
trying different models.}

\item{models}{Models to try. Support
\code{"nls"} (fitted by \code{\link[=nls]{nls()}}),
\code{"logistic"} (fitted by \code{\link[=glm]{glm()}}), and
\code{"lm"} (fitted by \code{\link[=lm]{lm()}}). By default,
all three models will be attempted,
in this order.}

\item{nls_options}{A named list of
options to be used by \code{\link[=power_curve]{power_curve()}}
to configure the use of \code{"nls"}. For
advanced use. See 'Details' for
available options.}

\item{x}{A \code{power_curve} object.}

\item{data_used}{Logical. Whether
the rejection rates data frame
used to fit the model is printed.}

\item{digits, right, row.names}{Arguments of the same names used
by the \code{print} method of a
\code{data.frame} object. Used when \code{data_used}
is \code{TRUE} and the rejection rates
data frame is printed.}

\item{...}{For the \code{print} method of
\code{power_curve} objects, they are optional
arguments to be passed to
\code{\link[=print.data.frame]{print.data.frame()}} when printing
the rejection rates data frame.}
}
\value{
It returns a list which is a
\code{power_curve} object, with the
following elements:
\itemize{
\item \code{fit}: The model fitted, which is the
output of \code{\link[stats:nls]{stats::nls()}},
\code{\link[stats:glm]{stats::glm()}}, or \code{\link[stats:lm]{stats::lm()}}.
\item \code{reject_df}: The table of reject
rates and other characteristics,
which is generated by
\code{\link[=rejection_rates]{rejection_rates()}}.
\item \code{predictor}: The predictor or the
power curve, ether \code{"n"} (sample
size) or \code{"es"} (population effect
size).
\item \code{call}: The call used to run this
function.
}

The \code{print} method of \code{power_curve}
object returns \code{x} invisibly. Called
for its side-effect.
}
\description{
Estimate the relation
between power and a characteristic,
such as sample size or population
effect size (population value of
a model parameter).
}
\details{
The function \code{\link[=power_curve]{power_curve()}}
retrieves the information
from the output of
\code{\link[=power4test_by_n]{power4test_by_n()}} or
\code{\link[=power4test_by_es]{power4test_by_es()}}, and
estimate the power curve: the
relation between the characteristic
varied, sample size for
\code{\link[=power4test_by_n]{power4test_by_n()}} and the
population effect size for
\code{\link[=power4test_by_es]{power4test_by_es()}}, and the
rejection rate of the test conducted
by \code{\link[=power4test_by_n]{power4test_by_n()}} or
\code{\link[=power4test_by_es]{power4test_by_es()}}. This
rejection rate is the power when the
null hypothesis is false (e.g., the
population value of the effect size
being tested is nonzero).

The model fitted is \emph{not} intended to
be a precise model for the relation
across a wide range. It is only a
crude estimate based on the limited
number of values of the
characteristic (e.g., sample size)
examined, which can be as small as
four or even smaller. The model is
intended to be
used for only for the range covered,
and for estimating the probable
sample size or effect size with a
desirable level of power. This value
should then be studied by higher
precision through simulation
using functions such as
\code{\link[=power4test]{power4test()}}.

These are the models to be tried,
in the following order:
\itemize{
\item One or nonlinear models, to be
fitted by \code{\link[stats:nls]{stats::nls()}}. If
several models are specified, all
will be fitted and the one with the
smallest deviance will be used.
\item If all the nonlinear models failed,
for whatever reason, a logistic
regression will be fitted by
\code{\link[stats:glm]{stats::glm()}} to predict the
binary significant test results.
\item If the logistic model also failed,
for whatever reason, a simple
linear regression model will be
fitted. Although the power curve is
nonlinear across a wide range of,
say, sample size, a linear model
can still be a good enough
approximation for a narrow range of
the predictor.
}

The output can then be plotted to
visualize the power curve, using
the \code{plot} method (\code{\link[=plot.power_curve]{plot.power_curve()}})
for the output
of \code{\link[=power_curve]{power_curve()}}.

This function can be used directly,
but is also used internally by
functions such as \code{\link[=x_from_power]{x_from_power()}}.
\subsection{'nls_options'}{

These are possible arguments.
\itemize{
\item \code{min_points}: If the object has
less than \code{min_points} rejection
rates, \code{nls} will not be used.
\item \code{regions}: A numeric vector with
values from 0 to 1. It defines
regions in which there must be
at least one rejection rate. If this
criterion is not met, \code{nls} will
not be used. This is to ensure that
\code{nls} will not be used when the
rejection rates are clustered around
a very narrow region.
}
}
}
\examples{

# Specify the population model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the effect sizes (population parameter values)

model_simple_med_es <-
"
y ~ m: l
m ~ x: m
y ~ x: s
"

# Simulate datasets to check the model
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

sim_only <- power4test(nrep = 10,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = FALSE)

# By n: Do a test for different sample sizes

out1 <- power4test_by_n(sim_only,
                        nrep = 10,
                        test_fun = test_parameters,
                        test_args = list(par = "y~x"),
                        n = c(25, 50, 100),
                        by_seed = 1234,
                        parallel = FALSE,
                        progress = FALSE)

pout1 <- power_curve(out1)
pout1
plot(pout1)

# By pop_es: Do a test for different population values of a model parameter

out2 <- power4test_by_es(sim_only,
                         nrep = 10,
                         test_fun = test_parameters,
                         test_args = list(par = "y~x"),
                         pop_es_name = "y ~ x",
                         pop_es_values = c(0, .3, .5),
                         by_seed = 1234,
                         parallel = FALSE,
                         progress = FALSE)

pout2 <- power_curve(out2)
pout2
plot(pout2)

}
\seealso{
\code{\link[=power4test_by_n]{power4test_by_n()}} and \code{\link[=power4test_by_es]{power4test_by_es()}}
for the output supported by
\code{\link[=power_curve]{power_curve()}}, \code{\link[=plot.power_curve]{plot.power_curve()}}
for the \code{plot} method and
\code{\link[=predict.power_curve]{predict.power_curve()}}
for the \code{predict} method of the output
of \code{\link[=power_curve]{power_curve()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize_test.R
\name{summarize_tests}
\alias{summarize_tests}
\alias{print.test_summary_list}
\alias{print.test_summary}
\alias{print.test_out_list}
\title{Summarize Test Results}
\usage{
summarize_tests(
  object,
  collapse = c("none", "all_sig", "at_least_one_sig", "at_least_k_sig"),
  at_least_k = 1,
  merge_all_tests = FALSE,
  p_adjust_method = "none",
  alpha = 0.05
)

\method{print}{test_summary_list}(x, digits = 3, ...)

\method{print}{test_summary}(x, digits = 2, ...)

\method{print}{test_out_list}(x, digits = 3, test_long = TRUE, ...)
}
\arguments{
\item{object}{A \code{power4test} object
or the element \code{test_all} in
a \code{power4test} object.}

\item{collapse}{Whether a single
decision (significant vs. not significant)
is made across all tests for a test
that consists of several tests
(e.g., the tests of several parameters).
If \code{"none"}, tests will be summarized
individually. If \code{"all_sig"}, then
the set of tests is considered significant
if all individual tests are significant.
If \code{"at_least_one_sig"}, then the set of
tests is considered significant if
at least one of the tests is significant.
If \code{"at_least_k_sig"}, then the set of
tests is considered significant if
at least \code{k} tests are significant,
\code{k} set by the argument \code{at_least_k}.}

\item{at_least_k}{Used by \code{collapse},
the number of tests required to be
significant for the set of tests to
be considered significant.}

\item{merge_all_tests}{If \code{TRUE}, all
the tests in each replication will be
merged into one test.}

\item{p_adjust_method}{The method to be
passed to \code{\link[=p.adjust]{p.adjust()}} to adjust the
\emph{p}-values when testing the effects.
Default is \code{"none"} and the \emph{p}-values
will not be adjusted. The unadjusted
\emph{p}-values will be stored in the column
\code{pvalue_org}. Ignored if some tests
do not have \emph{p}-values stored.
NOTE: Use this only if all tests
can be  conducted using \emph{p}-values.}

\item{alpha}{The level of significance
to use when using \code{p_adjust_method}.
The significance results (the column
\code{sig}) will be updated using the
adjusted \emph{p}-values. Used only if
\code{p_adjust_method} is not \code{"none"}.}

\item{x}{The object to be printed.}

\item{digits}{The numbers of digits
after the decimal when printing
numeric results.}

\item{...}{Optional arguments.
Not used.}

\item{test_long}{If \code{TRUE}, a detailed
report will be printed.}
}
\value{
The function \code{\link[=summarize_tests]{summarize_tests()}} returns
a list of the class \code{test_summary_list}.
Each element contains a summary of a
test stored. The elements are of
the class \code{test_summary}, with
these elements:
\itemize{
\item \code{test_attributes}: The stored
information of a test, for printing.
\item \code{nrep}: The number of datasets
(replications).
\item \code{mean}: The means of numeric
information. For significance
tests, these are the rejection
rates.
\item \code{nvalid}: The number of non-\code{NA}
replications used to compute each
mean.
}

The \code{print} methods returns \code{x} invisibly.
They are called for their side
effects.
}
\description{
Extract and summarize
test results.
}
\details{
The function \code{\link[=summarize_tests]{summarize_tests()}}
is used to extract
information from each test stored
in a \code{power4test} object.

The method \code{print.test_out_list()} is
used to print the content of a list
of test stored in a \code{power4test}
object, with the option to print
just the names of tests.
}
\section{The role of \code{summarize_tests()} and related functions}{
The function \code{\link[=summarize_tests]{summarize_tests()}} and
related print methods are used by
the all-in-one function
\code{\link[=power4test]{power4test()}} and its \code{summary}
method. Users usually do not
call them directly, though
developers can use this function to
develop other functions for power
analysis, or to build their own
workflows to do the power analysis.
}

\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Simulated datasets

sim_only <- power4test(nrep = 2,
                       model = mod,
                       pop_es = es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Test the parameters in each dataset

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

# Print the summary

summarize_tests(test_out)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim_out.R
\name{sim_out}
\alias{sim_out}
\alias{print.sim_out}
\title{Create a 'sim_out' Object}
\usage{
sim_out(data_all, ...)

\method{print}{sim_out}(x, digits = 3, digits_descriptive = 2, fit_to_all_args = list(), ...)
}
\arguments{
\item{data_all}{The output of
\code{\link[=sim_data]{sim_data()}}.}

\item{...}{Named arguments of
objects to be added to each
replication under the element
\code{extra}. For example, if set to
\code{fit = fit_all}, where \code{fit_all}
is the output of \code{\link[=fit_model]{fit_model()}},
then \code{data_all[[1]]$extra$fit}
will be set to the first output
in \code{fit_all}.}

\item{x}{The \code{sim_out} object
to be printed.}

\item{digits}{The numbers of digits
displayed after the decimal.}

\item{digits_descriptive}{The
number of digits displayed after
the decimal for the descriptive
statistics table.}

\item{fit_to_all_args}{A named list
of arguments to be passed to
\code{\link[lavaan:sem]{lavaan::sem()}} when the model is
fitted to a sample combined from
all samples stored.}
}
\value{
The function \code{\link[=sim_out]{sim_out()}} returns a \code{sim_out} object, which
is a list of length equal to the
length of \code{data_all}. Each element
of the list is a \code{sim_data} object
with the element \code{extra} added to
it. Other named elements will be
added under this name. For example.
the output of \code{\link[=fit_model]{fit_model()}}
for this replication can be added
to \code{fit}, under \code{extra}. See
the description of the argument
\code{...} for details.

The \code{print} method of \code{sim_out}
returns \code{x} invisibly. Called for
its side effect.
}
\description{
Combine the outputs
of \code{\link[=sim_data]{sim_data()}}, \code{\link[=fit_model]{fit_model()}},
and optionally \code{\link[=gen_mc]{gen_mc()}} and/or
\code{\link[=gen_boot]{gen_boot()}} to one
single object.
}
\details{
It merges into one object the output
of \code{\link[=sim_data]{sim_data()}}, which is a list of
\code{nrep} simulated datasets,
\code{\link[=fit_model]{fit_model()}}, which is a list of the
\code{\link[lavaan:sem]{lavaan::sem()}} output for the \code{nrep}
datasets, and optionally the output
of \code{\link[=gen_mc]{gen_mc()}} or \code{\link[=gen_boot]{gen_boot()}}, which is a list of the
\code{R} sets of Monte Carlo or bootstrap estimates
based on the results of
\code{\link[=fit_model]{fit_model()}}. The list has \code{nrep}
elements, each element with the data,
the model fit
results, and optionally the Monte
Carlo estimates matched.

This object can then be used for testing
effects of interests, which are
further processed to estimate the
power of this test.

The function \code{\link[=sim_out]{sim_out()}} is used by
the all-in-one function
\code{\link[=power4test]{power4test()}}. Users usually do not
call this function directly, though
developers can use this function to
develop other functions for power
analysis, or to build their own
workflows to do the power analysis.
}
\examples{

# Specify the model

mod <-
"m ~ x
 y ~ m + x"

# Specify the population values

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate the simulated datasets

dats <- sim_data(nrep = 5,
                 model = mod,
                 pop_es = es,
                 n = 100,
                 iseed = 1234)

# Fit the population model to each dataset

fits <- fit_model(dats)

# Combine the results to one object

sim_out_all <- sim_out(data_all = dats,
                       fit = fits)
sim_out_all

# Verify that the elements of fits are set to extra$fit

library(lavaan)
parameterEstimates(fits[[1]])
parameterEstimates(sim_out_all[[1]]$extra$fit)
parameterEstimates(fits[[2]])
parameterEstimates(sim_out_all[[2]]$extra$fit)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power4test_by_n.R
\name{power4test_by_n}
\alias{power4test_by_n}
\alias{c.power4test_by_n}
\alias{as.power4test_by_n}
\alias{print.power4test_by_n}
\title{Power By Sample Sizes}
\usage{
power4test_by_n(
  object,
  n = NULL,
  progress = TRUE,
  ...,
  by_seed = NULL,
  by_nrep = NULL,
  save_sim_all = TRUE
)

\method{c}{power4test_by_n}(
  ...,
  sort = TRUE,
  skip_checking_models = FALSE,
  tolerance_if_std_by_monte_carlo =
    getOption("power4mome.tolerance_if_std_by_monte_carlo", default = 0.01)
)

as.power4test_by_n(original_object)

\method{print}{power4test_by_n}(x, print_all = FALSE, ...)
}
\arguments{
\item{object}{A \code{power4test} object,
or a \code{power4test_by_n} object.
If it is a \code{power4test_by_n} object,
the first element, which is a
\code{power4test} object, will be used
as the value of this argument.}

\item{n}{A numeric vector of sample
sizes for which the simulation will
be conducted.}

\item{progress}{Logical. Whether
progress of the simulation will be
displayed.}

\item{...}{For \code{\link[=power4test_by_n]{power4test_by_n()}},
they are arguments to be passed
to \code{\link[=power4test]{power4test()}}. For \code{\link[=c.power4test_by_n]{c.power4test_by_n()}},
they are \code{\link[=power4test_by_n]{power4test_by_n()}} outputs
to be combined together.
For the \code{print} method of the output
of \code{\link[=power4test_by_n]{power4test_by_n()}}, they are
arguments to be passed to the
\code{print} method of the output of
\code{\link[=power4test]{power4test()}} (\code{\link[=print.power4test]{print.power4test()}}).}

\item{by_seed}{If set to a number,
it will be used to generate the
seeds for each call to \code{\link[=power4test]{power4test()}}.
If \code{NULL}, the default, then seeds
will still be randomly generated but
the results cannot be easily reproduced.}

\item{by_nrep}{If set to a number,
it will be used to generate the
number of replications (\code{nrep}) for
each call to \code{\link[=power4test]{power4test()}}.
If set to a numeric vector of the
same length as \code{n}, then these are
the \code{nrep} values for each of the
calls, allowing for different numbers
of replications for the sample sizes.
If \code{NULL}, the default, then the
original \code{nrep} will be used.
This argument is used by
\code{\link[=x_from_power]{x_from_power()}} for efficiency, and
is rarely used when calling
this function directly.}

\item{save_sim_all}{If \code{FALSE},
the dataset in the
\code{power4test} objects are not saved,
to reduce
the size of the output. Default
is \code{TRUE}.}

\item{sort}{When combining the
objects, whether they will be sorted
by sample sizes. Default is \code{TRUE}.}

\item{skip_checking_models}{Whether
the check of the data generation model
will be checked. Default is \code{TRUE}.
Should be set to \code{FALSE} only when
users are certain the they are based
on the same model, or when the model
is not saved (e.g., \code{save_sim_all}
set to \code{FALSE} when the objects
were generated).
This argument is used by
\code{\link[=x_from_power]{x_from_power()}} for efficiency, and
is rarely used when calling
the \code{c} method directly.}

\item{tolerance_if_std_by_monte_carlo}{The tolerance to be used by \code{\link[=all.equal]{all.equal()}}
when checking population values. If
standardization is conducted using
error variances estimated by
Monte Carlo simulation, then they may
not be exactly the same across replications.
This is the tolerance used to compare
population values, to allow for
minor variation due to simulation.}

\item{original_object}{The object
to be converted to a \code{power4test_by_n}
object.}

\item{x}{The object
to be printed.}

\item{print_all}{If \code{TRUE}, all
elements in \code{x}, that is, the results
of all sample sizes examined, will
be printed. If \code{FALSE}, then only
those of the first sample size
will be printed.}
}
\value{
The function
\code{\link[=power4test_by_n]{power4test_by_n()}} returns a
\code{power4test_by_n} object, which is a
list of \code{power4test} objects, one for
each sample size.

The method \code{\link[=c.power4test_by_n]{c.power4test_by_n()}} returns
a \code{power4test_by_n} object with
all the elements (tests for different
sample sizes) combined.

The function \code{\link[=as.power4test_by_n]{as.power4test_by_n()}} returns
a \code{power4test_by_n} object converted
from the input object.

The \code{print}-method of \code{power4test_by_n}
objects returns the object invisibly.
It is called for its side-effect.
}
\description{
Estimate power for a
set of sample sizes.
}
\details{
The function \code{\link[=power4test_by_n]{power4test_by_n()}}
regenerates
datasets for a set of sample sizes
and does the stored tests in each of
them.

Optionally, it can also be run
on a object with no stored tests.
In this case, additional arguments
must be set to instruct \code{\link[=power4test]{power4test()}}
on the tests to be conducted.

It is usually used to examine the
power over a set of sample sizes.

The \code{c} method of \code{power4test_by_n}
objects
is used to combine tests from different
runs of \code{\link[=power4test_by_n]{power4test_by_n()}}.

The function \code{\link[=as.power4test_by_n]{as.power4test_by_n()}}
is used to convert a \code{power4test}
object to a \code{power4test_by_n}
object, if it is not already one.
Useful when concatenating
\code{power4test} objects with
\code{power4test_by_n} objects.
}
\examples{

# Specify the model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the population values

model_simple_med_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 40,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

out <- power4test_by_n(test_out,
                       n = c(100, 110, 120))
out_reject <- rejection_rates(out)
out_reject

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power_curve_plot.R
\name{plot.power_curve}
\alias{plot.power_curve}
\alias{plot.power4test_by_n}
\alias{plot.power4test_by_es}
\title{Plot a Power Curve}
\usage{
\method{plot}{power_curve}(
  x,
  what = c("ci", "power_curve"),
  main = paste0("Power Curve ", "(Predictor: ", switch(x$predictor, n = "Sample Size", es
    = "Effect Size"), ")"),
  xlab = switch(x$predictor, n = "Sample Size", es = "Effect Size"),
  ylab = "Estimated Power",
  pars_ci = list(),
  type = "l",
  ylim = c(0, 1),
  ci_level = 0.95,
  ...
)

\method{plot}{power4test_by_n}(
  x,
  what = c("ci", "power_curve"),
  main = "Estimated Power vs. Sample Size",
  xlab = "Sample Size",
  ylab = "Estimated Power",
  pars_ci = list(),
  type = "l",
  ylim = c(0, 1),
  ci_level = 0.95,
  ...
)

\method{plot}{power4test_by_es}(
  x,
  what = c("ci", "power_curve"),
  main = paste0("Estimated Power vs. Effect Size / Parameter (", attr(x[[1]],
    "pop_es_name"), ")"),
  xlab = paste0("Effect Size / Parameter (", attr(x[[1]], "pop_es_name"), ")"),
  ylab = "Estiamted Power",
  pars_ci = list(),
  type = "l",
  ylim = c(0, 1),
  ci_level = 0.95,
  ...
)
}
\arguments{
\item{x}{The object to be plotted.
It can be a \code{power_curve} object,
the output of \code{\link[=power_curve]{power_curve()}}. It
can also be the output of
\code{\link[=power4test_by_n]{power4test_by_n()}} or
\code{\link[=power4test_by_es]{power4test_by_es()}}.}

\item{what}{A character vector of
what to include in the
plot. Possible values are
\code{"ci"} (confidence intervals
for the estimated sample size) and
\code{"power_curve"} (the crude power
curve, if available).
The default values depend on the
type of \code{x}.}

\item{main}{The title of the plot.}

\item{xlab, ylab}{The labels for the
horizontal and vertical axes,
respectively.}

\item{pars_ci}{A named list of
arguments to be passed to \code{\link[=arrows]{arrows()}}
to customize the drawing of the
confidence intervals.}

\item{type}{An argument of the
default plot method \code{\link[=plot.default]{plot.default()}}.
Default is \code{"l"}. See \code{\link[=plot.default]{plot.default()}}
for other options.}

\item{ylim}{A two-element numeric
vector of the range of the vertical
axis.}

\item{ci_level}{The level of
confidence of the confidence intervals,
if requested. Default is \code{.95}, denoting
95\%.}

\item{...}{Optional arguments.
Passed to \code{\link[=plot]{plot()}} when drawing
the base plot.}
}
\value{
The \code{plot}-methods return \code{x}
invisibly. They
are called for their side effects.
}
\description{
Plotting the results
in a 'power_curve' object, such as the
estimated power against sample size,
or the results of \code{\link[=power4test_by_n]{power4test_by_n()}}
or \code{\link[=power4test_by_es]{power4test_by_es()}}.
}
\details{
The \code{plot} method of \code{power_curve}
objects currently plots the relation
between estimated power and
the predictor. Other elements
can be requested (see the argument
\code{what}), and they can be customized
individually.
}
\examples{

# Specify the population model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the effect sizes (population parameter values)

model_simple_med_es <-
"
y ~ m: l
m ~ x: m
y ~ x: s
"

# Simulate datasets to check the model

sim_only <- power4test(nrep = 10,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = FALSE)

# By n: Do a test for different sample sizes
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

out1 <- power4test_by_n(sim_only,
                        nrep = 10,
                        test_fun = test_parameters,
                        test_args = list(par = "y~x"),
                        n = c(25, 50, 100),
                        by_seed = 1234,
                        parallel = FALSE,
                        progress = FALSE)

pout1 <- power_curve(out1)
pout1
plot(pout1)

# By pop_es: Do a test for different population values of a model parameter
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

out2 <- power4test_by_es(sim_only,
                             nrep = 10,
                             test_fun = test_parameters,
                             test_args = list(par = "y~x"),
                             pop_es_name = "y ~ x",
                             pop_es_values = c(0, .3, .5),
                             by_seed = 1234,
                             parallel = FALSE,
                             progress = FALSE)

pout2 <- power_curve(out2)
plot(pout2)

}
\seealso{
\code{\link[=power_curve]{power_curve()}},
\code{\link[=power4test_by_n]{power4test_by_n()}}, and
\code{\link[=power4test_by_es]{power4test_by_es()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power4test.R
\name{power4test}
\alias{power4test}
\alias{print.power4test}
\title{Estimate the Power of a Test}
\usage{
power4test(
  object = NULL,
  nrep = NULL,
  ptable = NULL,
  model = NULL,
  pop_es = NULL,
  standardized = TRUE,
  n = NULL,
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  x_fun = list(),
  e_fun = list(),
  process_data = NULL,
  fit_model_args = list(),
  R = NULL,
  ci_type = "mc",
  gen_mc_args = list(),
  gen_boot_args = list(),
  test_fun = NULL,
  test_args = list(),
  map_names = c(fit = "fit"),
  results_fun = NULL,
  results_args = list(),
  test_name = NULL,
  test_note = NULL,
  do_the_test = TRUE,
  sim_all = NULL,
  iseed = NULL,
  parallel = FALSE,
  progress = TRUE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  es1 = c(n = 0, nil = 0, s = 0.1, m = 0.3, l = 0.5, si = 0.141, mi = 0.361, li = 0.51,
    sm = 0.2, ml = 0.4),
  es2 = c(n = 0, nil = 0, s = 0.05, m = 0.1, l = 0.15, sm = 0.075, ml = 0.125),
  es_ind = c("si", "mi", "li"),
  n_std = 1e+05,
  std_force_monte_carlo = FALSE,
  rejection_rates_args = list(collapse = "none", at_least_k = 1, merge_all_tests = FALSE,
    p_adjust_method = "none", alpha = 0.05),
  n_ratio = 1
)

\method{print}{power4test}(
  x,
  what = c("data", "test"),
  digits = 3,
  digits_descriptive = 2,
  data_long = FALSE,
  test_long = TRUE,
  fit_to_all_args = list(),
  ...
)
}
\arguments{
\item{object}{Optional. If set to a
\code{power4test} object, it will be
updated using the value(s) in \code{n},
\code{pop_es}, and/or \code{nrep} if they changed,
or a new test will be conducted and
added to \code{objet}. See the help page
for details. Default is \code{NULL}.}

\item{nrep}{The number of replications
to generate the simulated datasets.
Default is \code{NULL}. Must be set when
called to create a \code{power4test}
object.}

\item{ptable}{The output of
\code{\link[=ptable_pop]{ptable_pop()}}, which is a
\code{ptable_pop} object, representing the
population model. If \code{NULL}, the
default, \code{\link[=ptable_pop]{ptable_pop()}} will be
called to generate the \code{ptable_pop}
object using \code{model} and \code{pop_es}.}

\item{model}{The \code{lavaan} model
syntax of the population model,
to be used by \code{\link[=ptable_pop]{ptable_pop()}}.
See 'Details' of
on how to specify the model.
Ignored if \code{ptable} is
specified.}

\item{pop_es}{The character vector or
multiline string to
specify population effect sizes
(population values of parameters). See
the help page on how to specify this
argument.
Ignored if \code{ptable} is
specified.}

\item{standardized}{Logical. If
\code{TRUE}, the default, variances and
error variances are scaled to ensure
the population variances of the
endogenous variables are close to
one, and hence the effect sizes
(population values) are
standardized effect sizes if the
variances of the continuous exogenous
variables are also equal to one.}

\item{n}{The sample size for each
dataset. Default is 100.}

\item{number_of_indicators}{A named
vector to specify the number of
indicators for each factors. See
the help page on how to set this
argument. Default is \code{NULL} and all
variables in the model syntax are
observed variables.
See the help page on how
to use this argument.}

\item{reliability}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the reliability coefficient
of each set of indicators. Default
is \code{NULL}.
See the help page on how
to use this argument.}

\item{loading_difference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the difference in factor
loadings between neighboring indicators
of each set of indicators. Default
is \code{NULL}, and all indicators of
a factor have
the same factor loadings. If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal.}

\item{reference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to indicate which indicator will
be the first indicator (and so
is the reference indicator, by default).
Default
is \code{NULL}, and for all factors,
the indicator with
the medium loading in a factor is
the first indicator. Has no effect
if loading difference is zero (and
so all indicators have the same
loadings). If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal. Accepted
values are \code{"medium"}, \code{"weakest"},
and \code{"strongest"}.}

\item{x_fun}{The function(s) used to
generate the exogenous variables or
error terms. If
not supplied, or set to \code{list()}, the
default, the variables are generated
from a multivariate normal
distribution. See the help page on how
to use this argument.}

\item{e_fun}{The function(s) used to
generate the error terms of indicators,
if any. If
not supplied, or set to \code{list()}, the
default, the error terms of indicators
are generated
from a multivariate normal
distribution. Specify in the same
way as \code{x_fun}. Refer to the help
page on \code{x_fun} on how to use this
argument.}

\item{process_data}{If not \code{NULL}, it
must be a named list with these
elements: \code{fun} (required), the function
to further process the simulated
data, such as generating missing data using
functions such as \code{\link[mice:ampute]{mice::ampute()}}; \code{args} (optional), a
named list of arguments to be passed
to \code{fun}, except the one for the
source data; \code{sim_data_name} (optional) the
name of the argument to receive the
simulated data (e.g., \code{"data"} for
\code{\link[mice:ampute]{mice::ampute()}}), default to
\code{"data"} if it is not set; \code{processed_data_name}
(optional), the name of the data frame
after being processed by \code{fun},
such as the data frame
with missing data in the output of
\code{fun} (e.g., \code{"amp"} for \code{\link[mice:ampute]{mice::ampute()}}),
if omitted, the output of \code{fun} should
be the data frame with missing data.}

\item{fit_model_args}{A list of the
arguments to be passed to \code{\link[=fit_model]{fit_model()}}
when fitting the
model.
Should be a named list
with names being the names of the
arguments.}

\item{R}{The number of replications
to generate the Monte Carlo or
bootstrapping estimates
for each fit output. No Monte Carlo
nor bootstrapping
estimates will be generated if \code{R}
is set to \code{NULL}.}

\item{ci_type}{The type of
simulation-based confidence
intervals to use. Can be either
\code{"mc"} for Monte Carlo method
(the default) or \code{"boot"} for
nonparametric bootstrapping method.
Relevant for test functions that
make use of estimates generate by
\code{\link[=gen_boot]{gen_boot()}} or \code{\link[=gen_mc]{gen_mc()}}, such as
\code{\link[=test_indirect_effect]{test_indirect_effect()}}.}

\item{gen_mc_args}{A list of
arguments to be passed to
\code{\link[manymome:do_mc]{manymome::do_mc()}} when generating
the Monte Carlo estimates.
Should be a named argument
with names being the names of the
arguments. Used only if
\code{ci_type} is \verb{"mc".}}

\item{gen_boot_args}{A list of
arguments to be passed to
\code{\link[manymome:do_boot]{manymome::do_boot()}} when generating
the bootstrap estimates.
Should be a named argument
with names being the names of the
arguments. Used only if
\code{ci_type} is `"boot".}

\item{test_fun}{A function to do the
test. See 'Details' for the requirement
of this function. There are some
built-in test functions in \code{power4mome},
described in 'Details'.}

\item{test_args}{A list of arguments
to be passed to the \code{test_fun}
function. Default is \code{list()}.}

\item{map_names}{A named character
vector specifying how the content of
the element \code{extra} in
each replication of \code{sim_all} map
to the argument of \code{test_fun}.
Default is \code{c(fit = "fit")},
indicating that the element \code{fit}
in the element \code{extra} is set to
the argument \code{fit} of \code{test_fun}.
That is, for the first replication,
\code{fit = sim_out[[1]]$extra$fit} when
calling \code{test_fun}.}

\item{results_fun}{The function to be
used to extract the test results.
See \code{Details} for the requirements
of this function. Default is \code{NULL},
assuming that the output of
\code{test_fun} can be used directly.}

\item{results_args}{A list of
arguments to be passed to the
\code{results_fun} function. Default is
\code{list()}.}

\item{test_name}{String. The name
of the test. Default is \code{NULL},
and the name will be created from
\code{test_fun}. Note that if \code{sim_out}
is a \code{power4test} object and already
has a test of this name stored, it
will be replaced by the new results.}

\item{test_note}{String. An optional
note for the test, stored in the
attribute \code{test_note} of the output
of \code{\link[=do_test]{do_test()}}. Default is \code{NULL}.}

\item{do_the_test}{If \code{TRUE},
\code{\link[=do_test]{do_test()}} will be called to do the
test specified by \code{test_fun} on the
fit output of each dataset.}

\item{sim_all}{If set to either a
\code{sim_out} object (the output of
\code{\link[=sim_out]{sim_out()}} or a \code{power4test} object
(the output of \code{\link[=power4test]{power4test()}}), the
stored datasets and fit outputs will
be used for doing the test. Setting
\code{object} to the output of \code{\link[=power4test]{power4test()}}
is now the preferred method, but this
argument is kept for backward
compatibility.}

\item{iseed}{The seed for the random
number generator. Default is \code{NULL}
and the seed is not changed. This
seed will be set only once, when
calling \code{\link[=sim_data]{sim_data()}}.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used when calling
other functions, if appropriate.
Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
of each step will be displayed.
Default is \code{TRUE}.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{es1}{Set the
values for each label of the effect
size (population value) for correlations and regression
paths.
Used only if \code{pop_es} is a named
vector or a multiline string.
See the help page on how to specify
this argument.}

\item{es2}{Set the
values for each label of the effect
size (population value) for product term.
Used only if \code{pop_es} is a named
vector or a multiline string.
See the help page on how to specify
this argument.}

\item{es_ind}{The names of labels
denoting the effect size of an
indirect effect. They will be
used to determine the population
values of the component paths along
an indirect path.}

\item{n_std}{The sample size used to
determine the error variances by
simulation when \code{std_force_monte_carlo}
is \code{TRUE}.}

\item{std_force_monte_carlo}{Logical.
If \code{FALSE}, the default,
standardization is done analytically
if the model has no product terms,
and by simulation if the model has
product terms. That is, error variances
required to ensure implied variances equal
to one are determined by simulation.
If \code{TRUE}, simulation
will be used whether the model has
product terms or not. Always fall
back to simulation if
analytical standardization failed.}

\item{rejection_rates_args}{Default
argument values to be used when
\code{\link[=rejection_rates]{rejection_rates()}} is called. Can
be overriden when calling
\code{\link[=rejection_rates]{rejection_rates()}}.}

\item{n_ratio}{If the model is a
multigroup model, and \code{n} is a single
number, this should be a numeric
vector used to determine the sample
size for each group. For example,
for a two-group model, if \code{n} is 100
and \code{n_ratio} is \code{c(1, 0.5)}, then
the sample sizes for the two groups
are 100 and 50, respectively. If
equal to 1, then all groups have the
same sample size.}

\item{x}{The object
to be printed.}

\item{what}{A string vector of
what to print, \code{"data"} for
simulated data and \code{"test"} for
stored test(s). Default is
\code{c("data", "test")}.}

\item{digits}{The numbers of digits
displayed after the decimal.}

\item{digits_descriptive}{The
number of digits displayed after
the decimal for the descriptive
statistics table.}

\item{data_long}{If \code{TRUE}, detailed
results will be printed when printing
the simulated data.}

\item{test_long}{If \code{TRUE}, detailed
results will be printed when printing
test(s).}

\item{fit_to_all_args}{A named list
of arguments to be passed to
\code{\link[lavaan:sem]{lavaan::sem()}} when the model is
fitted to a sample combined from
all samples stored.}

\item{...}{Optional arguments to
be passed to other print methods}
}
\value{
An object of the class \code{power4test},
which is a list with two elements:
\itemize{
\item \code{sim_all}: The output of \code{\link[=sim_out]{sim_out()}}.
\item \code{test_all}: A named list of the
output of \code{\link[=do_test]{do_test()}}. The names
are the values of \code{test_name}.
This list can have more than one
test because a call to
\code{\link[=power4test]{power4test()}} can add new tests
to a \code{power4test} object.
}

The \code{print} method of \code{power4test}
returns \code{x} invisibly. Called for
its side effect.
}
\description{
An all-in-one function
that receives a model specification,
generates datasets, fits a model, does
the target test, and returns the test
results.
}
\details{
The function \code{\link[=power4test]{power4test()}} is an
all-in-one function for
estimating the power of a test for
a model, given the sample size
and effect sizes (population values
of model parameters).
}
\section{Workflow}{
This is the workflow:
\itemize{
\item If \code{object} is an output
of the output
of a previous call to
\code{\link[=power4test]{power4test()}} with \code{do_the_test}
set to \code{FALSE} and so has only the
model and the simulated data,
the following steps
will be skipped and go directly to
doing the test.
\itemize{
\item Call \code{\link[=sim_data]{sim_data()}} to determine
the population model and
generate the datasets, using
arguments such as \code{model} and
\code{pop_es}.
\item Call \code{\link[=fit_model]{fit_model()}} to fit a
model to each of the datasets,
which is the population model
by default.
\item If \code{R} is not \code{NULL} and
\code{ci_type = "mc"}, call
\code{\link[=gen_mc]{gen_mc()}} to generate Monte
Carlo estimates using
\code{\link[manymome:do_mc]{manymome::do_mc()}}. The estimates
can be used by supported functions
such as \code{\link[=test_indirect_effect]{test_indirect_effect()}}.
\item If \code{R} is not \code{NULL} and
\code{ci_type = "boot"}, call
\code{\link[=gen_boot]{gen_boot()}} to generate
bootstrap estimates using
\code{\link[manymome:do_boot]{manymome::do_boot()}}. The estimates
can be used by supported functions
such as \code{\link[=test_indirect_effect]{test_indirect_effect()}}.
\item Merge the results into a
\code{sim_out} object by calling
\code{\link[=sim_out]{sim_out()}}.
\item If \code{do_the_test} is \code{FALSE},
skip the remaining steps and
return a \code{power4test} object,
which contains only the data
generated and optionally the
Monte Carlo or bootstrap
estimates.
}
\item If \code{do_the_test} is \code{TRUE}, do
the test.
\itemize{
\item \code{\link[=do_test]{do_test()}} will be called to do
the test in the fit output of
each dataset.
}
\item Return a \code{power4test} object which
include the output of \code{sim_out}
and, if \code{do_the_test} is \code{TRUE},
the output of \code{\link[=do_test]{do_test()}}.
}

This function is to be used when
users are interested only in the
power of one or several tests on a
particular aspect of the model, such
as a parameter, given a specific
effect sizes and sample sizes.

Detailed description on major
arguments can be found in sections
below.

NOTE: The technical internal workflow of
of \code{\link[=power4test]{power4test()}} can be found in
this page: \url{https://sfcheung.github.io/power4mome/articles/power4test_workflow.html}.
}

\section{Updating a Condition}{
The function \code{\link[=power4test]{power4test()}} can also be used to
update a condition when only some
selected aspects is to be changed.

For example,
instead of calling this function with
all the arguments set just to change
the sample size, it can be called
by supplying an existing
\code{power4test} object and set only
\code{n} to a new sample size. The data
and the tests will be updated
automatically. See the examples for
an illustration.
}

\section{Adding Another Test}{
The function \code{\link[=power4test]{power4test()}} can also be used to
add a test to the output from a
previous call to \code{\link[=power4test]{power4test()}}.

For example, after simulating the
datasets and doing one test,
the output can be set to \code{object}
of \code{\link[=power4test]{power4test()}}, and set only
\code{test_fun} and, optionally,
\code{test_fun_args} to do one more test
on the generated datasets. The output
will be the original
object with the results of the new
test added. See the examples for
an illustration.
}

\section{Model Fitting Arguments ('fit_model_args')}{
For power analysis, usually, the
population model (\code{model}) is to be
fitted, and there is no need to
set \code{fit_model_args}.

If power analysis is to be conducted
for fitting a model that is not the
population model, of if non-default
settings are desired when fitting
a model, then the argument \code{fit_model_args}
needed to be set to customize the
call to \code{\link[=fit_model]{fit_model()}}.

For example,
users may want to examine the power
of a test when a misspecified model
is fitted, or the power of a test
when MLR is used as the estimator
when calling \code{\link[lavaan:sem]{lavaan::sem()}}.

See the help page of \code{\link[=fit_model]{fit_model()}}
for some examples.
}

\section{Specify the Population Model by 'model'}{
\subsection{Single-Group Model}{

For a single-group model, \code{model}
should be a \code{lavaan} model syntax
string of the \emph{form} of the model.
The population values of the model
parameters are to be determined by
\code{pop_es}.

If the model has latent factors,
the syntax in \code{model} should specify
only the \emph{structural model} for the
\emph{latent factors}. There is no
need to specify the measurement
part. Other functions will generate
the measurement part on top of this
model.

For example, this is a simple mediation
model:

\preformatted{"m ~ x
 y ~ m + x"}

Whether \code{m}, \code{x}, and \code{y} denote
observed variables or latent factors
are determined by other functions,
such as \code{\link[=power4test]{power4test()}}.
}

\subsection{Multigroup Model}{

Because the model is the population
model, equality constraints are
irrelevant and the model syntax
specifies only the \emph{form} of the
model. Therefore, \code{model} is
specified as in the case of single
group models.
}
}

\section{Specify 'pop_es' Using Named Vectors}{
The argument \code{pop_es} is for specifying
the population values of model
parameters. This section describes
how to do this using named vectors.
\subsection{Single-Group Model}{

If \code{pop_es} is specified by a named
vector, it must follow the convention
below.
\itemize{
\item The names of the vectors are
\code{lavaan} names for the selected
parameters. For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
\item Alternatively, the names can be
either \code{".beta."} or \code{".cov."}.
Use \code{".beta."} to set the default
values for all regression coefficients.
Use \code{".cov."} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
\item The names can also be of this form:
\code{".ind.(<path>)"}, whether \verb{<path>}
denote path in the model. For
example, \code{".ind.(x->m->y)"} denotes
the path from \code{x} through \code{m} to
\code{y}. Alternatively, the \code{lavaan}
symbol \code{~} can also be used:
\code{".ind.(y~m~x)"}. This form is used
to set the indirect effect (standardized,
by default) along this path. The
value for this name will override
other settings.
\item If using \code{lavaan} names, can
specify more than one parameter
using \code{+}. For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
\item The value of each element can be
the label for the effect size: \code{n}
for nil, \code{s} for small, \code{m} for
medium, and \code{l} for large. The
value for each label is determined
by \code{es1} and \code{es2}. See the section
on specifying these two arguments.
\item The value of \code{pop_es} can also be
set to a value, but it must be
quoted as a string, such as \code{"y ~ x" = ".31"}.
}

This is an example:

\preformatted{c(".beta." = "s",
  "m1 ~ x" = "-m",
  "m2 ~ m1" = "l",
  "y ~ x:w" = "s")}

In this example,
\itemize{
\item All regression coefficients are
set to small (\code{s}) by default, unless
specified otherwise.
\item The path from \code{x} to \code{m1} is
set to medium and negative (\code{-m}).
\item The path from \code{m1} to \code{m2} is set
to large (\code{l}).
\item The coefficient of the product
term \code{x:w} when predicting \code{y} is
set to small (\code{s}).
}
\subsection{Indirect Effect}{

When setting an indirect effect to
a symbol (default: \code{"si"}, \code{"mi"},
\code{"li"}, with \code{"i"} added to differentiate
them from the labels for a direct path),
the corresponding value is used to
determine the population values of
\emph{all} component paths along the pathway.
All the values are assumed to be equal.
Therefore, \code{".ind.(x->m->y)" = ".20"}
is equivalent to setting \code{m ~ x}
and \code{y ~ m} to the square root of
.20, such that the corresponding
indirect effect is equal to the
designated value.

This behavior, though restricted,
is for quick manipulation of the
indirect effect. If different values
along a pathway, set the value for
each path directly.

Only nonnegative value is supported.
Therefore, \code{".ind.(x->m->y)" = "-si"}
and \code{".ind.(x->m->y)" = "-.20"} will
throw an error.
}

}

\subsection{Multigroup Model}{

The argument \code{pop_es} also supports multigroup
models.

For \code{pop_es}, instead of
named vectors, named \emph{list} of
named vectors should be used.
\itemize{
\item The names are the parameters, or
keywords such as \code{.beta.} and
\code{.cov.}, like specifying the
population values for a single
group model.
\item The elements are character vectors.
If it has only one element (e.g.,
a single string), then it is the
the population value for all groups.
If it has more than one element
(e.g., a vector of three strings),
then they are the population values
of the groups. For a model of \emph{k}
groups, each vector must have
either \emph{k} elements or one element.
}

This is an example:

\preformatted{list("m ~ x" = "m",
     "y ~ m" = c("s", "m", "l"))}

In this model, the population value
of the path \code{m ~ x} is medium (\code{m}) for
all groups, while the population
values for the path \code{y ~ m} are
small (\code{s}), medium (\code{m}), and large (\code{l}),
respectively.
}
}

\section{Specify 'pop_es' Using a Multiline String}{
When setting the argument \code{pop_es},
instead of using a named vector
or named list for
\code{pop_es}, the population values of
model parameters can also be
specified using a multiline string,
as illustrated below, to be parsed
by \code{\link[=pop_es_yaml]{pop_es_yaml()}}.
\subsection{Single-Group Model}{

This is an example of the multiline string
for a single-group model:

\preformatted{y ~ m: l
m ~ x: m
y ~ x: nil}

The string must follow this format:
\itemize{
\item Each line starts with \verb{tag:}.
\itemize{
\item \code{tag} can be the name of a
parameter, in \code{lavaan} model
syntax format.
\itemize{
\item For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
}
\item A tag in \code{lavaan} model syntax can
specify more than one parameter
using \code{+}.
\itemize{
\item For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
}
\item Alternatively, the \code{tag} can be
either \code{.beta.} or \code{.cov.}.
\itemize{
\item Use \code{.beta.} to set the default
values for all regression coefficients.
\item Use \code{.cov.} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
}
}
\item After each tag is the value of the
population value:

-\code{nil} for nil (zero),
\itemize{
\item \code{s} for small,
\item \code{m} for medium, and
\item \code{l} for large.
\item \code{si}, \code{mi}, and \code{li} for
small, medium, and large a
standardized indirect effect,
respectively.
}

Note: \code{n} \emph{cannot} be used in this mode.

The
value for each label is determined
by \code{es1} and \code{es2} as described
in \code{\link[=ptable_pop]{ptable_pop()}}.
\itemize{
\item The value can also be
set to a numeric value, such as
\code{.30} or \code{-.30}.
}
}

This is another example:

\preformatted{.beta.: s
y ~ m: l}

In this example, all regression
coefficients are \code{small}, while
the path from \code{m} to \code{y} is large.
}

\subsection{Multigroup Model}{

This is an example of the string
for a multigroup model:

\preformatted{y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil}

The format is similar to that for
a single-group model. If a parameter
has the same value for all groups,
then the line can be specified
as in the case of a single-group
model: \code{tag: value}.

If a parameter has different
values across groups, then it must
be in this format:
\itemize{
\item A line starts with the tag, followed
by two or more lines. Each line
starts with a hyphen \code{-} and the
value for a group.
}

For example:

\preformatted{m ~ x:
  - nil
  - s}

This denotes that the model has
two groups. The values of the path
from \code{x} to \code{m} for the two
groups are 0 (\code{nil}) and
small (\code{s}), respectively.

Another equivalent way to specify
the values are using \verb{[]}, on
the same line of a tag.

For example:

\preformatted{m ~ x: [nil, s]}

The number of groups is inferred
from the number of values for
a parameter. Therefore, if a tag
has more than one value, each tag
must has the same number of value,
or only one value.

The tag \code{.beta.} and \code{.cov.} can
also be used for multigroup models.
}

\subsection{Which Approach To Use}{

Note that using named vectors or
named lists is more reliable. However,
using a multiline string is
more user-friendly. If this method
failed, please use named vectors or
named list instead.
}

\subsection{Technical Details}{

The multiline string is parsed by \code{\link[yaml:read_yaml]{yaml::read_yaml()}}.
Therefore, the format requirement
is actually that of YAML. Users
knowledgeable of YAML can use other
equivalent way to specify the string.
}
}

\section{Set the Values for Effect Size Labels ('es1' and 'es2')}{
The vector \code{es1} is for correlations,
regression coefficients, and
indirect effect, and the
vector \code{es2} is for for standardized
moderation effect, the coefficients
of a product term. These labels
are to be used in interpreting
the specification in \code{pop_es}.
}

\section{Set 'number_of_indicators' and 'reliability'}{
The arguments \code{number_of_indicators}
and \code{reliability} are used to
specify the number of indicators
(e.g., items) for each factor,
and the population reliability
coefficient of each factor,
if the variables in the model
syntax are latent variables.

Optionally, \code{loading_difference} can
be used to generate indicators with
unequal standardized factor loadings,
and \code{reference} can be used to specify
the indicator with the medium,
strongest, or weakest standardized
factor loading as the first indicator,
which is used as the reference indicator
in \code{lavaan}.
\subsection{Single-Group Model}{

If a variable in the model is to be
replaced by indicators in the generated
data, set
\code{number_of_indicators} to a named
numeric vector. The names are the
variables of variables with
indicators, as appeared in the
\code{model} syntax. The value of each
name is the number of indicators.

The
argument \code{reliability} should then be
set a named numeric vector (or list,
see the section on multigroup models)
to specify the population reliability
coefficient ("omega") of each set of
indicators. The population standardized factor
loadings are then computed to ensure
that the population reliability
coefficient is of the target value.

These are examples for a single group
model:

\preformatted{number of indicator = c(m = 3, x = 4, y = 5)}

The numbers of indicators for \code{m},
\code{x}, and \code{y} are 3, 4, and 5,
respectively.

\preformatted{reliability = c(m = .90, x = .80, y = .70)}

The population reliability
coefficients of \code{m}, \code{x}, and \code{y} are
.90, .80, and .70, respectively.
}

\subsection{Multigroup Models}{

Multigroup models are supported.
The number of groups is inferred
from \code{pop_es} (see the help page
of \code{\link[=ptable_pop]{ptable_pop()}}), or directly
from \code{ptable}.

For a multigroup model, the number
of indicators for each variable
must be the same across groups.

However, the population reliability
coefficients can be different
across groups. For a multigroup model
of \emph{k} groups,
with one or more population reliability
coefficients differ across groups,
the argument \code{reliability} should be
set to a named list. The names are
the variables to which the population
reliability coefficients are to be
set. The element for each name is
either a single value for the common
reliability coefficient, or a
numeric vector of the reliability
coefficient of each group.

This is an example of \code{reliability}
for a model with 2 groups:

\preformatted{reliability = list(x = .80, m = c(.70, .80))}

The reliability coefficients of \code{x} are
.80 in all groups, while the
reliability coefficients of \code{m} are
.70 in one group and .80 in another.
}

\subsection{Equal Numbers of Indicators and/or Reliability Coefficients}{

If all variables in the model has
the same number of indicators,
\code{number_of_indicators} can be set
to one single value.

Similarly, if all sets of indicators
have the same population reliability
in all groups, \code{reliability} can also
be set to one single value.
}
}

\section{Specify The Distributions of Exogenous Variables Or Error Terms Using 'x_fun'}{
By default, variables and error
terms are generated
from a multivariate normal distribution.
If desired, users can supply the
function used to generate an exogenous
variable and error term by setting \code{x_fun} to
a named list.

The names of the list are the variables
for which a user function will be used
to generate the data.

Each element of the list must also
be a list. The first element of this
list, can be unnamed, is the
function to be used. If other arguments
need to be supplied, they should be
included as named elements of this list.

For example:

\preformatted{x_fun = list(x = list(power4mome::rexp_rs),
             w = list(power4mome::rbinary_rs,
                      p1 = .70)))}

The variables \code{x} and \code{w} will be
generated by user-supplied functions.

For \code{x}, the function is
\code{power4mome::rexp_rs}. No additional
argument when calling this function.

For \code{w}, the function is
\code{power4mome::rbinary_rx}. The argument
\code{p1 = .70} will be passed to this
function when generating the values
of \code{w}.

If a variable is an endogenous
variable (e.g., being predicted by
another variable in a model), then
\code{x_fun} is used to generate its
\emph{error term}. Its implied population
distribution may still be different
from that generate by \code{x_fun} because
the distribution also depends on the
distribution of other variables
predicting it.

These are requirements for the
user-functions:
\itemize{
\item They must return a numeric vector.
\item They mush has an argument \code{n} for
the number of values.
\item The \emph{population} mean and standard
deviation of the generated values
must be 0 and 1, respectively.
}

The package \code{power4mome} has
helper functions for generating
values from some common nonnormal
distributions and then scaling them
to have population mean and standard
deviation equal to 0 and 1 (by default), respectively.
These are some of them:
\itemize{
\item \code{\link[=rbinary_rs]{rbinary_rs()}}.
\item \code{\link[=rexp_rs]{rexp_rs()}}.
\item \code{\link[=rbeta_rs]{rbeta_rs()}}.
\item \code{\link[=rlnorm_rs]{rlnorm_rs()}}.
\item \code{\link[=rpgnorm_rs]{rpgnorm_rs()}}.
}

To use \code{x_fun}, the variables must
have zero covariances with other
variables in the population. It is
possible to generate nonnormal
multivariate data but we believe this
is rarely needed when estimating
power \emph{before} having the data.
}

\section{Major Test-Related Arguments}{
\subsection{The test function (test_fun)}{

The function set to \code{test_fun},
the \emph{test function}, usually
should work
on the output of \code{\link[lavaan:sem]{lavaan::sem()}},
\code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}, or
\code{\link[stats:lm]{stats::lm()}}, but can also be a
function that works on the output
of the function set to \code{fit_function}
when calling \code{\link[=fit_model]{fit_model()}} or
\code{\link[=power4test]{power4test()}} (see \code{fit_model_args}).

The function has two default
requirements.

First, it has an argument \code{fit}, to
be set to the output of
\code{\link[lavaan:sem]{lavaan::sem()}} or another output
stored in the element \code{extra$fit} of
a replication in the \code{sim_all}
object. (This requirement can be
relaxed; see the section on \code{map_names}.)

That is, the function definition
should be of this from: \code{FUN(fit, ...)}. This is the form of all
\verb{test_*} functions in \code{power4mome}.

If other arguments are to be passed
to the test function, they can be set
to \code{test_args} as a named list.

Second, the test function must
returns an output that (a) can be
processed by the results function
(see below), or (b) is of
the required format for the output of
a results function (see the next
section). If it already returns
an output of the required format,
then there is no need to set the
results function.
}

\subsection{The results function (results_fun)}{

The test results will be extracted
from the output of \code{test_fun} by the
function set to \code{results_fun},
the \emph{results function}. If
the \code{test_fun} already returns an
output of the expected format
(see below), then set \code{results_fun}
to \code{NULL}, the default. The output
of \code{test_fun} will be used for
estimating power.

The function set to \code{results_fun}
must accept the output of \code{test_fun},
as the first argument, and return a
named list (which can be a data frame)
or a named vector with some
of the following
elements:
\itemize{
\item \code{est}: Optional. The estimate of a
parameter, if applicable.
\item \code{se}: Optional. The standard error
of the estimate, if applicable.
\item \code{cilo}: Optional. The lower limit of the
confidence interval, if applicable.
\item \code{cihi}: Optional. The upper limit of the
confidence interval, if applicable.
\item \code{sig}: Required. If \code{1}, the test is
significant. If \code{0}, the test is not
significant. If the test cannot be
done for any reason, it should be
\code{NA}.
}

The results can then be used to
estimate the power or Type I
error of the test.

For example, if
the null hypothesis is false, then
the proportion of significant, that
is, the mean of the values of \code{sig}
across replications, is the power.
}

\subsection{Built-in test functions}{

The package \code{power4mome} has some ready-to-use
test functions:
\itemize{
\item \code{\link[=test_indirect_effect]{test_indirect_effect()}}
\item \code{\link[=test_cond_indirect]{test_cond_indirect()}}
\item \code{\link[=test_cond_indirect_effects]{test_cond_indirect_effects()}}
\item \code{\link[=test_moderation]{test_moderation()}}
\item \code{\link[=test_index_of_mome]{test_index_of_mome()}}
\item \code{\link[=test_parameters]{test_parameters()}}
}

Please refer to their help pages
for examples.
}

\subsection{The argument map_names}{

This argument is for developers using
a test function that has a different
name for the argument of the fit
object (\code{"fit"}, the default).

If \code{test_fun} is set to a function
that works on an output of, say,
\code{\link[lavaan:sem]{lavaan::sem()}} but the argument name
for the output is not \code{fit}, the
mapping can be changed by
\code{map_names}.

For example,
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}
receives an output of \code{\link[lavaan:sem]{lavaan::sem()}}
and reports the test results of model
parameters. However, the argument
name for the \code{lavaan} output is
\code{object.} To instruct \code{\link[=do_test]{do_test()}} to
do the test correctly when setting
\code{test_fun} to
\code{lavaan::parameterEstimates}, add
\code{map_names = c(object = "fit")}. The
element \code{fit} in a replication will
then set to the argument \code{object} of
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}.
}
}

\examples{

# Specify the model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the population values

model_simple_med_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

# Set nrep to a large number in real analysis, such as 400
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

out <- power4test(nrep = 10,
                  model = model_simple_med,
                  pop_es = model_simple_med_es,
                  n = 100,
                  test_fun = test_parameters,
                  test_args = list(pars = "m~x"),
                  iseed = 1234,
                  parallel = FALSE,
                  progress = TRUE)

print(out,
      test_long = TRUE)

# Change the sample size

out1 <- power4test(out,
                   n = 200,
                   iseed = 1234,
                   parallel = FALSE,
                   progress = TRUE)

print(out1,
      test_long = TRUE)

# Add one more test

out2 <- power4test(out,
                   test_fun = test_parameters,
                   test_args = list(pars = "y~x"),
                   parallel = FALSE,
                   progress = TRUE)

print(out2,
      test_long = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_mc.R
\name{gen_mc}
\alias{gen_mc}
\title{Generate Monte Carlo Estimates}
\usage{
gen_mc(
  fit_all,
  R = 100,
  ...,
  iseed = NULL,
  parallel = FALSE,
  progress = FALSE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  compute_implied_stats = FALSE,
  cl = NULL
)
}
\arguments{
\item{fit_all}{The output of
\code{\link[=fit_model]{fit_model()}} or an object of the
class \code{fit_out}.}

\item{R}{The number of replications
to generate the Monte Carlo estimates
for each fit output.}

\item{...}{Optional arguments to be
passed to \code{\link[manymome:do_mc]{manymome::do_mc()}} when
generating the Monte Carlo estimates.}

\item{iseed}{The seed for the random
number generator. Default is \code{NULL}
and the seed is not changed.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to generate
Monte Carlo estimates for the
fit outputs. Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
will be displayed.
Default is `FALSE.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{compute_implied_stats}{Whether
implied statistics are computed
in each Monte Carlo replication. Usually
not needed and so default to \code{FALSE}.}

\item{cl}{A cluster, such as one created
by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}. If \code{NULL},
a cluster will be created, but will be
stopped on exit. If set to an existing
cluster, it will not be stopped when
the function exits; users need to
stop it manually.}
}
\value{
An \code{mc_list} object, which is a list
of the output of \code{\link[manymome:do_mc]{manymome::do_mc()}}.
}
\description{
Get a list of the output
of \code{\link[lavaan:sem]{lavaan::sem()}} and generate
Monte Carlo estimates of model
parameters.
}
\details{
The function \code{\link[=gen_mc]{gen_mc()}} simply calls
\code{\link[manymome:do_mc]{manymome::do_mc()}}
on each output of
\code{\link[lavaan:sem]{lavaan::sem()}} in \code{fit_all}. The
simulated
estimates can then be used to test
effects such as indirect effects,
usually by functions from the
\code{manymome} package, such as
\code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}.
}
\section{The role of \code{gen_mc()}}{
This function is used by the
all-in-one function \code{\link[=power4test]{power4test()}}.
Users usually do not call this
function directly, though
developers can use this function to
customize the workflow of the
power analysis.
}

\examples{

# Specify the population model

mod <-
"m ~ x
 y ~ m + x"

# Specify the effect sizes (population parameter values)

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate several simulated datasets

data_all <- sim_data(nrep = 5,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     iseed = 1234)

# Fit the population model to each datasets

fit_all <- fit_model(data_all)

# Generate Monte Carlo estimates for each replication

mc_all <- gen_mc(fit_all,
                 R = 100,
                 iseed = 4567)

mc_all

}
\seealso{
See \code{\link[=power4test]{power4test()}} for
the all-in-one function that uses
this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_indirect_effect.R
\name{test_indirect_effect}
\alias{test_indirect_effect}
\title{Test an Indirect Effect}
\usage{
test_indirect_effect(
  fit = fit,
  x = NULL,
  m = NULL,
  y = NULL,
  mc_ci = TRUE,
  mc_out = NULL,
  boot_ci = FALSE,
  boot_out = NULL,
  check_post_check = TRUE,
  test_method = NULL,
  ...,
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}.}

\item{x}{The name of the \code{x}-variable,
the predictor.}

\item{m}{A character vector of the
name(s) of mediator(s). The path
moves from the first mediator in the
vector to the last mediator in the
vector. Can be \code{NULL} and the path
is a direct path without mediator.}

\item{y}{The name of the \code{y}-variable,
the outcome variable.}

\item{mc_ci}{Logical. If \code{TRUE}, the
default, Monte Carlo confidence
intervals will be formed. This argument
and \code{boot_ci} cannot be both \code{TRUE}.}

\item{mc_out}{The pre-generated
Monte Carlo estimates generated by
\link[manymome:do_mc]{manymome::do_mc}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{boot_ci}{Logical. If \code{TRUE},
the default, nonparametric bootstrap
confidence intervals will be formed.
This argument
and \code{mc_ci} cannot be both \code{TRUE}.}

\item{boot_out}{The pre-generated
bootstrap estimates generated by
\link[manymome:do_boot]{manymome::do_boot}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{test_method}{The method to do
the test. If \code{"ci"}, then the
confidence interval (e.g., bootstrap
confidence interval) will be used to
do the test. If
\code{"pvalue"}, then asymmetric \emph{p}-value
by Asparouhov & Muthén (2021) will
be used to do the test, and the
confidence interval will not be
computed. If \code{NULL}, its value will
be set to \code{"pvalue"} if the number of
simulated/bootstrap samples (\code{R}) is
a value that is supported by the
method by Boos and Zhang (2000), and
set to \code{"ci"} otherwise.}

\item{...}{Additional arguments to
be passed to \code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}.}

\item{fit_name}{The name of the
model fit object to be extracted.
Default is \code{"fit"}. Used only when
more than one model is fitted in
each replication. This should be
the name of the model on which the
test is to be conducted.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
a named numeric vector with the
following elements:
\itemize{
\item \code{est}: The mean of the estimated
indirect effect across datasets.
\item \code{cilo} and \code{cihi}: The means of the
lower and upper limits of the
confidence interval (95\% by
default), respectively.
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
}
}
\description{
Test an indirect effect
for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing an
indirect effect, by setting it
to the \code{test_fun} argument.

It uses \code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}
to do the test. It can be used on
models fitted by \code{\link[lavaan:sem]{lavaan::sem()}}
or fitted by a sequence of calls
to \code{\link[stats:lm]{stats::lm()}}, although only
nonparametric bootstrap confidence
interval is supported for models
fitted by regression using
\code{\link[stats:lm]{stats::lm()}}.
}
\examples{

# Specify the model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the population values

model_simple_med_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Simulate the data

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the test in each replication

test_ind <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        mc_ci = TRUE))
print(test_ind,
      test_long = TRUE)

}
\references{
Asparouhov, A., & Muthén, B. (2021). Bootstrap p-value computation.
Retrieved from https://www.statmodel.com/download/FAQ-Bootstrap\%20-\%20Pvalue.pdf
}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/x_from_power_for_x_summary.R
\name{summary.x_from_power}
\alias{summary.x_from_power}
\alias{summary.n_region_from_power}
\alias{print.summary.x_from_power}
\alias{print.summary.n_region_from_power}
\title{Summarize 'x_from_power' Results}
\usage{
\method{summary}{x_from_power}(object, ...)

\method{summary}{n_region_from_power}(object, ...)

\method{print}{summary.x_from_power}(x, digits = 3, ...)

\method{print}{summary.n_region_from_power}(x, digits = 3, ...)
}
\arguments{
\item{object}{An \code{x_from_power}-class
object, such as the output of
\code{\link[=x_from_power]{x_from_power()}}, or an object of the
class \code{n_region_from_power}, such as
the output of
\code{\link[=n_region_from_power]{n_region_from_power()}}.}

\item{...}{Additional arguments.
Not used for now.}

\item{x}{The output of
\code{\link[=summary.x_from_power]{summary.x_from_power()}}, the
\code{summary} method of
an \code{x_from_power} object,
which is the output of
\code{\link[=x_from_power]{x_from_power()}}, or the output
of \code{\link[=summary.n_region_from_power]{summary.n_region_from_power()}},
the \code{summary} method of an
\code{n_region_from_power} object (the
output of \code{\link[=n_region_from_power]{n_region_from_power()}}).}

\item{digits}{The number of digits
after the decimal when printing
the results.}
}
\value{
The \code{summary} method for
\code{x_from_power} objects returns an
object of the class
\code{summary.x_from_power}, which is
simply the output of \code{\link[=x_from_power]{x_from_power()}},
with a \code{print} method dedicated for
detailed summary. Please refer
to \code{\link[=x_from_power]{x_from_power()}} for the contents.

The \code{print}-method of \code{summary.x_from_power}
objects returns the object \code{x}
invisibly.
It is called for its side effect.

The \code{print}-method of \code{summary.n_region_from_power}
objects returns the object \code{x}
invisibly.
It is called for its side effect.
}
\description{
The summary method of
the output of \code{\link[=x_from_power]{x_from_power()}}.
}
\details{
The \code{summary} method simply prepares the
results of \code{\link[=x_from_power]{x_from_power()}}
to be printed in details.
}
\examples{

# Specify the population model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

# Generate the datasets

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 2345)

# Do a test

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(pars = "m~x"))

# Determine the sample size with a power of .80 (default)

power_vs_n <- x_from_power(test_out,
                           x = "n",
                           progress = TRUE,
                           target_power = .80,
                           final_nrep = 5,
                           max_trials = 1,
                           seed = 1234)
summary(power_vs_n)

}
\seealso{
\code{\link[=x_from_power]{x_from_power()}},
\code{\link[=n_region_from_power]{n_region_from_power()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_pars.R
\name{test_parameters}
\alias{test_parameters}
\alias{find_par_names}
\title{Test All Free Parameters}
\usage{
test_parameters(
  fit = fit,
  standardized = FALSE,
  pars = NULL,
  op = NULL,
  remove.nonfree = TRUE,
  check_post_check = TRUE,
  exclude_var = FALSE,
  compare_groups = FALSE,
  ...,
  omnibus = c("no", "all_sig", "at_least_one_sig", "at_least_k_sig"),
  at_least_k = 1,
  p_adjust_method = "none",
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)

find_par_names(object, fit_name = "fit")
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}},
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}}.}

\item{standardized}{Logical. If \code{TRUE},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}} will
be used. Can be used only with models
fitted by \code{lavaan}.}

\item{pars}{Optional. If set to
a character vector, only parameters
with \code{test_label} equal to values in
\code{pars} will be returned. See the
help page
on valid names.}

\item{op}{Optional. If set to a
character vector, only parameters with
operators (e.g., \code{"~"}, \code{"=~"}) will
be returned. If both \code{pars} and \code{op}
are specified, only parameters meeting
\emph{both} requirements will be returned.}

\item{remove.nonfree}{Logical. If
\code{TRUE}, the default, only free
parameters will be returned. Ignored
if \code{standardized} is \code{TRUE} or
if the model is not fitted by
\code{lavaan}.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{exclude_var}{Logical. If \code{TRUE},
exclude variances and error variances
from the test.}

\item{compare_groups}{Logical. If \code{TRUE},
the likelihood ratio test (by \code{\link[lavaan:lavTestLRT]{lavaan::lavTestLRT()}})
will be used to test the pairwise-equality
constraints for all selected free
parameters that appear in all groups.
These tests will be reported instead of
the tests for individual parameters.}

\item{...}{Additional arguments to
be passed to \code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}},
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}}.}

\item{omnibus}{If \code{"no"}, the default,
then the test results for all paths
are stored. If \code{"all_sig"}, then
only one row of test is stored, and
the test is declared significant if
\emph{all} paths are significant. If
\code{"at_least_one_sig"}, then only
one row of test is stored, and the
test is declared significant if
at least one of the paths is
significant. If \code{"at_least_k_sig"},
then only one row of test is stored,
and the test is declared significant
if at least \code{k} of the paths is
significant, \code{k} determined by the
argument \code{at_least_k}. If \code{"total"},
then the total indirect effect will
be tested.}

\item{at_least_k}{The minimum number
of paths required to be significant
for the omnibus test to be considered
significant. Used when
\code{omnibus} is \code{"at_least_k_sig"}.}

\item{p_adjust_method}{The method to be
passed to \code{\link[=p.adjust]{p.adjust()}} to adjust the
\emph{p}-values when testing the effects.
Default is \code{"none"} and the \emph{p}-values
will not be adjusted. The unadjusted
\emph{p}-values will be stored in the column
\code{pvalue_org}.}

\item{fit_name}{The name of the fit
results for which the parameter names
will be displayed. Default is \code{"fit"}.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}

\item{object}{A \code{power4test} object.}
}
\value{
In its normal usage, it returns
the output returned by
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}},
with the following modifications:
\itemize{
\item \code{est}: The parameter estimates,
even if standardized estimates
are requested (not \code{est.std}).
\item \code{cilo} and \code{cihi}: The
lower and upper limits of the
confidence interval (95\% by
default), respectively (not
\code{ci.lower} and \code{ci.upper}).
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
\item \code{test_label}: A column of labels
generated by
\code{\link[lavaan:lav_partable]{lavaan::lav_partable_labels()}},
which are usually the labels used by
\code{coef()} to label the parameters.
}
}
\description{
Test all free parameters,
including user-defined parameters,
for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing all
free and user-defined model
parameters, by
setting it to the \code{test_fun}
argument.

For models fitted by \code{lavaan},
it uses \code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}
to do the test. If bootstrapping was
requested (by setting \code{se = "boot"}),
then it supports bootstrap
confidence intervals returned by
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}.

It has preliminary, though limited,
supported for models fitted by
\code{\link[stats:lm]{stats::lm()}} (through
\code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}). Tests are
conducted by ordinary least squares
confidence intervals based on
the \emph{t} statistic, reported by
\code{\link[stats:confint]{stats::confint()}} applied to
the output of \code{\link[stats:lm]{stats::lm()}}.
}
\section{Find the names of parameters}{
To use the argument \code{pars}, the
names as appeared in the function
\code{\link[=coef]{coef()}} must be used. For the
output of \code{lavaan}, this can
usually be inferred from the
parameter syntax (e.g., \code{y~x},
no space). If not sure, call
\code{\link[=coef]{coef()}} on the output of \code{lavaan}.
If a parameter is labelled, then
the label should be used in \code{par}.

If not sure, the function
\code{\link[=find_par_names]{find_par_names()}} can be used to
find valid names.
}

\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Simulate the data

sim_only <- power4test(nrep = 2,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the tests in each replication

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters)

print(test_out,
      test_long = TRUE)

# Do the tests in each replication: Standardized solution
# Delta method SEs will be used to do the tests

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(standardized = TRUE))

print(test_out,
      test_long = TRUE)

# Do the tests in each replication: Parameters with the selected operator

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(op = "~"))

print(test_out,
      test_long = TRUE)


# Finding valid parameter names

find_par_names(sim_only)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pop_es_yaml.R
\name{pop_es_yaml}
\alias{pop_es_yaml}
\title{Parse YAML-Stye Values For 'pop_es'}
\usage{
pop_es_yaml(text)
}
\arguments{
\item{text}{The multiline string to be parsed
to a specification of population
values.}
}
\value{
Either a named vector (for a
single-group model) or a list of
named vector (for a multigroup model)
of the population values of the
parameters (the effect sizes).
}
\description{
Convert a YAML string
to a vector or list of \code{pop_es}
specification.
}
\details{
The function \code{\link[=pop_es_yaml]{pop_es_yaml()}}
allows users to specify population
values of a model using one single
string, as in 'lavaan' model syntax.
}
\section{Specify 'pop_es' Using a Multiline String}{
When setting the argument \code{pop_es},
instead of using a named vector
or named list for
\code{pop_es}, the population values of
model parameters can also be
specified using a multiline string,
as illustrated below, to be parsed
by \code{\link[=pop_es_yaml]{pop_es_yaml()}}.
\subsection{Single-Group Model}{

This is an example of the multiline string
for a single-group model:

\preformatted{y ~ m: l
m ~ x: m
y ~ x: nil}

The string must follow this format:
\itemize{
\item Each line starts with \verb{tag:}.
\itemize{
\item \code{tag} can be the name of a
parameter, in \code{lavaan} model
syntax format.
\itemize{
\item For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
}
\item A tag in \code{lavaan} model syntax can
specify more than one parameter
using \code{+}.
\itemize{
\item For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
}
\item Alternatively, the \code{tag} can be
either \code{.beta.} or \code{.cov.}.
\itemize{
\item Use \code{.beta.} to set the default
values for all regression coefficients.
\item Use \code{.cov.} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
}
}
\item After each tag is the value of the
population value:

-\code{nil} for nil (zero),
\itemize{
\item \code{s} for small,
\item \code{m} for medium, and
\item \code{l} for large.
\item \code{si}, \code{mi}, and \code{li} for
small, medium, and large a
standardized indirect effect,
respectively.
}

Note: \code{n} \emph{cannot} be used in this mode.

The
value for each label is determined
by \code{es1} and \code{es2} as described
in \code{\link[=ptable_pop]{ptable_pop()}}.
\itemize{
\item The value can also be
set to a numeric value, such as
\code{.30} or \code{-.30}.
}
}

This is another example:

\preformatted{.beta.: s
y ~ m: l}

In this example, all regression
coefficients are \code{small}, while
the path from \code{m} to \code{y} is large.
}

\subsection{Multigroup Model}{

This is an example of the string
for a multigroup model:

\preformatted{y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil}

The format is similar to that for
a single-group model. If a parameter
has the same value for all groups,
then the line can be specified
as in the case of a single-group
model: \code{tag: value}.

If a parameter has different
values across groups, then it must
be in this format:
\itemize{
\item A line starts with the tag, followed
by two or more lines. Each line
starts with a hyphen \code{-} and the
value for a group.
}

For example:

\preformatted{m ~ x:
  - nil
  - s}

This denotes that the model has
two groups. The values of the path
from \code{x} to \code{m} for the two
groups are 0 (\code{nil}) and
small (\code{s}), respectively.

Another equivalent way to specify
the values are using \verb{[]}, on
the same line of a tag.

For example:

\preformatted{m ~ x: [nil, s]}

The number of groups is inferred
from the number of values for
a parameter. Therefore, if a tag
has more than one value, each tag
must has the same number of value,
or only one value.

The tag \code{.beta.} and \code{.cov.} can
also be used for multigroup models.
}

\subsection{Which Approach To Use}{

Note that using named vectors or
named lists is more reliable. However,
using a multiline string is
more user-friendly. If this method
failed, please use named vectors or
named list instead.
}

\subsection{Technical Details}{

The multiline string is parsed by \code{\link[yaml:read_yaml]{yaml::read_yaml()}}.
Therefore, the format requirement
is actually that of YAML. Users
knowledgeable of YAML can use other
equivalent way to specify the string.
}
}

\examples{

mod_es <- c("y ~ m" = "l",
            "m ~ x" = "m",
            "y ~ x" = "nil")

mod_es_yaml <-
"
y ~ m: l
m ~ x: m
y ~ x: nil
"

pop_es_yaml(mod_es_yaml)


}
\seealso{
\code{\link[=ptable_pop]{ptable_pop()}},
\code{\link[=power4test]{power4test()}}, and other functions
that have the \code{pop_es} argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/do_test.R
\name{do_test}
\alias{do_test}
\title{Do a Test on Each Replication}
\usage{
do_test(
  sim_all,
  test_fun,
  test_args = list(),
  map_names = c(fit = "fit"),
  results_fun = NULL,
  results_args = list(),
  parallel = FALSE,
  progress = FALSE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  cl = NULL
)
}
\arguments{
\item{sim_all}{The output of
\code{\link[=sim_out]{sim_out()}}.}

\item{test_fun}{A function to do the
test. See 'Details' for the requirement
of this function. There are some
built-in test functions in \code{power4mome},
described in 'Details'.}

\item{test_args}{A list of arguments
to be passed to the \code{test_fun}
function. Default is \code{list()}.}

\item{map_names}{A named character
vector specifying how the content of
the element \code{extra} in
each replication of \code{sim_all} map
to the argument of \code{test_fun}.
Default is \code{c(fit = "fit")},
indicating that the element \code{fit}
in the element \code{extra} is set to
the argument \code{fit} of \code{test_fun}.
That is, for the first replication,
\code{fit = sim_out[[1]]$extra$fit} when
calling \code{test_fun}.}

\item{results_fun}{The function to be
used to extract the test results.
See \code{Details} for the requirements
of this function. Default is \code{NULL},
assuming that the output of
\code{test_fun} can be used directly.}

\item{results_args}{A list of
arguments to be passed to the
\code{results_fun} function. Default is
\code{list()}.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to do the
tests. Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
of tests will be displayed.
Default is `FALSE.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{cl}{A cluster, such as one created
by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}. If \code{NULL},
a cluster will be created, but will be
stopped on exit. If set to an existing
cluster, it will not be stopped when
the function exits; users need to
stop it manually.}
}
\value{
An object of the class \code{test_out},
which is a list of length equal to
\code{sim_all}, one element for each
replication. Each element of the list
has two elements:
\itemize{
\item \code{test}: The output of the function
set to \code{test_fun}.
\item \code{test_results}: The output of
the function set to \code{results_fun}.
}
}
\description{
Do a test on each
replication in the output of
\code{\link[=sim_out]{sim_out()}}.
}
\details{
The function \code{\link[=do_test]{do_test()}} does
an arbitrary test in each
replication using the function set to
\code{test_fun}.
}
\section{The role of \code{do_test()}}{
The function \code{\link[=do_test]{do_test()}} is used by
the all-in-one function
\code{\link[=power4test]{power4test()}}. Users usually do not
call this function directly, though
developers can use this function to
develop other functions for power
analysis, or to build their own
workflows to do the power analysis.
}

\section{Major Test-Related Arguments}{
\subsection{The test function (test_fun)}{

The function set to \code{test_fun},
the \emph{test function}, usually
should work
on the output of \code{\link[lavaan:sem]{lavaan::sem()}},
\code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}, or
\code{\link[stats:lm]{stats::lm()}}, but can also be a
function that works on the output
of the function set to \code{fit_function}
when calling \code{\link[=fit_model]{fit_model()}} or
\code{\link[=power4test]{power4test()}} (see \code{fit_model_args}).

The function has two default
requirements.

First, it has an argument \code{fit}, to
be set to the output of
\code{\link[lavaan:sem]{lavaan::sem()}} or another output
stored in the element \code{extra$fit} of
a replication in the \code{sim_all}
object. (This requirement can be
relaxed; see the section on \code{map_names}.)

That is, the function definition
should be of this from: \code{FUN(fit, ...)}. This is the form of all
\verb{test_*} functions in \code{power4mome}.

If other arguments are to be passed
to the test function, they can be set
to \code{test_args} as a named list.

Second, the test function must
returns an output that (a) can be
processed by the results function
(see below), or (b) is of
the required format for the output of
a results function (see the next
section). If it already returns
an output of the required format,
then there is no need to set the
results function.
}

\subsection{The results function (results_fun)}{

The test results will be extracted
from the output of \code{test_fun} by the
function set to \code{results_fun},
the \emph{results function}. If
the \code{test_fun} already returns an
output of the expected format
(see below), then set \code{results_fun}
to \code{NULL}, the default. The output
of \code{test_fun} will be used for
estimating power.

The function set to \code{results_fun}
must accept the output of \code{test_fun},
as the first argument, and return a
named list (which can be a data frame)
or a named vector with some
of the following
elements:
\itemize{
\item \code{est}: Optional. The estimate of a
parameter, if applicable.
\item \code{se}: Optional. The standard error
of the estimate, if applicable.
\item \code{cilo}: Optional. The lower limit of the
confidence interval, if applicable.
\item \code{cihi}: Optional. The upper limit of the
confidence interval, if applicable.
\item \code{sig}: Required. If \code{1}, the test is
significant. If \code{0}, the test is not
significant. If the test cannot be
done for any reason, it should be
\code{NA}.
}

The results can then be used to
estimate the power or Type I
error of the test.

For example, if
the null hypothesis is false, then
the proportion of significant, that
is, the mean of the values of \code{sig}
across replications, is the power.
}

\subsection{Built-in test functions}{

The package \code{power4mome} has some ready-to-use
test functions:
\itemize{
\item \code{\link[=test_indirect_effect]{test_indirect_effect()}}
\item \code{\link[=test_cond_indirect]{test_cond_indirect()}}
\item \code{\link[=test_cond_indirect_effects]{test_cond_indirect_effects()}}
\item \code{\link[=test_moderation]{test_moderation()}}
\item \code{\link[=test_index_of_mome]{test_index_of_mome()}}
\item \code{\link[=test_parameters]{test_parameters()}}
}

Please refer to their help pages
for examples.
}

\subsection{The argument map_names}{

This argument is for developers using
a test function that has a different
name for the argument of the fit
object (\code{"fit"}, the default).

If \code{test_fun} is set to a function
that works on an output of, say,
\code{\link[lavaan:sem]{lavaan::sem()}} but the argument name
for the output is not \code{fit}, the
mapping can be changed by
\code{map_names}.

For example,
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}
receives an output of \code{\link[lavaan:sem]{lavaan::sem()}}
and reports the test results of model
parameters. However, the argument
name for the \code{lavaan} output is
\code{object.} To instruct \code{\link[=do_test]{do_test()}} to
do the test correctly when setting
\code{test_fun} to
\code{lavaan::parameterEstimates}, add
\code{map_names = c(object = "fit")}. The
element \code{fit} in a replication will
then set to the argument \code{object} of
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}.
}
}

\section{The background for having the \code{results_fun} argument}{
In the early development of
\code{power4mome}, \code{test_fun} is designed
to accept existing functions from
other packages, such as
\code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}. Their
outputs are not of the required
format for power analysis, and so
results functions are needed to
process their outputs. In the current
version of \code{power4mome}, some
ready-to-use test functions, usually
wrappers of these existing functions
from other packages, have been
developed, and they no longer need
results functions to process the
output. The argument \code{results_fun} is
kept for backward compatibility and
advanced users to use test functions
from other packages.
}

\examples{

# Specify the population model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the effect sizes (population parameter values)

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate several simulated datasets

data_all <- sim_data(nrep = 5,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     iseed = 1234)

# Fit the population model to each datasets

fit_all <- fit_model(data_all)

# Generate Monte Carlo estimates for forming Monte Carlo confidence intervals

mc_all <- gen_mc(fit_all,
                 R = 50,
                 iseed = 4567)

# Combine the results to a 'sim_all' object
sim_all <- sim_out(data_all = data_all,
                   fit = fit_all,
                   mc_out = mc_all)

# Test the indirect effect in each replication
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

test_all <- do_test(sim_all,
                    test_fun = test_indirect_effect,
                    test_args = list(x = "x",
                                     m = "m",
                                     y = "y",
                                     mc_ci = TRUE),
                    parallel = FALSE,
                    progress = FALSE)

# The result for each dataset
lapply(test_all, function(x) x$test_results)


}
\seealso{
See \code{\link[=power4test]{power4test()}} for
the all-in-one function that uses
this function. See \code{\link[=test_indirect_effect]{test_indirect_effect()}},
\code{\link[=test_cond_indirect]{test_cond_indirect()}},
\code{\link[=test_cond_indirect_effects]{test_cond_indirect_effects()}},
\code{\link[=test_moderation]{test_moderation()}},
\code{\link[=test_index_of_mome]{test_index_of_mome()}}, and
\code{\link[=test_parameters]{test_parameters()}} for examples of
test functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power4test_by_pop_es.R
\name{power4test_by_es}
\alias{power4test_by_es}
\alias{c.power4test_by_es}
\alias{as.power4test_by_es}
\alias{print.power4test_by_es}
\title{Power By Effect Sizes}
\usage{
power4test_by_es(
  object,
  pop_es_name = NULL,
  pop_es_values = NULL,
  progress = TRUE,
  ...,
  by_seed = NULL,
  by_nrep = NULL,
  save_sim_all = TRUE
)

\method{c}{power4test_by_es}(..., sort = TRUE, skip_checking_models = FALSE)

as.power4test_by_es(original_object, pop_es_name)

\method{print}{power4test_by_es}(x, print_all = FALSE, digits = 3, ...)
}
\arguments{
\item{object}{A \code{power4test} object,
or a \code{power4test_by_es} object.
If it is a \code{power4test_by_es} object,
the first element, which is a
\code{power4test} object, will be used
as the value of this argument.}

\item{pop_es_name}{The name of the
parameter. See the help page
of \code{\link[=ptable_pop]{ptable_pop()}} on the names for
the argument \code{pop_es}.}

\item{pop_es_values}{A numeric
vector of the population values
of the parameter specified in
\code{pop_es_names}.}

\item{progress}{Logical. Whether
progress of the simulation will be
displayed.}

\item{...}{For \code{\link[=power4test_by_es]{power4test_by_es()}},
they are arguments to be passed
to \code{\link[=power4test]{power4test()}}. For \code{\link[=c.power4test_by_es]{c.power4test_by_es()}},
they are \code{\link[=power4test_by_es]{power4test_by_es()}} outputs
to be combined together.
For the \code{print} method of the output
of \code{\link[=power4test_by_es]{power4test_by_es()}}, they are
arguments to be passed to the
\code{print} method of the output of
\code{\link[=power4test]{power4test()}} (\code{\link[=print.power4test]{print.power4test()}}).}

\item{by_seed}{If set to a number,
it will be used to generate the
seeds for each call to \code{\link[=power4test]{power4test()}}.
If \code{NULL}, the default, then seeds
will still be randomly generated but
the results cannot be easily reproduced.}

\item{by_nrep}{If set to a number,
it will be used to generate the
number of replications (\code{nrep}) for
each call to \code{\link[=power4test]{power4test()}}.
If set to a numeric vector of the
same length as \code{pop_es_values}, then these are
the \code{nrep} values for each of the
calls, allowing for different numbers
of replications for the population values.
If \code{NULL}, the default, then the
original \code{nrep} will be used.
This argument is used by
\code{\link[=x_from_power]{x_from_power()}} for efficiency, and
is rarely used when calling
this function directly.}

\item{save_sim_all}{If \code{FALSE},
the dataset in the
\code{power4test} objects are not saved,
to reduce
the size of the output. Default
is \code{TRUE}.}

\item{sort}{WHen combining the
objects, whether they will be sorted
by population values. Default is \code{TRUE}.}

\item{skip_checking_models}{Whether
the check of the data generation model
will be checked. Default is \code{TRUE}.
Should be set to \code{FALSE} only when
users are certain the they are based
on the same model, or when the model
is not saved (e.g., \code{save_sim_all}
set to \code{FALSE} when the objects
were generated).
This argument is used by
\code{\link[=x_from_power]{x_from_power()}} for efficiency, and
is rarely used when calling
the \code{c} method directly.}

\item{original_object}{The object
to be converted to a \code{power4test_by_es}
object.}

\item{x}{The object
to be printed.}

\item{print_all}{If \code{TRUE}, all
elements in \code{x}, that is, the results
of all sample sizes examined, will
be printed. If \code{FALSE}, then only
those of the first value of the
parameter
will be printed.}

\item{digits}{The numbers of digits
displayed after the decimal.}
}
\value{
The function
\code{\link[=power4test_by_es]{power4test_by_es()}} returns a
\code{power4test_by_es} object, which is a
list of \code{power4test} objects, one for
each population value of the parameter.

The method \code{\link[=c.power4test_by_es]{c.power4test_by_es()}} returns
a \code{power4test_by_es} object with
all the elements (tests for different
values of \code{pop_es_values}) combined.

The function \code{\link[=as.power4test_by_es]{as.power4test_by_es()}} returns
a \code{power4test_by_es} object converted
from the input object.

The \code{print}-method of \code{power4test_by_es}
objects returns the object invisibly.
It is called for its side-effect.
}
\description{
Estimate power for a
set of effect sizes (population
values of a model parameter).
}
\details{
The function \code{\link[=power4test_by_es]{power4test_by_es()}}
regenerates
datasets for a set of effect sizes
(population values of a model parmeter)
and does the stored tests in each of
them.

Optionally, it can also be run
on a object with no stored tests.
In this case, additional arguments
must be set to instruct \code{\link[=power4test]{power4test()}}
on the tests to be conducted.

It is usually used to examine the
power over a sets of effect sizes
(population values).

The \code{c} method of \code{power4test_by_es}
objects
is used to combine tests from different
runs of \code{\link[=power4test_by_es]{power4test_by_es()}}.

The function \code{\link[=as.power4test_by_es]{as.power4test_by_es()}}
is used to convert a \code{power4test}
object to a \code{power4test_by_es}
object, if it is not already one.
Useful when concatenating
\code{power4test} objects with
\code{power4test_by_es} objects.
}
\examples{

# Specify the model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the population values

model_simple_med_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

sim_only <- power4test(nrep = 2,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 40,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234)

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

out <- power4test_by_es(test_out,
                            pop_es_name = "y ~ m",
                            pop_es_values = c(.10, .20))
out_reject <- rejection_rates(out)
out_reject

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/x_from_power.R
\name{x_from_power}
\alias{x_from_power}
\alias{n_from_power}
\alias{n_region_from_power}
\alias{print.x_from_power}
\alias{print.n_region_from_power}
\alias{arg_x_from_power}
\alias{pba_diagnosis}
\title{Sample Size and Effect Size Determination}
\usage{
x_from_power(
  object,
  x = arg_x_from_power(object, "x", arg_in = "call") \%||\% "n",
  pop_es_name = arg_x_from_power(object, "pop_es_name", arg_in = "call"),
  target_power = 0.8,
  what = arg_x_from_power(object, "what") \%||\% "point",
  goal = arg_x_from_power(object, "goal") \%||\% {
     switch(what, point = "ci_hit", ub
    = "close_enough", lb = "close_enough")
 },
  ci_level = 0.95,
  tolerance = NULL,
  x_interval = switch(x, n = c(50, 2000), es = NULL),
  extendInt = NULL,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  final_nrep = attr(object, "args")$nrep \%||\% (object$nrep_final \%||\% 400),
  final_R = attr(object, "args")$R \%||\% (object$args$final_R \%||\% 1000),
  seed = NULL,
  x_include_interval = FALSE,
  check_es_interval = TRUE,
  power_curve_args = list(power_model = NULL, start = NULL, lower_bound = NULL,
    upper_bound = NULL, nls_control = list(), nls_args = list()),
  save_sim_all = FALSE,
  algorithm = NULL,
  control = list(),
  internal_args = list(),
  rejection_rates_args = list(collapse = "all_sig", at_least_k = 1, p_adjust_method =
    "none", alpha = 0.05)
)

n_from_power(
  object,
  pop_es_name = NULL,
  target_power = 0.8,
  what = formals(x_from_power)$what,
  goal = formals(x_from_power)$goal,
  ci_level = 0.95,
  tolerance = NULL,
  x_interval = c(50, 2000),
  extendInt = NULL,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  final_nrep = formals(x_from_power)$final_nrep,
  final_R = formals(x_from_power)$final_R,
  seed = NULL,
  x_include_interval = FALSE,
  check_es_interval = TRUE,
  power_curve_args = list(power_model = NULL, start = NULL, lower_bound = NULL,
    upper_bound = NULL, nls_control = list(), nls_args = list()),
  save_sim_all = FALSE,
  algorithm = NULL,
  control = list(),
  internal_args = list(),
  rejection_rates_args = list(collapse = "all_sig", at_least_k = 1, p_adjust_method =
    "none", alpha = 0.05)
)

n_region_from_power(
  object,
  pop_es_name = NULL,
  target_power = 0.8,
  ci_level = 0.95,
  tolerance = NULL,
  x_interval = c(50, 2000),
  extendInt = NULL,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  final_nrep = formals(x_from_power)$final_nrep,
  final_R = formals(x_from_power)$final_R,
  seed = NULL,
  x_include_interval = FALSE,
  check_es_interval = TRUE,
  power_curve_args = list(power_model = NULL, start = NULL, lower_bound = NULL,
    upper_bound = NULL, nls_control = list(), nls_args = list()),
  save_sim_all = FALSE,
  algorithm = NULL,
  control = list(),
  internal_args = list(),
  rejection_rates_args = list(collapse = "all_sig", at_least_k = 1, p_adjust_method =
    "none", alpha = 0.05)
)

\method{print}{x_from_power}(x, digits = 3, call = TRUE, ...)

\method{print}{n_region_from_power}(x, digits = 3, ...)

arg_x_from_power(object, arg, arg_in = NULL)

pba_diagnosis(
  a_out,
  p_interval = c(0.05, 0.95),
  posterior_xlim = c(0.01, 0.99),
  which = NULL
)
}
\arguments{
\item{object}{A \code{power4test} object,
which is the output of \code{\link[=power4test]{power4test()}}.
Can also be a \code{power4test_by_n} object,
the output
of \code{\link[=power4test_by_n]{power4test_by_n()}}, or
a \code{power4test_by_es} object, the
output of
\code{\link[=power4test_by_es]{power4test_by_es()}}. For these
two types of objects, the attempt
with power closest to the
\code{target_power} will be used as
\code{object}, and all other attempts in
them will be included in the estimation
of subsequent attempts and the final
output. Last, it can also be the
output of a previous call to
\code{\link[=x_from_power]{x_from_power()}}, and the stored
trials will be retrieved.}

\item{x}{For \code{\link[=x_from_power]{x_from_power()}},
\code{x} set the value to
be searched. Can be \code{"n"}, the sample
size, or \code{"es"}, the population value
of a parameter (set by \code{pop_es_name}).
For the \code{print} method of \code{x_from_power}
objects, this is the output of
\code{\link[=x_from_power]{x_from_power()}}.}

\item{pop_es_name}{The name of the
parameter. Required if \code{x} is \code{"es"}.
See the help page
of \code{\link[=ptable_pop]{ptable_pop()}} on the names for
the argument \code{pop_es}.}

\item{target_power}{The target power,
a value greater than 0 and less than
one.}

\item{what}{The value for which is
searched: the estimate power (\code{"point"}),
the upper bound of the confidence
interval (\code{"ub"}), or the lower bound
of the confidence interval (\code{"lb"}).}

\item{goal}{The goal of the search.
If \code{"ci_hit"}, then the goal is to
find a value of \code{x} with the
confidence interval of the estimated
power including the target power.
If \code{"close_enough"}, then the goal
is to find a value of \code{x} with the
value in \code{what} "close enough" to
the target power, defined by having
an absolute difference with the
target power less than \code{tolerance}.}

\item{ci_level}{The level of confidence
of the confidence intervals computed
for the estimated power. Default is
.95, denoting 95\%.}

\item{tolerance}{Used when the goal
is \code{"close_enough"}.
If \code{NULL}, set automatically based
on the algorithm used.}

\item{x_interval}{A vector of
two values, the minimum value
and the maximum values of \code{x}, in
the search for the values
(sample sizes or population values).
If \code{NULL}, default when \code{x = "es"},
it will be determined internally.}

\item{extendInt}{Whether \code{x_interval}
can be expanded when estimating the
the values to try. The value will
be passed to the argument of the
same name in \code{\link[stats:uniroot]{stats::uniroot()}}.
If \code{x} is \code{"n"}, then the default
value is \code{"upX"}.
That is, a value higher than
the maximum in \code{x_interval} is
allowed, if predicted by the tentative
model. Otherwise, the default value
is \code{"no"}. See the help page of
\code{\link[stats:uniroot]{stats::uniroot()}} for further
information.}

\item{progress}{Logical. Whether
the searching progress is reported.}

\item{simulation_progress}{Logical.
Whether the progress in each call
to \code{\link[=power4test]{power4test()}}, \code{\link[=power4test_by_n]{power4test_by_n()}},
or \code{\link[=power4test_by_es]{power4test_by_es()}}
is shown. To be passed to
the \code{progress} argument of these
functions.
If \code{NULL}, set automatically based
on the algorithm used.}

\item{max_trials}{The maximum number
of trials in searching the value
with the target power. Rounded
up if not an integer.
If \code{NULL}, set automatically based
on the algorithm used.}

\item{final_nrep}{The number of
replications in the final stage,
also the maximum number of replications
in each call to \code{\link[=power4test]{power4test()}},
\code{\link[=power4test_by_n]{power4test_by_n()}}, or
\code{\link[=power4test_by_es]{power4test_by_es()}}.
If \code{object}
is an output of \code{\link[=power4test]{power4test()}}
or \code{\link[=x_from_power]{x_from_power()}} and
this argument is not set, \code{final_nrep}
will be set to \code{nrep} or \code{final_nrep}
stored in
\code{object}.}

\item{final_R}{The number of
Monte Carlo simulation or
bootstrapping samples in the final
stage. The \code{R} in calling
\code{\link[=power4test]{power4test()}}, \code{\link[=power4test_by_n]{power4test_by_n()}},
or \code{\link[=power4test_by_es]{power4test_by_es()}}
will be stepped up to this value
when approaching the target
power. Do not need to be very large
because the goal is to estimate
power by replications, not for high
precision in one single replication.
If \code{object}
is an output of \code{\link[=power4test]{power4test()}}
or \code{\link[=x_from_power]{x_from_power()}} and
this argument is not set, \code{final_R}
will be set to \code{R} or \code{final_R}
stored in
\code{object}.}

\item{seed}{If not \code{NULL}, \code{\link[=set.seed]{set.seed()}}
will be used to make the process
reproducible. This is not always
possible if many stages of
parallel processing is involved.}

\item{x_include_interval}{Logical.
Whether the minimum and maximum
values in \code{x_interval} are mandatory
to be included in the values
to be searched.}

\item{check_es_interval}{If \code{TRUE},
the default, and \code{x} is \code{"es"},
a conservative probable
range of valid values for the selected
parameter will be determined, and it
will be used instead of \code{x_interval}.
If the range spans both positive and
negative values, only the interval
of the same sign as the population
value in \code{object} will be used.}

\item{power_curve_args}{A named
list of arguments to be passed
\code{\link[=power_curve]{power_curve()}} when estimating
the relation between power and \code{x}
(sample size or effect size). Please
refer to \code{\link[=power_curve]{power_curve()}} on available
arguments. There is one except:
\code{power_model} is mapped to
the \code{formula} argument of
\code{\link[=power_curve]{power_curve()}}.}

\item{save_sim_all}{If \code{FALSE},
the default, the data in each
\code{power4test} object for each
value of \code{x} is not saved,
to reduce the size of the output.
If set to \code{TRUE}, the size of the
output can be very large in size.}

\item{algorithm}{The algorithm for
finding \code{x}. Can be \code{"power_curve"},
\code{"bisection"}, or
\code{"probabilistic_bisection"}. The default algorithm
depends on \code{x}.}

\item{control}{A named list of
additional
arguments to be passed to the
algorithm to be used. For advanced
users.}

\item{internal_args}{A named list
of internal arguments. For internal
testing. Do not use it.}

\item{rejection_rates_args}{Argument values to be used when
\code{\link[=rejection_rates]{rejection_rates()}} is called, used
to decide how rejection rates will
be estimated. Only one single test
is supported by \code{\link[=x_from_power]{x_from_power()}}.
Therefore, \code{merge_all_tests} is
always \code{TRUE} and cannot be changed.
The argument \code{collapse} can be
\code{"all_sig"}, \code{"at_least_one_sig"},
or \code{"at_least_k_sig"} (it cannot be
\code{"none"}). Please refer to
\code{\link[=rejection_rates]{rejection_rates()}} for other
possible arguments. These values,
if set, will overwrite any stored
settings in \code{object}.}

\item{digits}{The number of digits
after the decimal when printing
the results.}

\item{call}{Logical. Whether the
call is printed.}

\item{...}{Optional arguments.
Not used for now.}

\item{arg}{The name of element to
retrieve.}

\item{arg_in}{The name of the element
from which an element is to be
retrieved.}

\item{a_out}{The output of
\code{\link[=x_from_power]{x_from_power()}} and friends. The
algorithm used must be \code{"probabilistic_bisection"}.}

\item{p_interval}{The range of
the plot that "zooms" around the solution
(or the median of in the final posterior
probability distribution), expressed
in terms of the area of the distribution.}

\item{posterior_xlim}{The range of
the first plot of search history and the
plot of posterior probability distribution,
expressed
in terms of the area of the distribution.}

\item{which}{If \code{a_out} is a list with
more than one output of \code{\link[=x_from_power]{x_from_power()}},
such as the output of \code{\link[=n_region_from_power]{n_region_from_power()}},
\code{which} must be set to the name of
one such output (\code{"below"} or \code{"above"}
for the output of \code{\link[=n_region_from_power]{n_region_from_power()}},
or the output of \verb{q_power_mediation_*}).}
}
\value{
The function \code{\link[=x_from_power]{x_from_power()}}
returns an \code{x_from_power} object,
which is a list with the following
elements:
\itemize{
\item \code{power4test_trials}: The output of
\code{\link[=power4test_by_n]{power4test_by_n()}} for all sample
sizes examined, or of
\code{\link[=power4test_by_es]{power4test_by_es()}} for all
population values of the selected
parameter examined.
\item \code{rejection_rates}: The output of
\code{\link[=rejection_rates]{rejection_rates()}}.
\item \code{x_tried}: The sample sizes or
population values
examined.
\item \code{power_tried}: The estimated
rejection rates for all the values
examined.
\item \code{x_final}: The sample size or
population value in the
solution. \code{NA} if a solution not found.
\item \code{power_final}: The estimated power
of the value in the solution.
\code{NA} if a solution not found.
\item \code{i_final}: The position of the
solution in \code{power4test_trials}.
\code{NA} if a solution not found.
\item \code{ci_final}: The confidence interval
of the estimated power in the solution.
The method is determined
by the option \code{power4mome.ci_method}.
If \code{NULL} or \code{"wilson"}, Wilson's
(1927) method is used. If
\code{"norm"}, normal approximation
is used.
\item \code{ci_level}: The level of confidence
of \code{ci_final}.
\item \code{nrep_final}: The number of
replications (\code{nrep}) when estimating
the power in the solution.
\item \code{power_curve}: The output of
\code{\link[=power_curve]{power_curve()}} when estimating the
power curve.
\item \code{target_power}: The requested
target power.
\item \code{power_tolerance}: The allowed
difference between the solution's
estimated power and the target
power. Determined by the number
of replications and the level of
confidence of the confidence intervals.
\item \code{x_estimated}: The value
(sample size or population value)
with the target power, estimated by
\code{power_curve}. This is used, when
solution not found, to determine the
range of the values to search when
calling the function again.
\item \code{start}: The time and date when
the process started.
\item \code{end}: The time and date when the
process ended.
\item \code{time_spent}: The time spent in
doing the search.
\item \code{args}: A named list of the arguments
of \code{\link[=x_from_power]{x_from_power()}} used in the search.
\item \code{call}: The call when this function
is called.
}

The function \code{\link[=n_region_from_power]{n_region_from_power()}}
returns a named list of two output of
\code{\link[=n_from_power]{n_from_power()}}, of the class
\code{n_region_from_power}. The output
with \code{what = "ub"} is named \code{"below"},
and the output with \code{what = "lb"} is
namd \code{"above"}.

The \code{print}-method of \code{x_from_power}
objects returns the object \code{x}
invisibly.
It is called for its side effect.

The \code{print}-method of \code{x_from_power_region}
objects returns the object \code{x}
invisibly.
It is called for its side effect.

The function \code{\link[=arg_x_from_power]{arg_x_from_power()}}
returns the requested argument if
available. If not available, it
returns \code{NULL}.

The function \code{\link[=pba_diagnosis]{pba_diagnosis()}}
returns \code{NULL} invisibly. Called for
its side-effect.
}
\description{
It searches by simulation
the sample size (given other factors,
such as effect sizes) or effect size
(given other factors, such as sample
size) with power to
detect an effect close to a target
value.
}
\details{
This is how to use \code{\link[=x_from_power]{x_from_power()}}:
\itemize{
\item Specify the model by \code{\link[=power4test]{power4test()}},
with \code{do_the_test = FALSE}, and set
the magnitude of the effect sizes
to the minimum levels to detect.
\item Add the test using \code{\link[=power4test]{power4test()}}
using \code{test_fun} and \code{test_args}
(see the help page of \code{\link[=power4test]{power4test()}}
for details). Run it on the
starting sample size or
effect size.
\item Call \code{\link[=x_from_power]{x_from_power()}} on the output
of \code{\link[=power4test]{power4test()}} returned from
the previous step. This
function will iteratively repeat
the analysis on either other sample
sizes, or other values for a
selected model parameter (the
effect sizes),
trying to achieve a goal (\code{goal}) for
a value of interest (\code{what}).
}

If the \code{goal} is \code{"ci_hit"}, the
search will try to find a value (a sample
size, or a population value of
the selected model parameter) with
a power level close enough to the
target power, defined by having its
confidence interval for the power
including the target power.

If the \code{goal} is \code{"close_enough"},
then the search will try to find a
value of \code{x} with its level of
power (\code{"point"}), the upper bound
of the confidence interval for this
level of power (\code{"ub"}), or the
lower bound of the confidence interval
fro this level of power (\code{"lb"})
"close enough" to the target level of
power, defined by having an absolute
difference less than the \code{tolerance}.

If several values of \code{x} (sample
size or the population value of
a model parameter) have already been
examined by \code{\link[=power4test_by_n]{power4test_by_n()}} or
\code{\link[=power4test_by_es]{power4test_by_es()}}, the output
of these two functions can also be
used as \code{object} by \code{\link[=x_from_power]{x_from_power()}}.

Usually, the default values of the
arguments should be sufficient.

The results can be viewed using
\code{\link[=summary]{summary()}}, and the output has
a \code{plot} method (\code{\link[=plot.x_from_power]{plot.x_from_power()}}) to
plot the relation between power and
values (of \code{x}) examined.

A detailed illustration on how to
use this function for sample size can be found
from this page:

\url{https://sfcheung.github.io/power4mome/articles/x_from_power_for_n.html}

The function \code{\link[=n_from_power]{n_from_power()}} is just
a wrapper of \code{\link[=x_from_power]{x_from_power()}}, with
\code{x} set to \code{"n"}.

The function \code{\link[=n_region_from_power]{n_region_from_power()}} is just
a wrapper of \code{\link[=x_from_power]{x_from_power()}}, with
\code{x} set to \code{"n"}, with two passes, one
with \code{what = "ub"} and one with
\code{what = "lb"}.

The \code{print} method only prints
basic information. Call the
\code{summary} method of \code{x_from_power} objects
(\code{\link[=summary.x_from_power]{summary.x_from_power()}}) and its
\code{print} method for detailed results

The function \code{\link[=arg_x_from_power]{arg_x_from_power()}}
is a helper to set argument values
if \code{object} is an output
of \code{\link[=x_from_power]{x_from_power()}} or similar
functions.

The function \code{\link[=pba_diagnosis]{pba_diagnosis()}}
generates simple diagnostic plots
for the search history of probabilistic
bisection algorithm. This function is for advanced users
to examine the search history of
the probabilistic bisection algorithm.
It is for diagnostic purpose and has
limited support for customizing the plots.
}
\section{Algorithms}{
Three algorithms are currently available,
the simple (though sometimes inefficient)
bisection method, a method that
makes use of the estimated crude power
curve, and probabilistic bisection
algorithm (Waeber et al., 2013;
see Chalmers, 2024, for applying
this algorithm to power analysis).

Unlike typical root-finding problems,
the prediction of the level of power
is stochastic. Moreover, the computational
cost is high when Monte Carlo or
bootstrap confidence intervals are
used to do a test because the estimation
of the power for one single value of
\code{x} can sometimes take one minute or
longer. Therefore, in addition to
the simple bisection method, two methods,
named \emph{power curve} method
(which belongs to the family of
surrogate function approximation method
reviewed in Chalmers, 2024) and
probabilistic bisection method
(Chalmers, 2024; Waeber et al., 2013) were also
developed for this
scenario.
\subsection{Bisection Method}{

This method (called informat bisection
in Chalmers, 2024), enabled by
\code{algorithm = "bisection"},
basically starts with
an interval that probably encloses the
value of \code{x} that meets the goal,
and then successively narrows this
interval. A point inside this
interval, usually the mid-point but
can also be approximated by another
method (e.g., the power curve method),
is used as the estimate.
Though simple, there are cases in
which it can be slow. Nevertheless,
preliminary examination suggests that
this method is good enough for
scenarios in which only an approximate
sample size or range of sample sizes is
needed.

The internal workflow of
this method implemented in
\code{\link[=x_from_power]{x_from_power()}} can be found in
this technical vignette: \url{https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_bisection.html}.
}

\subsection{Power Curve Method}{

This method (belongs to the
surrogate function approximation family
reviewed in Chalmers, 2024),
enabled by \code{algorithm = "power_curve"},
starts with a crude
power curve based on a few points.
This tentative model is then used
to suggest the values to examine in
the next iteration. Unlike some other
implementations
of this family of methods, the form, not
just the parameters, of the
model can change across iterations,
as more and more data points are
available.

This method is the default method
for some scenarios, such as
\code{x = "es"} with \code{goal = "ci_hit"}
because the relation
between the power and the population
value of a parameter varies across
parameters, unlike the relation
between power and sample size, which
is monotonic. Therefore,
taking into account the working
power curve may help finding the
desired value of \code{x}.

Before version 0.1.1.33, this
method can be used only with
the goal \code{"ci_hit"}. Since
version 0.1.1.34, it supports all
goals, like the bisection method.

The internal workflow of
this method implemented in
\code{\link[=x_from_power]{x_from_power()}} can be found in
this technical vignette: \url{https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_power_curve.html}.
}

\subsection{Probabilistic Bisection}{

This method, proposed by Waeber
and others (2013) for stochastic
root-solving, has been adapted by
Chalmers (2024) for power analysis.
Similar to bisection, this method
starts with an interval, with a
initial probability for each value
(or range of values), such as sample sizes,
as the sample size with the target power.
A value is then selected to estimate
power (the median, by default).
Based on the estimated power
for this value,
the distribution of probabilities
is updated. This process is repeated
until some termination criteria are met.

Unlike bisection, each iteration can
be conducted with a small number of
replications (e.g., 50). The method
accumulate evidence in a Bayesian
approach, and so the certainty of the solution
is based on the accumulation of evidence
from successive iterations, not on
having strong evidence from a few iterations.

In \code{x_from_power}, the version
of probabilistic bisection proposed
by Waeber et al. (2013) was implemented,
with minor changes.

Most importantly, when some termination
criteria are met, the candidate value
of \code{x} (e.g., a sample size) will be
checked using a larger number of
replications (set by \code{final_nrep})
to ensure that the estimated power
is indeed close enough to the target
power (based on \code{tolerance} value,
determined internally but can also be
set directly). If yes, it will be
returned as the solution. If no, then
the search will continue, until a
maximum number of candidate values
has been checked.

Although the bisection method can be fast
in some situations (e.g., when the
interval is narrow and the solution
happens to be inside the interval),
the probabilistic bisection method
is nearly guaranteed to converge to
the solution (as long as the solution
is inside the interval). Therefore,
this is the default algorithm in
some scenarios.

The internal workflow of
this method implemented in
\code{\link[=x_from_power]{x_from_power()}} can be found in
this technical vignette: \url{https://sfcheung.github.io/power4mome/articles/x_from_power_workflow_pba.html}.
}
}

\examples{

# Specify the population model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

# Generate the datasets

sim_only <- power4test(nrep = 5,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 2345)

# Do a test

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(pars = "m~x"))

# Determine the sample size with a power of .80 (default)

# In real analysis, to have more stable results:
# - Use a larger final_nrep (e.g., 400).

# If the default values are OK, this call is sufficient:
# power_vs_n <- x_from_power(test_out,
#                            x = "n",
#                            seed = 4567)
power_vs_n <- x_from_power(test_out,
                           x = "n",
                           progress = TRUE,
                           target_power = .80,
                           final_nrep = 5,
                           max_trials = 1,
                           seed = 1234)
summary(power_vs_n)
plot(power_vs_n)


}
\references{
Chalmers, R. P. (2024). Solving variables
with Monte Carlo simulation experiments:
A stochastic root-solving approach.
\emph{Psychological Methods}.
Advance online publication.
\doi{10.1037/met0000689}

Waeber, R., Frazier, P. I., &
Henderson, S. G. (2013). Bisection
search with noisy responses.
\emph{SIAM Journal on Control and Optimization}, \emph{51}(3),
2261--2279. \doi{10.1137/120861898}

Wilson, E. B. (1927). Probable inference, the law of
succession, and statistical inference.
\emph{Journal of the American Statistical Association, 22}(158),
209-212. \doi{10.1080/01621459.1927.10502953}
}
\seealso{
\code{\link[=power4test]{power4test()}}, \code{\link[=power4test_by_n]{power4test_by_n()}},
and \code{\link[=power4test_by_es]{power4test_by_es()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runif_rs.R
\name{runif_rs}
\alias{runif_rs}
\title{Random Variable From a Uniform Distribution}
\usage{
runif_rs(n = 10, min = 0, max = 1, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{min}{min for runif.}

\item{max}{max for runif.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}
}
\value{
A vector of the generated random numbers.
}
\description{
Generate random numbers
from a uniform distribution, with
user-specified population mean and
standard deviation.
}
\details{
First, the user specifies
the parameters, min and max, and the
desired population mean and standard
deviation. Then the random numbers
will be generated and rescaled with
the desired population mean and
standard.
}
\examples{
set.seed(90870962)
x <- runif_rs(n = 5000,
              min = 2,
              max = 4,
              pmean = 3,
              psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power_curve_predict.R
\name{predict.power_curve}
\alias{predict.power_curve}
\title{Predict Method for a 'power_curve' Object}
\usage{
\method{predict}{power_curve}(object, newdata, ...)
}
\arguments{
\item{object}{A \code{power_curve}
object.}

\item{newdata}{A data frame with
a column named \code{x}. It can also
be a named list, with one element
named \code{x} and is a vector of the
values. If not supplied, values of
\code{x} stored in \code{object} will be used.}

\item{...}{Additional arguments.
Passed to the corresponding
\code{predict} method.}
}
\value{
It returns a numeric vector of
the predicted rejection rates.
}
\description{
Compute the predicted
values in a model fitted by
\code{\link[=power_curve]{power_curve()}}.
}
\details{
The \code{predict} method of \code{power_curve}
objects works in two modes.

If new
data is not supplied (through
\code{newdata}), it retrieves the stored
results and calls the corresponding
methods to compute the predicted
values, which are the predicted
rejection rates (power levels if
the null hypothesis is false,
e.g., the population effect size is
equal to zero).

If new data is supplied, such as
a named list with a vector of
sample sizes, they will be used to
compute the predicted rejection
rates.
}
\examples{

# Specify the population model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the effect sizes (population parameter values)

model_simple_med_es <-
"
y ~ m: l
m ~ x: m
y ~ x: s
"

# Simulate datasets to check the model
# Set `parallel` to TRUE for faster, usually much faster, analysis
# Set `progress` to TRUE to display the progress of the analysis

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 50,
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234,
                       parallel = FALSE,
                       progress = FALSE)

# By n: Do a test for different sample sizes

out1 <- power4test_by_n(sim_only,
                        nrep = 5,
                        test_fun = test_parameters,
                        test_args = list(par = "y~x"),
                        n = c(25, 100, 200, 1000),
                        by_seed = 1234,
                        parallel = FALSE,
                        progress = FALSE)

pout1 <- power_curve(out1)
pout1
predict(pout1,
        newdata = list(x = c(150, 250, 500)))


}
\seealso{
\code{\link[=power_curve]{power_curve()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rexp_rs.R
\name{rexp_rs}
\alias{rexp_rs}
\title{Random Variable From an Exponential Distribution}
\usage{
rexp_rs(n = 10, rate = 1, pmean = 0, psd = 1, rev = FALSE)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{rate}{\code{rate} for \code{\link[stats:Exponential]{stats::rexp()}}.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}

\item{rev}{If TRUE, the distribution
is revered to generate a negatively
skewed distribution. Default is
FALSE.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from an exponential distribution,
rescaled to have user-specified
population mean and standard
deviation.
}
\details{
First, specify the parameter,
\code{rate}, and the
desired population mean and standard
deviation. The random numbers, drawn
from an exponential distribution by
\code{\link[stats:Exponential]{stats::rexp()}}, will then be
rescaled with the desired population
mean and standard.
}
\examples{
set.seed(90870962)
x <- rexp_rs(n = 5000,
             rate = 4,
             pmean = 3,
             psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_indirect_effects.R
\name{test_k_indirect_effects}
\alias{test_k_indirect_effects}
\title{Test Several Indirect Effects}
\usage{
test_k_indirect_effects(
  fit = fit,
  x = NULL,
  m = NULL,
  y = NULL,
  mc_ci = TRUE,
  mc_out = NULL,
  boot_ci = FALSE,
  boot_out = NULL,
  check_post_check = TRUE,
  test_method = NULL,
  ...,
  omnibus = c("no", "all_sig", "at_least_one_sig", "at_least_k_sig", "total"),
  at_least_k = 1,
  p_adjust_method = "none",
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}.}

\item{x}{The name of the \code{x}-variable,
the predictor.}

\item{m}{Must be a list of character
vectors. Each character vector stores the
name(s) of mediator(s) along a path.
The path
moves from the first mediator in the
vector to the last mediator in the
vector. If \code{NULL}, the stored paths
will be used, which are all the
indirect paths in the model between
\code{x} and \code{y}, by default.}

\item{y}{The name of the \code{y}-variable,
the outcome variable.}

\item{mc_ci}{Logical. If \code{TRUE}, the
default, Monte Carlo confidence
intervals will be formed. This argument
and \code{boot_ci} cannot be both \code{TRUE}.}

\item{mc_out}{The pre-generated
Monte Carlo estimates generated by
\link[manymome:do_mc]{manymome::do_mc}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{boot_ci}{Logical. If \code{TRUE},
the default, nonparametric bootstrap
confidence intervals will be formed.
This argument
and \code{mc_ci} cannot be both \code{TRUE}.}

\item{boot_out}{The pre-generated
bootstrap estimates generated by
\link[manymome:do_boot]{manymome::do_boot}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{test_method}{The method to do
the test. If \code{"ci"}, then the
confidence interval (e.g., bootstrap
confidence interval) will be used to
do the test. If
\code{"pvalue"}, then asymmetric \emph{p}-value
by Asparouhov & Muthén (2021) will
be used to do the test, and the
confidence interval will not be
computed. If \code{NULL}, its value will
be set to \code{"pvalue"} if the number of
simulated/bootstrap samples (\code{R}) is
a value that is supported by the
method by Boos and Zhang (2000), and
set to \code{"ci"} otherwise.}

\item{...}{Additional arguments to
be passed to \code{\link[manymome:cond_indirect]{manymome::many_indirect_effects()}}.}

\item{omnibus}{If \code{"no"}, the default,
then the test results for all paths
are stored. If \code{"all_sig"}, then
only one row of test is stored, and
the test is declared significant if
\emph{all} paths are significant. If
\code{"at_least_one_sig"}, then only
one row of test is stored, and the
test is declared significant if
at least one of the paths is
significant. If \code{"at_least_k_sig"},
then only one row of test is stored,
and the test is declared significant
if at least \code{k} of the paths is
significant, \code{k} determined by the
argument \code{at_least_k}. If \code{"total"},
then the total indirect effect will
be tested.}

\item{at_least_k}{The minimum number
of paths required to be significant
for the omnibus test to be considered
significant. Used when
\code{omnibus} is \code{"at_least_k_sig"}.}

\item{p_adjust_method}{The method to be
passed to \code{\link[=p.adjust]{p.adjust()}} to adjust the
\emph{p}-values when testing the effects.
Default is \code{"none"} and the \emph{p}-values
will not be adjusted. The unadjusted
\emph{p}-values will be stored in the column
\code{pvalue_org}.}

\item{fit_name}{The name of the
model fit object to be extracted.
Default is \code{"fit"}. Used only when
more than one model is fitted in
each replication. This should be
the name of the model on which the
test is to be conducted.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
a data frame with the
following columns:
\itemize{
\item \code{est}: The estimated
indirect effect for each path.
\item \code{cilo} and \code{cihi}: The
lower and upper limits of the
confidence interval (95\% by
default), respectively,
for each indirect effect
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
\item \code{test_label}: A column of labels
generated to label the indirect
effects.
}

If \code{omnibus} is \code{"all_sig"} or
\verb{"at_least_one"sig"}, then
the data frame has only one row,
and the columns \code{"est"}, \code{"cilo"},
and \code{"cihi"} are \code{NA}. The column
\code{sig} is determined by whether
all paths are significant (\code{"all_sig"})
or whether at least one path is
significant (\code{"at_least_one_sig"}).
}
\description{
Test several indirect effects
for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing an
indirect effect, by setting it
to the \code{test_fun} argument.

It uses \code{\link[manymome:cond_indirect]{manymome::many_indirect_effects()}}
to do the test. It can be used on
models fitted by \code{\link[lavaan:sem]{lavaan::sem()}}
or fitted by a sequence of calls
to \code{\link[stats:lm]{stats::lm()}}, although only
nonparametric bootstrap confidence
interval is supported for models
fitted by regression using
\code{\link[stats:lm]{stats::lm()}}.

It also supports testing the total
indirect effect. Just set \code{omnibus}
to \code{"total"}.
}
\examples{

# Specify the model

model_simple_med <-
"
m1 ~ x
m2 ~ x
y ~ m1 + m2 + x
"

# Specify the population values

model_simple_med_es <-
"
y ~ m1: s
m1 ~ x: m
y ~ m2: s
m2 ~ x: l
y ~ x: n
"

# Simulate the data

sim_only <- power4test(nrep = 5,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the test in each replication

test_ind <- power4test(object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE))
print(test_ind,
      test_long = TRUE)

# Set omnibus = "all_sig" to declare
# significant only if all paths are
# significant

test_ind_all_sig <- power4test(
                       object = sim_only,
                       test_fun = test_k_indirect_effects,
                       test_args = list(x = "x",
                                        y = "y",
                                        mc_ci = TRUE,
                                        omnibus = "all_sig"))
print(test_ind_all_sig,
      test_long = TRUE)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_rejection_rates.R
\name{rejection_rates}
\alias{rejection_rates}
\alias{rejection_rates.default}
\alias{rejection_rates.power4test}
\alias{rejection_rates.power4test_by_es}
\alias{rejection_rates.power4test_by_n}
\alias{rejection_rates.x_from_power}
\alias{rejection_rates.n_region_from_power}
\alias{rejection_rates.q_power_mediation}
\alias{print.rejection_rates_df}
\title{Rejection Rates}
\usage{
rejection_rates(object, ...)

\method{rejection_rates}{default}(object, ...)

\method{rejection_rates}{power4test}(
  object,
  all_columns = FALSE,
  ci = TRUE,
  level = 0.95,
  se = FALSE,
  collapse = NULL,
  at_least_k = NULL,
  merge_all_tests = NULL,
  p_adjust_method = NULL,
  alpha = NULL,
  keep_nrep = FALSE,
  ...
)

\method{rejection_rates}{power4test_by_es}(
  object,
  all_columns = FALSE,
  ci = TRUE,
  level = 0.95,
  se = FALSE,
  nrep_if_diff = TRUE,
  ...
)

\method{rejection_rates}{power4test_by_n}(
  object,
  all_columns = FALSE,
  ci = TRUE,
  level = 0.95,
  se = FALSE,
  nrep_if_diff = TRUE,
  ...
)

\method{rejection_rates}{x_from_power}(object, ...)

\method{rejection_rates}{n_region_from_power}(object, ...)

\method{rejection_rates}{q_power_mediation}(object, ...)

\method{print}{rejection_rates_df}(x, digits = 3, annotation = TRUE, abbreviate_col_names = TRUE, ...)
}
\arguments{
\item{object}{The object
from which the rejection rates
are to be computed, such as
a \code{power4test} object,
a \code{power4test_by_n} object,
or a \code{power4test_by_es} object.}

\item{...}{Optional arguments. For
the \code{print} method, these arguments
will be passed to the \code{print} method
of \code{data.frame} objects \code{\link[=print.data.frame]{print.data.frame()}}.
For the \code{rejection_rates} method
of \code{x_from_power} and \code{n_region_from_power} objects, these are
arguments to be passed to the
\code{rejection_rates} method
for \code{power4test_by_n}
and \code{power4test_by_es} objects.
For the \code{rejection_rates} method for
\code{q_power_mediation} objects (the
output of \code{\link[=q_power_mediation]{q_power_mediation()}}
and friends), they are optional arguments
to be passed to the corresponding methods.}

\item{all_columns}{If \code{TRUE}, all
columns stored by a test will be
extracted. Default is \code{FALSE} and
only essential columns related to
power will be printed.}

\item{ci}{If \code{TRUE}, confidence
intervals for the rejection rates
(column \code{reject} or \code{sig}) will
be computed. The method is determined
by the option \code{power4mome.ci_method}.
If \code{NULL} or \code{"wilson"}, Wilson's
(1927) method is used. If
\code{"norm"}, normal approximation
is used.}

\item{level}{The level of confidence
for the confidence intervals, if
\code{ci} is \code{TRUE}. Default is .95,
denoting 95\%.}

\item{se}{If \code{TRUE}, standard errors
for the rejection rates
(column \code{reject} or \code{sig}) will
be computed. Normal approximation
is used to compute the standard
errors.}

\item{collapse}{Whether a single
decision (significant vs. not significant)
is made across all tests for a test
that consists of several tests
(e.g., the tests of several parameters).
If \code{"none"}, tests will be summarized
individually. If \code{"all_sig"}, then
the set of tests is considered significant
if all individual tests are significant.
If \code{"at_least_one_sig"}, then the set of
tests is considered significant if
at least one of the tests is significant.
If \code{"at_least_k_sig"}, then the set of
tests is considered significant if
at least \code{k} tests are significant,
\code{k} set by the argument \code{at_least_k}.
If \code{NULL}, will use the value stored
in \code{object} (default is \code{"none"}).}

\item{at_least_k}{Used by \code{collapse},
the number of tests required to be
significant for the set of tests to
be considered significant. If \code{NULL},
will use the value stored in \code{object}
(default is 1).}

\item{merge_all_tests}{If \code{TRUE}, all
the tests in each replication will be
merged into one test.
If \code{NULL}, will use the value stored
in \code{object} (default is \code{FALSE}).}

\item{p_adjust_method}{The method to be
passed to \code{\link[=p.adjust]{p.adjust()}} to adjust the
\emph{p}-values when testing the effects.
Default is \code{"none"} and the \emph{p}-values
will not be adjusted. Ignored if some tests
do not have \emph{p}-values stored.
NOTE: Use this only if all tests
can be  conducted using \emph{p}-values.
If \code{NULL}, will use the value stored in
\code{object} (default is \code{"none"}).}

\item{alpha}{The level of significance
to use when using \code{p_adjust_method}.
The significance results (the column
\code{sig}) will be updated using the
adjusted \emph{p}-values. Used only if
\code{p_adjust_method} is not \code{"none"}.
If \code{NULL}, will use the value stored
in \code{object} (default is .05).}

\item{keep_nrep}{If \code{TRUE}, the column
\code{nrep} will be kept.}

\item{nrep_if_diff}{If \code{TRUE} and
the numbers of replications across
rows are different, \code{nrep} will be
included in the output. If \code{FALSE},
\code{nrep} will not be included unless
\code{all_columns} is \code{TRUE}.}

\item{x}{The \code{rejection_rates_df}
object to be printed.}

\item{digits}{The number of digits
to be printed
after the decimal.}

\item{annotation}{Logical. Whether
additional notes will be printed.}

\item{abbreviate_col_names}{Logical.
Whether some column names will be
abbreviated.}
}
\value{
The \code{rejection_rates} method returns
a \code{rejection_rates_df} object,
with a \code{print} method.

If the input (\code{object}) is a
\code{power4test} object, the
\code{rejection_rates_df} object is
a data-frame like object with the
number of
rows equal to the number of tests.
Note that some tests, such as
the test by \code{\link[=test_parameters]{test_parameters()}},
conduct one test for each parameter.
Each such test is counted as one
test.

The data frame has at least these columns:
\itemize{
\item \code{test}: The name of the test.
\item \code{label}: The label for each
test, or \code{"Test"} if a test only
does one test (e.g., \code{\link[=test_indirect_effect]{test_indirect_effect()}}).
\item \code{pvalid}: The proportion of valid
tests across all replications.
\item \code{reject}: The rejection rate for
each test. If the null hypothesis
is false, then this is the power.
}

The \code{rejection_rates} method
for \code{power4test_by_es} objects
returns an object of the
class \code{rejection_rates_df_by_es},
which is a subclass of
\code{rejection_rates_df}.
It is a data frame which is
similar to the output of
\code{\link[=rejection_rates]{rejection_rates()}}, with two
columns added for the effect size (\code{pop_es_name} and
\code{pop_es_values})
for each test.

The \code{rejection_rates} method
for \code{power4test_by_n} objects
returns an object of the
class \code{rejection_rates_df_by_n},
which is a subclass of
\code{rejection_rates_df}.
It is a data frame which is
similar to the output of
a \code{power4test} object, with a
column \code{n} added for the sample size
for each test.

The \code{rejection_rates} method
for \code{x_from_power} objects
retrieves the stored
\code{power4test_by_n} or
\code{power4test_by_es} object,
and then
runs \code{rejection_rates} on it
and returns the result.

The \code{rejection_rates} method
for \code{n_region_from_power} objects
retrieves the stored
\code{power4test_by_n} object from
the \code{above} element (the search
for the region with power significantly
above the target power)
and then
runs \code{rejection_rates} on it
and returns the result.

The \code{rejection_rates} method
for \code{q_power_mediation} objects
retrieves the trials from
and then
runs \code{rejection_rates} on them
and returns the result. If \code{mode}
is \code{"n"}, then the stored output
from \code{\link[=n_from_power]{n_from_power()}} is used.
If \code{mode} is \code{"region"}, then
the stored output from \code{\link[=n_region_from_power]{n_region_from_power()}}
is used. If \code{mode} is \code{"power"}, then
the stored output from \code{\link[=power4test]{power4test()}}
is used.

The \code{print} method of a
\code{rejection_rates_df} object returns
the object invisibly. It is called
for its side-effect.
}
\description{
Get all rejection rates
of all tests stored in a \code{power4test}
object or other supported objects.
}
\details{
For a \code{power4test} object,
\link{rejection_rates} loops over the tests stored
in a \code{power4test} object and retrieves
the rejection rate of each test.

The \code{rejection_rates} method for
\code{power4test_by_es} objects
is used to compute the rejection
rates from a \code{power4test_by_es}
object, with effect sizes added to
the output.

The \code{rejection_rates} method for
\code{power4test_by_n} objects
is used to compute the rejection
rates, with sample sizes added to
the output.

The \code{rejection_rates} method for
\code{x_from_power} objects
is used to compute the rejection
rates for stored trials. It supports
the output of \code{\link[=x_from_power]{x_from_power()}} and
its wrappers, such as \code{\link[=n_from_power]{n_from_power()}}.

The \code{rejection_rates} method for
\code{n_region_from_power} objects
is used to compute the rejection
rates for stored trials. It supports
the output of \code{\link[=n_region_from_power]{n_region_from_power()}}.
It is sufficient to retrieve the
trials in searching for the upper bound
because they also include the trials
used in searching for the lower bound.

The \code{rejection_rates} method for
\code{q_power_mediation} objects
is used to compute the rejection
rates for stored trials.
}
\examples{

# Specify the population model

model_simple_med <-
"
m ~ x
y ~ m + x
"

# Specify the effect sizes (population parameter values)

model_simple_med_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Generate some datasets to check the model

sim_only <- power4test(nrep = 4,
                       model = model_simple_med,
                       pop_es = model_simple_med_es,
                       n = 100,
                       R = 50,
                       ci_type = "boot",
                       fit_model_args = list(fit_function = "lm"),
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the test 'test_indirect_effect' on each datasets

test_out <- power4test(object = sim_only,
                       test_fun = test_indirect_effect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        boot_ci = TRUE,
                                        mc_ci = FALSE))

# Do the test 'test_parameters' on each datasets
# and add the results to 'test_out'

test_out <- power4test(object = test_out,
                       test_fun = test_parameters)

# Compute and print the rejection rates for stored tests

rejection_rates(test_out)

# See the help pages of power4test_by_n() and power4test_by_es()
# for other examples.

}
\references{
Wilson, E. B. (1927). Probable inference, the law of
succession, and statistical inference.
\emph{Journal of the American Statistical Association, 22}(158),
209-212. \doi{10.1080/01621459.1927.10502953}
}
\seealso{
\code{\link[=power4test]{power4test()}},
\code{\link[=power4test_by_n]{power4test_by_n()}}, and
\code{\link[=power4test_by_es]{power4test_by_es()}}, which are
supported by this method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim_data.R
\name{sim_data}
\alias{sim_data}
\alias{print.sim_data}
\alias{pool_sim_data}
\title{Simulate Datasets Based on a Model}
\usage{
sim_data(
  nrep = 10,
  ptable = NULL,
  model = NULL,
  pop_es = NULL,
  ...,
  n = 100,
  iseed = NULL,
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  x_fun = list(),
  e_fun = list(),
  process_data = NULL,
  parallel = FALSE,
  progress = FALSE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  n_ratio = 1,
  cl = NULL
)

\method{print}{sim_data}(
  x,
  digits = 3,
  digits_descriptive = 2,
  data_long = TRUE,
  fit_to_all_args = list(),
  est_type = "standardized",
  variances = NULL,
  pure_x = TRUE,
  pure_y = TRUE,
  ...
)

pool_sim_data(object, as_list = FALSE)
}
\arguments{
\item{nrep}{The number of replications
to generate the simulated datasets.
Default is 10.}

\item{ptable}{The output of
\code{\link[=ptable_pop]{ptable_pop()}}, which is a
\code{ptable_pop} object, representing the
population model. If \code{NULL}, the
default, \code{\link[=ptable_pop]{ptable_pop()}} will be
called to generate the \code{ptable_pop}
object, using arguments such as
\code{model} and \code{pop_es}.}

\item{model}{The \code{lavaan} model
syntax of the population model.
Ignored if \code{ptable} is
specified. See \link{ptable_pop} on
how to specify this argument.}

\item{pop_es}{The character to
specify population effect sizes.
See \link{ptable_pop} on
how to specify this argument.
Ignored if \code{ptable} is
specified.}

\item{...}{For \link{sim_data}, parameters
to be passed to \code{\link[=ptable_pop]{ptable_pop()}}. For
\code{\link[=print.sim_data]{print.sim_data()}}, these arguments
are ignored.}

\item{n}{The sample size for each
dataset. Default is 100.}

\item{iseed}{The seed for the random
number generator. Default is \code{NULL}
and the seed is not changed.}

\item{number_of_indicators}{A named
vector to specify the number of
indicators for each factors. See
the help page on how to set this
argument. Default is \code{NULL} and all
variables in the model syntax are
observed variables.
See the help page on how
to use this argument.}

\item{reliability}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the reliability coefficient
of each set of indicators. Default
is \code{NULL}.
See the help page on how
to use this argument.}

\item{loading_difference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the difference in factor
loadings between neighboring indicators
of each set of indicators. Default
is \code{NULL}, and all indicators of
a factor have
the same factor loadings. If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal.}

\item{reference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to indicate which indicator will
be the first indicator (and so
is the reference indicator, by default).
Default
is \code{NULL}, and for all factors,
the indicator with
the medium loading in a factor is
the first indicator. Has no effect
if loading difference is zero (and
so all indicators have the same
loadings). If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal. Accepted
values are \code{"medium"}, \code{"weakest"},
and \code{"strongest"}.}

\item{x_fun}{The function(s) used to
generate the exogenous variables or
error terms. If
not supplied, or set to \code{list()}, the
default, the variables are generated
from a multivariate normal
distribution. See the help page on how
to use this argument.}

\item{e_fun}{The function(s) used to
generate the error terms of indicators,
if any. If
not supplied, or set to \code{list()}, the
default, the error terms of indicators
are generated
from a multivariate normal
distribution. Specify in the same
way as \code{x_fun}. Refer to the help
page on \code{x_fun} on how to use this
argument.}

\item{process_data}{If not \code{NULL}, it
must be a named list with these
elements: \code{fun} (required), the function
to further process the simulated
data, such as generating missing data using
functions such as \code{\link[mice:ampute]{mice::ampute()}}; \code{args} (optional), a
named list of arguments to be passed
to \code{fun}, except the one for the
source data; \code{sim_data_name} (optional) the
name of the argument to receive the
simulated data (e.g., \code{"data"} for
\code{\link[mice:ampute]{mice::ampute()}}), default to
\code{"data"} if it is not set; \code{processed_data_name}
(optional), the name of the data frame
after being processed by \code{fun},
such as the data frame
with missing data in the output of
\code{fun} (e.g., \code{"amp"} for \code{\link[mice:ampute]{mice::ampute()}}),
if omitted, the output of \code{fun} should
be the data frame with missing data.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to simulate
the datasets. Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
of data simulation will be displayed.
Default is `FALSE.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{n_ratio}{If the model is a
multigroup model, and \code{n} is a single
number, this should be a numeric
vector used to determine the sample
size for each group. For example,
for a two-group model, if \code{n} is 100
and \code{n_ratio} is \code{c(1, 0.5)}, then
the sample sizes for the two groups
are 100 and 50, respectively. If
equal to 1, then all groups have the
same sample size.}

\item{cl}{A cluster, such as one created
by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}. If \code{NULL},
a cluster will be created, but will be
stopped on exit. If set to an existing
cluster, it will not be stopped when
the function exits; users need to
stop it manually.}

\item{x}{The \code{sim_data} object
to be printed.}

\item{digits}{The numbers of digits
displayed after the decimal.}

\item{digits_descriptive}{The
number of digits displayed after
the decimal for the descriptive
statistics table.}

\item{data_long}{If \code{TRUE}, detailed
information will be printed.}

\item{fit_to_all_args}{A named list
of arguments to be passed to
\code{\link[lavaan:sem]{lavaan::sem()}} when the model is
fitted to a sample combined from
all samples stored.}

\item{est_type}{The type of estimates
to be printed. Can be a character
vector of one to two elements. If
only \code{"standardized"}, then the
standardized estimates are printed.
If only \code{"unstandardized"}, then the
unstandardized estimates are printed.
If a vector like
\code{c("standardized", "unstandardized")},
then both unstandardized and
standardized estimates are printed.}

\item{variances}{Logical. Whether
variances and error variances are printed.
Default depends on \code{est_type}. If
\code{"unstandardized"} is in \code{est_type},
then default is \code{TRUE} If
only \code{"standardized"} is in \code{est_type},
then default is \code{FALSE}.}

\item{pure_x, pure_y}{When Logical. When
printing indirect effects, whether
only "pure" x-variables (variables
not predicted by another other variables)
and/or "pure" y-variables (variables
that do not predict any other variables
other than indicators) will be included
in enumerating the paths.}

\item{object}{Either a \code{sim_data}
object or a \code{power4test} object.
It extracts the simulated data
and return them, combined to one
single data frame or, if \code{as_list}
is \code{TRUE}, as a list of data
frames.}

\item{as_list}{Logical. If \code{TRUE},
the simulated datasets is returned as one
single data frame. If \code{FALSE}, they
are returned as a list of data
frames.}
}
\value{
The function \code{\link[=sim_out]{sim_out()}} returns
a list of the class \code{sim_data},
with length \code{nrep}. Each element
is a \code{sim_data_i} object, with
the following major elements:
\itemize{
\item \code{ptable}: A \code{lavaan} parameter
table of the model, with population
values set in the column \code{start}.
(It is the output of the
function \code{\link[=ptable_pop]{ptable_pop()}}.)
\item \code{mm_out}: The population model
represented by model matrices
as in \code{lavaan}. (It is the output
of the function
\code{\link[=model_matrices_pop]{model_matrices_pop()}}.)
\item \code{mm_lm_out}: A list of regression
model formula, one for each
endogenous variable. (It is the
output of the internal function
\code{mm_lm()}.)
\item \code{mm_lm_dat_out}: A simulated dataset
generated from the population model.
(It is the output of the internal
function \code{mm_lm_data()}).
\item \code{model_original}: The original model
syntax (i.e., the argument \code{model}).
\item \code{model_final}: A modified model
syntax if the model is a latent
variable model. Indicators are added
to the syntax.
\item \code{fit0}: The output of \code{\link[lavaan:sem]{lavaan::sem()}}
with \code{ptable} as the model and
\code{do.fit} set to \code{FALSE}. Used for the
easy retrieval of information
about the model.
}

The \code{print} method of \code{sim_data}
returns \code{x} invisibly. It is called for
its side effect.

The function \code{pool_sim_data()} returns
either one data frame or a list
of data frames, depending on the
argument \code{as_list}
}
\description{
Get a model matrix and
effect size specification and
simulate a number of datasets,
along with other information.

The function
}
\details{
The function \code{\link[=sim_data]{sim_data()}} generates
a list of datasets based on a population
model.
}
\section{The role of \code{sim_data()}}{
The function \code{\link[=sim_data]{sim_data()}} is used by
the all-in-one function
\code{\link[=power4test]{power4test()}}. Users usually do not
call this function directly, though
developers can use this function to
develop other functions for power
analysis, or to build their own
workflows to do the power analysis.
}

\section{Workflow}{
The function \code{\link[=sim_data]{sim_data()}} does two tasks:
\itemize{
\item Determine the actual population
model with population values based
on:
\itemize{
\item A model syntax for the observed
variables (for a path model)
or the latent factors (for a
latent variable model).
\item A textual specification of the
effect sizes of parameters.
\item The number of indicators for
each latent factor if the model
is a latent variable model.
\item The reliability of each latent
factor as measured by the
indicators if the model is a
latent factor model.
}
\item Generate \emph{nrep} simulated datasets
from the population model.
}

The simulated datasets can then be
used to fit a model, test
parameters, and estimate power.

The output is usually used by
\code{\link[=fit_model]{fit_model()}} to fit a target model,
by default the population model, to each
of the dataset.
}

\section{Set 'number_of_indicators' and 'reliability'}{
The arguments \code{number_of_indicators}
and \code{reliability} are used to
specify the number of indicators
(e.g., items) for each factor,
and the population reliability
coefficient of each factor,
if the variables in the model
syntax are latent variables.

Optionally, \code{loading_difference} can
be used to generate indicators with
unequal standardized factor loadings,
and \code{reference} can be used to specify
the indicator with the medium,
strongest, or weakest standardized
factor loading as the first indicator,
which is used as the reference indicator
in \code{lavaan}.
\subsection{Single-Group Model}{

If a variable in the model is to be
replaced by indicators in the generated
data, set
\code{number_of_indicators} to a named
numeric vector. The names are the
variables of variables with
indicators, as appeared in the
\code{model} syntax. The value of each
name is the number of indicators.

The
argument \code{reliability} should then be
set a named numeric vector (or list,
see the section on multigroup models)
to specify the population reliability
coefficient ("omega") of each set of
indicators. The population standardized factor
loadings are then computed to ensure
that the population reliability
coefficient is of the target value.

These are examples for a single group
model:

\preformatted{number of indicator = c(m = 3, x = 4, y = 5)}

The numbers of indicators for \code{m},
\code{x}, and \code{y} are 3, 4, and 5,
respectively.

\preformatted{reliability = c(m = .90, x = .80, y = .70)}

The population reliability
coefficients of \code{m}, \code{x}, and \code{y} are
.90, .80, and .70, respectively.
}

\subsection{Multigroup Models}{

Multigroup models are supported.
The number of groups is inferred
from \code{pop_es} (see the help page
of \code{\link[=ptable_pop]{ptable_pop()}}), or directly
from \code{ptable}.

For a multigroup model, the number
of indicators for each variable
must be the same across groups.

However, the population reliability
coefficients can be different
across groups. For a multigroup model
of \emph{k} groups,
with one or more population reliability
coefficients differ across groups,
the argument \code{reliability} should be
set to a named list. The names are
the variables to which the population
reliability coefficients are to be
set. The element for each name is
either a single value for the common
reliability coefficient, or a
numeric vector of the reliability
coefficient of each group.

This is an example of \code{reliability}
for a model with 2 groups:

\preformatted{reliability = list(x = .80, m = c(.70, .80))}

The reliability coefficients of \code{x} are
.80 in all groups, while the
reliability coefficients of \code{m} are
.70 in one group and .80 in another.
}

\subsection{Equal Numbers of Indicators and/or Reliability Coefficients}{

If all variables in the model has
the same number of indicators,
\code{number_of_indicators} can be set
to one single value.

Similarly, if all sets of indicators
have the same population reliability
in all groups, \code{reliability} can also
be set to one single value.
}
}

\section{Specify The Distributions of Exogenous Variables Or Error Terms Using 'x_fun'}{
By default, variables and error
terms are generated
from a multivariate normal distribution.
If desired, users can supply the
function used to generate an exogenous
variable and error term by setting \code{x_fun} to
a named list.

The names of the list are the variables
for which a user function will be used
to generate the data.

Each element of the list must also
be a list. The first element of this
list, can be unnamed, is the
function to be used. If other arguments
need to be supplied, they should be
included as named elements of this list.

For example:

\preformatted{x_fun = list(x = list(power4mome::rexp_rs),
             w = list(power4mome::rbinary_rs,
                      p1 = .70)))}

The variables \code{x} and \code{w} will be
generated by user-supplied functions.

For \code{x}, the function is
\code{power4mome::rexp_rs}. No additional
argument when calling this function.

For \code{w}, the function is
\code{power4mome::rbinary_rx}. The argument
\code{p1 = .70} will be passed to this
function when generating the values
of \code{w}.

If a variable is an endogenous
variable (e.g., being predicted by
another variable in a model), then
\code{x_fun} is used to generate its
\emph{error term}. Its implied population
distribution may still be different
from that generate by \code{x_fun} because
the distribution also depends on the
distribution of other variables
predicting it.

These are requirements for the
user-functions:
\itemize{
\item They must return a numeric vector.
\item They mush has an argument \code{n} for
the number of values.
\item The \emph{population} mean and standard
deviation of the generated values
must be 0 and 1, respectively.
}

The package \code{power4mome} has
helper functions for generating
values from some common nonnormal
distributions and then scaling them
to have population mean and standard
deviation equal to 0 and 1 (by default), respectively.
These are some of them:
\itemize{
\item \code{\link[=rbinary_rs]{rbinary_rs()}}.
\item \code{\link[=rexp_rs]{rexp_rs()}}.
\item \code{\link[=rbeta_rs]{rbeta_rs()}}.
\item \code{\link[=rlnorm_rs]{rlnorm_rs()}}.
\item \code{\link[=rpgnorm_rs]{rpgnorm_rs()}}.
}

To use \code{x_fun}, the variables must
have zero covariances with other
variables in the population. It is
possible to generate nonnormal
multivariate data but we believe this
is rarely needed when estimating
power \emph{before} having the data.
}

\section{Specify the Population Model by 'model'}{
\subsection{Single-Group Model}{

For a single-group model, \code{model}
should be a \code{lavaan} model syntax
string of the \emph{form} of the model.
The population values of the model
parameters are to be determined by
\code{pop_es}.

If the model has latent factors,
the syntax in \code{model} should specify
only the \emph{structural model} for the
\emph{latent factors}. There is no
need to specify the measurement
part. Other functions will generate
the measurement part on top of this
model.

For example, this is a simple mediation
model:

\preformatted{"m ~ x
 y ~ m + x"}

Whether \code{m}, \code{x}, and \code{y} denote
observed variables or latent factors
are determined by other functions,
such as \code{\link[=power4test]{power4test()}}.
}

\subsection{Multigroup Model}{

Because the model is the population
model, equality constraints are
irrelevant and the model syntax
specifies only the \emph{form} of the
model. Therefore, \code{model} is
specified as in the case of single
group models.
}
}

\section{Specify 'pop_es' Using Named Vectors}{
The argument \code{pop_es} is for specifying
the population values of model
parameters. This section describes
how to do this using named vectors.
\subsection{Single-Group Model}{

If \code{pop_es} is specified by a named
vector, it must follow the convention
below.
\itemize{
\item The names of the vectors are
\code{lavaan} names for the selected
parameters. For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
\item Alternatively, the names can be
either \code{".beta."} or \code{".cov."}.
Use \code{".beta."} to set the default
values for all regression coefficients.
Use \code{".cov."} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
\item The names can also be of this form:
\code{".ind.(<path>)"}, whether \verb{<path>}
denote path in the model. For
example, \code{".ind.(x->m->y)"} denotes
the path from \code{x} through \code{m} to
\code{y}. Alternatively, the \code{lavaan}
symbol \code{~} can also be used:
\code{".ind.(y~m~x)"}. This form is used
to set the indirect effect (standardized,
by default) along this path. The
value for this name will override
other settings.
\item If using \code{lavaan} names, can
specify more than one parameter
using \code{+}. For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
\item The value of each element can be
the label for the effect size: \code{n}
for nil, \code{s} for small, \code{m} for
medium, and \code{l} for large. The
value for each label is determined
by \code{es1} and \code{es2}. See the section
on specifying these two arguments.
\item The value of \code{pop_es} can also be
set to a value, but it must be
quoted as a string, such as \code{"y ~ x" = ".31"}.
}

This is an example:

\preformatted{c(".beta." = "s",
  "m1 ~ x" = "-m",
  "m2 ~ m1" = "l",
  "y ~ x:w" = "s")}

In this example,
\itemize{
\item All regression coefficients are
set to small (\code{s}) by default, unless
specified otherwise.
\item The path from \code{x} to \code{m1} is
set to medium and negative (\code{-m}).
\item The path from \code{m1} to \code{m2} is set
to large (\code{l}).
\item The coefficient of the product
term \code{x:w} when predicting \code{y} is
set to small (\code{s}).
}
\subsection{Indirect Effect}{

When setting an indirect effect to
a symbol (default: \code{"si"}, \code{"mi"},
\code{"li"}, with \code{"i"} added to differentiate
them from the labels for a direct path),
the corresponding value is used to
determine the population values of
\emph{all} component paths along the pathway.
All the values are assumed to be equal.
Therefore, \code{".ind.(x->m->y)" = ".20"}
is equivalent to setting \code{m ~ x}
and \code{y ~ m} to the square root of
.20, such that the corresponding
indirect effect is equal to the
designated value.

This behavior, though restricted,
is for quick manipulation of the
indirect effect. If different values
along a pathway, set the value for
each path directly.

Only nonnegative value is supported.
Therefore, \code{".ind.(x->m->y)" = "-si"}
and \code{".ind.(x->m->y)" = "-.20"} will
throw an error.
}

}

\subsection{Multigroup Model}{

The argument \code{pop_es} also supports multigroup
models.

For \code{pop_es}, instead of
named vectors, named \emph{list} of
named vectors should be used.
\itemize{
\item The names are the parameters, or
keywords such as \code{.beta.} and
\code{.cov.}, like specifying the
population values for a single
group model.
\item The elements are character vectors.
If it has only one element (e.g.,
a single string), then it is the
the population value for all groups.
If it has more than one element
(e.g., a vector of three strings),
then they are the population values
of the groups. For a model of \emph{k}
groups, each vector must have
either \emph{k} elements or one element.
}

This is an example:

\preformatted{list("m ~ x" = "m",
     "y ~ m" = c("s", "m", "l"))}

In this model, the population value
of the path \code{m ~ x} is medium (\code{m}) for
all groups, while the population
values for the path \code{y ~ m} are
small (\code{s}), medium (\code{m}), and large (\code{l}),
respectively.
}
}

\section{Specify 'pop_es' Using a Multiline String}{
When setting the argument \code{pop_es},
instead of using a named vector
or named list for
\code{pop_es}, the population values of
model parameters can also be
specified using a multiline string,
as illustrated below, to be parsed
by \code{\link[=pop_es_yaml]{pop_es_yaml()}}.
\subsection{Single-Group Model}{

This is an example of the multiline string
for a single-group model:

\preformatted{y ~ m: l
m ~ x: m
y ~ x: nil}

The string must follow this format:
\itemize{
\item Each line starts with \verb{tag:}.
\itemize{
\item \code{tag} can be the name of a
parameter, in \code{lavaan} model
syntax format.
\itemize{
\item For example, \code{m ~ x}
denotes the path from \code{x} to \code{m}.
}
\item A tag in \code{lavaan} model syntax can
specify more than one parameter
using \code{+}.
\itemize{
\item For example, \code{y ~ m + x}
denotes the two paths from \code{m} and
\code{x} to \code{y}.
}
\item Alternatively, the \code{tag} can be
either \code{.beta.} or \code{.cov.}.
\itemize{
\item Use \code{.beta.} to set the default
values for all regression coefficients.
\item Use \code{.cov.} to set the default
values for all correlations of
exogenous variables (e.g., predictors).
}
}
\item After each tag is the value of the
population value:

-\code{nil} for nil (zero),
\itemize{
\item \code{s} for small,
\item \code{m} for medium, and
\item \code{l} for large.
\item \code{si}, \code{mi}, and \code{li} for
small, medium, and large a
standardized indirect effect,
respectively.
}

Note: \code{n} \emph{cannot} be used in this mode.

The
value for each label is determined
by \code{es1} and \code{es2} as described
in \code{\link[=ptable_pop]{ptable_pop()}}.
\itemize{
\item The value can also be
set to a numeric value, such as
\code{.30} or \code{-.30}.
}
}

This is another example:

\preformatted{.beta.: s
y ~ m: l}

In this example, all regression
coefficients are \code{small}, while
the path from \code{m} to \code{y} is large.
}

\subsection{Multigroup Model}{

This is an example of the string
for a multigroup model:

\preformatted{y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil}

The format is similar to that for
a single-group model. If a parameter
has the same value for all groups,
then the line can be specified
as in the case of a single-group
model: \code{tag: value}.

If a parameter has different
values across groups, then it must
be in this format:
\itemize{
\item A line starts with the tag, followed
by two or more lines. Each line
starts with a hyphen \code{-} and the
value for a group.
}

For example:

\preformatted{m ~ x:
  - nil
  - s}

This denotes that the model has
two groups. The values of the path
from \code{x} to \code{m} for the two
groups are 0 (\code{nil}) and
small (\code{s}), respectively.

Another equivalent way to specify
the values are using \verb{[]}, on
the same line of a tag.

For example:

\preformatted{m ~ x: [nil, s]}

The number of groups is inferred
from the number of values for
a parameter. Therefore, if a tag
has more than one value, each tag
must has the same number of value,
or only one value.

The tag \code{.beta.} and \code{.cov.} can
also be used for multigroup models.
}

\subsection{Which Approach To Use}{

Note that using named vectors or
named lists is more reliable. However,
using a multiline string is
more user-friendly. If this method
failed, please use named vectors or
named list instead.
}

\subsection{Technical Details}{

The multiline string is parsed by \code{\link[yaml:read_yaml]{yaml::read_yaml()}}.
Therefore, the format requirement
is actually that of YAML. Users
knowledgeable of YAML can use other
equivalent way to specify the string.
}
}

\section{Set the Values for Effect Size Labels ('es1' and 'es2')}{
The vector \code{es1} is for correlations,
regression coefficients, and
indirect effect, and the
vector \code{es2} is for for standardized
moderation effect, the coefficients
of a product term. These labels
are to be used in interpreting
the specification in \code{pop_es}.
}

\examples{

# Specify the model

mod <-
"m ~ x
 y ~ m + x"

# Specify the population values

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate the simulated datasets

data_all <- sim_data(nrep = 5,
                     model = mod,
                     pop_es = es,
                     n = 100,
                     iseed = 1234)

data_all

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbeta_rs.R
\name{rbeta_rs}
\alias{rbeta_rs}
\title{Random Variable From a Beta Distribution}
\usage{
rbeta_rs(n = 10, shape1 = 0.5, shape2 = 0.5, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{shape1}{\code{shape1} for
\code{\link[stats:Beta]{stats::rbeta()}}.}

\item{shape2}{\code{shape2} for
\code{\link[stats:Beta]{stats::rbeta()}}.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from a beta distribution, rescaled to
have user-specified population mean
and standard deviation.
}
\details{
First, specify the two parameters,
\code{shape1} and \code{shape2}, and the
desired population mean and standard
deviation. The random numbers, drawn
from a beta distribution by
\code{\link[stats:Beta]{stats::rbeta()}} will then be
rescaled with the desired population
mean and standard deviation.
}
\examples{
set.seed(90870962)
x <- rbeta_rs(n = 5000,
              shape1 = .5,
              shape2 = .5,
              pmean = 3,
              psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power4mome-package.R
\docType{package}
\name{power4mome-package}
\alias{power4mome}
\alias{power4mome-package}
\title{power4mome: Power Analysis for Moderation and Mediation}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Power analysis and sample size determination for moderation, mediation, and moderated mediation in models fitted by structural equation modelling using the 'lavaan' package by Rosseel (2012) \doi{10.18637/jss.v048.i02} or by multiple regression. The package 'manymome' by Cheung and Cheung (2024) \doi{10.3758/s13428-023-02224-z} is used to specify the indirect paths or conditional indirect paths to be tested.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://sfcheung.github.io/power4mome/}
  \item Report bugs at \url{https://github.com/sfcheung/power4mome/issues}
}

}
\author{
\strong{Maintainer}: Shu Fai Cheung \email{shufai.cheung@gmail.com} (\href{https://orcid.org/0000-0002-9871-9448}{ORCID})

Authors:
\itemize{
  \item Sing-Hang Cheung (\href{https://orcid.org/0000-0001-5182-0752}{ORCID})
  \item Wendie Yang (\href{https://orcid.org/0009-0000-8388-6481}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_conditional_indirect_effects.R
\name{test_cond_indirect_effects}
\alias{test_cond_indirect_effects}
\title{Test Several Conditional Indirect Effects}
\usage{
test_cond_indirect_effects(
  fit = fit,
  x = NULL,
  m = NULL,
  y = NULL,
  wlevels = NULL,
  mc_ci = TRUE,
  mc_out = NULL,
  boot_ci = FALSE,
  boot_out = NULL,
  check_post_check = TRUE,
  test_method = NULL,
  compare_groups = FALSE,
  ...,
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[manymome:cond_indirect]{manymome::cond_indirect_effects()}}.}

\item{x}{The name of the \code{x}-variable,
the predictor.}

\item{m}{A character vector of the
name(s) of mediator(s). The path
moves from the first mediator in the
vector to the last mediator in the
vector. Can be \code{NULL} and the path
is a direct path without mediator.}

\item{y}{The name of the \code{y}-variable,
the outcome variable.}

\item{wlevels}{The output of
\code{\link[manymome:merge_mod_levels]{manymome::merge_mod_levels()}}, or
the moderator(s) to be passed to
\code{\link[manymome:mod_levels]{manymome::mod_levels_list()}}.
If all the moderators can be
represented by one variable, that is,
each moderator is (a) a numeric
variable, (b) a dichotomous
categorical variable, or (c) a
factor or string variable used in
\code{\link[stats:lm]{stats::lm()}} in fit, then it is a
vector of the names of the moderators
as appeared in the data frame. If at
least one of the moderators is a
categorical variable represented by
more than one variable, such as
user-created dummy variables used in
\code{\link[lavaan:sem]{lavaan::sem()}}, then it must be a
list of the names of the moderators,
with such moderators represented by
a vector of names. For example:
\verb{list("w1", c("gpgp2", "gpgp3")},
the first moderator \code{w1} and the
second moderator a three-category
variable represented by \code{gpgp2} and
\code{gpgp3}. See the help page of
\code{\link[manymome:cond_indirect]{manymome::cond_indirect_effects()}}
for further details.}

\item{mc_ci}{Logical. If \code{TRUE}, the
default, Monte Carlo confidence
intervals will be formed. This argument
and \code{boot_ci} cannot be both \code{TRUE}.}

\item{mc_out}{The pre-generated
Monte Carlo estimates generated by
\link[manymome:do_mc]{manymome::do_mc}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{boot_ci}{Logical. If \code{TRUE},
the default, nonparametric bootstrap
confidence intervals will be formed.
This argument
and \code{mc_ci} cannot be both \code{TRUE}.}

\item{boot_out}{The pre-generated
bootstrap estimates generated by
\link[manymome:do_boot]{manymome::do_boot}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{test_method}{The method to do
the test. If \code{"ci"}, then the
confidence interval (e.g., bootstrap
confidence interval) will be used to
do the test. If
\code{"pvalue"}, then asymmetric \emph{p}-value
by Asparouhov & Muthén (2021) will
be used to do the test, and the
confidence interval will not be
computed. If \code{NULL}, its value will
be set to \code{"pvalue"} if the number of
simulated/bootstrap samples (\code{R}) is
a value that is supported by the
method by Boos and Zhang (2000), and
set to \code{"ci"} otherwise.}

\item{compare_groups}{If the model
is a multigroup model, compute and
test group differences for all pairwise
combinations of the groups. Ignored
if the model is a single-group model.}

\item{...}{Additional arguments to
be passed to \code{\link[manymome:cond_indirect]{manymome::cond_indirect_effects()}}.}

\item{fit_name}{The name of the
model fit object to be extracted.
Default is \code{"fit"}. Used only when
more than one model is fitted in
each replication. This should be
the name of the model on which the
test is to be conducted.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
the output returned by
\code{\link[manymome:cond_indirect]{manymome::cond_indirect_effects()}},
with the following modifications:
\itemize{
\item \code{est}: The estimated
conditional indirect effects.
\item \code{cilo} and \code{cihi}: The
lower and upper limits of the
confidence interval (95\% by
default), respectively.
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
\item \code{test_label}: A column of labels
generated to label the conditional
effects.
}
}
\description{
Test several conditional
indirect effects
for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing several
conditional
indirect effects, by setting it
to the \code{test_fun} argument.

It uses \code{\link[manymome:cond_indirect]{manymome::cond_indirect_effects()}}
to do the test. It can be used on
models fitted by \code{\link[lavaan:sem]{lavaan::sem()}}
or fitted by a sequence of calls
to \code{\link[stats:lm]{stats::lm()}}, although only
nonparametric bootstrap confidence
interval is supported for models
fitted by regression using
\code{\link[stats:lm]{stats::lm()}}.

It can also be used to test
conditional effects on a direct path
with no mediator. Just omit \code{m} when
calling the function.
}
\examples{

# Specify the model

model_simple_mod <-
"
m ~ x + w + x:w
y ~ m + x
"

# Specify the population values

model_simple_mod_es <-
"
y ~ m: l
y ~ x: n
m ~ x: m
m ~ w: n
m ~ x:w: l
"

# Simulate the data

# Set nrep to a larger value in real analysis, such as 400
sim_only <- power4test(nrep = 5,
                       model = model_simple_mod,
                       pop_es = model_simple_mod_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the tests in each replication

test_out <- power4test(object = sim_only,
                       test_fun = test_cond_indirect_effects,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wlevels = c("w"),
                                        mc_ci = TRUE))
print(test_out,
      test_long = TRUE)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bz_helpers.R
\name{bz_helpers}
\alias{bz_helpers}
\alias{Rs_bz_supported}
\alias{R_for_bz}
\title{Helpers for the Boos-and-Zhang (2000) Method}
\usage{
Rs_bz_supported(
  alpha = 0.05,
  Rmax = getOption("power4mome.bz_Rmax", default = 359)
)

R_for_bz(R_target, alpha = 0.05)
}
\arguments{
\item{alpha}{The level of significance,
two-tailed.}

\item{Rmax}{The maximum number of resamples
to be returned. Default is 359.
Though it is possible to use 1999
resamples or even more with Boos-Zhang-2000,
using such a large number of resamples defeats
the goal to reduce processing time.}

\item{R_target}{The target maximum
number of resamples.}
}
\value{
The function \code{\link[=Rs_bz_supported]{Rs_bz_supported()}} returns
a numeric vector of the numbers of
resamples supported.

The function \code{\link[=R_for_bz]{R_for_bz()}} returns
a scalar.
}
\description{
Helpers for the method
by Boos and Zhang (2000) for estimating
rejection rate for resampling-based
tests.
}
\details{
Boos and Zhang (2000) proposed
a method to estimate the rejection
rate (power, if the null hypothesis is false)
for methods based on resampling, such
as nonparametric bootstrapping. This
method is used by some functions
in \code{power4mome}, such as \code{\link[=x_from_power]{x_from_power()}}.

This method is implemented internally.
Some helper functions regarding this
method is exported for users.

The function \code{\link[=Rs_bz_supported]{Rs_bz_supported()}} returns
the number of bootstrap samples (for bootstrapping)
or simulated samples (for Monte Carlo),
both called resamples below for brevity,
usually specified by the argument \code{R},
that can be
used for the Boos-Zhang-2000 method,
given the desired two-tailed level of significance,
(.05 by default).
If possible, setting the number of resamples.
(e.g., setting \code{R} when calling \code{\link[=x_from_power]{x_from_power()}})
will automatically enable the Boos-Zhang-2000
method (unless explicitly turned off
by setting the option \code{"power4mome.bz"}
to \code{FALSE} by \code{options("power4mome.bz") <- FALSE}),
substantially reducing the processing
time. For now, only two-tailed tests are supported.

Given a target maximum number of
resamples and a level of significance,
the function \code{\link[=R_for_bz]{R_for_bz()}} returns
largest number of resamples that can be
used for the Boos-Zhang-2000 method.
This function can be used for
arguments such as \code{R} in \code{\link[=x_from_power]{x_from_power()}}
to automatically find the largest
value supported by the Boos-Zhang-2000
method.
}
\examples{
# === Rs_bz_supported ===

# alpha = .05
Rs_bz_supported()

# alpha = .01
Rs_bz_supported(alpha = .01)

# === R_for_bz ===

R_for_bz(200)
R_for_bz(500)

}
\references{
Boos, D. D., & Zhang, J. (2000).
Monte Carlo evaluation of
resampling-based hypothesis tests.
\emph{Journal of the American Statistical Association},
\emph{95}(450), 486--492.
\doi{10.1080/01621459.2000.10474226}
}
\seealso{
\code{\link[=x_from_power]{x_from_power()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_data_scale_scores.R
\name{scale_scores}
\alias{scale_scores}
\title{Process Data by Computing Scale Scores}
\usage{
scale_scores(data, method = c("mean", "sum"))
}
\arguments{
\item{data}{A data frame with the
indicator scores. It must has an
attribute \code{number_of_indicators}.
The same argument used by
\code{\link[=power4test]{power4test()}}. This attribute is
used to identify the factor names
and their indicators.}

\item{method}{The method to be used
to compute the scale scores. Can
be \code{"mean"} or \code{"sum"}. The default
\code{na.rm = FALSE} will be used.
Therefore, \code{data} must not have
missing data.}
}
\value{
It returns a data frame with the
scale scores computed.
}
\description{
For the \code{process_data}
argument. To compute scale scores
from indicators and replace the
indicators scores by computed
scale scores.
}
\details{
This function is to be used in
the \code{process_data} argument of
\code{\link[=power4test]{power4test()}}.

It retrieves the attribute
\code{"number_of_indicators"},
stored by \code{\link[=power4test]{power4test()}}, to identify
factors with indicators, computes
the scale scores based on \code{method},
and replace the indicators by the
scale scores.

All subsequent steps, such as the
test functions, will see only the
scale scores, or original scores if
a variable has no indicator. The
model will also be fitted on the
scale scores, not on the indicators.

It can be used to estimate power
for analyzing the scale scores,
taking into account the measurement
error due to imperfect reliability.
}
\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Specify the numbers of indicators and reliability coefficients

k <- c(y = 3,
       m = 4,
       x = 5)
rel <- c(y = .70,
         m = .70,
         x = .70)

# Simulate the data

out <- power4test(
         nrep = 2,
         model = mod,
         pop_es = mod_es,
         n = 200,
         number_of_indicators = k,
         reliability = rel,
         process_data = list(fun = "scale_scores"),
         test_fun = test_parameters,
         test_args = list(op = "~"),
         parallel = FALSE,
         iseed = 1234)

dat <- pool_sim_data(out)
head(dat)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rt_rs.R
\name{rt_rs}
\alias{rt_rs}
\title{Random Variable From a t Distribution}
\usage{
rt_rs(n = 10, df = 5, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{df}{\code{df} for \code{\link[stats:TDist]{stats::rt()}}.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from a \emph{t} distribution, rescaled to
have user-specified population mean
and standard deviation.
}
\details{
First, specify the parameter \code{df}
and the desired population
mean and standard deviation. The
random numbers, drawn from the
generalized normal distribution by
\code{\link[stats:TDist]{stats::rt()}}, will then be
rescaled with the desired population
mean and standard.
}
\examples{
set.seed(90870962)
x <- rt_rs(n = 5000,
           df = 5,
           pmean = 3,
           psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_boot.R
\name{gen_boot}
\alias{gen_boot}
\title{Generate Bootstrap Estimates}
\usage{
gen_boot(
  fit_all,
  R = 100,
  ...,
  iseed = NULL,
  parallel = FALSE,
  progress = FALSE,
  ncores = max(1, parallel::detectCores(logical = FALSE) - 1),
  compute_implied_stats = FALSE,
  cl = NULL
)
}
\arguments{
\item{fit_all}{The output of
\code{\link[=fit_model]{fit_model()}} or an object of the
class \code{fit_out}.}

\item{R}{The number of replications
to generate the bootstrap estimates
for each fit output.}

\item{...}{Optional arguments to be
passed to \code{\link[manymome:do_boot]{manymome::do_boot()}} when
generating the bootstrap estimates.}

\item{iseed}{The seed for the random
number generator. Default is \code{NULL}
and the seed is not changed.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to generate
bootstrap estimates for the
fit outputs. Default is \code{FALSE}.}

\item{progress}{If \code{TRUE}, the progress
will be displayed.
Default is `FALSE.}

\item{ncores}{The number of CPU
cores to use if parallel processing
is used.}

\item{compute_implied_stats}{Whether
implied statistics are computed
in each bootstrap samples. Usually
not needed and so default to \code{FALSE}}

\item{cl}{A cluster, such as one created
by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}. If \code{NULL},
a cluster will be created, but will be
stopped on exit. If set to an existing
cluster, it will not be stopped when
the function exits; users need to
stop it manually.}
}
\value{
A \code{boot_list} object, which is a list
of the output of \code{\link[manymome:do_boot]{manymome::do_boot()}}.
}
\description{
Get a list of the output
of \code{\link[lavaan:sem]{lavaan::sem()}} or \code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}
and generate
bootstrap estimates of model
parameters.
}
\details{
The function \code{\link[=gen_boot]{gen_boot()}}
simply calls \code{\link[manymome:do_boot]{manymome::do_boot()}}
on each output of
\code{\link[lavaan:sem]{lavaan::sem()}} or \code{\link[lmhelprs:many_lm]{lmhelprs::many_lm()}}
in \code{fit_all}. The
simulated
estimates can then be used to test
effects such as indirect effects,
usually by functions from the
\code{manymome} package, such as
\code{\link[manymome:cond_indirect]{manymome::indirect_effect()}}.
}
\section{The role of \code{gen_boot()}}{
This function is used by the
all-in-one function \code{\link[=power4test]{power4test()}}.
Users usually do not call this
function directly, though
developers can use this function to
customize the workflow of the
power analysis.
}

\examples{

# Specify the population model

mod <-
"m ~ x
 y ~ m + x"

# Specify the effect sizes (population parameter values)

es <-
"
y ~ m: m
m ~ x: m
y ~ x: n
"

# Generate several simulated datasets

data_all <- sim_data(nrep = 2,
                     model = mod,
                     pop_es = es,
                     n = 50,
                     iseed = 1234)

# Fit the population model to each datasets

fit_all <- fit_model(data_all)

# Generate bootstrap estimates for each replication

boot_all <- gen_boot(fit_all,
                     R = 10,
                     iseed = 4567)
boot_all


}
\seealso{
See \code{\link[=power4test]{power4test()}} for
the all-in-one function that uses
this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_ordinal.R
\name{ordinal_variables}
\alias{ordinal_variables}
\alias{cut_patterns}
\title{Process Data by Creating Ordinal Variables}
\usage{
ordinal_variables(data, cut_patterns = NULL, cuts = NULL)

cut_patterns(which = NULL)
}
\arguments{
\item{data}{A data frame.}

\item{cut_patterns}{A named vector.
The names are the names of the latent
variables for which indicator scores
will be converted. Each value must be
the name of one of the built-in
patterns (call \code{cut_patterns()} to
list the patterns and their names).
Can be used with \code{cuts} but a latent
variable should appear only either
in \code{cut_patterns} or \code{cuts}, not both.}

\item{cuts}{A named list.
The names are the names of the latent
variables for which indicator scores
will be converted.
Each element is a vector of the
thresholds for the conversion. \code{-Inf}
and \code{Inf} will be automatically
included during the conversion.
Can be used with \code{cut_patterns} but a latent
variable should appear only either
in \code{cut_patterns} or \code{cuts}, not both.}

\item{which}{A name of a built-in pattern.}
}
\value{
The function \code{\link[=ordinal_variables]{ordinal_variables()}}
returns a data frame with the
the converted scores.

The function \code{\link[=cut_patterns]{cut_patterns()}}
returns a named list of the built-in
patterns if \code{which = NULL}, and
a numeric vector of the thresholds
of a built-in pattern if \code{which} is
set to one of the names of the built-in
patterns.
}
\description{
For the \code{process_data}
argument. It converts continuous
indicator variable to ordinal variables.
}
\details{
This function is to be used in
the \code{process_data} argument of
\code{\link[=power4test]{power4test()}}.

It is used for converting the
continuous indicator scores generated
to ordinal scores (with values 1, 2, 3, etc.).

For example, if the cut points (thresholds) are
-2 and 2, then sores will be converted
to three categories: (-Inf to -2],
(-2 to 2], and (2 to Inf]. The intervals
are closed on the right. That is,
a score of -2 is in the interval
(-Inf to -2], not in the interval
(-2 to 2).

The conversion is implemented by \code{\link[=cut]{cut()}}.

There are two ways to specify the
conversion. The first one is to use
some built-in thresholds, based on
Savalei and Rhemtulla (2013), Table 1.
Call \code{\link[=cut_patterns]{cut_patterns()}} with no argument
to list all the built-in patterns
and their names.

Alternatively, users can specify the
thresholds directly through the argument
\code{cuts}.

Currently, all indicators of a latent
variable must be converted in the same
way.
}
\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
y ~ m: l
m ~ x: m
y ~ x: n
"

# Specify the numbers of indicators and reliability coefficients

k <- c(y = 3,
       m = 4,
       x = 5)
rel <- c(y = .70,
         m = .70,
         x = .70)

# Simulate the data

out <- power4test(
         nrep = 2,
         model = mod,
         pop_es = mod_es,
         n = 200,
         number_of_indicators = k,
         reliability = rel,
         process_data = list(fun = "ordinal_variables",
                             args = list(cut_patterns = c(x = "s3"),
                                         cuts = list(m = c(-2, 0, 2)))),
         test_fun = test_parameters,
         test_args = list(op = "~"),
         parallel = FALSE,
         iseed = 1234)

dat <- pool_sim_data(out)
head(dat, 50)

# The built-in patterns

cut_patterns()

}
\references{
Savalei, V., & Rhemtulla, M. (2013).
The performance of robust test
statistics with categorical data.
\emph{British Journal of Mathematical and Statistical Psychology},
\emph{66}(2), 201--223.
\doi{10.1111/j.2044-8317.2012.02049.x}
}
\seealso{
\code{\link[=power4test]{power4test()}}. See also
\code{\link[=cut]{cut()}} for the implementation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_conditional_indirect_effect.R
\name{test_cond_indirect}
\alias{test_cond_indirect}
\title{Test a Conditional Indirect Effect}
\usage{
test_cond_indirect(
  fit = fit,
  x = NULL,
  m = NULL,
  y = NULL,
  wvalues = NULL,
  mc_ci = TRUE,
  mc_out = NULL,
  boot_ci = FALSE,
  boot_out = NULL,
  check_post_check = TRUE,
  test_method = NULL,
  ...,
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[manymome:cond_indirect]{manymome::cond_indirect()}}.}

\item{x}{The name of the \code{x}-variable,
the predictor.}

\item{m}{A character vector of the
name(s) of mediator(s). The path
moves from the first mediator in the
vector to the last mediator in the
vector. Can be \code{NULL} and the path
is a direct path without mediator.}

\item{y}{The name of the \code{y}-variable,
the outcome variable.}

\item{wvalues}{A numeric vector of
named elements. The names are the
variable names of the moderators,
and the values are the values to
which the moderators will be set to.
Default is \code{NULL}.}

\item{mc_ci}{Logical. If \code{TRUE}, the
default, Monte Carlo confidence
intervals will be formed. This argument
and \code{boot_ci} cannot be both \code{TRUE}.}

\item{mc_out}{The pre-generated
Monte Carlo estimates generated by
\link[manymome:do_mc]{manymome::do_mc}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{boot_ci}{Logical. If \code{TRUE},
the default, nonparametric bootstrap
confidence intervals will be formed.
This argument
and \code{mc_ci} cannot be both \code{TRUE}.}

\item{boot_out}{The pre-generated
bootstrap estimates generated by
\link[manymome:do_boot]{manymome::do_boot}, stored in
a \code{power4test} object. Users should
not set this argument and should let
\code{\link[=power4test]{power4test()}} to set it automatically.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{test_method}{The method to do
the test. If \code{"ci"}, then the
confidence interval (e.g., bootstrap
confidence interval) will be used to
do the test. If
\code{"pvalue"}, then asymmetric \emph{p}-value
by Asparouhov & Muthén (2021) will
be used to do the test, and the
confidence interval will not be
computed. If \code{NULL}, its value will
be set to \code{"pvalue"} if the number of
simulated/bootstrap samples (\code{R}) is
a value that is supported by the
method by Boos and Zhang (2000), and
set to \code{"ci"} otherwise.}

\item{...}{Additional arguments to
be passed to \code{\link[manymome:cond_indirect]{manymome::cond_indirect()}}.}

\item{fit_name}{The name of the
model fit object to be extracted.
Default is \code{"fit"}. Used only when
more than one model is fitted in
each replication. This should be
the name of the model on which the
test is to be conducted.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
a named numeric vector with the
following elements:
\itemize{
\item \code{est}: The mean of the estimated
indirect effect across datasets.
\item \code{cilo} and \code{cihi}: The means of the
lower and upper limits of the
confidence interval (95\% by
default), respectively.
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
}
}
\description{
Test a conditional
indirect effect
for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing a
conditional
indirect effect, by setting it
to the \code{test_fun} argument.

It uses \code{\link[manymome:cond_indirect]{manymome::cond_indirect()}}
to do the test. It can be used on
models fitted by \code{\link[lavaan:sem]{lavaan::sem()}}
or fitted by a sequence of calls
to \code{\link[stats:lm]{stats::lm()}}, although only
nonparametric bootstrap confidence
interval is supported for models
fitted by regression using
\code{\link[stats:lm]{stats::lm()}}.

It can also be used to test
a conditional effect on a direct path
with no mediator. Just omit \code{m} when
calling the function.
}
\examples{

# Specify the model

model_simple_mod <-
"
m ~ x + w + x:w
y ~ m + x
"

# Specify the population values

model_simple_mod_es <-
"
y ~ m: l
y ~ x: n
m ~ x: m
m ~ w: n
m ~ x:w: l
"

# Simulate the data

sim_only <- power4test(nrep = 5,
                       model = model_simple_mod,
                       pop_es = model_simple_mod_es,
                       n = 100,
                       R = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the test in each replication

test_ind <- power4test(object = sim_only,
                       test_fun = test_cond_indirect,
                       test_args = list(x = "x",
                                        m = "m",
                                        y = "y",
                                        wvalues = c(w = 1),
                                        mc_ci = TRUE))
print(test_ind,
      test_long = TRUE)

}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/x_from_power_for_x_plot.R
\name{plot.x_from_power}
\alias{plot.x_from_power}
\alias{plot.n_region_from_power}
\title{Plot The Results of 'x_from_power'}
\usage{
\method{plot}{x_from_power}(
  x,
  what = c("ci", "power_curve", "final_x", "final_power", "target_power", switch(x$x, n =
    "sig_area", es = NULL)),
  text_what = c("final_x", "final_power", switch(x$x, n = "sig_area", es = NULL)),
  digits = 3,
  main = NULL,
  xlab = NULL,
  ylab = "Estimated Power",
  ci_level = 0.95,
  pars_ci = list(),
  pars_power_curve = list(),
  pars_ci_final_x = list(lwd = 3, length = 0.2, col = "blue"),
  pars_target_power = list(lty = "dashed", lwd = 2, col = "black"),
  pars_final_x = list(lty = "dotted"),
  pars_final_power = list(lty = "dotted", col = "blue"),
  pars_text_final_x = list(y = 0, pos = 3, cex = 1),
  pars_text_final_power = list(pos = 3, cex = 1),
  pars_sig_area = list(col = adjustcolor("lightblue", alpha.f = 0.1)),
  pars_text_sig_area = list(cex = 1),
  prop_of_trials = NULL,
  min_trials_for_prop = NULL,
  override_for_pba = TRUE,
  ...
)

\method{plot}{n_region_from_power}(
  x,
  what = c("ci", "power_curve", "final_x", "final_power", "target_power", "sig_area"),
  text_what = c("final_x", "final_power", "sig_area"),
  digits = 3,
  main = paste0("Power Curve ", "(Target Power: ", formatC(x$below$target_power, digits =
    digits, format = "f"), ")"),
  xlab = NULL,
  ylab = "Estimated Power",
  ci_level = 0.95,
  pars_ci = list(),
  pars_power_curve = list(),
  pars_ci_final_x = list(lwd = 2, length = 0.2, col = "blue"),
  pars_target_power = list(lty = "dashed", lwd = 2, col = "black"),
  pars_final_x = list(lty = "dotted"),
  pars_final_power = list(lty = "dotted", col = "blue"),
  pars_text_final_x = list(pos = 3, cex = 1),
  pars_text_final_x_lower = pars_text_final_x,
  pars_text_final_x_upper = pars_text_final_x,
  pars_text_final_power = list(cex = 1),
  pars_sig_area = list(col = adjustcolor("lightblue", alpha.f = 0.1)),
  pars_text_sig_area = list(cex = 1),
  ...
)
}
\arguments{
\item{x}{An \code{x_from_power} object,
the output of \code{\link[=x_from_power]{x_from_power()}}.}

\item{what}{A character vector of
what to include in the
plot. Possible values are
\code{"ci"} (confidence intervals
for the estimated value of the
predictor),
\code{"power_curve"} (the crude power
curve, if available), \code{"final_x"}
(a vertical line for
the value of the predictor with estimated
power close enough to the target
power by confidence interval),
\code{"final_power"} (a horizontal line
for the estimated power of the
final value of the predictor),
\code{"target_power"} (a horizontal
line for the target power),
and \code{"sig_area"} (the area significantly
higher or lower than the target
power, if \code{goal} is \code{"close_enough"}
and \code{what} is \code{"lb"} or \code{"ub"}).
By default, all these elements will
be plotted.}

\item{text_what}{A character vector
of what numbers to be added as
labels. Possible values are
\code{"final_x"} (the value of the
predictor with
estimated power close enough to
the target power by confidence interval)
\code{"final_power"} (the estimated
power of the final value of the
predictor), and
\code{"sig_area"} (labeling the area significantly
higher or lower than the target
power, if \code{goal} is \code{"close_enough"}
and \code{what} is \code{"lb"} or \code{"ub"}).
By default, all these labels will
be added.}

\item{digits}{The number of digits
after the decimal that will be used
when adding numbers.}

\item{main}{The title of the plot.
If \code{NULL}, it will be generated
automatically.}

\item{xlab, ylab}{The labels for the
horizontal and vertical axes,
respectively.}

\item{ci_level}{The level of
confidence of the confidence intervals,
if requested. Default is \code{.95}, denoting
95\%.}

\item{pars_ci}{A named list of
arguments to be passed to \code{\link[=arrows]{arrows()}}
to customize the drawing of the
confidence intervals.}

\item{pars_power_curve}{A named list of
arguments to be passed to \code{\link[=points]{points()}}
to customize the drawing of the
power curve.}

\item{pars_ci_final_x}{A named list of
arguments to be passed to \code{\link[=arrows]{arrows()}}
to customize the drawing of the
confidence interval of the final
value of the predictor.}

\item{pars_target_power}{A named list
of arguments to be passed to \code{\link[=abline]{abline()}}
when drawing the horizontal line
for the target power.}

\item{pars_final_x}{A
named list of arguments to be passed
to \code{\link[=abline]{abline()}} when drawing the
vertical line for the final value
of the predictor.}

\item{pars_final_power}{A
named list of arguments to be passed
to \code{\link[=abline]{abline()}} when drawing the
horizontal line for the estimated
power at the final value of the
predictor.}

\item{pars_text_final_x}{A
named list of arguments to be passed
to \code{\link[=text]{text()}} when adding the
label for the final value of the
predictor.}

\item{pars_text_final_power}{A
named list of arguments to be passed
to \code{\link[=text]{text()}} when adding the
label for the estimated power
of final value of the predictor.}

\item{pars_sig_area}{A named list
of arguments to be passed to
\code{\link[=rect]{rect()}} when shading the area
significantly higher or lower than
the target power.}

\item{pars_text_sig_area}{A named list
of arguments to be passed to
\code{\link[=text]{text()}} when labelling the area
significantly higher or lower than
the target power.}

\item{prop_of_trials}{The proportion
of trials to be included in the plot.
If \code{NULL}, it will be determined
based on the algorithm used.}

\item{min_trials_for_prop}{The minimum
number of trials for \code{prop_of_trials}
to be used.
If \code{NULL}, it will be determined
based on the algorithm used.}

\item{override_for_pba}{If \code{TRUE},
the default, the values of
some arguments will
be overriden internally if the
algorithm used is \code{"probabilistic_bisection"},
to make the plot suitable for
this algorithm. For example, the
power curve and the confidence intervals
for trials other than the solution
will not be plotted, even if requested.
If \code{FALSE}, then argument values
will be not be changed internally.}

\item{...}{Optional arguments.
Passed to \code{\link[=plot]{plot()}} when drawing
the estimated power against the
predictor.}

\item{pars_text_final_x_lower, pars_text_final_x_upper}{If two values of the predictor are to
be printed, these are the named list
of the arguments to be passed to \code{\link[=text]{text()}}
when adding the labels for these two
values.}
}
\value{
The \code{plot}-method of \code{x_from_power}
returns \code{x} invisibly.
It is called for its side effect.

The \code{plot}-method of \code{n_region_from_power}
returns \code{x} invisibly.
It is called for its side effect.
}
\description{
It plots the results
of 'x_from_power', such as the
estimated power against sample size.
}
\details{
The \code{plot} method of \code{x_from_power}
objects currently plots the relation
between estimated power and
the values examined by \code{\link[=x_from_power]{x_from_power()}}.
Other elements
can be requested (see the argument
\code{what}), and they can be customized
individually.

The \code{plot}-method for
\code{n_region_from_power} objects is
a modified version of the \code{plot}-method
for \code{x_from_power}. It plots the
results of two runs of \code{\link[=n_from_power]{n_from_power()}}
in one plot. It is otherwise similar
to the \code{plot}-method for \code{x_from_power}.
}
\examples{

# Specify the population model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
m ~ x: m
y ~ m: l
y ~ x: n
"

# Generate the datasets

sim_only <- power4test(nrep = 10,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do a test

test_out <- power4test(object = sim_only,
                       test_fun = test_parameters,
                       test_args = list(pars = "m~x"))

# Determine the sample size with a power of .80 (default)

power_vs_n <- x_from_power(test_out,
                           x = "n",
                           progress = TRUE,
                           target_power = .80,
                           final_nrep = 10,
                           max_trials = 1,
                           seed = 2345)
plot(power_vs_n)

}
\seealso{
\code{\link[=x_from_power]{x_from_power()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_group_constraints.R
\name{test_group_equal}
\alias{test_group_equal}
\title{Test Group Constraints}
\usage{
test_group_equal(
  fit = fit,
  group.equal = NULL,
  group.partial = NULL,
  check_post_check = TRUE,
  ...,
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object. Must be
the output of \code{\link[lavaan:lavaan]{lavaan::lavaan()}} or
its wrappers, such as \code{\link[lavaan:sem]{lavaan::sem()}}
and \code{\link[lavaan:cfa]{lavaan::cfa()}}. The model must
be a multigroup model.}

\item{group.equal}{The same argument
used by \code{lavaan}. A character vector
with one or more of these values:
\code{"regressions"}, \code{"loadings"},
\code{"lv.covariances"}, \code{"lv.variances"},
\code{"intercepts"}, \code{"means"}, \code{"thresholds"},
\code{"residual.covariances"}, \code{"composite.weights"},
and \code{"residuals"}.}

\item{group.partial}{The same argument
used by \code{lavaan}. The parameters that
should be free across groups. Used with
\code{group.equal} to exclude some parameters
from those requested to be equal
across groups by \code{group.equal}.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{...}{Optional arguments to be
passed to \code{\link[lavaan:lavTestLRT]{lavaan::lavTestLRT()}}.}

\item{fit_name}{The name of the
model fit object to be extracted.
Default is \code{"fit"}. Used only when
more than one model is fitted in
each replication. This should be
the name of the model on which the
test is to be conducted.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
a one-row data frame with the
following columns:
\itemize{
\item \code{est}: The chi-square difference.
\item \code{cilo} and \code{cihi}: \code{NA}. Not used.
\item \code{sig}: Whether the chi-square
difference test is significant
\item \code{test_label}: The constraints
imposted.
}
}
\description{
Test the model fit change
when one or more between-group
constraints are imposed.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing
the difference in model fit
when one or more between-group
constraints are imposed
, by
setting it to the \code{test_fun}
argument.
}
\examples{

# Specify the model

mod <-
"
m ~ x
y ~ m + x
"

# Specify the population values

mod_es <-
"
y ~ m: l
m ~ x:
  - nil
  - s
y ~ x: nil
"

# Simulate the data

sim_only <- power4test(nrep = 2,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       iseed = 1234)

# Do the tests in each replication

test_out <- power4test(object = sim_only,
                       test_fun = test_group_equal,
                       test_args = list(group.equal = "regressions"))

print(test_out,
      test_long = TRUE)


}
\seealso{
\code{\link[=power4test]{power4test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_moderation.R
\name{test_moderation}
\alias{test_moderation}
\title{Test All Moderation Effects}
\usage{
test_moderation(
  fit = fit,
  standardized = FALSE,
  check_post_check = TRUE,
  ...,
  p_adjust_method = "none",
  fit_name = "fit",
  get_map_names = FALSE,
  get_test_name = FALSE
)
}
\arguments{
\item{fit}{The fit object, to be
passed to \code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}},
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}}.}

\item{standardized}{Logical. If \code{TRUE},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}} will
be used. Can be used only with models
fitted by \code{lavaan}.}

\item{check_post_check}{Logical. If
\code{TRUE}, the default, and the model
is fitted by \code{lavaan}, the test
will be conducted only if the model
passes the \code{post.check} conducted
by \code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} (with
\code{what = "post.check"}).}

\item{...}{Additional arguments to
be passed to \code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}},
\code{\link[lavaan:standardizedSolution]{lavaan::standardizedSolution()}},
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}}.}

\item{p_adjust_method}{The method to be
passed to \code{\link[=p.adjust]{p.adjust()}} to adjust the
\emph{p}-values when testing the effects.
Default is \code{"none"} and the \emph{p}-values
will not be adjusted. The unadjusted
\emph{p}-values will be stored in the column
\code{pvalue_org}.}

\item{fit_name}{The name of the fit
results for which the parameter names
will be displayed. Default is \code{"fit"}.}

\item{get_map_names}{Logical. Used
by \code{\link[=power4test]{power4test()}} to determine how
to extract stored information and
assign them to this function. Users
should not use this argument.}

\item{get_test_name}{Logical. Used
by \code{\link[=power4test]{power4test()}} to get the default
name of this test. Users should not
use this argument.}
}
\value{
In its normal usage, it returns
the output returned by
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}
or \code{\link[lmhelprs:lm_list_to_partable]{lmhelprs::lm_list_to_partable()}},
with the following modifications:
\itemize{
\item \code{est}: The parameter estimates,
even if standardized estimates
are requested (not \code{est.std}).
\item \code{cilo} and \code{cihi}: The
lower and upper limits of the
confidence interval (95\% by
default), respectively (not
\code{ci.lower} and \code{ci.upper}).
\item \code{sig}: Whether a test by confidence
interval is significant (\code{1}) or
not significant (\code{0}).
\item \code{test_label}: A column of labels
generated by
\code{\link[lavaan:lav_partable]{lavaan::lav_partable_labels()}},
which are usually the labels used by
\code{coef()} to label the parameters.
}
}
\description{
Test all moderation
effects by testing all product
terms for a \code{power4test} object.
}
\details{
This function is to be used in
\code{\link[=power4test]{power4test()}} for testing all
product terms, by
setting it to the \code{test_fun}
argument.

It is just a wrapper to
\code{\link[=test_parameters]{test_parameters()}}. It will
first identifies all product
terms (terms with \code{:} in the names),
and then call \code{\link[=test_parameters]{test_parameters()}},
with \code{pars} set to select the
regression coefficients of these
terms.
}
\examples{

# Specify the model

mod <-
"
m ~ x + w1 + x:w1
y ~ m + x
"

# Specify the population values

mod_es <-
"
m ~ x: n
y ~ x: m
m ~ w1: n
m ~ x:w1: l
"

# Simulate the data

sim_only <- power4test(nrep = 4,
                       model = mod,
                       pop_es = mod_es,
                       n = 100,
                       do_the_test = FALSE,
                       iseed = 1234)

# Do the test in each replication

test_out <- power4test(object = sim_only,
                       test_fun = test_moderation)

print(test_out,
      test_long = TRUE)

}
\seealso{
\code{\link[=power4test]{power4test()}},
\code{\link[=test_parameters]{test_parameters()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbinary_rs.R
\name{rbinary_rs}
\alias{rbinary_rs}
\title{Random Binary Variable}
\usage{
rbinary_rs(n = 10, p1 = 0.5, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{p1}{The probability of being
1, before rescaling.}

\item{pmean}{Population mean.}

\item{psd}{Population standard
deviation.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from a distribution of 0 or 1,
rescaled to have user-specified
population mean and standard
deviation.
}
\details{
First, specify probability of 1
(\code{p1}), and the desired population
mean and standard deviation. The
random numbers, drawn from a
distribution of 0 (\code{1 - p1}
probability) and 1 (\code{p1}
probability), will then be rescaled
with the desired population mean and
standard.
}
\examples{
set.seed(90870962)
x <- rbinary_rs(n = 5000,
                p1 = .5,
                pmean = 3,
                psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rlnorm_rs.R
\name{rlnorm_rs}
\alias{rlnorm_rs}
\title{Random Variable From a Lognormal Distribution}
\usage{
rlnorm_rs(n = 10, mui = 0, sigma = 1, rev = FALSE, pmean = 0, psd = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{mui}{The parameter \code{mui} to
be used by \code{\link[stats:Lognormal]{stats::rlnorm()}}.}

\item{sigma}{The parameter \code{sigma} to
be used by \code{\link[stats:Lognormal]{stats::rlnorm()}}.}

\item{rev}{If TRUE, the distribution
is revered to generate a negatively
skewed distribution. Default is
FALSE.}

\item{pmean}{Population mean.}

\item{psd}{Population standard deviation.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from a lognormal distribution,
rescaled to have user-specified
population mean and standard
deviation.
}
\details{
First, specify the parameter,
\code{mui} and \code{sigma}, and the
desired population mean and standard
deviation. The random numbers, drawn
from a lognormal distribution by
\code{\link[stats:Lognormal]{stats::rlnorm()}}, will then be
rescaled with the desired population
mean and standard.
}
\examples{
set.seed(90870962)
x <- rlnorm_rs(n = 5000,
               mui = 0,
               sigma = 1,
               pmean = 0,
               psd = 1)
mean(x)
sd(x)
hist(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/q_functions.R
\name{q_power_mediation}
\alias{q_power_mediation}
\alias{print.q_power_mediation}
\alias{plot.q_power_mediation}
\alias{summary.q_power_mediation}
\alias{q_power_mediation_simple}
\alias{q_power_mediation_serial}
\alias{q_power_mediation_parallel}
\title{All-in-One Power Estimation For Mediation Models}
\usage{
q_power_mediation(
  model = NULL,
  pop_es = NULL,
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_fun = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
)

\method{print}{q_power_mediation}(x, mode = c("all", "region", "n", "power"), ...)

\method{plot}{q_power_mediation}(x, ...)

\method{summary}{q_power_mediation}(object, ...)

q_power_mediation_simple(
  a = "m",
  b = "m",
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
)

q_power_mediation_serial(
  ab = c("m", "m"),
  ab_other = "n",
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
)

q_power_mediation_parallel(
  as = c("m", "m"),
  bs = c("m", "m"),
  cp = "n",
  number_of_indicators = NULL,
  reliability = NULL,
  loading_difference = NULL,
  reference = NULL,
  omnibus = c("all_sig", "at_least_one_sig", "at_least_k_sig"),
  at_least_k = 1,
  test_more_args = list(),
  target_power = 0.8,
  nrep = NULL,
  n = NULL,
  R = 1000,
  ci_type = c("mc", "boot"),
  seed = NULL,
  iseed = NULL,
  parallel = TRUE,
  progress = TRUE,
  simulation_progress = NULL,
  max_trials = NULL,
  algorithm = NULL,
  ...,
  mode = c("power", "region", "n")
)
}
\arguments{
\item{model}{The \code{lavaan} model
syntax of the population model,
to be used by \code{\link[=ptable_pop]{ptable_pop()}}.
See 'Details' of
on how to specify the model.
Ignored if \code{ptable} is
specified.}

\item{pop_es}{The character vector or
multiline string to
specify population effect sizes
(population values of parameters). See
the help page on how to specify this
argument.
Ignored if \code{ptable} is
specified.}

\item{number_of_indicators}{A named
vector to specify the number of
indicators for each factors. See
the help page on how to set this
argument. Default is \code{NULL} and all
variables in the model syntax are
observed variables.
See the help page on how
to use this argument.}

\item{reliability}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the reliability coefficient
of each set of indicators. Default
is \code{NULL}.
See the help page on how
to use this argument.}

\item{loading_difference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to set the difference in factor
loadings between neighboring indicators
of each set of indicators. Default
is \code{NULL}, and all indicators of
a factor have
the same factor loadings. If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal.}

\item{reference}{A named vector
(for a single-group model) or a
named list of named vectors
(for a multigroup model)
to indicate which indicator will
be the first indicator (and so
is the reference indicator, by default).
Default
is \code{NULL}, and for all factors,
the indicator with
the medium loading in a factor is
the first indicator. Has no effect
if loading difference is zero (and
so all indicators have the same
loadings). If specified,
must be specified for all factors
named in \code{reliability}, even for
those with all loadings equal. Accepted
values are \code{"medium"}, \code{"weakest"},
and \code{"strongest"}.}

\item{test_fun}{A function to do the
test. See 'Details' of \code{\link[=power4test]{power4test()}}
for the requirement of this function.}

\item{test_more_args}{A named list of
additional arguments to be passed
to the test function
(\code{\link[=test_indirect_effect]{test_indirect_effect()}} for
simple and serial mediation models,
and \code{\link[=test_k_indirect_effects]{test_k_indirect_effects()}}
for parallel mediation models).
Similar to \code{test_args} in
\code{\link[=power4test]{power4test()}}.}

\item{target_power}{The target power,
a value greater than 0 and less than
one.}

\item{nrep}{The number of replications
to generate the simulated datasets.
Default is \code{NULL} and will be determined
internally.}

\item{n}{The sample size for the
first run of \code{\link[=power4test]{power4test()}}. Must be
set to a sample size if \code{mode} is
\code{"power"}. For the other modes,
if \code{NULL}, the default, it will be
determined internally.}

\item{R}{The number of replications
to generate the Monte Carlo or
bootstrapping estimates
for each fit output. No Monte Carlo
nor bootstrapping
estimates will be generated if \code{R}
is set to \code{NULL}.}

\item{ci_type}{The type of
simulation-based confidence
intervals to use. Can be either
\code{"mc"} for Monte Carlo method
(the default) or \code{"boot"} for
nonparametric bootstrapping method.
Relevant for test functions that
make use of estimates generate by
\code{\link[=gen_boot]{gen_boot()}} or \code{\link[=gen_mc]{gen_mc()}}, such as
\code{\link[=test_indirect_effect]{test_indirect_effect()}}.}

\item{seed}{The seed for the random
number generator. Used by
\code{\link[=n_region_from_power]{n_region_from_power()}}.}

\item{iseed}{The seed for the random
number generator. Used by
\code{\link[=power4test]{power4test()}}.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used when calling
other functions, if appropriate.}

\item{progress}{If \code{TRUE}, the progress
of each step will be displayed.
Default is \code{TRUE}.}

\item{simulation_progress}{Logical.
Whether the progress in each call
to \code{\link[=power4test]{power4test()}}, \code{\link[=power4test_by_n]{power4test_by_n()}},
or \code{\link[=power4test_by_es]{power4test_by_es()}}
is shown. To be passed to
the \code{progress} argument of these
functions.
If \code{NULL}, set automatically based
on the algorithm used.}

\item{max_trials}{The maximum number
of trials in searching the value
with the target power. Rounded
up if not an integer.
If \code{NULL}, set automatically based
on the algorithm used.}

\item{algorithm}{The algorithm to be
used in mode \code{"region"} and \code{"n"}.
If \code{NULL}, then it will be determined
internally based on the \code{mode}. (The
default may be different from that
of \code{\link[=n_region_from_power]{n_region_from_power()}} and
\code{\link[=n_from_power]{n_from_power()}})}

\item{...}{For \verb{q_power_mediation_*},
these are optional arguments to
be passed to \code{\link[=power4test]{power4test()}} and
\code{\link[=n_region_from_power]{n_region_from_power()}}.
For the \code{print} method, these
are optional arguments to
be passed to \code{\link[=rejection_rates]{rejection_rates()}}
as well as other print methods
(see \code{\link[=print.power4test]{print.power4test()}} and
\code{\link[=print.n_region_from_power]{print.n_region_from_power()}}).
For the \code{plot} method, these
are optional arguments to
be passed to \code{\link[=plot.n_region_from_power]{plot.n_region_from_power()}}.
For the \code{summary} method, these
are optional arguments to
be passed to \code{\link[=summary.n_region_from_power]{summary.n_region_from_power()}}.}

\item{mode}{For \verb{q_power_mediation_*, this is the mode of what to do. If }"power"\verb{, then only [power4test()] will be called, and the level of power will be estimated. If }"region"\verb{, then the region of sample sizes with levels of power not significantly different from the target power will be searched by calling [n_region_from_power()]. For the }print\verb{method, this indicates what to print. If}"region"\verb{and the output of [n_region_from_power()] is available, it will print the results of [n_region_from_power()] if available. If}"n"\verb{, then the output of [n_from_power()] will be printed if available. If }"power"\verb{, then the output of [power4test()] will be printed. If }"all"`, then all available outputs
will be printed.}

\item{x}{The object
for the relevant methods.}

\item{object}{For the \code{summary}
method of \code{\link[=q_power_mediation]{q_power_mediation()}}
outputs.}

\item{a}{For a simple mediation
model, this is the population effect
size for the path from \code{x} to \code{m} .}

\item{b}{For a simple mediation
model, this is the population effect
size for the path from \code{m} to \code{y} .}

\item{cp}{For a simple mediation
model, this is the population effect
size for the direct path from
\code{c} to \code{y} .}

\item{ab}{For a serial mediation
model, this is a numeric vector
of the population effect
sizes along the path
\code{x->m1->m2->...->y}.}

\item{ab_other}{Should be one single
value. This is the population effect
sizes of all other paths not along
\code{x->m1->m2->...->y}, except for the
direct path from \code{x} to \code{y}.}

\item{as}{For a parallel mediation
model, this is a numeric vector
of the population effect
sizes for the paths from \code{x} to
the mediators: \code{x->m1}, \code{x->m2},
... \code{x->mp}, for a parallel mediation
model with \code{p} mediators.}

\item{bs}{For a parallel mediation
model, this is a numeric vector
of the population effect
sizes for the paths from the
mediators to \code{y}: \code{m1->y}, \code{m2->y},
... \code{mp->y}, for a parallel mediation
model with \code{p} mediators.}

\item{omnibus}{\code{"all_sig"}, the
default, then
the test is declared significant if
\emph{all} paths are significant. If
\code{"at_least_one_sig"}, then only
one row of test is stored, and the
test is declared significant if
at least one of the paths is
significant. If \code{"at_least_k_sig"},
then only one row of test is stored,
and the test is declared significant
if at least \code{k} of the paths is
significant, \code{k} determined by the
argument \code{at_least_k}.}

\item{at_least_k}{The minimum number
of paths required to be significant
for the omnibus test to be considered
significant. Used when
\code{omnibus} is \code{"at_least_k_sig"}.}
}
\value{
A named list of outputs.
If \code{mode} is \code{power}, then a
\code{power4test} object is set to the
element \code{power4test}.
If \code{mode} is \code{region}, then a
\code{n_region_from_power} object is
set of the element \code{n_region_from_power}.
If \code{mode} is \code{n_from_power}, then a
\code{n_from_power} object is
set to the \code{n_from_power} element.

The \code{print} method of \code{q_power_mediation}
returns \code{x} invisibly. Called for
its side effect.

The \code{plot}-method of \code{q_power_mediation}
returns \code{NULL}.
It will plot either the output of
\code{\link[=n_region_from_power]{n_region_from_power()}}
(mode \code{"region"}) or the output of
\code{\link[=n_from_power]{n_from_power()}} (mode \code{"n"}).
If no output from
either of these functions is available,
nothing wil be plotted.

The \code{summary} method for
\code{q_power_mediation} objects returns
the output of \code{\link[=summary]{summary()}} for
either the output of \code{\link[=n_region_from_power]{n_region_from_power()}}
(mode \code{"region"}) or the output of
\code{\link[=n_from_power]{n_from_power()}} (mode \code{"n"}).
Return \code{NULL} if no output from
either of these functions is available.
}
\description{
All-in-one functions for
estimating power or finding the
region with target power for common
mediation models.
}
\section{All-in-One Functions for Common Mediation Models}{
These functions are wrappers
that call \code{\link[=power4test]{power4test()}} and
\code{\link[=n_region_from_power]{n_region_from_power()}} to (a)
estimate the level of power for
a mediation model, given the population
effects and the sample size, and
(b) find the region of sample sizes
with the levels of power not
significantly different from the
target power.

They are convenient functions that
set the argument values automatically
for common mediation models before
calling \code{\link[=power4test]{power4test()}} and
\code{\link[=n_region_from_power]{n_region_from_power()}}. Please refer
to the help pages of these two
functions for the details on how
the estimation and the search are
conducted.

For some arguments not described in
details here, please refer to the
help pages of \code{\link[=power4test]{power4test()}}
and \code{\link[=n_region_from_power]{n_region_from_power()}},
}

\section{Simple Mediation Model}{
The function \code{\link[=q_power_mediation_simple]{q_power_mediation_simple()}}
can be used for the power analysis of
a simple mediation model with only
one mediator.

This function will fit
the following model:

\preformatted{"m ~ x
 y ~ m + x"}
}

\section{Serial Mediation Model}{
The function \code{\link[=q_power_mediation_serial]{q_power_mediation_serial()}}
can be used for the power analysis of
a serial mediation model with only
any number of mediators.

This is the model being fitted if
the model has two mediators:

\preformatted{"m1 ~ x
 m2 ~ m1 + x
 y ~ m2 + m1 + x"}
}

\section{Parallel Mediation Model}{
The function \code{\link[=q_power_mediation_parallel]{q_power_mediation_parallel()}}
can be used for the power analysis of
a parallel mediation model with only
any number of mediators.

This is the model being fitted if
the model has two mediators:

\preformatted{"m1 ~ x
 m2 ~ x
 y ~ m2 + m1 + x"}
}

\section{An Arbitrary Mediation Model}{
The function \code{\link[=q_power_mediation]{q_power_mediation()}},
an advanced function,
can be used for the power analysis of
an arbitrary mediation model.
The model and the population effect
sizes are specified as in \code{\link[=power4test]{power4test()}}.

This is an example of a model with
both parallel paths and serial paths:

\preformatted{model <-
 "
 m1 ~ x
 m21 ~ m1
 m22 ~ m1
 y ~ m21 + m22 + x
 "}

\preformatted{pop_es <-
"
m1 ~ x: m
m21 ~ m1: m
m22 ~ m1: m
y ~ m21: m
y ~ m22: m
"}

Knowledge of using \code{\link[=power4test]{power4test()}}
is required to use this advanced
function.

If this advanced function is used,
users need to specify \code{test_fun}
as when using \code{\link[=power4test]{power4test()}}, and
need to set \code{test_args} correctly
}

\examples{

\dontrun{

# An arbitrary mediation model

model <-
"
m1 ~ x
m21 ~ m1
m22 ~ m1
y ~ m21 + m22
"
pop_es <-
"
m1 ~ x: m
m21 ~ m1: m
m22 ~ m1: m
y ~ m21: m
y ~ m22: m
"

# NOTE: In real power analysis:
# - Set R to an appropriate value.
# - Remove nrep or set nrep to the desired value.
# - Remove parallel or set it to TRUE to enable parallel processing.
# - Remove progress or set it to TRUE to see the progress.

outa1 <- q_power_mediation(
    model = model,
    pop_es = pop_es,
    n = 100,
    R = 199,
    test_fun = test_k_indirect_effects,
    test_more_args = list(x = "x",
                          y = "y",
                          omnibus = "all"),
    seed = 1234,
    mode = "region",
    nrep = 20,
    parallel = FALSE,
    progress = FALSE
  )
outa1
summary(outa1)
plot(outa1)

}


# Simple mediation model

# NOTE: In real power analysis:
# - Set R to an appropriate value.
# - Remove nrep or set nrep to the desired value.
# - Remove parallel or set it to TRUE to enable parallel processing.
# - Remove progress or set it to TRUE to see the progress.

out <- q_power_mediation_simple(
    a = "m",
    b = "m",
    cp = "n",
    n = 50,
    R = 79,
    seed = 1234,
    nrep = 10,
    parallel = FALSE,
    progress = FALSE
  )
out

# If mode = "region" is added, can call the following
# summary(out)
# plot(out)


\dontrun{
# Serial mediation model

# NOTE: In real power analysis:
# - Set R to an appropriate value.
# - Remove nrep or set nrep to the desired value.
# - Remove parallel or set it to TRUE to enable parallel processing.
# - Remove progress or set it to TRUE to see the progress.

outs <- q_power_mediation_serial(
    ab = c("s", "m", "l"),
    ab_others = "n",
    cp = "s",
    n = 50,
    R = 199,
    seed = 1234,
    mode = "region",
    nrep = 20,
    parallel = FALSE,
    progress = FALSE
  )
outs
summary(outs)
plot(outs)
}


\dontrun{
# Parallel mediation model

# NOTE: In real power analysis:
# - Set R to an appropriate value.
# - Remove nrep or set nrep to the desired value.
# - Remove parallel or set it to TRUE to enable parallel processing.
# - Remove progress or set it to TRUE to see the progress.

outp <- q_power_mediation_parallel(
    as = c("s", "m"),
    bs = c("m", "s"),
    cp = "n",
    n = 100,
    R = 199,
    seed = 1234,
    mode = "region",
    nrep = 20,
    parallel = FALSE,
    progress = FALSE
  )
outp
summary(outp)
plot(outp)
}

}
\seealso{
See \code{\link[=power4test]{power4test()}} and
\code{\link[=n_region_from_power]{n_region_from_power()}} for full
details on how these functions
work.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbeta_rs2.R
\name{rbeta_rs2}
\alias{rbeta_rs2}
\title{Random Variable From a Beta Distribution (User Range)}
\usage{
rbeta_rs2(n = 10, bmean, bsd, blow = 0, bhigh = 1)
}
\arguments{
\item{n}{The number of random numbers
to generate.}

\item{bmean}{The population mean.}

\item{bsd}{The population standard
deviation. If \code{bsd} is zero or
negative, all random numbers will be
equal to \code{bmean}.}

\item{blow}{The lower bound of the
target range.}

\item{bhigh}{The upper bound of the
target range.}
}
\value{
A vector of the generated
random numbers.
}
\description{
Generate random numbers
from a beta distribution, rescaled to
have user-specified population mean
and standard deviation, and within a
specific range.
}
\details{
First, specify the two parameters,
\code{shape1} and \code{shape2}, and the
desired population mean and standard
deviation. The random numbers, drawn
from a beta distribution by
\code{\link[stats:Beta]{stats::rbeta()}} will then be
rescaled to the desired population range.
}
\examples{
set.seed(90870962)
x <- rbeta_rs2(n = 5000,
               bmean = .80,
               bsd = .10,
               blow = .00,
               bhigh = .95)
mean(x)
sd(x)
hist(x)
y <- rbeta_rs2(n = 5000,
               bmean = 4,
               bsd = 3,
               blow = -10,
               bhigh = 10)
mean(y)
sd(y)
hist(y)

}
